Package: metaseqR2
Type: Package
Title: An R package for the analysis and result reporting of RNA-Seq
        data by combining multiple statistical algorithms
Authors@R: c(person(given="Panagiotis", family="Moulos",
    email="moulos@fleming.gr", role=c("aut", "cre")))
Depends: R (>= 4.0.0), DESeq2, limma, locfit, splines
Imports: ABSSeq, Biobase, BiocGenerics, BiocParallel, biomaRt,
        Biostrings, corrplot, DSS, DT, EDASeq, edgeR, genefilter,
        Seqinfo, GenomeInfoDb, GenomicAlignments, GenomicFeatures,
        GenomicRanges, gplots, graphics, grDevices, harmonicmeanp,
        heatmaply, htmltools, httr, IRanges, jsonlite, lattice, log4r,
        magrittr, MASS, Matrix, methods, NBPSeq, pander, parallel,
        qvalue, rmarkdown, rmdformats, Rsamtools, RSQLite, rtracklayer,
        S4Vectors, stats, stringr, SummarizedExperiment, survcomp,
        txdbmaker, utils, VennDiagram, vsn, yaml, zoo
Suggests: BiocStyle, BiocManager, BSgenome, knitr, RMySQL, RUnit
Enhances: TCC
Description: Provides an interface to several normalization and
    statistical testing packages for RNA-Seq gene expression data.
    Additionally, it creates several diagnostic plots, performs
    meta-analysis by combinining the results of several statistical
    tests and reports the results in an interactive way.
License: GPL (>= 3)
Encoding: UTF-8
LazyData: false
NeedsCompilation: yes
URL: http://www.fleming.gr
biocViews: Software, GeneExpression, DifferentialExpression,
        WorkflowStep, Preprocessing, QualityControl, Normalization,
        ReportWriting, RNASeq, Transcription, Sequencing,
        Transcriptomics, Bayesian, Clustering, CellBiology,
        BiomedicalInformatics, FunctionalGenomics, SystemsBiology,
        ImmunoOncology, AlternativeSplicing, DifferentialSplicing,
        MultipleComparison, TimeCourse, DataImport, ATACSeq,
        Epigenetics, Regression, ProprietaryPlatforms,
        GeneSetEnrichment, BatchEffect, ChIPSeq
VignetteBuilder: knitr
BugReports: https://github.com/pmoulos/metaseqR2/issues
Version: 1.25.0
Date: 2026-03-21
Collate: 'annotation.R' 'argcheck.R' 'count.R' 'deseq.R' 'export.R'
        'external.R' 'filter.R' 'json.R' 'main.R' 'meta.R' 'noiseq.R'
        'norm.R' 'metaseqr2-data.R' 'metaseqR2-package.R' 'plot.R'
        'query.R' 'reportdb.R' 'sim.R' 'stat.R' 'tracks.R' 'util.R'
        'zzz.R'
git_url: https://git.bioconductor.org/packages/metaseqR2
git_branch: devel
git_last_commit: 0bd17c4
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-29
Packaged: 2026-04-29 22:52:02 UTC; biocbuild
Author: Panagiotis Moulos [aut, cre]
Maintainer: Panagiotis Moulos <moulos@fleming.gr>

import(Biobase)
import(biomaRt)
import(DT)
import(edgeR)
import(GenomicRanges)
import(graphics)
import(lattice)
import(limma)
import(methods)
#import(magrittr)
import(NBPSeq)
import(parallel)
import(Rsamtools)
import(RSQLite)
import(splines)
import(VennDiagram)

import(grDevices,except=windows)
import(locfit,except=dat)

importFrom("ABSSeq",ABSDataSet)
importFrom("ABSSeq",ABSSeq)
importFrom("ABSSeq",ABSSeqlm)
importFrom("ABSSeq",aFoldcomplexDesign)
importFrom("ABSSeq",callDEs)
importFrom("ABSSeq",callParameter)
importFrom("ABSSeq",callParameterwithoutReplicates)
importFrom("ABSSeq",estimateSizeFactorsForMatrix)
importFrom("ABSSeq",excounts)
importFrom("ABSSeq","excounts<-")
importFrom("ABSSeq",maxRates)
importFrom("ABSSeq",minimalDispersion)
importFrom("ABSSeq",minRates)
importFrom("ABSSeq",normalFactors)
importFrom("ABSSeq",normMethod)

#importFrom("baySeq",getLibsizes)
#importFrom("baySeq",getPriors.NB)
#importFrom("baySeq",getLikelihoods.NB)
#importFrom("baySeq",groups)
#importFrom("baySeq",libsizes)
#importFrom("baySeq",replicates)
#importFrom("baySeq",topCounts)

importFrom("BiocGenerics",counts)
importFrom("BiocGenerics","counts<-")
importFrom("BiocGenerics",conditions)
importFrom("BiocGenerics","conditions<-")
importFrom("BiocGenerics",dispTable)
importFrom("BiocGenerics","dispTable<-")
importFrom("BiocGenerics",sizeFactors)
importFrom("BiocGenerics","sizeFactors<-")

importFrom("BiocParallel",bpparam)

importFrom("Biostrings",alphabetFrequency)
importFrom("Biostrings",uniqueLetters)
importFrom("Biostrings",readDNAStringSet)

importFrom("corrplot",corrplot)

importFrom("DESeq2",DESeq)
importFrom("DESeq2",DESeqDataSetFromMatrix)
importFrom("DESeq2",estimateBetaPriorVar)
importFrom("DESeq2",estimateDispersions)
importFrom("DESeq2",estimateSizeFactors)
importFrom("DESeq2",nbinomLRT)
importFrom("DESeq2",nbinomWaldTest)
importFrom("DESeq2",normalizationFactors)
importFrom("DESeq2",normalizeGeneLength)
importFrom("DESeq2",normTransform)
importFrom("DESeq2",sizeFactors)

importFrom("DSS",callDML)
importFrom("DSS",callDMR)
importFrom("DSS",dispersion)
importFrom("DSS",DMLfit.multiFactor)
importFrom("DSS",DMLtest)
importFrom("DSS",DMLtest.multiFactor)
importFrom("DSS",DSS.DE)
importFrom("DSS",estDispersion)
importFrom("DSS",estNormFactors)
importFrom("DSS",newSeqCountSet)
importFrom("DSS",normalizationFactor)
importFrom("DSS",waldTest)

importFrom("EDASeq",betweenLaneNormalization)
importFrom("EDASeq",biasPlot)
importFrom("EDASeq",MDPlot)
importFrom("EDASeq",meanVarPlot)
importFrom("EDASeq",newSeqExpressionSet)
importFrom("EDASeq",normCounts)
importFrom("EDASeq",withinLaneNormalization)

importFrom("genefilter",rowVars)

importFrom("Seqinfo",seqinfo)
importFrom("Seqinfo",Seqinfo)
importFrom("Seqinfo",seqlevels)
importFrom("Seqinfo",seqlengths)
importFrom("Seqinfo","seqlengths<-")

importFrom("GenomeInfoDb",dropSeqlevels)
#importFrom("GenomeInfoDb",fetchExtendedChromInfoFromUCSC)
importFrom("GenomeInfoDb",getChromInfoFromUCSC)
importFrom("GenomeInfoDb",seqlevelsStyle)
importFrom("GenomeInfoDb","seqlevelsStyle<-")

importFrom("GenomicAlignments",readGAlignments)
importFrom("GenomicAlignments",summarizeOverlaps)

importFrom("txdbmaker",makeTxDbFromGFF)
#importFrom("GenomicFeatures",makeTxDbFromGFF)
importFrom("GenomicFeatures",exons)
importFrom("GenomicFeatures",genes)
importFrom("GenomicFeatures","seqlevels<-")
importFrom("GenomicFeatures",transcripts)
importFrom("GenomicFeatures",threeUTRsByTranscript)

importFrom("gplots",bluered)
importFrom("gplots",col2hex)
importFrom("gplots",colorpanel)
importFrom("gplots",heatmap.2)

importFrom("harmonicmeanp",hmp.stat)
importFrom("harmonicmeanp",p.hmp)

importFrom("heatmaply",heatmaply)

importFrom("htmltools",tags)

importFrom("httr",config)
importFrom("httr",set_config)

importFrom("IRanges",IRanges)
importFrom("IRanges",slice)

importFrom("jsonlite",fromJSON)
importFrom("jsonlite",toJSON)
importFrom("jsonlite",unbox)

importFrom("log4r",create.logger)
importFrom("log4r",debug)
importFrom("log4r",error)
importFrom("log4r",fatal)
importFrom("log4r",info)
importFrom("log4r",level)
importFrom("log4r",logfile)
importFrom("log4r",warn)

importFrom(magrittr,"%>%")

importFrom("MASS",negative.binomial)

importFrom("Matrix",rankMatrix)

importFrom("pander",pandoc.header)

importFrom("qvalue",qvalue)

importFrom("rmarkdown",render)

importFrom("rmdformats",material)

importFrom("rtracklayer",export.2bit)
importFrom("rtracklayer",export.bw)
importFrom("rtracklayer",import)
importFrom("rtracklayer",import.bed)

importFrom("S4Vectors",DataFrame)
importFrom("S4Vectors",Rle)
importFrom("S4Vectors",queryHits)
importFrom("S4Vectors",subjectHits)

importFrom("stats",aggregate)
importFrom("stats",approxfun)
importFrom("stats",as.dist)
importFrom("stats",as.formula)
importFrom("stats",cmdscale)
importFrom("stats",coefficients)
importFrom("stats",cor)
importFrom("stats",density)
importFrom("stats",deviance)
importFrom("stats",df)
importFrom("stats",df.residual)
importFrom("stats",dnbinom)
importFrom("stats",Gamma)
importFrom("stats",glm)
importFrom("stats",glm.fit)
importFrom("stats",fitted.values)
importFrom("stats",kmeans)
importFrom("stats",ks.test)
importFrom("stats",lowess)
importFrom("stats",mad)
importFrom("stats",median)
importFrom("stats",model.matrix)
importFrom("stats",na.exclude)
importFrom("stats",na.omit)
importFrom("stats",optimize)
importFrom("stats",p.adjust)
importFrom("stats",p.adjust.methods)
importFrom("stats",pchisq)
importFrom("stats",prop.test)
importFrom("stats",quantile)
importFrom("stats",rexp)
importFrom("stats",rnbinom)
importFrom("stats",rnorm)
importFrom("stats",rmultinom)
importFrom("stats",runif)
importFrom("stats",sd)
importFrom("stats",splinefun)
importFrom("stats",supsmu)
importFrom("stats",var)
importFrom("stats",wilcox.test)

importFrom("stringr",str_match)

importFrom("SummarizedExperiment",assays)

importFrom("survcomp",combine.test)

importFrom("utils",capture.output)
importFrom("utils",combn)
importFrom("utils",download.file)
importFrom("utils",head)
importFrom("utils",packageVersion)
importFrom("utils",read.delim)
importFrom("utils",read.table)
importFrom("utils",write.table)
importFrom("utils",zip)

importFrom("vsn",vsn2)
importFrom("vsn",predict)

importFrom("yaml",read_yaml)
importFrom("yaml",write_yaml)

importFrom("zoo",rollmean)

useDynLib("metaseqR2",calc_pvals)

export(buildAnnotationDatabase)
export(buildCustomAnnotation)
export(createSignalTracks)
export(combineBonferroni)
export(combineHarmonic)
export(combineMaxp)
export(combineMinp)
export(combineSimes)
export(combineWeight)
export(diagplotAvgFtd)
export(diagplotBoxplot)
export(diagplotCor)
export(diagplotDeHeatmap)
export(diagplotEdaseq)
export(diagplotFiltered)
export(diagplotFtd)
export(diagplotMds)
export(diagplotNoiseq)
export(diagplotPairs)
export(diagplotRoc)
export(diagplotVenn)
export(diagplotVolcano)
export(downsampleCounts)
export(estimateAufcWeights)
export(estimateSimParams)
export(getAnnotation)
export(getDefaults)
export(getInstalledAnnotations)
export(getWeights)
export(importCustomAnnotation)
export(loadAnnotation)
export(makeSimDataTcc)
export(makeSimDataSd)
export(metaseqr2)
export(metaseqrPlot)
export(metaTest)
export(normalizeAbsseq)
export(normalizeDeseq)
export(normalizeDeseq2)
export(normalizeDss)
export(normalizeEdaseq)
export(normalizeEdger)
export(normalizeNbpseq)
export(normalizeNoiseq)
export(read2count)
export(readTargets)
export(statAbsseq)
#export(statBayseq)
export(statDeseq)
export(statDeseq2)
export(statDss)
export(statEdger)
export(statLimma)
export(statNbpseq)
export(statNoiseq)

<!-- badges: start -->
  ![Bioconductor build](http://www.bioconductor.org/shields/build/devel/bioc/metaseqR2.svg)
  ![Bioconductor platforms](http://www.bioconductor.org/shields/availability/devel/metaseqR2.svg)
  ![Bioconductor dependencies](http://www.bioconductor.org/shields/dependencies/devel/metaseqR2.svg)
  </br>
  ![GitHub](https://img.shields.io/github/license/pmoulos/metaseqR2)
  ![GitHub repo size](https://img.shields.io/github/repo-size/pmoulos/metaseqR2)
  ![GitHub issues](https://img.shields.io/github/issues/pmoulos/metaseqR2)
<!-- badges: end -->

# metaseqR2

An R package for the analysis, meta-analysis and result reporting of RNA-Seq 
gene expression data - Next Generation!

## Citation

metaseqR2 along with further research regarding the abilities of the PANDORA 
algorithm was published in [Briefings in Bioinformatics](https://academic.oup.com/bib/advance-article/doi/10.1093/bib/bbaa156/5890504). If you use metaseqR2 in your research, please cite:

> Dionysios Fanidis, Panagiotis Moulos: **Integrative, normalization-insusceptible statistical analysis of RNA-Seq data, with improved differential expression and unbiased downstream functional analysis**, *Briefings in Bioinformatics*, 2020, bbaa156, DOI: [10.1093/bib/bbaa156](https://doi.org/10.1093/bib/bbaa156)

## Installation from Bioconductor

```
if (!requireNamespace("BiocManager",quietly=TRUE))
    install.packages("BiocManager")

library(BiocManager)

BiocManager::install("metaseqR2")

# or for development version to be installed
# BiocManager::install("metaseqR2",version="devel")
```

## Installation from GitHub

Use with caution as the latest version may be unstable, although typical
Bioconductor checks are executed before each push.

```
if (!requireNamespace("devtools",quietly=TRUE))
    install.packages("devtools")

library(devtools)
install_github("pmoulos/metaseqR2")
```

## Installation from source

The same things apply regarding stability.

```
git clone https://github.com/pmoulos/metaseqR2.git
mkdir metaseqR2-build
rsync -avr --exclude=README.md --exclude=.git --exclude=.gitignore  \
    ./metaseqR2-local/ ./metaseqR2-build/metaseqR2
cd ./metaseqR2-build
R CMD build ./metaseqR2
```

This will take some time to build the vignettes. If you do not need them:

```
R CMD build --no-build-vignettes ./metaseqR2
```

And then install

```
R CMD INSTALL ./metaseqR2_x.y.z.tar.gz
```

Please report any issues [here](https://github.com/pmoulos/metaseqR2-local/issues). 

## metaseqR2 annotation database

If you do not wish to build annotation databases on your own using the
```buildAnnotationDatabase``` function, you can find complete pre-built 
annotation SQLite databases
[here](https://drive.google.com/drive/folders/15lOY9PBggCcaoohO_0rQTvExXenqah55?usp=sharing). 
New versions will be constructed from time to time, most probably whenever a new
Ensembl release comes live.

The prebuilt annotations contain:

* Annotations for all supported organsisms for all types of metaseqR2 analyses.
* For every supported organism:
  + For the latest version of each genome, the latest two Ensembl required
  annotations.
  + For all other versions of each genome, the latest Ensembl required
  annotations supporting that particular version.
  + UCSC and RefSeq annotations as fetched in the day of the build (denoted
  by the folder name in the above link).
  
The SQLite database must be placed in ```system.file(package="metaseqR2")``` and
named ```annotation.sqlite```, that is
```file.path(system.file(package="metaseqR2"),"annotation.sqlite")```. Otherwise
you will have to provide your desired location in each ```metaseqr2``` call.
Alternatively, on-the-fly download is still supported but is inneficient.

## List of required packages

metaseqR2 would benefit from the existence of all the following packages:

* ABSSeq
* Biobase
* BiocGenerics
* BiocManager
* BiocParallel
* BiocStyle
* biomaRt
* Biostrings
* BSgenome
* corrplot
* DESeq2
* DSS
* DT
* EDASeq
* edgeR
* harmonicmeanp
* genefilter
* GenomeInfoDb
* GenomicAlignments
* GenomicFeatures
* GenomicRanges
* gplots
* graphics
* grDevices
* heatmaply
* htmltools
* httr
* IRanges
* jsonlite
* knitr
* limma
* log4r
* magrittr
* Matrix
* methods
* NBPSeq
* pander
* parallel
* qvalue
* rmarkdown
* rmdformats
* RMySQL
* Rsamtools
* RSQLite
* rtracklayer
* RUnit
* S4Vectors
* Seqinfo
* splines
* stats
* stringr
* SummarizedExperiment
* survcomp
* TCC
* utils
* VennDiagram
* vsn
* zoo

A recent version of [Pandoc](https://pandoc.org/) is also required, ideally
above 2.0.

The files C.bam and T.bam are also included in the Bioconductor package recoup
and are part of the data described in Nikolaou et al., 2017, PMID: 28746875 
(https://www.ncbi.nlm.nih.gov/pubmed/28746875)
Specifically, each file comprises 50000 random reads from H4K20me3 methylation
data in wild-type adult mice (C.bam) and the respective knock-out of the Kmt5a 
gene (T.bam).
These are not RNA-Seq data but are well suited to demonstrate the counting and
other functions they are used with.

normalizeEdaseq <- function(geneCounts,sampleList,normArgs=NULL,
    geneData=NULL,output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","edaseq")
    if (!is.matrix(geneCounts))
        geneCounts <- as.matrix(geneCounts)
    if (!is.null(geneData) && is(geneData,"GenomicRanges")) {
        gl <- NULL
        if (!is.null(attr(geneData,"geneLength")))
            gl <- attr(geneData,"geneLength")
        geneData <- as.data.frame(geneData)
        geneData <- geneData[,c(1,2,3,6,7,5,8,9)]
        if (!is.null(gl))
            attr(geneData,"geneLength") <- gl
    }
    if (!is.null(geneData) && is.null(attr(geneData,"geneLength")))
        attr(geneData,"geneLength") <- rep(1,nrow(geneCounts))
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    classes <- asClassVector(sampleList)
    if (is.null(geneData)) {
        seqGenes <- newSeqExpressionSet(
            geneCounts,
            phenoData=AnnotatedDataFrame(
                data.frame(
                    conditions=classes,
                    row.names=colnames(geneCounts)
                )
            ),
            featureData=AnnotatedDataFrame(
                data.frame(
                    length=rep(1,nrow(geneCounts)),
                    row.names=rownames(geneCounts)
                )
            )
        )
        seqGenes <- betweenLaneNormalization(withinLaneNormalization(seqGenes,
            "length",which=normArgs$within.which),
            which=normArgs$between.which)
    }
    else {
        seqGenes <- newSeqExpressionSet(
            geneCounts,
            phenoData=AnnotatedDataFrame(
                data.frame(
                    conditions=classes,
                    row.names=colnames(geneCounts)
                )
            ),
            featureData=AnnotatedDataFrame(
                data.frame(
                    gc=geneData$gc_content,
                    length=attr(geneData,"geneLength"),
                    row.names=if (is.data.frame(geneData)) rownames(geneData)
                        else names(geneData)
                )
            )
        )
        seqGenes <- betweenLaneNormalization(withinLaneNormalization(seqGenes,
            "gc",which=normArgs$within.which),which=normArgs$between.which)
    }
    if (output=="matrix")
        return(normCounts(seqGenes)) # Class: matrix
    else if (output=="native")
        return(seqGenes) # Class: SeqExpressionSet
}

normalizeDeseq <- function(geneCounts,sampleList,normArgs=NULL,
    output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","deseq")
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    classes <- asClassVector(sampleList)
    cds <- newCountDataSet(geneCounts,classes)
    #cds <- DESeq::estimateSizeFactors(cds,locfunc=normArgs$locfunc)
    cds <- estimateSizeFactors(cds,locfunc=normArgs$locfunc)
    if (output=="native")
        return(cds) # Class: .CountDataSet
    else if (output=="matrix")
        #return(round(DESeq::counts(cds,normalized=TRUE))) # Class: matrix
        return(round(counts(cds,normalized=TRUE))) # Class: matrix
}

normalizeDeseq2 <- function(geneCounts,sampleList,normArgs=NULL,
    output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","deseq2")
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    classes <- asClassVector(sampleList)
    conditions=as.factor(classes)                 
    colData=DataFrame(conditions)
    design= as.formula(c("~", names(colData[1]))) 
    dds <- DESeqDataSetFromMatrix(geneCounts,colData,design=design,
        tidy=normArgs$tidy)
    dds <- DESeq2::estimateSizeFactors(dds,type=normArgs$type,
        locfunc=normArgs$locfunc)
    if (output=="native")
        return(dds) # Class: DESeqDataSet
    else if (output=="matrix")
        return(round(DESeq2::counts(dds,normalized=TRUE))) # Class: matrix
}

normalizeEdger <- function(geneCounts,sampleList,normArgs=NULL,
    output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","edger")
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    classes <- asClassVector(sampleList)
    dge <- DGEList(counts=geneCounts,group=classes)
    dge <- calcNormFactors(dge,method=normArgs$method,
        refColumn=normArgs$refColumn,logratioTrim=normArgs$logratioTrim,
        sumTrim=normArgs$sumTrim,doWeighting=normArgs$doWeighting,
        Acutoff=normArgs$Acutoff,p=normArgs$p)
    if (output=="native")
        return(dge) # Class: DGEList
    else if (output=="matrix") {
        scl <- dge$samples$lib.size * dge$samples$norm.factors
        return(round(t(t(dge$counts)/scl)*mean(scl)))
        #return(round(dge$pseudo.counts)) # Class: matrix
    }
}

normalizeNoiseq <- function(geneCounts,sampleList,normArgs=NULL,
    geneData=NULL,logOffset=1,output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","noiseq")
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    classes <- asClassVector(sampleList)
    if (!is.null(geneData)) {
        if (is(geneData,"GenomicRanges")) {
            gl <- NULL
            if (!is.null(attr(geneData,"geneLength")))
                gl <- attr(geneData,"geneLength")
            else
                gl <- width(geneData)
            geneData <- as.data.frame(geneData)
            geneData <- geneData[,c(1,2,3,6,7,5,8,9)]
            if (!is.null(gl))
                attr(geneData,"geneLength") <- gl
        }
        else {
            if (is.null(attr(geneData,"geneLength"))) {
                gl <- geneData$end - geneData$start + 1
                attr(geneData,"geneLength") <- gl
            }
        }
    }
    if (is.null(geneData)) {
        #nsObj <- NOISeq::readData(
        nsObj <- readData(
            data=geneCounts,
            factors=data.frame(class=classes)
        )
    }
    else {
        gcContent <- geneData$gc_content
        geneLength <- attr(geneData,"geneLength")
        biotype <- as.character(geneData$biotype)
        names(gcContent) <- names(biotype) <- names(geneLength) <- 
            if (is.data.frame(geneData)) rownames(geneData) else names(geneData)
        #nsObj <- NOISeq::readData(
        nsObj <- readData(
            data=geneCounts,
            length=geneLength,
            gc=gcContent,
            chromosome=geneData[,c(1,2,3)],
            factors=data.frame(class=classes),
            biotype=biotype
        )
    }
    normArgs$k <- logOffset # Set the zero fixing constant
    switch(normArgs$method,
        rpkm = {
            #M <- NOISeq::rpkm(assayData(nsObj)$exprs,long=normArgs$long,
            #    k=normArgs$k,lc=normArgs$lc)
            M <- noirpkm(assayData(nsObj)$exprs,long=normArgs$long)
        },
        uqua = {
            M <- noiuqua(assayData(nsObj)$exprs,long=normArgs$long,
                k=normArgs$k,lc=normArgs$lc)
        },
        tmm = {
            M <- noitmm(assayData(nsObj)$exprs,long=normArgs$long,
                k=normArgs$k,lc=normArgs$lc,refColumn=normArgs$refColumn,
                logratioTrim=normArgs$logratioTrim,sumTrim=normArgs$sumTrim,
                doWeighting=normArgs$doWeighting,Acutoff=normArgs$Acutoff)
        }
    )
    if (output=="native") {
        if (is.null(geneData))
            #return(NOISeq::readData(
            return(readData(
                data=M,
                factors=data.frame(class=classes)
        )) # Class: CD
        else    
            #return(NOISeq::readData(
            return(readData(
                data=M,
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )) # Class: CD
    }
    else if (output=="matrix")
        return(as.matrix(round(M))) # Class: matrix
}

normalizeNbpseq <- function(geneCounts,sampleList,normArgs=NULL,
    libsizeList=NULL,output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","nbpseq")
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    classes <- asClassVector(sampleList)
    if (is.null(libsizeList)) {
        libsizeList <- vector("list",length(classes))
        names(libsizeList) <- unlist(sampleList,use.names=FALSE)
        for (n in names(libsizeList))
            libsizeList[[n]] <- sum(geneCounts[,n])
    }
    libSizes <- unlist(libsizeList)
    norm.factors <- estimate.norm.factors(geneCounts,lib.sizes=libSizes,
        method=normArgs$method)
    #if (normArgs$main.method=="nbpseq")
    #    nbData <- prepare.nbData(geneCounts,libSizes=libSizes,
    #    norm.factors=norm.factors)
    #else if (normArgs$main.method=="nbsmyth")
        nbData <- prepare.nbp(geneCounts,classes,lib.sizes=libSizes,
            norm.factors=norm.factors,thinning=normArgs$thinning)
    if (output=="native")
        return(nbData) # Class: list or nbp
    else if (output=="matrix") {
        #if (normArgs$main.method=="nbpseq") {
        #    normCounts <- matrix(0,nrow(geneCounts),ncol(geneCounts))
        #    for (i in seq_len(ncol(geneCounts)))
        #        normCounts[,i] <- norm.factors[i]*geneCounts[,i]
        #}
        #else if (normArgs$main.method=="nbsmyth") 
            normCounts <- nbData$pseudo.counts
        return(as.matrix(round(normCounts))) # Class: matrix
    }
}

normalizeAbsseq <- function(geneCounts,sampleList,normArgs=NULL,
    output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","absseq")
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    #classes <- asClassVector(sampleList)
    #groups= as.factor(classes)
    abs <- ABSDataSet(geneCounts,normMethod= normArgs$normMethod)  
    abs <- normalFactors(abs)
    excounts(abs) <- ABSSeq::counts(abs,norm=TRUE)
    if (output=="native")
        return(abs) # Class: ABSDataSet
    else if (output=="matrix")
        return(as.matrix(round(excounts(abs)))) # Class: matrix
}

normalizeDss <- function(geneCounts,sampleList,normArgs=NULL,
    output=c("matrix","native")) {
    if (is.null(normArgs))
        normArgs <- getDefaults("normalization","dss")
    output <- tolower(output[1])
    checkTextArgs("output",output,c("matrix","native"))
    
    # make the design
    classes <- asClassVector(sampleList)
    design <- as.data.frame(classes)
    colnames(design) <- "designs"
    
    # newSeqCountSet takes only matrices as gene.counts
    if (is(geneCounts,"data.frame")) {
        geneCountsTmp <- geneCounts
        samCols <- match(unlist(sampleList),colnames(geneCountsTmp))
        samCols <- samCols[which(!is.na(samCols))]
        geneCountsTmp <- as.matrix(geneCounts[,samCols,drop=FALSE])
    }
    if (is(geneCounts,"matrix"))
        geneCountsTmp <- geneCounts
    
    seqD <- newSeqCountSet(geneCountsTmp,design) # create the class
    # estimate normalization factors
    seqD <- estNormFactors(seqD,method=normArgs$method)
    
    if (output=="native")
        return(seqD) # Class: SeqCountSet
    else if (output=="matrix") {
        # I cannot extract normalized counts from DSS and thus I do this small 
        # dribble in order to output a matrix
        theDesign <- data.frame(condition=classes,row.names=colnames(seqD)) 
        cds <- newCountDataSet(as.matrix(round(assayData(seqD)$exprs)),
            theDesign$condition)
        #DESeq::sizeFactors(cds) <- normalizationFactor(seqD)
        sizeFactors(cds) <- normalizationFactor(seqD)
        #counts <- as.matrix(DESeq::counts(cds,normalized=TRUE))
        counts <- as.matrix(counts(cds,normalized=TRUE))
        return(as.matrix(round(counts))) # Class: matrix
    }
}

buildAnnotationDatabase <- function(organisms,sources,
    #db=file.path(path.expand("~"),".metaseqR","annotation.sqlite"),
    db=file.path(system.file(package="metaseqR2"),"annotation.sqlite"),
    forceDownload=TRUE,rc=NULL) {
    if (missing(organisms))
        organisms <- getSupportedOrganisms()
    if (missing(sources))
        sources <- getSupportedRefDbs()
    
    orgIsList <- FALSE
    if (!is.list(organisms) && is.character(organisms)
        && "ensembl" %in% sources)
        warning("When ensembl is in the annotation sources to download, it is ",
            "advised to provide organisms as\na named list with names the ",
            "requested organisms and list members the Ensembl versions.\n",
            "Otherwise, the latest Ensembl version for each organism will be ",
            "used.",immediate.=TRUE)
    if (is.list(organisms)) {
        if (is.null(names(organisms)))
            stop("When organisms is a list, it must be named!")
        orgList <- organisms
        organisms <- names(organisms)
        orgIsList <- TRUE
    }
    
    checkTextArgs("organisms",organisms,getSupportedOrganisms(),multiarg=TRUE)
    checkTextArgs("sources",sources,getSupportedRefDbs(),multiarg=TRUE)
    
    if (!requireNamespace("GenomeInfoDb"))
        stop("R package GenomeInfoDb is required to construct annotation ",
            "stores!")
    
    # Check database path
    if (!dir.exists(dirname(db)))
        dir.create(dirname(db),recursive=TRUE,mode="0755")
    
    # Initialize or open the annotation SQLite datatabase
    message("Opening metaseqR2 SQLite database ",db)
    con <- initDatabase(db)
    
    for (s in sources) {
        for (o in organisms) {
            # Retrieving genome info. We will be inserting the seqinfo to the
            # sqlite multiple times for the sake of simplicity regarding later
            # deletion by foreign key cascade.
            message("Retrieving genome information for ",o," from ",s)
            sf <- getSeqInfo(o)
            
            # Now, we must derive versioning. For Ensembl it's obvious as it has
            # official versions. For UCSC we need to keep a date track. Then 
            # inside metaseqR, if date not provided, it will automatically
            # detect the latest one (as in Ensembl with versions, if not
            # provided).
            if (s == "ensembl") {
                if (orgIsList)
                    vs <- orgList[[o]]
                else {
                    vss <- getUcscToEnsembl(o)
                    vs <- vss[length(vss)]
                }
                vs <- .validateEnsemblVersions(o,vs)
                if (is.null(vs)) { # Something went wrong... Get latest again
                    vss <- getUcscToEnsembl(o)
                    vs <- vss[length(vss)]
                }
            }
            else if (s %in% getSupportedUcscDbs())
                vs <- format(Sys.Date(),"%Y%m%d")
                        
            for (v in vs) {
                #storePath <- file.path(home,s,o,v)
                #if (!dir.exists(storePath))
                #    dir.create(storePath,recursive=TRUE,mode="0755")
                    
                # Retrieve gene annotations
                if (.annotationExists(con,o,s,v,"gene") && !forceDownload)
                #if (file.exists(file.path(storePath,"gene.rda")) 
                #    && !forceDownload)
                    message("Gene annotation for ",o," from ",s," version ",v,
                        " has already been created and will be skipped.\nIf ",
                        "you wish to recreate it choose forceDownload = TRUE.")
                else {
                    message("Retrieving gene annotation for ",o," from ",s,
                        " version ",v)
                    ann <- getAnnotation(o,"gene",refdb=s,ver=v,rc=rc)
                    #gene <- makeGRangesFromDataFrame(
                    #    df=ann,
                    #    seqinfo=sf,
                    #    keep.extra.columns=TRUE,
                    #    seqnames.field="chromosome"
                    #)
                    #save(gene,file=file.path(storePath,"gene.rda"),
                    #    compress=TRUE)
                    
                    # First drop if previously exists
                    nr <- .dropAnnotation(con,o,s,v,"gene")
                    # Then insert to the contents table so as to get the content
                    # id to attach in the annotation table
                    nr <- .insertContent(con,o,s,v,"gene")
                    nid <- .annotationExists(con,o,s,v,"gene",out="id")
                    # If something happens, the whole procedure will break
                    # anyway
                    # Add content_id
                    ann$content_id <- rep(nid,nrow(ann))
                    sfGene <- sf
                    sfGene$content_id <- rep(nid,nrow(sfGene))
                    # Write genes and seqinfo
                    dbWriteTable(con,"gene",ann,row.names=FALSE,append=TRUE)
                    dbWriteTable(con,"seqinfo",sfGene,row.names=FALSE,
                        append=TRUE)
                }
                
                # Retrieve transcript annotations
                if (.annotationExists(con,o,s,v,"transcript") && !forceDownload)
                    message("Transcript annotation for ",o," from ",s,
                        " version ",v," has already been created and will be ",
                        "skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    message("Retrieving transcript annotation for ",o,
                        " from ",s," version ",v)
                    ann <- getAnnotation(o,"transcript",refdb=s,ver=v,rc=rc)
                    nr <- .dropAnnotation(con,o,s,v,"transcript")
                    nr <- .insertContent(con,o,s,v,"transcript")
                    nid <- .annotationExists(con,o,s,v,"transcript",out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfTranscript <- sf
                    sfTranscript$content_id <- rep(nid,nrow(sfTranscript))
                    dbWriteTable(con,"transcript",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfTranscript,row.names=FALSE,
                        append=TRUE)
                }
                
                # Then summarize the transcripts and write again with type 
                # sum_transcript
                if (.annotationExists(con,o,s,v,"summarized_transcript") 
                    && !forceDownload)
                    message("Summarized transcript annotation for ",o," from ",
                        s," version ",v," has already been created and will ",
                        "be skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"transcript")) 
                        stop("Transcript annotation for ",o," from ",s,
                            " version ",v," is required in order to build ",
                            "predefined merged transcript regions for read ",
                            "counting.\nPlease rerun the ",
                            "buildAnnotationDatabase function with ",
                            "appropriate parameters.")
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"transcript",
                        "gene")
                    message("Merging transcripts for ",o," from ",s,
                        " version ",v)
                    #annGr <- reduceTranscripts(annGr)
                    #ann <- as.data.frame(annGr)
                    annList <- reduceTranscripts(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    
                    nr <- .dropAnnotation(con,o,s,v,"summarized_transcript")
                    nr <- .insertContent(con,o,s,v,"summarized_transcript")
                    nid <- .annotationExists(con,o,s,v,"summarized_transcript",
                        out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumTranscript <- sf
                    sfSumTranscript$content_id <- rep(nid,nrow(sfSumTranscript))
                    dbWriteTable(con,"summarized_transcript",ann,
                        row.names=FALSE,append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumTranscript,row.names=FALSE,
                        append=TRUE)
                }
                
                # Retrieve 3' UTR annotations
                if (.annotationExists(con,o,s,v,"utr") && !forceDownload)
                    message("3' UTR annotation for ",o," from ",s," version ",
                        v," has already been created and will be skipped.\nIf ",
                        "you wish to recreate it choose forceDownload = TRUE.")
                else {
                    message("Retrieving 3' UTR annotation for ",o,
                        " from ",s," version ",v)
                    ann <- getAnnotation(o,"utr",refdb=s,ver=v,rc=rc)
                    nr <- .dropAnnotation(con,o,s,v,"utr")
                    nr <- .insertContent(con,o,s,v,"utr")
                    nid <- .annotationExists(con,o,s,v,"utr",out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfUtr <- sf
                    sfUtr$content_id <- rep(nid,nrow(sfUtr))
                    dbWriteTable(con,"utr",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfUtr,row.names=FALSE,
                        append=TRUE)
                }
                
                # Then summarize the 3'utrs per gene and write again with type 
                # sum_utr
                if (.annotationExists(con,o,s,v,"summarized_3utr")
                    && !forceDownload)
                    message("Summarized 3' UTR annotation for ",o," from ",
                        s," version ",v," has already been created and will ",
                        "be skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"utr")) 
                        stop("3' UTR annotation for ",o," from ",s," version ",
                            vs," is required in order to build predefined ",
                            "merged 3' UTR regions for read counting.\nPlease ",
                            "rerun the buildAnnotationDatabase function with ",
                            "appropriate parameters.")
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"gene","utr")
                    message("Merging gene 3' UTRs for ",o," from ",s,
                        " version ",v)
                    #annGr <- reduceTranscripts(annGr)
                    #ann <- as.data.frame(annGr)
                    annList <- reduceTranscripts(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- .dropAnnotation(con,o,s,v,"summarized_3utr")
                    nr <- .insertContent(con,o,s,v,"summarized_3utr")
                    nid <- .annotationExists(con,o,s,v,"summarized_3utr",
                        out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumUtr <- sf
                    sfSumUtr$content_id <- rep(nid,nrow(sfSumUtr))
                    dbWriteTable(con,"summarized_3utr",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumUtr,row.names=FALSE,
                        append=TRUE)
                    
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_utr_length")
                    nr <- .insertContent(con,o,s,v,"active_utr_length")
                    nid <- .annotationExists(con,o,s,v,"active_utr_length",
                        out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_utr_length",active,row.names=FALSE,
                        append=TRUE)
                }
                
                # Then summarize the 3'utrs per transcript and write again with 
                # type sum_utr
                if (.annotationExists(con,o,s,v,"summarized_3utr_transcript")
                    && !forceDownload)
                    message("Summarized 3' UTR annotation per transcript for ",
                        o," from ",s," version ",v," has already been created ",
                        "and will be skipped.\nIf you wish to recreate it ",
                        "choose forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"utr"))
                        stop("3' UTR annotation per transcript for ",o," from ",
                            s," version ",v," is required in order to build ",
                            "predefined merged 3'UTR regions for read ",
                            "counting.\nPlease rerun the ",
                            "buildAnnotationDatabase function with ",
                            "appropriate parameters.")
                    annGr <- 
                        .loadPrebuiltAnnotation(con,o,s,v,"transcript","utr")
                    message("Merging transcript 3' UTRs for ",o," from ",s,
                        " version ",v)
                    #annGr <- reduceTranscriptsUtr(annGr)
                    #ann <- as.data.frame(annGr)
                    annList <- reduceTranscriptsUtr(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- 
                        .dropAnnotation(con,o,s,v,"summarized_3utr_transcript")
                    nr <- .insertContent(con,o,s,v,"summarized_3utr_transcript")
                    nid <- .annotationExists(con,o,s,v,
                        "summarized_3utr_transcript",out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumUtrTranscript <- sf
                    sfSumUtrTranscript$content_id <- 
                        rep(nid,nrow(sfSumUtrTranscript))
                    dbWriteTable(con,"summarized_3utr_transcript",ann,
                        row.names=FALSE,append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumUtrTranscript,
                        row.names=FALSE,append=TRUE)
                    
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_trans_utr_length")
                    nr <- .insertContent(con,o,s,v,"active_trans_utr_length")
                    nid <- .annotationExists(con,o,s,v,
                        "active_trans_utr_length",out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_trans_utr_length",active,
                        row.names=FALSE,append=TRUE)
                }
                
                # Retrieve exon annotations
                if (.annotationExists(con,o,s,v,"exon") && !forceDownload)
                    message("Exon annotation for ",o," from ",s," version ",v,
                        " has already been created and will be skipped.\nIf ",
                        "you wish to recreate it choose forceDownload = TRUE.")
                else {
                    message("Retrieving exon annotation for ",o," from ",s,
                        " version ",v)
                    ann <- getAnnotation(o,"exon",refdb=s,ver=v,rc=rc)
                    nr <- .dropAnnotation(con,o,s,v,"exon")
                    nr <- .insertContent(con,o,s,v,"exon")
                    nid <- .annotationExists(con,o,s,v,"exon",out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfExon <- sf
                    sfExon$content_id <- rep(nid,nrow(sfExon))
                    dbWriteTable(con,"exon",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfExon,row.names=FALSE,
                        append=TRUE)
                }
                
                # Retrieve extended annotations
                if (.annotationExists(con,o,s,v,"transexon") && !forceDownload)
                    message("Extended exon annotation for ",o," from ",s,
                        " version ",v," has already been created and will be ",
                        "skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    message("Retrieving extended exon annotation for ",o,
                        " from ",s," version ",v)
                    ann <- getAnnotation(o,"transexon",refdb=s,ver=v,rc=rc)
                    nr <- .dropAnnotation(con,o,s,v,"transexon")
                    nr <- .insertContent(con,o,s,v,"transexon")
                    nid <- .annotationExists(con,o,s,v,"transexon",out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfTrExon <- sf
                    sfTrExon$content_id <- rep(nid,nrow(sfExon))
                    dbWriteTable(con,"transexon",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfTrExon,row.names=FALSE,
                        append=TRUE)
                }
                
                # Then summarize the exons per gene and write again with type 
                # summarized_exon
                if (.annotationExists(con,o,s,v,"summarized_exon")
                    && !forceDownload)
                    message("Summarized exon annotation for ",o," from ",s,
                        " version ",v," has already been created and will be ",
                        "skipped.\nIf you wish to recreate it choose ",
                        "forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"exon")) 
                        stop("Exon annotation for ",o," from ",s," version ",v,
                            " is required in order to build predefined merged ",
                            "exon regions for RNA-Seq (exon) coverage ",
                            "calculations.\nPlease rerun the ",
                            "buildAnnotationDatabase function with ",
                            "appropriate parameters.")
                            
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"exon","exon")
                    message("Merging exons for ",o," from ",s," version ",v)
                    annList <- reduceExons(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- .dropAnnotation(con,o,s,v,"summarized_exon")
                    nr <- .insertContent(con,o,s,v,"summarized_exon")
                    nid <- .annotationExists(con,o,s,v,"summarized_exon",
                        out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumExon <- sf
                    sfSumExon$content_id <- rep(nid,nrow(sfSumExon))
                    dbWriteTable(con,"summarized_exon",ann,row.names=FALSE,
                        append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumExon,row.names=FALSE,
                        append=TRUE)        
                            
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_length")
                    nr <- .insertContent(con,o,s,v,"active_length")
                    nid <- .annotationExists(con,o,s,v,"active_length",
                        out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_length",active,row.names=FALSE,
                        append=TRUE)
                }
                
                # Then summarize the exons per transcript and write again with 
                # type summarized_transcript_exon
                if (.annotationExists(con,o,s,v,"summarized_transcript_exon")
                    && !forceDownload)
                    message("Summarized exon annotation per transcript for ",o,
                        " from ",s," version ",v," has already been created ",
                        "and will be skipped.\nIf you wish to recreate it ",
                        "choose forceDownload = TRUE.")
                else {
                    if (!.annotationExists(con,o,s,v,"transexon")) 
                        stop("Extended exon annotation for ",o," from ",s,
                            " version ",v," is required in order to build ",
                            "summarized exon annotation per transcript\n",
                            "Please rerun the buildAnnotationDatabase ",
                            "function with appropriate parameters.")
                
                    annGr <- .loadPrebuiltAnnotation(con,o,s,v,"transcript",
                        "exon")
                    message("Merging exons for ",o," from ",s," version ",v)
                    annList <- reduceTranscriptsExons(annGr)
                    ann <- as.data.frame(annList$model)
                    ann <- ann[,c(1,2,3,6,7,5,8,9)]
                    names(ann)[1] <- "chromosome"
                    ann$chromosome <- as.character(ann$chromosome)
                    ann <- ann[order(ann$chromosome,ann$start),]
                    nr <- .dropAnnotation(con,o,s,v,
                        "summarized_transcript_exon")
                    nr <- .insertContent(con,o,s,v,"summarized_transcript_exon")
                    nid <- .annotationExists(con,o,s,v,
                        "summarized_transcript_exon",out="id")
                    ann$content_id <- rep(nid,nrow(ann))
                    sfSumTrExon <- sf
                    sfSumTrExon$content_id <- rep(nid,nrow(sfSumTrExon))
                    dbWriteTable(con,"summarized_transcript_exon",ann,
                        row.names=FALSE,append=TRUE)
                    dbWriteTable(con,"seqinfo",sfSumTrExon,row.names=FALSE,
                        append=TRUE)        
                            
                    activeLength <- annList$length
                    nr <- .dropAnnotation(con,o,s,v,"active_trans_exon_length")
                    nr <- .insertContent(con,o,s,v,"active_trans_exon_length")
                    nid <- .annotationExists(con,o,s,v,
                        "active_trans_exon_length",out="id")
                    active <- data.frame(
                        name=names(activeLength),
                        length=activeLength,
                        content_id=rep(nid,length(activeLength))
                    )
                    dbWriteTable(con,"active_trans_exon_length",active,
                        row.names=FALSE,append=TRUE)
                }
            }
        }
    }
    
    dbDisconnect(con)
}

# GTF only!
buildCustomAnnotation <- function(gtfFile,metadata,
    db=file.path(system.file(package="metaseqR2"),"annotation.sqlite"),
    #db=file.path(path.expand("~"),".metaseqR","annotation.sqlite"),
    rewrite=TRUE) {
    # Check metadata
    if (is.null(metadata$organism))
        stop("An organism name must be provided with metadata!")
    if (is.null(metadata$source)) {
        warning("A source should be provided with metadata! Using 'inhouse'...",
            immediate.=TRUE)
        metadata$source <- "inhouse"
    }
    if (is.null(metadata$version)) {
        warning("A version should be provided with metadata! Using today...",
            immediate.=TRUE)
        metadata$version <- format(Sys.Date(),"%Y%m%d")
    }
    if (is.null(metadata$chromInfo)) {
        warning("Chromosomal lengths should be provided with metadata! ",
            "Only chromosome names will be available... ",immediate.=TRUE)
        metadata$chromInfo <- NULL
    }
    else {
        str <- metadata$chromInfo
        if (is.character(str)) {
            out <- tryCatch(open(Rsamtools::BamFile(str)),error=function(e) e)
            if (inherits(out,"error")) # Not a BAM file, try to read.delim
                metadata$chromInfo <- read.delim(str,row.names=1)
            else
                metadata$chromInfo <- .chromInfoFromBAM(str)
        }
        if (!is.data.frame(metadata$chromInfo))
            stop("metadata$chromInfo must be a data frame!")
    }
    
    # Check database path
    if (!dir.exists(dirname(db)))
        dir.create(dirname(db),recursive=TRUE,mode="0755")
    
    # Initialize or open the annotation SQLite datatabase
    message("Opening metaseqR2 SQLite database ",db)
    con <- initDatabase(db)
    
    parsed <- parseCustomGtf(gtfFile)
    
    s <- tolower(metadata$source)
    o <- tolower(metadata$organism)
    v <- metadata$version
        
    # Retrieve gene annotations
    if (.annotationExists(con,o,s,v,"gene") && !rewrite)
        message("Gene annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving gene annotation for ",o," from ",s," version ",v,
            " from ",gtfFile)
        ann <- annotationFromCustomGtf(parsed,level="gene",type="gene",
            asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"gene")
        nr <- .insertContent(con,o,s,v,"gene",1)
        nid <- .annotationExists(con,o,s,v,"gene",out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfGene <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfGene$content_id <- rep(nid,nrow(sfGene))
        dbWriteTable(con,"gene",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfGene,row.names=FALSE,append=TRUE)
    }
    
    # Retrieve transcript annotations
    if (.annotationExists(con,o,s,v,"transcript") && !rewrite)
        message("Transcript annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving transcript annotation for ",o," from ",s,
            " version ",v)
        ann <- annotationFromCustomGtf(parsed,level="transcript",type="gene",
            asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"transcript")
        nr <- .insertContent(con,o,s,v,"transcript",1)
        nid <- .annotationExists(con,o,s,v,"transcript",out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfTranscript <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfTranscript$content_id <- rep(nid,nrow(sfTranscript))
        dbWriteTable(con,"transcript",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfTranscript,row.names=FALSE,append=TRUE)
    }
    
    # Then summarize the transcripts and write again with type sum_transcript
    if (.annotationExists(con,o,s,v,"summarized_transcript") && !rewrite)
        message("Summarized transcript annotation for ",o," from ",s,
            " version ",v," has already been created and will be skipped.\nIf ",
            "you wish to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized transcript annotation for ",o," from ",s,
            " version ",v)
        ann <- annotationFromCustomGtf(parsed,level="transcript",type="gene",
            summarized=TRUE,asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"summarized_transcript")
        nr <- .insertContent(con,o,s,v,"summarized_transcript",1)
        nid <- .annotationExists(con,o,s,v,"summarized_transcript",out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfSumTranscript <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfSumTranscript$content_id <- rep(nid,nrow(sfSumTranscript))
        dbWriteTable(con,"summarized_transcript",ann,row.names=FALSE,
            append=TRUE)
        dbWriteTable(con,"seqinfo",sfSumTranscript,row.names=FALSE,append=TRUE)
    }
    
    # Retrieve 3' UTR annotations
    if (.annotationExists(con,o,s,v,"utr") && !rewrite)
        message("3' UTR annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving 3' UTR annotation for ",o," from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,level="gene",type="utr",asdf=TRUE)
        if (nrow(ann) > 0) {
            nr <- .dropAnnotation(con,o,s,v,"utr")
            nr <- .insertContent(con,o,s,v,"utr",1)
            nid <- .annotationExists(con,o,s,v,"utr",out="id")
            ann$content_id <- rep(nid,nrow(ann))
            sfUtr <- .chromInfoToSeqInfoDf(metadata$chromInfo)
            sfUtr$content_id <- rep(nid,nrow(sfUtr))
            dbWriteTable(con,"utr",ann,row.names=FALSE,append=TRUE)
            dbWriteTable(con,"seqinfo",sfUtr,row.names=FALSE,append=TRUE)
        }
        else
            message("3' UTR annotation for ",o," from ",s," version ",v," is ",
                "not available in the provided GTF file.")
    }
    
    # Then summarize the 3'UTRs and write again with type sum_transcript
    if (.annotationExists(con,o,s,v,"summarized_3utr") && !rewrite)
        message("Summarized 3' UTR annotation for ",o," from ",s," version ",v,
            " has already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized 3' UTR annotation per gene for ",o,
            " from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,level="gene",type="utr",
            summarized=TRUE,asdf=TRUE)
        
        if (nrow(ann) > 0) {
            activeLength <- attr(ann,"activeLength")
            
            nr <- .dropAnnotation(con,o,s,v,"summarized_3utr")
            nr <- .insertContent(con,o,s,v,"summarized_3utr",1)
            nid <- .annotationExists(con,o,s,v,"summarized_3utr",out="id")
            ann$content_id <- rep(nid,nrow(ann))
            sfSumUtr <- .chromInfoToSeqInfoDf(metadata$chromInfo)
            sfSumUtr$content_id <- rep(nid,nrow(sfSumUtr))
            dbWriteTable(con,"summarized_3utr",ann,row.names=FALSE,append=TRUE)
            dbWriteTable(con,"seqinfo",sfSumUtr,row.names=FALSE,append=TRUE)
            
            nr <- .dropAnnotation(con,o,s,v,"active_utr_length")
            nr <- .insertContent(con,o,s,v,"active_utr_length",1)
            nid <- .annotationExists(con,o,s,v,"active_utr_length",out="id")
            active <- data.frame(
                name=names(activeLength),
                length=activeLength,
                content_id=rep(nid,length(activeLength))
            )
            dbWriteTable(con,"active_utr_length",active,row.names=FALSE,
                append=TRUE)
        }
        else
            message("3' UTR annotation for ",o," from ",s," version ",v," is ",
                "not available in the provided GTF file.")
    }
    
    # Then summarize the 3'utrs per transcript and write again with 
    # type sum_utr
    if (.annotationExists(con,o,s,v,"summarized_3utr_transcript") && !rewrite)
        message("Summarized 3' UTR annotation per transcript for ",o," from ",s,
            " version ",v," has already been created and will be skipped.\nIf ",
            "you wish to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized 3' UTR annotation per transcript for ",o,
            " from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,level="transcript",type="utr",
            summarized=TRUE,asdf=TRUE)
        
        if (nrow(ann) > 0) {
            activeLength <- attr(ann,"activeLength")
            nr <- .dropAnnotation(con,o,s,v,"summarized_3utr_transcript")
            nr <- .insertContent(con,o,s,v,"summarized_3utr_transcript",1)
            nid <- .annotationExists(con,o,s,v,"summarized_3utr_transcript",
                out="id")
            ann$content_id <- rep(nid,nrow(ann))
            sfSumUtrTranscript <- .chromInfoToSeqInfoDf(metadata$chromInfo)
            sfSumUtrTranscript$content_id <- rep(nid,nrow(sfSumUtrTranscript))
            dbWriteTable(con,"summarized_3utr_transcript",ann,row.names=FALSE,
                append=TRUE)
            dbWriteTable(con,"seqinfo",sfSumUtrTranscript,row.names=FALSE,
                append=TRUE)
            
            nr <- .dropAnnotation(con,o,s,v,"active_trans_utr_length")
            nr <- .insertContent(con,o,s,v,"active_trans_utr_length",1)
            nid <- .annotationExists(con,o,s,v,"active_trans_utr_length",
                out="id")
            active <- data.frame(
                name=names(activeLength),
                length=activeLength,
                content_id=rep(nid,length(activeLength))
            )
            dbWriteTable(con,"active_trans_utr_length",active,row.names=FALSE,
                append=TRUE)
        }
        else
            message("3' UTR annotation for ",o," from ",s," version ",v," is ",
                "not available in the provided GTF file.")
    }
    
    # Retrieve exon annotations
    if (.annotationExists(con,o,s,v,"exon") && !rewrite)
        message("Exon annotation for ",o," from ",s," version ",v," has ",
            "already been created and will be skipped.\nIf you wish to ",
            "recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving exon annotation for ",o," from ",s," version ",v)
        ann <- annotationFromCustomGtf(parsed,level="exon",type="exon",
            asdf=TRUE)
        nr <- .dropAnnotation(con,o,s,v,"exon")
        nr <- .insertContent(con,o,s,v,"exon",1)
        nid <- .annotationExists(con,o,s,v,"exon",out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfExon <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfExon$content_id <- rep(nid,nrow(sfExon))
        dbWriteTable(con,"exon",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfExon,row.names=FALSE,append=TRUE)
    }
    
    # Then summarize the exons and write again with type sum_exon
    if (.annotationExists(con,o,s,v,"summarized_exon") && !rewrite)
        message("Summarized exon annotation for ",o," from ",s," version ",v,
            " has already been created and will be skipped.\nIf you wish to ",
            "to recreate it choose rewrite = TRUE.")
    else {
        message("Retrieving summarized exon annotation for ",o," from ",s,
            " version ",v)
        ann <- annotationFromCustomGtf(parsed,level="gene",type="exon",
            summarized=TRUE,asdf=TRUE)
        activeLength <- attr(ann,"activeLength")
        
        nr <- .dropAnnotation(con,o,s,v,"summarized_exon")
        nr <- .insertContent(con,o,s,v,"summarized_exon",1)
        nid <- .annotationExists(con,o,s,v,"summarized_exon",out="id")
        ann$content_id <- rep(nid,nrow(ann))
        sfSumExon <- .chromInfoToSeqInfoDf(metadata$chromInfo)
        sfSumExon$content_id <- rep(nid,nrow(sfSumExon))
        dbWriteTable(con,"summarized_exon",ann,row.names=FALSE,append=TRUE)
        dbWriteTable(con,"seqinfo",sfSumExon,row.names=FALSE,append=TRUE)

        nr <- .dropAnnotation(con,o,s,v,"active_length")
        nr <- .insertContent(con,o,s,v,"active_length",1)
        nid <- .annotationExists(con,o,s,v,"active_length",out="id")
        active <- data.frame(
            name=names(activeLength),
            length=activeLength,
            content_id=rep(nid,length(activeLength))
        )
        dbWriteTable(con,"active_length",active,row.names=FALSE,append=TRUE)
    }
}

.chromInfoFromBAM <- function(bam) {
    # Danger of including non-canonical chromosomes
    b <- scanBamHeader(bam)
    ci <- as.data.frame(b[[bam]]$targets)
    names(ci) <- "length"
    return(ci)
}

.chromInfoFromSeqinfo <- function(sf) {
    sf <- as.data.frame(sf)
    sf <- sf[,1,drop=FALSE]
    names(sf) <- "length"
    return(sf)
}

# Load annotation must be capable of reading custom annotation files if imported
# with metaseqR facilities
loadAnnotation <- function(genome,refdb,level=c("gene","transcript","exon"),
    type=c("gene","exon","utr"),version="auto",
    db=file.path(system.file(package="metaseqR2"),"annotation.sqlite"),
    summarized=FALSE,asdf=FALSE,rc=NULL) {
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to load annotation from database!")
    
    genome <- tolower(genome[1])
    refdb <- tolower(refdb[1])
    level <- level[1]
    checkTextArgs("level",level,c("gene","transcript","exon"),multiarg=FALSE)
    type <- type[1]
    checkTextArgs("type",type,c("gene","exon","utr"),multiarg=FALSE)
    if (version != "auto")
        checkNumArgs("version",version,"numeric")
    
    # Check if local storage has been set
    onTheFly <- FALSE
    if (file.exists(db)) {
        # Open the connection
        drv <- dbDriver("SQLite")
        con <- dbConnect(drv,dbname=db)
        
        # Is the general resource (organism, source) installed?
        if (!.annotationExists(con,genome,refdb)) {
            warning("The requested annotation does not seem to exist in the ",
                "database! It will be loaded on the fly.\nConsider importing ",
                "it by using buildAnnotationDatabase.")
            onTheFly <- TRUE
        }
        
        # If main source exists, decide on version
        if (!onTheFly) {
            if (version != "auto") {
                if (!.annotationExists(con,genome,refdb,version,
                    .annotationTypeFromInputArgs(level,type))) {
                    warning("The requested annotation version does not seem ",
                        "to exist! Have you run buildAnnotationDatabase or ",
                        "possibly mispelled? Will use newest existing version.",
                        immediate.=TRUE)
                    version <- "auto"
                }
            }
            if (version == "auto") {
                # Check if annotation exists, has been performed before
                vers <- .installedVersions(con,genome,refdb)
                vers <- sort(vers,decreasing=TRUE)
                version <- vers[1]
            }
            ann <- .loadPrebuiltAnnotation(con,genome,refdb,version,level,type,
                summarized)
            dbDisconnect(con)
            
            if (asdf) {
                a <- attr(ann,"activeLength")
                ann <- as.data.frame(unname(ann))
                ann <- ann[,c(1,2,3,6,7,5,8,9)]
                names(ann)[1] <- "chromosome"
                if (!is.null(a))
                    attr(ann,"activeLength") <- a
                return(ann)
            }
            else
                return(ann)
        }
    }
    else
        onTheFly <- TRUE
        
    if (onTheFly) {
        if (genome %in% getSupportedOrganisms()
            && refdb %in% getSupportedRefDbs()) {
            message("Getting latest annotation on the fly for ",genome," from ",
                refdb)
            ann <- .loadAnnotationOnTheFly(genome,refdb,level,type,rc)
            if (asdf) {
                a <- attr(ann,"activeLength")
                ann <- as.data.frame(unname(ann))
                ann <- ann[,c(1,2,3,6,7,5,8,9)]
                names(ann)[1] <- "chromosome"
                if (!is.null(a))
                    attr(ann,"activeLength") <- a
                return(ann)
            }
            else
                return(ann)
        }
        else {
            stop("genome and refdb not in supported automatically download ",
                "annotation options. Please use importCustomAnnotation.")
        }
    }
}

.loadPrebuiltAnnotation <- function(con,genome,refdb,version,level,type,
    summarized=FALSE) {
    metaType <- .annotationTypeFromInputArgs(level,type,summarized)
    cid <- .annotationExists(con,genome,refdb,version,metaType,out="id")
    if (metaType == "summarized_exon")
        tName <- "active_length"
    else if (metaType == "summarized_3utr")
        tName <- "active_utr_length"
    else if (metaType == "summarized_3utr_transcript")
        tName <- "active_trans_utr_length"
    else if (metaType == "summarized_transcript_exon")
        tName <- "active_trans_exon_length"
    cida <- .annotationExists(con,genome,refdb,version,"active_length",out="id")
        
    querySet <- .makeAnnotationQuerySet(metaType,cid,cida)
    
    preAnn <- dbGetQuery(con,querySet$main)
    preAnn$`_id` <- NULL
    preAnn$content_id <- NULL
    ann <- GRanges(preAnn)
    seqlevels(ann) <- unique(preAnn$chromosome)
    
    preSf <- dbGetQuery(con,querySet$seqinfo)
    preSf$`_id` <- NULL
    preSf$content_id <- NULL
    rownames(preSf) <- as.character(preSf[,1])
    if (genome %in% getSupportedOrganisms())
        sfg <- getUcscOrganism(genome)
    else
        sfg <- genome
    sf <- Seqinfo(seqnames=preSf[,1],seqlengths=preSf[,2],
        isCircular=rep(FALSE,nrow(preSf)),genome=sfg)
    
    if (length(sf) > 0) {
        if (length(seqlevels(ann)) != length(seqlevels(sf)))
            # If a subset, this is enough
            seqinfo(ann) <- sf[intersect(seqlevels(ann),seqlevels(sf))]
        else if (!all(seqlevels(ann) == seqlevels(sf))) {
            # Must also be sorted in the same way
            seqlevels(ann) <- seqlevels(sf)
            seqinfo(ann) <- sf
        }
        else
            seqinfo(ann) <- sf
    }
    
    ann <- .nameAnnotationFromMetaType(ann,metaType)
    
    preActive <- NULL
    if (!is.null(querySet$active)) {
        preActive <- dbGetQuery(con,querySet$active)
        preActive$`_id` <- NULL
        preActive$content_id <- NULL
        active <- as.integer(preActive$length)
        names(active) <- as.character(preActive$name)
        # FIXME: Will have to take care later of this
        #active <- active[names(ann)]
        attr(ann,"activeLength") <- active
    }
    
    return(ann)
}

.loadAnnotationOnTheFly <- function(genome,refdb,level,type,rc=NULL) {
    message("Retrieving genome information for ",genome," from ",refdb)
    sf <- getSeqInfo(genome,asSeqinfo=TRUE)
    
    switch(level,
        gene = {
            switch(type,
                gene = {
                    message("Retrieving latest gene annotation for ",genome,
                        " from ",refdb)
                    ann <- getAnnotation(genome,"gene",refdb=refdb,rc=rc)
                    annGr <- makeGRangesFromDataFrame(
                        df=ann,
                        seqinfo=sf,
                        keep.extra.columns=TRUE,
                        seqnames.field="chromosome"
                    )
                    names(annGr) <- as.character(annGr$gene_id)
                },
                exon = {
                    message("Retrieving latest exon annotation for ",genome,
                        " from ",refdb)
                    ann <- getAnnotation(genome,"exon",refdb=refdb,rc=rc)
                    tmpGr <- makeGRangesFromDataFrame(
                        df=ann,
                        seqinfo=sf,
                        keep.extra.columns=TRUE,
                        seqnames.field="chromosome"
                    )
                    message("Merging exons for ",genome," from ",refdb)
                    annList <- reduceExons(tmpGr)
                    annGr <- annList$model
                    names(annGr) <- as.character(annGr$exon_id)
                    activeLength <- annList$length
                    names(activeLength) <- unique(annGr$gene_id)
                    attr(annGr,"activeLength") <- activeLength
                },
                utr = {
                    message("Retrieving latest 3' UTR annotation for ",genome,
                        " from ",refdb)
                    ann <- getAnnotation(genome,"utr",refdb=refdb,rc=rc)
                    tmpGr <- makeGRangesFromDataFrame(
                        df=ann,
                        seqinfo=sf,
                        keep.extra.columns=TRUE,
                        seqnames.field="chromosome"
                    )
                    message("Merging 3' UTRs for ",genome," from ",refdb)
                    #annGr <- reduceTranscripts(tmpGr)
                    #names(annGr) <- as.character(annGr$transcript_id)
                    annList <- reduceExons(tmpGr)
                    annGr <- annList$model
                    names(annGr) <- as.character(annGr$transcript_id)
                    activeLength <- annList$length
                    names(activeLength) <- unique(annGr$gene_id)
                    attr(annGr,"activeLength") <- activeLength
                }
            )
        },
        transcript = {
            switch(type,
                gene = {
                    message("Retrieving latest transcript annotation for ",
                        genome," from ",refdb)
                    ann <- getAnnotation(genome,"transcript",refdb=refdb,rc=rc)
                    annGr <- makeGRangesFromDataFrame(
                        df=ann,
                        seqinfo=sf,
                        keep.extra.columns=TRUE,
                        seqnames.field="chromosome"
                    )
                    names(annGr) <- as.character(annGr$transcript_id)
                },
                exon = {
                    # Stub
                    # TODO: "summarized_exon_by_transcript.rda"
                },
                utr = {
                    message("Retrieving latest 3' UTR annotation for ",genome,
                        " from ",refdb)
                    ann <- getAnnotation(genome,"utr",refdb=refdb,rc=rc)
                    annGr <- makeGRangesFromDataFrame(
                        df=ann,
                        seqinfo=sf,
                        keep.extra.columns=TRUE,
                        seqnames.field="chromosome"
                    )
                    message("Merging 3' UTRs for ",genome," from ",refdb)
                    #annGr <- reduceTranscriptsUtr(annGr)
                    #names(annGr) <- as.character(annGr$transcript_id)
                    annList <- reduceTranscriptsUtr(tmpGr)
                    annGr <- annList$model
                    names(annGr) <- as.character(annGr$transcript_id)
                    activeLength <- annList$length
                    names(activeLength) <- unique(annGr$transcript_id)
                    attr(annGr,"activeLength") <- activeLength
                }
            )
        },
        exon = {
            switch(type,
                exon = {
                    message("Retrieving latest exon annotation for ",genome,
                        " from ",refdb)
                    ann <- getAnnotation(genome,"exon",refdb=refdb,rc=rc)
                    annGr <- makeGRangesFromDataFrame(
                        df=ann,
                        seqinfo=sf,
                        keep.extra.columns=TRUE,
                        seqnames.field="chromosome"
                    )
                    names(annGr) <- as.character(annGr$exon_id)
                }
            )
        }
    )
    
    return(annGr)
}

.nameAnnotationFromMetaType <- function(ann,type) {
    switch(type,
        gene = {
            names(ann) <- as.character(ann$gene_id)
        },
        summarized_exon = {
            names(ann) <- as.character(ann$exon_id)
        },
        exon = {
            names(ann) <- as.character(ann$exon_id)
        },
        summarized_3utr = {
            names(ann) <- as.character(ann$transcript_id)
        },
        utr = {
            names(ann) <- as.character(ann$transcript_id)
        },
        summarized_transcript = {
            names(ann) <- as.character(ann$transcript_id)
        },
        transcript = {
            names(ann) <- as.character(ann$transcript_id)
        },
        summarized_3utr_transcript = {
            names(ann) <- as.character(ann$transcript_id)
        }
    )
    return(ann)
}

.annotationTypeFromInputArgs <- function(level,type,summarized=FALSE) {
    switch(level,
        gene = {
            switch(type,
                gene = {
                    return("gene")
                },
                exon = {
                    if (summarized)
                        return("summarized_exon")
                    else
                        return("exon")
                },
                utr = {
                    if (summarized)
                        return("summarized_3utr")
                    else
                        return("utr")
                }
            )
        },
        transcript = {
            switch(type,
                gene = {
                    if (summarized)
                        return("summarized_transcript")
                    else
                        return("transcript")
                },
                exon = {
                    if (summarized)
                        return("summarized_transcript_exon")
                    else
                        return("transexon")
                },
                utr = {
                    if (summarized)
                        return("summarized_3utr_transcript")
                    else
                        return("utr")
                }
            )
        },
        exon = {
            switch(type,
                gene = {
                    if (summarized)
                        return("summarized_exon")
                    else
                        return("exon")
                },
                exon = {
                    if (summarized)
                        return("summarized_exon")
                    else
                        return("exon")
                },
                utr = {
                    return("utr")
                }
            )
        }
    )
}

importCustomAnnotation <- function(gtfFile,metadata,
    level=c("gene","transcript","exon"),type=c("gene","exon","utr")) {
    level <- tolower(level[1])
    type <- tolower(type[1])
    # Check metadata
    if (is.null(metadata$organism)) {
        tmpOrg <- paste("species",format(Sys.Date(),"%Y%m%d"),sep="_")
        warning("An organism name must be provided with metadata for ",
            "reporting purposes! Using ",tmpOrg,immediate.=TRUE)
        metadata$organism <- tmpOrg
    }

    if (is.null(metadata$source)) {
        tmpSource <- paste("source",format(Sys.Date(),"%Y%m%d"),sep="_")
        warning("A source should be provided with metadata for reporting ",
            "purposes ! Using ",tmpSource,immediate.=TRUE)
        metadata$source <- tmpSource
    }
    if (is.null(metadata$version)) {
        tmpVer <- paste("version",format(Sys.Date(),"%Y%m%d"),sep="_")
        warning("A version should be provided with metadata for reporting ",
            "purposes! Using ",tmpVer,immediate.=TRUE)
        metadata$version <- tmpVer
    }
    if (is.null(metadata$chromInfo)) {
        warning("Chromosomal lengths should be provided with metadata! ",
            "Only chromosome names will be available... ",immediate.=TRUE)
        metadata$chromInfo <- NULL
    }
    else {
        str <- metadata$chromInfo
        if (is.character(str) && file.exists(str)) {
            out <- tryCatch(open(Rsamtools::BamFile(str)),error=function(e) e)
            if (inherits(out,"error")) # Not a BAM file, try to read.delim
                metadata$chromInfo <- read.delim(str,row.names=1)
            else
                metadata$chromInfo <- .chromInfoFromBAM(str)
        }
        if (!is.data.frame(metadata$chromInfo))
            stop("metadata$chromInfo must be a data frame!")
    }
    
    # For display meta information
    s <- metadata$source
    o <- metadata$organism
    v <- metadata$version
    
    # Parse the GTF file... If something wrong, it will crash here
    parsed <- parseCustomGtf(gtfFile)
    
    switch(level,
        gene = {
            switch(type,
                gene = {
                    message("Retrieving gene annotation for ",o," from ",s,
                        " version ",v," from ",gtfFile)
                    annGr <- annotationFromCustomGtf(parsed,level="gene",
                        type="gene")
                },
                exon = {
                    message("Retrieving summarized exon annotation for ",o,
                        " from ",s," version ",v," from ",gtfFile)
                    annGr <- annotationFromCustomGtf(parsed,level="gene",
                        type="exon",summarized=TRUE)
                },
                utr = {
                    message("Retrieving summarized 3' UTR annotation per gene ",
                        "for ",o," from ",s," version ",v," from ",gtfFile)
                    annGr <- annotationFromCustomGtf(parsed,level="gene",
                        type="utr",summarized=TRUE)
                }
            )
        },
        transcript = {
            switch(type,
                gene = {
                    message("Retrieving transcript annotation for ",o," from ",
                        s," version ",v," from ",gtfFile)
                    annGr <- annotationFromCustomGtf(parsed,level="transcript",
                        type="gene")
                },
                exon = {
                    # Stub
                    # TODO: "summarized_exon_by_transcript.rda"
                },
                utr = {
                    message("Retrieving summarized 3' UTR annotation per ",
                        "transcript for ",o," from ",s," version ",v," from ",
                        gtfFile)
                    annGr <- annotationFromCustomGtf(parsed,level="transcript",
                        type="utr",summarized=TRUE)
                }
            )
        },
        exon = {
            switch(type,
                exon = {
                    message("Retrieving exon annotation for ",o," from ",s,
                        " version ",v," from ",gtfFile)
                    annGr <- annotationFromCustomGtf(parsed,level="exon",
                        type="exon")
                }
            )
        }
    )
    
    return(annGr)
}

.validateDbCon <- function(obj) {
    # obj can be an already opened connection or the db file or missing. In the
    # latter case, the function looks at the package filesystem location
    if (is.null(obj)) {
        db <- file.path(system.file(package="metaseqR"),"annotation.sqlite")
        drv <- dbDriver("SQLite")
        con <- tryCatch(dbConnect(drv,dbname=db),error=function(e) {
            message("Caught error: ",e)
            stop("Have you constructed the metaseqR annotation database?")
        },finally="")
    }
    if (file.exists(obj)) {
        drv <- dbDriver("SQLite")
        con <- tryCatch(dbConnect(drv,dbname=obj),error=function(e) {
            message("Caught error: ",e)
            stop("Is obj an SQLite database?")
        },finally="")
    }
    else if (is(obj,"SQLiteConnection"))
        con <- obj
    return(con)
}

getInstalledAnnotations <- function(obj=NULL) {
    con <- .validateDbCon(obj)
    content <- .browseContent(con)
    dbDisconnect(con)   
    return(content[,-1])
}

getUserAnnotations <- function(obj=NULL) {
    con <- .validateDbCon(obj)
    userContent <- .browseUserContent(con)
    dbDisconnect(con)
    return(userContent[,-1])
}

correctTranscripts <- function(ann) {
    rownames(ann) <- paste("T",seq_len(nrow(ann)),sep="_")
    len <- ann[,3] - ann[,2]
    len <- len[-which(is.na(len))]
    len[len==0] <- 1
    defUtrLen <- round(2^mean(log2(len)))
    nas <- which(is.na(ann$start))
    annNa <- ann[nas,]
    annNa$start <- annNa$tstart
    annNa$end <- annNa$tend
    tmp <- makeGRangesFromDataFrame(df=annNa)
    tmp <- flank(resize(tmp,width=1,fix="end"),start=FALSE,width=defUtrLen)
    ann[names(tmp),"start"] <- start(tmp)
    ann[names(tmp),"end"] <- end(tmp)
    return(ann)
}

reduceTranscripts <- function(gr) {
    # Get the gene ids to use as split factor
    gene <- unique(as.character(gr$gene_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
    
    # There will be a transcript_id
    ir <- which(names(meta) == "transcript_id")
    map <- meta[,-ir]
    
    # Remove duplicates and name the map
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,]
    # In some cases, transcripts are recorded with multiple biotypes...
    if (length(map$gene_id) != length(unique(map$gene_id))) {
        mapp <- map
        ir <- which(names(mapp) == "biotype")
        mapp <- mapp[,-ir]
        d <- which(duplicated(mapp))
        if (length(d) > 0)
            map <- map[-d,]
    }
    rownames(map) <- map$gene_id
    map <- map[gene,]
    
    # Gene names and biotypes for later reconstruction
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$gene_id)
    # Re-order for common reference with the map
    grList <- grList[gene]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    grNew <- unname(unlist(grList))
    gene_id <- rep(gene,lens)
    transcript_id <- paste(gene_id,"MET",inds,sep="_")
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        transcript_id=transcript_id,
        gene_id=gene_id,
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$gene_id)
    tmp <- tmp[gene]
    len <- vapply(width(tmp),sum,integer(1))
    
    #return(grNew)
    return(list(model=grNew,length=len))
}

reduceTranscriptsUtr <- function(gr) {
    # Get the transcript ids to use ordering element
    trans <- unique(as.character(gr$transcript_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
    
    # Make the map and remove duplicates
    map <- meta
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,,drop=FALSE]
    rownames(map) <- map$transcript_id
    map <- map[trans,]
    
    # Gene ids, names and biotypes for later reconstruction
    gi <- as.character(map$gene_id)
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$transcript_id)
    # Re-order for common reference with the map
    grList <- grList[trans]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    grNew <- unname(unlist(grList))
    transcript_id <- paste(rep(trans,lens),"MEU",inds,sep="_")
    gene_id <- rep(gi,lens)
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        transcript_id=transcript_id,
        #gene_id=gene_id,
        gene_id=rep(trans,lens),
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$gene_id)
    tmp <- tmp[trans]
    len <- vapply(width(tmp),sum,integer(1))
    
    #return(grNew)
    return(list(model=grNew,length=len))
}

reduceExons <- function(gr) {
    # Get the gene ids to use as split factor
    gene <- unique(as.character(gr$gene_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
    
    # There will be an exon_id
    ir <- which(names(meta) == "exon_id")
    map <- meta[,-ir]
    
    # Remove duplicates and name the map
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,]
    rownames(map) <- map$gene_id
    map <- map[gene,]
    
    # Gene names and biotypes for later reconstruction
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$gene_id)
    # Re-order for common reference with the map
    grList <- grList[gene]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    grNew <- unname(unlist(grList))
    gene_id <- rep(gene,lens)
    exon_id <- paste(gene_id,"MEX",inds,sep="_")
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        exon_id=exon_id,
        gene_id=gene_id,
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$gene_id)
    tmp <- tmp[gene]
    len <- vapply(width(tmp),sum,integer(1))
    
    return(list(model=grNew,length=len))
}

reduceTranscriptsExons <- function(gr) {
    # Get the transcript ids to use ordering element
    trans <- unique(as.character(gr$transcript_id))
    
    # Get the GRanges metadata to create a map for later
    meta <- elementMetadata(gr)
    if (is.null(gr$gene_name))
        meta$gene_name <- meta$gene_id
    if (is.null(gr$biotype))
        meta$biotype <- rep("gene",nrow(meta))
        
    # There will be an exon_id
    ir <- which(names(meta) == "exon_id")
    map <- meta[,-ir]
    
    # Make the map and remove duplicates
    d <- which(duplicated(map))
    if (length(d) > 0)
        map <- map[-d,,drop=FALSE]
    rownames(map) <- map$transcript_id
    map <- map[trans,]
    
    # Gene ids, names and biotypes for later reconstruction
    #gi <- as.character(map$gene_id)
    gn <- as.character(map$gene_name)
    bt <- as.character(map$biotype)
    
    # Split the initial GRanges to apply rest operations
    grList <- split(gr,gr$transcript_id)
    # Re-order for common reference with the map
    grList <- grList[trans]
    # Now, reduce to merge overlaping exons
    grList <- reduce(grList)
    
    # Start the reconstruction by getting new lengths and create names
    lens <- lengths(grList)
    inds <- unlist(lapply(lens,function(j) return(seq_len(j))),use.names=FALSE)
    grNew <- unname(unlist(grList))
    exon_id <- paste(rep(trans,lens),"MTE",inds,sep="_")
    #transcript_id <- rep(gi,lens)
    #gene_id <- rep(gi,lens)
    gene_name <- rep(gn,lens)
    biotype <- rep(bt,lens)
    newMeta <- DataFrame(
        exon_id=exon_id,
        #gene_id=gene_id,
        transcript_id=rep(trans,lens),
        gene_name=gene_name,
        biotype=biotype
    )
    mcols(grNew) <- newMeta
    
    # grNew is the GRanges to return. In order to get the activeLength, we split
    # again per gene_id in a temp variable
    tmp <- split(grNew,grNew$transcript_id)
    tmp <- tmp[trans]
    len <- vapply(width(tmp),sum,integer(1))
    
    #return(grNew)
    return(list(model=grNew,length=len))
}

getAnnotation <- function(org,type,refdb="ensembl",ver=NULL,rc=NULL) {
    org <- tolower(org)
    switch(refdb,
        ensembl = { return(getEnsemblAnnotation(org,type,ver)) },
        ucsc = { return(getUcscAnnotation(org,type,refdb,rc=rc)) },
        refseq = { return(getUcscAnnotation(org,type,refdb,rc=rc)) }
    )
}

getEnsemblAnnotation <- function(org,type,ver=NULL) {
    if (org=="tair10")
        dat <- "plants_mart"
    else
        dat <- "ENSEMBL_MART_ENSEMBL"
    host <- getHost(org,ver)
    message("Using Ensembl host ",host)
    mart <- useMart(biomart=dat,host=host,dataset=getDataset(org))

    chrsExp <- paste("^",getValidChrs(org),"$",sep="",collapse="|")
    if (type=="gene") {
        bm <- tryCatch(
            getBM(attributes=.getGeneAttributes(org),mart=mart),
            error=function(e) {
                message("Caught error: ",e)
                message("This error is most probably related to httr package ",
                    "internals! Using fallback solution...")
                .myGetBM(attributes=.getGeneAttributes(org),mart=mart)
            },
            finally=""
        )
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$start_position,
            end=bm$end_position,
            gene_id=bm$ensembl_gene_id,
            gc_content=if (org %in% 
                c("hg18","hg19","mm9","rn5","dm3","danrer7")) 
                bm$percentage_gc_content else bm$percentage_gene_gc_content,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) bm$external_gene_id 
                else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        rownames(ann) <- ann$gene_id
    }
    else if (type=="transcript") {
        bm <- tryCatch(
            getBM(attributes=.getTranscriptAttributes(org),mart=mart),
            error=function(e) {
                message("Caught error: ",e)
                message("This error is most probably related to httr package ",
                    "internals! Using fallback solution...")
                .myGetBM(attributes=.getTranscriptAttributes(org),mart=mart)
            },
            finally=""
        )
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$transcript_start,
            end=bm$transcript_end,
            transcript_id=bm$ensembl_transcript_id,
            gene_id=bm$ensembl_gene_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        rownames(ann) <- as.character(ann$transcript_id)
    }
    else if (type=="utr") {
        bm <- tryCatch(
            getBM(attributes=.getTranscriptUtrAttributes(org),mart=mart),
            error=function(e) {
                message("Caught error: ",e)
                message("This error is most probably related to httr package ",
                    "internals! Using fallback solution...")
                .myGetBM(attributes=.getTranscriptUtrAttributes(org),mart=mart)
            },
            finally=""
        )
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$`3_utr_start`,
            end=bm$`3_utr_end`,
            tstart=bm$transcript_start,
            tend=bm$transcript_end,
            transcript_id=bm$ensembl_transcript_id,
            gene_id=bm$ensembl_gene_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9","tair10")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        ann <- correctTranscripts(ann)
        ann <- ann[,c("chromosome","start","end","transcript_id","gene_id",
            "strand","gene_name","biotype")]
    }
    else if (type=="exon") {
        bm <- tryCatch(
            getBM(attributes=.getExonAttributes(org),mart=mart),
            error=function(e) {
                message("Caught error: ",e)
                message("This error is most probably related to httr package ",
                    "internals! Using fallback solution...")
                .myGetBM(attributes=.getExonAttributes(org),mart=mart)
            },
            finally=""
        )
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$exon_chrom_start,
            end=bm$exon_chrom_end,
            exon_id=bm$ensembl_exon_id,
            gene_id=bm$ensembl_gene_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
        rownames(ann) <- ann$exon_id
    }
    else if (type=="transexon") {
        bm <- tryCatch(
            getBM(attributes=.getTranscriptExonAttributes(org),mart=mart),
            error=function(e) {
                message("Caught error: ",e)
                message("This error is most probably related to httr package ",
                    "internals! Using fallback solution...")
                .myGetBM(attributes=.getTranscriptExonAttributes(org),mart=mart)
            },
            finally=""
        )
        ann <- data.frame(
            chromosome=paste("chr",bm$chromosome_name,sep=""),
            start=bm$exon_chrom_start,
            end=bm$exon_chrom_end,
            exon_id=bm$ensembl_exon_id,
            transcript_id=bm$ensembl_transcript_id,
            #gene_id=bm$ensembl_gene_id,
            strand=ifelse(bm$strand==1,"+","-"),
            gene_name=if (org %in% c("hg18","hg19","mm9")) 
                bm$external_gene_id else bm$external_gene_name,
            biotype=bm$gene_biotype
        )
    }
    ann <- ann[order(ann$chromosome,ann$start),]
    ann <- ann[grep(chrsExp,ann$chromosome),]
    ann$chromosome <- as.character(ann$chromosome)
    
    return(ann)
}

getUcscAnnotation <- function(org,type,refdb="ucsc",chunkSize=500,rc=NULL) {
    if (!requireNamespace("RMySQL")) {
        rmysqlPresent <- FALSE
        warning("R package RMySQL is not present! Annotation will be ",
            "retrieved by downloading temporary files from UCSC and the usage
            of a temporary SQLite database...",immediate.=TRUE)
    }
    else
        rmysqlPresent <- TRUE
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to use annotation from UCSC!")
    
    if (org=="tair10") {
        warnwrap("Arabidopsis thaliana genome is not supported by UCSC Genome ",
            "Browser database! Switching to Ensembl...")
        return(getEnsemblAnnotation("tair10",type))
    }
    
    validChrs <- getValidChrs(org)
    chrsExp <- paste("^",paste(validChrs,collapse="$|^"),"$",sep="")

    dbOrg <- getUcscOrganism(org)
    if (rmysqlPresent) {
        # The UTR case is handled later
        if (type != "utr") {
            dbCreds <- .getUcscCredentials()
            drv <- dbDriver("MySQL")
            con <- dbConnect(drv,user=dbCreds[2],password=NULL,dbname=dbOrg,
                host=dbCreds[1])
            type2 <- ifelse(type=="transexon","exon",type)
            query <- getUcscQuery(org,type2,refdb)
            rawAnn <- dbGetQuery(con,query)
            dbDisconnect(con)
        }
    }
    else {
        # This should return the same data frame as the db query
        #if (type == "transcript")
        #    tmpSqlite <- getUcscDbl(org,"gene",refdb)
        #else if (type %in% c("gene","exon"))
        #    tmpSqlite <- getUcscDbl(org,type,refdb)
        
        tmpSqlite <- getUcscDbl(org,refdb)
        
        if (type != "utr") { # Otherwise diect download is used
            drv <- dbDriver("SQLite")
            con <- dbConnect(drv,dbname=tmpSqlite)
            query <- getUcscQuery(org,type,refdb)
            rawAnn <- dbGetQuery(con,query)
            dbDisconnect(con)
        }
    }
    if (type=="gene") {
        ann <- rawAnn
        ann <- ann[grep(chrsExp,ann$chromosome,perl=TRUE),]
        ann$chromosome <- as.character(ann$chromosome)
        rownames(ann) <- ann$gene_id
        #rownames(ann) <- ann$transcript_id
        #tmpAnn <- rawAnn
        #tmpAnn <- tmpAnn[grep(chrsExp,tmpAnn$chromosome,perl=TRUE),]
        #tmpAnn$chromosome <- as.character(tmpAnn$chromosome)
        #rownames(tmpAnn) <- tmpAnn$transcript_id
        ## Split the UCSC transcripts per gene name
        ##tmp <- split(tmpAnn,tmpAnn$gene_name)
        # Retrieve the longest transcript (as per Ensembl convention)
        #ann <- do.call("rbind",cmclapply(tmp,function(x) {
        #    size <- x$end - x$start
        #    selected <- which(size == max(size))
        #    return(x[selected[1],,drop=FALSE])
        #},rc=rc))
        #names(ann)[4] <- "gene_id"
        gcContent <- getGcContent(ann,org)
        ann$gc_content <- gcContent
        # rownames are leftover from splitting above
        rownames(ann) <- ann$gene_id
    }
    else if (type=="transcript") {
        ann <- rawAnn
        ann <- ann[grep(chrsExp,ann$chromosome,perl=TRUE),]
        d <- which(duplicated(ann))
        if (length(d) > 0)
            ann <- ann[-d,]
        ann$chromosome <- as.character(ann$chromosome)
        ann$transcript_id <- as.character(ann$transcript_id)
        ann$gene_id <- as.character(ann$transcript_id)
        # There are still duplicated transcript ids
        d <- which(duplicated(ann$transcript_id))
        iter <- 1
        while(length(d) > 0) {
            ann$transcript_id[d] <- paste(ann$transcript_id[d],iter,sep="_")
            iter <- iter + 1
            d <- which(duplicated(ann$transcript_id))
        }
        rownames(ann) <- ann$transcript_id
        ann <- ann[,c(1,2,3,4,8,5:7)]
    }
    else if (type=="exon" || type=="transexon") {
        rawAnn <- rawAnn[grep(chrsExp,rawAnn$chromosome,perl=TRUE),]
        exList <- cmclapply(as.list(seq_len(nrow(rawAnn))),function(x,d) {
            r <- d[x,,drop=FALSE]
            starts <- as.numeric(strsplit(r[,"start"],",")[[1]])
            ends <- as.numeric(strsplit(r[,"end"],",")[[1]])
            nexons <- length(starts)
            ret <- data.frame(
                rep(r[,"chromosome"],nexons),
                starts,ends,
                paste(r[,"exon_id"],"_e",seq_len(nexons),sep=""),
                rep(r[,"strand"],nexons),
                rep(r[,"gene_id"],nexons),
                rep(r[,"gene_name"],nexons),
                rep(r[,"biotype"],nexons)
            )
            names(ret) <- names(r)
            rownames(ret) <- ret$exon_id
            return(ret)
        },rawAnn,rc=rc)
        
        # For some reason rbind takes ages for large datasets... We have to 
        # split in chunks of 1000
        N <- length(exList)
        mo <- N%%chunkSize
        if (mo == 0) {
            fl <- N/chunkSize
            fac <- factor(rep(seq_len(fl),each=chunkSize))
        }
        else {
            fl <- (N-mo)/chunkSize
            fac <- factor(c(rep(seq_len(fl),each=chunkSize),rep(fl,mo)))
        }
        exChunkedList <- split(exList,fac)
        # Merge the chunks
        tmp <- cmclapply(names(exChunkedList),function(n,X) {
            message("Binding chunk ",n,"...")
            return(do.call("rbind",X[[n]]))
        },exChunkedList,rc=rc)
        # Final merge
        message("Binding all chunks...")
        tmpAnn <- do.call("rbind",tmp)
        
        ann <- data.frame(
            chromosome=as.character(tmpAnn$chromosome),
            start=tmpAnn$start,
            end=tmpAnn$end,
            exon_id=as.character(tmpAnn$exon_id),
            gene_id=as.character(tmpAnn$gene_id),
            strand=as.character(tmpAnn$strand),
            gene_name=as.character(tmpAnn$gene_name),
            biotype=as.character(tmpAnn$biotype)
        )
        rownames(ann) <- ann$exon_id
        
        if (type == "transexon")
            names(ann)[5] <- "transcript_id"
    }
    else if (type=="utr") {
        # We are already supposed to have the necessary elements to construct
        # the required data frame. Should be something like
        # 1. Read the GTF as TxDb
        # 2. Import the GTF with rtracklayer
        # 3. Call the 3UTR function of TxDb on (1)
        # 4. Construct a map to add gene_name and other things (essentially
        #    the biotype if we make it, if not we have to connect it from the
        #    respective gene annotation during the pipeline execution)
        # 5. Add the mapped elements to the GRanges/data.frame
        # 6. Return a data.frame
        #
        # All the process is streamlined in getUcscUtr function
        
        preAnn <- getUcscUtr(org,refdb)
        preAnn <- as.data.frame(preAnn)
        preAnn <- preAnn[,c(1,2,3,6,7,5,8,9)]
        preAnn <- preAnn[grep(chrsExp,preAnn$seqnames,perl=TRUE),]
        names(preAnn) <- c("chromosome","start","end","transcript_id","gene_id",
            "strand","gene_name","biotype")
        # preAnn now has exon names as rownames... OK...
        #rownames(preAnn) <- paste("T",seq_len(nrow(preAnn)),sep="_")
        ann <- preAnn
    }
    
    ann <- ann[order(ann$chromosome,ann$start),]
    return(ann)
}

getUcscUtr <- function(org,refdb="ucsc") {
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    checkTextArgs("refdb",refdb,getSupportedUcscDbs())
    
    if (!requireNamespace("GenomicFeatures"))
        stop("Bioconductor package GenomicFeatures is required!")
    
    httpBase <- paste("http://hgdownload.soe.ucsc.edu/goldenPath/",
        getUcscOrganism(org),"/database/",sep="")
    tableUtr <- getUcscTableNameUtr(org,refdb) # Need one table
    message("  Retrieving table ",tableUtr," for 3'UTR annotation generation ",
        "from ",refdb," for ",org)
    download.file(paste(httpBase,tableUtr,sep=""),file.path(tempdir(),
        paste(tableUtr,".txt.gz",sep="")),quiet=TRUE)
    
    # Do the conversion stuff. AS there is no easy way to check if genePredToGtf
    # exists in the system, we should download it on the fly (once for the 
    # session). If no Linux machine, then problem.
    genePredToGtfEnv <- Sys.getenv("GENEPREDTOGTF_BINARY")
    if (genePredToGtfEnv == "") {
        genePredToGtf <- file.path(customDir,"genePredToGtf")
    } else {
        genePredToGtf <- file.path(genePredToGtfEnv)
    }
    if (!file.exists(genePredToGtf)) {
        message("  Retrieving genePredToGtf tool")
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
            genePredToGtf,quiet=TRUE
        )
        system(paste("chmod 775",genePredToGtf))
    }
    
    # Then do the conversion... No need for windows case as Kent tools do not
    # work in Windows anyway
    gtfFile <- file.path(tempdir(),paste(tableUtr,".gtf",sep=""))
    message("  Converting ",tableUtr," to GTF")
    tmpName <- file.path(tempdir(),paste(format(Sys.time(),"%Y%m%d%H%M%S"),
        "tgtf",sep="."))
    commandUcsc <- paste(
        "zcat ",file.path(tempdir(),paste(tableUtr,".txt.gz",sep=""))," | ",
        "cut -f1-10 - | ",genePredToGtf," file stdin ",tmpName," -source=",
        tableUtr," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",gtfFile,
        sep=""
    )
    commandRefSeq <- paste(
        "zcat ",file.path(tempdir(),paste(tableUtr,".txt.gz",sep=""))," | ",
        "cut -f2- | ",genePredToGtf," file stdin ",tmpName," -source=",
        tableUtr," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",gtfFile,
        sep=""
    )
    command <- commandRefSeq
    # There is an exception for organisms that do not exist in UCSC databases
    # so we must use RefSeq
    ucscUnsup <- c("rn5","rn6","dm3","dm6","danrer7","danrer10","danrer11",
        "pantro4","pantro5","susscr3","susscr11","equcab3")
    if (refdb == "ucsc" && !(org %in% ucscUnsup))
        command <- commandUcsc
    
    # Run the command and process the data
    message("Executing: ",command)
    system(command)
    parsed <- parseCustomGtf(gtfFile)
    return(.makeGeneUtrFromTxDb(parsed$txdb,parsed$map,asdf=FALSE))
}

getGcContent <- function(ann,org) {
    if (missing(ann))
        stop("A valid annotation data frame must be provided in order to ",
            "retrieve GC-content.")
    org <- tolower(org[1])
    checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    # Convert annotation to GRanges
    message("Converting annotation to GenomicRanges object...")
    annGr <- tryCatch(GRanges(ann),error=function(e) {
        if (packageVersion("GenomicRanges")<1.14)
            return(GRanges(
                seqnames=Rle(ann[,1]),
                ranges=IRanges(start=ann[,2],end=ann[,3]),
                strand=Rle(ann[,6]),
                name=as.character(ann[,4])
            ))
        else
            return(makeGRangesFromDataFrame(
                df=ann,
                keep.extra.columns=TRUE,
                seqnames.field="chromosome"
            ))
    },finaly="")
    
    bsg <- loadBsGenome(org)
    if (is(bsg,"BSgenome")) {
        message("Getting DNA sequences...")
        seqs <- tryCatch(getSeq(bsg,names=annGr),error=function(e) {
            warning("Caught error ",e,immediate.=TRUE)
            message("Cannot get ",org," sequences! Returning NA GC content...")
            return(NA)
        },finally="")
        if (!is.na(seqs[1])) {
            message("Getting GC content...")
            freqMatrix <- alphabetFrequency(seqs,as.prob=TRUE,baseOnly=TRUE)
            gcContent <- apply(freqMatrix,1,function(x) round(100*sum(x[2:3]),
                digits=2))
        }
        else
            gcContent <- rep(NA,nrow(ann))
    }
    else
        gcContent <- rep(NA,nrow(ann))
    names(gcContent) <- as.character(ann[,4])
    return(gcContent)
}

getSeqInfo <- function(org,asSeqinfo=FALSE) {
    sf <- tryCatch(.chromInfoWrapperGID(getUcscOrganism(org)),
        error=function(e) {
            message("GenomeInfoDb chrom info fetch mechanism failed with the ",
                "following error: ")
            message(e)
            message("")
            message("Trying a direct download...")
            getChromInfo(getUcscOrganism(org))
        },finally="")
    rownames(sf) <- as.character(sf[,1])
    sf <- sf[getValidChrs(org),]
    if (asSeqinfo)
        return(Seqinfo(seqnames=sf[,1],seqlengths=sf[,2],
            isCircular=sf[,"circular"],genome=getUcscOrganism(org)))
    else
        return(data.frame(chromosome=sf[,1],length=as.integer(sf[,2])))
}

getUcscOrganism <- function(org) {
    switch(org,
        hg18 = { return("hg18") },
        hg19 = { return("hg19") },
        hg38 = { return("hg38") },
        mm9 = { return("mm9") },
        mm10 = { return("mm10") },
        mm39 = { return("mm39") },
        rn5 = { return("rn5") },
        rn6 = { return("rn6") },
        dm3 = { return("dm3") },
        dm6 = { return("dm6") },
        danrer7 = { return("danRer7") },
        danrer10 = { return("danRer10") },
        danrer11 = { return("danRer11") },
        pantro4 = { return("panTro4") },
        pantro5 = { return("panTro5") },
        susscr3 = { return("susScr3") },
        susscr11 = { return("susScr11") },
        equcab3 = { return("equCab3") },
        tair10 = { return("TAIR10") }
    )
}

getBsOrganism <- function(org) {
    switch(org,
        hg18 = {
            return("BSgenome.Hsapiens.UCSC.hg18")
        },
        hg19 = {
            return("BSgenome.Hsapiens.UCSC.hg19")
        },
        hg38 = {
            return("BSgenome.Hsapiens.UCSC.hg38")
        },
        mm9 = {
            return("BSgenome.Mmusculus.UCSC.mm9")
        },
        mm10 = {
            return("BSgenome.Mmusculus.UCSC.mm10")
        },
        mm39 = {
            return("BSgenome.Mmusculus.UCSC.mm39")
        },
        rn5 = {
            return("BSgenome.Rnorvegicus.UCSC.rn5")
        },
        rn6 = {
            return("BSgenome.Rnorvegicus.UCSC.rn6")
        },
        dm3 = {
            return("BSgenome.Dmelanogaster.UCSC.dm3")
        },
        dm6 = {
            return("BSgenome.Dmelanogaster.UCSC.dm6")
        },
        danrer7 = {
            return("BSgenome.Drerio.UCSC.danRer7")
        },
        danrer10 = {
            return("BSgenome.Drerio.UCSC.danRer10")
        },
        danrer11 = {
            warning("danRer11 is not supported by BSgenome! Please use Ensembl",
                " as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
            #return("BSgenome.Drerio.UCSC.danRer11")
        },
        pantro4 = {
            warning("panTro4 is not supported by BSgenome! Please use Ensembl ",
                "as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
        },
        pantro5 = {
            return("BSgenome.Ptroglodytes.UCSC.panTro5")
        },
        susscr3 = {
            return("BSgenome.Sscrofa.UCSC.susScr3")
        },
        susscr11 = {
            warning("susScr11 is not supported by BSgenome! Please use ",
                "Ensembl as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
        },
        equcab3 = {
            warning("equCab3 is not supported by BSgenome! Please use Ensembl ",
                "as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
        },
        tair10 = {
            warning("TAIR10 is not supported by BSgenome! Please use Ensembl ",
                "as annotation source if GC content is important.",
                immediate.=TRUE)
            return(NA)
        }
    )
}

loadBsGenome <- function(org) {
    if (!requireNamespace("BiocManager"))
        stop("The Bioconductor package BiocManager is required to ",
            "proceed!")
    if (!requireNamespace("BSgenome"))
        stop("The Bioconductor package BSgenome is required to proceed!")
    bsOrg <- getBsOrganism(org)
    if (!is.na(bsOrg)) {
        if (bsOrg %in% BSgenome::installed.genomes())
            bsObj <- BSgenome::getBSgenome(getUcscOrganism(org))
        else {
            BiocManager::install(bsOrg,update=FALSE,ask=FALSE)
            bsObj <- BSgenome::getBSgenome(getUcscOrganism(org))
        }
        return(bsObj)
    }
    else
        return(NA)
}

getChromInfo <- function(org,
    goldenPath="http://hgdownload.cse.ucsc.edu/goldenPath/") {
    download.file(paste(goldenPath,org,"/database/chromInfo.txt.gz",sep=""),
        file.path(tempdir(),"chromInfo.txt.gz"),quiet=TRUE)
    chromInfo <- read.delim(file.path(tempdir(),"chromInfo.txt.gz"),
        header=FALSE)
    chromInfo <- chromInfo[,1,2]
    chromInfo[,1] <- as.character(chromInfo[,1])
    chromInfo$V3 <- rep(FALSE,nrow(chromInfo))
    m <- grep("M",chromInfo[,1])
    if (length(m) > 0)
        chromInfo$V3[m] <- TRUE
    return(chromInfo)
}

getHost <- function(org,ver=NULL) {
    if (!requireNamespace("biomaRt"))
        stop("The Bioconductor package biomaRt is required to proceed!")
    
    org <- tolower(org[1])
    checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    if (!is.null(ver) 
        && (!is.numeric(ver) || is.na(suppressWarnings(as.numeric(ver)))))
        stop("ver must be numeric or coercible to numeric if not NULL!")
        
    if (org == "tair10")
        return("plants.ensembl.org")
    if (org == "hg19")
        return("grch37.ensembl.org")
    
    aver <- getUcscToEnsembl(org)
    if (!is.null(ver) && !(ver %in% aver)) {
        warning("Version ",ver," not available/existing for ",org,"! Will ",
            "use the latest available version...",immediate.=TRUE)
        ver <- NULL
    }
    
    if (is.null(ver)) {
        u2e <- ucscToEnsembl()
        vers <- u2e[[org]]
        ver <- vers[length(vers)]
    }
    
    ea <- biomaRt::listEnsemblArchives()
    i <- grep(as.character(ver),ea[,"version"])
    if (length(i) > 0) {
        if (ea[i,"current_release"] == "*")
            return("http://www.ensembl.org")
        else
            return(ea[i,"url"])
    }
    else {
        warning("Version ",ver," not found in biomaRt archives for ",org,"! ",
            "Will use the latest available version for ",org,"...",
            immediate.=TRUE)
        u2e <- ucscToEnsembl()
        vers <- u2e[[org]]
        for (v in rev(vers)) {
            newi <- grep(as.character(v),ea[,"version"])
            if (length(newi) > 0)
                return(ea[newi,"url"])
        }
    }
    # If everything fails...
    return(NULL)
}

getUcscToEnsembl <- function(org) {
    u2e <- ucscToEnsembl()
    return(u2e[[org]])
}

checkUcscToEnsembl <- function(org,ver) {
    u2e <- getUcscToEnsembl()
    return(ver %in% u2e[[org]])
}

ucscToEnsembl <- function() {
    return(list(
        hg18=54,
        hg19=75,
        hg38=110:115,
        mm9=54,
        mm10=102,
        mm39=110:115,
        rn5=77,
        rn6=110:115,
        dm3=77,
        dm6=110:115,
        danrer7=77,
        danrer10=80,
        danrer11=110:115,
        pantro4=80,
        pantro5=110:115,
        #pantro6=,
        susscr3=80,
        susscr11=110:115,
        equcab3=110:115
    ))
}

getDataset <- function(org) {
    switch(org,
        hg18 = { return("hsapiens_gene_ensembl") },
        hg19 = { return("hsapiens_gene_ensembl") },
        hg38 = { return("hsapiens_gene_ensembl") },
        mm9 = { return("mmusculus_gene_ensembl") },
        mm10 = { return("mmusculus_gene_ensembl") },
        mm39 = { return("mmusculus_gene_ensembl") },
        rn5 = { return("rnorvegicus_gene_ensembl") },
        rn6 = { return("rnorvegicus_gene_ensembl") },
        dm3 = { return("dmelanogaster_gene_ensembl") },
        dm6 = { return("dmelanogaster_gene_ensembl") },
        danrer7 = { return("drerio_gene_ensembl") },
        danrer10 = { return("drerio_gene_ensembl") },
        danrer11 = { return("drerio_gene_ensembl") },
        pantro4 = { return("ptroglodytes_gene_ensembl") },
        pantro5 = { return("ptroglodytes_gene_ensembl") },
        #pantro6 = { return("ptroglodytes_gene_ensembl") },
        susscr3 = { return("sscrofa_gene_ensembl") },
        susscr11 = { return("sscrofa_gene_ensembl") },
        equcab3 = { return("ecaballus_gene_ensembl") },
        tair10 = { return("athaliana_eg_gene") }
    )
}

getValidChrs <- function(org) {
    switch(org,
        hg18 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        hg19 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        hg38 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        mm9 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY"
            ))
        },
        mm10 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY"
            ))
        },
        mm39 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY"
            ))
        },
        rn5 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX"
            ))
        },
        rn6 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX"
            ))
        },
        dm3 = {
            return(c(
                "chr2L","chr2LHet","chr2R","chr2RHet","chr3L","chr3LHet",
                "chr3R","chr3RHet","chr4","chrU","chrUextra","chrX","chrXHet",
                "chrYHet"
            ))
        },
        dm6 = {
            return(c(
                "chr2L","chr2R","chr3L","chr3R","chr4","chrX"
            ))
        },
        danrer7 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        danrer10 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        danrer11 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        pantro4 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        pantro5 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        pantro6 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY"
            ))
        },
        susscr3 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr2","chr3","chr4","chr5","chr6","chr7",
                "chr8","chr9","chrX","chrY"
            ))
        },
        susscr11 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr2","chr3","chr4","chr5","chr6","chr7",
                "chr8","chr9","chrX","chrY"
            ))
        },
        equcab3 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr26","chr27","chr28","chr29","chr3","chr30",
                "chr31","chr4","chr5","chr6","chr7","chr8","chr9","chrX"#,"chrY"
            ))
        },
        tair10 = {
            return(c(
                "chr1","chr2","chr3","chr4","chr5"
            ))
        }
    )
}

getValidChrsWithMit <- function(org) {
    switch(org,
        hg18 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY","chrM"
            ))
        },
        hg19 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY","chrM"
            ))
        },
        hg38 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr3",
                "chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY","chrM"
            ))
        },
        mm9 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY","chrM"
            ))
        },
        mm10 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY","chrM"
            ))
        },
        mm39 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrY","chrM"
            ))
        },
        rn5 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrM"
            ))
        },
        rn6 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr3","chr4","chr5","chr6",
                "chr7","chr8","chr9","chrX","chrM"
            ))
        },
        dm3 = {
            return(c(
                "chr2L","chr2LHet","chr2R","chr2RHet","chr3L","chr3LHet",
                "chr3R","chr3RHet","chr4","chrU","chrUextra","chrX","chrXHet",
                "chrYHet"
            ))
        },
        dm6 = {
            return(c(
                "chr2L","chr2LHet","chr2R","chr2RHet","chr3L","chr3LHet",
                "chr3R","chr3RHet","chr4","chrU","chrUextra","chrX","chrXHet",
                "chrYHet"
            ))
        },
        danrer7 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        danrer10 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        danrer11 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr3","chr4","chr5","chr6","chr7","chr8","chr9"
            ))
        },
        pantro4 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY",
                "chrM"
            ))
        },
        pantro5 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY",
                "chrM"
            ))
        },
        pantro6 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr20","chr21","chr22","chr2A","chr2B",
                "chr3","chr4","chr5","chr6","chr7","chr8","chr9","chrX","chrY",
                "chrM"
            ))
        },
        susscr3 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr2","chr3","chr4","chr5","chr6","chr7",
                "chr8","chr9","chrX","chrY","chrM"
            ))
        },
        susscr11 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr2","chr3","chr4","chr5","chr6","chr7",
                "chr8","chr9","chrX","chrY","chrM"
            ))
        },
        equcab3 = {
            return(c(
                "chr1","chr10","chr11","chr12","chr13","chr14","chr15","chr16",
                "chr17","chr18","chr19","chr2","chr20","chr21","chr22","chr23",
                "chr24","chr25","chr26","chr27","chr28","chr29","chr3","chr30",
                "chr31","chr4","chr5","chr6","chr7","chr8","chr9","chrX",
                #"chrY",
                "chrM"
            ))
        },
        tair10 = {
            return(c(
                "chr1","chr2","chr3","chr4","chr5"
            ))
        }
    )
}

.getGeneAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "start_position",
            "end_position",
            "ensembl_gene_id",
            "percentage_gc_content",
            "strand",
            "external_gene_id",
            "gene_biotype"
        ))
    else if (org %in% c("rn5","danrer7","dm3"))
        return(c(
            "chromosome_name",
            "start_position",
            "end_position",
            "ensembl_gene_id",
            "percentage_gc_content",
            "strand",
            "external_gene_name",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "start_position",
            "end_position",
            "ensembl_gene_id",
            "percentage_gene_gc_content",
            "strand",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getTranscriptAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getTranscriptUtrAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "3_utr_start",
            "3_utr_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "transcript_start",
            "transcript_end",
            "3_utr_start",
            "3_utr_end",
            "ensembl_transcript_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getExonAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_gene_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

.getTranscriptExonAttributes <- function(org) {
    if (org %in% c("hg18","hg19","mm9"))
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_transcript_id",
            #"ensembl_gene_id",
            "external_gene_id",
            "gene_biotype"
        ))
    else
        return(c(
            "chromosome_name",
            "exon_chrom_start",
            "exon_chrom_end",
            "ensembl_exon_id",
            "strand",
            "ensembl_transcript_id",
            #"ensembl_gene_id",
            "external_gene_name",
            "gene_biotype"
        ))
}

getBiotypes <- function(org) {
    if (!(org %in% getSupportedOrganisms()))
        return(NULL)
    switch(org,
        hg18 = {
            return(c("unprocessed_pseudogene","pseudogene","miRNA",
                "retrotransposed","protein_coding","processed_pseudogene",
                "snRNA","snRNA_pseudogene","Mt_tRNA_pseudogene",
                "miRNA_pseudogene","misc_RNA","tRNA_pseudogene","snoRNA",
                "scRNA_pseudogene","rRNA_pseudogene","snoRNA_pseudogene","rRNA",
                "misc_RNA_pseudogene","IG_V_gene","IG_D_gene","IG_J_gene",
                "IG_C_gene","IG_pseudogene","scRNA"))
        },
        hg19 = {
            return(c("pseudogene","lincRNA","protein_coding","antisense",
                "processed_transcript","snRNA","sense_intronic","miRNA",
                "misc_RNA","snoRNA","rRNA","polymorphic_pseudogene",
                "sense_overlapping","3prime_overlapping_ncrna","TR_V_gene",
                "TR_V_pseudogene","TR_D_gene","TR_J_gene","TR_C_gene",
                "TR_J_pseudogene","IG_C_gene","IG_C_pseudogene","IG_J_gene",
                "IG_J_pseudogene","IG_D_gene","IG_V_gene","IG_V_pseudogene"))
        },
        hg38 = {
            return(c("protein_coding","polymorphic_pseudogene","lincRNA",
                "unprocessed_pseudogene","processed_pseudogene","antisense",
                "processed_transcript","transcribed_unprocessed_pseudogene",
                "sense_intronic","unitary_pseudogene","IG_V_gene",
                "IG_V_pseudogene","TR_V_gene","sense_overlapping",
                "transcribed_processed_pseudogene","miRNA","snRNA","misc_RNA",
                "rRNA","snoRNA","IG_J_pseudogene","IG_J_gene","IG_D_gene",
                "3prime_overlapping_ncrna","IG_C_gene","IG_C_pseudogene",
                "pseudogene","TR_V_pseudogene","Mt_tRNA","Mt_rRNA",
                "translated_processed_pseudogene","TR_J_gene","TR_C_gene",
                "TR_D_gene","TR_J_pseudogene","LRG_gene"))
        },
        mm9 = {
            return(c("pseudogene","snRNA","protein_coding","antisense","miRNA",
                "lincRNA","snoRNA","processed_transcript","misc_RNA","rRNA",
                "sense_overlapping","sense_intronic","polymorphic_pseudogene",
                "non_coding","3prime_overlapping_ncrna","IG_C_gene",
                "IG_J_gene","IG_D_gene","IG_V_gene","ncrna_host"))
        },
        mm10 = {
            return(c("pseudogene","snRNA","protein_coding","antisense","miRNA",
                "snoRNA","lincRNA","processed_transcript","misc_RNA","rRNA",
                "sense_intronic","sense_overlapping","polymorphic_pseudogene",
                "IG_C_gene","IG_J_gene","IG_D_gene","IG_LV_gene","IG_V_gene",
                "IG_V_pseudogene","TR_V_gene","TR_V_pseudogene",
                "3prime_overlapping_ncrna"))
        },
        mm39 = {
            return(c("IG_C_gene","IG_C_pseudogene","IG_D_gene",
                "IG_D_pseudogene","IG_J_gene","IG_LV_gene","IG_pseudogene",
                "IG_V_gene","IG_V_pseudogene","lncRNA","miRNA","misc_RNA",
                "Mt_rRNA","Mt_tRNA","processed_pseudogene","protein_coding",
                "pseudogene","ribozyme","rRNA","scaRNA","scRNA","snoRNA",
                "snRNA","sRNA","TEC","transcribed_processed_pseudogene",
                "transcribed_unitary_pseudogene",
                "transcribed_unprocessed_pseudogene",   
                "translated_unprocessed_pseudogene",
                "TR_C_gene","TR_D_gene","TR_J_gene","TR_J_pseudogene",
                "TR_V_gene","TR_V_pseudogene","unitary_pseudogene",
                "unprocessed_pseudogene"))
        },
        dm3 = {
            return(c("protein_coding","ncRNA","snoRNA","pre_miRNA","pseudogene",
                "snRNA","tRNA","rRNA"))
        },
        dm6 = {
            return(c("protein_coding","ncRNA","snoRNA","pre_miRNA","pseudogene",
                "snRNA","tRNA","rRNA"))
        },
        rn5 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","misc_RNA"))
        },
        rn6 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","misc_RNA"))
        },
        danrer7 = {
            return(c("antisense","protein_coding","miRNA","snoRNA","rRNA",
                "lincRNA","processed_transcript","snRNA","pseudogene",
                "sense_intronic","misc_RNA","polymorphic_pseudogene",
                "IG_V_pseudogene","IG_C_pseudogene","IG_J_pseudogene",
                "non_coding","sense_overlapping"
            ))
        },
        danrer10 = {
            return(c("antisense","protein_coding","miRNA","snoRNA","rRNA",
                "lincRNA","processed_transcript","snRNA","pseudogene",
                "sense_intronic","misc_RNA","polymorphic_pseudogene",
                "IG_V_pseudogene","IG_C_pseudogene","IG_J_pseudogene",
                "non_coding","sense_overlapping"
            ))
        },
        danrer11 = {
            return(c("antisense","protein_coding","miRNA","snoRNA","rRNA",
                "lincRNA","processed_transcript","snRNA","pseudogene",
                "sense_intronic","misc_RNA","polymorphic_pseudogene",
                "IG_V_pseudogene","IG_C_pseudogene","IG_J_pseudogene",
                "non_coding","sense_overlapping"
            ))
        },
        pantro4 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","snRNA","snoRNA","misc_RNA"))
        },
        pantro5 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","snRNA","snoRNA","misc_RNA"))
        },
        pantro6 = {
            return(c("protein_coding","pseudogene","processed_pseudogene",
                "miRNA","rRNA","snRNA","snoRNA","misc_RNA"))
        },
        susscr3 = {
            return(c("antisense","protein_coding","lincRNA","pseudogene",
                "processed_transcript","miRNA","rRNA","snRNA","snoRNA",
                "misc_RNA","non_coding","IG_C_gene","IG_J_gene",
                "IG_V_gene","IG_V_pseudogene"))
        },
        equcab3 = {
            return(c("miRNA","misc_RNA","protein_coding","pseudogene","rRNA",
                "processed_pseudogene","snoRNA","snRNA"))
        },
        tair10 = {
            return(c("miRNA","ncRNA","protein_coding","pseudogene","rRNA",
                "snoRNA","snRNA","transposable_element","tRNA"))
        }
    )
}

getSupportedRefDbs <- function() {
    return(c("ensembl","ucsc","refseq"))
}

getSupportedOrganisms <- function() {
    return(c("hg18","hg19","hg38","mm9","mm10","mm39","rn5","rn6","dm3","dm6",
        "danrer7","danrer10","danrer11","pantro4","pantro5","susscr3",
        #"pantro6",
        "susscr11","equcab3","tair10"))
}

getSupportedUcscDbs <- function() {
    return(c("ucsc","refseq"))
}

importCustomGtf <- function(gtfFile,level=c("gene","transcript","exon"),
    type=c("gene","exon","utr")) {
    if (!requireNamespace("GenomicFeatures"))
        stop("Bioconductor package GenomicFeatures is required to build ",
            "custom annotation!")
        
    # Some argument checking
    level <- level[1]
    type <- type[1]
    checkTextArgs("level",level,c("gene","transcript","exon"),multiarg=FALSE)
    checkTextArgs("type",type,c("gene","exon","utr"),multiarg=FALSE)
    
    # Import the GTF with rtracklayer to create a map of available metadata
    message("  Importing GTF ",gtfFile," as GTF to make id map")
    desiredColumns <- c("type","gene_id","transcript_id","exon_id",
        "gene_name","gene_biotype")
    gr <- import(gtfFile,format="gtf",colnames=desiredColumns,
        feature.type=.GFF_FEATURE_TYPES)
    grdf <- as.data.frame(gr)
    grdf <- grdf[grdf$type=="exon",]

    # Need to map gene ids to names and biotypes. We need a collapsed 
    # structure exon_id - transcript_id - gene_id - gane_name - biotype
    message("  Making id map")
    map <- .makeIdMap(grdf)
    
    message("  Importing GTF ",gtfFile," as TxDb")
    txdb <- suppressWarnings(makeTxDbFromGFF(gtfFile))
    
    switch(level,
        gene = {
            switch(type,
                gene = {
                    return(.makeGeneGeneFromTxDb(txdb,map))
                },
                exon = {
                    return(.makeGeneExonFromTxDb(txdb,map))
                },
                utr = {
                    return(.makeGeneUtrFromTxDb(txdb,map))
                }
            )
        },
        transcript = {
            switch(type,
                gene = {
                    return(.makeTranscriptGeneFromTxDb(txdb,map))
                },
                exon = {
                    # Stub
                },
                utr = {
                    return(.makeTranscriptUtrFromTxDb(txdb,map))
                }
            )
        },
        exon = {
            switch(type,
                exon = {
                    return(.makeExonExonFromTxDb(txdb,map))
                }
            )
        }
    )
}

# parsed <- parseCustomGtf(gffFile)
parseCustomGtf <- function(gtfFile) {
    if (!requireNamespace("GenomicFeatures"))
        stop("Bioconductor package GenomicFeatures is required to parse ",
            "custom GTF file!")
    
    # Import the GTF with rtracklayer to create a map of available metadata
    message("  Importing GTF ",gtfFile," as GTF to make id map")
    desiredColumns <- c("type","gene_id","transcript_id","exon_id",
        "gene_name","gene_biotype")
    # Let it recognize automatically GFF/GTF
    gr <- import(gtfFile,colnames=desiredColumns,
        feature.type=.GFF_FEATURE_TYPES)
    grdf <- as.data.frame(gr)
    grdf <- grdf[grdf$type=="exon",]

    # Need to map gene ids to names and biotypes. We need a collapsed 
    # structure exon_id - transcript_id - gene_id - gane_name - biotype
    message("  Making id map")
    map <- .makeIdMap(grdf)
    
    message("  Importing GTF ",gtfFile," as TxDb")
    txdb <- suppressWarnings(makeTxDbFromGFF(gtfFile))
    
    return(list(txdb=txdb,map=map))
}

annotationFromCustomGtf <- function(parsed,level=c("gene","transcript","exon"),
    type=c("gene","exon","utr"),summarized=FALSE,asdf=FALSE) {
    # Some argument checking
    if (!is.logical(summarized))
        stop("summarized must be TRUE or FALSE")
    level <- level[1]
    type <- type[1]
    checkTextArgs("level",level,c("gene","transcript","exon"),multiarg=FALSE)
    checkTextArgs("type",type,c("gene","exon","utr"),multiarg=FALSE)
    
    txdb <- parsed$txdb
    map <- parsed$map
    
    switch(level,
        gene = {
            switch(type,
                gene = {
                    return(.makeGeneGeneFromTxDb(txdb,map,asdf))
                },
                exon = {
                    if (summarized)
                        return(.makeSumGeneExonFromTxDb(txdb,map,asdf))
                    else
                        return(.makeGeneExonFromTxDb(txdb,map,asdf))
                },
                utr = {
                    if (summarized)
                        return(.makeSumGeneUtrFromTxDb(txdb,map,asdf))
                    else
                        return(.makeGeneUtrFromTxDb(txdb,map,asdf))
                }
            )
        },
        transcript = {
            switch(type,
                gene = {
                    if (summarized)
                        return(.makeSumTranscriptGeneFromTxDb(txdb,map,asdf))
                    else
                        return(.makeTranscriptGeneFromTxDb(txdb,map,asdf))
                },
                exon = {
                    # Stub
                },
                utr = {
                    if (summarized)
                        return(.makeSumTranscriptUtrFromTxDb(txdb,map,asdf))
                    else
                        return(.makeTranscriptUtrFromTxDb(txdb,map,asdf))
                }
            )
        },
        exon = {
            switch(type,
                exon = {
                    return(.makeExonExonFromTxDb(txdb,map,asdf))
                }
            )
        }
    )
}

# ann <- .makeGeneGeneFromTxDb(txdb,map,FALSE)
.makeGeneGeneFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- genes(txdb)
        # We need to add gc_content, gene_name, biotype from the map
        # Remove the exon_id and transcript_id columns from the map
        # for the gene case
        ir <- c(which(names(map)=="exon_id"),
            which(names(map)=="transcript_id"))
        if (length(ir) > 0)
            smap <- map[,-ir,drop=FALSE]
        dgt <- which(duplicated(smap))
        if (length(dgt) > 0)
            smap <- smap[-dgt,]
        
        # Add metadata from the map
        rownames(smap) <- smap$gene_id
        smap <- smap[names(gr),,drop=FALSE]
        gr$gene_name <- smap$gene_name
        gr$biotype <- smap$biotype
    }
    else {
        gr <- genes(txdb,columns=c("gene_id","tx_type"))
        gr$gene_id <- as.character(gr$gene_id)
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(gr$tx_type)
        gr$tx_type <- NULL
    }
    
    gr$gc_content = rep(50,length(gr))
    ann <- as.data.frame(gr)
    ann <- ann[,c(1:3,6,9,5,7,8)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeGeneExonFromTxDb(txdb,map,FALSE)
.makeGeneExonFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- exons(txdb,columns="exon_name")
        
        # There are cases where exons are unnamed and just the structure 
        useMap <- TRUE
        if (any(is.na(gr$exon_name)))
            useMap <- FALSE
        
        if (useMap) {
            names(gr) <- gr$exon_name
        
            # We need to add gene_id, gene_name, biotype from the map
            # Remove the transcript_id column from the map for the gene
            # case
            ir <- which(names(map)=="transcript_id")
            if (length(ir) > 0)
                smap <- map[,-ir,drop=FALSE]
            dgt <- which(duplicated(smap))
            if (length(dgt) > 0)
                smap <- smap[-dgt,]
            
            # Add metadata from the map
            rownames(smap) <- smap$exon_id
            smap <- smap[names(gr),,drop=FALSE]
            gr$gene_id <- smap$gene_id
            gr$gene_name <- smap$gene_name
            gr$biotype <- smap$biotype
        }
        else {
            # In order to create the exon annotation, we need to manually 
            # overlap and assign to genes
            ge <- genes(txdb,columns=c("gene_id","tx_type"))
            ov <- findOverlaps(gr,ge)
            exonPos <- queryHits(ov)
            genePos <- subjectHits(ov)
            levs <- unique(genePos)
            
            dup <- which(duplicated(exonPos))
            if (length(dup) > 0) {
                exonPos <- exonPos[-dup]
                genePos <- genePos[-dup]
            }
            
            S <- split(exonPos,factor(genePos,levels=levs))
            gr$gene_id <- gr$gene_name <- rep(names(ge)[levs],lengths(S))
            gr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],
                lengths(S))
            #gr$biotype <- rep("gene",length(gr))
            if (is(gr$gene_id,"CharacterList"))
                gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
            else
                gtmp <- gr$gene_id
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gtmp,sep="")
            gr$exon_name <- names(gr)
        }
    }
    else {
        gr <- exons(txdb,columns=c("exon_name","gene_id","tx_type"))
        if (is(gr$gene_id,"CharacterList"))
            gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
        else
            gtmp <- gr$gene_id
        if (any(is.na(gr$exon_name)))            
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gtmp,sep="")
        else
            names(gr) <- gr$exon_name
        gr$gene_id <- gtmp
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(lapply(gr$tx_type,function(x) x[1]))
        gr$tx_type <- NULL
    }
    
    ann <- as.data.frame(unname(gr))
    ann <- ann[,c(1,2,3,6,7,5,8,9)]
    names(ann)[c(1,4)] <- c("chromosome","exon_id")
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    if (length(unique(ann$exon_id)) == length(ann$exon_id))
        # Should always be TRUE
        rownames(ann) <- as.character(ann$exon_id)
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumGeneExonFromTxDb(txdb,map,FALSE)
.makeSumGeneExonFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- exons(txdb,columns="exon_name")
        
        useMap <- TRUE
        if (any(is.na(gr$exon_name)))
            useMap <- FALSE
        
        if (useMap) {
            names(gr) <- gr$exon_name
            
            # We need to add gene_id, gene_name, biotype from the map
            # Remove the transcript_id column from the map for the gene
            # case
            ir <- which(names(map)=="transcript_id")
            if (length(ir) > 0)
                smap <- map[,-ir,drop=FALSE]
            dgt <- which(duplicated(smap))
            if (length(dgt) > 0)
                smap <- smap[-dgt,]
            
            # Add metadata from the map
            rownames(smap) <- smap$exon_id
            smap <- smap[names(gr),,drop=FALSE]
            gr$gene_id <- smap$gene_id
            gr$gene_name <- smap$gene_name
            gr$biotype <- smap$biotype
        }
        else {
            # In order to create the exon annotation, we need to manually 
            # overlap and assign to genes
            ge <- genes(txdb,columns=c("gene_id","tx_type"))
            ov <- findOverlaps(gr,ge)
            exonPos <- queryHits(ov)
            genePos <- subjectHits(ov)
            levs <- unique(genePos)
            
            dup <- which(duplicated(exonPos))
            if (length(dup) > 0) {
                exonPos <- exonPos[-dup]
                genePos <- genePos[-dup]
            }
            
            S <- split(exonPos,factor(genePos,levels=levs))
            gr$gene_id <- gr$gene_name <- rep(names(ge)[levs],lengths(S))
            gr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],
                lengths(S))
            if (is(gr$gene_id,"CharacterList"))
                gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
            else
                gtmp <- gr$gene_id
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gtmp,sep="")
            gr$exon_name <- names(gr)
        }
    }
    else {
        gr <- exons(txdb,columns=c("exon_name","gene_id","tx_type"))
        if (is(gr$gene_id,"CharacterList"))
            gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
        else
            gtmp <- gr$gene_id
        if (any(is.na(gr$exon_name)))            
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gtmp,sep="")
        else
            names(gr) <- gr$exon_name
        gr$gene_id <- gtmp
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(lapply(gr$tx_type,function(x) x[1]))
        gr$tx_type <- NULL
    }
    
    ann <- as.data.frame(unname(gr))
    ann <- ann[,c(1,2,3,6,8,5,7,9)]
    names(ann)[c(1,4)] <- c("chromosome","exon_id")
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    message("  summarizing exons per gene for imported GTF")
    annList <- reduceExons(GRanges(ann))
    sexon <- annList$model
    names(sexon) <- as.character(sexon$exon_id)
    activeLength <- annList$length
    names(activeLength) <- unique(sexon$gene_id)
    attr(sexon,"activeLength") <- activeLength
    
    if (asdf) {
        eann <- as.data.frame(sexon)
        eann <- eann[,c(1,2,3,6,8,5,7,9)]
        names(eann)[c(1,4)] <- c("chromosome","exon_id")
        attr(eann,"activeLength") <- activeLength
        return(eann)
    }
    else
        return(sexon)
}

# ann <- .makeGeneUtrFromTxDb(txdb,map,FALSE)
.makeGeneUtrFromTxDb <- function(txdb,map,asdf) {
    utrList <- threeUTRsByTranscript(txdb,use.names=TRUE)
    utrGr <- unlist(utrList)
    
    # There may be no sufficient information to create UTRs
    if (length(utrGr) == 0) {
        warning("No UTR information was found in the provided GTF file! Will ",
            "return an empty object...",immediate.=TRUE)
        ann <- data.frame(chromosome=1,start=1,end=1,
            transcript_id=1,gene_id=1,strand=1,gene_name=1,
            biotype=1,row.names=1)
        # Strange but required to be compatible with GRanges
        ann <- ann[-1,]
        if (asdf)
            return(ann)
        else
            return(GRanges(ann))
    }
    
    utrGr$transcript_id <- names(utrGr)
    utrTmp <- as.data.frame(unname(utrGr))
    keep <- c("seqnames","start","end","transcript_id",
        "exon_rank","strand","exon_name")
    utr <- utrTmp[,keep]
    
    useMap <- TRUE
    if (any(is.na(utr$exon_name)) || is.null(map))
        # Impossible to take info from map...
        useMap <- FALSE
    else
        rownames(utr) <- paste(utr$exon_name,utr$transcript_id,sep="_")
        
    if (useMap) {
        rownames(map) <- paste(map$exon_id,map$transcript_id,sep="_")

        # Different case with map here
        smap <- map[rownames(utr),]
        
        # We need to add gene_id, gene_name, biotype from the map
        # Remove the exon_id column from the map for the gene case
        ir <- which(names(smap)=="exon_id")
        if (length(ir) > 0)
            smap <- smap[,-ir,drop=FALSE]
        
        # Add metadata from the map
        utr$gene_id <- smap$gene_id
        utr$gene_name <- smap$gene_name
        utr$biotype <- smap$biotype
    }
    else {
        # Add metadata from the utr frame
        utr$gene_id <- utr$transcript_id
        utr$gene_name <- utr$gene_id
        
        # To get biotypes
        ge <- genes(txdb,columns=c("gene_id","tx_type"))
        ov <- findOverlaps(utrGr,ge)
        utrPos <- queryHits(ov)
        genePos <- subjectHits(ov)
        levs <- unique(genePos)
        
        dup <- which(duplicated(utrPos))
        if (length(dup) > 0) {
            utrPos <- utrPos[-dup]
            genePos <- genePos[-dup]
        }
        
        S <- split(utrPos,factor(genePos,levels=levs))
        utr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],lengths(S))
        #utr$biotype <- rep("gene",nrow(utr))
    }
    
    ann <- utr[,c(1,2,3,4,8,6,9,10)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumGeneUtrFromTxDb(txdb,map,FALSE)
.makeSumGeneUtrFromTxDb <- function(txdb,map,asdf) {
    utrList <- threeUTRsByTranscript(txdb,use.names=TRUE)
    utrGr <- unlist(utrList)
    
    if (length(utrGr) == 0) {
        warning("No UTR information was found in the provided GTF file! Will ",
            "return an empty object...",immediate.=TRUE)
        ann <- data.frame(chromosome=1,start=1,end=1,
            transcript_id=1,gene_id=1,strand=1,gene_name=1,
            biotype=1,row.names=1)
        # Strange but required to be compatible with GRanges
        ann <- ann[-1,]
        if (asdf)
            return(ann)
        else
            return(GRanges(ann))
    }
    
    utrGr$transcript_id <- names(utrGr)
    utrTmp <- as.data.frame(unname(utrGr))
    keep <- c("seqnames","start","end","transcript_id",
        "exon_rank","strand","exon_name")
    utr <- utrTmp[,keep]
    
    useMap <- TRUE
    if (any(is.na(utr$exon_name)) || is.null(map))
        useMap <- FALSE
    else
        rownames(utr) <- paste(utr$exon_name,utr$transcript_id,sep="_")
    
    if (useMap) {
        rownames(map) <- paste(map$exon_id,map$transcript_id,sep="_")
    
        # Different case with map here
        smap <- map[rownames(utr),]
        
        # We need to add gene_id, gene_name, biotype from the map
        # Remove the exon_id column from the map for the gene case
        ir <- which(names(smap)=="exon_id")
        if (length(ir) > 0)
            smap <- smap[,-ir,drop=FALSE]
        
        # Add metadata from the map
        utr$gene_id <- smap$gene_id
        utr$gene_name <- smap$gene_name
        utr$biotype <- smap$biotype
    }
    else {
        # Add metadata from the utr frame
        utr$gene_id <- utr$transcript_id
        utr$gene_name <- utr$gene_id
        
        # To get biotypes
        ge <- genes(txdb,columns=c("gene_id","tx_type"))
        ov <- findOverlaps(utrGr,ge)
        utrPos <- queryHits(ov)
        genePos <- subjectHits(ov)
        levs <- unique(genePos)
        
        dup <- which(duplicated(utrPos))
        if (length(dup) > 0) {
            utrPos <- utrPos[-dup]
            genePos <- genePos[-dup]
        }
        
        S <- split(utrPos,factor(genePos,levels=levs))
        utr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],lengths(S))
    }
    ann <- utr[,c(1,2,3,4,8,6,9,10)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    message("  summarizing UTRs per gene for imported GTF")
    #s3utr <- reduceTranscripts(GRanges(ann))
    annList <- reduceTranscripts(GRanges(ann))
    s3utr <- annList$model
    names(s3utr) <- as.character(s3utr$transcript_id)
    activeLength <- annList$length
    names(activeLength) <- unique(as.character(s3utr$gene_id))
    
    if (asdf) {
        sann <- as.data.frame(s3utr)
        sann <- sann[,c(1,2,3,6,8,5,7,9)]
        if (useMap)
            names(sann)[c(1,4)] <- c("chromosome","gene_id")
        else {
            sann <- sann[,c(1,2,3,7,5,6,4,8)]
            names(sann)[c(1,4,7)] <- c("chromosome","gene_id","transcript_id")
        }
        attr(sann,"activeLength") <- activeLength
        return(sann)
    }
    else
        return(s3utr)
}

# ann <- .makeTranscriptGeneFromTxDb(txdb,map,FALSE)
.makeTranscriptGeneFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- transcripts(txdb,columns="tx_name")
        names(gr) <- gr$tx_name
        
        # We need to add gene_id, gene_name, biotype from the map
        # Remove the exon_id column from the map for the transcript 
        # case
        ir <- which(names(map)=="exon_id")
        if (length(ir) > 0)
            smap <- map[,-ir,drop=FALSE]
        dgt <- which(duplicated(smap))
        if (length(dgt) > 0)
            smap <- smap[-dgt,]
        
        # Add metadata from the map
        rownames(smap) <- smap$transcript_id
        smap <- smap[names(gr),,drop=FALSE]
        gr$gene_id <- smap$gene_name
        gr$gene_name <- smap$gene_name
        gr$biotype <- smap$biotype
    }
    else {
        gr <- transcripts(txdb,columns=c("tx_name","gene_id","tx_type"))
        if (is(gr$gene_id,"CharacterList"))
            gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
        else
            gtmp <- as.character(gr$gene_id)
        gr$gene_id <- gtmp
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(lapply(gr$tx_type,function(x) x[1]))
        gr$tx_type <- NULL
    }
    
    ann <- as.data.frame(gr)
    ann <- ann[,c(1,2,3,6,7,5,8,9)]
    names(ann)[c(1,4)] <- c("chromosome","transcript_id")
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumTranscriptGeneFromTxDb(txdb,map,FALSE)
.makeSumTranscriptGeneFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- transcripts(txdb,columns="tx_name")
        names(gr) <- gr$tx_name
        
        # We need to add gene_id, gene_name, biotype from the map
        # Remove the exon_id column from the map for the transcript 
        # case
        ir <- which(names(map)=="exon_id")
        if (length(ir) > 0)
            smap <- map[,-ir,drop=FALSE]
        dgt <- which(duplicated(smap))
        if (length(dgt) > 0)
            smap <- smap[-dgt,]
        
        # Add metadata from the map
        rownames(smap) <- smap$transcript_id
        smap <- smap[names(gr),,drop=FALSE]
        gr$gene_id <- smap$gene_name
        gr$gene_name <- smap$gene_name
        gr$biotype <- smap$biotype
    }
    else {
        gr <- transcripts(txdb,columns=c("tx_name","gene_id","tx_type"))
        if (is(gr$gene_id,"CharacterList"))
            gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
        else
            gtmp <- as.character(gr$gene_id)
        gr$gene_id <- gtmp
        #gr$gene_id <- as.character(gr$gene_id)
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(lapply(gr$tx_type,function(x) x[1]))
        gr$tx_type <- NULL
    }
    
    ann <- as.data.frame(gr)
    ann <- ann[,c(1,2,3,6,7,5,8,9)]
    names(ann)[c(1,4)] <- c("chromosome","transcript_id")
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    #stranscript <- reduceTranscripts(GRanges(ann))
    annList <- reduceTranscripts(GRanges(ann))
    stranscript <- annList$model
    names(stranscript) <- as.character(stranscript$transcript_id)
    activeLength <- annList$length
    #names(activeLength) <- as.character(stranscript$transcript_id)
    names(activeLength) <- unique(stranscript$gene_id)
    
    if (asdf) {
        sann <- as.data.frame(stranscript)
        sann <- sann[,c(1,2,3,6,8,5,7,9)]
        names(sann)[c(1,4)] <- c("chromosome","transcript_id")
        attr(sann,"activeLength") <- activeLength
        return(sann)
    }
    else
        return(stranscript)
}

# ann <- .makeTranscriptExonFromTxDb(txdb,map,FALSE)
.makeTranscriptExonFromTxDb <- function(txdb,map,asdf) {
    # Stub
    # TODO: "summarized_exon_by_transcript.rda"
}

# ann <- .makeTranscriptUtrFromTxDb(txdb,map,FALSE)
.makeTranscriptUtrFromTxDb <- function(txdb,map,asdf) {
    utrList <- threeUTRsByTranscript(txdb,use.names=TRUE)
    utrGr <- unlist(utrList)
    
    if (length(utrGr) == 0) {
        warning("No UTR information was found in the provided GTF file! Will ",
            "return an empty object...",immediate.=TRUE)
        ann <- data.frame(chromosome=1,start=1,end=1,
            transcript_id=1,gene_id=1,strand=1,gene_name=1,
            biotype=1,row.names=1)
        # Strange but required to be compatible with GRanges
        ann <- ann[-1,]
        if (asdf)
            return(ann)
        else
            return(GRanges(ann))
    }
    
    utrGr$transcript_id <- names(utrGr)
    utrTmp <- as.data.frame(unname(utrGr))
    keep <- c("seqnames","start","end","transcript_id",
        "exon_rank","strand","exon_name")
    utr <- utrTmp[,keep]
    
    useMap <- TRUE
    if (any(is.na(utr$exon_name)) || is.null(map))
        useMap <- FALSE
    else
        rownames(utr) <- paste(utr$exon_name,utr$transcript_id,sep="_")
    
    if (useMap) {
        rownames(map) <- paste(map$exon_id,map$transcript_id,sep="_")
    
        # Different case with map here
        smap <- map[rownames(utr),]
        
        # We need to add gene_id, gene_name, biotype from the map
        # Remove the exon_id column from the map for the gene case
        ir <- which(names(map)=="exon_id")
        if (length(ir) > 0)
            smap <- smap[,-ir,drop=FALSE]
        
        # Add metadata from the map
        #rownames(smap) <- smap$transcript_id
        #smap <- smap[rownames(utr),]
        utr$gene_id <- smap$gene_id
        utr$gene_name <- smap$gene_name
        utr$biotype <- smap$biotype
    }
    else {
        # Add metadata from the utr frame
        utr$gene_id <- utr$transcript_id
        utr$gene_name <- utr$gene_id
        
        # To get biotypes
        ge <- genes(txdb,columns=c("gene_id","tx_type"))
        ov <- findOverlaps(utrGr,ge)
        utrPos <- queryHits(ov)
        genePos <- subjectHits(ov)
        levs <- unique(genePos)
        
        dup <- which(duplicated(utrPos))
        if (length(dup) > 0) {
            utrPos <- utrPos[-dup]
            genePos <- genePos[-dup]
        }
        
        S <- split(utrPos,factor(genePos,levels=levs))
        utr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],lengths(S))
        #utr$biotype <- rep("gene",nrow(utr))
    }
    
    #ann <- utr[,c(1,2,3,4,7,6,8,9)]
    ann <- utr[,c(1,2,3,4,8,6,9,10)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

# ann <- .makeSumTranscriptUtrFromTxDb(txdb,map,FALSE)
.makeSumTranscriptUtrFromTxDb <- function(txdb,map,asdf) {
    utrList <- threeUTRsByTranscript(txdb,use.names=TRUE)
    utrGr <- unlist(utrList)
    
    if (length(utrGr) == 0) {
        warning("No UTR information was found in the provided GTF file! Will ",
            "return an empty object...",immediate.=TRUE)
        ann <- data.frame(chromosome=1,start=1,end=1,
            transcript_id=1,gene_id=1,strand=1,gene_name=1,
            biotype=1,row.names=1)
        # Strange but required to be compatible with GRanges
        ann <- ann[-1,]
        if (asdf)
            return(ann)
        else
            return(GRanges(ann))
    }
    
    utrGr$transcript_id <- names(utrGr)
    utrTmp <- as.data.frame(unname(utrGr))
    keep <- c("seqnames","start","end","transcript_id",
        "exon_rank","strand","exon_name")
    utr <- utrTmp[,keep]
    
    useMap <- TRUE
    if (any(is.na(utr$exon_name)) || is.null(map))
        useMap <- FALSE
    else
        rownames(utr) <- paste(utr$exon_name,utr$transcript_id,sep="_")
    
    if (useMap) {
        rownames(map) <- paste(map$exon_id,map$transcript_id,sep="_")
        
        # Different case with map here
        smap <- map[rownames(utr),]
        
        # We need to add gene_id, gene_name, biotype from the map
        # Remove the exon_id column from the map for the gene case
        ir <- which(names(map)=="exon_id")
        if (length(ir) > 0)
            smap <- smap[,-ir,drop=FALSE]
        
        # Add metadata from the map
        utr$gene_id <- smap$gene_id
        utr$gene_name <- smap$gene_name
        utr$biotype <- smap$biotype
    }
    else {
        # Add metadata from the utr frame
        utr$gene_id <- utr$transcript_id
        utr$gene_name <- utr$gene_id
        
        # To get biotypes
        ge <- genes(txdb,columns=c("gene_id","tx_type"))
        ov <- findOverlaps(utrGr,ge)
        utrPos <- queryHits(ov)
        genePos <- subjectHits(ov)
        levs <- unique(genePos)
        
        dup <- which(duplicated(utrPos))
        if (length(dup) > 0) {
            utrPos <- utrPos[-dup]
            genePos <- genePos[-dup]
        }
        
        S <- split(utrPos,factor(genePos,levels=levs))
        utr$biotype <- rep(unlist(ge$tx_type,use.names=FALSE)[levs],lengths(S))
    }
    
    #ann <- utr[,c(1,2,3,4,7,6,8,9)]
    ann <- utr[,c(1,2,3,4,8,6,9,10)]
    names(ann)[1] <- "chromosome"
    ann$chromosome <- as.character(ann$chromosome)
    ann <- ann[order(ann$chromosome,ann$start),]
    
    message("  summarizing UTRs per gene for imported GTF")
    #s3utrTranscript <- reduceTranscriptsUtr(GRanges(ann))
    annList <- reduceTranscriptsUtr(GRanges(ann))
    s3utrTranscript <- annList$model
    names(s3utrTranscript) <- 
        as.character(s3utrTranscript$transcript_id)
    activeLength <- annList$length
    #names(activeLength) <- as.character(s3utrTranscript$transcript_id)
    # Remember, gene_id in UTR's per transcript is the transcript
    names(activeLength) <- unique(as.character(s3utrTranscript$gene_id))
    
    if (asdf) {
        sann <- as.data.frame(s3utrTranscript)
        sann <- sann[,c(1,2,3,6,8,5,7,9)]
        names(sann)[c(1,4)] <- c("chromosome","transcript_id")
        attr(sann,"activeLength") <- activeLength
        return(sann)
    }
    else
        return(s3utrTranscript)
}

# ann <- .makeExonExonFromTxDb(txdb,map,FALSE)
.makeExonExonFromTxDb <- function(txdb,map,asdf) {
    if (!is.null(map)) {
        gr <- exons(txdb,columns="exon_name")
        
        # There are cases where exons are unnamed and just the structure 
        useMap <- TRUE
        if (any(is.na(gr$exon_name)))
            useMap <- FALSE
        
        if (useMap) {
            names(gr) <- gr$exon_name
            # We need to add gene_id, gene_name, biotype from the map
            # Remove the transcript_id column from the map for the gene
            # case
            ir <- which(names(map)=="transcript_id")
            if (length(ir) > 0)
                smap <- map[,-ir,drop=FALSE]
            dgt <- which(duplicated(smap))
            if (length(dgt) > 0)
                smap <- smap[-dgt,]
            
            # Add metadata from the map
            rownames(smap) <- smap$exon_id
            smap <- smap[names(gr),,drop=FALSE]
            gr$gene_id <- smap$gene_id
            gr$gene_name <- smap$gene_name
            gr$biotype <- smap$biotype
            
            ann <- as.data.frame(unname(gr))
            #ann <- ann[,c(1,2,3,6,8,5,7,9)]
            ann <- ann[,c(1,2,3,6,7,5,8,9)]
            names(ann)[c(1,4)] <- c("chromosome","exon_id")
            ann$chromosome <- as.character(ann$chromosome)
            ord <- order(ann$chromosome,ann$start)
            ann <- ann[ord,]
            smap <- smap[ord,]
            if (length(unique(ann$exon_id)) == length(ann$exon_id))
                rownames(ann) <- as.character(ann$exon_id)
            else
                rownames(ann) <- rownames(smap)
        }
        else {
            # In order to create the exon annotation, we need to manually 
            # overlap and assign to genes
            ge <- genes(txdb)
            ov <- findOverlaps(gr,ge)
            exonPos <- queryHits(ov)
            genePos <- subjectHits(ov)
            levs <- unique(genePos)
            
            dup <- which(duplicated(exonPos))
            if (length(dup) > 0) {
                exonPos <- exonPos[-dup]
                genePos <- genePos[-dup]
            }
            
            S <- split(exonPos,factor(genePos,levels=levs))
            gr$gene_id <- gr$gene_name <- rep(names(ge)[levs],lengths(S))
            gr$biotype <- rep("gene",length(gr))
            if (is(gr$gene_id,"CharacterList"))
                gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
            else
                gtmp <- gr$gene_id
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gtmp,sep="")
            gr$exon_name <- names(gr)
            
            ann <- as.data.frame(unname(gr))
            ann <- ann[,c(1,2,3,6,7,5,8,9)]
            names(ann)[c(1,4)] <- c("chromosome","exon_id")
            ann$chromosome <- as.character(ann$chromosome)
            ord <- order(ann$chromosome,ann$start)
            ann <- ann[ord,]
            if (length(unique(ann$exon_id)) == length(ann$exon_id))
                # Should always be TRUE
                rownames(ann) <- as.character(ann$exon_id)
        }
    }
    else {
        gr <- exons(txdb,columns=c("exon_name","gene_id","tx_type"))
        if (is(gr$gene_id,"CharacterList"))
            gtmp <- unlist(lapply(gr$gene_id,function(x) x[1]))
        else
            gtmp <- gr$gene_id
        if (any(is.na(gr$exon_name)))            
            names(gr) <- paste(seqnames(gr),":",start(gr),"-",end(gr),"_",
                gtmp,sep="")
        else
            names(gr) <- gr$exon_name
        gr$gene_id <- gtmp
        gr$gene_name <- gr$gene_id
        gr$biotype <- unlist(lapply(gr$tx_type,function(x) x[1]))
        gr$tx_type <- NULL
        ann <- as.data.frame(unname(gr))
        ann <- ann[,c(1,2,3,6,7,5,8,9)]
        names(ann)[c(1,4)] <- c("chromosome","exon_id")
        ann$chromosome <- as.character(ann$chromosome)
        ord <- order(ann$chromosome,ann$start)
        ann <- ann[ord,]
        if (length(unique(ann$exon_id)) == length(ann$exon_id))
            # Should always be TRUE
            rownames(ann) <- as.character(ann$exon_id)
    }
    
    if (asdf)
        return(ann)
    else
        return(GRanges(ann))
}

.makeIdMap <- function(grdf) {
    # Some basic names must be present in order to make a map
    if (!any(c("exon_id","transcript_id","gene_id") %in% names(grdf))
        || any(is.na(grdf$exon_id))
        || any(is.na(grdf$transcript_id))
        || any(is.na(grdf$gene_id)))
        return(NULL)

    #hasGeneName <- hasBiotype <- FALSE
    if (!all(is.na(grdf$gene_name)) && !all(is.na(grdf$gene_biotype))) {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_name",
            "gene_biotype")]
        names(map)[5] <- "biotype"
        #hasGeneName <- hasBiotype <- TRUE
    }
    else if (all(is.na(grdf$gene_name)) && !all(is.na(grdf$gene_biotype))) {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_id",
            "gene_biotype")]
        names(map)[4:5] <- c("gene_name","biotype")
        #hasBiotype <- TRUE
    }
    else if (!all(is.na(grdf$gene_name)) && all(is.na(grdf$gene_biotype))) {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_name")]
        map$gene_biotype <- rep("gene",nrow(map))
        names(map)[5] <- "biotype"
        #hasGeneName <- TRUE
    }
    else {
        map <- grdf[,c("exon_id","transcript_id","gene_id","gene_id")]
        map$gene_biotype <- rep("gene",nrow(map))
        names(map)[4:5] <- c("gene_name","biotype")
    }
    nag <- which(is.na(map$gene_name))
    if (length(nag) > 0)
        map$gene_name[nag] <- map$gene_id[nag]
    nab <- which(is.na(map$biotype))
    if (length(nab) > 0)
        map$biotype[nag] <- "gene"
    return(map)
}

initDatabase <- function(db) {
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to build the annotation database!")
    if (missing(db))
        stop("A database file must be provided!")
    drv <- dbDriver("SQLite")
    if (file.exists(db))
        # The database has been created at least with the tables defined
        con <- dbConnect(drv,dbname=db)
    else {
        # Create database file and define tables
        con <- dbConnect(drv,dbname=db)
        rs <- .initTables(con)
    }
    return(con)
}

.initTables <- function(con) {
    queries <- .localTblDef()
    rs <- dbSendQuery(con,queries[[1]])
    if (dbHasCompleted(rs))
        dbClearResult(rs)
    for (n in names(queries)) {
        rs <- dbSendStatement(con,queries[[n]])
        if (dbHasCompleted(rs))
            dbClearResult(rs)
    }
}

.chromInfoToSeqInfoDf <- function(ci,o="custom",circ=FALSE,asSeqinfo=FALSE) {
    if (asSeqinfo)
        return(Seqinfo(seqnames=rownames(ci),seqlengths=ci[,1],
            isCircular=rep(circ,nrow(ci)),genome=o))
    else
        return(data.frame(chromosome=rownames(ci),length=as.integer(ci[,1])))
}

.chromInfoWrapperGID <- function(o) {
    #if (packageVersion("GenomeInfoDb")>=1.23)
    #    return(GenomeInfoDb::getChromInfoFromUCSC(o))
    #else
    #    return(GenomeInfoDb::getChromInfoFromUCSC(o))
    if (o == "TAIR10") {
        g <- GenomeInfoDb::getChromInfoFromEnsembl(o,division="plants")
        g$name = paste("chr",g$name,sep="")
        return(g)
    }
    else
        return(GenomeInfoDb::getChromInfoFromUCSC(o))
}

.validateEnsemblVersions <- function(o,v) {
    ea <- biomaRt::listEnsemblArchives()
    found <- as.character(v) %in% ea[,"version"]
    if (any(!found)) {
        nf <- which(!found)
        warning("Version(s) ",paste0(v[nf],collapse=", ")," not found in ",
            "biomaRt archives for current organism!\nWill use the latest ",
            "available version(s)...",immediate.=TRUE)
        newv <- v[-nf]
        if (length(newv) == 0) {
            warning("No Ensembl versions left after validation! Will return ",
                "only the latest available...",immediate.=TRUE)
            u2e <- ucscToEnsembl()
            vers <- u2e[[o]]
            for (newv in rev(vers)) {
                newi <- grep(as.character(newv),ea[,"version"])
                if (length(newi) > 0)
                    return(newv)
            }
        }
        return(newv)
    }
    return(v)
}

# The new function with the new weights.
getWeights <- function(org=c("human","chimpanzee","mouse","fruitfly",
    "arabidopsis","rat")) {
    org <- tolower(org)
    checkTextArgs("org",org,c("human","chimpanzee","mouse","fruitfly",
        "arabidopsis","rat"))
    switch(org,
        human = {
            #return(c(
            #    deseq=0.03077317,
            #    deseq2=0.16791844,
            #    edger=0.07695718,
            #    limma=0.18600075,
            #    nbpseq=0.03315111,
            #    noiseq=0.05058633,
            #    bayseq=0.15290457,
            #    absseq=0.12843044,
            #    dss=0.17327802
            #))
            return(c(
                deseq=0.04988624,
                deseq2=0.18703151,
                edger=0.09607025,
                limma=0.20511382,
                nbpseq=0.05226418,
                noiseq=0.06969940,
                absseq=0.14754351,
                dss=0.19239109
            ))
        },
        chimpanzee = {
            #return(c(
            #    deseq=0.03345859,
            #    deseq2=0.17691534,
            #    edger=0.08214923,
            #    limma=0.18127409,
            #    nbpseq=0.03209271,
            #    noiseq=0.05261594,
            #    bayseq=0.15328300,
            #    absseq=0.13449179,
            #    dss=0.15371930
            #))
            return(c(
                deseq=0.05261897,
                deseq2=0.19607572,
                edger=0.10130960,
                limma=0.20043447,
                nbpseq=0.05125309,
                noiseq=0.07177632,
                absseq=0.15365217,
                dss=0.17287968
            ))
        },
        mouse = {
            #return(c(
            #    deseq=0.02707875,
            #    deseq2=0.16434559,
            #    edger=0.12614496,
            #    limma=0.17038977,
            #    nbpseq=0.02614409,
            #    noiseq=0.02064747,
            #    bayseq=0.15508511,
            #    absseq=0.14800242,
            #    dss=0.16216184
            #))
            return(c(
                deseq=0.04646439,
                deseq2=0.18373123,
                edger=0.14553060,
                limma=0.18977541,
                nbpseq=0.04552973,
                noiseq=0.04003311,
                absseq=0.16738806,
                dss=0.18154748
            ))
        },
        fruitfly = {
            #return(c(
            #    deseq=0.006526282,
            #    deseq2=0.172079991,
            #    edger=0.064114940,
            #    limma=0.193692581,
            #    nbpseq=0.005907387,
            #    noiseq=0.019310076,
            #    bayseq=0.177483603,
            #    absseq=0.166022212,
            #    dss=0.194862929
            #))
            return(c(
                deseq=0.02871173,
                deseq2=0.19426544,
                edger=0.08630039,
                limma=0.21587803,
                nbpseq=0.02809284,
                noiseq=0.04149553,
                absseq=0.18820766,
                dss=0.21704838
            ))
        },
        arabidopsis = {
            #return(c(
            #    deseq=0.02745255,
            #    deseq2=0.18604304,
            #    edger=0.05530561,
            #    limma=0.20806899,
            #    nbpseq=0.02567801,
            #    noiseq=0.04377844,
            #    bayseq=0.16104719,
            #    absseq=0.09621622,
            #    dss=0.19640995
            #))
            return(c(
                deseq=0.04758345,
                deseq2=0.20617394,
                edger=0.07543651,
                limma=0.22819989,
                nbpseq=0.04580891,
                noiseq=0.06390934,
                absseq=0.11634712,
                dss=0.21654085
            ))
        },
        rat = {
            #return(c(
            #    deseq=0.03409202,
            #    deseq2=0.15203180,
            #    edger=0.14542665,
            #    limma=0.15327172,
            #    nbpseq=0.03501810,
            #    noiseq=0.01954866,
            #    bayseq=0.15587551,
            #    absseq=0.15289613,
            #    dss=0.15183941
            #))
            return(c(
                deseq=0.05357646,
                deseq2=0.17151624,
                edger=0.16491109,
                limma=0.17275616,
                nbpseq=0.05450254,
                noiseq=0.03903310,
                absseq=0.17238057,
                dss=0.17132385
            ))
        }
    )
}

getDefaults <- function(what,method=NULL) {
    if (what %in% c("normalization","statistics") && is.null(method))
        stopwrap("The method argument must be provided when what is ",
            "\"normalization\" or \"statistics\"!")
    switch(what,
        normalization = {
            switch(method,
                edaseq = {
                    return(list(within.which="loess",between.which="full"))
                },
                deseq = {
                    return(list(locfunc=median))
                },
                deseq2 = {
                    return(list(
                        tidy=FALSE, # DESeqDataSet object creation args
                        type="ratio",locfunc=stats::median
                    ))
                },                
                edger = {
                    return(list(
                        method="TMM",refColumn=NULL,logratioTrim=0.3,
                        sumTrim=0.05,doWeighting=TRUE,Acutoff=-1e10,p=0.75
                    ))
                },
                noiseq = {
                    return(list(
                        method="tmm", # which normalization
                        long=1000,lc=1,k=1, # common arguments
                        refColumn=1,logratioTrim=0.3,sumTrim=0.05,
                        # TMM normalization arguments
                        doWeighting=TRUE,Acutoff=-1e+10 
                    ))
                },
                nbpseq = {
                    return(list(main.method="nbsmyth",method="AH2010",
                        thinning=TRUE))
                },
                absseq = {
                    # method for estimating sizeFactors
                    return(list(normMethod="qtotal"))
                },
                dss = {
                    return(list(method="lr"))                        
                }
            )
        },           
        statistics = {
            switch(method,
                deseq = {
                    return(list(method="blind",sharingMode="fit-only",
                        fitType="local"))
                },
                deseq2 = {
                    # If sth is changed from here, change it also from the 
                    # switch beggining at line 804
                    return(list(      
                        # DESeqDataSetFromMatrix default args
                        tidy=FALSE,
                        # sizeFactors default args won't be passed because user 
                        # doesn't have to change anything at this point
                        # estimateDispersions default args
                        fitType= "parametric", maxit= 100, quiet= FALSE, 
                        modelMatrix= NULL, 
                        #minmu= 0.5, #removed because it caused 
                        # "Error ...bla-bla... unused argument minmu= 0.5"
                        #nbinomWaldTest default args
                        betaPrior= FALSE,
                        #betaPriorVar, modelMatrixType, 
                        betaTol=1e-08,useOptim=TRUE,
                        useT=FALSE,useQR=TRUE,
                        #results default args
                        lfcThreshold=0,altHypothesis="greaterAbs",
                        independentFiltering=TRUE, alpha=0.1,pAdjustMethod="BH",
                        format= "DataFrame",addMLE= FALSE,parallel=FALSE#,
                        #BPPARAM=bpparam()
                        #nbinomLRT default args are all passed above
                    ))
                },  
                edger = {
                    return(list(
                        main.method="classic", # classic or glm fit
                        rowsum.filter=5,prior.df=10,
                        trend="movingave",span=NULL,
                        # classic estimateCommonDisp arguments
                        tag.method="grid",grid.length=11,grid.range=c(-6,6),
                        # classic estimateTagwiseDisp arguments
                        offset=NULL,glm.method="CoxReid",subset=10000,
                        # glm estimateGLMCommonDisp and estimateGLMTrendedDisp 
                        # arguments
                        AveLogCPM=NULL,trend.method="auto",
                        # glm estimateGLMTagwiseDisp arguments
                        dispersion=NULL,offset=NULL,weights=NULL, 
                        # glmFit arguments
                        lib.size=NULL,prior.count=0.125,start=NULL,
                        method="auto",test="chisq",
                        # glmLRT arguments
                        abundance.trend=TRUE,robust=FALSE,
                        winsor.tail.p=c(0.05,0.1) # glmLFTest arguments
                    ))
                },
                noiseq = {
                    return(list(
                        k=0.5,norm="n",replicates="biological",
                        factor="class",conditions=NULL,pnr=0.2,
                        nss=5,v=0.02,lc=1, 
                        # noiseq general and specific arguments
                        nclust=15,r=100,adj=1.5,
                        a0per=0.9,filter=0,depth=NULL,        
                        cv.cutoff=500,cpm=1 # noiseqbio specific arguments
                        
                    ))
                },
                #bayseq = {
                #    return(list(samplesize=10000,samplingSubset=NULL,
                #        equalDispersions=TRUE,estimation="QL",zeroML=FALSE,
                #        consensus=FALSE,moderate=TRUE,pET="BIC",
                #        marginalise=FALSE,subset=NULL,priorSubset=NULL,
                #        bootStraps=1,conv=1e-4,nullData=FALSE,returnAll=FALSE,
                #        returnPD=FALSE,discardSampling=FALSE,cl=NULL))
                #},
                limma = {
                    return(list(normalize.method="none"))
                },
                nbpseq = {
                    return(list(
                        main.method="nbsmyth",
                        model=list(nbpseq="log-linear-rel-mean",nbsmyth="NBP"),
                        tests="HOA",
                        alternative="two.sided"
                    ))
                },
                absseq = {
                    # if sth is changed from here, change it also from the
                    # switch beggining at line 802
                    return(list(
                        #ABSDataSet default args
                        paired = FALSE, minDispersion = NULL, minRates = 0.1,
                        maxRates = 0.3, LevelstoNormFC = 100,
                        #ABSSeq default args
                        adjmethod = "BH", replaceOutliers = TRUE, 
                        useaFold = FALSE, quiet = FALSE,
                        #ABSSeqlm default args
                        lmodel = TRUE, preval = 0.05,
                        qforkappa = 0, scale = FALSE
                    ))
                },
                dss = {
                    # if sth is changed from here, change it also from the 
                    # switch beggining at line 802
                    return(list(
                        #estDispersion default args
                        trend=FALSE,
                        #waldTest default args
                        equal.var=FALSE
                    ))
                }
            )
        },
        utrOpts = {
            return(list(
                frac=1,
                minLength=300,
                downstram=50
            ))
        },
        geneFilter = {
            return(list(
                length=list(
                    length=500
                ),
                avgReads=list(
                    average.per.bp=100,
                    quantile=0.75
                ),
                expression=list(
                    median=TRUE,
                    mean=FALSE,
                    quantile=NA,
                    known=NA,
                    custom=NA
                ),
                biotype=getDefaults("biotypeFilter",method[1]),
                presence=list(
                    frac=0.25,
                    min.count=10,
                    per.condition=FALSE
                )
            ))
        },
        exonFilter = {
            return(list(
                mnrpx=list(
                    exons.per.gene=5,
                    min.exons=2,
                    frac=1/5
                )
            ))
        },
        biotypeFilter = {
            switch(method,
                hg18 = {
                    return(list(
                        unprocessed_pseudogene=TRUE,
                        pseudogene=FALSE,
                        miRNA=FALSE,
                        retrotransposed=FALSE,
                        protein_coding=FALSE,
                        processed_pseudogene=FALSE,
                        snRNA=FALSE,
                        snRNA_pseudogene=TRUE,
                        Mt_tRNA_pseudogene=TRUE,
                        miRNA_pseudogene=TRUE,
                        misc_RNA=FALSE,
                        tRNA_pseudogene=TRUE,
                        snoRNA=FALSE,
                        scRNA_pseudogene=TRUE,
                        rRNA_pseudogene=TRUE,
                        snoRNA_pseudogene=TRUE,
                        rRNA=TRUE,
                        misc_RNA_pseudogene=TRUE,
                        IG_V_gene=FALSE,
                        IG_D_gene=FALSE,
                        IG_J_gene=FALSE,
                        IG_C_gene=FALSE,
                        IG_pseudogene=TRUE,
                        scRNA=FALSE
                    ))
                },
                hg19 = {
                    return(list(
                        pseudogene=FALSE,
                        lincRNA=FALSE,
                        protein_coding=FALSE,
                        antisense=FALSE,
                        processed_transcript=FALSE,
                        snRNA=FALSE,
                        sense_intronic=FALSE,
                        miRNA=FALSE,
                        misc_RNA=FALSE,
                        snoRNA=FALSE,
                        rRNA=TRUE,
                        polymorphic_pseudogene=FALSE,
                        sense_overlapping=FALSE,
                        three_prime_overlapping_ncrna=FALSE,
                        TR_V_gene=FALSE,
                        TR_V_pseudogene=TRUE,
                        TR_D_gene=FALSE,
                        TR_J_gene=FALSE,
                        TR_C_gene=FALSE,
                        TR_J_pseudogene=TRUE,
                        IG_C_gene=FALSE,
                        IG_C_pseudogene=TRUE,
                        IG_J_gene=FALSE,
                        IG_J_pseudogene=TRUE,
                        IG_D_gene=FALSE,
                        IG_V_gene=FALSE,
                        IG_V_pseudogene=TRUE
                    ))
                },
                hg38 = {
                    return(list(
                        protein_coding=FALSE,
                        polymorphic_pseudogene=FALSE,
                        lincRNA=FALSE,
                        unprocessed_pseudogene=TRUE,
                        processed_pseudogene=FALSE,
                        antisense=FALSE,
                        processed_transcript=FALSE,
                        transcribed_unprocessed_pseudogene=FALSE,
                        sense_intronic=FALSE,
                        unitary_pseudogene=TRUE,
                        IG_V_gene=FALSE,
                        IG_V_pseudogene=TRUE,
                        TR_V_gene=FALSE,
                        sense_overlapping=FALSE,
                        transcribed_processed_pseudogene=FALSE,
                        miRNA=FALSE,
                        snRNA=FALSE,
                        misc_RNA=FALSE,
                        rRNA=TRUE,
                        snoRNA=FALSE,
                        IG_J_pseudogene=TRUE,
                        IG_J_gene=FALSE,
                        IG_D_gene=FALSE,
                        three_prime_overlapping_ncrna=FALSE,
                        IG_C_gene=FALSE,
                        IG_C_pseudogene=TRUE,
                        pseudogene=TRUE,
                        TR_V_pseudogene=TRUE,
                        Mt_tRNA=TRUE,
                        Mt_rRNA=TRUE,
                        translated_processed_pseudogene=FALSE,
                        TR_J_gene=FALSE,
                        TR_C_gene=FALSE,
                        TR_D_gene=FALSE,
                        TR_J_pseudogene=TRUE,
                        LRG_gene=FALSE
                    ))
                },
                mm9 = {
                    return(list(
                        pseudogene=FALSE,
                        snRNA=FALSE,
                        protein_coding=FALSE,
                        antisense=FALSE,
                        miRNA=FALSE,
                        lincRNA=FALSE,
                        snoRNA=FALSE,
                        processed_transcript=FALSE,
                        misc_RNA=FALSE,
                        rRNA=TRUE,
                        sense_overlapping=FALSE,
                        sense_intronic=FALSE,
                        polymorphic_pseudogene=FALSE,
                        non_coding=FALSE,
                        three_prime_overlapping_ncrna=FALSE,
                        IG_C_gene=FALSE,
                        IG_J_gene=FALSE,
                        IG_D_gene=FALSE,
                        IG_V_gene=FALSE,
                        ncrna_host=FALSE
                    ))
                },
                mm10 = {
                    return(list(
                        pseudogene=FALSE,
                        snRNA=FALSE,
                        protein_coding=FALSE,
                        antisense=FALSE,
                        miRNA=FALSE,
                        snoRNA=FALSE,
                        lincRNA=FALSE,
                        processed_transcript=FALSE,
                        misc_RNA=FALSE,
                        rRNA=TRUE,
                        sense_intronic=FALSE,
                        sense_overlapping=FALSE,
                        polymorphic_pseudogene=FALSE,
                        IG_C_gene=FALSE,
                        IG_J_gene=FALSE,
                        IG_D_gene=FALSE,
                        IG_LV_gene=FALSE,
                        IG_V_gene=FALSE,
                        IG_V_pseudogene=TRUE,
                        TR_V_gene=FALSE,
                        TR_V_pseudogene=TRUE,
                        three_prime_overlapping_ncrna=FALSE
                    ))
                },
                dm3 = {
                    return(list(
                        protein_coding=FALSE,
                        ncRNA=FALSE,
                        snoRNA=FALSE,
                        pre_miRNA=FALSE,
                        pseudogene=FALSE,
                        snRNA=FALSE,
                        tRNA=FALSE,
                        rRNA=TRUE
                    ))
                },
                rn5 = {
                    return(list(
                        protein_coding=FALSE,
                        pseudogene=FALSE,
                        processed_pseudogene=FALSE,
                        miRNA=FALSE,
                        rRNA=TRUE,
                        misc_RNA=FALSE
                    ))
                },
                rn6 = {
                    return(list(
                        antisense=FALSE , 
                        lincRNA=FALSE, 
                        miRNA=FALSE, 
                        misc_RNA=FALSE, 
                        Mt_rRNA=TRUE, 
                        Mt_tRNA=TRUE, 
                        processed_pseudogene=FALSE, 
                        processed_transcript=FALSE, 
                        protein_coding=FALSE, 
                        pseudogene=FALSE, 
                        ribozyme=FALSE, 
                        rRNA=TRUE, 
                        scaRNA=FALSE, 
                        sense_intronic=FALSE, 
                        snoRNA=FALSE, 
                        snRNA=FALSE, 
                        sRNA=FALSE, 
                        TEC=FALSE, 
                        transcribed_processed_pseudogene=FALSE, 
                        transcribed_unprocessed_pseudogene=FALSE, 
                        unprocessed_pseudogene=TRUE
                    ))
                },
                danrer7 = {
                    return(list(
                        antisense=FALSE,
                        protein_coding=FALSE,
                        miRNA=FALSE,
                        snoRNA=FALSE,
                        rRNA=TRUE,
                        lincRNA=FALSE,
                        processed_transcript=FALSE,
                        snRNA=FALSE,
                        pseudogene=FALSE,
                        sense_intronic=FALSE,
                        misc_RNA=FALSE,
                        polymorphic_pseudogene=FALSE,
                        IG_V_pseudogene=TRUE,
                        IG_C_pseudogene=TRUE,
                        IG_J_pseudogene=TRUE,
                        non_coding=FALSE,
                        sense_overlapping=FALSE
                    ))
                },
                pantro4 = {
                    return(list(
                        protein_coding=FALSE,
                        pseudogene=FALSE,
                        processed_pseudogene=FALSE,
                        miRNA=FALSE,
                        rRNA=TRUE,
                        snRNA=FALSE,
                        snoRNA=FALSE,
                        misc_RNA=FALSE
                    ))
                },
                susscr3 = {
                    return(list(
                        antisense=FALSE,
                        protein_coding=FALSE,
                        lincRNA=FALSE,
                        pseudogene=FALSE,
                        processed_transcript=FALSE,
                        miRNA=FALSE,
                        rRNA=TRUE,
                        snRNA=FALSE,
                        snoRNA=FALSE,
                        misc_RNA=FALSE,
                        non_coding=FALSE,
                        IG_C_gene=FALSE,
                        IG_J_gene=FALSE,
                        IG_V_gene=FALSE,
                        IG_V_pseudogene=TRUE
                    ))
                },
                tair10 = {
                    return(list(
                        miRNA=FALSE,
                        ncRNA=FALSE,
                        protein_coding=FALSE,
                        pseudogene=FALSE,
                        rRNA=TRUE,
                        snoRNA=FALSE,
                        snRNA=FALSE,
                        transposable_element=FALSE,
                        tRNA=FALSE
                    ))
                },
                equcab3 = {
                    return(list(
                        miRNA=FALSE,
                        misc_RNA=FALSE,
                        protein_coding=FALSE,
                        pseudogene=FALSE,
                        processed_pseudogene=FALSE,
                        rRNA=TRUE,
                        snoRNA=FALSE,
                        snRNA=FALSE
                    ))
                }
            )
        }
    )
}

validateAlgArgs <- function(normalization,statistics,normArgs,statArgs) {
    if (any(is.na(statistics))) # Locally fake an option as it will not be used
        statistics <- "deseq2"
    if (normalization=="each") {
        if (!is.null(normArgs)) {
            for (s in statistics) {
                if (!is.null(normArgs[[s]])) {
                    switch(s,
                        deseq = {
                            tmp <- normArgs[[s]]
                            tmp <- validateListArgs("normalization",s,tmp)
                            normArgs[[s]] <- getDefaults("normalization",s)
                            if (length(tmp)>0)
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        },
                        deseq2 = {
                            tmp <- normArgs[[s]]
                            tmp <- validateListArgs("normalization",s,tmp)
                            normArgs[[s]] <- getDefaults("normalization",s)
                            if (length(tmp)>0)
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        },
                        edger = {
                            tmp <- normArgs[[s]]
                            tmp <- validateListArgs("statistics",s,tmp)
                            normArgs[[s]] <- getDefaults("normalization",s)
                            if (length(tmp)>0)
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        },
                        limma = {
                            tmp <- normArgs[[s]]
                            tmp <- validateListArgs("statistics","edger",tmp)
                            normArgs[[s]] <- getDefaults("normalization",
                                "edger")
                            if (length(tmp)>0)
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        },
                        nbpseq = {
                            tmp <- normArgs[[s]]
                            tmp <- validateListArgs("statistics",s,tmp)
                            normArgs[[s]] <- getDefaults("normalization",s)
                            if (length(tmp)>0)
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        },
                        noiseq = {
                            tmp <- normArgs[[s]]
                            tmp <- validateListArgs("statistics",s,tmp)
                            normArgs[[s]] <- getDefaults("normalization",s)
                            if (length(tmp)>0)
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        },
                        #bayseq = {
                        #    tmp <- normArgs[[s]]
                        #    tmp <- validateListArgs("statistics","edger",tmp)
                        #    normArgs[[s]] <- getDefaults("normalization",
                        #        "edger")
                        #    if (length(tmp)>0)
                        #        normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        #},
                        absseq = {
                            #the arguments' list passed by the user
                            tmp <- normArgs[[s]]
                            #it validates the arguments given
                            tmp <- validateListArgs("statistics",s,tmp) 
                            #it gets defaults and thus ignores everything else
                            normArgs[[s]] <- getDefaults("normalization", 
                                s) 
                            if (length(tmp)>0)
                                # it changes default values of valid arguments 
                                # to user-specified 
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp) 
                        },
                        dss = {
                            tmp <- normArgs[[s]]
                            tmp <- validateListArgs("statistics",s,tmp)
                            normArgs[[s]] <- getDefaults("normalization",s)
                            if (length(tmp)>0)
                                normArgs[[s]] <- setArg(normArgs[[s]],tmp)
                        }                        
                    )
                }
                else {
                    switch(s,
                        deseq = {
                            normArgs[[s]] <- getDefaults(normalization,
                                "deseq")
                        },
                        deseq2 = {
                            normArgs[[s]] <- getDefaults(normalization,
                                "deseq2")
                        },
                        edger = {
                            normArgs[[s]] <- getDefaults(normalization,
                                "edger")
                        },
                        limma = {
                            normArgs[[s]] <- getDefaults(normalization,
                                "edger")
                        },
                        nbpseq = {
                            normArgs[[s]] <- getDefaults(normalization,
                                "nbpseq")
                        },
                        noiseq = {
                            normArgs[[s]] <- getDefaults(normalization,
                                "noiseq")
                        },
                        #bayseq = {
                        #    normArgs[[s]] <- getDefaults(normalization,
                        #        "edger")
                        #},    
                        absseq = {
                            normArgs[[s]] <- getDefaults(normalization,
                                "absseq")
                        },
                        dss = {
                            normArgs[[s]] <- getDefaults(normalization,"dss")
                        }
                    )
                }
            }
        }
        else {
            normArgs <- vector("list",length(statistics))
            names(normArgs) <- statistics
            for (s in statistics) {
                switch(s,
                    deseq = {
                        normArgs[[s]] <- getDefaults(normalization,"deseq")
                    },
                    deseq2 = {
                        normArgs[[s]] <- getDefaults(normalization,"deseq2")
                    },    
                    edger = {
                        normArgs[[s]] <- getDefaults(normalization,"edger")
                    },
                    limma = {
                        normArgs[[s]] <- getDefaults(normalization,"edger")
                    },
                    nbpseq = {
                        normArgs[[s]] <- getDefaults(normalization,"nbpseq")
                    },
                    noiseq = {
                        normArgs[[s]] <- getDefaults(normalization,"noiseq")
                    },
                    #bayseq = {
                    #    normArgs[[s]] <- getDefaults(normalization,"edger")
                    #},
                    absseq = {
                        normArgs[[s]] <- getDefaults(normalization,"absseq")
                    },   
                    dss = {
                        norm.args[[s]] <- getDefaults(normalization,"dss")
                    }
                )
            }
        }
    }
    else {
        if (!is.null(normArgs))
        {
            tmp <- normArgs
            tmp <- validateListArgs("normalization",normalization,tmp)
            normArgs <- getDefaults("normalization",normalization)
            if (length(tmp)>0)
                normArgs <- setArg(normArgs,tmp)
        }
        else
            normArgs <- getDefaults("normalization",normalization)
    }
    for (s in statistics) {
        if (!is.null(statArgs[[s]])) {
            tmp <- statArgs[[s]]
            tmp <- validateListArgs("statistics",s,tmp)
            statArgs[[s]] <- getDefaults("statistics",s)
            if (length(tmp)>0)
                statArgs[[s]] <- setArg(statArgs[[s]],tmp)
        }
        else
            statArgs[[s]] <- getDefaults("statistics",s)
    }
    return(list(normArgs=normArgs,statArgs=statArgs))
}

validateListArgs <- function(what,method=NULL,argList) {
    what <- tolower(what)
    checkTextArgs("what",what,c("normalization","statistics","geneFilter",
        "exonFilter","biotype.filter"))
    if (what %in% c("normalization","statistics") && is.null(method))
        stopwrap("The method argument must be provided when what is ",
            "\"normalization\" or \"statistics\"!")
    switch(what,
        normalization = {
            switch(method,
                edaseq = { 
                    valid <- names(argList) %in% c("within.which",
                        "between.which")
                    not.valid <- which(!valid)
                },
                deseq = {
                    valid <- names(argList) %in% c("locfunc")
                    not.valid <- which(!valid) 
                },
                deseq2 = {
                    valid <- names(argList) %in% c("tidy","type","locfunc")
                    not.valid <- which(!valid) 
                },
                edger = {
                    valid <- names(argList) %in% c("method","refColumn",
                        "logratioTrim","sumTrim","doWeighting","Acutoff","p")
                    not.valid <- which(!valid)
                },
                noiseq = {
                    valid <- names(argList) %in% c("method","long","lc","k",
                        "refColumn","logratioTrim","sumTrim","doWeighting",
                        "Acutoff")
                    not.valid <- which(!valid)
                },
                nbpseq = {
                    valid <- names(argList) %in% c("main.method","method",
                        "thinning")
                    not.valid <- which(!valid)
                },
                absseq = {
                    valid <- names(argList) %in% c("normMethod")
                    not.valid <- which(!valid)
                },
                dss = {
                    valid <- names(argList) %in% c("method")
                    notValid <- which(!valid)
                }
            )
            if (length(not.valid)>0) {
                warnwrap(paste("The following",method,what,"argument names",
                    "are invalid and will be ignored:",
                    paste(names(argList)[not.valid],collapse=", ")))
                argList[not.valid] <- NULL
            }
            return(argList)
        },
        statistics = {
            switch(method,
                deseq = {
                    valid <- names(argList) %in% c("method","sharingMode",
                        "fitType")
                    not.valid <- which(!valid)
                },
                deseq2 = {
                    valid <- names(argList) %in% c("tidy","fitType","maxit",
                        "quiet","modelMatrix","minmu","betaPrior",
                        #"betaPriorVar","modelMatrixType", #removed because they
                        # are also removed in get defaults
                        "betaTol","useOptim","useT","df","useQR","lfcThreshold",
                        "altHypothesis","independentFiltering","alpha",
                        "pAdjustMethod","format","addMLE","parallel","BPPARAM")
                    not.valid <- which(!valid)
                },
                edger = {
                    valid <- names(argList) %in% c("main.method",
                        "rowsum.filter","prior.df","trend","span","tag.method",
                        "grid.length","grid.range","offset","glm.method",
                        "subset","AveLogCPM","trend.method","dispersion",
                        "offset","weights","lib.size","prior.count","start",
                        "method","abundance.trend","robust",
                        "winsor.tail.p")
                    not.valid <- which(!valid)
                },
                noiseq = {
                    valid <- names(argList) %in% c("k","norm","replicates",
                        "factor","conditions","pnr","nss","v","lc","nclust","r",
                        "adj","a0per","filter","depth","cv.cutoff","cpm")
                    not.valid <- which(!valid)
                },
                #bayseq = {
                #    valid <- names(argList) %in% c("samplesize",
                #        "samplingSubset","equalDispersions","estimation",
                #        "zeroML","consensus","moderate","pET","marginalise",
                #        "subset","priorSubset","bootStraps","conv","nullData",
                #        "returnAll","returnPD","discardSampling","cl")
                #    not.valid <- which(!valid)
                #},
                limma = {
                    valid <- names(argList) %in% c("normalize.method")
                    not.valid <- which(!valid)
                },
                nbpseq = {
                    valid <- names(argList) %in% c("main.method","method",
                        "tests","alternative")
                    not.valid <- which(!valid)
                },
                absseq = {
                    valid <- names(argList) %in% c("paired","minDispersion",
                        "minRates","maxRates","LevelstoNormFC","adjmethod",
                        "replaceOutliers","useaFold","quiet","lmodel","preval",
                        "qforkappa","scale")
                    not.valid <- which(!valid)
                },
                dss = {
                    valid <- names(argList) %in% c("trend","equal.var")
                    not.valid <- which(!valid)
                }
            )
            if (length(not.valid)>0) {
                warnwrap(paste("The following",method,what,"argument names",
                    "are invalid and will be ignored:",
                    paste(names(argList)[not.valid],collapse=", ")))
                argList[not.valid] <- NULL
            }
            return(argList)
        },
        geneFilter = {
            valid.1 <- names(argList) %in% c("length","avgReads","expression",
                "biotype","presence")
            not.valid.1 <- which(!valid.1)
            if (length(not.valid.1)>0) {
                warnwrap(paste("The following",method,what,"argument names",
                    "are invalid and will be ignored:",
                    paste(names(argList)[not.valid.1],collapse=", ")))
                argList[not.valid.1] <- NULL
            }
            if (length(argList)>0) {
                for (n in names(argList)) {
                    switch(n,
                        length = {
                            valid.2 <- names(argList[[n]]) %in% c("length")
                            not.valid.2 <- which(!valid.2)
                        },
                        avgReads = {
                            valid.2 <- names(argList[[n]]) %in%
                                c("average.per.bp","quantile")
                            not.valid.2 <- which(!valid.2)
                        },
                        expression = {
                            valid.2 <- names(argList[[n]]) %in% c("median",
                                "mean","quantile","known","custom")
                            not.valid.2 <- which(!valid.2)
                        },
                        presence = {
                            valid.2 <- names(argList[[n]]) %in% c("frac",
                                "min.count","per.condition")
                            not.valid.2 <- which(!valid.2)
                        }
                    )
                    if (length(not.valid.2)>0) {
                        warnwrap(paste("The following",method,what,
                            "sub-argument names are invalid and will be",
                            "ignored:",paste(names(argList[[n]])[not.valid.2],
                            collapse=", ")))
                        argList[[n]][not.valid.2] <- NULL
                    }
                }
            }
            return(argList)
        },
        exonFilter = {
            valid.1 <- names(argList) %in% c("mnrpx")
            not.valid.1 <- which(!valid.1)
            if (length(not.valid.1)>0) {
                warnwrap(paste("The following",method,what,"argument names",
                    "are invalid and will be ignored:",
                    paste(names(argList)[not.valid.1],collapse=", ")))
                argList[not.valid.1] <- NULL
            }
            if (length(argList)>0) {
                for (n in names(argList)) {
                    switch(n,
                        mnrpx = {
                            valid.2 <- names(argList[[n]]) %in%
                                c("exons.per.gene","min.exons","frac")
                            not.valid.2 <- which(!valid.2)
                        }
                    )
                    if (length(not.valid.2)>0) {
                        warnwrap(paste("The following",method,what,
                            "sub-argument names are invalid and will be",
                            "ignored:",paste(names(argList[[n]])[not.valid.2],
                            collapse=", ")))
                        argList[[n]][not.valid.2] <- NULL
                    }
                }
            }
            return(argList)
        },
        biotypeFilter = {
            switch(method,
                hg18 = {
                    valid <- names(argList) %in% c("unprocessed_pseudogene",
                        "pseudogene","miRNA","retrotransposed","protein_coding",
                        "processed_pseudogene","snRNA","snRNA_pseudogene",
                        "Mt_tRNA_pseudogene","miRNA_pseudogene","misc_RNA",
                        "tRNA_pseudogene","snoRNA","scRNA_pseudogene",
                        "rRNA_pseudogene","snoRNA_pseudogene","rRNA",
                        "misc_RNA_pseudogene","IG_V_gene","IG_D_gene",
                        "IG_J_gene","IG_C_gene","IG_pseudogene","scRNA")
                    not.valid <- which(!valid)
                },
                hg19 = {
                    valid <- names(argList) %in% c("pseudogene","lincRNA",
                        "protein_coding","antisense","processed_transcript",
                        "snRNA","sense_intronic","miRNA","misc_RNA","snoRNA",
                        "rRNA","polymorphic_pseudogene","sense_overlapping",
                        "three_prime_overlapping_ncrna","TR_V_gene",
                        "TR_V_pseudogene","TR_D_gene","TR_J_gene","TR_C_gene",
                        "TR_J_pseudogene","IG_C_gene","IG_C_pseudogene",
                        "IG_J_gene","IG_J_pseudogene","IG_D_gene","IG_V_gene",
                        "IG_V_pseudogene")
                    not.valid <- which(!valid)
                },
                mm9 = {
                    valid <- names(argList) %in% c("pseudogene","snRNA",
                        "protein_coding","antisense","miRNA","lincRNA","snoRNA",
                        "processed_transcript","misc_RNA","rRNA",
                        "sense_overlapping","sense_intronic",
                        "polymorphic_pseudogene","non_coding",
                        "three_prime_overlapping_ncrna","IG_C_gene","IG_J_gene",
                        "IG_D_gene","IG_V_gene","ncrna_host")
                    not.valid <- which(!valid)
                },
                mm10 = {
                    valid <- names(argList) %in% c("pseudogene","snRNA",
                        "protein_coding","antisense","miRNA","snoRNA","lincRNA",
                        "processed_transcript","misc_RNA","rRNA",
                        "sense_intronic","sense_overlapping",
                        "polymorphic_pseudogene","IG_C_gene","IG_J_gene",
                        "IG_D_gene","IG_LV_gene","IG_V_gene","IG_V_pseudogene",
                        "TR_V_gene","TR_V_pseudogene",
                        "three_prime_overlapping_ncrna")
                    not.valid <- which(!valid)
                },
                dm3 = {
                    valid <- names(argList) %in% c("protein_coding","ncRNA",
                        "snoRNA","pre_miRNA","pseudogene","snRNA","tRNA","rRNA")
                    not.valid <- which(!valid)
                },
                rn5 = {
                    valid <- names(argList) %in% c("protein_coding",
                        "pseudogene","processed_pseudogene","miRNA","rRNA",
                        "misc_RNA")
                    not.valid <- which(!valid)
                },
                rn6 = {
                    valid <- names(argList) %in% c("antisense","lincRNA",
                        "miRNA","misc_RNA","Mt_rRNA","Mt_tRNA",
                        "processed_pseudogene","processed_transcript",
                        "protein_coding","pseudogene","ribozyme","rRNA",
                        "scaRNA","sense_intronic","snoRNA","snRNA","sRNA",
                        "TEC","transcribed_processed_pseudogene",
                        "transcribed_unprocessed_pseudogene",
                        "unprocessed_pseudogene")
                    not.valid <- which(!valid)
                },
                danrer7 = {
                    valid <- names(argList) %in% c("antisense",
                        "protein_coding","miRNA","snoRNA","rRNA","lincRNA",
                        "processed_transcript","snRNA","pseudogene",
                        "sense_intronic","misc_RNA","polymorphic_pseudogene",
                        "IG_V_pseudogene","IG_C_pseudogene","IG_J_pseudogene",
                        "non_coding","sense_overlapping")
                    not.valid <- which(!valid)
                },
                pantro4 = {
                    valid <- names(argList) %in% c("protein_coding",
                        "pseudogene","processed_pseudogene","miRNA","rRNA",
                        "snRNA","snoRNA","misc_RNA")
                    not.valid <- which(!valid)
                },
                susscr3 = {
                    valid <- names(argList) %in% c("antisense",
                        "protein_coding","lincRNA","pseudogene",
                        "processed_transcript","miRNA","rRNA","snRNA","snoRNA",
                        "misc_RNA","non_coding","IG_C_gene","IG_J_gene",
                        "IG_V_gene","IG_V_pseudogene")
                    not.valid <- which(!valid)
                },
                tair10 = {
                    valid <- names(argList) %in% c("miRNA","ncRNA",
                        "protein_coding","pseudogene","rRNA","snoRNA",
                        "snRNA","transposable_element","tRNA")
                    not.valid <- which(!valid)
                },
                equcab3 = {
                    valid <- names(argList) %in% c("miRNA","misc_RNA",
                        "protein_coding","pseudogene","processed_pseudogene",
                        "rRNA","snoRNA","snRNA")
                }
            )
            if (length(not.valid)>0) {
                warnwrap(paste("The following",method,what,"argument names",
                    "are invalid and will be ignored:",
                    paste(names(argList)[not.valid],collapse=", ")))
                argList[not.valid] <- NULL
            }
            return(argList)
        }
    )
}

getStrictBiofilter <- function(org) {
    switch(org,
        hg18 = {
            return(list(
                unprocessed_pseudogene=TRUE,
                pseudogene=TRUE,
                miRNA=FALSE,
                retrotransposed=FALSE,
                protein_coding=FALSE,
                processed_pseudogene=TRUE,
                snRNA=FALSE,
                snRNA_pseudogene=TRUE,
                Mt_tRNA_pseudogene=TRUE,
                miRNA_pseudogene=TRUE,
                misc_RNA=TRUE,
                tRNA_pseudogene=TRUE,
                snoRNA=TRUE,
                scRNA_pseudogene=TRUE,
                rRNA_pseudogene=TRUE,
                snoRNA_pseudogene=TRUE,
                rRNA=TRUE,
                misc_RNA_pseudogene=TRUE,
                IG_V_gene=FALSE,
                IG_D_gene=FALSE,
                IG_J_gene=FALSE,
                IG_C_gene=FALSE,
                IG_pseudogene=TRUE,
                scRNA=FALSE
            ))
        },
        hg19 = {
            return(list(
                pseudogene=TRUE,
                lincRNA=FALSE,
                protein_coding=FALSE,
                antisense=FALSE,
                processed_transcript=FALSE,
                snRNA=FALSE,
                sense_intronic=FALSE,
                miRNA=FALSE,
                misc_RNA=FALSE,
                snoRNA=TRUE,
                rRNA=TRUE,
                polymorphic_pseudogene=TRUE,
                sense_overlapping=FALSE,
                three_prime_overlapping_ncrna=FALSE,
                TR_V_gene=FALSE,
                TR_V_pseudogene=TRUE,
                TR_D_gene=FALSE,
                TR_J_gene=FALSE,
                TR_C_gene=FALSE,
                TR_J_pseudogene=TRUE,
                IG_C_gene=FALSE,
                IG_C_pseudogene=TRUE,
                IG_J_gene=FALSE,
                IG_J_pseudogene=TRUE,
                IG_D_gene=FALSE,
                IG_V_gene=FALSE,
                IG_V_pseudogene=TRUE
            ))
        },
        hg38 = {
            return(list(
                protein_coding=FALSE,
                polymorphic_pseudogene=TRUE,
                lincRNA=FALSE,
                unprocessed_pseudogene=TRUE,
                processed_pseudogene=TRUE,
                antisense=FALSE,
                processed_transcript=FALSE,
                transcribed_unprocessed_pseudogene=TRUE,
                sense_intronic=FALSE,
                unitary_pseudogene=TRUE,
                IG_V_gene=FALSE,
                IG_V_pseudogene=TRUE,
                TR_V_gene=FALSE,
                sense_overlapping=FALSE,
                transcribed_processed_pseudogene=TRUE,
                miRNA=FALSE,
                snRNA=FALSE,
                misc_RNA=FALSE,
                rRNA=TRUE,
                snoRNA=TRUE,
                IG_J_pseudogene=TRUE,
                IG_J_gene=FALSE,
                IG_D_gene=FALSE,
                three_prime_overlapping_ncrna=FALSE,
                IG_C_gene=FALSE,
                IG_C_pseudogene=TRUE,
                pseudogene=TRUE,
                TR_V_pseudogene=TRUE,
                Mt_tRNA=TRUE,
                Mt_rRNA=TRUE,
                translated_processed_pseudogene=TRUE,
                TR_J_gene=FALSE,
                TR_C_gene=FALSE,
                TR_D_gene=FALSE,
                TR_J_pseudogene=TRUE,
                LRG_gene=FALSE
            ))
        },
        mm9 = {
            return(list(
                pseudogene=TRUE,
                snRNA=FALSE,
                protein_coding=FALSE,
                antisense=FALSE,
                miRNA=FALSE,
                lincRNA=FALSE,
                snoRNA=TRUE,
                processed_transcript=FALSE,
                misc_RNA=TRUE,
                rRNA=TRUE,
                sense_overlapping=FALSE,
                sense_intronic=FALSE,
                polymorphic_pseudogene=TRUE,
                non_coding=FALSE,
                three_prime_overlapping_ncrna=FALSE,
                IG_C_gene=FALSE,
                IG_J_gene=FALSE,
                IG_D_gene=FALSE,
                IG_V_gene=FALSE,
                ncrna_host=FALSE
            ))
        },
        mm10 = {
            return(list(
                pseudogene=TRUE,
                snRNA=FALSE,
                protein_coding=FALSE,
                antisense=FALSE,
                miRNA=FALSE,
                snoRNA=TRUE,
                lincRNA=FALSE,
                processed_transcript=FALSE,
                misc_RNA=TRUE,
                rRNA=TRUE,
                sense_intronic=FALSE,
                sense_overlapping=FALSE,
                polymorphic_pseudogene=TRUE,
                IG_C_gene=FALSE,
                IG_J_gene=FALSE,
                IG_D_gene=FALSE,
                IG_LV_gene=FALSE,
                IG_V_gene=FALSE,
                IG_V_pseudogene=TRUE,
                TR_V_gene=FALSE,
                TR_V_pseudogene=TRUE,
                three_prime_overlapping_ncrna=FALSE
            ))
        },
        dm3 = {
            return(list(
                protein_coding=FALSE,
                ncRNA=FALSE,
                snoRNA=TRUE,
                pre_miRNA=FALSE,
                pseudogene=TRUE,
                snRNA=FALSE,
                tRNA=FALSE,
                rRNA=TRUE
            ))
        },
        rn5 = {
            return(list(
                protein_coding=FALSE,
                pseudogene=TRUE,
                processed_pseudogene=FALSE,
                miRNA=FALSE,
                rRNA=TRUE,
                misc_RNA=TRUE
            ))
        },
        rn6 = {
            return(list(
                antisense=FALSE, 
                lincRNA=FALSE, 
                miRNA=FALSE, 
                misc_RNA=TRUE, 
                Mt_rRNA=TRUE, 
                Mt_tRNA=TRUE, 
                processed_pseudogene=FALSE, 
                processed_transcript=FALSE, 
                protein_coding=FALSE, 
                pseudogene=FALSE, 
                ribozyme=FALSE, 
                rRNA=TRUE, 
                scaRNA=FALSE, 
                sense_intronic=FALSE, 
                snoRNA=TRUE, 
                snRNA=FALSE, 
                sRNA=FALSE, 
                TEC=FALSE, 
                transcribed_processed_pseudogene=FALSE, 
                transcribed_unprocessed_pseudogene=FALSE, 
                unprocessed_pseudogene=TRUE
            ))
        },
        danrer7 = {
            return(list(
                antisense=FALSE,
                protein_coding=FALSE,
                miRNA=FALSE,
                snoRNA=TRUE,
                rRNA=TRUE,
                lincRNA=FALSE,
                processed_transcript=FALSE,
                snRNA=FALSE,
                pseudogene=TRUE,
                sense_intronic=FALSE,
                misc_RNA=TRUE,
                polymorphic_pseudogene=TRUE,
                IG_V_pseudogene=TRUE,
                IG_C_pseudogene=TRUE,
                IG_J_pseudogene=TRUE,
                non_coding=FALSE,
                sense_overlapping=FALSE
            ))
        },
        pantro4 = {
            return(list(
                protein_coding=FALSE,
                pseudogene=TRUE,
                processed_pseudogene=TRUE,
                miRNA=FALSE,
                rRNA=TRUE,
                snRNA=TRUE,
                snoRNA=TRUE,
                misc_RNA=TRUE
            ))
        },
        susscr3 = {
            return(list(
                antisense=FALSE,
                protein_coding=FALSE,
                lincRNA=FALSE,
                pseudogene=TRUE,
                processed_transcript=FALSE,
                miRNA=FALSE,
                rRNA=TRUE,
                snRNA=TRUE,
                snoRNA=TRUE,
                misc_RNA=TRUE,
                non_coding=FALSE,
                IG_C_gene=TRUE,
                IG_J_gene=TRUE,
                IG_V_gene=TRUE,
                IG_V_pseudogene=FALSE
            ))
        },
        tair10 = {
            return(list(
                miRNA=FALSE,
                ncRNA=FALSE,
                protein_coding=FALSE,
                pseudogene=TRUE,
                rRNA=TRUE,
                snoRNA=TRUE,
                snRNA=TRUE,
                transposable_element=FALSE,
                tRNA=TRUE
            ))
        },
        equcab3 = {
            return(list(
                miRNA=FALSE,
                misc_RNA=TRUE,
                protein_coding=FALSE,
                pseudogene=FALSE,
                processed_pseudogene=FALSE,
                rRNA=TRUE,
                snoRNA=TRUE,
                snRNA=TRUE
            ))
        }
    )
}

getPresetOpts <- function(preset,org) {
    # Override filter rules and maybe normArgs and statArgs
    switch(preset,
        all_basic = {
            exonFilters <- NULL
            geneFilters <- NULL
            pcut <- NA
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change")
            exportScale <- c("natural","log2")
            exportValues <- c("normalized")
            exportStats <- c("mean")
        },
        all_normal = {
            exonFilters <- NULL
            geneFilters <- NULL
            pcut <- NA
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change","stats",
                "counts")
            exportScale <- c("natural","log2")
            exportValues <- c("normalized")
            exportStats <- c("mean","sd","cv")
        },
        all_full = {
            exonFilters <- NULL
            geneFilters <- NULL
            pcut <- NA
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change","stats",
                "counts","flags")
            exportScale <- c("natural","log2","log10","vst")
            exportValues <- c("raw","normalized")
            exportStats <- c("mean","median","sd","mad","cv","rcv")
        },
        medium_basic = {
            exonFilters <- list(
                minActiveExons=list(
                    exonsPerGene=5,
                    minExons=2,
                    frac=1/5
                )
            )
            geneFilters <- list(
                length=list(
                    length=500
                ),
                avgReads=list(
                    averagePerBp=100,
                    quantile=0.25
                ),
                expression=list(
                    median=TRUE,
                    mean=FALSE,
                    quantile=NA,
                    known=NA,
                    custom=NA
                ),
                biotype=getDefaults("biotypeFilter",org[1])
            )
            pcut <- 0.05
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change")
            exportScale <- c("natural","log2")
            exportValues <- c("normalized")
            exportStats <- c("mean")
        },
        medium_normal = {
            exonFilters <- list(
                minActiveExons=list(
                    exonsPerGene=5,
                    minExons=2,
                    frac=1/5
                )
            )
            geneFilters <- list(
                length=list(
                    length=500
                ),
                avgReads=list(
                    averagePerBp=100,
                    quantile=0.25
                ),
                expression=list(
                    median=TRUE,
                    mean=FALSE,
                    quantile=NA,
                    known=NA,
                    custom=NA
                ),
                biotype=getDefaults("biotypeFilter",org[1])
            )
            pcut <- 0.05
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change","stats",
                "counts")
            exportScale <- c("natural","log2")
            exportValues <- c("normalized")
            exportStats <- c("mean","sd","cv")
        },
        medium_full = {
            exonFilters <- list(
                minActiveExons=list(
                    exonsPerGene=5,
                    minExons=2,
                    frac=1/5
                )
            )
            geneFilters <- list(
                length=list(
                    length=500
                ),
                avgReads=list(
                    averagePerBp=100,
                    quantile=0.25
                ),
                expression=list(
                    median=TRUE,
                    mean=FALSE,
                    quantile=NA,
                    known=NA,
                    custom=NA
                ),
                biotype=getDefaults("biotypeFilter",org[1])
            )
            pcut <- 0.05
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change","stats",
                "counts","flags")
            exportScale <- c("natural","log2","log10","vst")
            exportValues <- c("raw","normalized")
            exportStats <- c("mean","median","sd","mad","cv","rcv")
        },
        strict_basic = {
            exonFilters=list(
                minActiveExons=list(
                    exons_per_gene=4,
                    min_exons=2,
                    frac=1/4
                )
            )
            geneFilters=list(
                length=list(
                    length=750
                ),
                avgReads=list(
                    averagePerBp=100,
                    quantile=0.5
                ),
                expression=list(
                    median=TRUE,
                    mean=FALSE,
                    quantile=NA,
                    known=NA,
                    custom=NA
                ),
                biotype=getStrictBiofilter(org[1])
            )
            pcut <- 0.01
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change")
            exportScale <- c("natural","log2")
            exportValues <- c("normalized")
            exportStats <- c("mean")
        },
        strict_normal = {
            exonFilters=list(
                minActiveExons=list(
                    exonsPerGene=4,
                    minExons=2,
                    frac=1/4
                )
            )
            geneFilters=list(
                length=list(
                    length=750
                ),
                avgReads=list(
                    averagePerBp=100,
                    quantile=0.5
                ),
                expression=list(
                    median=TRUE,
                    mean=FALSE,
                    quantile=NA,
                    known=NA,
                    custom=NA
                ),
                biotype=getStrictBiofilter(org[1])
            )
            pcut <- 0.01
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change","stats",
                "counts")
            exportScale <- c("natural","log2")
            exportValues <- c("normalized")
            exportStats <- c("mean","sd","cv")
        },
        strict_full = {
            exonFilters=list(
                minActiveExons=list(
                    exonsPerGene=4,
                    minExons=2,
                    frac=1/4
                )
            )
            geneFilters=list(
                length=list(
                    length=750
                ),
                avgReads=list(
                    averagePerBp=100,
                    quantile=0.5
                ),
                expression=list(
                    median=TRUE,
                    mean=FALSE,
                    quantile=NA,
                    known=NA,
                    custom=NA
                ),
                biotype=getStrictBiofilter(org[1])
            )
            pcut <- 0.01
            exportWhat <- c("annotation","p_value","adj_p_value",
                "meta_p_value","adj_meta_p_value","fold_change","stats",
                "counts","flags")
            exportScale <- c("natural","log2","log10","vst")
            exportValues <- c("raw","normalized")
            exportStats <- c("mean","median","sd","mad","cv","rcv")
        }
    )
    presetOpts <- list(
        exonFilters=exonFilters,
        geneFilters=geneFilters,
        pcut=pcut,
        exportWhat=exportWhat,
        exportScale=exportScale,
        exportValues=exportValues,
        exportStats=exportStats
    )
    return(presetOpts)
}

makeFoldChange <- function(contrast,sampleList,dataMatrix,logOffset=1) {
    conds <- strsplit(contrast,"_vs_")[[1]]
    foldMat <- matrix(0,nrow(dataMatrix),length(conds)-1)
    for (i in seq_len(length(conds)-1)) { # Last condition is ALWAYS reference
    #for (i in 2:length(conds)) { # First condition is ALWAYS reference
        samplesNom <- sampleList[[conds[i]]]
        samplesDenom <- sampleList[[conds[length(conds)]]]
        #samplesDenom <- sampleList[[conds[1]]]
        nom <- dataMatrix[,match(samplesNom,colnames(dataMatrix)),drop=FALSE]
        denom <- dataMatrix[,match(samplesDenom,colnames(dataMatrix)),
            drop=FALSE]
        if (!is.matrix(nom)) 
            nom <- as.matrix(nom) # Cover the case with no replicates...
        if (!is.matrix(denom)) 
            denom <- as.matrix(denom)
        meanNom <- apply(nom,1,mean)
        meanDenom <- apply(denom,1,mean)
        #meanNom <- ifelse(meanNom==0,logOffset,meanNom)
        if (any(meanNom==0)) 
            meanNom <- meanNom + logOffset
        #meanDenom <- ifelse(meanDenom==0,logOffset,meanDenom)
        if (any(meanDenom==0)) 
            meanDenom <- meanDenom + logOffset
        #foldMat[,i-1] <- meanNom/meanDenom
        foldMat[,i] <- meanNom/meanDenom
    }
    rownames(foldMat) <- rownames(dataMatrix)
    #colnames(foldMat) <- paste(conds[1],"_vs_",conds[2:length(conds)],sep="")
    colnames(foldMat) <- paste(conds[seq_len(length(conds)-1)],"_vs_",
        conds[length(conds)],sep="")
    return(foldMat)
}

makeA <- function(contrast,sampleList,dataMatrix,logOffset=1) {
    conds <- strsplit(contrast,"_vs_")[[1]]
    aMat <- matrix(0,nrow(dataMatrix),length(conds)-1)
    for (i in seq_len(length(conds)-1)) { # Last condition is ALWAYS reference
    #for (i in 2:length(conds)) { # First condition is ALWAYS reference
        samplesTrt <- sampleList[[conds[i]]]
        samplesCnt <- sampleList[[conds[length(conds)]]]
        #samplesCnt <- sampleList[[conds[1]]]
        trt <- dataMatrix[,match(samplesTrt,colnames(dataMatrix)),drop=FALSE]
        cnt <- dataMatrix[,match(samplesCnt,colnames(dataMatrix)),drop=FALSE]
        meanTrt <- apply(trt,1,mean)
        meanCnt <- apply(cnt,1,mean)
        if (any(meanTrt==0)) 
            meanTrt <- meanTrt + logOffset
        if (any(meanCnt==0)) 
            meanCnt <- meanCnt + logOffset
        aMat[,i] <- 0.5*(log2(meanTrt)+log2(meanCnt))
        #aMat[,i-1] <- 0.5*(log2(meanTrt)+log2(meanCnt))
    }
    rownames(aMat) <- rownames(dataMatrix)
    colnames(aMat) <- paste(conds[seq_len(length(conds)-1)],"_vs_",
        conds[length(conds)],sep="")
    #colnames(aMat) <- paste(conds[2:length(conds)],"_vs_",conds[1],sep="")
    return(aMat)
}

makeAvgExpression <- function(contrast,sampleList,dataMatrix,logOffset=1) {
    conds <- strsplit(contrast,"_vs_")[[1]]
    aMat <- matrix(0,nrow(dataMatrix),length(conds)-1)
    for (i in seq_len(length(conds)-1)) { # Last condition is ALWAYS reference
    #for (i in 2:length(conds)) { # First condition is ALWAYS reference
        samplesNom <- sampleList[[conds[i]]]
        #samplesDenom <- sampleList[[conds[1]]]
        samplesDenom <- sampleList[[conds[length(conds)]]]
        nom <- dataMatrix[,match(samplesNom,colnames(dataMatrix)),drop=FALSE]
        denom <- dataMatrix[,match(samplesDenom,colnames(dataMatrix)),
            drop=FALSE]
        if (!is.matrix(nom)) # Cover the case with no replicates...
            nom <- as.matrix(nom) 
        if (!is.matrix(denom)) 
            denom <- as.matrix(denom)
        meanNom <- apply(nom,1,mean)
        meanDenom <- apply(denom,1,mean)
        if (any(meanNom==0)) 
            meanNom <- meanNom + logOffset
        if (any(meanDenom==0)) 
            meanDenom <- meanDenom + logOffset
        aMat[,i] <- 0.5*(log2(meanNom)+log2(meanDenom))
        #aMat[,i-1] <- 0.5*(log2(meanNom)+log2(meanDenom))
    }
    rownames(aMat) <- rownames(dataMatrix)
    #colnames(aMat) <- paste(conds[1],"_vs_",conds[2:length(conds)],sep="")
    colnames(aMat) <- paste(conds[seq_len(length(conds)-1)],"_vs_",
        conds[length(conds)],sep="")
    return(aMat)
}

.makeHtmlCells <- function(mat,type="numeric",digits=3) {
    if (type=="numeric")
        tmp <- format(mat,digits=digits)
        #tmp <- formatC(mat,digits=digits,format="f")
    else
        tmp <- mat
    if (!is.matrix(tmp)) tmp <- as.matrix(tmp)
    tmp <- apply(tmp,c(1,2),function(x) paste("<td>",x,"</td>",sep=""))
    return(tmp)
}

.makeHtmlRows <- function(mat) {
    tmp <- apply(mat,1,paste,collapse="")
    tmp <- paste("<tr>",tmp,"</tr>",sep="")
    return(tmp)
}

.makeHtmlHeader <- function(h) {
    tmp <- paste("<th>",h,"</th>",sep="")
    tmp <- paste(tmp,collapse="")
    tmp <- paste("<tr>",tmp,"</tr>",sep="")
    return(tmp)
}

.makeHtmlBody <- function(mat) {
    tmp <- paste(mat,collapse="")
    return(tmp)
}

.makeHtmlTable <- function(b,h=NULL,id=NULL) {
    if (!is.null(id))
        html <- paste("<table id=\"",id,"\" class=\"datatable\">",sep="")
    else
        html <- "<table class=\"datatable\">"
    if (!is.null(h))
        html <- paste(html,"<thead>",h,"</thead>",sep="")
    html <- paste(html,"<tbody>",b,"</tbody></table>",sep="")
    return(html)
}    

makeTransformation <- function(dataMatrix,exportScale,
    scf=NULL,logOffset=1) {
    mat <- vector("list",length(exportScale))
    names(mat) <- exportScale
    if (!is.matrix(dataMatrix)) dataMatrix <- as.matrix(dataMatrix)
    if (is.null(scf)) scf <- rep(1,nrow(dataMatrix))
    for (scl in exportScale) {
        switch(scl,
            natural = {
                mat[[scl]] <- dataMatrix
            },
            log2 = {
                mat[[scl]] <- nat2log(dataMatrix,base=2,off=logOffset)
            },
            log10 = {
                mat[[scl]] <- nat2log(dataMatrix,base=10,off=logOffset)
            },
            vst = {
                fit <- vsn2(dataMatrix,verbose=FALSE)
                mat[[scl]] <- predict(fit,newdata=dataMatrix)
            },
            rpgm = {
                mat[[scl]] <- dataMatrix
                for (i in seq_len(ncol(dataMatrix)))
                    mat[[scl]][,i] <- dataMatrix[,i]/scf
            }
        )
    }
    return(mat)
}

makeStat <- function(samples,dataList,stat,exportScale) {
    statResult <- vector("list",length(exportScale))
    names(statResult) <- exportScale
    for (scl in exportScale) {
        statData <- dataList[[scl]][,match(samples,colnames(dataList[[scl]]))]
        if (!is.matrix(statData))
            statData <- as.matrix(statData)
        switch(stat,
            mean = {
                statResult[[scl]] <- apply(statData,1,function(x,s) {
                    if (s=="natural")
                        return(round(mean(x)))
                    else
                        return(mean(x))
                },scl)
            },
            median = {
                statResult[[scl]] <- apply(statData,1,function(x,s) {
                    if (s=="natural")
                        return(round(median(x)))
                    else
                        return(median(x))
                },scl)
            },
            sd = {
                statResult[[scl]] <- apply(statData,1,function(x,s) {
                    if (s=="natural")
                        return(ceiling(sd(x)))
                    else
                        return(sd(x))
                },scl)
            },
            mad = {
                statResult[[scl]] <- apply(statData,1,function(x,s) {
                    if (s=="natural")
                        return(ceiling(mad(x)))
                    else
                        return(mad(x))
                },scl)
            },
            cv = {
                statResult[[scl]] <- apply(statData,1,function(x,s) {
                    if (s=="natural")
                        return(ceiling(sd(x))/round(mean(x)))
                    else
                        return(sd(x)/mean(x))
                },scl)
            },
            rcv = {
                statResult[[scl]] <- apply(statData,1,function(x,s) {
                    if (s=="natural")
                        return(ceiling(mad(x))/round(median(x))) 
                    else
                        return(mad(x)/median(x))
                },scl)
            }
        )
    }
    return(do.call("cbind",statResult))
}

makeMatrix <- function(samples,dataList,exportScale="natural") {
    mat <- vector("list",length(exportScale))
    names(mat) <- exportScale
    for (scl in exportScale) {
        mat.data <- dataList[[scl]][,match(samples,
            colnames(dataList[[scl]]))]
        if (!is.matrix(mat.data)) {
            mat.data <- as.matrix(mat.data)
            colnames(mat.data) <- samples
        }
        mat[[scl]] <- mat.data
    }
    return(do.call("cbind",mat))
}

makeContrastList <- function(contrast,sampleList) {
    # Construction
    contrastList <- vector("list",length(contrast))
    names(contrastList) <- contrast
    # First break the contrast vector
    cnts <- strsplit(contrast,"_vs_")
    names(cnts) <- names(contrastList)
    # Create list members
    for (n in names(contrastList)) {
        contrastList[[n]] <- vector("list",length(cnts[[n]]))
        for (i in seq_len(length(cnts[[n]]))) {
            contrastList[[n]][[i]] <- rep(cnts[[n]][i],
                length(sampleList[[cnts[[n]][i]]]))
            names(contrastList[[n]][[i]]) <- sampleList[[cnts[[n]][[i]]]]
        }
    }
    return(contrastList)
}

makeSampleList <- function(input,type=c("simple","targets")) {
    if (missing(input) || !file.exists(input))
        stopwrap("File to make sample list from should be a valid existing ",
            "text file!")
    type <- tolower(type[1])
    checkTextArgs("type",type,c("simple","targets"),multiarg=FALSE)
    tab <- read.delim(input)
    samples <- as.character(tab[,1])
    if (type=="simple") {
        ii <- 2
        conditions <- unique(as.character(tab[,ii]))
    }
    else if (type=="targets") {
        ii <- 3
        conditions <- unique(as.character(tab[,ii]))
    }
    if (length(samples) != length(unique(samples)))
        stopwrap("Sample names must be unique for each sample!")
    sampleList <- vector("list",length(conditions))
    names(sampleList) <- conditions
    for (n in conditions)
        sampleList[[n]] <- samples[which(as.character(tab[,ii])==n)]
    return(sampleList)
}

makeProjectPath <- function(path,f=NULL) {
    if (is.na(path) || is.null(path)) {
        if (!is.data.frame(f) && !is.null(f) && !is.list(f) && file.exists(f)) {
            if (length(grep(".RData$",f))>0)
                mainPath <- file.path(getwd(),paste("metaseqr_result_",
                    format(Sys.time(),format="%Y%m%d%H%M%S"),sep=""))
            else
                mainPath <- file.path(dirname(f),paste("metaseqr_result_",
                    format(Sys.time(),format="%Y%m%d%H%M%S"),sep=""))
        }
        else
            mainPath <- file.path(getwd(),paste("metaseqr_result_",
                format(Sys.time(),format="%Y%m%d%H%M%S"),sep=""))
        projectPath <- makePathStruct(mainPath)
    }
    else {
        success <- tryCatch(
            if (!file.exists(path)) dir.create(path,recursive=TRUE) else TRUE,
            error=function(e) {
                disp("Cannot create ",path,"! Is it a valid system path? Is ",
                    "there a write permissions problem? Reverting to ",
                    "automatic creation...")
                return(FALSE)
            },
            finally=""
        )
        if (success)
            projectPath <- makePathStruct(path)
        else
            projectPath <- makeProjectPath(NA,f)
    }
    return(projectPath)
}

makePathStruct <- function(mainPath) {
    projectPath <- list(
        main=mainPath,
        js=file.path(mainPath,"js"),
        media=file.path(mainPath,"media"),
        data=file.path(mainPath,"data"),
        logs=file.path(mainPath,"logs"),
        lists=file.path(mainPath,"lists"),
        tracks=file.path(mainPath,"tracks"),
        plots=file.path(mainPath,"plots"),
        qc=file.path(mainPath,"plots","qc"),
        normalization=file.path(mainPath,"plots","normalization"),
        statistics=file.path(mainPath,"plots","statistics")
    )
    for (p in names(projectPath))
        if (!file.exists(projectPath[[p]]))
            dir.create(projectPath[[p]],recursive=TRUE)
    return(projectPath)
}

makeExportList <- function(con) {
    if (!is.null(con)) {
        f <- vector("list",length(con))
        names(f) <- con
    }
    else
        f <- list(NT=NULL)
    return(f)
}

makeGrid <- function(n) {
    m <- 0
    while (n > m*m)
        m <- m+1
    if (n < m*m) {
        k <- m-1
        if (n > m*k)
            k <- k+1
        else {
            while (n > m*k)
                k <- k-1
        }
    }
    else
        k <- m
    return(c(m,k))
}

makeReportMessages <- function(lang) {
    switch(lang,
        en = {
            messages <- list(
                org=list(
                    hg18=paste("human (<em>Homo sapiens</em>),",
                        "genome version alias hg18"),
                    hg19=paste("human (<em>Homo sapiens</em>),",
                        "genome version alias hg19"),
                    hg38=paste("human (<em>Homo sapiens</em>),",
                        "genome version alias hg38"),
                    mm9=paste("mouse (<em>Mus musculus</em>),",
                        "genome version alias mm9"),
                    mm10=paste("mouse (<em>Mus musculus</em>),",
                        "genome version alias mm10"),
                    rn5=paste("rat (<em>Rattus norvegicus</em>),",
                        "genome version  alias rn5"),
                    rn6=paste("rat (<em>Rattus norvegicus</em>),",
                        "genome version alias rn6"),
                    dm3=paste("fruitfly (<em>Drosophila melanogaster</em>),",
                        "genome version alias dm3"),
                    dm6=paste("fruitfly (<em>Drosophila melanogaster</em>),",
                        "genome version alias dm6"),
                    danrer7=paste("zebrafish (<em>Danio rerio</em>),",
                        "genome version alias danRer7"),
                    danrer10=paste("zebrafish (<em>Danio rerio</em>),",
                        "genome version alias danRer10"),
                    danrer11=paste("zebrafish (<em>Danio rerio</em>),",
                        "genome version alias danRer11"),
                    pantro4=paste("chimpanzee (<em>Pan troglodytes</em>),",
                        "genome version alias panTro4"),
                    pantro5=paste("chimpanzee (<em>Pan troglodytes</em>),",
                        "genome version alias panTro5"),
                    susscr3=paste("pig (<em>Sus scrofa</em>),",
                        "genome version alias susScr3"),
                    susscr11=paste("pig (<em>Sus scrofa</em>),",
                        "genome version alias susScr11"),
                    equcab3=paste("horse (<em>Equus cabalus</em>),",
                        "genome version alias equcab3"),
                    tair10=paste("arabidopsis (<em>Arabidobsis thaliana</em>)",
                        ",","genome version alias TAIR10")
                ),
                refdb=list(
                    ensembl="Ensembl genomes",
                    ucsc="UCSC genomes database",
                    refseq="RefSeq database",
                    other="User provided"
                ),
                whenfilter=list(
                    prenorm="before normalization",
                    postnorm="after normalization"
                ),
                norm=list(
                    edaseq="EDASeq",
                    deseq="DESeq",
                    deseq2="DESeq2",
                    edger="edgeR",
                    noiseq="NOISeq",
                    nbpseq="NBPSeq",
                    absseq="ABSSeq",
                    dss="DSS",
                    each="the same as the corresponding statistical test"
                ),
                stat=list(
                    deseq="DESeq",
                    deseq2="DESeq2",
                    edger="edgeR",
                    noiseq="NOISeq",
                    #bayseq="baySeq",
                    limma="limma",
                    nbpseq="NBPSeq",
                    absseq="ABSSeq",
                    dss="DSS"
                ),
                meta=list(
                    intersection="intersection of individual results",
                    union="union of individual results",
                    fisher="Fisher's method",
                    fperm="Fisher's method with permutations",
                    dperm_min=paste("samples permutation based method with",
                        "minimum p-values"),
                    dperm_max=paste("samples permutation based method with",
                        "maximum p-values"),
                    dperm_weight=paste("samples permutation based method with",
                        "weighted p-values"),
                    minp="minimum p-value across results",
                    maxp="maximum p-value across results",
                    weight="PANDORA weighted p-value across results",
                    pandora="PANDORA weighted p-value across results",
                    simes="Simes correction and combination method",
                    whitlock=paste("Whitlock's Z-transformation method",
                        "(Bioconductor package survcomp)"),
                    harmonic="Wilson's Harmonic mean of p-values",
                    none=paste("no meta-analysis, reported p-values from the",
                        "first supplied statistical algorithm")
                ),
                adjust=list(
                    holm="Holm FWER",
                    hochberg="Hochberg DFR",
                    hommel="Hommel FWER",
                    bonferroni="Bonferroni FWER",
                    bh="Benjamini-Hochberg FDR",
                    by="Benjamini-Yekutiely FDR",
                    fdr="Benjamini-Hochberg FDR",
                    none="no multiple test correction",
                    qvalue="Storey-Tibshirani FDR"
                ),
                plots=list(
                    mds="multidimensional scaling",
                    biodetection="biotype detection",
                    countsbio="biotype counts",
                    saturation="sample and biotype saturation",
                    rnacomp="RNA composition",
                    boxplot="boxplots",
                    gcbias="GC-content bias",
                    lengthbias="transcript length bias",
                    meandiff="mean-difference plot",
                    meanvar="mean-variance plot",
                    deheatmap="DEG heatmap",
                    volcano="volcano plot",
                    biodist="DEG biotype detection",
                    filtered="filtered biotypes",
                    correl="correlation heatmap and correlogram",
                    pairwise="pairwise scatterplots between samples",
                    statvenn="statistical algorithm Venn diagrams",
                    foldvenn="fold change Venn diagrams",
                    deregulogram="deregulogram",
                    mastat="staistical significance MA plot"
                ),
                export=list(
                    annotation="Annotation",
                    p_value="p-value",
                    meta_p_value="Combined p-value",
                    adj_p_value="Adjusted p-value (FDR)",
                    adj_meta_p_value="Adjusted combined p-value (FDR)",
                    fold_change="Fold change",
                    stats="Statistics",
                    counts="Read counts",
                    natural="Natural scale",
                    log2="log2 scale",
                    log10="log10 scale",
                    vst="Variance stabilization transformation",
                    rpgm="Reads per Gene Model",
                    raw="Raw values",
                    normalized="Normalized values",
                    mean="Mean",
                    median="Median",
                    sd="Standard deviation",
                    mad="Median Absolute Deviation (MAD)",
                    cv="Coefficient of Variation",
                    rcv="Robust Coefficient of Variation"
                ),
                preset=list(
                    all_basic=paste("use all genes and export all genes and",
                        "basic annotation and statistics elements"),
                    all_normal=paste("use all genes and export all genes and",
                        "normal annotation and statistics elements"),
                    all_full=paste("use all genes and export all genes and all",
                        "available annotation and statistics elements"),
                    medium_basic=paste("apply a medium set of filters and",
                        "export statistically significant genes and basic",
                        "annotation and statistics elements"),
                    medium_normal=paste("apply a medium set of filters and",
                        "export statistically significant genes and normal",
                        "annotation and statistics elements"),
                    medium_full=paste("apply a medium set of filters and",
                        "export statistically significant genes and all",
                        "available annotation and statistics elements"),
                    strict_basic=paste("apply a strict set of filters and",
                        "export statistically significant genes and basic",
                        "annotation and statistics elements"),
                    strict_normal=paste("apply a medium set of filters and",
                        "export statistically significant genes and normal",
                        "annotation and statistics elements"),
                    strict_full=paste("apply a medium set of filters and",
                        "export statistically significant genes and all",
                        "available annotation and statistics elements")
                ),
                explain=list(
                    mds=paste(
                "Multidimensional Scaling (MDS) plots constitute a means",
                "of visualizing the level of similarity of individual cases",
                "of a dataset. It is similar to Principal Component Analysis",
                "(PCA), but instead of using the covariance matrix to find",
                "similarities between cases, MDS uses absolute distance",
                "metrics such as the classical Euclidean distance. Because",
                "of the relative linear relations between sequencing samples,",
                "it provides a more realistic clustering of samples. MDS",
                "serves quality control and it can be interpreted as follows:",
                "when the distance between samples of the same biological",
                "condition in the MDS space is small, this is an indication",
                "of high correlation and reproducibility between them. When",
                "this distance is larger or heterogeneous (e.g. the 3rd",
                "sample of a triplicate set is further from the other 2),",
                "this constitutes an indication of low correlation and",
                "reproducibility between samples. It can help exclude poor",
                "samples from further analysis.",collapse=" "
                    ),
                    biodetection=paste(
                "The biotype detection bar diagrams are a set of quality",
                "control charts that show the percentage of each biotype",
                "in the genome (i.e. in the whole set of features provided,",
                "for example, protein coding genes, non coding RNAs or",
                "pseudogenes) in red bars, the proportion of which has been",
                "detected in a sample before normalization and after",
                "basic filtering by removing features with zero counts in",
                "green bars, and the percentage of each biotype within the",
                "sample in blue bars. The difference between red bars and",
                "blue bars is that red bars show the percentage of a",
                "feature in the genome while blue bars show the percentage",
                "in the sample. Thus, the blue bars may sometimes be higher",
                "than the green bars because certain features (e.g. protein",
                "coding genes) may be detected within a sample with a higher",
                "proportion relative to their presence in the genome, as",
                "compared with other features. For example, while the",
                "percentage of protein coding genes in the whole genome is",
                "already higher than other biotypes, this percentage is",
                "expected to be even higher in an RNA-Seq experiment where one",
                "expects protein-coding genes to exhibit greater abundance.",
                "The vertical line separates the most abundant (yellow band)",
                "biotypes (on the left-hand side, corresponding to the",
                "left axis scale) from the rest (on the right-hand side,",
                "corresponding to the right axis scale, red band). Otherwise,",
                "lower abundance biotypes would be indistinguishable.",
                "Unexpected outcomes in this quality control chart (e.g.",
                "very low detection of protein coding genes) would signify",
                "possible low quality of a sample.",collapse=" "
                    ),
                    countsbio=paste(
                "Biotype detection counts boxplots are a set of quality",
                "control charts that depict both the biological classification",
                "for the detected features and the actual distribution of",
                "the read counts for each biological type. The boxplot",
                "comprises a means of summarizing the read counts distribution",
                "of a sample in the form of a bar with extending lines,",
                "as a commonly used way of graphically presenting groups of",
                "numerical data. A boxplot also indicates which observations,",
                "if any, might be considered outliers and is able to visually",
                "show different types of populations, without making any",
                "assumptions about the underlying statistical distribution.",
                "The spacing between the different parts of the box help",
                "indicate variance, skewness and identify outliers. The",
                "thick bar inside the colored box is the median of the",
                "observations while the box extends over the Interquartile",
                "Range of the observations. The whiskers extend up (down)",
                "to +/-1.5xIQR. Unexpected outcomes (e.g. protein coding",
                "read count distribution similar to pseudogene read count",
                "distribution) indicates poor sample quality.",collapse=" "
                    ),
                    saturation=paste(
                "Read and biotype saturation plots are a set of quality",
                "control charts that depict the read count saturation",
                "levels at several sequencing depths. Thus, they comprise",
                "a means of assessing whether the sequencing depth of an",
                "RNA-Seq experiment is sufficient in order to detect the",
                "biological features under investigation. These quality",
                "control charts are separated in two subgroups: the first",
                "(read saturation per biotype for all samples)",
                "is a set of plots, one for each biological feature (e.g.",
                "protein coding, pseudogene, lincRNA, etc.), that depict",
                "the number of detected features in different sequencing",
                "depths and for all samples in the same plot. The second",
                "subgroup (read saturation per sample for all biotypes)",
                "is a set of plots similar to the above, but with",
                "one pair of plots with two panels for each sample,",
                "presenting all biological features. The left panel depicts",
                "the saturation levels for the less abundatnt features,",
                "while the right panel, the saturation for the more abundant",
                "features, as placing them all together would make the",
                "less abundant features indistinguishable. All the saturation",
                "plots should be interpreted as follows: if the read counts",
                "for a biotype tend to be saturated, the respective curve",
                "should tend to reach a plateau at higher depths. Otherwise,",
                "more sequencing is needed for the specific biotype.",
                        collapse=" "
                    ),
                    readnoise=paste(
                "The read noise plots depict the percentage of biological",
                "features detected when subsampling the total number of",
                "reads. Very steep curves in read noise plots indicate",
                "that although the sequencing depth reaches its maximum,",
                "a relatively small percentage of total features is detected,",
                "indicating that the level of background noise is relatively",
                "high. Less steep RNA composition curves, indicate less noise.",
                "When a sample's curve deviate from the rest, it may",
                "indicate lower or higher quality, depending on the curves",
                "of the rest of the samples.",collapse=" "
                    ),
                    correl=paste(
                "Sample correlation plots depict the accordance of",
                "RNA-Seq samples, as this is manifested through the",
                "read counts table used with the metaseqR2 pipeline, with",
                "representations that both use the correlation matrix",
                "(a matrix which depicts all the pairwise correlations",
                "between each pair of samples) of the read counts matrix.",
                "The correlation representation is a clustered heatmap which",
                "depicts the correlations of samples as color-scaled",
                "images and the hierarchical clustering tree depicts the",
                "grouping of the samples according to their correlation.",
                "If samples from the same group not being clustered together",
                "provides an indication that there might be a quality",
                "problem with the dataset.",collapse=" "
                    ),
                    pairwise=paste(
                "Pairwise comparison plots are split in two parts:",
                "the upper part consists of a simple scatterplot for",
                "all pairwise sample comparisons. It is a simple measure",
                "of between sample correlation using all the available",
                "data points instead of only the correlation matrix. The",
                "lower part consists of mean-difference plots for all",
                "pairwise sample comparisons. A mean-difference plot (or",
                "a Bland-Altman plots) is a method of data plotting used",
                "in analyzing the agreement between two different",
                "assays/variables. In this graphical method the differences",
                "(or alternatively the ratios) between the two variables",
                "are plotted against the averages of the two. Such a plot",
                "is useful, for example, for analyzing data with strong",
                "correlation between the x and y axes, when the (x,y) dots on",
                "the plot are close to the diagonal x=y. In this case, the",
                "value of the transformed variable X is approximately the same",
                "as x and y and the variable Y shows the difference between",
                "x and y. In both represantations, irregular shapes of the",
                "red smoother lines are an indication of poor correlation",
                "between samples or of other systematic bias sources,",
                "which is usually corrected through data normalization.",
                collapse=" "
                    ),
                    rnacomp=paste(
                "The RNA composition plots depict differences in the",
                "distributions of reads in the same biological features",
                "across samples. The following is taken from the NOISeq",
                "vignette: <em>'...when two samples have different RNA",
                "composition, the distribution of sequencing reads across",
                "the features is different in such a way that although",
                "a feature had the same number of read counts in both",
                "samples, it would not mean that it was equally expressed",
                "in both... To check if this bias is present in the data,",
                "the RNA composition plot and the correponding diagnostic",
                "test can be used. In this case, each sample s is compared",
                "to the reference sample r (which can be arbitrarily",
                "chosen). To do that, M values are computed as",
                "log2(counts_sample = counts_reference). If no bias is",
                "present, it should be expected that the median of M",
                "values for each comparison is 0. Otherwise, it would be",
                "indicating that expression levels in one of the samples",
                "tend to be higher than in the other, and this could lead",
                "to false discoveries when computing differencial expression.",
                "Confidence intervals for the M median are also computed by",
                "bootstrapping. If value 0 does not fall inside the interval,",
                "it means that the deviation of the sample with regard",
                "to the reference sample is statistically significant.",
                "Therefore, a normalization procedure is required.'</em>",
                collapse=" "
                    ),
                boxplot=paste(
                "The boxplot comprises a means of summarizing the read",
                "counts distribution of a sample in the form of a bar",
                "with extending lines, as a commonly used way of",
                "graphically presenting groups of numerical data. A",
                "boxplot also indicates which observations, if any, might",
                "be considered outliers and is able to visually show",
                "different types of populations, without making any",
                "assumptions about the underlying statistical distribution.",
                "The spacings between the different parts of the box help",
                "indicate variance, skewness and identify outliers. The",
                "thick bar inside the colored box is the median of the",
                "observations while the box extends over the Interquartile",
                "Range of the observations. The whiskers extend up (down)",
                "to +/-1.5xIQR. Similar boxplots indicate good quality of",
                "normalization. If boxplots remain dissimilar after",
                "normalization, another normalization algorithm may have to be",
                "examined. The un-normalized boxplots show the need for data",
                "normalization in order for the data from different",
                "samples to follow the same underlying distribution and",
                "statistical testing to become possible.",collapse=" "
                ),
                    gcbias=paste(
                "The GC-content bias plot is a quality control chart that",
                "shows the possible dependence of the read counts (in log2",
                "scale) under a gene to the GC content percentage of that",
                "gene. In order for the statistical tests to be able to",
                "detect statistical significance which occurs due to real",
                "biological effects and not through other systematic biases",
                "present in the data (e.g. possible GC-content bias),",
                "the latter should be accounted for by the applied",
                "normalization algorithm. Although the tests are performed",
                "for each gene across biological conditions one could assume",
                "that the GC content does not represent a bias, as it is the",
                "same for the tested gene across samples and conditions.",
                "However, Risso et al. (2011) showed that GC-content could",
                "could have an impact on the statistical testing procedure.",
                "The GC-content bias plot depicts the dependence of the",
                "read counts to the GC content before and after normalization.",
                "The smoothing lines for each sample, should be as 'straight'",
                "as possible after normalization. In addition, if the",
                "smoothing lines differ significantly between biological",
                "conditions, this would constitute a possible quality warning.",
                collapse=" "
                    ),
                    lengthbias=paste(
                "The gene/transcript length bias plot is a quality control",
                "chart that shows the possible dependence of read counts (in",
                "log2 scale) under a gene to the length of that gene (whole",
                "gene or sum of exons depending on the analysis). In order",
                "for the statistical tests to be able to detect statistical",
                "significance which occurs due to real biological effects",
                "and not by other systematic biases present in the data",
                "(e.g. possible length bias), the latter should be accounted",
                "for by the applied normalization algorithm. Although the",
                "tests are performed for each gene across biological",
                "conditions, one could assume that the gene length does not",
                "represent a bias, as it is the same for the tested gene",
                "across samples and conditions. However, it has been shown in",
                "several studies that gene length could have an impact on the",
                "statistical testing procedure. The length bias plot",
                "depicts the dependence of the read counts to the",
                "gene/transcript length before and after normalization.",
                "The smoothing lines for each sample, should be as 'straight'",
                "as possible after normalization. In addition, if the",
                "smoothing lines differ significantly between biological",
                "conditions, this would constitute a possible quality warning.",
                collapse=" "
                    ),
                    meandiff=paste(
                "A mean-difference plot (or a Bland-Altman plot) is a",
                "method of data plotting used in analyzing the agreement",
                "between two different assays/variables. In this graphical",
                "method the differences (or alternatively the ratios)",
                "between the two variables are plotted against the averages",
                "of the two. Such a plot is useful, for example, for analyzing",
                "data with strong correlation between the x and y axes, when",
                "the (x,y) dots on the plot are close to the diagonal x=y.",
                "In this case, the value of the transformed variable X is",
                "approximately the same as x and y and variable Y shows the",
                "difference between x and y. When the data cloud in a mean",
                "difference plot is centered around the horizontal zero line,",
                "this is an indication of good data quality and good",
                "normalization results. On the other hand, when the data",
                "cloud deviates from the center line or has a 'banana'",
                "shape, this constitutes an indication of systematic biases",
                "present in the data and that either the chosen normalization",
                "algorithm has not worked well, or that data are not",
                "normalized. The smoothing curve that traverses the data",
                "(red curve) summarizes the above trends.",collapse=" "
                    ),
                    meanvar=paste(
                "The mean-variance plot comprises a graphical means of",
                "displaying a possible relationship between the means of",
                "gene expression (counts) values and their variances",
                "across replicates of a gene expression experiment. Thus",
                "data can be inspected for possible overdispersion (greater",
                "variability in a dataset than would be expected based on",
                "a given simple statistical model). In such plots for",
                "RNA-Seq data, overdispersion is usually manifested as",
                "increasing variance with increasing gene expression",
                "(counts) and it is summarized through a smoothing curve",
                "(red curve). The following is taken from the EDASeq package",
                "vignette: '<em>...although the Poisson distribution",
                "is a natural and simple way to model count data, it has",
                "the limitation of assuming equality of the mean and",
                "variance. For this reason, the negative binomial",
                "distribution has been proposed as an alternative when the",
                "data show over-dispersion...'</em> If overdispersion is",
                "not present, the data cloud is expected to be evenly",
                "scattered around the smoothing curve.",collapse=" "
                    ),
                    deheatmap=paste(
                    "Differentially Expressed Genes (DEGs) heatmaps depict",
                    "how well samples from different conditions cluster",
                    "together according to their expression values after",
                    "normalization and statistical testing, for each requested",
                    "statistical contrast. If samples from the same biological",
                    "condition do not cluster together, this would constitute",
                    "a warning sign regarding the quality of the samples. In",
                    "addition, DEG heatmaps provide an initial view of",
                    "possible clusters of co-expressed genes.",collapse=" "
                    ),
                    volcano=paste(
                "A volcano plot is a scatterplot that is often used when",
                "analyzing high-throughput -omics data (e.g. microarray",
                "data, RNA-Seq data) to give an overview of interesting",
                "genes. The log2 fold change is plotted on the x-axis and",
                "the negative log10 p-value is plotted on the y-axis. A",
                "volcano plot combines the results of a statistical test",
                "(aka, p-values) with the magnitude of the change enabling",
                "quick visual identification of those genes that display",
                "large-magnitude changes and that are also statistically",
                "significant. The horizontal dashed line sets the threshold",
                "for statistical significance, while the vertical dashed",
                "lines set the thresholds for biological significance. It",
                "should be noted that volcano plots become harder to",
                "interpret when using more than one statistical algorithm",
                "and performing meta-analysis. This happens because the genes",
                "that have stronger evidence of being differentially",
                "expressed obtain lower p-values while the rest either",
                "remain at similar levels or obtain higher p-values.",
                "The result is a 'warped' volcano plot, with two",
                "main data clouds: one in the upper part of the plot, and",
                "one in the lower part of the plot. You can always zoom in",
                "when using interactive mode (the default).",collapse=" "
                    ),
                mastat=paste(
                "A mean-difference (or MA) plot with overlaid statistical",
                "information (p-value and fold change thresholds manifested",
                "as points with different colors) is a very useful graphic",
                "that enables the visualization of the results of",
                "differential expression analysis. It differs from the",
                "volcano plot regarding what is displayed in the axes system.",
                "While a volcano plot displays the fold change (x-axis)",
                "versus the statistical significance (y-axis), an MA plot",
                "with statistical scores depicts average expression over the",
                "biological conditions that are compared (x-axis) versus the",
                "fold change of the comparison. Statistical significance",
                "categorization is added as point coloring and statistical",
                "significance is indicated only by different colors and not",
                "by the position to the axes system as in the volcano plot.",
                "This plot is useful when it is of little interest how",
                "statistically significant a gene/transcript is (we are",
                "interested only in the fact that it is) but someone is",
                "interested in actual expression and fold change values",
                "instead.",collapse=" "
                ),
                deregulogram=paste(
                "The de-regulogram is a scatterplot of fold changes between ",
                "two different contrasts. It depicts whether the DEGs between",
                "the two selected contrasts follow a concordant or discordant",
                "regulation pattern. For each (common) DEG, the x-axis and",
                "y-axis represent the log<sub>2</sub> fold change of the two",
                "contrasts. The location of each point along the four",
                "quartiles can directly show its regulation pattern in the two",
                "comparisons. Therefore, the dots localized in the second or",
                "the fourth quartile, illustrate DEGs with a common regulation",
                "pattern, while those localized in the first or third quartile",
                "represent DEGs with opposite patterns of regulation."
                ),
                biodist=paste(
                "The chromosome and biotype distributions bar diagram for",
                "Differentially Expressed Genes (DEGs) is split in two",
                "panels: i)in the upper panel DEGs are distributed per",
                "chromosome and the percentage of each chromosome in the",
                "genome is presented in red bars, the percentage of DEGs",
                "in each chromosome is presented in green bars and the",
                "percentage of certain chromosomes in the distribution of",
                "DEGs is presented in blue bars; ii)in the lower panel,",
                "DEGs are distributed per biotype and the percentage of",
                "each biotype in the genome (i.e. in the whole set of",
                "features provided, for example, protein coding genes, non",
                "coding RNAs or pseudogenes) is presented in red bars,",
                "the percentage of DEGs in each biotype is presented in",
                "green bars and the percentage of each biotype in DEGs is",
                "presented in blue lines. The vertical line separates the most",
                "abundant biotypes (on the left-hand side, corresponding to",
                "the left axis scale), from the rest(on the right-hand side,",
                "corresponding to the right axis scale). Otherwise, the lower",
                "abundance, biotypes would be indistinguishable.",collapse=" "
                    ),
                filtered=paste(
                "The chromosome and biotype distribution of filtered genes",
                "is a quality control chart with two rows and four panels:",
                "on the left panel of the first row, the bar chart depicts",
                "the numbers of filtered genes per chromosome (actual numbers",
                "shown above the bars). On the right panel of the first row,",
                "the bar chart depicts the numbers of filtered genes per",
                "biotype (actual numbers shown above the bars). On the left",
                "panel of the second row, the bar chart depicts the fraction",
                "of filtered genes to the total genes per chromosome",
                "(actual percentages shown above the bars). On the right",
                "panel of the second row, the bar chart depicts the fraction",
                "of the filtered genes to the total genes per biotype",
                "(actual percentages shown above the bars). This plot",
                "should indicate possible quality problems when for example",
                "the filtered genes for a specific chromosome (or the",
                "fraction) is much higher than the rest. Generally,",
                "the fractions per chromosome should be uniform and the",
                "fractions per biotype should be proportional to the biotype",
                "fraction relative to the genome.",collapse=" "
                    ),
                    statvenn=paste(
                    "Venn diagrams are an intuitive way of presenting",
                    "overlaps between lists, based on the overlap of basic",
                    "geometrical shapes. The numbers of overlapping genes per",
                    "statistical algorithm are shown in the different areas",
                    "of the Venn diagrams, one for each contrast. Apart from",
                    "a p-value cutoff, a fold change threshold of 0.5 in",
                    "log<sub>2</sub> scale is applied for each contrast. For",
                    "multi-condition contrasts, the first condition is used to",
                    "calculate the fold change against the reference.",
                    collapse=" "
                    ),
                    foldvenn=paste(
                    "Venn diagrams are an intuitive way of presenting",
                    "overlaps between lists, based on the overlap of basic",
                    "geometrical shapes. The numbers of overlapping genes per",
                    "statistical contrast are shown in the different areas",
                    "of the Venn diagrams, one for each contrast. Apart from",
                    "a p-value cutoff, a fold change threshold of 0.5 in",
                    "log<sub>2</sub> scale is applied for each contrast. For",
                    "multi-condition contrasts, the first condition is used to",
                    "calculate the fold change against the reference.",
                    collapse=" "
                    )
                ),
                references=list(
                    main=paste("Moulos, P., Hatzis, P. (2015). Systematic",
                        "integration of RNA-Seq statistical algorithms for",
                        "accurate detection of differential gene expression",
                        "patterns. Nucleic Acids Research 43(4), e25."),
                    filein=list(
                        sam=paste("Statham, A.L., Strbenac, D., Coolen, M.W.,",
                            "Stirzaker, C., Clark, S.J., Robinson, M.D.",
                            " (2010). Repitools: an R package for the analysis",
                            "of enrichment-based epigenomic data.",
                            "Bioinformatics 26(13), 1662-1663."),
                        bam=paste("Statham, A.L., Strbenac, D., Coolen, M.W.,",
                            "Stirzaker, C., Clark, S.J., Robinson, M.D. (2010)",
                            "Repitools: an R package for the analysis of",
                            "enrichment-based epigenomic data. Bioinformatics",
                            "26(13), 1662-1663."),
                        bed=paste("Lawrence, M., Gentleman, R., Carey, V.",
                            "(2009). rtracklayer: an R package for interfacing",
                            "with genome browsers. Bioinformatics 25(14),",
                            "1841-1842.")
                    ),
                    norm=list(
                        edaseq=paste("Risso, D., Schwartz, K., Sherlock, G.,",
                            "and Dudoit, S. (2011). GC-content normalization",
                            "for RNA-Seq data. BMC Bioinformatics 12, 480."),
                        deseq=paste("Anders, S., and Huber, W. (2010).",
                            "Differential expression analysis for sequence",
                            "count data. Genome Biol 11, R106."),
                        deseq2=paste("Love, M.I., Huber, W., Anders, S.",
                            "Moderated estimation of fold change and",
                            "dispersion for RNA-seq data with DESeq2.",
                            "Genome Biology 15(12):550 (2014)"), 
                        edger=paste("Robinson, M.D., McCarthy, D.J., and",
                            "Smyth, G.K. (2010). edgeR: a Bioconductor package",
                            "for differential expression analysis of digital",
                            "gene expression data. Bioinformatics 26,",
                            "139-140."),
                        noiseq=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        nbpseq=paste("Di, Y, Schafer, D., Cumbie, J.S., and",
                            "Chang, J.H. (2011). The NBP Negative Binomial",
                            "Model for Assessing Differential Gene Expression",
                            "from RNA-Seq. Statistical Applications in",
                            "Genetics and Molecular Biology 10(1), 1-28."),
                        absseq=paste("Wentao Yang, Philip Rosenstiel and",
                            "Hinrich Schulenburg: ABSSeq: a new RNA-Seq",
                            "analysis method based on modelling absolute",
                            "expression differences. BMC Genomics 2016;",
                            "17: 541"),
                        dss=paste("Hao Wu, Chi Wang, Zhijin Wu (2013):",
                            "A new shrinkage estimator for dispersion",
                            "improves differential expression detection in",
                            "RNA-seq data. Biostatistics, 14(2):232-43."),
                        none=NULL
                    ),
                    stat=list(
                        deseq=paste("Anders, S., and Huber, W. (2010).",
                            "Differential expression analysis for sequence",
                            "count data. Genome Biol 11, R106."),
                        deseq2=paste("Love, M.I., Huber, W., Anders, S.",
                            "Moderated estimation of fold change and",
                            "dispersion for RNA-seq data with DESeq2.",
                            "Genome Biology 15(12):550 (2014)"),
                        edger=paste("Robinson, M.D., McCarthy, D.J., and",
                            "Smyth, G.K. (2010). edgeR: a Bioconductor package",
                            "for differential expression analysis of digital",
                            "gene expression data. Bioinformatics 26,",
                            "139-140."),
                        noiseq=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        limma=paste("Smyth, G. (2005). Limma: linear models",
                            "for microarray data. In Bioinformatics and",
                            "Computational Biology Solutions using R and",
                            "Bioconductor, G. R., C. V., D. S., I. R., and",
                            "H. W., eds. (New York, Springer), pp. 397-420."),
                        #bayseq=paste("Hardcastle, T.J., and Kelly, K.A.",
                        #    "(2010). baySeq: empirical Bayesian methods for",
                        #    "identifying differential expression in sequence",
                        #    "count data. BMC Bioinformatics 11, 422."),
                        nbpseq=paste("Di, Y, Schafer, D., Cumbie, J.S., and",
                            "Chang, J.H. (2011). The NBP Negative Binomial",
                            "Model for Assessing Differential Gene Expression",
                            "from RNA-Seq. Statistical Applications in",
                            "Genetics and Molecular Biology 10(1), 1-28."),
                        ebseq=paste("Leng, N., Dawson, J.A., Thomson, J.A.,",
                            "Ruotti, V., Rissman, A.I., Smits, B.M., Haag,",
                            "J.D., Gould, M.N., Stewart, R.M., and",
                            "Kendziorski, C. (2013). EBSeq: an empirical",
                            "Bayes hierarchical model for inference in",
                            "RNA-seq experiments. Bioinformatics 29, ",
                            "1035-1043"),
                        absseq=paste("Wentao Yang, Philip Rosenstiel and",
                            "Hinrich Schulenburg: ABSSeq: a new RNA-Seq",
                            "analysis method based on modelling absolute",
                            "expression differences BMC Genomics 2016; 17:",
                            "541"),
                        dss=paste("Hao Wu, Chi Wang, Zhijin Wu (2013):",
                            "A new shrinkage estimator for dispersion",
                            "improves differential expression detection in",
                            "RNA-seq data. Biostatistics, 14(2):232-43.",
                            "doi:10.1093/biostatistics/kxs033")
                    ),
                    meta=list(
                        fisher=paste("Fisher, R.A. (1932). Statistical",
                            "Methods for Research Workers (Edinburgh, Oliver",
                            "and Boyd)."),
                        fperm=paste("Fisher, R.A. (1932). Statistical",
                            "Methods for Research Workers (Edinburgh, Oliver",
                            "and Boyd)."),
                        whitlock=c(
                            paste("Whitlock, M.C. (2005). Combining",
                                "probability from independent tests:",
                                "the weighted Z-method is superior to Fisher's",
                                "approach. J Evol Biol 18, 1368-1373."),
                            paste("Schroder, M.S., Culhane, A.C., Quackenbush,",
                                "J., and Haibe-Kains, B. (2011). survcomp:",
                                "an R/Bioconductor package for performance",
                                "assessment and comparison of survival",
                                "models. Bioinformatics 27, 3206-3208.")
                        ),
                        weight=paste("Genovese, C.R., Roeder, K., Wasserman,",
                            "L. (2006). False discovery control with p-value",
                            "weighting. Biometrika 93 (3): 509-524."),
                        simes=paste("Simes, R. J. (1986). An improved",
                            "Bonferroni procedure for multiple tests of",
                            "significance. Biometrika 73 (3): 751-754."),
                        harmonic=paste("Wilson, D. J. (2019). The harmonic",
                            "mean p-value for combining dependent tests.",
                            "PNAS 116 (4): 1995-1200."),
                        none=NULL
                    ),
                    multiple=list(
                        BH=paste("Benjamini, Y., and Hochberg, Y. (1995). ",
                            "Controlling the False Discovery Rate: A Practical",
                            "and Powerful Approach to Multiple Testing.",
                            "Journal of the Royal Statistical Society Series",
                            "B (Methodological) 57, 289-300."),
                        fdr=paste("Benjamini, Y., and Hochberg, Y. (1995). ",
                            "Controlling the False Discovery Rate: A Practical",
                            "and Powerful Approach to Multiple Testing.",
                            "Journal of the Royal Statistical Society Series",
                            "B (Methodological) 57, 289-300."),
                        BY=paste("Benjamini, Y., and Yekutieli, D. (2001). The",
                            "control of the false discovery rate in multiple",
                            "testing under dependency. Annals of Statistics",
                            "26, 1165-1188."),
                        bonferroni=paste("Shaffer, J.P. (1995). Multiple",
                            "hypothesis testing. Annual Review of",
                            "Psychology 46, 561-576."),
                        holm=paste("Holm, S. (1979). A simple sequentially",
                            "rejective multiple test procedure. Scandinavian",
                            "Journal of Statistics 6, 65-70."),
                        hommel=paste("Hommel, G. (1988). A stagewise rejective",
                            "multiple test procedure based on a modified",
                            "Bonferroni test. Biometrika 75, 383-386."),
                        hochberg=paste("Hochberg, Y. (1988). A sharper",
                            "Bonferroni procedure for multiple tests of",
                            "significance. Biometrika 75, 800-803."),
                        qvalue=paste("Storey, J.D., and Tibshirani, R. (2003).",
                            "Statistical significance for genomewide studies.",
                            "Proc Natl Acad Sci U S A 100, 9440-9445.")
                    ),
                    figure=list(
                        mds=paste("Planet, E., Attolini, C.S., Reina, O.,",
                            "Flores, O., and Rossell, D. (2012). htSeqTools:",
                            "high-throughput sequencing quality control,",
                            "processing and visualization in R. Bioinformatics",
                            "28, 589-590."),
                        biodetection=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        countsbio=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        saturation=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        readnoise=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        gcbias=paste("Risso, D., Schwartz, K., Sherlock, G.,",
                            "and Dudoit, S. (2011). GC-content normalization",
                            "for RNA-Seq data. BMC Bioinformatics 12, 480."),
                        lengthbias=paste("Risso, D., Schwartz, K., Sherlock,",
                            "G., and Dudoit, S. (2011). GC-content",
                            "normalization for RNA-Seq data.",
                            "BMC Bioinformatics 12, 480."),
                        meandiff=paste("Risso, D., Schwartz, K., Sherlock, G.,",
                            "and Dudoit, S. (2011). GC-content normalization",
                            "for RNA-Seq data. BMC Bioinformatics 12, 480."),
                        meanvar=paste("Risso, D., Schwartz, K., Sherlock, G.,",
                            "and Dudoit, S. (2011). GC-content normalization",
                            "for RNA-Seq data. BMC Bioinformatics 12, 480."),
                        rnacomp=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        biodist=paste("Tarazona, S., Garcia-Alcalde, F.,",
                            "Dopazo, J., Ferrer, A., and Conesa, A. (2011).",
                            "Differential expression in RNA-seq: a matter of",
                            "depth. Genome Res 21, 2213-2223."),
                        statvenn=paste("Chen, H., and Boutros, P.C. (2011).",
                            "VennDiagram: a package for the generation of",
                            "highly-customizable Venn and Euler diagrams in R.",
                            "BMC Bioinformatics 12, 35."),
                        foldvenn=paste("Chen, H., and Boutros, P.C. (2011).",
                            "VennDiagram: a package for the generation of",
                            "highly-customizable Venn and Euler diagrams in R.",
                            "BMC Bioinformatics 12, 35."),
                        filtered=NULL
                    )
                )
            )
        }
    )
    return(messages)
}

#makeHighchartsPoints <- function(x,y,a=NULL) {
#    if (length(x)>0) {
#        n <- names(x)
#        x <- unname(x)
#        y <- unname(y)
#        stru <- vector("list",length(x))
#        if (is.null(a)) {
#            for (i in seq_len(length(x)))
#                stru[[i]] <- list(
#                    x=round(x[i],digits=3),
#                    y=round(y[i],digits=3),
#                    name=n[i]
#                )
#        }
#        else {
#            for (i in seq_len(length(x)))
#                stru[[i]] <- list(
#                    x=round(x[i],digits=3),
#                    y=round(y[i],digits=3),
#                    name=n[i],
#                    alt_name=a[i]
#                )
#        }
#    }
#    else
#        stru <- list(x=NULL,y=NULL,name=NULL,alt_name=NULL)
#    return(stru)
#}

makeHighchartsPoints <- function(x,y,a=NULL,p=NULL,simple=FALSE) {
    if (length(x)>0) {
        n <- names(x)
        x <- unname(x)
        y <- unname(y)
        if (!is.null(a))
            a <- unname(a)
        if (!is.null(p))
            p <- unname(p)
        #stru <- vector("list",length(x))
        if (simple) {
            return(lapply(seq_along(x),function(i,x,y) {
                return(c(x[i],y[i]))
            },x,y))
        }
        if (is.null(a) && is.null(p)) {
            stru <- lapply(seq_along(x),function(i,x,y,n) {
                return(list(
                    x=round(x[i],digits=3),
                    y=round(y[i],digits=3),
                    name=n[i]
                ))
            },x,y,n)
        }
        else if (!is.null(a) && is.null(p)) {
            stru <- lapply(seq_along(x),function(i,x,y,n,a) {
                return(list(
                    x=round(x[i],digits=3),
                    y=round(y[i],digits=3),
                    name=n[i],
                    alt_name=a[i]
                ))
            },x,y,n,a)
        }
        else if (is.null(a) && !is.null(p)) {
            stru <- lapply(seq_along(x),function(i,x,y,n,p) {
                return(list(
                    x=round(x[i],digits=3),
                    y=round(y[i],digits=3),
                    name=n[i],
                    sig=round(p[i],digits=3)
                ))
            },x,y,n,p)
        }
        else if (!is.null(a) && !is.null(p)) {
            stru <- lapply(seq_along(x),function(i,x,y,n,a,p) {
                return(list(
                    x=round(x[i],digits=3),
                    y=round(y[i],digits=3),
                    name=n[i],
                    alt_name=a[i],
                    sig=round(p[i],digits=3)
                ))
            },x,y,n,a,p)
        }
    }
    else
        stru <- list(x=NULL,y=NULL,name=NULL,alt_name=NULL,sig=NULL)
    return(stru)
}

asClassVector <- function(sampleList) {
    classes <- vector("list",length(sampleList))
    names(classes) <- names(sampleList)
    for (n in names(sampleList))
        classes[[n]] <- rep(n,times=length(sampleList[[n]]))
    classes <- unlist(classes,use.names=FALSE)
    names(classes) <- unlist(sampleList,use.names=FALSE)
    return(classes)
}

getArg <- function(argList,argName) {
    return(argList[argName])
}

setArg <- function(argList,argName,argValue=NULL) {
    if (is.list(argName))
        argList[names(argName)] <- argName
    else if (is.character(argName)) {
        tmp <- vector("list",length(argName))
        names(tmp) <- argName
        i <- 0
        for (n in argName) {
            i <- i + 1
            tmp[[n]] <- argValue[i]
        }
        argList[argName] <- tmp
    }
    return(argList)
}

wpAdjust <- function(p,m) {
    if (m=="qvalue")
        return(qvalue(p))
    else
        return(p.adjust(p,method=m))
}

cmclapply <- function(...,rc) {
    if (suppressWarnings(!requireNamespace("parallel")) 
        || .Platform$OS.type!="unix")
        m <- FALSE
    else {
        m <- TRUE
        ncores <- parallel::detectCores()
        if (ncores==1) 
            m <- FALSE
        else {
            if (!missing(rc) && !is.null(rc))
                ncores <- ceiling(rc*ncores)
            else 
                m <- FALSE
        }
    }
    if (m)
        return(mclapply(...,mc.cores=ncores,mc.set.seed=FALSE))
    else
        return(lapply(...))
}

filterLow <- function(x,f) { return(all(x<=f)) }

filterHigh <- function(x,f) { return(all(x>=f)) }

disp <- function(...) {
    #verbose <- get("VERBOSE",envir=metaEnv)
    verbose <- TRUE
    if (!is.null(verbose) && verbose) {
        message("\n",...,appendLF=FALSE)
    }
    logger <- get("LOGGER",envir=metaEnv)
    levalias <- c("one","two","three","four","five")
    if (!is.null(logger)) {
        switch(levalias[level(logger)],
            one = { debug(logger,paste0(...)) },
            two = { info(logger,gsub("\\n","",paste0(...))) },
            three = { warn(logger,gsub("\\n","",paste0(...))) },
            four = { error(logger,gsub("\\n","",paste0(...))) },
            five = { fatal(logger,gsub("\\n","",paste0(...))) }
        )
    }
}

stopwrap <- function(...,t="fatal") {
    logger <- get("LOGGER",envir=metaEnv)
    if (!is.null(logger)) {
        if (t=="fatal")
            fatal(logger,gsub("\\n","",paste0(...)))
        else
            error(logger,gsub("\\n","",paste0(...)))
    }
    stop(paste0(...))
}

warnwrap <- function(...,now=FALSE) {
    logger <- get("LOGGER",envir=metaEnv)
    if (!is.null(logger))
        warn(logger,gsub("\\n","",paste0(...)))
    if (now)
        warning(paste0(...),call.=FALSE,immediate.=TRUE)
    else
        warning(paste0(...),call.=FALSE)
}

exampleCountData <- function(ngenes) {
    if (missing(ngenes))
        ngenes <- 10000
    q0 <- rexp(ngenes,rate=1/250)
    is_DE <- runif(ngenes) < 0.3
    lfc <- rnorm(ngenes,sd=2)
    q0A <- ifelse(is_DE,q0*2^(lfc/2),q0)
    q0B <- ifelse(is_DE,q0*2^(-lfc/2),q0)
    true_sf <- c(1.0, 1.3,0.7,0.9,1.6)
    conds <- c("A","A","B","B","B")
    m <- t(vapply(seq_len(ngenes),function(i)
        vapply(seq_len(5),function(j)
            rnbinom(1,mu=true_sf[j]*ifelse(conds[j]=="A",q0A[i],q0B[i] ),
                size=1/0.2),numeric(1)),numeric(5)))
    colnames(m) <- c("A1","A2","B1","B2","B3")
    rownames(m) <- paste("gene",seq_len(ngenes),
        ifelse(is_DE,"T","F"),sep="_")
    return(m)
}

elap2human <- function(start.time) {
    start.time <- as.POSIXct(start.time)
    dt <- difftime(Sys.time(),start.time,units="secs")
    ndt <- as.numeric(dt)
    if (ndt<60)
        format(.POSIXct(dt,tz="GMT"),"%S seconds")
    else if (ndt>=60 && ndt<3600)
        format(.POSIXct(dt,tz="GMT"),"%M minutes %S seconds")
    else if (ndt>=3600 && ndt<86400)
        format(.POSIXct(dt,tz="GMT"),"%H hours %M minutes %S seconds")
    else if (ndt>=86400)
        format(.POSIXct(dt,tz="GMT"),"%d days %H hours %M minutes %S seconds")
}

.deprecatedWarning <- function(func) {
    switch(func,
        readTargets = {
            warnwrap("\"yes\" and \"no\" for read strandedness have been ",
                "deprecated. Please use \"forward\", \"forward\" or \"no\". ",
                "Replacing \"yes\" with \"forward\"...")
        }
    )
}

## The old function. It has NULLs in the new fields that are filled in
## getWeights2 (see furhter down).
#..getWeightsOld <- function(org=c("human","chimpanzee","mouse","fruitfly",
#    "arabidopsis","rat")) {
#    org <- tolower(org)
#    checkTextArgs("org",org,c("human","chimpanzee","mouse","fruitfly",
#        "arabidopsis","rat"))
#    switch(org,
#        human = {
#            return(c(
#                deseq=0.05772458,
#                deseq2=NULL,
#                edger=0.14321672,
#                limma=0.34516089,
#                nbpseq=0.06108182,
#                noiseq=0.11595169,
#                bayseq=0.27686431,
#                absseq=NULL,
#                dss=NULL
#            ))
#        },
#        chimpanzee = {
#            return(c(
#                deseq=0.06026782,
#                deseq2=NULL,
#                edger=0.14964358,
#                limma=0.33500306,
#                nbpseq=0.05814585,
#                noiseq=0.11337043,
#                bayseq=0.28356925,
#                absseq=NULL,
#                dss=NULL
#            ))
#        },
#        mouse = {
#            return(c(
#                deseq=0.05257695,
#                deseq2=NULL,
#                edger=0.24161354,
#                limma=0.29957277,
#                nbpseq=0.04914485,
#                noiseq=0.06847809,
#                bayseq=0.28861381,
#                absseq=NULL,
#                dss=NULL
#            ))
#        },
#        fruitfly = {
#            return(c(
#                deseq=0.01430269,
#                deseq2=NULL,
#                edger=0.12923339,
#                limma=0.38315685,
#                nbpseq=0.01265952,
#                noiseq=0.06778537,
#                bayseq=0.39286218,
#                absseq=NULL,
#                dss=NULL
#            ))
#        },
#        arabidopsis = {
#            return(c(
#                deseq=0.04926122,
#                deseq2=NULL,
#                edger=0.10130858,
#                limma=0.40842011,
#                nbpseq=0.04596652,
#                noiseq=0.09336509,
#                bayseq=0.30167848,
#                absseq=NULL,
#                dss=NULL
#            ))
#        },
#        chimp = {
#            return(c(
#                deseq=NULL,
#                deseq2=NULL,
#                edger=NULL,
#                limma=NULL,
#                nbpseq=NULL,
#                noiseq=NULL,
#                bayseq=NULL,
#                absseq=NULL,
#                dss=NULL
#            ))
#        },
#        rat = {
#            return(c(
#                deseq=NULL,
#                deseq2=NULL,
#                edger=NULL,
#                limma=NULL,
#                nbpseq=NULL,
#                noiseq=NULL,
#                bayseq=NULL,
#                absseq=NULL,
#                dss=NULL
#            ))
#        }
#    )
#}

.myCreateFilterXMLchunk <- function(filterChunk,mart) {
    individualFilters <- vapply(names(filterChunk),
        FUN=function(filter,values,mart) {
            # if the filter exists and is boolean we do this
            if (filter %in% listFilters(mart,what="name") && 
                grepl('boolean',filterType(filter=filter,mart=mart))) {
                if (!is.logical(values[[filter]])) 
                    stop("'",filter, "' is a boolean filter and needs a ",
                        "corresponding logical value of TRUE or FALSE to ",
                        "indicate if the query should retrieve all data that ",
                        "fulfill the boolean or alternatively that all data ", 
                        "that not fulfill the requirement should be retrieved.",
                        call. = FALSE)
                val <- ifelse(values[[filter]],yes=0,no=1)
                val <- paste0("\" excluded = \"",val,"\" ")
                
            } else { 
                # otherwise the filter isn't boolean, or doesn't exist
                if(is.numeric(values[[filter]])) 
                    values[[filter]] <- as.integer(values[[filter]])
                val <- paste0(values[[filter]], collapse = ",")
                # convert " to ' to avoid truncating the query string
                val <- gsub(x=val,pattern="\"",replacement="'",
                    fixed=TRUE)
                val <- paste0('" value = "',val,'" ')
            }
            filterXML <- paste0("<Filter name = \"",filter,val,"/>")
            return(filterXML)
        },FUN.VALUE=character(1),filterChunk,mart,USE.NAMES=FALSE)
    filterXML <- paste0(individualFilters,collapse="")
    return(filterXML)
}

.mySubmitQueryXML <- function (host, query) {
    httr::set_config(httr::config(http_version = 1L))
    res <- httr::POST(url = host, body = list(query = query),
        httr::timeout(2000))
    if (httr::status_code(res) == 302) {
        host <- stringr::str_match(string=res$all_headers[[1]]$headers$location,
            pattern = "//([a-zA-Z./]+)\\??;?redirectsrc")[, 2]
        res <- httr::POST(url = host, body = list(query = query),
            config = list(httr::timeout(2000)))
    }
    return(suppressMessages(httr::content(res)))
}

.myGenerateFilterXML <- function(filters="",values,mart) {
    # return empty string if no filter specified & this isn't ensembl
    # specifying no filter is generally bad, as it will get 'everything'
    # and we might encounter the time out problem
    if (filters[1] == "") {
        return("")
    }
    # if we have multiple filters, the values must be specified as a list.
    if (length(filters) > 1 && !is(values,"list")) {
        stop("If using multiple filters, the 'value' has to be a list.\n",
            "For example, a valid list for 'value' could be: ",
            "list(affyid=c('1939_at','1000_at'),chromosome='16')\n",
            "Here we select on Affymetrix identifier and chromosome, only ",
            "results that pass both filters will be returned");
    }
    # it's easy to not realise you're passing a data frame here, so check
    if (is.data.frame(values) && ncol(values == 1)) {
        values <- values[,1]
    }
    if (!is.list(values)) {
        values <- list(values)
    }
    names(values) <- filters
    values <- .mySplitValues(list(values),maxChunkSize=1000)
    filterXML_list <- lapply(values,.myCreateFilterXMLchunk,mart)
    return(filterXML_list)
}

.mySplitValues <- function(valuesList,maxChunkSize=500) {
    vLength <- vapply(valuesList[[1]],FUN=length,FUN.VALUE=integer(1))
    
    if(all(vLength <= maxChunkSize)) {
        return(valuesList)
    } else {
        # pick the next filter to split
        vIdx <- min(which(vLength > maxChunkSize))
        nchunks <- (vLength[vIdx] %/% maxChunkSize) + 1
        splitIdx <- rep(seq_len(nchunks),
            each=ceiling(vLength[vIdx]/nchunks))[seq_len(vLength[vIdx])]
        
        # a new list we will populate with the chunks
        tmpList <- list()
        for(i in seq_len(nchunks)) {
            for(j in seq_along(valuesList)) {
                listIdx <- ((i - 1) * length(valuesList)) + j
                tmpList[[listIdx]] <- valuesList[[j]]
                tmpList[[listIdx]][[vIdx]] <-
                    tmpList[[listIdx]][[vIdx]][which(splitIdx == i)]
            }
        }
        # recursively call the function to process next filter
        valuesList <- .mySplitValues(tmpList,maxChunkSize=maxChunkSize)
    }
    return(valuesList)
}

.myGetBM <- function(attributes,filters="",values="",mart,curl=NULL,
    checkFilters=TRUE,verbose=FALSE,uniqueRows=TRUE,bmHeader=FALSE,quote="\"") {
    .myMartCheck(mart)
    if(missing(attributes))
        stop("Argument 'attributes' must be specified.")
    
    if (is.list(filters) && !missing(values))
        warning("Argument 'values' should not be used when argument 'filters'",
            "is a list and will be ignored.")
    if (is.list(filters) && is.null(names(filters)))
        stop("Argument 'filters' must be a named list when sent as a list.")
    if (!is.list(filters) && all(filters != "") && missing(values))
        stop("Argument 'values' must be specified.")
    if (length(filters) > 0 && length(values) == 0)
        stop("Values argument contains no data.")
    
    if (is.list(filters)) {
        values <- filters
        filters <- names(filters)
    }
    
    if (!is(uniqueRows,"logical"))
        stop("Argument 'uniqueRows' must be a logical value, so either TRUE ",
            "or FALSE")
    
    # force the query to return the 'english text' header names with the result
    # we use these later to match and order attribute/column names    
    callHeader <- TRUE
    xmlQuery <- paste0("<?xml version='1.0' encoding='UTF-8'?><!DOCTYPE Query>",
        #"<Query  virtualSchemaName = '",biomaRt:::martVSchema(mart),
        "<Query  virtualSchemaName = '",mart@vschema,
        "' uniqueRows = '",as.numeric(uniqueRows),
        "' count = '0' datasetConfigVersion = '0.6' header='",
        as.numeric(callHeader),"' requestid= 'biomaRt'> <Dataset name = '",
        #biomaRt:::martDataset(mart),"'>")
        mart@dataset,"'>")
    
    # checking the Attributes
    invalid <- !(attributes %in% listAttributes(mart, what="name"))
    if(any(invalid))
        stop(paste("Invalid attribute(s):", paste(attributes[invalid],
            collapse=", "),"\nPlease use the function 'listAttributes' to get ",
            "valid attribute names"))
    
    # attribute are ok lets add them to the query
    attributeXML = paste("<Attribute name = '",attributes, "'/>",collapse="",
        sep="")
    
    #checking the filters
    if (filters[1] != "" && checkFilters) {
        invalid <- !(filters %in% listFilters(mart, what="name"))
        if(any(invalid))
            stop(paste("Invalid filters(s):", paste(filters[invalid],
                collapse=", "),"\nPlease use the function 'listFilters' to ",
                "get valid filter names"))
    }
    
    # filterXML is a list containing filters with reduced numbers of values
    # to meet the 500 value limit in BioMart queries
    filterXmlList <- .myGenerateFilterXML(filters,values,mart)
    
    resultList <- list()
    
    # we submit a query for each chunk of the filter list
    for (i in seq_along(filterXmlList)) {
        filterXML <- filterXmlList[[i]]
        fullXmlQuery <- paste(xmlQuery,attributeXML,filterXML,
            "</Dataset></Query>",sep="")
        
        if(verbose)
            message(fullXmlQuery)
        
        # we choose a separator based on whether '?redirect=no' is present
        #sep <- ifelse(grepl(x=biomaRt:::martHost(mart),
        sep <- ifelse(grepl(x=mart@host,
            pattern=".+\\?.+"), "&", "?")
        
        #postRes <- .mySubmitQueryXML(host=paste0(biomaRt:::martHost(mart),sep),
        postRes <- .mySubmitQueryXML(host=paste0(mart@host,sep),
            query=fullXmlQuery)
        
        if (verbose) {
            writeLines("#################\nResults from server:")
            print(postRes)
        }
        
        if (!(is.character(postRes) && (length(postRes)==1L)))
            stop("The query to the BioMart webservice returned an invalid ",
                "result: biomaRt expected a character string of length 1.\n",
                "Please report this on the support site at", 
                "http://support.bioconductor.org")
        
        if (gsub("\n","",postRes,fixed=TRUE,useBytes=TRUE)== "") { 
            # meaning an empty result
            result <- as.data.frame(matrix("",ncol=length(attributes),nrow=0),
                stringsAsFactors=FALSE)
        } else {
            if (length(grep("^Query ERROR",postRes)) > 0L)
                stop(postRes)
            
            # convert the serialized table into a dataframe
            con <- textConnection(postRes)
            result <- read.table(con,sep="\t",header=callHeader,quote=quote,
                comment.char="",check.names=FALSE,stringsAsFactors=FALSE)
            if(verbose) {
                writeLines("#################\nParsed results:")
                print(result)
            }
            close(con)
            
            if (!(is(result,"data.frame") 
                && (ncol(result)==length(attributes)))) {
                print(head(result))
                stop("The query to the BioMart webservice returned an invalid ",
                    "result: the number of columns in the result table does ",
                    "not equal the number of attributes in the query.\n",
                    "Please report this on the support site at",
                    "http://support.bioconductor.org")
            }
        }
        
        resultList[[i]] <- .mySetResultColNames(result,mart=mart,
            attributes=attributes,bmHeader=bmHeader)
    }
    
    # collate results
    result <- do.call('rbind',resultList)
    return(result)
}

.myMartCheck = function(mart,biomart=NULL) {
    if (missing(mart) || !is(mart,'Mart')) {
        stop("You must provide a valid Mart object. To create a Mart ",
            "object use the function: useMart. Check ?useMart for more ",
            "information.")
    }
    if(!is.null(biomart)){
        martcheck <- mart@biomart
        bmok <- FALSE
        for (k in seq_len(length(biomart))) {
            if(martcheck[1] == biomart[k]) {
                bmok <- TRUE
            }
        }
        if (!bmok) {
            stop(paste("This function only works when used with the ",
                biomart," BioMart.",sep=""))
        }
    }
    if (mart@dataset == "") {
        stop("No dataset selected, please select a dataset first.\nYou can ",
            "see the available datasets by using the listDatasets function\n",
            "See ?listDatasets for more information. Then you should create\n",
            "the Mart object by using the useMart function.\nSee ?useMart ",
            "for more information")
    }
}

.mySetResultColNames <- function(result,mart,attributes,bmHeader=FALSE) {
    # get all avaialble sttributes and 
    # filter only for the ones we've actually asked for
    att <- listAttributes(mart,what=c("name","description"))
    att <- att[which(att[,'name'] %in% attributes),]
    if (length(which(duplicated(att[,'description']))) > 
        length(which(duplicated(att)))) {
        warning("Cannot unambiguously match attribute names\nIgnoring ",
            "bmHeader argument and using biomart description field")
        return(result)
    }
    
    resultNames = colnames(result)
    # match the returned column names with the attribute names
    matches <- match(resultNames,att[,2],NA)
    if (any(is.na(matches))) {
        warning("Problems assigning column names. Currently using ",
            "the biomart description field. You may wish to set these ",
            "manually.")
        return(result)
    }
    # if we want to use the attribute names we specified, do this, 
    # otherwise we use the header returned with the query
    if (!bmHeader) {
        colnames(result) <- att[matches,1]
    }
    # now put things in the order we actually asked for the attributes in
    result <- result[,match(att[matches,1],attributes),drop=FALSE]
    
    return(result)
}

.GENE_TYPES <- c("gene","pseudogene","transposable_element_gene")
.TX_TYPES <- c("transcript","pseudogenic_transcript","primary_transcript",
    "mRNA","ncRNA","rRNA","snoRNA","snRNA","tRNA","tmRNA","miRNA",
    "miRNA_primary_transcript","RNase_P_RNA","RNase_MRP_RNA","SRP_RNA",
    "misc_RNA","antisense_RNA", "antisense","lnc_RNA","antisense_lncRNA",
    "transcript_region","pseudogenic_tRNA","scRNA","guide_RNA",
    "telomerase_RNA","vault_RNA","Y_RNA")
.EXON_TYPES <- c("exon","pseudogenic_exon","coding_exon",
    "five_prime_coding_exon","interior_coding_exon",
    "three_prime_coding_exon","exon_of_single_exon_gene","interior_exon",
    "noncoding_exon","five_prime_noncoding_exon",
    "three_prime_noncoding_exon")
.CDS_TYPES <- c("CDS","transposable_element_CDS","CDS_predicted",
    "edited_CDS")
.STOP_CODON_TYPES <- "stop_codon"
.GFF_FEATURE_TYPES <- c(.GENE_TYPES,.TX_TYPES,.EXON_TYPES,.CDS_TYPES,
    .STOP_CODON_TYPES)

buildExport <- function(geneData,rawGeneCounts,normGeneCounts,flags,
    sampleList,cnt=NULL,statistics=NA,rawList,normList,
    pMat=if (is(geneData,"GenomicRanges")) 
        matrix(NA,length(geneData),length(statistics))
        else matrix(NA,nrow(geneData),length(statistics)),
    adjpMat=if (is(geneData,"GenomicRanges"))
        matrix(NA,length(geneData),length(statistics))
        else matrix(NA,nrow(geneData),length(statistics)),
    sumP=if (is(geneData,"GenomicRanges")) rep(NA,length(geneData))
        else rep(NA,nrow(geneData)),
    adjSumP=if (is(geneData,"GenomicRanges")) rep(NA,length(geneData))
        else rep(NA,nrow(geneData)),
    exportWhat=c("annotation","p_value","adj_p_value","meta_p_value",
        "adj_meta_p_value","fold_change","stats","counts","flags"),
    exportScale=c("natural","log2","log10","rpgm","vst"),
    exportValues=c("raw","normalized"),
    exportStats=c("mean","median","sd","mad","cv","rcv"),
    logOffset=1,report=TRUE) {
    if (is(geneData,"GenomicRanges")) {
        geneData <- as.data.frame(geneData)
        gci <- grep("gc_content",colnames(geneData))
        nmi <- grep("gene_name",colnames(geneData))
        bti <- grep("biotype",colnames(geneData))
        baseInd <- c(1,2,3,6)
        if (length(gci) > 0)
            baseInd <- c(baseInd,gci,5)
        else
            baseInd <- c(baseInd,5)
        if (length(nmi) > 0)
            baseInd <- c(baseInd,nmi)
        if (length(bti) > 0)
            baseInd <- c(baseInd,bti)
        geneData <- geneData[,baseInd]
    }
    
    if (is.null(cnt)) { # No p-value or fold change can be exported
        if ("p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "p_value")]
        if ("adj_p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "adj_p_value")]
        if ("meta_p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "meta_p_value")]
        if ("adj_meta_p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "adj_meta_p_value")]
        if ("fold_change" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "fold_change")]
    }
    
    if (any(is.na(statistics))) { # No p-value can be exported
        if ("p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "p_value")]
        if ("adj_p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "adj_p_value")]
        if ("meta_p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "meta_p_value")]
        if ("adj_meta_p_value" %in% exportWhat)
            exportWhat <- exportWhat[-which(exportWhat == "adj_meta_p_value")]
    }

    if (is.null(colnames(pMat)) && "p_value" %in% exportWhat)
        colnames(pMat) <- statistics
    if (is.null(adjpMat) && "adj_p_value" %in% exportWhat)
        adjpMat=matrix(NA,nrow(geneData),length(statistics))
    if (is.null(colnames(adjpMat)) && "adj_p_value" %in% exportWhat)
        colnames(adjpMat) <- statistics
    
    export <- data.frame(row.names=rownames(geneData))
    if (report)
        exportHtml <- as.matrix(export)
    theNames <- character(0)
    if ("annotation" %in% exportWhat) {
        disp("      binding annotation...")
        export <- cbind(export,geneData)
        if (report)
            exportHtml <- cbind(exportHtml,.makeHtmlCells(geneData,
                type="text"))
        theNames <- c(theNames,colnames(geneData))
    }
    if ("p_value" %in% exportWhat) {
        disp("      binding p-values...")
        export <- cbind(export,pMat)
        if (report) 
            exportHtml <- cbind(exportHtml,.makeHtmlCells(pMat))
        theNames <- c(theNames,paste("p-value_",colnames(pMat),sep=""))
    }
    if ("adj_p_value" %in% exportWhat) {
        disp("      binding FDRs...")
        export <- cbind(export,adjpMat)
        if (report) 
            exportHtml <- cbind(exportHtml,.makeHtmlCells(adjpMat))
        theNames <- c(theNames,paste("FDR_",colnames(adjpMat),sep=""))
    }
    if ("meta_p_value" %in% exportWhat && length(statistics)>1) { 
        # Otherwise, it does not exist
        disp("      binding meta p-values...")
        export <- cbind(export,sumP)
        if (report) 
            exportHtml <- cbind(exportHtml,.makeHtmlCells(sumP))
        theNames <- c(theNames,paste("meta_p-value_",cnt,sep=""))
    }
    if ("adj_meta_p_value" %in% exportWhat && length(statistics)>1) {
        disp("      binding adjusted meta p-values...")
        export <- cbind(export,adjSumP)
        if (report) 
            exportHtml <- cbind(exportHtml,.makeHtmlCells(adjSumP))
        theNames <- c(theNames,paste("meta_FDR_",cnt,sep=""))
    }
    if ("fold_change" %in% exportWhat) {
        if ("normalized" %in% exportValues) {
            tmp <- makeFoldChange(cnt,sampleList,normGeneCounts,logOffset)
            if (!("natural" %in% exportScale || "natural" %in% exportScale)) {
                # In case only "rpgm" output requested, at least log2 fold
                # changes must be exported if requested
                disp("      binding log2 normalized fold changes...")
                export <- cbind(export,log2(tmp))
                if (report) 
                    exportHtml <- cbind(exportHtml,.makeHtmlCells(log2(tmp)))
                theNames <- c(theNames,paste("log2_normalized_fold_change_",
                    colnames(tmp),sep=""))
            }
            else {
                if ("natural" %in% exportScale) {
                disp("      binding natural normalized fold changes...")
                    export <- cbind(export,tmp)
                    if (report) 
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,
                        paste("natural_normalized_fold_change_",colnames(tmp),
                            sep=""))
                }
                if ("log2" %in% exportScale) {
                    disp("      binding log2 normalized fold changes...")
                    export <- cbind(export,log2(tmp))
                    if (report) 
                        exportHtml <- cbind(exportHtml,
                            .makeHtmlCells(log2(tmp)))
                    theNames <- c(theNames,paste("log2_normalized_fold_change_",
                        colnames(tmp),sep=""))
                }
            }
        }
        if ("raw" %in% exportValues) {
            tmp <- makeFoldChange(cnt,sampleList,rawGeneCounts,logOffset)
            if (!("natural" %in% exportScale || "natural" %in% exportScale)) {
                disp("      binding log2 raw fold changes...")
                export <- cbind(export,log2(tmp))
                if (report) 
                    exportHtml <- cbind(exportHtml,.makeHtmlCells(log2(tmp)))
                theNames <- c(theNames,paste("log2_raw_fold_change_",
                    colnames(tmp),sep=""))
            }
            else {
                if ("natural" %in% exportScale) {
                    disp("      binding natural raw fold changes...")
                    export <- cbind(export,tmp)
                    if (report) exportHtml <- cbind(exportHtml,
                        .makeHtmlCells(tmp))
                    theNames <- c(theNames,paste("natural_raw_fold_change_",
                        colnames(tmp),sep=""))
                }
                if ("log2" %in% exportScale) {
                    disp("      binding log2 raw fold changes...")
                    export <- cbind(export,log2(tmp))
                    if (report) 
                        exportHtml <- cbind(exportHtml,
                            .makeHtmlCells(log2(tmp)))
                    theNames <- c(theNames,paste("log2_raw_fold_change_",
                        colnames(tmp),sep=""))
                }
            }
            
        }
    }
    if ("stats" %in% exportWhat) {
        if (!is.null(cnt))
            conds <- strsplit(cnt,"_vs_")[[1]]
        else
            conds <- names(sampleList)
        for (cond in conds) {
            if ("normalized" %in% exportValues) {
                if ("mean" %in% exportStats) {
                    disp("      binding normalized mean counts...")
                    tmp <- makeStat(sampleList[[cond]],normList,"mean",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report) 
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_normalized_mean_counts_",cond,sep=""))
                }
                if ("median" %in% exportStats) {
                    disp("      binding normalized median counts...")
                    tmp <- makeStat(sampleList[[cond]],normList,"median",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report) 
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_normalized_median_counts_",cond,sep=""))
                }
                if ("sd" %in% exportStats) {
                    disp("      binding normalized count sds...")
                    tmp <- makeStat(sampleList[[cond]],normList,"sd",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report) 
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_normalized_sd_counts_",cond,sep=""))
                }
                if ("mad" %in% exportStats) {
                    disp("      binding normalized count MADs...")
                    tmp <- makeStat(sampleList[[cond]],normList,"mad",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_normalized_mad_counts_",cond,sep=""))
                }
                if ("cv" %in% exportStats) {
                    disp("      binding normalized count CVs...")
                    tmp <- makeStat(sampleList[[cond]],normList,"cv",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_normalized_cv_counts_",cond,sep=""))
                }
                if ("rcv" %in% exportStats) {
                    disp("      binding normalized counts RCVs...")
                    tmp <- makeStat(sampleList[[cond]],normList,"rcv",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_normalized_rcv_counts_",cond,sep=""))
                }
            }
            if ("raw" %in% exportValues) {
                if ("mean" %in% exportStats) {
                    disp("      binding raw mean counts...")
                    tmp <- makeStat(sampleList[[cond]],rawList,"mean",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_raw_mean_counts_",cond,sep=""))
                }
                if ("median" %in% exportStats) {
                    disp("      binding raw median counts...")
                    tmp <- makeStat(sampleList[[cond]],rawList,"median",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_raw_median_counts_",cond,sep=""))
                }
                if ("sd" %in% exportStats) {
                    disp("      binding raw counts sds...")
                    tmp <- makeStat(sampleList[[cond]],rawList,"sd",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_raw_sd_counts_",cond,sep=""))
                }
                if ("mad" %in% exportStats) {
                    disp("      binding raw counts MADs...")
                    tmp <- makeStat(sampleList[[cond]],rawList,"mad",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_raw_mad_counts_",cond,sep=""))
                }
                if ("cv" %in% exportStats) {
                    disp("      binding raw counts CVs...")
                    tmp <- makeStat(sampleList[[cond]],rawList,"cv",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_raw_cv_counts_",cond,sep=""))
                }
                if ("rcv" %in% exportStats) {
                    disp("      binding raw counts RCVs...")
                    tmp <- makeStat(sampleList[[cond]],rawList,"rcv",
                        exportScale)
                    export <- cbind(export,tmp)
                    if (report)
                        exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                    theNames <- c(theNames,paste(colnames(tmp),
                        "_raw_rcv_counts_",cond,sep=""))
                }
            }
        }
    }
    if ("counts" %in% exportWhat) {
        if (!is.null(cnt))
            conds <- strsplit(cnt,"_vs_")[[1]]
        else
            conds <- names(sampleList)
        for (cond in conds) {
            if ("normalized" %in% exportValues) {                    
                disp("      binding all normalized counts for ",cond,"...")
                tmp <- makeMatrix(sampleList[[cond]],normList,exportScale)
                export <- cbind(export,tmp)
                if (report)
                    exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                part1 <- rep(paste(exportScale,"_normalized_counts_",sep=""),
                    each=length(sampleList[[cond]]))
                part2 <- paste(part1,colnames(tmp),sep="")
                theNames <- c(theNames,part2)
            }
            if ("raw" %in% exportValues) {
                disp("      binding all raw counts for ",cond,"...")
                tmp <- makeMatrix(sampleList[[cond]],rawList,exportScale)
                export <- cbind(export,tmp)
                if (report)
                    exportHtml <- cbind(exportHtml,.makeHtmlCells(tmp))
                part1 <- rep(paste(exportScale,"_raw_counts_",sep=""),
                    each=length(sampleList[[cond]]))
                part2 <- paste(part1,colnames(tmp),sep="")
                theNames <- c(theNames,part2)
            }
        }
    }
    if ("flags" %in% exportWhat && !is.null(flags)) {
        disp("      binding filtering flags...")
        export <- cbind(export,as.data.frame(flags))
        if (report)
            exportHtml <- cbind(exportHtml,.makeHtmlCells(flags))
        theNames <- c(theNames,colnames(flags))
    }
    names(export) <- theNames

    if (!report)
        exportHtml <- NULL

    return (list(textTable=export,htmlTable=exportHtml,headers=theNames))
}

statDeseq <- function(object,sampleList,contrastList=NULL,statArgs=NULL) {
    if (is.null(statArgs) && !is(object,".CountDataSet"))
        statArgs <- getDefaults("statistics","deseq")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
            collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    classes <- asClassVector(sampleList)
    theDesign <- data.frame(condition=classes,row.names=colnames(object))
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)
    # Check if there is no replication anywhere
    if (all(vapply(sampleList,function(x) ifelse(length(x)==1,TRUE,FALSE),
        logical(1)))) {
        warnwrap("No replication detected! There is a possibility that ",
            "DESeq will fail to estimate dispersions...")
        methodDisp <- "blind"
        sharingModeDisp <- "fit-only"
        fitTypeDisp <- "local"
    }
    else {
        methodDisp <- statArgs$method
        sharingModeDisp <- statArgs$sharingMode
        fitTypeDisp <- statArgs$fitType
    }
    switch(class(object)[1],
        .CountDataSet = { # Has been normalized with DESeq
            cds <- object
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp,fitType=fitTypeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp,fitType=fitTypeDisp)
        },
        DESeqDataSet = { # Has been normalized with DESeq2
            #cds <- newCountDataSet(round(DESeq::counts(object,
            #   normalized=TRUE)),
            #    theDesign$condition)
            cds <- newCountDataSet(round(counts(object,normalized=TRUE)),
                theDesign$condition)
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp)
        },
        DGEList = { # Has been normalized with edgeR
            # Trick found at http://cgrlucb.wikispaces.com/edgeR+spring2013
            scl <- object$samples$lib.size * object$samples$norm.factors
            cds <- newCountDataSet(round(t(t(object$counts)/scl)*mean(scl)),
                theDesign$condition)
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp)
        },
        matrix = { # Has been normalized with EDASeq or NOISeq
            cds <- newCountDataSet(object,theDesign$condition)
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp)
        },
        nbData = { # Has been normalized with NBPSeq and main method was 
            # "nbpseq"
            cds <- newCountDataSet(as.matrix(round(sweep(object$counts,2,
                object$norm.factors,"*"))),theDesign$condition)
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp)
        },
        nbp = { # Has been normalized with NBPSeq and main method was 
            # "nbsmyth"...
            cds <- newCountDataSet(as.matrix(round(object$pseudo.counts)),
                theDesign$condition)
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp)
        },
        ABSDataSet = { # Has been normalized with ABSSeq
            cds <- newCountDataSet(as.matrix(round(excounts(object)),
                theDesign$condition))
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp)
        },
        SeqCountSet = { # Has been normalized with DSS
            cds <- newCountDataSet(as.matrix(round(assayData(object)$exprs),
                theDesign$condition))
            #DESeq::sizeFactors(cds) <- normalizationFactor(object)
            sizeFactors(cds) <- normalizationFactor(object)
            #cds <- DESeq::estimateDispersions(cds,method=methodDisp,
            #    sharingMode=sharingModeDisp)
            cds <- estimateDispersions(cds,method=methodDisp,
                sharingMode=sharingModeDisp)
        }
    )
    for (conName in names(contrastList)) {
        disp("  Contrast: ", conName)
        con <- contrastList[[conName]]
        cons <- unique(unlist(con))
        if (length(con)==2) {
            res <- nbinomTest(cds,cons[1],cons[2])
            p[[conName]] <- res$pval
        }
        else {
            #cind <- match(cons,theDesign$condition)
            #if (any(is.na(cind)))
            #    cind <- cind[which(!is.na(cind))]
            cc <- names(unlist(con))
            cdsTmp <- cds[,cc]
            fit0 <- fitNbinomGLMs(cdsTmp,count~1)
            fit1 <- fitNbinomGLMs(cdsTmp,count~condition)
            p[[conName]] <- nbinomGLMTest(fit1,fit0)
        }
        names(p[[conName]]) <- rownames(object)
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    return(p)
}

statDeseq2 <- function(object,sampleList,contrastList=NULL,statArgs=NULL) {
    if (is.null(statArgs))
        statArgs <- getDefaults("statistics","deseq2")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
        collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    
    classes <- asClassVector(sampleList)
    # Transform and rename classes it into a factor so as to use it to construct
    # a colData DataFrame
    conditions <- as.factor(classes)                 
    # colData constructed based on the sampleList given by the user 
    colData <- DataFrame(conditions)
    # Design constructed based on the colData made in the above line
    design <- as.formula(c("~", names(colData[1]))) 
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)

    switch(class(object)[1],
        DESeqDataSet = { # Has been normalized with DESeq2
            dds <- object
            dds <- DESeq2::estimateDispersions(dds,fitType=statArgs$fitType, 
                maxit=statArgs$maxit,quiet=statArgs$quiet,
                modelMatrix=statArgs$modelMatrix)                   
        },
        .CountDataSet = { # Has been normalized with DESeq
            #countData <- round(DESeq::counts(object, normalized=TRUE))
            countData <- round(counts(object, normalized=TRUE))
            dds <- DESeqDataSetFromMatrix(countData,colData,design=design,
                tidy=statArgs$tidy) 
            DESeq2::sizeFactors(dds) <- rep(1,ncol(countData))
            dds= DESeq2::estimateDispersions(dds,fitType=statArgs$fitType, 
                    maxit=statArgs$maxit,quiet=statArgs$quiet, 
                    modelMatrix=statArgs$modelMatrix)
        },
        DGEList = { # Has been normalized with edgeR              
            dds <- DESeqDataSetFromMatrix(object$counts,colData,
                design=design,tidy=statArgs$tidy)
            # edgeR has already normalized for library size. So what is done
            # is to pass size factors from edgeR to sizeFactors of DESeq2
            DESeq2::sizeFactors(dds) <- object$samples$norm.factors 
            dds <- DESeq2::estimateDispersions(dds,fitType=statArgs$fitType, 
                maxit=statArgs$maxit,quiet=statArgs$quiet, 
                modelMatrix=statArgs$modelMatrix)
        },
        SeqExpressionSet = { # Has been normalized with EDASeq
            # A matrix with the normalized counts (both within- and 
            # between-lane)
            countData <- normCounts(object) 
            dds <- DESeqDataSetFromMatrix(countData,colData,design=design,
                tidy=statArgs$tidy)
            # Because EDAseq has done library size norm, we set all 
            # sizeFactors equal to 1
            DESeq2::sizeFactors(dds) <- rep(1,ncol(dds)) 
            dds <- DESeq2::estimateDispersions(dds,fitType=statArgs$fitType,
                maxit=statArgs$maxit,quiet=statArgs$quiet,
                modelMatrix=statArgs$modelMatrix)
        },
        matrix = { # Has been normalized with EDASeq or NOISeq or ABSSeq
            # I take the matrix ouput of these tools that contains 
            # normalized counts
            dds <- DESeqDataSetFromMatrix(object,colData,design=design,
                tidy=statArgs$tidy)
            # Because EDAseq has done library size norm (and NOISeq too, 
            # and EdgeR too), we set all sizeFactors equal 1
            DESeq2::sizeFactors(dds) <- rep(1,ncol(dds))
            dds <- DESeq2::estimateDispersions(dds,fitType=statArgs$fitType,
                maxit=statArgs$maxit,quiet=statArgs$quiet,
                modelMatrix=statArgs$modelMatrix)
        },
        nbData = { # Normalized with NBPSeq and main method was "nbpseq".
            # Using sweep to multiply raw counts with the respective 
            # norm.factors -> normCounts generated
            dds <- DESeqDataSetFromMatrix(
                round(sweep(object$counts,2,object$norm.factors,"*")),
                colData,design=design,tidy=statArgs$tidy) 
            DESeq2::sizeFactors(dds)= rep(1,ncol(dds))
            dds <- DESeq2::estimateDispersions(dds,fitType=statArgs$fitType,
                maxit=statArgs$maxit,quiet=statArgs$quiet,
                modelMatrix=statArgs$modelMatrix)
        },
        nbp = { # Normalized with NBPSeq and main method was "nbsmyth"; 
            #"nbsmyth" refers to prepare.nbp. 
            # We take the pseudocounts of the nbp object because they are the 
            # normalized ones for the lib.size
            dds <- DESeqDataSetFromMatrix(round(object$pseudo.counts),
                colData,design=design,tidy=statArgs$tidy) 
            DESeq2::sizeFactors(dds)= rep(1,ncol(dds))
            dds <- DESeq2::estimateDispersions(dds,fitType=statArgs$fitType,
                maxit=statArgs$maxit,quiet=statArgs$quiet,
                modelMatrix=statArgs$modelMatrix)
        },
        ABSDataSet = { # Has been normalized with ABSSeq
            dds <- DESeqDataSetFromMatrix(round(excounts(object)),colData,
                design=design,tidy=statArgs$tidy)
            DESeq2::sizeFactors(dds) <- rep(1,ncol(dds))
            dds <- DESeq2::estimateDispersions(dds,fitType=statArgs$fitType,
                maxit=statArgs$maxit,quiet=statArgs$quiet,
            modelMatrix=statArgs$modelMatrix)
        },
        SeqCountSet = { # Has been normalized with DSS
            dds <- DESeqDataSetFromMatrix(round(assayData(object)$exprs),
                colData,design=design,tidy=statArgs$tidy)
            DESeq2::sizeFactors(dds)=normalizationFactor(object)
            dds <-  DESeq2::estimateDispersions(dds,fitType=statArgs$fitType,
                maxit=statArgs$maxit,quiet=statArgs$quiet,
                modelMatrix=statArgs$modelMatrix)
        }
    )
    
    for (conName in names(contrastList)) {
        disp(" Contrast: ",conName)
        con <- contrastList[[conName]]
        cons <- unique(unlist(con))
        # nbinomWaldTest performs ALL possible pairwise comparisons between the 
        # levels of the factor specified in the dds's design.
        dds <- nbinomWaldTest(dds,betaPrior=statArgs$betaPrior,
            modelMatrix=statArgs$modelMatrix,betaTol=statArgs$betaTol, 
            maxit=statArgs$maxit,useOptim=statArgs$useOptim,
            quiet=statArgs$quiet,useT=statArgs$useT, 
            useQR=statArgs$useQR)
        # Then, one has just to ask for the results of the comparison he wants.
        if (length(con)==2) {
            # Because in conditions factor may be >2 levels, in a pairwise 
            # comparison I have to specify which results to show
            res <- DESeq2::results(dds,contrast=c("conditions",cons[[1]],
                cons[[2]]))
            p[[conName]] <- res$pvalue
        }
        else { # DE analysis of >2 levels in conditions factor
            cc <- names(unlist(con))
            # This assignment, except from keeping only the samples we want to 
            # compare, it also keeps only the respective levels under 
            # conditions!
            ddsTmp <- dds[,cc] 
            ddsTmp <- nbinomLRT(ddsTmp,full=design,reduced=~1,
                betaTol=statArgs$betaTol,maxit=statArgs$maxit,
                useOptim=statArgs$useOptim,quiet=statArgs$quiet, 
                useQR=statArgs$useQR)
            # Using the LRT to compare all the levels inside a factor 
            # (like an ANOVA)
            res <- DESeq2::results(ddsTmp)
            p[[conName]] <- res$pvalue
        }
        names(p[[conName]]) <- rownames(object)
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    
    return(p)
}                 

statEdger <- function(object,sampleList,contrastList=NULL,statArgs=NULL) {
    if (is.null(statArgs))
        statArgs <- getDefaults("statistics","edger")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
            collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    classes <- asClassVector(sampleList)
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)
    switch(class(object)[1],
        .CountDataSet = { # Has been normalized with DESeq
            #dge <- DGEList(counts=DESeq::counts(object,normalized=TRUE),
            #    group=classes)
            dge <- DGEList(counts=counts(object,normalized=TRUE),group=classes)
        },
        DESeqDataSet = { # Has been normalized with DESeq2
            dge <- DGEList(counts=DESeq2::counts(object,normalized=TRUE),
                group=classes)
        },  
        DGEList = { # Has been normalized with edgeR
            dge <- object    
        },
        matrix = { # Has been normalized with EDASeq or NOISeq or ABSSeq
            dge <- DGEList(object,group=classes)
        },
        nbData = { # Has been normalized with NBPSeq and method was "nbpseq"
            dge <- DGEList(counts=as.matrix(round(sweep(object$counts,2,
                object$norm.factors,"*"))),group=classes)
        },
        nbp = { # Has been normalized with NBPSeq and main method was "nbsmyth"
            dge <- DGEList(counts=as.matrix(round(object$pseudo.counts)),
                group=classes)
        },
        ABSDataSet = { # Has been normalized with ABSSeq
            dge <- DGEList(counts=as.matrix(round(excounts(object))),
                group=classes)
        },
        SeqCountSet = { # Has been normalized with DSS
            dge <- DGEList(counts=as.matrix(round(assayData(object)$exprs)),
                norm.factors=normalizationFactor(object),group=classes)
        }
    )
    # Dispersion estimate step
    # Check if there is no replication anywhere
    repli = TRUE
    if (all(vapply(sampleList,function(x) ifelse(length(x)==1,TRUE,FALSE),
        logical(1)))) {
        warnwrap("No replication when testing with edgeR! Consider using ",
            "another statistical test or just performing empirical analysis. ",
            "Setting bcv to 0.2...")
        repli <- FALSE
        bcv <- 0.2
    }
    if (repli) {
        if (statArgs$main.method=="classic") {
            dge <- estimateCommonDisp(dge,rowsum.filter=statArgs$rowsum.filter)
            dge <- estimateTagwiseDisp(dge,prior.df=statArgs$prior.df,
                trend=statArgs$trend,span=statArgs$span,
                    method=statArgs$tag.method,
                grid.length=statArgs$grid.length,
                    grid.range=statArgs$grid.range)
        }
        else if (statArgs$main.method=="glm") {
            design <- model.matrix(~0+classes,data=dge$samples)
            dge <- estimateGLMCommonDisp(dge,design=design,
                offset=statArgs$offset,
                method=statArgs$glm.method,subset=statArgs$subset,
                AveLogCPM=statArgs$AveLogCPM)
            dge <- estimateGLMTrendedDisp(dge,design=design,
                offset=statArgs$offset,
                method=statArgs$trend.method,AveLogCPM=statArgs$AveLogCPM)
            dge <- estimateGLMTagwiseDisp(dge,design=design,
                offset=statArgs$offset,
                dispersion=statArgs$dispersion,prior.df=statArgs$prior.df,
                span=statArgs$span,AveLogCPM=statArgs$AveLogCPM)
        }
    }
    # Actual statistical test
    for (conName in names(contrastList)) {
        disp("  Contrast: ", conName)
        con <- contrastList[[conName]]
        if (length(con)==2) {
            if (repli) {
                if (statArgs$main.method=="classic") {
                    res <- exactTest(dge,pair=unique(unlist(con)))
                }
                else if (statArgs$main.method=="glm") {
                    s <- unlist(con)
                    us <- unique(s)
                    ms <- match(names(s),rownames(dge$samples))
                    if (any(is.na(ms)))
                        ms <- ms[which(!is.na(ms))]
                    design <- model.matrix(~0+s,data=dge$samples[ms,])
                    colnames(design) <- us
                    #fit <- glmFit(dge[,ms],design=design,
                    #    offset=statArgs$offset,
                    #    weights=statArgs$weights,lib.size=statArgs$lib.size,
                    #    prior.count=statArgs$prior.count,
                    #    start=statArgs$start,method=statArgs$method)
                    fit <- glmFit(dge[,ms],design=design,
                        prior.count=statArgs$prior.count,
                        start=statArgs$start,method=statArgs$method)
                    co <- makeContrasts(paste(us[2],us[1],sep="-"),
                        levels=design)
                    lrt <- glmLRT(fit,contrast=co)
                    res <- topTags(lrt,n=nrow(dge))
                }
            }
            else {
                if (statArgs$main.method=="classic") {
                    res <- exactTest(dge,pair=unique(unlist(con)),
                        dispersion=bcv^2)
                }
                else if (statArgs$main.method=="glm") {
                    s <- unlist(con)
                    us <- unique(s)
                    ms <- match(names(s),rownames(dge$samples))
                    if (any(is.na(ms)))
                        ms <- ms[which(!is.na(ms))]
                    design <- model.matrix(~0+s,data=dge$samples[ms,])
                    colnames(design) <- us
                    #fit <- glmFit(dge[,ms],design=design,
                    #    offset=statArgs$offset,weights=statArgs$weights,
                    #    lib.size=statArgs$lib.size,
                    #    prior.count=statArgs$prior.count,
                    #    start=statArgs$start,
                    #    method=statArgs$method,dispersion=bcv^2)
                    fit <- glmFit(dge[,ms],design=design,dispersion=bcv^2,
                        prior.count=statArgs$prior.count,start=statArgs$start,
                        method=statArgs$method)
                    co <- makeContrasts(paste(us[2],us[1],sep="-"),
                        levels=design)
                    lrt <- glmLRT(fit,contrast=co)
                    res <- topTags(lrt,n=nrow(dge))
                }
            }
        }
        else { # GLM only
            s <- unlist(con)
            us <- unique(s)
            #design <- model.matrix(~0+s,data=dge$samples) # Ouch!
            ms <- match(names(s),rownames(dge$samples))
            if (any(is.na(ms)))
                ms <- ms[which(!is.na(ms))]
            design <- model.matrix(~s,data=dge$samples[ms,])
            if (repli)
                #fit <- glmFit(dge[,ms],design=design,offset=statArgs$offset,
                #    weights=statArgs$weights,lib.size=statArgs$lib.size,
                #    prior.count=statArgs$prior.count,start=statArgs$start,
                #    method=statArgs$method)
                fit <- glmFit(dge[,ms],design=design,start=statArgs$start,
                    prior.count=statArgs$prior.count)
            else
                #fit <- glmFit(dge[,ms],design=design,offset=statArgs$offset,
                #    weights=statArgs$weights,lib.size=statArgs$lib.size,
                #    prior.count=statArgs$prior.count,start=statArgs$start,
                #    method=statArgs$method,dispersion=bcv^2)
                #    dispersion=bcv^2)
                fit <- glmFit(dge[,ms],design=design,dispersion=bcv^2,
                    prior.count=statArgs$prior.count,start=statArgs$start)
                #lrt <- glmLRT(fit,coef=2:ncol(fit$design))
                lrt <- glmLRT(fit,coef=2:ncol(fit$design))
                res <- topTags(lrt,n=nrow(dge))
        }
        p[[conName]] <- res$table[,"PValue"]
        names(p[[conName]]) <- rownames(res$table)
        p[[conName]] <- p[[conName]][rownames(dge)]
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    return(p)
}

statLimma <- function(object,sampleList,contrastList=NULL,statArgs=NULL) {
    if (is.null(statArgs))
        statArgs <- getDefaults("statistics","limma")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
            collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    classes <- asClassVector(sampleList)
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)
    switch(class(object)[1],
        .CountDataSet = { # Has been normalized with DESeq
            #dge <- DGEList(DESeq::counts(object,normalized=TRUE),group=classes)
            dge <- DGEList(counts(object,normalized=TRUE),group=classes)
        },
        DESeqDataSet = { # Has been normalized with DESeq2
            dge <- DGEList(DESeq2::counts(object,normalized=TRUE),group=classes)
        },
        DGEList = { # Has been normalized with edgeR
            dge <- object
        },
        matrix = { # Has been normalized with EDASeq or NOISeq or EDASeq
            dge <- DGEList(object,group=classes)
        },
        nbData = { # Has been normalized with NBPSeq and method was "nbpseq"
            dge <- DGEList(counts=as.matrix(round(sweep(object$counts,2,
                object$norm.factors,"*"))),group=classes)
        },
        nbp = { # Has been normalized with NBPSeq and main method was "nbsmyth"
            dge <- DGEList(counts=as.matrix(round(object$pseudo.counts)),
                group=classes)
        },
        ABSDataSet = { # Has been normalized with ABSSeq
            dge <- DGEList(counts=as.matrix(round(excounts(object))),
                group=classes)
        },
        SeqCountSet = { # Has been normalized with DSS
            dge <- DGEList(counts=as.matrix(round(assayData(object)$exprs)),
                norm.factors=normalizationFactor(object),group=classes)
        }
    )
    for (conName in names(contrastList)) {
        disp("  Contrast: ", conName)
        con <- contrastList[[conName]]
        s <- unlist(con)
        us <- unique(s)
        ms <- match(names(s),rownames(dge$samples))
        if (any(is.na(ms)))
            ms <- ms[which(!is.na(ms))]
        if (length(con)==2) {
            design <- model.matrix(~0+s,data=dge$samples[ms,])
            colnames(design) <- us
            vom <- voom(dge[,ms],design,
                normalize.method=statArgs$normalize.method)
            fit <- lmFit(vom,design)
            fit <- eBayes(fit)
            co <- makeContrasts(contrasts=paste(us[2],us[1],sep="-"),
                levels=design)
            fit <- eBayes(contrasts.fit(fit,co))
            p[[conName]] <- fit$p.value[,1]
        }
        else {
            design <- model.matrix(~s,data=dge$samples[ms,])
            vom <- voom(dge[,ms],design,
                normalize.method=statArgs$normalize.method)
            fit <- lmFit(vom,design)
            fit <- eBayes(fit)
            res <- topTable(fit,coef=2:ncol(fit$design),number=nrow(vom))
            p[[conName]] <- res[,"P.Value"]
            names(p[[conName]]) <- rownames(res)
            p[[conName]] <- p[[conName]][rownames(dge)]
        }
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    return(p)
}

statNoiseq <- function(object,sampleList,contrastList=NULL,statArgs=NULL,
    geneData=NULL,logOffset=1) {
    if (is.null(statArgs))
        statArgs <- getDefaults("statistics","noiseq")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
            collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    if (!is.null(geneData) && is(geneData,"GenomicRanges")) {
        gl <- NULL
        if (!is.null(attr(geneData,"geneLength")))
            gl <- attr(geneData,"geneLength")
        geneData <- as.data.frame(geneData)
        geneData <- geneData[,c(1,2,3,6,7,5,8,9)]
        if (!is.null(gl))
            attr(geneData,"geneLength") <- gl
    }
    if (is.null(geneData)) {
        gcContent <- NULL
        chromosome <- NULL
        biotype <- NULL
        geneLength <- NULL
    }
    else {
        gcContent <- geneData$gc_content
        if (is.null(geneData$gc_content))
            gcContent <- rep(0.5,nrow(geneData))
        biotype <- as.character(geneData$biotype)
        names(gcContent) <- names(biotype) <- rownames(geneData)
        if (is.null(attr(geneData,"geneLength")))
            geneLength <- NULL
        else {
            geneLength <- attr(geneData,"geneLength")
            names(geneLength) <- rownames(geneData)
        }
    }
    classes <- asClassVector(sampleList)
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)
    switch(class(object)[1],
        .CountDataSet = { # Has been normalized with DESeq
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                #data=DESeq::counts(object,normalized=TRUE),
                data=counts(object,normalized=TRUE),
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        },
        DESeqDataSet = { # Has been normalized with DESeq2
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                data=DESeq2::counts(object,normalized=TRUE),
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        },
        DGEList = { # Has been normalized with edgeR
            # Trick found at http://cgrlucb.wikispaces.com/edgeR+spring2013
            scl <- object$samples$lib.size * object$samples$norm.factors
            dm <- round(t(t(object$counts)/scl)*mean(scl))            
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                data=dm,
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        },
        ExpressionSet = { # Has been normalized with NOISeq
            nsObj <- object
        },
        matrix = { # Has been normalized with EDASeq
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                data=object,
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        },
        nbData = { # Has been normalized with NBPSeq and main method "nbpseq"
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                data=as.matrix(round(sweep(object$counts,2,
                    object$norm.factors,"*"))),
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        },
        nbp = { # Has been normalized with NBPSeq and main method was "nbsmyth"
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                data=as.matrix(round(object$pseudo.counts)),
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        },
        ABSDataSet = { # Has been normalized with ABSSeq
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                data=as.matrix(round(excounts(object))),
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        },
        SeqCountSet = { # Has been normalized with DSS; 
            # Because NOISeq class does not have any slot for norm.factors and 
            # because DSS doesn't have any function to return the normalized 
            # counts on their own, I will transform the SeqCountSet into a cds 
            # object and then I will get the normalized counts
            theDesign <- data.frame(condition=classes,
                row.names=colnames(object))
            cds <- newCountDataSet(as.matrix(round(assayData(object)$exprs)),
                theDesign$condition)
            #DESeq::sizeFactors(cds) <- normalizationFactor(object)
            sizeFactors(cds) <- normalizationFactor(object)
            #nsObj <- NOISeq::readData(
            nsObj <- readData(
                #data=DESeq::counts(cds,normalized=TRUE),
                data=counts(cds,normalized=TRUE),
                length=geneLength,
                gc=gcContent,
                chromosome=geneData[,c(1,2,3)],
                factors=data.frame(class=classes),
                biotype=biotype
            )
        }
    )
    for (conName in names(contrastList)) {
        disp("  Contrast: ", conName)
        con <- contrastList[[conName]]
        if (length(con)==2) {
            statArgs$conditions=unique(unlist(con))
            if (any(vapply(sampleList,function(x) ifelse(length(x)==1,
                TRUE,FALSE),logical(1))))
                # At least one condition does not have replicates
                statArgs$replicates <- "no" 
            if (statArgs$replicates %in% c("technical","no"))
                res <- noiseq(nsObj,k=logOffset,norm="n",
                    replicates=statArgs$replicates,factor=statArgs$factor,
                    conditions=statArgs$conditions,pnr=statArgs$pnr,
                    nss=statArgs$nss,v=statArgs$v,lc=statArgs$lc)
            else
                res <- noiseqbio(nsObj,k=logOffset,norm="n",
                    nclust=statArgs$nclust,factor=statArgs$factor,
                    lc=statArgs$lc,conditions=statArgs$conditions,
                    r=statArgs$r,adj=statArgs$adj,a0per=statArgs$a0per,
                    cpm=statArgs$cpm,
                    filter=statArgs$filter,depth=statArgs$depth,
                    cv.cutoff=statArgs$cv.cutoff)
            # Beware! This is not the classical p-value!
            p[[conName]] <- 1 - res@results[[1]]$prob
        }
        else {
            warnwrap(paste("NOISeq differential expression algorithm does not ",
                "support multi-factor designs (with more than two conditions ",
                "to be compared)! Switching to DESeq for this comparison: ",
                conName))
            M <- assayData(nsObj)$exprs
            cds <- newCountDataSet(round(M),data.frame(condition=unlist(con),
                row.names=names(unlist(con))))
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method="blind",
            #    sharingMode="fit-only")
            cds <- estimateDispersions(cds,method="blind",
                sharingMode="fit-only")
            fit0 <- fitNbinomGLMs(cds,count~1)
            fit1 <- fitNbinomGLMs(cds,count~condition)
            p[[conName]] <- nbinomGLMTest(fit1,fit0)
        }
        names(p[[conName]]) <- rownames(nsObj)
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    return(p)
}

#statBayseq <- function(object,sampleList,contrastList=NULL,statArgs=NULL,
#    libsizeList=NULL) {
#    if (is.null(statArgs))
#        statArgs <- getDefaults("statistics","bayseq")
#    if (is.null(contrastList))
#        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
#            collapse="_vs_"),sampleList)
#    if (!is.list(contrastList))
#        contrastList <- makeContrastList(contrastList,sampleList)
#    classes <- asClassVector(sampleList)
#    p <- vector("list",length(contrastList))
#    names(p) <- names(contrastList)
#    switch(class(object)[1],
#        .CountDataSet = { # Has been normalized with DESeq
#            #bayesData <- DESeq::counts(object,normalized=TRUE)
#            bayesData <- counts(object,normalized=TRUE)
#        },
#        DESeqDataSet = { # Has been normalized with DESeq2
#            bayesData <- DESeq2::counts(object,normalized=TRUE)
#        },  
#        DGEList = { # Has been normalized with edgeR
#            scl <- object$samples$lib.size * object$samples$norm.factors
#            bayesData <- round(t(t(object$counts)/scl)*mean(scl))
#        },
#        matrix = { # Has been normalized with EDASeq or NOISeq or ABSSeq
#            bayesData <- object
#        },
#        nbData = {
#            bayesData <- as.matrix(round(sweep(object$counts,2,
#                object$norm.factors,"*")))
#        },
#        nbp = {
#            bayesData <- as.matrix(round(object$pseudo.counts))
#        },
#        ABSDataSet = {
#            bayesData <- as.matrix(round(excounts(object)))
#        },
#        SeqCountSet = { # Again the same trick I did back in NOISeq
#            theDesign <- data.frame(condition=classes,
#                row.names=colnames(object))
#            cds <- newCountDataSet(as.matrix(round(assayData(object)$exprs)),
#                theDesign$condition)
#            #DESeq::sizeFactors(cds) <- normalizationFactor(object)
#            sizeFactors(cds) <- normalizationFactor(object)
#            #bayesData <- as.matrix(DESeq::counts(cds,normalized=TRUE))
#            bayesData <- as.matrix(counts(cds,normalized=TRUE))
#        }
#    )
#    CD <- new("countData",data=bayesData,replicates=classes)
#    CD@annotation <- data.frame(name=rownames(bayesData))
#    if (is.null(libsizeList))
#        baySeq::libsizes(CD) <- baySeq::getLibsizes(CD)
#    else
#        baySeq::libsizes(CD) <- unlist(libsizeList)
#    for (conName in names(contrastList)) {
#        disp("  Contrast: ", conName)
#        con <- contrastList[[conName]]
#        #cd <- CD[,names(unlist(con))]
#        cd <- CD[,match(names(unlist(con)),colnames(CD@data))]
#        if (length(con)==2)
#            baySeq::groups(cd) <- list(NDE=rep(1,length(unlist(con))),
#                DE=c(rep(1,length(con[[1]])),rep(2,length(con[[2]]))))
#        else
#            baySeq::groups(cd) <- list(NDE=rep(1,length(unlist(con))),
#                DE=unlist(con,use.names=FALSE)) # Maybe this will not work
#        baySeq::replicates(cd) <- as.factor(classes[names(unlist(con))])
#        cd <- baySeq::getPriors.NB(cd,samplesize=statArgs$samplesize,
#            samplingSubset=statArgs$samplingSubset,
#            equalDispersions=statArgs$equalDispersions,
#            estimation=statArgs$estimation,zeroML=statArgs$zeroML,
#            consensus=statArgs$consensus,cl=statArgs$cl)
#        cd <- baySeq::getLikelihoods(cd,pET=statArgs$pET,
#            marginalise=statArgs$marginalise,subset=statArgs$subset,
#            priorSubset=statArgs$priorSubset,bootStraps=statArgs$bootStraps,
#            conv=statArgs$conv,nullData=statArgs$nullData,
#            returnAll=statArgs$returnAll,returnPD=statArgs$returnPD,
#            discardSampling=statArgs$discardSampling,cl=statArgs$cl)
#        tmp <- baySeq::topCounts(cd,group="DE",number=nrow(cd))
#        #p[[conName]] <- 1 - as.numeric(tmp[,"Likelihood"])
#        p[[conName]] <- 1 - as.numeric(tmp[,"likes"])
#        names(p[[conName]]) <- as.character(tmp$name)
#        p[[conName]] <- p[[conName]][rownames(CD@data)]
#        p[[conName]][which(is.na(p[[conName]]))] <- 1
#    }
#    return(p)
#}

statNbpseq <- function(object,sampleList,contrastList=NULL,statArgs=NULL,
    libsizeList=NULL) {
    if (is.null(statArgs) && !is(object,"list"))
        statArgs <- getDefaults("statistics","nbpseq")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
            collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    classes <- asClassVector(sampleList)
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)
    switch(class(object)[1],
        .CountDataSet = { # Has been normalized with DESeq
            #counts <- round(DESeq::counts(object,normalized=TRUE))
            counts <- round(counts(object,normalized=TRUE))
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(counts[,n])
            }
            libSizes <- unlist(libsizeList)
        },
        DESeqDataSet = { # Has been normalized with DESeq2
            counts <- round(DESeq2::counts(object,normalized=TRUE))
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(counts[,n])
            }
            libSizes <- unlist(libsizeList)
        },
        DGEList = { # Has been normalized with edgeR
            # Trick found at http://cgrlucb.wikispaces.com/edgeR+spring2013
            scl <- object$samples$lib.size * object$samples$norm.factors
            counts <- round(t(t(object$counts)/scl)*mean(scl))
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(counts[,n])
            }
            libSizes <- unlist(libsizeList)
        },
        matrix = { # Has been normalized with EDASeq or NOISeq or ABSSeq
            counts <- object
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(counts[,n])
            }
            libSizes <- unlist(libsizeList)
        },
        nbData = { # Has been normalized with NBPSeq
            object$counts <- as.matrix(object$counts)
            nbData <- object
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(nbData$counts[,n])
            }
            libSizes <- unlist(libsizeList)
            nbData$pseudo.lib.sizes=rep(1e+7,dim(object$counts)[2])
        },
        nbp = { # Same...
            object$pseudo.counts <- as.matrix(object$pseudo.counts)
            nbData <- object
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(nbData$counts[,n])
            }
            libSizes <- unlist(libsizeList)
        },
        ABSDataSet = { # Has been normalized with ABSSeq
            counts <- as.matrix(round(excounts(object)))
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(counts[,n])
            }
            libSizes <- unlist(libsizeList)
        },
        SeqCountSet = { # Again the trick of NOISeq
            theDesign <- data.frame(condition=classes,
                row.names=colnames(object)) 
            cds <- newCountDataSet(as.matrix(round(assayData(object)$exprs)),
                theDesign$condition)
            #DESeq::sizeFactors(cds) <- normalizationFactor(object)
            sizeFactors(cds) <- normalizationFactor(object)
            #counts <- as.matrix(DESeq::counts(cds,normalized=TRUE))
            counts <- as.matrix(counts(cds,normalized=TRUE))
            if (is.null(libsizeList)) {
                libsizeList <- vector("list",length(classes))
                names(libsizeList) <- unlist(sampleList,use.names=FALSE)
                for (n in names(libsizeList))
                    libsizeList[[n]] <- sum(counts[,n])
            }
            libSizes <- unlist(libsizeList)
        }
    )
    # To avoid repeating the following chunk in the above
    if ((!is(object,"list") && !is(object,"nbp")) || is(object,"DGEList")) {
        #if (statArgs$main.method=="nbpseq") {
        #    nbData <- list(
        #        counts=as.matrix(counts),
        #        lib.sizes=libSizes,
        #        norm.factors=rep(1,dim(counts)[2]),
        #        eff.lib.sizes=libSizes*rep(1,dim(counts)[2]),
        #        rel.frequencies=as.matrix(sweep(counts,2,
        #            libSizes*rep(1,dim(counts)[2]),"/")),
        #        tags=matrix(row.names(counts),dim(counts)[1],1)
        #    )
        #}
        #else if (statArgs$main.method=="nbsmyth") {
        #    nbData <- new("nbp",list(
        #        counts=as.matrix(counts),
        #        lib.sizes=libSizes,
        #        grp.ids=classes,
        #        eff.lib.sizes=libSizes*rep(1,dim(counts)[2]),
        #        pseudo.counts=as.matrix(counts),
        #        pseudo.lib.sizes=colSums(as.matrix(counts))*rep(1,
        #            dim(counts)[2])
        #    ))
            nbData <- list(
                counts=as.matrix(counts),
                lib.sizes=libSizes,
                grp.ids=classes,
                eff.lib.sizes=libSizes*rep(1,dim(counts)[2]),
                pseudo.counts=as.matrix(counts),
                #pseudo.lib.sizes=colSums(as.matrix(counts)) *
                #    rep(1,dim(counts)[2])
                pseudo.lib.sizes=rep(1e+7,dim(counts)[2])
            )
            class(nbData) <- "nbp"
        #}
    }
    for (conName in names(contrastList)) {
        disp("  Contrast: ", conName)
        con <- contrastList[[conName]]
        cons <- unique(unlist(con))
        if (length(con)==2) {
            #if (statArgs$main.method=="nbpseq") {
            #    dispersions <- estimate.dispersion(nbData,
            #    model.matrix(~classes),
            #        method=statArgs$method$nbpseq)
            #    res <- test.coefficient(nbData,dispersion=dispersions,
            #        x=model.matrix(~classes),beta0=c(NA,0),
            #        tests=statArgs$tests,
            #        alternative=statArgs$alternative,print.level=1)
            #    #res <- nb.glm.test(nbData$counts,x=model.matrix(~classes),
            #        beta0=c(NA,0),libSizes=libSizes,
            #    #    dispersion.method=statArgs$method$nbpseq,
            #         tests=statArgs$tests)
            #    p[[conName]] <- res[[statArgs$tests]]$p.values
            #    #p[[conName]] <- res$test[[statArgs$tests]]$p.values
            #}
            #else if (statArgs$main.method=="nbsmyth") {
                obj <- suppressWarnings(estimate.disp(nbData,
                    model=statArgs$model$nbsmyth,print.level=0))
                obj <- exact.nb.test(obj,cons[1],cons[2],print.level=0)
                p[[conName]] <- obj$p.values
            #}
        }
        else {
            warnwrap(paste("NBPSeq differential expression algorithm does not ",
                "support ANOVA-like designs with more than two conditions to ",
                "be compared! Switching to DESeq for this comparison:",
                conName))
            cds <- newCountDataSet(nbData$counts,
                data.frame(condition=unlist(con),
                row.names=names(unlist(con))))
            #DESeq::sizeFactors(cds) <- rep(1,ncol(cds))
            sizeFactors(cds) <- rep(1,ncol(cds))
            #cds <- DESeq::estimateDispersions(cds,method="blind",
            #    sharingMode="fit-only")
            cds <- estimateDispersions(cds,method="blind",
                sharingMode="fit-only")
            fit0 <- fitNbinomGLMs(cds,count~1)
            fit1 <- fitNbinomGLMs(cds,count~condition)
            p[[conName]] <- nbinomGLMTest(fit1,fit0)
        }
        names(p[[conName]]) <- rownames(nbData$counts)
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    return(p)
}

statAbsseq <- function(object,sampleList,contrastList=NULL,statArgs=NULL) {
    if (is.null(statArgs))
        statArgs <- getDefaults("statistics","absseq")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
        collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)

    # In statAbsseq "switch()" is inside "for" loop. Actually, there are two 
    # different "switch()" sections inside the loop. This happens because this 
    # tool takes only 2-level factors for analysis. There is a feature where we
    # can give a factor with more than 2 levels, but then the tool-specific
    # object must be created without the "groups" argument (see tools 
    # reference manual and vignette) and the function called for differential 
    # expression differs from that of the 2-level factors. This is why we place 
    # switch inside the "for" loop, in order to give the user the freedom to use
    # the tool to check between 2 or more levels of a factor during the same 
    # analysis.
    
    for(conName in names(contrastList)){
    disp(" Contrast: ",conName)
    con <- contrastList[[conName]]
    cons= unique(unlist(con))
        if(length(con)==2){
            # Checking if we have a pairwise comparison, because then we must 
            # resize a sampleList to only the 2 levels of interest and pass them
            # in the groups argument of ABSDataSet().
            sampleListTmp <- sampleList[cons]
            # In any other case, of length(groups)!=2 it will through an ERROR
            classes <- asClassVector(sampleListTmp)
            # Sample description as wanted by ABSDataSet for pairwise 
            # comparisons
            groupsTmp <- as.factor(classes)

            # Check for replication
            # We cannot do anything for the case that there is no replication, 
            # but to call separately callParameterWithoutReplicates (which by
            # the way has no other args except from the ABSDataSet) instead of 
            # callParameter and then continue with callDEs
            # All the above are automatically done by ABSSeq()
            # However, ABSSeqlm() cannot run without providing replicates (see
            # further down for another comment)
            switch(class(object)[1],
                ABSDataSet = {
                    abs <- ABSDataSet(excounts(object)[,names(unlist(con))],
                        groupsTmp,normMethod="user",
                        # "user" so as to pass my own sizeFactors
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },
                .CountDataSet = { # Has been normalized with DESeq
                    #countsTmp <- DESeq::counts(object)
                    countsTmp <- counts(object)
                    abs <- ABSDataSet(countsTmp[,names(unlist(con))],
                        # "user" so as to pass my own sizeFactors
                        groupsTmp,normMethod="user",
                        sizeFactor=sizeFactors(object)[names(unlist(con))], 
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                        # Here I used the DESeq's raw counts and size factors
                        # because that way I got a little better p-values
                },
                DESeqDataSet = { # Has been normalized with DESeq2
                    countsTmp <- DESeq2::counts(object)
                    abs <- ABSDataSet(countsTmp[,names(unlist(con))],
                        # "user" so as to pass my own sizeFactors
                        groupsTmp,normMethod= "user",
                        sizeFactor=sizeFactors(object)[names(unlist(con))], 
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC) 
                        # Here I used the DESeq's raw counts and size factors
                        #  because that way I got a little better p-values 
                },
                DGEList = { # Has been normalized with edgeR                
                    abs <- ABSDataSet(object$counts[,names(unlist(con))],
                        groupsTmp,normMethod="user", 
                        sizeFactor=object$samples[names(unlist(con)),3],
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },
                SeqExpressionSet = { # Has been normalized with EDASeq
                    abs <- ABSDataSet(normCounts(object)[,
                        names(unlist(con))],groupsTmp,normMethod= "user", 
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },  
                matrix = { # Has been normalized with EDASeq or NOISeq or Edger
                    abs <- ABSDataSet(object[,names(unlist(con))],groupsTmp,
                        normMethod= "user",
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC) 
                        # Because EDAseq has done library size norm (and 
                        # NOISeq too), we set all sizeFactors equal to 1
                },              
                nbData = { # Has been normalized with NBPSeq and main method was
                    # "nbpseq"; 
                    # "nbpseq" refers to prepare.nbData that seems to have 
                    # stopped running in metaseqR (see norm.R lines 366-368)
                    normCounts=round(sweep(object$counts,2,
                        object$norm.factors,"*"))
                    abs <- ABSDataSet(object[,names(unlist(con))],groupsTmp,
                        normMethod="user", 
                        sizeFactor= rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },
                nbp = { # Has been normalized with NBPSeq and main method was 
                    #"nbsmyth"; 
                    #"nbsmyth" refers to prepare.nbp (see norm.R lines 370-371) 
                    abs <- ABSDataSet(round(object$pseudo.counts[,
                        names(unlist(con))]),groupsTmp,normMethod= "user", 
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC) 
                        # I take the pseudocounts of the nbp object because 
                        # they are the normalized ones for the lib.size
                },
                SeqCountSet = {
                    # isolate normalization factors so as to give them names and
                    # be able to subset them at will
                    nF <- normalizationFactor(object) 
                    names(nF) <- names(unlist(sampleList))
                    abs <- ABSDataSet(round(
                        assayData(object)$exprs[,names(unlist(con))]),
                        groupsTmp,normMethod= "user", 
                        sizeFactor=nF[names(unlist(con))],
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                }
            )
        
            abs <- ABSSeq(abs,adjmethod=statArgs$adjmethod,
                replaceOutliers=statArgs$replaceOutliers, 
                useaFold=statArgs$useaFold,quiet= statArgs$quiet) 
                # It linearly calls normalFactors(),callParameter(),callDEs(); 
                # normalFactors() will use those given.
            res <- ABSSeq::results(abs)
            p[[conName]] <- res[,"pvalue"]
            names(p[[conName]]) <- rownames(res) 
            # ATTENTION: p[[con.names]] had to be named by the subsetted count 
            # table of each contrast and not at the end. If not thus there was a
            # problem with line 2975 of main.R
        }
        else { 
            # Here we build the ABSDataSet object WITHOUT the groups argument so
            # as to perform ANOVA. Otherwise, if I try to build a groups
            # argument with more than 2 levels to run the ANOVA, it will through
            # an ERROR 
            sampleListTmp <- sampleList[cons]
            classes <- asClassVector(sampleListTmp)
            groupsTmp <- as.factor(classes)
            design <- model.matrix(~0+groupsTmp)

            switch(class(object)[1],
                ABSDataSet = {
                    abs <- ABSDataSet(excounts(object)[,names(unlist(con))],
                        normMethod="user", 
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },
                .CountDataSet = { # Has been normalized with DESeq
                    #countsTmp <- DESeq::counts(object)
                    countsTmp <- counts(object)
                    abs <- ABSDataSet(countsTmp[,names(unlist(con))],
                        normMethod="user", 
                        sizeFactor=sizeFactors(object)[names(unlist(con))],
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },
                DESeqDataSet = { # Has been normalized with DESeq2
                        countsTmp= DESeq2::counts(object)
                    abs <- ABSDataSet(countsTmp[,names(unlist(con))],
                        normMethod="user", 
                        sizeFactor=sizeFactors(object)[names(unlist(con))],
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                        # Here we use the DESeq's raw counts and size factors
                        # because that way I got a little better p-values 
                },
                DGEList = { # Has been normalized with edgeR                
                    abs <- ABSDataSet(object$counts[,names(unlist(con))],
                        normMethod="user",
                        sizeFactor=object$samples[names(unlist(con)),3],
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },
                SeqExpressionSet = { # Has been normalized with EDASeq
                    abs <- ABSDataSet(
                        EDASeq::normCounts(object)[,names(unlist(con))],
                        normMethod="user",
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)              
                },  
                matrix = { # Has been normalized with EDASeq or NOISeq
                    abs <- ABSDataSet(object[,names(unlist(con))],
                        normMethod="user", 
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC) 
                        # Because EDAseq has done library size norm (and 
                        # NOISeq too), we set all sizeFactors equal to 1
                },
                nbData = { 
                    # Has been normalized with NBPSeq and main method was 
                    # "nbpseq"; 
                    # "nbpseq" refers to prepare.nbData that seems to have 
                    # stopped running in metaseqR (see norm.R lines 366-368)
                    normCounts <- round(sweep(object$counts,2,
                        object$norm.factors,"*"))
                    abs <- ABSDataSet(object[,names(unlist(con))],
                        normMethod="user",
                        sizeFactor=rep(1,length(unlist(con))),
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                },
                nbp = { # Has been normalized with NBPSeq and main method was
                    # "nbsmyth"; 
                    # "nbsmyth" refers to prepare.nbp 
                    # (see norm.R lines 370-371) 
                    abs <- ABSDataSet(round(
                        object$pseudo.counts[,names(unlist(con))]),
                            normMethod="user", 
                            sizeFactor=rep(1,length(unlist(con))),
                            paired=statArgs$paired,
                            minDispersion=statArgs$minDispersion, 
                            minRates=statArgs$minRates,
                            maxRates=statArgs$maxRates, 
                            LevelstoNormFC=statArgs$LevelstoNormFC) 
                            # I take the pseudocounts of the nbp object because
                            # they are the normalized ones for the lib.size
                },
                SeqCountSet = {
                    # isolate normalization factors so as to give them names and
                    # be able to subset them at will
                    nF <- normalizationFactor(object) 
                    names(nF) <- names(unlist(sampleList))
                    abs <- ABSDataSet(round(assayData(
                        object)$exprs[,names(unlist(con))]),
                        normMethod="user", 
                        sizeFactor=nF[names(unlist(con))],
                        paired=statArgs$paired,
                        minDispersion=statArgs$minDispersion, 
                        minRates=statArgs$minRates,
                        maxRates=statArgs$maxRates, 
                        LevelstoNormFC=statArgs$LevelstoNormFC)
                }
            )   
            abs <- ABSSeqlm(abs,design=design,condA=colnames(design),
                lmodel=statArgs$lmodel,preval=statArgs$preval,
                qforkappa=statArgs$qforkappa,adjmethod=statArgs$adjmethod,
                scale=statArgs$scale,quiet=statArgs$quiet) 
                # It performs ANOVA because I have passed the design to condA 
                # and condB is NULL (default;I don't let the user to change it).
                # normalFactors() and aFoldcomplexDesign() are called in row by 
                # this function
            p[[conName]] <- abs[,"pvalue"]
            names(p[[conName]]) <- rownames(abs)
        }           
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    return(p)       
}

statDss <- function(object,sampleList,contrastList=NULL,statArgs=NULL) {
    if (is.null(statArgs))
        statArgs <- getDefaults("statistics","dss")
    if (is.null(contrastList))
        contrastList <- makeContrastList(paste(names(sampleList)[c(1,2)],
            collapse="_vs_"),sampleList)
    if (!is.list(contrastList))
        contrastList <- makeContrastList(contrastList,sampleList)
    
    classes <- asClassVector(sampleList)
    design <- as.data.frame(classes)
    # if colname is not designs then it throughs an error
    colnames(design) <- "designs" 
    p <- vector("list",length(contrastList))
    names(p) <- names(contrastList)
    
    switch(class(object)[1],
        .CountDataSet = { # Has been normalized with DESeq
            #countData <- round(DESeq::counts(object, normalized=TRUE))
            countData <- round(counts(object, normalized=TRUE))
            seqData <- newSeqCountSet(countData,design,
            normalizationFactor=rep(1,ncol(countData)))
            # estimating dispersions
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        },
        DESeqDataSet = { # Has been normalized with DESeq2
            countData <- round(DESeq2::counts(object,normalized=TRUE))
            seqData <- newSeqCountSet(countData,design,
            normalizationFactor=rep(1,ncol(countData)))
            # estimating dispersions
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        },
        DGEList = { # Has been normalized with edgeR              
            seqData <- newSeqCountSet(object$counts,design,
                normalizationFactor=object$samples$norm.factors)
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        },
        SeqExpressionSet = { # Has been normalized with EDASeq
            seqData <- newSeqCountSet(normCounts(object),design,
                normalizationFactor=rep(1,ncol(normCounts))) 
            # Getting the normalized counts and passing in normFactors the 
            # value 1
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        },
        matrix = { # Has been normalized with EDASeq or NOISeq
            seqData <- newSeqCountSet(object,design,
            normalizationFactor=rep(1,ncol(object)))
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        },
        nbData ={ # Has been normalized with NBPSeq and method was "nbpseq".
            #seqData <- newSeqCountSet(round(sweep(object$counts,2,
            #   object$norm.factors,"*")),design,
            #   normalizationFactor=rep(1,ncol(object$counts)))
            seqData <- newSeqCountSet(normCounts(object),design,
                normalizationFactor=rep(1,ncol(object)))
            # Getting the normalized counts and passing in normFactors 
            # the value 1
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        }, 
        nbp = { # Has been normalized with NBPSeq and method was "nbsmyth".
            seqData <- newSeqCountSet(round(object$pseudo.counts),design,
                normalizationFactor=rep(1,ncol(object$pseudo.counts)))
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        },
        ABSDataSet = { # Has been normalized with ABSSeq
            seqData <- newSeqCountSet(round(excounts(object)),design,
                normalizationFactor=rep(1,ncol(object$pseudo.counts)))
            seqData <- estDispersion(seqData,trend=statArgs$trend)
        },
        SeqCountSet = {
            seqData <- object
            seqData <- estDispersion(seqData,trend=statArgs$trend)       
        }
    )

    for (conName in names(contrastList)) {
        disp(" Contrast: ",conName)
        # It assignes all element values of contrastList to con (number of 
        # conditions)
        con <- contrastList[[conName]]
        # It unlists con and picks the unique values
        cons <- unique(unlist(con)) 

        if(length(con)==2) { # DE analysis of 2 conditions
            # We choose from all levels the two that will be used for the 
            # pairwise comparison: "cons[1],cons[2]"
            res <- waldTest(seqData,cons[1],cons[2],
                equal.var=statArgs$equal.var)
            # Order according to initial object! waldTest returns results sorted
            # in decreasing p-values...
            res <- res[rownames(object),,drop=FALSE]
            p[[conName]] <- res$pval
        }
        else { # DE analysis of >2 levels in conditions factor
            warnwrap(paste("DSS differential expression algorithm does not ",
                "support multi-level designs (with more than two levels in a ",
                "factor to be compared)! Switching to DESeq. ",
                "Comparison: ",conName))
            statArgs <- getDefaults("statistics","deseq")
            colData <- DataFrame(design)
            colnames(colData) <- "conditions"
            designTmp <- as.formula(c("~",names(colData[1])))
            
            # The same chunck with the norm.R script to get a cds file
            theDesign <- data.frame(condition=classes,
                row.names=colnames(seqData)) 
            cds <- newCountDataSet(as.matrix(round(assayData(seqData)$exprs)),
                conditions=theDesign$condition)
            # retrieve DSS-calculated normalizationFactors
            #DESeq::sizeFactors(cds) <- normalizationFactor(seqData)     
            sizeFactors(cds) <- normalizationFactor(seqData)     
            #cds <- DESeq::estimateDispersions(cds,method=statArgs$method,
            #        sharingMode=statArgs$sharingMode)
            cds <- estimateDispersions(cds,method=statArgs$method,
                    sharingMode=statArgs$sharingMode)

            cc <- names(unlist(con))
            cdsTmp <- cds[,cc]
            fit0 <- fitNbinomGLMs(cdsTmp,count~1)
            fit1 <- fitNbinomGLMs(cdsTmp,count~condition)
            p[[conName]] <- nbinomGLMTest(fit1,fit0)
        }
        names(p[[conName]]) <- rownames(object)
        p[[conName]][which(is.na(p[[conName]]))] <- 1
    }
    return(p)
}                  

## ----install-0, eval=FALSE, echo=TRUE-----------------------------------------
# if(!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# BiocManager::install("metaseqR2")

## ----load-library, echo=TRUE, eval=TRUE, message=FALSE, warning=FALSE---------
library(metaseqR2)

## ----seed-0, eval=TRUE, echo=TRUE---------------------------------------------
set.seed(42)

## ----data-1, eval=TRUE, echo=TRUE---------------------------------------------
data("mm9GeneData",package="metaseqR2")

## ----head-1, eval=TRUE, echo=TRUE---------------------------------------------
head(mm9GeneCounts)

## ----random-1, eval=TRUE, echo=TRUE-------------------------------------------
sampleListMm9

## ----random-2, eval=TRUE, echo=TRUE-------------------------------------------
libsizeListMm9

## ----example-1, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
library(metaseqR2)

data("mm9GeneData",package="metaseqR2")

# You can explore the results in the session's temporary directory
print(tempdir())

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    embedCols=list(
        idCol=4,
        gcCol=5,
        nameCol=8,
        btCol=7
    ),
    org="mm9",
    countType="gene",
    normalization="edger",
    statistics="edger",
    pcut=0.05,
    qcPlots=c(
        "mds","filtered","correl","pairwise","boxplot","gcbias",
        "lengthbias","meandiff","meanvar","deheatmap","volcano",
        "mastat"
    ),
    figFormat=c("png","pdf"),
    exportWhat=c("annotation","p_value","adj_p_value","fold_change"),
    exportScale=c("natural","log2"),
    exportValues="normalized",
    exportStats=c("mean","sd","cv"),
    exportWhere=file.path(tempdir(),"test1"),
    restrictCores=0.01,
    geneFilters=list(
         length=list(
                length=500
         ),
         avgReads=list(
                averagePerBp=100,
                quantile=0.25
         ),
         expression=list(
                median=TRUE,
                mean=FALSE,
                quantile=NA,
                known=NA,
                custom=NA
         ),
         biotype=getDefaults("biotypeFilter","mm9")
    ),
    outList=TRUE
)

## ----head-2, eval=TRUE, echo=TRUE---------------------------------------------
head(result[["data"]][["adult_8_weeks_vs_e14.5"]])

## ----example-2, eval=TRUE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE----
library(metaseqR2)

data("mm9GeneData",package="metaseqR2")

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    embedCols=list(
        idCol=4,
        gcCol=5,
        nameCol=8,
        btCol=7
    ),
    org="mm9",
    countType="gene",
    whenApplyFilter="prenorm",
    normalization="edaseq",
    statistics=c("deseq","edger"),
    metaP="fisher",
    #qcPlots=c(
    #    "mds","biodetection","countsbio","saturation","readnoise","filtered",
    #    "correl","pairwise","boxplot","gcbias","lengthbias","meandiff",
    #    "meanvar","rnacomp","deheatmap","volcano","mastat","biodist","statvenn"
    #),
    qcPlots=c(
        "mds","filtered","correl","pairwise","boxplot","gcbias",
        "lengthbias","meandiff","meanvar","deheatmap","volcano",
        "mastat"
    ),
    restrictCores=0.01,
    figFormat=c("png","pdf"),
    preset="medium_normal",
    exportWhere=file.path(tempdir(),"test2"),
    outList=TRUE
)

## ----example-3, eval=TRUE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE----
library(metaseqR2)

data("mm9GeneData",package="metaseqR2")

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    embedCols=list(
        idCol=4,
        gcCol=5,
        nameCol=8,
        btCol=7
    ),
    org="mm9",
    countType="gene",
    normalization="edaseq",
    statistics=c("deseq","edger"),
    metaP="fisher",
    qcPlots=c(
        "mds","filtered","correl","pairwise","boxplot","gcbias",
        "lengthbias","meandiff","meanvar","deheatmap","volcano",
        "mastat"
    ),
    restrictCores=0.01,
    figFormat=c("png","pdf"),
    preset="medium_normal",
    outList=TRUE,
    exportWhere=file.path(tempdir(),"test3")
)

## ----example-4, eval=FALSE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE----
# library(metaseqR2)
# 
# data("mm9GeneData",package="metaseqR2")
# 
# result <- metaseqr2(
#     counts=mm9GeneCounts,
#     sampleList=sampleListMm9,
#     contrast=c("adult_8_weeks_vs_e14.5"),
#     libsizeList=libsizeListMm9,
#     annotation="embedded",
#     embedCols=list(
#         idCol=4,
#         gcCol=5,
#         nameCol=8,
#         btCol=7
#     ),
#     org="mm9",
#     countType="gene",
#     normalization="edaseq",
#     statistics=c("edger","limma"),
#     metaP="fisher",
#     figFormat="png",
#     preset="medium_basic",
#     qcPlots=c(
#         "mds","filtered","correl","pairwise","boxplot","gcbias",
#         "lengthbias","meandiff","meanvar","deheatmap","volcano",
#         "mastat"
#     ),
#     restrictCores=0.01,
#     outList=TRUE,
#     exportWhere=file.path(tempdir(),"test4")
# )

## ----example-5, eval=FALSE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE----
# data("mm9GeneData",package="metaseqR2")
# weights <- estimateAufcWeights(
#     counts=as.matrix(mm9GeneCounts[,9:12]),
#     normalization="edaseq",
#     statistics=c("edger","limma"),
#     nsim=1,N=10,ndeg=c(2,2),top=4,modelOrg="mm39",
#     rc=0.01,libsizeGt=1e+5
# )

## ----head-3, eval=TRUE, echo=TRUE---------------------------------------------
weights

## ----example-6, eval=TRUE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE----
data("hg19pvalues",package="metaseqR2")

# Examine the data
head(hg19pvalues)

# Now combine the p-values using the Simes method
pSimes <- apply(hg19pvalues,1,combineSimes)

# The harmonic mean method with PANDORA weights
w <- getWeights("human")
pHarm <- apply(hg19pvalues,1,combineHarmonic,w)

# The PANDORA method
pPandora <- apply(hg19pvalues,1,combineWeight,w)

## ----help-2, eval=TRUE, echo=TRUE, message=FALSE------------------------------
help(statEdger)

## ----si-1, eval=TRUE, echo=TRUE-----------------------------------------------
sessionInfo()

mdsToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- out[1]
    
    x <- obj$x
    y <- obj$y
    xlim <- obj$xlim
    ylim <- obj$ylim
    samples <- obj$samples
    cols <- .getColorScheme()
    
    # Construct series
    counter <- 0
    series <- vector("list",length(samples))
    names(series) <- names(samples)
    for (n in names(series)) {
        counter <- counter + 1
        series[[n]] <- list()
        series[[n]]$name=n
        series[[n]]$type="scatter"
        series[[n]]$color=cols$fill[counter]
        series[[n]]$marker <- list(
            lineWidth=1,
            states=list(
                hover=list(
                    enabled=TRUE,
                    lineColor=cols$border[counter]
                ),
                select=list(
                    fillColor=cols$selected[counter],
                    lineColor=cols$border[counter],
                    lineWidth=2
                )
            )
        )
        m <- match(samples[[n]],names(x))
        if (length(m)>0) {            
            series[[n]]$data <- makeHighchartsPoints(x[m],y[m])
        }
    }
    
    switch(jl,
        highcharts = {
            point.format=paste("<strong>Sample name: </strong>{point.name}<br>",
                "<strong>Principal coordinate 1: </strong>{point.x}<br>",
                "<strong>Principal coordinate 2: </strong>{point.y}",sep="")
                
                json <- list(
                    chart=list(
                        type="scatter",
                        zoomType="xy"
                    ),
                    title=list(
                        text=paste("Multidimensional Scaling")
                    ),
                    xAxis=list(
                        title=list(
                            useHTML=TRUE,
                            text="1<sup>st</sup> Principal Coordinate",
                            margin=20,
                            style=list(
                                color="#000000",
                                fontSize="1.2em"
                            )
                        ),
                        labels=list(
                            style=list(
                                color="#000000",
                                fontSize="1.1em",
                                fontWeight="bold"
                            )
                        ),
                        startOnTick=TRUE,
                        endOnTick=TRUE,
                        showLastLabel=TRUE,
                        gridLineWidth=1,
                        min=round(xlim[1],3),
                        max=round(xlim[2],3)
                    ),
                    yAxis=list(
                        title=list(
                            useHTML=TRUE,
                            text="2<sup>nd</sup> Principal Coordinate",
                            margin=25,
                            style=list(
                                color="#000000",
                                fontSize="1.2em"
                            )
                        ),
                        labels=list(
                            style=list(
                                color="#000000",
                                fontSize="1.1em",
                                fontWeight="bold"
                            )
                        ),
                        startOnTick=TRUE,
                        endOnTick=TRUE,
                        showLastLabel=TRUE,
                        gridLineWidth=1,
                        min=round(ylim[1],3),
                        max=round(ylim[2],3)
                    ),
                    plotOptions=list(
                        scatter=list(
                            allowPointSelect=TRUE,
                            tooltip=list(
                                headerFormat=paste("<span style=",
                                    "\"font-size:1.1em;color:{series.color};",
                                    "font-weight:bold\">{series.name}<br>",
                                    sep=""),
                                pointFormat=point.format
                            )
                        )
                    ),
                    series=unname(series)
                )
        }
    )
    
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else
        return(json)
}

countsBioToJSON <- function(obj,by=c("sample","biotype"),jl=c("highcharts"),
    out=c("json","list")) {
    by <- tolower(by[1])
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    samples <- obj$samples
    status <- obj$status
    altnames <- obj$altnames
    counts <- round(2^obj$user$counts - 1)
    #counts[counts==0] <- 0.001
    counts <- counts + 1 # Offset for log
    #counts <- obj$user$counts
    
    covars <- obj$user$covars
    biotypes <- unique(as.character(covars$biotype))
    if (!is.null(altnames))
        names(altnames) <- rownames(counts)
    
    grouped <- FALSE
    if (is.null(samples)) {
        if (is.null(colnames(counts)))
            samplenames <- paste("Sample",seq_len(ncol(counts)),sep=" ")
        else
            samplenames <- colnames(counts)
        samples <- list(Samples=samplenames)
    }
    else if (is.list(samples)) {
        samplenames <- unlist(samples,use.names=FALSE)
        grouped <- TRUE
    }
    
    # y label formatter for logarithmic axis
    y.label.formatter <- paste('function() {if(this.value === 0.001)',
        '{return 0;} else {return Highcharts.Axis.prototype.',
        'defaultLabelFormatter.call(this);}}',sep="")
        
    tooltip.point.formatter <- paste("function() {",
        "   var min = this.low === 0.001 ? 0 : this.low;" ,
        "   var q1 = this.q1 === 0.001 ? 0 : this.q1;" ,
        "   var med = this.median === 0.001 ? 0 : this.median;",
        "   var q3 = this.q3 === 0.001 ? 0 : this.q3;",
        "   var max = this.high === 0.001 ? 0 : this.high;",
        "   var str = 'Maximum: ' + max + '<br/>' +",
        "       'Upper quartile: ' + q3 + '<br/>' +",
        "       'Median: ' + med + '<br/>' +",
        "       'Lower quartile: ' + q1 + '<br/>' +",
        "       'Minimum: ' + min + '<br/>';",
        "   return  str;",
        "}",sep="")
        
    # Legend clicker
    boxplot.onclick <- paste("function() { ",
        "   var chart =  this.chart;",
        "   var outlier_id =  chart.get(this.name);",
        "   if (!outlier_id.visible) {",
        "       outlier_id.show();",
        "   } else {",
        "       outlier_id.hide();",
        "   }",
        "}",sep="")
    
    # Outliers tooltip
    if (is.null(obj$altnames)) {
        outlier.pointformat <- paste(
            '<strong>Sample {point.category}</strong><br/>',
            'Gene ID: {point.name}<br/>',
            'Value: {point.y}<br/>',sep=""
        )
    }
    else {
        outlier.pointformat <- paste(
            '<strong>Sample {point.category}</strong><br/>',
            'Gene ID: {point.name}<br/>',
            'Gene name: {point.alt_name}<br/>',
            'Value: {point.y}<br/>',
            sep=""
        )
    }
    
    if (by=="sample") {
        cols <- .getColorScheme(length(biotypes))
        boxList <- json <- vector("list",length(samplenames))
        names(boxList) <- names(json) <- samplenames
        for (n in samplenames) {
            boxList[[n]] <- vector("list",length(biotypes))
            names(boxList[[n]]) <- biotypes
            for (b in biotypes)
                boxList[[n]][[b]] <- counts[covars$biotype==b,n]
            
            B <- boxplot(boxList[[n]],plot=FALSE)$stats
            colnames(B) <- biotypes
            oList <- lapply(names(boxList[[n]]),function(x,M,b) {
                v <- b[,x]
                o <- which(M[[x]]<v[1] | M[[x]]>v[5])
                if (length(o)>0)
                    return(M[[x]][o])
                else
                    return(NULL)
            },boxList[[n]],B)
            names(oList) <- biotypes
    
            # Data series
            BB <- matrix(0,nrow(B),ncol(B)) # Workaround of strange problem...
            colnames(BB) <- colnames(B)
            for (jj in seq_len(ncol(B)))
                BB[,jj] <- round(B[,jj],3)
            d <- as.data.frame(BB)
            ids <- 0:(ncol(d)-1)
            d <- rbind(ids,d)
            names(ids) <- colnames(d)
            counter <- 0
            series <- vector("list",length(biotypes))
            names(series) <- biotypes
            for (s in names(series)) {
                counter <- counter + 1
                series[[s]] <- list()
                series[[s]]$name <- s
                series[[s]]$color <- cols$fill[counter]
                series[[s]]$data <- list(unname(as.list(d[,s])))
                r <- round(d[,s])
                series[[s]]$tooltip=list(
                    pointFormat=paste('<strong>Population: ',
                        length(boxList[[n]][[s]]),'</strong><br/>',
                        'Maximum: ',r[6],'<br/>',
                        'Upper quartile: ',r[5],'<br/>',
                        'Median: ',r[4],'<br/>',
                        'Lower quartile: ',r[3],'<br/>',
                        'Minimum: ',r[2],'<br/>',sep="")
                )
            }
            
            # Outlier series (if any)
            counter <- 0
            outliers <- vector("list",length(biotypes))
            names(outliers) <- biotypes
            for (o in names(outliers)) {
                counter <- counter + 1
                outliers[[o]] <- list()
                outliers[[o]]$id <- o
                outliers[[o]]$name <- o
                outliers[[o]]$type <- "scatter"
                outliers[[o]]$showInLegend <- FALSE
                outliers[[o]]$color <- cols$fill[counter]
                outliers[[o]]$marker <- list(
                    fillColor=cols$fill[counter],
                    symbol="circle",
                    lineWidth=1,
                    lineColor=cols$border[counter]
                )
                outliers[[o]]$data <- list()
                x <- rep(d[1,o],length(oList[[o]]))
                names(x) <- names(oList[[o]])
                if (is.null(obj$altnames)) {
                    outliers[[o]]$data <- 
                        makeHighchartsPoints(x,oList[[o]])
                }
                else {
                    outliers[[o]]$data <- 
                        makeHighchartsPoints(x,oList[[o]],
                            unname(altnames[names(x)]))
                }
            }
            
            json[[n]] <- switch(jl,
                highcharts = {
                    list(
                        chart=list(
                            type="boxplot"
                        ),
                        title=list(
                            text=paste("Biotype detection for sample ",n,
                                sep="")
                        ),
                        legend=list(
                            enabled=TRUE,
                            itemHoverStyle=list(
                                color="#B40000"
                            )
                        ),
                        xAxis=list(
                            categories=biotypes,
                            title=list(
                                text="Biotype",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        yAxis=list(
                            type="logarithmic",
                            showFirstLabel=FALSE,
                            min=1e-4,
                            tickInterval=1,
                            title=list(
                                useHTML=TRUE,
                                #text="Read count (log<sub>2</sub>)",
                                text="Expression (read count)",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                ),
                                formatter=y.label.formatter
                            )
                        ),
                        plotOptions=list(
                            boxplot=list(
                                fillColor="#F0F0E0",
                                lineWidth=2,
                                medianColor="#000000",
                                medianWidth=3,
                                stemColor="#000000",
                                stemDashStyle="dash",
                                stemWidth=1,
                                whiskerColor="#000000",
                                whiskerLength="75%",
                                whiskerWidth=1,
                                grouping=FALSE,
                                turboThreshold=10000,
                                tooltip=list(
                                    headerFormat=paste(
                                        '<span style="font-size:1.1em;',
                                        'color:{series.color};',
                                        'font-weight:bold">',
                                        '\u25CF </span>',
                                        '<span style="font-size:1.1em;',
                                        'font-weight:bold">',
                                        'Biotype {series.name}</span><br/>',
                                        sep=""
                                    )
                                ),
                                events=list(
                                    legendItemClick=boxplot.onclick
                                )
                            ),
                            scatter=list(
                                allowPointSelect=TRUE,
                                turboThreshold=10000,
                                tooltip=list(
                                    headerFormat=paste(
                                        '<span style="font-weight:bold;',
                                        'color:{series.color};">',
                                        '\u25CF </span>',
                                        '<span style="font-weight:bold">',
                                        'Biotype {series.name}</span><br/>',
                                        sep=""
                                    ),
                                    pointFormat=outlier.pointformat
                                ),
                                states=list(
                                    hover=list(
                                        marker=list(
                                            enabled=FALSE
                                        )
                                    )
                                )
                            )
                        ),
                        series=c(unname(series),unname(outliers))
                    )
                }
            )
        }
        if (out=="json") {
            for (i in seq_along(json))
                #json[[i]] <- .unquote_js_fun(toJSON(json[[i]],
                #   auto_unbox=TRUE,null="null"))
                json[[i]] <- toJSON(json[[i]],auto_unbox=TRUE,null="null")
            return(json)
        }
        else
            return(json)
    }
    else if (by=="biotype") {
        cols <- .getColorScheme(length(samples))
        boxList <- json <- vector("list",length(biotypes))
        names(boxList) <- names(json) <- biotypes
        for (b in biotypes) {
            boxList[[b]] <- vector("list",length(samplenames))
            names(boxList[[b]]) <- samplenames
            for (n in samplenames)
                boxList[[b]][[n]] <- counts[covars$biotype==b,n]
            
            B <- boxplot(boxList[[b]],plot=FALSE)$stats
            colnames(B) <- samplenames
            oList <- lapply(names(boxList[[b]]),function(x,M,b) {
                v <- b[,x]
                o <- which(M[[x]]<v[1] | M[[x]]>v[5])
                if (length(o)>0)
                    return(M[[x]][o])
                else
                    return(NULL)
            },boxList[[b]],B)
            names(oList) <- samplenames
    
            # Data series
            BB <- matrix(0,nrow(B),ncol(B)) # Workaround of strange problem...
            colnames(BB) <- colnames(B)
            for (jj in seq_len(ncol(B)))
                BB[,jj] <- round(B[,jj],3)
            d <- as.data.frame(BB)
            ids <- 0:(ncol(d)-1)
            d <- rbind(ids,d)
            names(ids) <- colnames(d)
            counter <- 0
            series <- vector("list",length(samples))
            names(series) <- names(samples)
            for (s in names(series)) {
                counter <- counter + 1
                series[[s]] <- list()
                series[[s]]$name=s
                if (grouped)
                    series[[s]]$color=cols$fill[counter]
                else
                    series[[s]]$color=cols$fill[1]
                m <- match(samples[[s]],colnames(d))
                if (length(m) > 1)
                    series[[s]]$data <- unname(as.list(d[,m]))
                else
                    series[[s]]$data <- list(unname(as.list(d[,m])))
            }
            
            # Outlier series (if any)
            counter <- 0
            outliers <- vector("list",length(samples))
            names(outliers) <- names(samples)
            for (o in names(outliers)) {
                counter <- counter + 1
                outliers[[o]] <- list()
                outliers[[o]]$id <- o
                outliers[[o]]$name <- o
                outliers[[o]]$type <- "scatter"
                outliers[[o]]$showInLegend <- FALSE
                if (grouped) {
                    outliers[[o]]$color <- cols$fill[counter]
                    outliers[[o]]$marker <- list(
                        fillColor=cols$fill[counter],
                        symbol="circle",
                        lineWidth=1,
                        lineColor=cols$border[counter]
                    )
                }
                else {
                    outliers[[o]]$color <- cols$fill[1]
                    outliers[[o]]$marker <- list(
                        fillColor=cols$fill[1],
                        symbol="circle",
                        lineWidth=1,
                        lineColor=cols$border[1]
                    )
                }
                outliers[[o]]$data <- list()
                m <- match(samples[[o]],colnames(d))
                if (length(m)>0) {
                    for (i in m) {
                        x <- rep(d[1,i],length(oList[[i]]))
                        names(x) <- names(oList[[i]])
                        if (is.null(obj$altnames)) {
                            outliers[[o]]$data <- 
                                makeHighchartsPoints(x,oList[[i]])
                        }
                        else {
                            outliers[[o]]$data <- c(outliers[[o]]$data,
                                makeHighchartsPoints(x,oList[[i]],
                                unname(altnames)))
                        }
                    }
                }
            }
            
            json[[b]] <- switch(jl,
                highcharts = {
                    list(
                        chart=list(
                            type="boxplot"
                        ),
                        title=list(
                            text=paste("Detection for biotype ",b,
                                " (population: ",lengths(boxList[[b]])[1],
                                ")",sep="")
                        ),
                        legend=list(
                            enabled=TRUE
                        ),
                        xAxis=list(
                            categories=samplenames,
                            title=list(
                                text="Sample name",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        yAxis=list(
                            type="logarithmic",
                            showFirstLabel=FALSE,
                            min=1e-4,
                            tickInterval=1,
                            title=list(
                                text="Expression (read count)",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                ),
                                formatter=y.label.formatter
                            )
                        ),
                        plotOptions=list(
                            boxplot=list(
                                fillColor="#F0F0E0",
                                lineWidth=2,
                                medianColor="#000000",
                                medianWidth=3,
                                stemColor="#000000",
                                stemDashStyle="dash",
                                stemWidth=1,
                                whiskerColor="#000000",
                                whiskerLength="75%",
                                whiskerWidth=1,
                                grouping=FALSE,
                                turboThreshold=10000,
                                tooltip=list(
                                    headerFormat=paste(
                                        '<span style="font-size:1.1em;',
                                        'color:{series.color};',
                                        'font-weight:bold">',
                                        '\u25CF </span>',
                                        '<span style="font-size:1.1em;',
                                        'font-weight:bold">',
                                        'Condition {series.name}</span>',
                                        '<br/>',
                                        '<span style="font-weight:bold;">',
                                        'Sample {point.key}',
                                        '</span><br/>',sep=""
                                    ),
                                    pointFormatter=tooltip.point.formatter
                                ),
                                events=list(
                                    legendItemClick=boxplot.onclick
                                )
                            ),
                            scatter=list(
                                allowPointSelect=TRUE,
                                turboThreshold=10000,
                                tooltip=list(
                                    headerFormat=paste(
                                        '<span style="font-weight:bold;',
                                        'color:{series.color};">',
                                        '\u25CF </span>',
                                        '<span style="font-weight:bold">',
                                        'Condition {series.name}</span>',
                                        '<br/>',sep=""
                                    ),
                                    pointFormat=outlier.pointformat
                                ),
                                states=list(
                                    hover=list(
                                        marker=list(
                                            enabled=FALSE
                                        )
                                    )
                                )
                            )
                        ),
                        series=c(unname(series),unname(outliers))
                    )
                }
            )
        }
        if (out=="json") {
            for (i in seq_along(json))
                #json[[i]] <- .unquote_js_fun(toJSON(json[[i]],
                #   auto_unbox=TRUE,null="null"))
                json[[i]] <- toJSON(json[[i]],auto_unbox=TRUE,null="null")
            return(json)
        }
        else
            return(json)
    }
}

bioDetectionToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    samples <- obj$samples
    status <- obj$status
    plotdata <- obj$user$plotdata
    covars <- obj$user$covars
    
    if (!is.null(samples) && is.list(samples)) {
        samplenames <- unlist(samples,use.names=FALSE)
        names(plotdata$biotables) <- samplenames
    }
    # Otherwise we are using the names present in the input object
    
    abu <- which(plotdata$genome>7)
    nabu <- which(plotdata$genome<=7)
    
    cols <- .getColorScheme()
    json <- vector("list",length(samplenames))
    names(json) <- samplenames
    for (n in samplenames) {
        # Data series
        seriesAbu <- vector("list",3)
        names(seriesAbu) <- c("genome","detectionVSgenome","detectionVSsample")
        seriesAbu$genome <- list()
        seriesAbu$genome$id <- "abu_genome"
        seriesAbu$genome$name <- "% in genome"
        seriesAbu$genome$color <- cols$trans[1]
        seriesAbu$genome$pointPlacement <- -0.2
        seriesAbu$genome$data <- round(as.numeric(plotdata$genome[abu]),3)
        seriesAbu$detectionVSgenome <- list()
        seriesAbu$detectionVSgenome$id <- "abu_detected"
        seriesAbu$detectionVSgenome$name <- "% detected"
        seriesAbu$detectionVSgenome$color <- cols$trans[2]
        seriesAbu$detectionVSgenome$pointPlacement <- 0
        seriesAbu$detectionVSgenome$data <- round(as.numeric(
            plotdata$biotables[[n]][1,abu]),3)
        seriesAbu$detectionVSsample <- list()
        seriesAbu$detectionVSsample$id <- "abu_sample"
        seriesAbu$detectionVSsample$name <- "% in sample"
        seriesAbu$detectionVSsample$color <- cols$trans[3]
        seriesAbu$detectionVSsample$pointPlacement <- 0.2
        seriesAbu$detectionVSsample$data <- round(as.numeric(
            plotdata$biotables[[n]][2,abu]),3)
        seriesNabu <- vector("list",3)
        names(seriesNabu) <- c("genome","detectionVSgenome",
            "detectionVSsample")
        seriesNabu$genome <- list()
        seriesNabu$genome$name <- "% in genome"
        seriesNabu$genome$yAxis <- 1
        seriesNabu$genome$pointStart <- length(abu)
        seriesNabu$genome$linkedTo <- "abu_genome"
        seriesNabu$genome$color <- cols$trans[1]
        seriesNabu$genome$pointPlacement <- -0.2
        seriesNabu$genome$data <- round(as.numeric(plotdata$genome[nabu]),3)
        seriesNabu$detectionVSgenome <- list()
        seriesNabu$detectionVSgenome$name <- "% detected"
        seriesNabu$detectionVSgenome$yAxis <- 1
        seriesNabu$detectionVSgenome$pointStart <- length(abu)
        seriesNabu$detectionVSgenome$linkedTo <- "abu_detected"
        seriesNabu$detectionVSgenome$color <- cols$trans[2]
        seriesNabu$detectionVSgenome$pointPlacement <- 0
        seriesNabu$detectionVSgenome$data <- round(as.numeric(
            plotdata$biotables[[n]][1,nabu]),3)
        seriesNabu$detectionVSsample <- list()
        seriesNabu$detectionVSsample$name <- "% in sample"
        seriesNabu$detectionVSsample$yAxis <- 1
        seriesNabu$detectionVSsample$pointStart <- length(abu)
        seriesNabu$detectionVSsample$linkedTo <- "abu_sample"
        seriesNabu$detectionVSsample$color <- cols$trans[3]
        seriesNabu$detectionVSsample$pointPlacement <- 0.2
        seriesNabu$detectionVSsample$data <- round(as.numeric(
            plotdata$biotables[[n]][2,nabu]),3)
        
        json[[n]] <- switch(jl,
            highcharts = {
                list(
                    chart=list(
                        type="column",
                        alignTicks=FALSE
                    ),
                    title=list(
                        text=paste("Comparative biotype detection for ",
                            "sample ",n,sep="")
                    ),
                    legend=list(
                        enabled=TRUE,
                        itemHoverStyle=list(
                            color="#B40000"
                        )
                    ),
                    tooltip=list(
                        shared=TRUE
                    ),  
                    xAxis=list(
                        categories=names(plotdata$genome)[c(abu,nabu)],
                        title=list(
                            text="Biotype",
                            margin=25,
                            style=list(
                                color="#000000",
                                fontSize="1.2em"
                            )
                        ),
                        labels=list(
                            style=list(
                                color="#000000",
                                fontWeight="bold"
                            )
                        ),
                        plotLines=list(
                            list(
                                color="#8A8A8A",
                                width=1.5,
                                dashStyle="Dash",
                                value=length(abu)-0.5
                            )
                        ),
                        plotBands=list(
                            list(
                                color="#FFFFE0",
                                from=-0.5,
                                to=length(abu)-0.5
                            ),
                            list(
                                color="#FFECEB",
                                from=length(abu)-0.5,
                                to=length(plotdata$genome)
                            )
                        )
                    ),
                    yAxis=list(
                        list(
                            min=0,
                            max=70,
                            title=list(
                                text="% of abundant features",
                                margin=20,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),                          
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        list(
                            min=0,
                            max=7,
                            title=list(
                                text="% of non-abundant features",
                                margin=20,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),                          
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                )
                            ),
                            opposite=TRUE
                        )
                    ),
                    plotOptions=list(
                        column=list(
                            grouping=FALSE,
                            shadow=FALSE,
                            groupPadding=0.3,
                            pointPadding=0.25,
                            tooltip=list(
                                headerFormat=paste(
                                    '<span style="font-size:1.1em;',
                                    'font-weight:bold">',
                                    '{point.key}</span><br/>',sep=""
                                )
                            )
                        )
                    ),
                    series=c(unname(seriesAbu),unname(seriesNabu))
                )
            }
        )
    }
    
    if (out=="json") {
        for (i in seq_along(json))
            json[[i]] <- toJSON(json[[i]],auto_unbox=TRUE,null="null")
        return(json)
    }
    else if (out=="list")
        return(json)
}

bioSaturationToJSON <- function(obj,by=c("sample","biotype"),
    jl=c("highcharts"),out=c("json","list")) {
    by <- tolower(by[1])
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    samples <- obj$samples
    plotdata <- obj$user$plotdata
    
    if (!is.null(samples)&& is.list(samples)) {
        samplenames <- unlist(samples,use.names=FALSE)
        names(plotdata) <- samplenames
    }
    
    if (by=="sample") {
        json <- vector("list",length(samplenames))
        names(json) <- samplenames
        for (n in samplenames) {
            depth <- round(plotdata[[n]][,1]/1e+6)
            global <- round(plotdata[[n]][,2])
            M <- plotdata[[n]][,3:ncol(plotdata[[n]]),drop=FALSE]
            
            # To determine the separation
            ord <- sort(M[nrow(M),],decreasing=TRUE,index.return=TRUE)
            abu <- ord$ix[c(1,2)]
            names(abu) <- names(ord$x[c(1,2)])
            nabu <- ord$ix[3:length(ord$ix)]
            names(nabu) <- names(ord$x[3:length(ord$x)])
    
            cols <- .getColorScheme(ncol(plotdata[[n]])-1)
            counter <- 1
            
            global.series <- list(
                global=list(
                    id="global",
                    name="global",
                    color=cols$fill[counter],
                    data=makeHighchartsPoints(depth,global)
                )
            )
            
            abuSeries <- vector("list",2)
            names(abuSeries) <- names(abu)
            for (s in names(abuSeries)) {
                counter <- counter + 1
                abuSeries[[s]] <- list()
                abuSeries[[s]]$id <- s
                abuSeries[[s]]$name <- s
                abuSeries[[s]]$color <- cols$fill[counter]
                abuSeries[[s]]$data <- makeHighchartsPoints(depth,
                    round(M[,s]))
            }
            
            nabuSeries <- vector("list",length(3:ncol(M)))
            names(nabuSeries) <- names(nabu)
            for (s in names(nabuSeries)) {
                counter <- counter + 1
                nabuSeries[[s]] <- list()
                nabuSeries[[s]]$id <- s
                nabuSeries[[s]]$name <- s
                nabuSeries[[s]]$color <- cols$fill[counter]
                nabuSeries[[s]]$data <- makeHighchartsPoints(depth,
                    round(M[,s]))
            }
            
            json[[n]] <- switch(jl,
                highcharts = {
                    list(
                        chart=list(
                            type="scatter",
                            zoomType="xy"
                        ),
                        title=list(
                            text=paste("Biotype saturations for sample ",n,
                                sep="")
                        ),
                        legend=list(
                            enabled=TRUE,
                            itemHoverStyle=list(
                                color="#B40000"
                            )
                        ),
                        xAxis=list(
                            title=list(
                                text="Read depth (millions of reads)",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        yAxis=list(
                            min=0,
                            max=max(global),
                            title=list(
                                text="Detected features",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        plotOptions=list(
                            series=list(
                                lineWidth=2
                            ),
                            scatter=list(
                                tooltip=list(
                                    headerFormat=paste(
                                        '<span style="font-weight:bold;',
                                        'color:{series.color};">',
                                        '\u25CF </span>',
                                        '<span style="font-weight:bold">',
                                        'Biotype {series.name}</span><br/>',
                                        sep=""
                                    ),
                                    pointFormat=paste(
                                        "Depth: {point.x}M<br/>",
                                        "Detected features: {point.y}",
                                        sep="")
                                )
                            )
                        ),
                        series=c(unname(global.series),
                            unname(abuSeries),unname(nabuSeries))
                    )
                }
            )
        }
        
        if (out=="json") {
            for (i in seq_along(json))
                json[[i]] <- toJSON(json[[i]],auto_unbox=TRUE,null="null")
            return(json)
        }
        else if (out=="list")
            return(json)
    }
    else if (by=="biotype") {
        biotypes <- colnames(plotdata[[1]])[2:ncol(plotdata[[1]])]
        depths <- vector("list",length(plotdata))
        names(depths) <- samplenames
        for (n in samplenames)
            depths[[n]] <- round(plotdata[[n]][,1]/1e+6)
        json <- vector("list",length(biotypes))
        names(json) <- biotypes
        
        for (b in biotypes) {
            series <- vector("list",length(plotdata))
            names(series) <- samplenames
            cols <- .getColorScheme(length(samplenames))
            counter <- 0
            for (s in names(series)) {
                counter <- counter + 1
                series[[s]] <- list()
                series[[s]]$id <- s
                series[[s]]$name <- s
                series[[s]]$color <- cols$fill[counter]
                series[[s]]$data <- makeHighchartsPoints(depths[[s]],
                    round(plotdata[[s]][,b]))
            }
            
            json[[b]] <- switch(jl,
                highcharts = {
                    list(
                        chart=list(
                            type="scatter",
                            zoomType="xy"
                        ),
                        title=list(
                            text=paste("Sample saturations for biotype ",b,
                                sep="")
                        ),
                        legend=list(
                            enabled=TRUE,
                            itemHoverStyle=list(
                                color="#B40000"
                            )
                        ),
                        xAxis=list(
                            title=list(
                                text="Read depth (millions of reads)",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        yAxis=list(
                            title=list(
                                text="Detected features",
                                margin=25,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        plotOptions=list(
                            series=list(
                                lineWidth=2
                            ),
                            scatter=list(
                                tooltip=list(
                                    headerFormat=paste(
                                        '<span style="font-weight:bold;',
                                        'color:{series.color};">',
                                        '\u25CF </span>',
                                        '<span style="font-weight:bold">',
                                        'Sample {series.name}</span><br/>',
                                        sep=""
                                    ),
                                    pointFormat=paste(
                                        "Depth: {point.x}M<br/>",
                                        "Detected features: {point.y}",
                                        sep="")
                                )
                            )
                        ),
                        series=c(unname(series))
                    )
                }
            )
        }
        if (out=="json") {
            for (i in seq_along(json))
                json[[i]] <- toJSON(json[[i]],auto_unbox=TRUE,null="null")
            return(json)
        }
        else if (out=="list") {
            names(json) <- biotypes
            return(json)
        }
    }
}

readNoiseToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    d <- obj$user
    samples <- obj$samples
    
    # Too many points for a lot of curves of interactive data
    if (nrow(d)>1000) {
        ii <- sort(sample(seq_len(nrow(d)),998))
        ii <- c(1,ii,nrow(d))
        d <- cbind(d[ii,1],d[ii,2:ncol(d)])
    }

    if (is.null(samples)) 
        samples <- seq_len((ncol(d)-1))
    if (is.numeric(samples)) 
        samplenames = colnames(dat)[samples+1]
    if (is.list(samples))
        samplenames <- unlist(samples)
    
    cols <- .getColorScheme(length(samplenames))
    
    # Construct series
    counter <- 0
    series <- vector("list",length(samplenames))
    names(series) <- samplenames
    for (n in names(series)) {
        counter <- counter + 1
        series[[n]] <- list()
        series[[n]]$name=n
        series[[n]]$color=cols$fill[counter]
        series[[n]]$data <- makeHighchartsPoints(d[,1],d[,n],simple=TRUE)
        series[[n]]$tooltip=list(
            headerFormat=paste("<span style=",
                "\"font-size:1.1em;color:{series.color};",
                "font-weight:bold\">{series.name}<br>",
                sep=""),
            pointFormat=NULL
        )
    }
    
    switch(jl,
        highcharts = {
            json <- list(
                chart=list(
                    type="line",
                    zoomType="xy"
                ),
                title=list(
                    text=paste("RNA-Seq mapped reads noise")
                ),
                xAxis=list(
                    title=list(
                        text="% detected features",
                        margin=20,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=0,
                    max=100
                ),
                yAxis=list(
                    title=list(
                        text="% of total reads",
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    tickPositions=seq(0,110,10)
                ),
                plotOptions=list(
                    line=list(
                        allowPointSelect=TRUE,
                        lineWidth=1,
                        marker=list(
                            enabled=FALSE
                        ),
                        tooltip=list(
                            headerFormat=paste("<span style=",
                                "\"font-size:1.1em;color:{series.color};",
                                "font-weight:bold\">{series.name}<br>",
                                sep=""),
                            pointFormat=NULL
                        ),
                        turboThreshold=50000
                    )
                ),
                series=unname(series)
            )
        }
    )
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out=="list")
        return(json)
}

boxplotToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    b <- obj$plot
    name <- obj$samples
    status <- obj$status
    altnames <- obj$altnames
    oList <- obj$user
    
    grouped <- FALSE
    if (is.null(name)) {
        if (is.null(colnames(b$stats)))
            nams <- paste("Sample",seq_len(ncol(b$stats)),sep=" ")
        else
            nams <- colnames(b$stats)
        name <- list(Samples=nams)
    }
    else if (length(name)==1 && name=="none") {
        nams <- rep("",ncol(b$stats))
        name <- list(Samples=nams)
    }
    else if (is.list(name)) { # Is sampleList
        nams <- unlist(name,use.names=FALSE)
        grouped <- TRUE
    }
    cols <- .getColorScheme()
    
    # Data series
    d <- as.data.frame(round(b$stat,3))
    ids <- 0:(ncol(d)-1)
    d <- rbind(ids,d)
    colnames(d) <- nams
    names(ids) <- colnames(d)
    counter <- 0
    series <- vector("list",length(name))
    names(series) <- names(name)
    for (n in names(series)) {
        counter <- counter + 1
        series[[n]] <- list()
        series[[n]]$name=n
        if (grouped)
            series[[n]]$color=cols$fill[counter]
        else
            series[[n]]$color=cols$fill[1]
        m <- match(name[[n]],colnames(d))
        if (length(m) > 1)
            series[[n]]$data <- unname(as.list(d[,m]))
        else
            series[[n]]$data <- list(unname(as.list(d[,m])))
    }
    # Outlier series (if any)
    counter <- 0
    outliers <- vector("list",length(name))
    names(outliers) <- names(name)
    for (n in names(outliers)) {
        counter <- counter + 1
        outliers[[n]] <- list()
        outliers[[n]]$id <- n
        outliers[[n]]$name <- n
        outliers[[n]]$type <- "scatter"
        outliers[[n]]$showInLegend <- FALSE
        if (grouped) {
            outliers[[n]]$color <- cols$fill[counter]
            outliers[[n]]$marker <- list(
                fillColor=cols$fill[counter],
                symbol="circle",
                lineWidth=1,
                lineColor=cols$border[counter]
            )
        }
        else {
            outliers[[n]]$color <- cols$fill[1]
            outliers[[n]]$marker <- list(
                fillColor=cols$fill[1],
                symbol="circle",
                lineWidth=1,
                lineColor=cols$border[1]
            )
        }
        outliers[[n]]$data <- list()
        m <- match(name[[n]],colnames(d))
        if (length(m)>0) {
            for (i in m) {
                x <- rep(d[1,i],length(oList[[i]]))
                names(x) <- names(oList[[i]])
                outliers[[n]]$data <- c(outliers[[n]]$data,
                    makeHighchartsPoints(x,oList[[i]],unname(altnames)))
            }
        }
    }
        
    # Boxplot tooltip point formatter for the case of zeros
    tooltip.point.formatter <- paste("function() {",
        "   var min = this.low === 0.001 ? 0 : this.low;" ,
        "   var q1 = this.q1 === 0.001 ? 0 : this.q1;" ,
        "   var med = this.median === 0.001 ? 0 : this.median;",
        "   var q3 = this.q3 === 0.001 ? 0 : this.q3;",
        "   var max = this.high === 0.001 ? 0 : this.high;",
        "   var str = 'Maximum: ' + max + '<br/>' +",
        "       'Upper quartile: ' + q3 + '<br/>' +",
        "       'Median: ' + med + '<br/>' +",
        "       'Lower quartile: ' + q1 + '<br/>' +",
        "       'Minimum: ' + min + '<br/>';",
        "   return  str;",
        "}",sep="")
    
    # Legend clicker
    boxplot.onclick <- paste("function() {",
        "   var chart =  this.chart;",
        "   var outlier_id =  chart.get(this.name);",
        "   if (!outlier_id.visible) {",
        "       outlier_id.show();",
        "   } else {",
        "       outlier_id.hide();",
        "   }",
        "}",sep="")
    
    if (is.null(obj$altnames)) {
        outlier.pointformat=paste(
            '<strong>Sample {point.category}</strong><br/>',
            'Gene ID: {point.name}<br/>',
            'Value: {point.y}<br/>',sep=""
        )
    }
    else {
        outlier.pointformat=paste(
            '<strong>Sample {point.category}</strong><br/>',
            'Gene ID: {point.name}<br/>',
            'Gene name: {point.alt_name}<br/>',
            'Value: {point.y}<br/>',sep=""
        )
    }
    
    json <- switch(jl,
        highcharts = {
            list(
                chart=list(
                    type="boxplot"
                ),
                title=list(
                    text=paste("Boxplot ",status,sep="")
                ),
                legend=list(
                    enabled=TRUE
                ),
                xAxis=list(
                    categories=nams,
                    title=list(
                        text="Sample name",
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontWeight="bold"
                        )
                    )
                ),
                yAxis=list(
                    title=list(
                        useHTML=TRUE,
                        text="Read count (log<sub>2</sub>)",
                        margin=30,
                        style=list(
                            color="#000000",
                            fontSize="1.1em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    )
                ),
                plotOptions=list(
                    boxplot=list(
                        fillColor="#F0F0E0",
                        lineWidth=2,
                        medianColor="#000000",
                        medianWidth=3,
                        stemColor="#000000",
                        stemDashStyle="dash",
                        stemWidth=1,
                        whiskerColor="#000000",
                        whiskerLength="75%",
                        whiskerWidth=1,
                        grouping=FALSE,
                        tooltip=list(
                            headerFormat=paste(
                                '<span style="font-size:1.1em;',
                                'color:{series.color};',
                                'font-weight:bold">',
                                '\u25CF </span>',
                                '<span style="font-size:1.1em;',
                                'font-weight:bold">',
                                'Condition {series.name}</span><br/>',
                                '<span style="font-weight:bold">',
                                'Sample {point.key}</span><br/>',sep=""
                            ),
                            pointFormatter=tooltip.point.formatter
                        ),
                        events=list(
                            legendItemClick=boxplot.onclick
                        )
                    ),
                    scatter=list(
                        allowPointSelect=TRUE,
                        tooltip=list(
                            headerFormat=paste(
                                '<span style="font-weight:bold;',
                                'color:{series.color};">',
                                '\u25CF </span>',
                                '<span style="font-weight:bold">',
                                'Condition {series.name}</span><br/>',
                                sep=""
                            ),
                            pointFormat=outlier.pointformat
                        ),
                        states=list(
                            hover=list(
                                marker=list(
                                    enabled=FALSE
                                )
                            )
                        )
                        
                    )
                ),
                series=c(unname(series),unname(outliers))
            )
        }
    )
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else
        return(json)
}

biasPlotToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    counts <- round(nat2log(obj$user$counts),3)
    status <- obj$status
    covar <- obj$user$covar
    covarname <- obj$user$covarname
    samples <- obj$samples
    
    # Too many points for a lot of curves of interactive data
    if (nrow(counts)>2000) {
        #set.seed(seed)
        ii <- sample(seq_len(nrow(counts)),2000)
        counts <- counts[ii,]
        covar <- covar[ii]
    }
    
    # If length bias, not nice to have x-axis at -200k
    if (covarname == "lengthbias")
        minX <- ifelse(max(covar>100),0,"undefined")
    else
        minX <- 0.1

    if (is.null(samples)) {
        if (is.null(colnames(x)))
            samplenames <- paste("Sample",seq_len(ncol(counts)),sep=" ")
        else
            samplenames <- colnames(counts)
        samples <- list(Samples=samplenames)
        cols <- .getColorScheme(length(samples))
    }
    else if (is.list(samples)) { # Is sampleList
        samplenames <- unlist(samples,use.names=FALSE)
        grouped <- TRUE
        cols <- .getColorScheme(length(samplenames))
    }
    colnames(counts) <- samplenames
    
    # Construct series
    counter <- 0
    series <- vector("list",length(samplenames))
    names(series) <- samplenames
    for (n in names(series)) {
        counter <- counter + 1
        x <- counts[,n]
        fit <- lowess(covar,x)
        series[[n]] <- list()
        series[[n]]$name <- n
        series[[n]]$color <- cols$fill[counter]
        series[[n]]$data <- lapply(seq_len(length(x)),function(i,x,y) {
            return(c(x[i],y[i])) },round(fit$x,3),round(fit$y,3))
    }
    
    switch(jl,
        highcharts = {
            json <- list(
                chart=list(
                    type="line",
                    zoomType="xy"
                ),
                title=list(
                    text=paste(covarname," bias detection - ",status)
                ),
                xAxis=list(
                    min=minX,
                    title=list(
                        text=covarname,
                        margin=20,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE
                ),
                yAxis=list(
                    title=list(
                        useHTML=TRUE,
                        text="Read count (log<sub>2</sub>)",
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE
                ),
                plotOptions=list(
                    line=list(
                        marker=list(
                            enabled=FALSE,
                            states=list(
                                hover=list(
                                    enabled=FALSE
                                )
                            )
                        ),
                        tooltip=list(
                            headerFormat=paste("<span style=",
                                "\"font-size:1.1em;color:{series.color};",
                                "font-weight:bold\">{series.name}<br>",
                                sep=""),
                            pointFormat=NULL
                        ),
                        turboThreshold=50000
                    )
                ),
                series=unname(series)
            )
        }
    )
    
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out=="list")
        return(json)
}

filteredToJSON <- function(obj,by=c("chromosome","biotype"),
    jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    by <- tolower(by[1])
    out <- tolower(out[1])
    
    filtered <- obj$user$filtered
    total <- obj$user$total
    cols <- .getColorScheme(2)
    
    if (by=="chromosome") {
        chrTab <- table(as.character(filtered$chromosome))
        chr <- as.numeric(chrTab)
        names(chr) <- names(chrTab)
        chrAllTab <- table(as.character(total$chromosome))
        chrAll <- as.numeric(chrAllTab)
        names(chrAll) <- names(chrAllTab)
        barlabChr <- as.character(chr)        
        perChr <- round(chr/chrAll[names(chr)],3)
        perChr[perChr>1] <- 1
        
        series <- vector("list",2)
        names(series) <- c("number","fraction")
                
        series$number <- list()
        series$number$id <- "chr_number"
        series$number$name <- "Number of genes"
        series$number$color <- cols$fill[1]
        series$number$pointPlacement <- -0.2
        series$number$data <- unname(chr)
        
        series$fraction <- list()
        series$fraction$id <- "chr_fraction"
        series$fraction$name <- "Fraction of total genes"
        series$fraction$color <- cols$fill[2]
        series$fraction$pointPlacement <- 0.2
        series$fraction$yAxis <- 1
        series$fraction$data <- unname(perChr)
        
        what <- chr
    }
    else if (by=="biotype") {
        btTab <- table(as.character(filtered$biotype))
        bt <- as.numeric(btTab)
        names(bt) <- names(btTab)
        btAllTab <- table(as.character(total$biotype))
        btAll <- as.numeric(table(as.character(total$biotype)))
        names(btAll) <- names(btAllTab)
        barlabBt <- as.character(bt)
        perBt <- round(bt/btAll[names(bt)],3)
        perBt[perBt>1] <- 1
        
        series <- vector("list",2)
        names(series) <- c("number","fraction")
                
        series$number <- list()
        series$number$id <- "bt_number"
        series$number$name <- "Number of genes"
        series$number$color <- cols$fill[1]
        series$number$pointPlacement <- -0.2
        series$number$data <- unname(bt)
        
        series$fraction <- list()
        series$fraction$id <- "bt_fraction"
        series$fraction$name <- "Fraction of total genes"
        series$fraction$color <- cols$fill[2]
        series$fraction$pointPlacement <- 0.2
        series$fraction$yAxis <- 1
        series$fraction$data <- unname(perBt)
        
        what <- bt
    }
    
    json <- switch(jl,
        highcharts = {
            list(
                chart=list(
                    type="column",
                    alignTicks=FALSE
                ),
                title=list(
                    text=paste("Filtered genes per ",by,sep="")
                ),
                legend=list(
                    enabled=TRUE,
                    itemHoverStyle=list(
                        color="#B40000"
                    )
                ),
                tooltip=list(
                    shared=TRUE
                ),
                xAxis=list(
                    categories=names(what),
                    title=list(
                        text=paste(toupper(substr(by,1,1)),substr(by,2,
                            nchar(by)),sep=""),
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontWeight="bold"
                        )
                    )
                ),
                yAxis=list(
                    list(
                        lineColor=cols$fill[1],
                        lineWidth=2,
                        min=0,
                        tickAmount=11,
                        title=list(
                            text="Number of genes",
                            margin=20,
                            style=list(
                                color="#000000",
                                fontSize="1.2em"
                            )
                        ),                 
                        labels=list(
                            style=list(
                                color="#000000",
                                fontSize="1.1em",
                                fontWeight="bold"
                            )
                        ),
                        offset=10
                    ),
                    list(
                        lineColor=cols$fill[2],
                        lineWidth=2,
                        min=0,
                        max=1,
                        tickAmount=11,
                        #tickInterval=0.1,
                        title=list(
                            text="Fraction of total genes",
                            margin=20,
                            style=list(
                                color="#000000",
                                fontSize="1.2em"
                            )
                        ),                          
                        labels=list(
                            style=list(
                                color="#000000",
                                fontSize="1.1em",
                                fontWeight="bold"
                            )
                        ),
                        opposite=TRUE,
                        offset=10
                    )
                ),
                plotOptions=list(
                    column=list(
                        grouping=FALSE,
                        shadow=FALSE,
                        groupPadding=0.2,
                        pointPadding=0.2,
                        tooltip=list(
                            headerFormat=paste(
                                '<span style="font-size:1.1em;',
                                'font-weight:bold">',
                                '{point.key}</span><br/>',sep=""
                            )
                        )
                    )
                ),
                series=c(unname(series))
            )
        }
    )
    
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out=="list")
        return(json)
}

volcanoToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    f <- obj$x
    p <- obj$y
    xlim <- obj$xlim
    ylim <- obj$ylim
    altNames <- obj$altnames
    pcut <- obj$pcut
    fcut <- obj$fcut
    up <- obj$user$up
    down <- obj$user$down
    ff <- obj$user$unf
    pp <- obj$user$unp
    altNamesNeutrall <- obj$user$ualt
    con <- obj$user$con
    
    switch(jl,
        highcharts = {
            if (is.null(altNames))
                point.format=paste("<strong>Gene ID: </strong>{point.name}<br>",
                    "<strong>Fold change: </strong>{point.x}<br>",
                    "<strong>Significance: </strong>{point.y}",sep="")
            else
                point.format=paste("<strong>Gene name: </strong>",
                    "{point.alt_name}<br>",
                    "<strong>Gene ID: </strong>{point.name}<br>",
                    "<strong>Fold change: </strong>{point.x}<br>",
                    "<strong>Significance: </strong>{point.y}",sep="")
                json <- list(
                    chart=list(
                    type="scatter",
                    zoomType="xy"
                ),
                title=list(
                    text=paste("Volcano plot for",con)
                ),
                xAxis=list(
                    title=list(
                        text="Fold change",
                        margin=20,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(xlim[1],3),
                    max=round(xlim[2],3)
                ),
                yAxis=list(
                    title=list(
                        useHTML=TRUE,
                        text="Significance (-log<sub>10</sub>(p-value))",
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(ylim[1]-2,3),
                    max=round(ylim[2],3)
                ),
                #legend=list(
                #    layout="vertical",
                #    align="left",
                #    verticalAlign="top",
                #    floating=TRUE,
                #    backgroundColor="#FFFFFF",
                #    borderWidth=1
                #),
                plotOptions=list(
                    scatter=list(
                        allowPointSelect=TRUE,
                        marker=list(
                            radius=2,
                            states=list(
                                hover=list(
                                    enabled=TRUE,
                                    lineColor="#333333"
                                )
                            )
                        ),
                        states=list(
                            hover=list(
                                marker=list(
                                    enabled=FALSE
                                )
                            )
                        ),
                        tooltip=list(
                            headerFormat=paste("<span style=",
                                "\"font-size:1.1em;color:{series.color};",
                                "font-weight:bold\">{series.name}<br>",
                                sep=""),
                            pointFormat=point.format
                        ),
                        turboThreshold=50000
                    )
                ),
                series=list(
                    list(
                        name="Up-regulated",
                        color="#EE0000",
                        marker=list(
                            symbol="circle"
                        ),
                        data=makeHighchartsPoints(f[up],-log10(p[up]),
                            unname(altNames[up]))
                    ),
                    list(
                        name="Down-regulated",
                        marker=list(
                            symbol="circle"
                        ),
                        color="#00CD00",
                        data=makeHighchartsPoints(f[down],-log10(p[down]),
                            unname(altNames[down]))
                    ),
                    list(
                        name="Unregulated",
                        marker=list(
                            symbol="circle"
                        ),
                        color="#0000EE",
                        data=makeHighchartsPoints(ff,-log10(pp),
                            unname(altNamesNeutrall))
                    ),
                    list(
                        name="Downfold threshold",
                        color="#000000",
                        type="line",
                        dashStyle="dash",
                        marker=list(
                            enabled=FALSE
                        ),
                        tooltip=list(
                            headerFormat=paste('<strong>{series.name}',
                                '</strong><br/>',sep=""),
                            pointFormat=paste('<strong>Threshold: ',
                                '</strong>{point.x}<br/>',sep="")
                        ),
                        data=list(round(c(-fcut,ylim[1]-5),3),
                            round(c(-fcut,ylim[2]),3))
                    ),
                    list(
                        name="Upfold threshold",
                        color="#000000",
                        type="line",
                        dashStyle="Dash",
                        marker=list(
                            enabled=FALSE
                        ),
                        tooltip=list(
                            headerFormat=paste('<strong>{series.name}',
                                '</strong><br/>',sep=""),
                            pointFormat=paste('<strong>Threshold: ',
                                '</strong>{point.x}<br/>',sep="")
                        ),
                        data=list(round(c(fcut,ylim[1]-5),3),
                            round(c(fcut,ylim[2]),3))
                    ),
                    list(
                        name="Significance threshold",
                        color="#000000",
                        type="line",
                        dashStyle="DashDot",
                        marker=list(
                            enabled=FALSE
                        ),
                        tooltip=list(
                            headerFormat=paste('<strong>{series.name}',
                                '</strong><br/>',sep=""),
                            pointFormat=paste('<strong>Threshold: ',
                                '</strong>{point.y}<br/>',sep="")
                        ),
                        data=list(round(c(xlim[1],-log10(pcut)),3),
                            round(c(xlim[2],-log10(pcut)),3))
                    )
                )
            )
        }
    )
    
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out=="list")
        return(json)
}

scatterToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    x <- obj$x
    y <- obj$y
    altNames <- obj$altnames
    type <- obj$user$covarname
    status <- obj$status
    samples <- obj$samples
    
    if (type=="Mean-Difference") {
        xLab <- "Mean"
        yLab <- "Difference"
    }
    else if (type=="Mean-Variance") {
        xLab <- "Mean"
        yLab <- "Variance"
    }
    else if (type=="X-Y") {
        xLab <- paste("Expression for",samples[1])
        yLab <- paste("Expression for",samples[2])
    }
    
    # Too many points for mostly static, QC plots
    if (length(x)>10000) {
        ii <- sample(seq_len(length(x)),10000)
        x <- x[ii]
        y <- y[ii]
    }
    
    fit <- lowess(x,y)
    
    stMsg <- ""
    if (!is.null(status))
        stMsg <- status
    
    switch(jl,
        highcharts = {
            #if (is.null(altNames))
            #   point.format=paste("<strong>Gene ID: </strong>{point.name}<br>",
            #        "<strong>",xLab,": </strong>{point.x}<br>",
            #        "<strong>",yLab,": </strong>{point.y}",sep="")
            #else
            #    point.format=paste("<strong>Gene name: </strong>",
            #        "{point.alt_name}<br>",
            #        "<strong>Gene ID: </strong>{point.name}<br>",
            #        "<strong>",xLab,": </strong>{point.x}<br>",
            #        "<strong>",yLab,": </strong>{point.y}",sep="")
            point.format=paste("<strong>",xLab,": </strong>{point.x}<br>",
                "<strong>",yLab,": </strong>{point.y}",sep="")
                json <- list(
                    chart=list(
                    type="scatter",
                    zoomType="xy"
                ),
                title=list(
                    text=paste(type,"plot for",stMsg,"samples",samples[1],
                        "and",samples[2])
                ),
                xAxis=list(
                    title=list(
                        text=xLab,
                        margin=20,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(min(x),3),
                    max=round(max(x),3)
                ),
                yAxis=list(
                    title=list(
                        useHTML=TRUE,
                        text=yLab,
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(min(y),3),
                    max=round(max(y),3)
                ),
                plotOptions=list(
                    scatter=list(
                        turboThreshold=25000
                    ),
                    line=list(
                        turboThreshold=25000
                    )
                ),
                series=list(
                    list(
                        type="scatter",
                        name="Genes/Transcripts",
                        #color="#00ABEE",
                        color="rgba(0,171,238,0.5)",
                        marker=list(
                            radius=0.5
                        ),
                        #data=makeHighchartsPoints(x,y,unname(altNames)),
                        data=makeHighchartsPoints(x,y,simple=TRUE),
                        tooltip=list(
                            followPointer=FALSE,
                            headerFormat=paste("<span style=",
                                "\"font-size:1.1em;color:{series.color};",
                                "font-weight:bold\">{series.name}<br>",
                                sep=""),
                            pointFormat=point.format
                        )
                    ),
                    list(
                        type="line",
                        name="Trend",
                        marker=list(
                            enabled=FALSE
                        ),
                        color="#E40000",
                        data=lapply(seq_len(length(x)),function(i,x,y) {
                            return(c(x[i],y[i])) 
                        },round(fit$x,3),round(fit$y,3))
                    )
                )
            )
        }
    )
    
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out=="list")
        return(json)
}

rnacompToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- tolower(out[1])
    
    status <- obj$status
    dat <- obj$user$plotdata$data2plot
    samples <- obj$samples
    refColumn <- obj$user$plotdata$refColumn
    
    # Too many points for a lot of curves of interactive data
    if (nrow(dat)>1000) {
        ii <- sort(sample(seq_len(nrow(dat)),998))
        ii <- c(1,ii,nrow(dat))
        dat <- cbind(dat[ii,1],dat[ii,2:ncol(dat)])
    }

    if (is.list(samples))
        samplenames <- unlist(samples)
    
    samplenames <- setdiff(samplenames,refColumn)
    dat <- as.matrix(dat[,samplenames])
    datDens <- apply(dat,2,density,adjust=1.5)
    limY <- c(0,max(vapply(datDens,function (x) max(x$y,na.rm=TRUE),
        numeric(1))))  
    cols <- .getColorScheme(length(samplenames))
    
    # Construct series
    counter <- 0
    densSeries <- vector("list",length(samplenames))
    names(densSeries) <- samplenames
    for (n in names(densSeries)) {
        counter <- counter + 1
        densSeries[[n]] <- list()
        densSeries[[n]]$name=n
        densSeries[[n]]$color=cols$fill[counter]
        densSeries[[n]]$data <- 
            makeHighchartsPoints(datDens[[n]]$x,datDens[[n]]$y)
        densSeries[[n]]$tooltip=list(
            headerFormat=paste("<span style=",
                "\"font-size:1.1em;color:{series.color};",
                "font-weight:bold\">{series.name}<br>",
                sep=""),
            pointFormat=NULL
        )
    }
    counter <- 0
    abSeries <- vector("list",length(samplenames))
    names(abSeries) <- samplenames
    for (n in names(abSeries)) {
        counter <- counter + 1
        abSeries[[n]] <- list()
        abSeries[[n]]$name=paste(n,"median")
        abSeries[[n]]$color=cols$fill[counter]
        abSeries[[n]]$data <- 
            makeHighchartsPoints(rep(median(dat[,n],na.rm=TRUE),nrow(dat)),
                seq(limY[1],limY[2],length.out=nrow(dat)))
        abSeries[[n]]$enableMouseTracking=FALSE
        abSeries[[n]]$dashStyle="Dash"
    }
    
    switch(jl,
        highcharts = {
            json <- list(
                chart=list(
                    type="line",
                    zoomType="xy"
                ),
                title=list(
                    text=paste("RNA-composition for",status,"mapped reads")
                ),
                xAxis=list(
                    title=list(
                        text="log<sub>2</sub>(sample/reference sample)",
                        margin=20,
                        useHTML=TRUE,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1
                ),
                yAxis=list(
                    title=list(
                        text="Density",
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=limY[1],
                    max=limY[2]
                ),
                plotOptions=list(
                    line=list(
                        allowPointSelect=TRUE,
                        lineWidth=1,
                        marker=list(
                            enabled=FALSE
                        ),
                        tooltip=list(
                            headerFormat=paste("<span style=",
                                "\"font-size:1.1em;color:{series.color};",
                                "font-weight:bold\">{series.name}<br>",
                                sep=""),
                            pointFormat=NULL
                        ),
                        turboThreshold=50000
                    )
                ),
                series=c(unname(densSeries),unname(abSeries))
            )
        }
    )
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out=="list")
        return(json)
}

biodistToJSON <- function(obj,jl=c("highcharts"),by=c("chromosome","biotype"),
    out=c("json","list")) {
    jl <- tolower(jl[1])
    by <- by[1]
    out <- tolower(out[1])
    
    plotdataChromosome <- obj$user$plotdata$chromosome
    plotdataBiotype <- obj$user$plotdata$biotype
    
    cols <- .getColorScheme()
    
    if (by == "chromosome") {
        # Data series
        series <- vector("list",3)
        names(series) <- c("genome","chromosome_deg","deg_chromosome")
        series$genome <- list()
        series$genome$id <- "genome"
        series$genome$name <- "% chromosome in genome"
        series$genome$color <- cols$trans[1]
        series$genome$pointPlacement <- -0.2
        series$genome$data <- round(as.numeric(plotdataChromosome[1,]),3)
        series$chromosome_deg <- list()
        series$chromosome_deg$id <- "chromosome_deg"
        series$chromosome_deg$name <- "% chromosome in DEG"
        series$chromosome_deg$color <- cols$trans[2]
        series$chromosome_deg$pointPlacement <- 0
        series$chromosome_deg$data <- 
            round(as.numeric(plotdataChromosome[2,]),3)
        series$deg_chromosome <- list()
        series$deg_chromosome$id <- "deg_chromosome"
        series$deg_chromosome$name <- "% DEG in chromosome"
        series$deg_chromosome$color <- cols$trans[3]
        series$deg_chromosome$pointPlacement <- 0.2
        series$deg_chromosome$data <- 
            round(as.numeric(plotdataChromosome[3,]),3)
            
        json <- switch(jl,
            highcharts = {
                list(
                    chart=list(
                        type="column",
                        alignTicks=FALSE
                    ),
                    title=list(
                        text="DEG distribution across chromosomes"
                    ),
                    legend=list(
                        enabled=TRUE,
                        itemHoverStyle=list(
                            color="#B40000"
                        )
                    ),
                    tooltip=list(
                        shared=TRUE
                    ),  
                    xAxis=list(
                        categories=names(plotdataChromosome),
                        title=list(
                            text="Chromosome",
                            margin=25,
                            style=list(
                                color="#000000",
                                fontSize="1.2em"
                            )
                        ),
                        labels=list(
                            style=list(
                                color="#000000",
                                fontWeight="bold"
                            )
                        )
                    ),
                    yAxis=list(
                        list(
                            min=0,
                            max=ceiling(max(plotdataChromosome[1,])),
                            title=list(
                                text="% of features",
                                margin=20,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),                          
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                )
                            )
                        )
                    ),
                    plotOptions=list(
                        column=list(
                            grouping=FALSE,
                            shadow=FALSE,
                            groupPadding=0.3,
                            pointPadding=0.25,
                            tooltip=list(
                                headerFormat=paste(
                                    '<span style="font-size:1.1em;',
                                    'font-weight:bold">',
                                    '{point.key}</span><br/>',sep=""
                                )
                            )
                        )
                    ),
                    series=unname(series)
                )
            }
        )
    }
    if (by == "biotype") {
        # For biotype
        abu <- which(plotdataBiotype[1,]>7)
        nabu <- which(plotdataBiotype[1,]<=7)
        
        # Data series
        seriesAbu <- vector("list",3)
        names(seriesAbu) <- c("genome","biotype_deg","deg_biotype")
        seriesAbu$genome <- list()
        seriesAbu$genome$id <- "abu_genome"
        seriesAbu$genome$name <- "% biotype in genome"
        seriesAbu$genome$color <- cols$trans[1]
        seriesAbu$genome$pointPlacement <- -0.2
        seriesAbu$genome$data <- round(as.numeric(plotdataBiotype[1,abu]),3)
        seriesAbu$biotype_deg <- list()
        seriesAbu$biotype_deg$id <- "abu_biotype_deg"
        seriesAbu$biotype_deg$name <- "% biotype in DEG"
        seriesAbu$biotype_deg$color <- cols$trans[2]
        seriesAbu$biotype_deg$pointPlacement <- 0
        seriesAbu$biotype_deg$data <- 
            round(as.numeric(plotdataBiotype[2,abu]),3)
        seriesAbu$deg_biotype <- list()
        seriesAbu$deg_biotype$id <- "abu_deg_biotype"
        seriesAbu$deg_biotype$name <- "% DEG in biotype"
        seriesAbu$deg_biotype$color <- cols$trans[3]
        seriesAbu$deg_biotype$pointPlacement <- 0.2
        seriesAbu$deg_biotype$data <- 
            round(as.numeric(plotdataBiotype[3,abu]),3)
        seriesNabu <- vector("list",3)
        names(seriesNabu) <- c("genome","biotype_deg","deg_biotype")
        seriesNabu$genome <- list()
        seriesNabu$genome$name <- "% biotype in genome"
        seriesNabu$genome$yAxis <- 1
        seriesNabu$genome$pointStart <- length(abu)
        seriesNabu$genome$linkedTo <- "abu_genome"
        seriesNabu$genome$color <- cols$trans[1]
        seriesNabu$genome$pointPlacement <- -0.2
        seriesNabu$genome$data <- round(as.numeric(plotdataBiotype[1,nabu]),3)
        seriesNabu$biotype_deg <- list()
        seriesNabu$biotype_deg$name <- "% biotype in DEG"
        seriesNabu$biotype_deg$yAxis <- 1
        seriesNabu$biotype_deg$pointStart <- length(abu)
        seriesNabu$biotype_deg$linkedTo <- "abu_detected"
        seriesNabu$biotype_deg$color <- cols$trans[2]
        seriesNabu$biotype_deg$pointPlacement <- 0
        seriesNabu$biotype_deg$data <- 
            round(as.numeric(plotdataBiotype[2,nabu]),3)
        seriesNabu$deg_biotype <- list()
        seriesNabu$deg_biotype$name <- "% DEG in biotype"
        seriesNabu$deg_biotype$yAxis <- 1
        seriesNabu$deg_biotype$pointStart <- length(abu)
        seriesNabu$deg_biotype$linkedTo <- "abu_sample"
        seriesNabu$deg_biotype$color <- cols$trans[3]
        seriesNabu$deg_biotype$pointPlacement <- 0.2
        seriesNabu$deg_biotype$data <- 
            round(as.numeric(plotdataBiotype[3,nabu]),3)
        
        json <- switch(jl,
            highcharts = {
                list(
                    chart=list(
                        type="column",
                        alignTicks=FALSE
                    ),
                    title=list(
                        text="DEG distribution across biotypes"
                    ),
                    legend=list(
                        enabled=TRUE,
                        itemHoverStyle=list(
                            color="#B40000"
                        )
                    ),
                    tooltip=list(
                        shared=TRUE
                    ),  
                    xAxis=list(
                        categories=colnames(plotdataBiotype)[c(abu,nabu)],
                        title=list(
                            text="Biotype",
                            margin=25,
                            style=list(
                                color="#000000",
                                fontSize="1.2em"
                            )
                        ),
                        labels=list(
                            style=list(
                                color="#000000",
                                fontWeight="bold"
                            )
                        ),
                        plotLines=list(
                            list(
                                color="#8A8A8A",
                                width=1.5,
                                dashStyle="Dash",
                                value=length(abu)-0.5
                            )
                        ),
                        plotBands=list(
                            list(
                                color="#FFFFE0",
                                from=-0.5,
                                to=length(abu)-0.5
                            ),
                            list(
                                color="#FFECEB",
                                from=length(abu)-0.5,
                                to=ncol(plotdataBiotype)
                            )
                        )
                    ),
                    yAxis=list(
                        list(
                            min=0,
                            max=70,
                            title=list(
                                text="% of abundant features",
                                margin=20,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),                          
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                )
                            )
                        ),
                        list(
                            min=0,
                            max=7,
                            title=list(
                                text="% of non-abundant features",
                                margin=20,
                                style=list(
                                    color="#000000",
                                    fontSize="1.2em"
                                )
                            ),                          
                            labels=list(
                                style=list(
                                    color="#000000",
                                    fontSize="1.1em",
                                    fontWeight="bold"
                                )
                            ),
                            opposite=TRUE
                        )
                    ),
                    plotOptions=list(
                        column=list(
                            grouping=FALSE,
                            shadow=FALSE,
                            groupPadding=0.3,
                            pointPadding=0.25,
                            tooltip=list(
                                headerFormat=paste(
                                    '<span style="font-size:1.1em;',
                                    'font-weight:bold">',
                                    '{point.key}</span><br/>',sep=""
                                )
                            )
                        )
                    ),
                    series=c(unname(seriesAbu),unname(seriesNabu))
                )
            }
        )
    }
    
    if (out=="json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out=="list")
        return(json)
}

maStatToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- out[1]
    
    a <- obj$x
    f <- obj$y
    p <- obj$user$p
    xlim <- obj$xlim
    ylim <- obj$ylim
    pcut <- obj$pcut
    fcut <- obj$fcut
    altNames <- obj$altnames
    con <- obj$user$con
    #stat <- obj$user$stat
    
    conlab <- strsplit(con,"_vs_")
    conlab <- paste(conlab[[1]][1],"against",conlab[[1]][2])
    #conlab <- paste(conlab[[1]][2],"against",conlab[[1]][1])
    #statMap <- .getStatMap()
    #statlab <- statMap[[stat]]
    
    if (!is.null(p)) {
        upstat <- which(f>=fcut & p<pcut)
        downstat <- which(f<=-fcut & p<pcut)
        up <- which(f>=fcut & p>=pcut)
        down <- which(f<=-fcut & p>=pcut)
        poor <- which(p<pcut & abs(f)<fcut)
        neutral <- setdiff(seq_len(length(a)),
            Reduce("union",list(upstat,downstat,up,down,poor)))
    }
    else {
        upstat <- downstat <- poor <- integer(0)
        up <- which(f>=fcut)
        down <- which(f<=-fcut)
        neutral <- setdiff(seq_len(length(a)),union(up,down))
    }
    
    switch(jl,
        highcharts = {
            series <- list()
            counter <- 0
            if (length(upstat)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significantly up-regulated",
                    color="#EE0000",
                    marker=list(
                        radius=3
                    ),
                    data=makeHighchartsPoints(
                        x=a[upstat],
                        y=f[upstat],
                        a=unname(altNames[upstat]),
                        p=if (!is.null(p)) -log10(p[upstat]) else NULL
                    )
                )
            }
            if (length(downstat)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significantly down-regulated",
                    color="#00EE00",
                    marker=list(
                        radius=3
                    ),
                    data=makeHighchartsPoints(
                        x=a[downstat],
                        y=f[downstat],
                        a=unname(altNames[downstat]),
                        p=if (!is.null(p)) -log10(p[downstat]) else NULL
                    )
                )
            }
            if (length(up)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Up-regulated",
                    color="#B80000",
                    marker=list(
                        radius=2
                    ),
                    data=makeHighchartsPoints(
                        x=a[up],
                        y=f[up],
                        a=unname(altNames[up]),
                        p=if (!is.null(p)) -log10(p[up]) else NULL
                    )
                )
            }
            if (length(down)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Down-regulated",
                    color="#008000",
                    marker=list(
                        radius=2
                    ),
                    data=makeHighchartsPoints(
                        x=a[down],
                        y=f[down],
                        a=unname(altNames[down]),
                        p=if (!is.null(p)) -log10(p[down]) else NULL
                    )
                )
            }
            if (length(poor)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significant only",
                    color="#FFA600",
                    data=makeHighchartsPoints(
                        x=a[poor],
                        y=f[poor],
                        a=unname(altNames[poor]),
                        p=if (!is.null(p)) -log10(p[poor]) else NULL
                    )
                )
            }
            if (length(neutral)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Neutral",
                    color="#858585",
                    data=makeHighchartsPoints(
                        x=a[neutral],
                        y=f[neutral],
                        a=unname(altNames[neutral]),
                        p=if (!is.null(p)) -log10(p[neutral]) else NULL
                    )
                )
            }
        
            if (is.null(altNames))
                point.format=paste("<strong>Gene ID: </strong>{point.name}<br>",
                    "<strong>Average expression: </strong>{point.x}<br>",
                    "<strong>Fold change: </strong>{point.y}<br>",
                    "<strong>Significance: </strong>{point.sig}",sep="")
            else
                point.format=paste("<strong>Gene name: </strong>",
                    "{point.alt_name}<br>",
                    "<strong>Gene ID: </strong>{point.name}<br>",
                    "<strong>Average expression: </strong>{point.x}<br>",
                    "<strong>Fold change: </strong>{point.y}<br>",
                    "<strong>Significance: </strong>{point.sig}",sep="")
                json <- list(
                    chart=list(
                    type="scatter",
                    zoomType="xy"
                ),
                title=list(
                    text=paste("Mean-Difference (MA) plot for",conlab)
                ),
                xAxis=list(
                    title=list(
                        text="Average expression",
                        margin=20,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(xlim[1],3),
                    max=round(xlim[2],3)
                ),
                yAxis=list(
                    title=list(
                        useHTML=TRUE,
                        text="Fold change (log<sub>2</sub>)",
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(ylim[1],3),
                    max=round(ylim[2],3)
                ),
                plotOptions=list(
                    scatter=list(
                        allowPointSelect=TRUE,
                        marker=list(
                            radius=2,
                            symbol="circle",
                            states=list(
                                hover=list(
                                    enabled=TRUE,
                                    lineColor="#333333"
                                )
                            )
                        ),
                        states=list(
                            hover=list(
                                marker=list(
                                    enabled=FALSE
                                )
                            )
                        ),
                        #events=list(
                        #   legendItemClick=paste("function() {",
                        #       "return false; }")
                        #),
                        tooltip=list(
                            headerFormat=paste("<span style=",
                                "\"font-size:1.1em;color:{series.color};",
                                "font-weight:bold\">{series.name}<br>",
                                sep=""),
                            pointFormat=point.format
                        ),
                        turboThreshold=50000
                    )
                ),
                series=series
            )
        }
    )
    if (out == "json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out == "list")
        return(json)
}

dereguloToJSON <- function(obj,jl=c("highcharts"),out=c("json","list")) {
    jl <- tolower(jl[1])
    out <- out[1]
    
    xlim <- obj$xlim
    ylim <- obj$ylim
    pcut <- obj$pcut
    fcut <- obj$fcut
    altNames <- obj$altnames
    fmat <- obj$user$fmat
    pmat <- obj$user$pmat
    
    conlabX <- strsplit(colnames(fmat)[1],"_vs_")
    conlabX <- paste(conlabX[[1]][1],"against",conlabX[[1]][2])
    conlabY <- strsplit(colnames(fmat)[2],"_vs_")
    conlabY <- paste(conlabY[[1]][1],"against",conlabY[[1]][2])
    
    # red
    upupstat <- which(apply(fmat,1,function(x) all(x >= fcut)) &
        apply(pmat,1,function(x) all(x < pcut)))
    # green
    downdownstat <- which(apply(fmat,1,function(x) all(x <= -fcut)) &
        apply(pmat,1,function(x) all(x < pcut)))
    # red3
    upup <- which(apply(fmat,1,function(x) all(x >= fcut)) &
        apply(pmat,1,function(x) any(x >= pcut)))
    # green3
    downdown <- which(apply(fmat,1,function(x) all(x <= -fcut)) &
        apply(pmat,1,function(x) any(x >= pcut)))
    # orange
    updownstat <- which(apply(fmat,1,
        function(x) x[1] >= fcut & x[2] <= -fcut) &
            apply(pmat,1,function(x) all(x < pcut)))
    # blue
    downupstat <- which(apply(fmat,1,
        function(x) x[1] <= -fcut & x[2] >= fcut) &
            apply(pmat,1,function(x) all(x < pcut)))
    # orange3
    updown <- which(apply(fmat,1,function(x) x[1] >= fcut & x[2] <= -fcut) &
        apply(pmat,1,function(x) any(x >= pcut)))
    # blue3
    downup <- which(apply(fmat,1,function(x) x[1] <= -fcut & x[2] >= fcut) &
        apply(pmat,1,function(x) any(x >= pcut)))
    # black
    poor <- which(apply(pmat,1,function(x) all(x < pcut)) &
        apply(fmat,1,function(x) any(abs(x) < fcut)))
    # gray70
    nones <- which(apply(pmat,1,function(x) any(x >= pcut)) &
        apply(fmat,1,function(x) all(abs(x) < fcut)))
    # gray40
    neutral <- setdiff(seq_len(nrow(fmat)),
        Reduce("union",list(upupstat,downdownstat,upup,downdown,updownstat,
            downupstat,updown,downup,poor,nones)))
    
    switch(jl,
        highcharts = {
            series <- list()
            counter <- 0
            if (length(upupstat)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significantly both up-regulated",
                    color="#EE0000",
                    marker=list(
                        radius=3
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[upupstat,1],
                        y=fmat[upupstat,2],
                        a=unname(altNames[upupstat])
                    )
                )
            }
            if (length(downdownstat)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significantly both down-regulated",
                    color="#00EE00",
                    marker=list(
                        radius=3
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[downdownstat,1],
                        y=fmat[downdownstat,2],
                        a=unname(altNames[downdownstat])
                    )
                )
            }
            if (length(upup)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Both up-regulated",
                    color="#B80000",
                    marker=list(
                        radius=2
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[upup,1],
                        y=fmat[upup,2],
                        a=unname(altNames[upup])
                    )
                )
            }
            if (length(downdown)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Both down-regulated",
                    color="#008000",
                    marker=list(
                        radius=2
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[downdown,1],
                        y=fmat[downdown,2],
                        a=unname(altNames[downdown])
                    )
                )
            }
            if (length(updownstat)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significantly up-down-regulated",
                    color="#FFB700",
                    marker=list(
                        radius=3
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[updownstat,1],
                        y=fmat[updownstat,2],
                        a=unname(altNames[updownstat])
                    )
                )
            }
            if (length(downupstat)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significantly down-up-regulated",
                    color="#0000FF",
                    marker=list(
                        radius=3
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[downupstat,1],
                        y=fmat[downupstat,2],
                        a=unname(altNames[downupstat])
                    )
                )
            }
            if (length(updown)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Up-down-regulated",
                    color="#C68100",
                    marker=list(
                        radius=2
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[updown,1],
                        y=fmat[updown,2],
                        a=unname(altNames[updown])
                    )
                )
            }
            if (length(downup)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Down-up-regulated",
                    color="#0000AB",
                    marker=list(
                        radius=2
                    ),
                    data=makeHighchartsPoints(
                        x=fmat[downup,1],
                        y=fmat[downup,2],
                        a=unname(altNames[downup])
                    )
                )
            }
            if (length(poor)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Significant only",
                    color="#1F1F1F",
                    data=makeHighchartsPoints(
                        x=fmat[poor,1],
                        y=fmat[poor,2],
                        a=unname(altNames[poor])
                    )
                )
            }
            if (length(neutral)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="Neutral",
                    color="#00DADE",
                    data=makeHighchartsPoints(
                        x=fmat[neutral,1],
                        y=fmat[neutral,2],
                        a=unname(altNames[neutral])
                    )
                )
            }
            if (length(nones)>0) {
                counter <- counter + 1
                series[[counter]] <- list(
                    name="No regulation",
                    color="#CCCCCC",
                    data=makeHighchartsPoints(
                        x=fmat[nones,1],
                        y=fmat[nones,2],
                        a=unname(altNames[nones])
                    )
                )
            }
            
            # Add fold lines
            counter <- counter + 1
            series[[counter]] <- list(
                name="X up fold threshold",
                color="#000000",
                type="line",
                dashStyle="Dash",
                lineWidth=1,
                marker=list(
                    enabled=FALSE
                ),
                tooltip=list(
                    headerFormat=paste('<strong>{series.name}',
                        '</strong><br/>',sep=""),
                    pointFormat=paste('<strong>Threshold: ',
                        '</strong>{point.x}<br/>',sep="")
                ),
                data=list(round(c(fcut,floor(ylim[1])),3),
                    round(c(fcut,ceiling(ylim[2])),3))
            )
            counter <- counter + 1
            series[[counter]] <- list(
                name="X down fold threshold",
                color="#000000",
                type="line",
                dashStyle="Dash",
                lineWidth=1,
                marker=list(
                    enabled=FALSE
                ),
                tooltip=list(
                    headerFormat=paste('<strong>{series.name}',
                        '</strong><br/>',sep=""),
                    pointFormat=paste('<strong>Threshold: ',
                        '</strong>{point.x}<br/>',sep="")
                ),
                data=list(round(c(-fcut,floor(ylim[1])),3),
                    round(c(-fcut,ceiling(ylim[2])),3))
            )
            counter <- counter + 1
            series[[counter]] <- list(
                name="Y up fold threshold",
                color="#000000",
                type="line",
                dashStyle="Dash",
                marker=list(
                    enabled=FALSE
                ),
                tooltip=list(
                    headerFormat=paste('<strong>{series.name}',
                        '</strong><br/>',sep=""),
                    pointFormat=paste('<strong>Threshold: ',
                        '</strong>{point.y}<br/>',sep="")
                ),
                data=list(round(c(floor(xlim[1]),fcut),3),
                    round(c(ceiling(xlim[2]),fcut),3))
            )
            counter <- counter + 1
            series[[counter]] <- list(
                name="Y down fold threshold",
                color="#000000",
                type="line",
                dashStyle="Dash",
                marker=list(
                    enabled=FALSE
                ),
                tooltip=list(
                    headerFormat=paste('<strong>{series.name}',
                        '</strong><br/>',sep=""),
                    pointFormat=paste('<strong>Threshold: ',
                        '</strong>{point.y}<br/>',sep="")
                ),
                data=list(round(c(floor(xlim[1]),-fcut),3),
                    round(c(ceiling(xlim[2]),-fcut),3))
            )
        
            if (is.null(altNames))
                point.format=paste("<strong>Gene ID: </strong>{point.name}<br>",
                    "<strong>Fold change X: </strong>{point.x}<br>",
                    "<strong>Fold change Y: </strong>{point.y}<br>",sep="")
            else
                point.format=paste("<strong>Gene name: </strong>",
                    "{point.alt_name}<br>",
                    "<strong>Gene ID: </strong>{point.name}<br>",
                    "<strong>Fold change X: </strong>{point.x}<br>",
                    "<strong>Fold change Y: </strong>{point.y}<br>",sep="")
                json <- list(
                    chart=list(
                    type="scatter",
                    zoomType="xy"
                ),
                title=list(
                    text=paste("Deregulogram")
                ),
                xAxis=list(
                    title=list(
                        text=paste0("Fold change (log<sub>2</sub>)",conlabX),
                        margin=20,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(xlim[1],3),
                    max=round(xlim[2],3)
                ),
                yAxis=list(
                    title=list(
                        useHTML=TRUE,
                        text=paste0("Fold change (log<sub>2</sub>)",conlabY),
                        margin=25,
                        style=list(
                            color="#000000",
                            fontSize="1.2em"
                        )
                    ),
                    labels=list(
                        style=list(
                            color="#000000",
                            fontSize="1.1em",
                            fontWeight="bold"
                        )
                    ),
                    startOnTick=TRUE,
                    endOnTick=TRUE,
                    showLastLabel=TRUE,
                    gridLineWidth=1,
                    min=round(ylim[1],3),
                    max=round(ylim[2],3)
                ),
                plotOptions=list(
                    scatter=list(
                        allowPointSelect=TRUE,
                        marker=list(
                            radius=2,
                            symbol="circle",
                            states=list(
                                hover=list(
                                    enabled=TRUE,
                                    lineColor="#333333"
                                )
                            )
                        ),
                        states=list(
                            hover=list(
                                marker=list(
                                    enabled=FALSE
                                )
                            )
                        ),
                        tooltip=list(
                            headerFormat=paste("<span style=",
                                "\"font-size:1.1em;color:{series.color};",
                                "font-weight:bold\">{series.name}<br>",
                                sep=""),
                            pointFormat=point.format
                        ),
                        turboThreshold=50000
                    )
                ),
                series=series
            )
        }
    )
    if (out == "json")
        return(toJSON(json,auto_unbox=TRUE,null="null"))
    else if (out == "list")
        return(json)
}

.unquote_js_fun <- function(js) {
    if (is.list(js))
        js <- lapply(js,.unquote_js_fun)
    else {
        op <- gregexpr(pattern="function",js)
        cl <- gregexpr(pattern="}\\\"",js)
        if (length(op)>0) {
            starts <- as.numeric(op[[1]])
            for (i in seq_len(length(starts)))
                substr(js,starts[i]-1,starts[i]-1) <- " "
            ends <- as.numeric(cl[[1]])
            for (i in seq_len(length(starts)))
                substr(js,ends[i]+1,ends[i]+1) <- " "
        }
    }
    return(js)
}

.getGroupColorScheme <- function(group) {
    cols <- .getColorScheme(length(group))
    classes <- as.factor(asClassVector(group))
    design <- as.numeric(classes)
    return(lapply(cols,function(x,classes,design) {
        return(x[seq_len(length(levels(classes)))][design])
    },classes,design))
}

.getColorScheme <- function(n=NULL) {
    if (missing(n) || is.null(n))
        return(.getColors())
    else {
        cols <- .getColors()
        if (n > length(cols$fill)) {
            cols$fill <- rep(cols$fill,length.out=n)
            cols$border <- rep(cols$border,length.out=n)
            cols$select <- rep(cols$select,length.out=n)
            cols$trans <- rep(cols$trans,length.out=n)
        }
        return(cols)
    }
}

.getColors <- function() {
    return(list(
        fill=c("#CD0000","#00CD00","#0000EE","#FFD700","#87CEEB","#CD8500",
            "#DEB887","#FF0000","#0000FF","#00FF00","#FFA500","#A9A9A9",
            "#008B00","#313131","#FFC0CB","#A52A2A","#FF00FF","#9ACD32",
            "#8B636C","#2E8B57","#008B8B"),
        border=c("#850000","#006B00","#000085","#927C00","#156280","#5A3A00",
            "#8B7457","#935E18","#000080","#008500","#603E00","#454545",
            "#073E07","#000000","#896067","#691111","#7C007C","#3A4D14",
            "#5B1726","#0C2517","#062A2A"),
        selected=c("#FF0000","#00FF00","#0066FF","#FFD700","#FFEB77","#FFB428",
            "#FFD9A5","#FF326D","#0089FF","#B3FF00","#FFC352","#D9D9D9",
            "#00EC00","#8E8E8E","#FFDAE0","#F94444","#FF87FF","#C2FF45",
            "#EA889D","#4EE590","#00DADA"),
        trans=c("rgba(205,0,0,0.6)","rgba(0,205,0,0.6)","rgba(0,0,238,0.6)",
            "rgba(255,215,0,0.6)","rgba(135,206,235,0.6)","rgba(205,133,0,0.6)",
            "rgba(222,184,135,0.6)","rbga(255,0,0,0.5)","rgba(0,0,255,0.5)",
            "rgba(0,255,0,0.5)","rgba(255,165,0,0.6)","rgba(169,169,169,0.5)",
            "rgba(0,139,0,0.6)","rgba(49,49,49,0.6)","rgba(255,192,203,0.5)",
            "rgba(165,42,42,0.6)","rgba(255,0,255,0.6)","rgba(154,205,50,0.6)",
            "rgba(139,99,108,0.6)","rgba(46,139,87,0.6)","rgba(0,139,139,0.6)")
    ))
}

.getStatMap <- function() {
    return(list(
        deseq="DEseq",
        deseq2="DEseq2",
        edger="edgeR",
        limma="voom",
        nbpseq="NBPSeq",
        noiseq="NOISeq",
        #bayseq="baySeq",
        absseq="ABSSeq",
        dss="DSS"
    ))
}

createSignalTracks <- function(targets,org,urlBase=NULL,stranded=FALSE,
    normTo=1e+9,exportPath=".",hubInfo=list(name="MyHub",shortLabel="My hub",
    longLabel="My hub",email="someone@example.com"),fasta=NULL,gtf=NULL,
    forceHub=FALSE,overwrite=FALSE,rc=NULL) {
    if (!requireNamespace("rtracklayer"))
        stopwrap("Bioconductor package rtracklayer is required to build ",
            "tracks!")
    
    if (!is.null(fasta) && !requireNamespace("Biostrings"))
        stopwrap("Bioconductor package Biostrings is required to read ",
            "FASTA files!")

    if (!is.list(targets)) {
        if (file.exists(targets))
            targets <- readTargets(targets)
        else
            stopwrap("targets must be either the result of readTargets ",
                "function or a valid targets file!")
    }

    # Read in BAMs and make Rles
    message("Creating tracks...")
    if (stranded) {
        message("  stranded mode")
        .createTrackHub(targets,org,normTo,stranded=TRUE,urlBase,
            exportPath,hubInfo,fasta,gtf,overwrite,rc=rc)
        return(paste0(urlBase,"/hub.txt"))
        #return(.createStrandedSignalTracks(targets,org,normTo,urlBase,
        #    exportPath,hubInfo,overwrite,rc=rc))
    }
    else {
        message("  unstranded mode")
        if (forceHub) {
            .createTrackHub(targets,org,normTo,stranded=FALSE,urlBase,
                exportPath,hubInfo,fasta,gtf,overwrite,rc=rc)
            return(paste0(urlBase,"/hub.txt"))
        }
        else {
            trackLines <- .createUnstrandedSignalTracks(targets,org,normTo,
                urlBase,exportPath,overwrite,asLines=TRUE,rc=rc)
            tracksFile <- file.path(exportPath,"tracks.txt")
            writeLines(trackLines,tracksFile)
            tracksLink <- paste0(urlBase,"/tracks.txt")
            return(tracksLink)
        }
        
    }
}

.createTrackHub <- function(targets,org,normTo,stranded=FALSE,urlBase,
    exportPath,hubInfo,fasta=NULL,gtf=NULL,overwrite=FALSE,rc=NULL) {
    supOrg <- org %in% getSupportedOrganisms()
    if (supOrg)
        org <- getUcscOrganism(org)
    trackDbPath <- file.path(exportPath,org)
    
    # The BAM files
    bams <- unlist(targets$files,use.names=FALSE)
    
    # Get seqinfo form first BAM
    preSf <- .chromInfoFromBAM(bams[1])
    if (supOrg) {
        vchrs <- getValidChrs(org)
        preSf <- preSf[intersect(vchrs,rownames(preSf)),,drop=FALSE]
    }
    sf <- .chromInfoToSeqInfoDf(preSf,o=org,asSeqinfo=TRUE)
    
    # Start creating the hub
    groups <- rep(names(targets$files),lengths(targets$files))
    if (is.null(groups))
        groups <- rep("Signal",length(bams))
    
    # First the dir structure
    if (!dir.exists(trackDbPath))
        dir.create(trackDbPath,recursive=TRUE)
    
    # Has a custom FASTA file been provided?
    has2bit <- FALSE
    if (!supOrg && !is.null(fasta) && is.character(fasta) 
        && file.exists(fasta)) {
        has2bit <- TRUE
        message("Reading file ",fasta)
        ss <- readDNAStringSet(fasta)
        names(ss) <- 
            vapply(strsplit(names(ss)," "),function(x) x[1],character(1))
        message("Writing file ",file.path(trackDbPath,paste0(org,".2bit")))
        export.2bit(ss,file.path(trackDbPath,paste0(org,".2bit")))
    }
    
    # Has a custom GTF file been provided? More complex...
    hasGtf <- FALSE
    if (!supOrg && !is.null(gtf) && is.character(gtf)
        && file.exists(gtf)) {
        hasGtf <- TRUE
        message("Reading file ",gtf," and creating bigBed")
        bb <- .externalGtfToBigBed(gtf,preSf)
        file.copy(bb,file.path(trackDbPath,paste0(org,".bigBed")),
            overwrite=TRUE)
    }
    
    # Write the genomes file
    gf <- .makeTrackhubGenomesFile(org,bams[1],has2bit)
    writeLines(paste(names(gf)," ",gf,sep=""),
        file.path(exportPath,"genomes.txt"))
        
    # Write the groups file
    if (file.exists(file.path(trackDbPath,"groups.txt")))
        unlink(file.path(trackDbPath,"groups.txt"))
    grp <- .makeTrackhubGroupsFile(targets,hasGtf)
    for (g in grp) {
        grm <- paste(paste(names(g)," ",g,sep=""),collapse="\n")
        aGroup <- c(grm,"\n\n")
        cat(aGroup,file=file.path(trackDbPath,"groups.txt"),append=TRUE)
    }
    
    # Create the bigWigs and write the trackDb file
    # Clear previous file if exists, otherwise append will cause problems
    if (file.exists(file.path(trackDbPath,"trackDb.txt")))
        unlink(file.path(trackDbPath,"trackDb.txt"))
    if (stranded) {
        # BigWig attributes
        pretdb <- .createStrandedSignalTracks(targets,org,normTo,urlBase,
            exportPath,overwrite=overwrite,rc=rc)
        for (p in pretdb) {
            tracks <- p$tracks
            p$tracks <- NULL
            meta <- paste(paste(names(p)," ",p,sep=""),collapse="\n")
            post <- paste(names(tracks$positive)," ",tracks$positive,sep="")
            post <- paste(paste0("    ",post),sep="",collapse="\n")
            negt <- paste(names(tracks$negative)," ",tracks$negative,sep="")
            negt <- paste(paste0("    ",negt),sep="",collapse="\n")
            aTrack <- c(meta,"\n\n",post,"\n\n",negt,"\n\n")
            cat(aTrack,file=file.path(trackDbPath,"trackDb.txt"),sep="",
                append=TRUE)
        }
    }
    else {
        pretdb <- .createUnstrandedSignalTracks(targets,org,normTo,urlBase,
            exportPath,overwrite=overwrite,asLines=FALSE,rc=rc)
        for (p in pretdb) {
            meta <- paste(paste(names(p)," ",p,sep=""),collapse="\n")
            aTrack <- c(meta,"\n\n")
            cat(aTrack,file=file.path(trackDbPath,"trackDb.txt"),sep="",
                append=TRUE)
        }
    }
    
    # Is there an annotation bigBed track?
    if (hasGtf) {
        annTrack <- list(
            track=paste0(org,"_genes"),
            type="bigGenePred",
            shortLabel=paste0(org," genes"),
            longLabel=paste0(org," gene models"),
            boxedCfg="on",
            color="184,0,212",
            exonArrows="on",
            visibility="pack",
            group="annotation",
            bigDataUrl=paste0(urlBase,"/",org,"/",paste0(org,".bigBed"))
        )
        annTrack <- paste(paste(names(annTrack)," ",annTrack,sep=""),
            collapse="\n")
        annTrack <- c(annTrack,"\n\n")
        cat(annTrack,file=file.path(trackDbPath,"trackDb.txt"),sep="",
            append=TRUE)
    }
    
    # Finally, write and return the hub file
    hub <- .makeTrackhubEntrypoint(hubInfo)
    
    writeLines(paste(names(hub)," ",hub,sep=""),
        file.path(exportPath,"hub.txt"))
}

.createStrandedSignalTracks <- function(targets,org,normTo,urlBase,
    exportPath,hubInfo,fasta=NULL,gtf=NULL,overwrite=FALSE,rc=NULL) {
    # First check if bigWig files already exist
    supOrg <- org %in% getSupportedOrganisms()
    if (supOrg)
        org <- getUcscOrganism(org)
    
    #trackDbPath <- file.path(exportPath,getUcscOrganism(org))
    trackDbPath <- file.path(exportPath,org)
    posCheck <- file.path(trackDbPath,
        paste0(unlist(targets$samples,use.names=FALSE),"_plus.bigWig"))
    negCheck <- file.path(trackDbPath,
        paste0(unlist(targets$samples,use.names=FALSE),"_minus.bigWig"))
    posEx <- file.exists(posCheck)
    negEx <- file.exists(negCheck)
    if (all(posEx) && all(negEx) && !overwrite) {
        message("Requested tracks have already been created! ",
            "Use overwrite=TRUE to re-create them.")
        return("")
    }

    # Positive and negative color options
    posBaseColours <- .getPosBaseColors()
    negBaseColours <- .getNegBaseColors()
    posCol <- rep(posBaseColours,length.out=length(targets$samples))
    posCol <- rep(posCol,lengths(targets$samples))
    negCol <- rep(negBaseColours,length.out=length(targets$samples))
    negCol <- rep(negCol,lengths(targets$samples))
    names(posCol) <- names(negCol) <- unlist(targets$samples,use.names=FALSE)

    # The BAM files
    bams <- unlist(targets$files,use.names=FALSE)
    groups <- rep(names(targets$files),lengths(targets$files))
    if (is.null(groups))
        groups <- rep("Signal",length(bams))

    # Get seqinfo form first BAM
    preSf <- .chromInfoFromBAM(bams[1])
    if (supOrg) {
        vchrs <- getValidChrs(org)
        preSf <- preSf[intersect(vchrs,rownames(preSf)),,drop=FALSE]
    }
    else
        vchrs <- rownames(preSf)
    sf <- .chromInfoToSeqInfoDf(preSf,o=org,asSeqinfo=TRUE)

    # Get coverage and assign seqinfo for bigwig
    message("Reading positive strand reads from BAM files to Rle...")
    pbg <- cmclapply(bams,function(b,v,s) {
        message("  reading ",b)
        reads <- trim(unlist(grglist(readGAlignments(file=b,
            param=ScanBamParam(scanBamFlag(isMinusStrand=FALSE))))))
        cov <- coverage(reads)
        seqs <- as.character(seqlevels(reads))
        vv <- intersect(v,seqs)
        cov <- cov[vv]
        gr <- as(IRanges::slice(cov,1),"GRanges")
        seqinfo(gr) <- s
        return(gr)
    },vchrs,sf,rc=rc)
    names(pbg) <- names(posCol)

    message("Reading negative strand reads from BAM files to Rle...")
    nbg <- cmclapply(bams,function(b,v,s) {
        message("  reading ",b)
        reads <- trim(unlist(grglist(readGAlignments(file=b,
            param=ScanBamParam(scanBamFlag(isMinusStrand=TRUE))))))
        cov <- coverage(reads)
        seqs <- as.character(seqlevels(reads))
        vv <- intersect(v,seqs)
        cov <- cov[vv]
        gr <- as(IRanges::slice(cov,1),"GRanges")
        # Inverse the coverage
        gr$score <- -gr$score
        seqinfo(gr) <- s
        return(gr)
    },vchrs,sf,rc=rc)
    names(nbg) <- names(negCol)

    # Calculate normalization factors
    rawPosSums <- vapply(pbg,function(x) sum(x$score),numeric(1))
    rawNegSums <- vapply(nbg,function(x) sum(x$score),numeric(1))

    rat <- rawPosSums/-rawNegSums
    posNormTo <- rat*0.5*normTo
    negNormTo <- normTo - posNormTo

    posNormFacs <- 0.5*posNormTo/rawPosSums
    negNormFacs <- -0.5*negNormTo/rawNegSums

    names(posNormFacs) <- names(pbg)
    names(negNormFacs) <- names(nbg)

    # Normalize (should be quick)
    message("Normalizing positive strand...")
    npbg <- cmclapply(names(pbg),function(n,B,N) {
        B[[n]]$score <- B[[n]]$score*N[n]
        return(B[[n]])
    },pbg,posNormFacs)
    names(npbg) <- names(pbg)
    
    message("Normalizing negative strand...")
    nnbg <- cmclapply(names(nbg),function(n,B,N) {
        B[[n]]$score <- B[[n]]$score*N[n]
        return(B[[n]])
    },nbg,negNormFacs)
    names(nnbg) <- names(nbg)
    
    # Start creating the hub
    # First the dir structure
    if (!dir.exists(trackDbPath))
        dir.create(trackDbPath,recursive=TRUE)
        
    # Put the bigWig files there
    message("Exporting bigWig files...")
    posBwFiles <- file.path(trackDbPath,paste0(names(npbg),"_plus.bigWig"))
    names(posBwFiles) <- names(npbg)
    for (n in names(npbg)) {
        message("  exporting + strand for sample ",n," to ",posBwFiles[n])
        export.bw(npbg[[n]],posBwFiles[n])
    }
    negBwFiles <- file.path(trackDbPath,paste0(names(nnbg),"_minus.bigWig"))
    names(negBwFiles) <- names(nnbg)
    for (n in names(nnbg)) {
        message("  exporting - strand for sample ",n," to ",negBwFiles[n])
        export.bw(nnbg[[n]],negBwFiles[n])
    }
    
    ## Write the genomes file
    #gf <- .makeTrackhubGenomesFile(org)
    #writeLines(paste(names(gf)," ",gf,sep=""),
    #    file.path(exportPath,"genomes.txt"))
    
    ## Write the hub file
    #hub <- .makeTrackhubEntrypoint(hubInfo)
    #writeLines(paste(names(hub)," ",hub,sep=""),
    #    file.path(exportPath,"hub.txt"))
    
    # Write the trackDb file...
    pretdb <- .makeTrackhubMultiwig(names(posBwFiles),posBwFiles,negBwFiles,
        posCol,negCol,urlBase,org,groups)
    
    return(pretdb)
    
    ## Clear previous file if exists, otherwise append will cause problems
    #if (file.exists(file.path(trackDbPath,"trackDb.txt")))
    #    unlink(file.path(trackDbPath,"trackDb.txt"))
    #for (p in pretdb) {
    #    tracks <- p$tracks
    #    p$tracks <- NULL
    #    meta <- paste(paste(names(p)," ",p,sep=""),collapse="\n")
    #    post <- paste(names(tracks$positive)," ",tracks$positive,sep="")
    #    post <- paste(paste0("    ",post),sep="",collapse="\n")
    #    negt <- paste(names(tracks$negative)," ",tracks$negative,sep="")
    #    negt <- paste(paste0("    ",negt),sep="",collapse="\n")
    #    aTrack <- c(meta,"\n\n",post,"\n\n",negt,"\n\n")
    #    cat(aTrack,file=file.path(trackDbPath,"trackDb.txt"),sep="",
    #        append=TRUE)
    #}
    
    ## Generate the hub link
    #return(paste0(urlBase,"/hub.txt"))
}

.createUnstrandedSignalTracks <- function(targets,org,normTo,urlBase,
    exportPath,overwrite,asLines=TRUE,rc=NULL) {
    supOrg <- org %in% getSupportedOrganisms()
    if (supOrg)
        org <- getUcscOrganism(org)
    if (!asLines) # Belong to a trackhub, org is attached to the exportPath
        exportPath <- file.path(exportPath,org)
    
    # First check if bigWig files already exist
    bCheck <- file.path(exportPath,paste0(unlist(targets$samples,
        use.names=FALSE),".bigWig"))
    bEx <- file.exists(bCheck)
    if (all(bEx) && !overwrite) {
        message("Requested tracks have already been created! ",
            "Use overwrite=TRUE to re-create them.")
        return("")
    }
    
    # Create color scheme
    posBaseColours <- .getPosBaseColors()
    posCol <- rep(posBaseColours,length.out=length(targets$samples))
    posCol <- rep(posCol,lengths(targets$samples))
    names(posCol) <- unlist(targets$samples,use.names=FALSE)
    
    # The BAM files
    bams <- unlist(targets$files,use.names=FALSE)
    groups <- rep(names(targets$files),lengths(targets$files))
    if (is.null(groups))
        groups <- rep("Signal",length(bams))
    
    # Get seqinfo form first BAM
    preSf <- .chromInfoFromBAM(bams[1])
    if (supOrg) {
        vchrs <- getValidChrs(org)
        preSf <- preSf[intersect(vchrs,rownames(preSf)),,drop=FALSE]
    }
    else
        vchrs <- rownames(preSf)
    sf <- .chromInfoToSeqInfoDf(preSf,o=org,asSeqinfo=TRUE)
    
    # Get standed coverage and assign seqinfo for bigwig
    message("Reading BAM files to Rle...")
    bg <- cmclapply(bams,function(b,v,s) {
        message("  reading ",b)
        reads <- trim(unlist(grglist(readGAlignments(file=b))))
        cov <- coverage(reads)
        seqs <- as.character(seqlevels(reads))
        vv <- intersect(v,seqs)
        cov <- cov[vv]
        gr <- as(IRanges::slice(cov,1),"GRanges")
        seqinfo(gr) <- s
        return(gr)
    },vchrs,sf,rc=rc)
    names(bg) <- names(posCol)
    
    # Calculate normalization factors
    rawSums <- vapply(bg,function(x) sum(x$score),numeric(1))
    normFacs <- normTo/rawSums
    names(normFacs) <- names(bg)
    
    # Normalize (should be quick)
    message("Normalizing...")
    nbg <- cmclapply(names(bg),function(n,B,N) {
        B[[n]]$score <- B[[n]]$score*N[n]
        return(B[[n]])
    },bg,normFacs)
    names(nbg) <- names(bg)
    
    message("Exporting bigWig files...")
    bwFiles <- file.path(exportPath,paste0(names(nbg),".bigWig"))
    names(bwFiles) <- names(nbg)
    for (n in names(nbg)) {
        message("  exporting for sample ",n," to ",bwFiles[n])
        export.bw(nbg[[n]],bwFiles[n])
    }
    
    if (asLines) { # For simple usage
        message("Creating track lines...")
        opts <- vector("list",length(posCol))
        trackLines <- character(length(bwFiles))
        for (i in seq_along(bwFiles)) {
            opts[[i]]$type <- "bigWig"
            opts[[i]]$name <- sub("^([^.]*).*","\\1",basename(bwFiles[i]))
            opts[[i]]$name <- gsub(" ","_",opts[[i]]$name)
            opts[[i]]$description <- paste0("\"",opts[[i]]$name," signal\"")
            opts[[i]]$color <- paste(t(col2rgb(posCol[i])),collapse=",")
            opts[[i]]$visibility <- "full"
            opts[[i]]$maxHeightPixels <- "128:64:16"
            opts[[i]]$bigDataUrl <- paste0(urlBase,"/",basename(bwFiles[i]))
            popts <- paste(names(opts[[i]]),"=",opts[[i]],sep="")
            trackLines[i] <- paste(popts,collapse=" ")
            trackLines[i] <- paste("track",trackLines[i])
        }
        return(trackLines)
    }
    else { # For trackhub usage
        opts <- .makeTrackhubSinglewig(names(bwFiles),bwFiles,posCol,urlBase,
            org,groups)
        return(opts)
    }
    
    #tracksFile <- file.path(exportPath,"tracks.txt")
    #writeLines(trackLines,tracksFile)
    #tracksLink <- paste0(urlBase,"/tracks.txt")
    
    #return(tracksLink)
}

.makeTrackhubSinglewig <- function(tnames,files,cols,url,org,groups) {
    if (org %in% getSupportedOrganisms())
        org <- getUcscOrganism(org)
    opts <- vector("list",length(cols))
    for (i in seq_along(files)) {
        opts[[i]]$track <- tolower(sub("^([^.]*).*","\\1",basename(files[i])))
        opts[[i]]$type <- "bigWig"
        opts[[i]]$shortLabel <- sub("^([^.]*).*","\\1",basename(files[i]))
        opts[[i]]$shortLabel <- gsub(" ","_",opts[[i]]$shortLabel)
        opts[[i]]$longLabel <- paste0(opts[[i]]$shortLabel," signal")
        opts[[i]]$color <- paste(t(col2rgb(cols[i])),collapse=",")
        opts[[i]]$visibility <- "full"
        opts[[i]]$maxHeightPixels <- "128:64:16"
        opts[[i]]$autoScale <- "on"
        opts[[i]]$group <- tolower(groups[i])
        opts[[i]]$bigDataUrl <- paste0(url,"/",org,"/",basename(files[i]))
    }
    return(opts)
}

.makeTrackhubMultiwig <- function(tnames,pfiles,nfiles,pcols,ncols,
    url,org,groups) {
    if (org %in% getSupportedOrganisms())
        org <- getUcscOrganism(org)
    opts <- vector("list",length(pcols))
    for (i in seq_along(pfiles)) {
        opts[[i]]$track <- paste0(tnames[i],"_stranded")
        opts[[i]]$type <- "bigWig"
        opts[[i]]$container <- "multiWig"
        opts[[i]]$aggregate <- "transparentOverlay"
        opts[[i]]$showSubtrackColorOnUi <- "on"
        opts[[i]]$shortLabel <- tnames[i]
        opts[[i]]$longLabel <- paste0(tnames[i]," signal")
        opts[[i]]$boxedCfg <- "on"
        opts[[i]]$autoScale <- "on"
        opts[[i]]$visibility <- "full"
        opts[[i]]$maxHeightPixels <- "128:64:16"
        opts[[i]]$group <- tolower(groups[i])
        opts[[i]]$tracks <- list(
            positive=list(
                track=paste0(tnames[i],"_plus"),
                parent=paste0(tnames[i],"_stranded"),
                type="bigWig",
                color=paste(t(col2rgb(pcols[i])),collapse=","),
                bigDataUrl=paste0(url,"/",org,"/",paste0(tnames[i],
                    "_plus.bigWig"))
            ),
            negative=list(
                track=paste0(tnames[i],"_minus"),
                parent=paste0(tnames[i],"_stranded"),
                type="bigWig",
                color=paste(t(col2rgb(ncols[i])),collapse=","),
                bigDataUrl=paste0(url,"/",org,"/",paste0(tnames[i],
                    "_minus.bigWig"))
            )
        )
    }
    return(opts)
}

.makeTrackhubGroupsFile <- function(targets,hasGtf=FALSE) {
    groups <- names(targets$files)
    if (is.null(groups))
        groups <- "Signal"
    
    if (hasGtf)
        grp <- vector("list",length(groups)+1)
    else
        grp <- vector("list",length(groups))
        
    for (i in seq_len(length(groups))) {
        grp[[i]]$name <- tolower(gsub(" ","_",groups[i]))
        grp[[i]]$label <- groups[i]
        grp[[i]]$priority <- i
        grp[[i]]$defaultIsClosed <- 0
    }
    
    n <- length(grp)
    if (hasGtf) {
        grp[[n]]$name <- "annotation"
        grp[[n]]$label <- "Annotation"
        grp[[n]]$priority <- n
        grp[[n]]$defaultIsClosed <- 0
    }
    
    return(grp)
}

.makeTrackhubGenomesFile <- function(org,b=NULL,has2bit=FALSE) {
    supOrg <- FALSE
    if (org %in% getSupportedOrganisms()) {
        supOrg <- TRUE
        org <- getUcscOrganism(org)
    }
    gf <- list(
        genome=org,
        groups=paste0(org,"/groups.txt"),
        trackDb=paste0(org,"/trackDb.txt")
    )
    if (!supOrg) {
        ci <- .chromInfoFromBAM(b)
        chr <- rownames(ci)[1]
        len <- as.integer(ci[1,1])
        middle <- round(len/2)
        start <- sample.int(middle,1)
        end <- sample((middle+1):(len-1),1)
        gf$organism <- org
        gf$description <- paste0("Hub for ",org)
        gf$defaultPos <- paste0(chr,":",start,"-",end)
    }
    if (has2bit)
        gf$twoBitPath <- paste0(org,"/",org,".2bit")
    return(gf)
}

.makeTrackhubEntrypoint <- function(h) {
    return(list(
        hub=h$name,
        shortLabel=h$shortLabel,
        longLabel=h$longLabel,
        genomesFile="genomes.txt",
        email=h$email
    ))
}

.getPosBaseColors <- function() {
    return(c("#B40000","#00B400","#0000B4","#B45200","#9B59B6","#21BCBF",
        "#BC4800","#135C34","#838F00","#4900B5"))
}

.getNegBaseColors <- function() {
    return(c("#FF7575","#79FF79","#8484FF","#FFB77C","#EBB7FF","#63E5E7",
        "#FFB88B","#5BFFA5","#F4FF75","#C69FFF"))
}

.externalGtfToBigBed <- function(gtf,ci) {
    # 1. Get the required tools
    # 1.1. gtfToGenePred
    gtfToGenePred <- file.path(tempdir(),"gtfToGenePred")
    if (!file.exists(file.path(tempdir(),"gtfToGenePred"))) {
        message("  Retrieving gtfToGenePred tool")
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/gtfToGenePred",
            gtfToGenePred,quiet=TRUE
        )
        system(paste("chmod 775",gtfToGenePred))
    }
    
    # 1.2. genePredToBigGenePred
    genePredToBigGenePred <- file.path(tempdir(),"genePredToBigGenePred")
    if (!file.exists(file.path(tempdir(),"genePredToBigGenePred"))) {
        message("  Retrieving genePredToBigGenePred tool")
        download.file(paste0("http://hgdownload.soe.ucsc.edu/admin/exe/",
            "linux.x86_64/genePredToBigGenePred"),genePredToBigGenePred,
            quiet=TRUE
        )
        system(paste("chmod 775",genePredToBigGenePred))
    }
    
    # 1.3. bedToBigBed
    bedToBigBed <- file.path(tempdir(),"bedToBigBed")
    if (!file.exists(file.path(tempdir(),"bedToBigBed"))) {
        message("  Retrieving bedToBigBed tool")
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/bedToBigBed",
            bedToBigBed,quiet=TRUE
        )
        system(paste("chmod 775",bedToBigBed))
    }
    
    # 1.4. bigGenePred.as
    bigGenePred.as <- file.path(tempdir(),"bigGenePred.as")
    if (!file.exists(file.path(tempdir(),"bigGenePred.as"))) {
        message("  Retrieving bigGenePred.as definition")
        download.file(
            "https://genome.ucsc.edu/goldenPath/help/examples/bigGenePred.as",
            bigGenePred.as,quiet=TRUE
        )
        system(paste("chmod 775",bigGenePred.as))
    }
    
    # 2. Run gtfToGenePred
    #gtfFile <- file.path(tempdir(),basename(gtf))
    message("  Converting ",basename(gtf)," to genePred")
    tmpGenePred <- file.path(tempdir(),paste(format(Sys.time(),"%Y%m%d%H%M%S"),
        "tgp",sep="."))
    command1 <- paste(gtfToGenePred,"-genePredExt",gtf,tmpGenePred)
    message("Executing: ",command1)
    system(command1)
    
    # 3. Run genePredToBigGenePred
    message("  Converting ",basename(tmpGenePred)," to bigGenePred")
    tmpBigGenePred <- file.path(tempdir(),paste(format(Sys.time(),
        "%Y%m%d%H%M%S"),"tbgp",sep="."))
    command2 <- paste(genePredToBigGenePred,tmpGenePred,tmpBigGenePred)
    message("Executing: ",command2)
    system(command2)
    
    # 4. Write the chromosome info file
    chromInfo <- file.path(tempdir(),"chromInfo.txt")
    write.table(ci,file=chromInfo,sep="\t",col.names=FALSE,quote=FALSE)
    
    # 5. Run sort
    message("  Sorting ",tmpBigGenePred)
    tmpSort <- file.path(tempdir(),paste(format(Sys.time(),
        "%Y%m%d%H%M%S"),"txt",sep="."))
    command3 <- paste("sort -k1,1 -k2n,2",tmpBigGenePred,">",tmpSort)
    message("Executing: ",command3)
    system(command3)
    
    # 6. Run bedToBigBed
    tmpBigBed <- file.path(tempdir(),paste(format(Sys.time(),
        "%Y%m%d%H%M%S"),"bigBed",sep="."))
    message("  Converting sorted ",basename(tmpSort)," to bigBed")
    command4 <- paste(bedToBigBed," -type=bed12+8 -tab -as=",bigGenePred.as,
        " ",tmpSort," ",chromInfo," ",tmpBigBed,sep="")
    message("Executing: ",command4)
    system(command4)
    
    return(tmpBigBed)
}

#' @docType data
#' @name mm9GeneCounts
#' @title mouse RNA-Seq data with two conditions, four samples
#' @description This data set contains RNA-Seq gene read counts for 3 
#' chromosomes.
#' The data were downloaded from the ENCODE public repository and are derived
#' from the study of Mortazavi et al., 2008 (Mortazavi A, Williams BA, McCue K, 
#' Schaeffer L, Wold B. Mapping and quantifying mammalian transcriptomes by 
#' RNA-Seq.
#' Nat Methods. 2008 Jul;5(7):621-8). In their experiment, the authors studied
#' among others genes expression at two developmental stages of mouse liver 
#' cells.
#' It has two conditions-developmental stages (e14.5,
#' adult_8_weeks) and four samples (e14.5_1, e14.5_2, a8w_1, a8w_2). It also
#' contains a predefined \code{sampleList} and \code{libsizeList}
#' named \code{sampleListMm9} and \code{libsizeListMm9}.
#' @usage mm9.gene.counts
#' @format a \code{data.frame} with gene read counts and some embedded 
#' annotation, one row per gene.
#' @source ENCODE (http://genome.ucsc.edu/encode/)
#' @author Panagiotis Moulos
NULL
#' @docType data
#' @name sampleListMm9
#' @title Mouse RNA-Seq data with two conditions, four samples
#' @description The sample list for \code{mm9GeneCounts}. See the data set
#' description.
#' @usage sampleListMm9
#' @format a named \code{list} with condition and sample names.
#' @source ENCODE (http://genome.ucsc.edu/encode/)
#' @author Panagiotis Moulos
NULL
#' @docType data
#' @name libsizeListMm9
#' @title Mouse RNA-Seq data with two conditions, four samples
#' @description The library size list for \code{mm9GeneCounts}. See the data set
#' description.
#' @usage libsizeListMm9
#' @format a named \code{list} with library sizes.
#' @source ENCODE (http://genome.ucsc.edu/encode/)
#' @author Panagiotis Moulos
NULL

#' An R package for the analysis and result reporting of RNA-Seq gene expression
#' data, using multiple statistical algorithms.
#'
#' \tabular{ll}{
#' Package: \tab metaseqR2\cr
#' Type: \tab Package\cr
#' Version: \tab 0.0.20\cr
#' Date: \tab 2020-03-13\cr
#' Depends: \tab R (>= 4.0.0), limma, qvalue\cr
#' Encoding: \tab UTF-8\cr
#' License: \tab GPL (>= 3)\cr
#' LazyLoad: \tab yes\cr
#' URL: \tab http://www.fleming.gr\cr
#' }
#'
#' Provides an interface to several normalization and statistical testing 
#' packages for RNA-Seq gene expression data. Additionally, it creates several
#' diagnostic plots, performs meta-analysis by combinining the results of 
#' several statistical tests and reports the results in an interactive way.
#'
#' @aliases metaseqR2-package metaseqR2
#' @name metaseqR2-package
#' @docType package
#' @title The metaseqR2 Package
#' @author Panagiotis Moulos \email{moulos@@fleming.gr}
#' @keywords package
NULL

.makeReportEnv <- function(e) {
    re <- new.env(parent=globalenv())
    
    re$theList <- e$theList
    re$tLink <- e$tLink
    re$createTracks <- e$createTracks
    re$trackExportPath <- e$trackExportPath
    re$trackInfo <- e$trackInfo
    re$offlineReport <- e$offlineReport
    re$reportDb <- e$reportDb
    re$org <- e$org
    re$refdb <- e$refdb
    re$utrOpts <- e$utrOpts
    re$qcPlots <- e$qcPlots
    re$sampleList <- e$sampleList
    re$geneData <- e$geneData
    re$geneCounts <- e$geneCounts
    re$geneDataExpr <- e$geneDataExpr
    re$normGenesExpr <- e$normGenesExpr
    re$fromRaw <- e$fromRaw
    re$transLevel <- e$transLevel
    re$countType <- e$countType
    re$fileType <- e$fileType
    re$geneFilters <- e$geneFilters
    re$exonFilters <- e$exonFilters
    re$normalization <- e$normalization
    re$statistics <- e$statistics
    re$metaP <- e$metaP
    re$pcut <- e$pcut
    re$contrast <- e$contrast
    re$adjustMethod <- e$adjustMethod
    re$sumpList <- e$sumpList
    re$logOffset <- e$logOffset
    re$countsName <- e$countsName
    re$excludeList <- e$excludeList
    re$libsizeList <- e$libsizeList
    re$reportMessages <- e$reportMessages
    re$preset <- e$preset
    re$whenApplyFilter <- e$whenApplyFilter
    re$normArgs <- e$normArgs
    re$statArgs <- e$statArgs
    re$figFormat <- e$figFormat
    re$exportWhere <- e$exportWhere
    re$exportWhat <- e$exportWhat
    re$exportScale <- e$exportScale
    re$exportValues <- e$exportValues
    re$exportStats <- e$exportStats
    re$execTime <- e$execTime
    re$theZeros <- e$theZeros
    re$theDead <- e$theDead
    re$exonFilterResult <- e$exonFilterResult
    re$geneFilterResult <- e$geneFilterResult
    re$geneFilterCutoff <- e$geneFilterCutoff
    re$runLog <- e$runLog
    re$geneDataFiltered <- e$geneDataFiltered
    re$totalGeneData <- e$totalGeneData
    re$normGenes <- e$normGenes
    re$geneCountsZero <- e$geneCountsZero
    re$contrastList <- e$contrastList
    re$reportTop <- e$reportTop
    re$cpList <- e$cpList
    re$reportTables <- e$reportTables
    re$exportCountsTable <- e$exportCountsTable
    
    re$FUN_CALL <- e$FUN_CALL
    re$PROJECT_PATH <- e$PROJECT_PATH
    
    return(re)
}

.createSqlPlotDb <- function(obj) {
    sampleList <- obj$sampleList
    contrastList <- obj$contrastList
    qcPlots <- obj$qcPlots
    geneCounts <- obj$geneCounts
    geneData <- obj$geneData
    geneDataExpr <- obj$geneDataExpr
    geneDataFiltered <- obj$geneDataFiltered
    totalGeneData <- obj$totalGeneData
    normGenes <- obj$normGenes
    normGenesExpr <- obj$normGenesExpr
    cpList <- obj$cpList
    sumpList <- obj$sumpList
    pcut <- obj$pcut
    metaP <- obj$metaP
    whenApplyFilter <- obj$whenApplyFilter
    PROJECT_PATH <- obj$PROJECT_PATH
    
    samples <- unlist(sampleList)
    nsa <- length(samples)
    
    covarsRaw <- covarsStat <- NULL
    if (any(qcPlots %in% c("biodetection","countsbio","saturation",
        "rnacomp","readnoise"))) {
        covarsRaw <- list(
            data=geneCounts,
            length=width(geneData),
            gc=as.numeric(geneData$gc_content),
            chromosome=data.frame(
                chromosome=as.character(seqnames(geneData)),
                start=start(geneData),
                end=end(geneData)
            ),
            factors=data.frame(class=asClassVector(sampleList)),
            biotype=as.character(geneData$biotype),
            gene_name=as.character(geneData$gene_name)
        )
    }

    if ("biodist" %in% qcPlots) {
        covarsStat <- list(
            data=normGenesExpr,
            length=width(geneDataExpr),
            gc=as.numeric(geneDataExpr$gc_content),
            chromosome=data.frame(
                chromosome=as.character(seqnames(geneDataExpr)),
                start=start(geneDataExpr),
                end=end(geneDataExpr)
            ),
            factors=data.frame(class=asClassVector(sampleList)),
            biotype=as.character(geneDataExpr$biotype),
            gene_name=as.character(geneDataExpr$gene_name)
        )
    }

    rdb <- file.path(PROJECT_PATH$data,"reportdb.sqlite")
    disp("Opening plot database in ",rdb)
    con <- dbConnect(dbDriver("SQLite"),dbname=rdb)
    rs <- .initReportDbTables(con)

    if ("mds" %in% qcPlots) {
        disp("Importing mds...")
        json <- diagplotMds(geneCounts,sampleList,output="json")
        .dbImportPlot(con,"MDS","mds","generic",json)
    }
    if ("biodetection" %in% qcPlots) {
        disp("  Importing biodetection...")
        cap <- capture.output({
            json <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
                whichPlot="biodetection",output="json")
        })
        for (s in samples) {
            disp("    ",s)
            name <- paste("biodetection",s,sep="_")
            .dbImportPlot(con,name,"biodetection","generic",json[[s]])
        }
    }
    if ("countsbio" %in% qcPlots) {
        disp("  Importing countsbio...")
        cap <- capture.output({
            jsonList <- diagplotNoiseq(geneCounts,sampleList,
                covars=covarsRaw,whichPlot="countsbio",output="json")
        })
        for (s in samples) {
            disp("    ",s)
            name <- paste("countsbio",s,sep="_")
            .dbImportPlot(con,name,"countsbio","sample",
                jsonList[["sample"]][[s]])
        }
        for (b in names(jsonList[["biotype"]])) {
            disp("    ",b)
            name <- paste("countsbio",b,sep="_")
            .dbImportPlot(con,name,"countsbio","biotype",
                jsonList[["biotype"]][[b]])
        }
    }
    if ("saturation" %in% qcPlots) {
        disp("  Importing saturation...")
        jsonList <- diagplotNoiseq(geneCounts,sampleList,
            covars=covarsRaw,whichPlot="saturation",output="json")
        for (s in unlist(sampleList)) {
            disp("    ",s)
            name <- paste("saturation",s,sep="_")
            .dbImportPlot(con,name,"saturation","sample",
                jsonList[["sample"]][[s]])
        }
        for (b in names(jsonList[["biotype"]])) {
            disp("    ",b)
            name <- paste("saturation",b,sep="_")
            .dbImportPlot(con,name,"saturation","biotype",
                jsonList[["biotype"]][[b]])
        }
    }
    if ("readnoise" %in% qcPlots) {
        disp("  Importing readnoise...")
        json <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
            whichPlot="readnoise",output="json")
        .dbImportPlot(con,"ReadNoise","readnoise","generic",json)
    }
    if ("pairwise" %in% qcPlots) {
        disp("  Importing pairwise...")
        jsonList <- diagplotPairs(geneCounts,output="json")
        for (name in names(jsonList$xy)) {
            disp("    ",name)
            .dbImportPlot(con,name,"pairwise","xy",jsonList$xy[[name]])
            .dbImportPlot(con,name,"pairwise","md",jsonList$md[[name]])
        }
    }
    if ("filtered" %in% qcPlots) {
        disp("  Importing filtered...")
        jsonList <- diagplotFiltered(geneDataFiltered,totalGeneData,
            output="json")
        .dbImportPlot(con,"filtered_chromosome","filtered","chromosome",
            jsonList[["chromosome"]])
        .dbImportPlot(con,"filtered_biotype","filtered","biotype",
            jsonList[["biotype"]])
    }

    if ("boxplot" %in% qcPlots) {
        disp("  Importing boxplot...")
        jsonUnorm <- diagplotBoxplot(geneCounts,name=sampleList,
            isNorm=FALSE,output="json")
        jsonNorm <- diagplotBoxplot(normGenes,name=sampleList,
            isNorm=FALSE,output="json")
        .dbImportPlot(con,"Boxplot","boxplot","unorm",jsonUnorm)
        .dbImportPlot(con,"Boxplot","boxplot","norm",jsonNorm)
    }
    if ("gcbias" %in% qcPlots) {
        disp("  Importing gcbias...")
        covarRaw <- as.numeric(geneData$gc_content)
        # Covers whenApplyFilter="prenorm" case
        covarNorm <- as.numeric(geneData[rownames(normGenes)]$gc_content)
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,covar=covarRaw,
            isNorm=FALSE,whichPlot="gcbias",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,covar=covarNorm,
            isNorm=TRUE,whichPlot="gcbias",output="json")
        .dbImportPlot(con,"GCBias","gcbias","unorm",jsonUnorm)
        .dbImportPlot(con,"GCBias","gcbias","norm",jsonNorm)
    }

    if ("lengthbias" %in% qcPlots) {
        disp("  Importing lengthbias...")
        covarRaw <- as.numeric(geneData$gc_content)
        # Covers whenApplyFilter="prenorm" case
        covarNorm <- as.numeric(geneData[rownames(normGenes)]$gc_content)
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,covar=covarRaw,
            isNorm=FALSE,whichPlot="lengthbias",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,covar=covarNorm,
            isNorm=TRUE,whichPlot="lengthbias",output="json")
        .dbImportPlot(con,"LengthBias","lengthbias","unorm",jsonUnorm)
        .dbImportPlot(con,"LengthBias","lengthbias","norm",jsonNorm)
    }

    if ("meandiff" %in% qcPlots) {
        disp("  Importing meandif...")
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,isNorm=FALSE,
            whichPlot="meandiff",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,isNorm=TRUE,
            whichPlot="meandiff",output="json")
        for (n in names(jsonUnorm)) {
            for (s in names(jsonUnorm[[n]])) {
                disp("    ",s)
                .dbImportPlot(con,s,"meandiff","unorm",
                    jsonUnorm[[n]][[s]])
                .dbImportPlot(con,s,"meandiff","norm",
                    jsonNorm[[n]][[s]])
            }
        }
    }
    if ("meanvar" %in% qcPlots) {
        disp("  Importing meanvar...")
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,isNorm=FALSE,
            whichPlot="meanvar",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,isNorm=TRUE,
            whichPlot="meanvar",output="json")
        .dbImportPlot(con,"MeanVar","meanvar","unorm",jsonUnorm)
        .dbImportPlot(con,"MeanVar","meanvar","orm",jsonNorm)
    }
    if ("rnacomp" %in% qcPlots) {
        disp("  Importing rnacomp...")
        cap1 <- capture.output({
            jsonUnorm <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
                whichPlot="rnacomp",isNorm=FALSE,output="json")
        })
        cap2 <- capture.output({
            jsonNorm <- diagplotNoiseq(normGenes,sampleList,
            covars=if (whenApplyFilter=="prenorm") covarsStat else covarsRaw,
                whichPlot="rnacomp",isNorm=TRUE,output="json")
        })
        .dbImportPlot(con,"RnaComp","rnacomp","unorm",jsonUnorm)
        .dbImportPlot(con,"RnaComp","rnacomp","norm",jsonNorm)
    }
    if ("volcano" %in% qcPlots) {
        disp("  Importing volcano")
        nn <- names(contrastList)
        for (n in nn) {
            fc <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            for (contr in colnames(fc)) {
                disp("    ",n," ",contr)
                json <- diagplotVolcano(fc[,contr],sumpList[[n]],contr,
                    altNames=geneDataExpr$gene_name,output="json")
                .dbImportPlot(con,paste("volcano",contr,sep="_"),
                    "volcano","generic",json)
            }
        }
    }
    if ("mastat" %in% qcPlots) {
        disp("  Importing mastat")
        nn <- names(contrastList)
        for (n in nn) {
            m <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            a <- makeA(n,sampleList,normGenesExpr,1)
            for (contr in colnames(m)) {
                disp("    ",n," ",contr)
                json <- diagplotMa(m[,contr],a[,contr],sumpList[[n]],
                    contr,altNames=geneDataExpr$gene_name,output="json")
                .dbImportPlot(con,paste("mastat",contr,sep="_"),
                    "mastat","generic",json)
            }
        }
    }
    if ("biodist" %in% qcPlots) {
        disp("  Importing biodist")
        nn <- names(contrastList)
        for (n in nn) {
            disp("    ",n)
            cap <- capture.output({
                json <- diagplotNoiseq(normGenesExpr,sampleList,
                    covars=covarsStat,whichPlot="biodist",
                    biodistOpts=list(p=cpList[[n]],pcut=pcut,name=n),
                    output="json")
            })
            .dbImportPlot(con,paste("biodist",n,sep="_"),"biodist",
                "chromosome",json$chromosome)
            .dbImportPlot(con,paste("biodist",n,sep="_"),"biodist",
                "biotype",json$biotype)
        }
    }
    if ("statvenn" %in% qcPlots) {
        disp("  Importing statvenn")
        nn <- names(contrastList)
        geneNames <- as.character(geneDataExpr$gene_name)
        names(geneNames) <- names(geneDataExpr)
        for (n in nn) {
            disp("    ",n)
            fc <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            # To narrow We take always the first condition versus the reference
            # to narrow down a bit the list of DE genes
            fc <- fc[,1,drop=FALSE]
            fcmat <- matrix(data=apply(fc,2,function(x) {
                rep(x,length(colnames(cpList[[n]])))
            }),ncol=ncol(fc)*ncol(cpList[[n]]))
            colnames(fcmat) <- colnames(cpList[[n]])
            # Deregulated
            json <- makeJVennStatData(cpList[[n]],fcmat=fcmat,pcut=pcut,
                direction="dereg",altNames=geneNames[rownames(cpList[[n]])])
            .dbImportPlot(con,paste("statvenn",n,"dereg",sep="_"),"statvenn",
                "stat_dereg",toJSON(json,auto_unbox=TRUE,null="null"))
            # Up
            json <- makeJVennStatData(cpList[[n]],fcmat=fcmat,pcut=pcut,
                direction="up",altNames=geneNames[rownames(cpList[[n]])])
            .dbImportPlot(con,paste("statvenn",n,"up",sep="_"),"statvenn",
                "stat_up",toJSON(json,auto_unbox=TRUE,null="null"))
            # Down
            json <- makeJVennStatData(cpList[[n]],fcmat=fcmat,pcut=pcut,
                direction="down",altNames=geneNames[rownames(cpList[[n]])])
            .dbImportPlot(con,paste("statvenn",n,"down",sep="_"),"statvenn",
                "stat_down",toJSON(json,auto_unbox=TRUE,null="null"))
        }
    }
    
    if ("foldvenn" %in% qcPlots) {
        disp("  Importing foldvenn")
        nn <- names(contrastList)
        
        ## Fold venn cannot be created for ANOVA-like cases, only p-values can
        ## be used in such cases, so a failsafe
        #tmp <- strsplit(nn,"_vs_")
        #tmpl <- lengths(tmp)
        #nn2 <- NULL
        #if (any(tmpl) > 2) {
        #   jj1 <- which(tmpl==2)
        #   jj2 <- which(tmpl>2)
        #   nn2 <- nn[jj2]
        #   nn <- nn[jj1]
        #}
        
        # We need to construct a matrix with p-values for each contrast and
        # each statistical algorithm
        if (ncol(cpList[[1]]) > 1) {
            N <- names(cpList)
            prePmat <- lapply(N,function(n,C,S) {
                many <- C[[n]]
                one <- S[[n]]
                mat <- cbind(many,one)
                colnames(mat)[ncol(mat)] <- metaP
                return(mat)
            },cpList,sumpList)
            names(prePmat) <- N
        }
        else
            prePmat <- sumpList
        
        # Now we need to reformat prePmat to a list names according to each
        # algorithm
        algs <- colnames(prePmat[[1]])
        pmat <- lapply(algs,function(a,P) {
            algMat <- matrix(NA,nrow(P[[1]]),length(P))
            colnames(algMat) <- names(P)
            rownames(algMat) <- rownames(P[[1]])
            for (cnt in names(P))
                algMat[,cnt] <- prePmat[[cnt]][,a]
            return(algMat)
        },prePmat)
        
        # We need to construct a matrix with fold changes for each contrast
        fList <- vector("list",length(nn))
        names(fList) <- nn
        for (n in nn) {
            fc <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            fc <- fc[,1,drop=FALSE]
            fList[[n]] <- fc
        }
        fcmat <- do.call("cbind",fList)
        fcmat <- fcmat[rownames(pmat),]
        
        # Based on pmat and fcmat, we calculate Venns
        geneNames <- geneDataExpr$gene_name
        names(geneNames) <- names(geneDataExpr)
        for (n in names(pmat)) {
            disp("    ",n)
            jsonList[[n]] <- list(dereg=NULL,up=NULL,down=NULL)
            for (d in c("dereg","up","down"))
                json <- makeJVennFoldData(pmat=pmat,fcmat=fcmat,pcut=pcut,
                    direction=d,altNames=geneNames[rownames(pmat[[n]])])
                .dbImportPlot(con,paste("foldvenn",n,d,sep="_"),"foldvenn",
                    paste0("fold_",d),toJSON(json,auto_unbox=TRUE,null="null"))
        }
    }
    
    if ("deregulogram" %in% qcPlots) {
        disp("  Importing deregulogram")
        cntPairs <- combn(names(contrastList),2)
        for (i in seq_len(ncol(cntPairs))) {
            disp("    ",cntPairs[1,i]," and ",cntPairs[2,i])
            namc <- paste0(cntPairs[1,i],"__",cntPairs[2,i])
            fmat <- cbind(
                log2(makeFoldChange(cntPairs[1,i],sampleList,
                    normGenesExpr,1))[,1,drop=FALSE],
                log2(makeFoldChange(cntPairs[2,i],sampleList,
                    normGenesExpr,1))[,1,drop=FALSE]
            )
            pmat <- do.call("cbind",sumpList[c(cntPairs[1,i],
                cntPairs[2,i])])
            colnames(pmat) <- colnames(fmat)
            json <- diagplotDeregulogram(fmat,pmat,pcut=pcut,fcut=1,
                output="json",altNames=geneDataExpr$gene_name)
            .dbImportPlot(con,paste("deregulogram",namc,sep="_"),
                "deregulogram","generic",json)
        }
    }

    # Close SQLite connection
    dbDisconnect(con)
}

.createDexiePlotDb <- function(obj) {
    sampleList <- obj$sampleList
    contrastList <- obj$contrastList
    qcPlots <- obj$qcPlots
    geneCounts <- obj$geneCounts
    geneData <- obj$geneData
    geneDataExpr <- obj$geneDataExpr
    geneDataFiltered <- obj$geneDataFiltered
    totalGeneData <- obj$totalGeneData
    normGenes <- obj$normGenes
    normGenesExpr <- obj$normGenesExpr
    cpList <- obj$cpList
    sumpList <- obj$sumpList
    pcut <- obj$pcut
    metaP <- obj$metaP
    whenApplyFilter <- obj$whenApplyFilter
    PROJECT_PATH <- obj$PROJECT_PATH
    
    samples <- unlist(sampleList)
    nsa <- length(samples)
    
    covarsRaw <- covarsStat <- NULL
    if (any(qcPlots %in% c("biodetection","countsbio","saturation",
        "rnacomp","readnoise"))) {
        covarsRaw <- list(
            data=geneCounts,
            length=width(geneData),
            gc=as.numeric(geneData$gc_content),
            chromosome=data.frame(
                chromosome=as.character(seqnames(geneData)),
                start=start(geneData),
                end=end(geneData)
            ),
            factors=data.frame(class=asClassVector(sampleList)),
            biotype=as.character(geneData$biotype),
            gene_name=as.character(geneData$gene_name)
        )
    }

    if ("biodist" %in% qcPlots) {
        covarsStat <- list(
            data=normGenesExpr,
            length=width(geneDataExpr),
            gc=as.numeric(geneDataExpr$gc_content),
            chromosome=data.frame(
                chromosome=as.character(seqnames(geneDataExpr)),
                start=start(geneDataExpr),
                end=end(geneDataExpr)
            ),
            factors=data.frame(class=asClassVector(sampleList)),
            biotype=as.character(geneDataExpr$biotype),
            gene_name=as.character(geneDataExpr$gene_name)
        )
    }

    plots <- list()
    listIndex <- 0
    
    if ("mds" %in% qcPlots) {
        disp("  Importing mds...")
        json <- diagplotMds(geneCounts,sampleList,output="json")
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="MDS",
            type="mds",
            subtype="generic",
            json=fromJSON(json)
        )
    }
    
    if ("biodetection" %in% qcPlots) {
        disp("  Importing biodetection...")
        cap <- capture.output({
            json <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
                whichPlot="biodetection",output="json")
        })
        for (i in seq_along(json)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("biodetection",samples[i],sep="_"),
                type="biodetection",
                subtype="generic",
                json=fromJSON(json[[samples[i]]])
            )
        }
    }
    
    if ("countsbio" %in% qcPlots) {
        disp("  Importing countsbio...")
        cap <- capture.output({
            jsonList <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
                whichPlot="countsbio",output="json")
        })
        for (i in seq_along(jsonList$sample)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("countsbio",samples[i],sep="_"),
                type="countsbio",
                subtype="sample",
                json=fromJSON(jsonList$sample[[samples[i]]])
            )
        }   
        biotypes <- names(jsonList$biotype)
        for (i in seq_along(jsonList$biotype)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("countsbio",biotypes[i],sep="_"),
                type="countsbio",
                subtype="biotype",
                json=fromJSON(jsonList$biotype[[biotypes[i]]])
            )
        }
    }
    
    if ("saturation" %in% qcPlots) {
        disp("  Importing saturation...")
        jsonList <- diagplotNoiseq(geneCounts,sampleList,
            covars=covarsRaw,whichPlot="saturation",output="json")
        for (i in seq_along(jsonList$sample)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("saturation",samples[i],sep="_"),
                type="saturation",
                subtype="sample",
                json=fromJSON(jsonList$sample[[samples[i]]])
            )
        }
        biotypes <- names(jsonList$biotype)
        for (i in seq_along(jsonList$biotype)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("saturation",biotypes[i],sep="_"),
                type="saturation",
                subtype="biotype",
                json=fromJSON(jsonList$biotype[[biotypes[i]]])
            )
        }
    }
    
    if ("readnoise" %in% qcPlots) {
        disp("  Importing readnoise...")
        json <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
            whichPlot="readnoise",output="json")
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="ReadNoise",
            type="readnoise",
            subtype="generic",
            json=fromJSON(json)
        )
    }
    
    if ("pairwise" %in% qcPlots) {
        disp("  Importing pairwise...")
        jsonList <- diagplotPairs(geneCounts,output="json")
        nams <- names(jsonList$xy)
        for (i in seq_along(jsonList$xy)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=nams[i],
                type="pairwise",
                subtype="xy",
                json=fromJSON(jsonList$xy[[nams[i]]])
            )
        }
        for (i in seq_along(jsonList$md)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=nams[i],
                type="pairwise",
                subtype="md",
                json=fromJSON(jsonList$md[[nams[i]]])
            )
        }
    }
    
    if ("filtered" %in% qcPlots) {
        disp("  Importing filtered...")
        jsonList <- diagplotFiltered(geneDataFiltered,totalGeneData,
            output="json")
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="filtered_chromosome",
            type="filtered",
            subtype="chromosome",
            json=fromJSON(jsonList[["chromosome"]])
        )
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="filtered_biotype",
            type="filtered",
            subtype="biotype",
            json=fromJSON(jsonList[["biotype"]])
        )
    }

    if ("boxplot" %in% qcPlots) {
        disp("  Importing boxplot...")
        jsonUnorm <- diagplotBoxplot(geneCounts,name=sampleList,
            isNorm=FALSE,output="json")
        jsonNorm <- diagplotBoxplot(normGenes,name=sampleList,
            isNorm=FALSE,output="json")
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="Boxplot",
            type="boxplot",
            subtype="unorm",
            json=fromJSON(jsonUnorm)
        )
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="Boxplot",
            type="boxplot",
            subtype="norm",
            json=fromJSON(jsonNorm)
        )
    }
    
    if ("gcbias" %in% qcPlots) {
        disp("  Importing gcbias...")
        covarRaw <- as.numeric(geneData$gc_content)
        # Covers whenApplyFilter="prenorm" case
        covarNorm <- as.numeric(geneData[rownames(normGenes)]$gc_content)
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,covar=covarRaw,
            isNorm=FALSE,whichPlot="gcbias",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,covar=covarNorm,
            isNorm=TRUE,whichPlot="gcbias",output="json")
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="GCBias",
            type="gcbias",
            subtype="unorm",
            json=fromJSON(jsonUnorm)
        )
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="GCBias",
            type="gcbias",
            subtype="norm",
            json=fromJSON(jsonNorm)
        )
    }

    if ("lengthbias" %in% qcPlots) {
        disp("  Importing lengthbias...")
        covarRaw <- width(geneData[rownames(geneCounts)])
        # Covers whenApplyFilter="prenorm" case
        covarNorm <- as.numeric(geneData[rownames(normGenes)]$gc_content)
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,covar=covarRaw,
            isNorm=FALSE,whichPlot="lengthbias",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,covar=covarNorm,
            isNorm=TRUE,whichPlot="lengthbias",output="json")
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="LengthBias",
            type="lengthbias",
            subtype="unorm",
            json=fromJSON(jsonUnorm)
        )
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="LengthBias",
            type="lengthbias",
            subtype="norm",
            json=fromJSON(jsonNorm)
        )
    }

    if ("meandiff" %in% qcPlots) {
        disp("  Importing meandif...")
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,isNorm=FALSE,
            whichPlot="meandiff",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,isNorm=TRUE,
            whichPlot="meandiff",output="json")
        np <- sum(lengths(jsonUnorm))
        for (i in seq_along(jsonUnorm)) {
            for (j in seq_along(jsonUnorm[[i]])) {
                listIndex <- listIndex + 1
                plots[[listIndex]] <- list(
                    name=names(jsonUnorm[[i]])[j],
                    type="meandiff",
                    subtype="unorm",
                    json=fromJSON(jsonUnorm[[i]][[j]])
                )
            }
        }
        for (i in seq_along(jsonNorm)) {
            for (j in seq_along(jsonNorm[[i]])) {
                listIndex <- listIndex + 1
                plots[[listIndex]] <- list(
                    name=names(jsonNorm[[i]])[j],
                    type="meandiff",
                    subtype="norm",
                    json=fromJSON(jsonNorm[[i]][[j]])
                )
            }
        }
    }
    
    if ("meanvar" %in% qcPlots) {
        disp("  Importing meanvar...")
        jsonUnorm <- diagplotEdaseq(geneCounts,sampleList,isNorm=FALSE,
            whichPlot="meanvar",output="json")
        jsonNorm <- diagplotEdaseq(normGenes,sampleList,isNorm=TRUE,
            whichPlot="meanvar",output="json")
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="MeanVar",
            type="meanvar",
            subtype="unorm",
            json=fromJSON(jsonUnorm)
        )
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="MeanVar",
            type="meanvar",
            subtype="norm",
            json=fromJSON(jsonNorm)
        )
    }
    
    if ("rnacomp" %in% qcPlots) {
        disp("  Importing rnacomp...")
        cap1 <- capture.output({
            jsonUnorm <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
                whichPlot="rnacomp",isNorm=FALSE,output="json")
        })
        cap2 <- capture.output({
            jsonNorm <- diagplotNoiseq(normGenes,sampleList,
            covars=if (whenApplyFilter=="prenorm") covarsStat else covarsRaw,
                whichPlot="rnacomp",isNorm=TRUE,output="json")
        })
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="RnaComp",
            type="rnacomp",
            subtype="unorm",
            json=fromJSON(jsonUnorm)
        )
        listIndex <- listIndex + 1
        plots[[listIndex]] <- list(
            name="RnaComp",
            type="rnacomp",
            subtype="norm",
            json=fromJSON(jsonNorm)
        )
    }
    
    if ("volcano" %in% qcPlots) {
        disp("  Importing volcano")
        nn <- names(contrastList)
        json <- list()
        namc <- character(0)
        index <- 0
        for (n in nn) {
            fc <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            for (contr in colnames(fc)) {
                disp("    ",n," ",contr)
                index <- index + 1
                namc <- c(namc,contr)
                json[[index]] <- diagplotVolcano(fc[,contr],sumpList[[n]],contr,
                    altNames=geneDataExpr$gene_name,output="json")
            }
        }
        names(json) <- namc
        for (i in seq_along(json)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("volcano",namc[i],sep="_"),
                type="volcano",
                subtype="generic",
                json=fromJSON(json[[i]])
            )
        }
    }
    
    if ("mastat" %in% qcPlots) {
        disp("  Importing mastat")
        nn <- names(contrastList)
        json <- list()
        namc <- character(0)
        index <- 0
        for (n in nn) {
            m <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            a <- makeA(n,sampleList,normGenesExpr,1)
            for (contr in colnames(m)) {
                disp("    ",n," ",contr)
                index <- index + 1
                namc <- c(namc,contr)
                json[[index]] <- diagplotMa(m[,contr],a[,contr],sumpList[[n]],
                    contr,altNames=geneDataExpr$gene_name,output="json")
            }
        }
        names(json) <- namc
        for (i in seq_along(json)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("mastat",namc[i],sep="_"),
                type="mastat",
                subtype="generic",
                json=fromJSON(json[[i]])
            )
        }
    }
    
    if ("biodist" %in% qcPlots) {
        disp("  Importing biodist")
        nn <- names(contrastList)
        jsonList <- vector("list",length(nn))
        names(jsonList) <- nn
        for (n in nn) {
            disp("    ",n)
            cap <- capture.output({
                jsonList[[n]] <- diagplotNoiseq(normGenesExpr,sampleList,
                    covars=covarsStat,whichPlot="biodist",
                    biodistOpts=list(p=cpList[[n]],pcut=pcut,name=n),
                    output="json")
            })
        }
        for (i in seq_along(jsonList)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("biodist",nn[i],sep="_"),
                type="biodist",
                subtype="chromosome",
                json=fromJSON(jsonList[[i]]$chromosome)
            )
        }
        for (i in seq_along(jsonList)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("biodist",nn[i],sep="_"),
                type="biodist",
                subtype="biotype",
                json=fromJSON(jsonList[[i]]$biotype)
            )
        }
    }
    
    if ("statvenn" %in% qcPlots) {
        disp("  Importing statvenn")
        nn <- names(contrastList)
        jsonList <- vector("list",length(nn))
        names(jsonList) <- nn
        geneNames <- geneDataExpr$gene_name
        names(geneNames) <- names(geneDataExpr)
        for (n in nn) {
            disp("    ",n)
            jsonList[[n]] <- list(dereg=NULL,up=NULL,down=NULL)
            fc <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            # To narrow We take always the first condition versus the reference
            # to narrow down a bit the list of DE genes
            fc <- fc[,1,drop=FALSE]
            fcmat <- matrix(data=apply(fc,2,function(x) {
                rep(x,length(colnames(cpList[[n]])))
            }),ncol=ncol(fc)*ncol(cpList[[n]]))
            colnames(fcmat) <- colnames(cpList[[n]])
            jsonList[[n]]$dereg <- makeJVennStatData(cpList[[n]],fcmat=fcmat,
                pcut=pcut,direction="dereg",
                altNames=geneNames[rownames(cpList[[n]])])
            jsonList[[n]]$up <- makeJVennStatData(cpList[[n]],fcmat=fcmat,
                pcut=pcut,direction="up",
                altNames=geneNames[rownames(cpList[[n]])])
            jsonList[[n]]$down <- makeJVennStatData(cpList[[n]],fcmat=fcmat,
                pcut=pcut,direction="down",
                altNames=geneNames[rownames(cpList[[n]])])
        }
        for (i in seq_along(jsonList)) {
            for (d in names(jsonList[[i]])) {
                listIndex <- listIndex + 1
                plots[[listIndex]] <- list(
                    name=paste("statvenn",nn[i],d,sep="_"),
                    type="statvenn",
                    subtype=paste0("stat_",d),
                    json=jsonList[[i]][[d]]
                )
            }
        }
    }
    
    if ("foldvenn" %in% qcPlots) {
        disp("  Importing foldvenn")
        nn <- names(contrastList)
        
        # We need to construct a matrix with p-values for each contrast and
        # each statistical algorithm
        if (ncol(cpList[[1]]) > 1) {
            N <- names(cpList)
            prePmat <- lapply(N,function(n,C,S) {
                many <- C[[n]]
                one <- S[[n]]
                mat <- cbind(many,one)
                colnames(mat)[ncol(mat)] <- metaP
                return(mat)
            },cpList,sumpList)
            names(prePmat) <- N
        }
        else
            prePmat <- sumpList
        
        # Now we need to reformat prePmat to a list names according to each
        # algorithm
        algs <- colnames(prePmat[[1]])
        pmat <- lapply(algs,function(a,P) {
            algMat <- matrix(NA,nrow(P[[1]]),length(P))
            colnames(algMat) <- names(P)
            rownames(algMat) <- rownames(P[[1]])
            for (cnt in names(P))
                algMat[,cnt] <- prePmat[[cnt]][,a]
            return(algMat)
        },prePmat)
        names(pmat) <- algs
        
        # We need to construct a matrix with fold changes for each contrast
        fList <- vector("list",length(nn))
        names(fList) <- nn
        for (n in nn) {
            fc <- log2(makeFoldChange(n,sampleList,normGenesExpr,1))
            fc <- fc[,1,drop=FALSE]
            fList[[n]] <- fc
        }
        fcmat <- do.call("cbind",fList)
        fcmat <- fcmat[rownames(pmat[[1]]),]
        
        # Based on pmat and fcmat, we calculate Venns
        jsonList <- vector("list",length(pmat))
        names(jsonList) <- names(pmat)
        geneNames <- geneDataExpr$gene_name
        names(geneNames) <- names(geneDataExpr)
        for (n in names(jsonList)) {
            disp("    ",n)
            jsonList[[n]] <- list(dereg=NULL,up=NULL,down=NULL)
            for (d in c("dereg","up","down"))
                jsonList[[n]][[d]] <- makeJVennFoldData(pmat=pmat[[n]],
                    fcmat=fcmat,pcut=pcut,direction=d,
                    altNames=geneNames[rownames(pmat[[n]])])
        }
        for (n in names(jsonList)) {
            for (d in names(jsonList[[n]])) {
                listIndex <- listIndex + 1
                plots[[listIndex]] <- list(
                    name=paste("foldvenn",n,d,sep="_"),
                    type="foldvenn",
                    subtype=paste0("fold_",d),
                    json=jsonList[[n]][[d]]
                )
            }
        }
    }
    
    if ("deregulogram" %in% qcPlots) {
        disp("  Importing deregulogram")
        cntPairs <- combn(names(contrastList),2)
        json <- vector("list",ncol(cntPairs))
        namc <- character(ncol(cntPairs))
        counter <- 0
        for (i in seq_len(ncol(cntPairs))) {
            counter <- counter + 1
            disp("    ",cntPairs[1,i]," and ",cntPairs[2,i])
            namc[counter] <- paste0(cntPairs[1,i],"__",cntPairs[2,i])
            fmat <- cbind(
                log2(makeFoldChange(cntPairs[1,i],sampleList,
                    normGenesExpr,1))[,1,drop=FALSE],
                log2(makeFoldChange(cntPairs[2,i],sampleList,
                    normGenesExpr,1))[,1,drop=FALSE]
            )
            pmat <- do.call("cbind",sumpList[c(cntPairs[1,i],
                cntPairs[2,i])])
            colnames(pmat) <- colnames(fmat)
            json[[counter]] <- diagplotDeregulogram(fmat,pmat,fcut=0.3,
                pcut=pcut,output="json",altNames=geneDataExpr$gene_name)
        }
        for (i in seq_along(json)) {
            listIndex <- listIndex + 1
            plots[[listIndex]] <- list(
                name=paste("deregulogram",namc[i],sep="_"),
                type="deregulogram",
                subtype="generic",
                json=fromJSON(json[[i]])
            )
        }
    }
    
    disp("Writing plot database in ",file.path(PROJECT_PATH$data,
        "reportdb.js"))
    jsonPlots <- toJSON(plots,auto_unbox=TRUE,null="null",pretty=TRUE)
    code <- paste0("var plotData = ",jsonPlots)
    cat(code,file=file.path(PROJECT_PATH$data,"reportdb.js"),sep="\n")
}

.initReportDbTables <- function(con) {
    queries <- .reportDbTblDef()
    rs <- dbSendQuery(con,queries[[1]])
    if (dbHasCompleted(rs))
        dbClearResult(rs)
    for (n in names(queries)) {
        rs <- dbSendStatement(con,queries[[n]])
        if (dbHasCompleted(rs))
            dbClearResult(rs)
    }
}

.reportDbTblDef <- function() {
    return(list(
        enable_fkey="PRAGMA foreign_keys=1;",
        plot=paste(
            "CREATE TABLE IF NOT EXISTS plot (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "type TEXT,",
            "subtype TEXT,",
            "json TEXT",
            ");"
        ),
        clear="DELETE FROM plot;"
    ))
}

.dbImportPlot <- function(con,name,type,subtype=NULL,json) {
    if (is.null(subtype))
        query <- paste("INSERT INTO plot (name, type, subtype, json) ",
            "VALUES (","'",name,"', ","'",type,"', NULL, :j)",sep="")
    else
        query <- paste("INSERT INTO plot (name, type, subtype, json) ",
            "VALUES (","'",name,"', ","'",type,"', ","'",
            subtype,"', :j)",sep="")
    #print(query)
    nr <- dbExecute(con,query,params=list(j=json))
    return(nr)
}

.downloadJsLibs <- function(prPath,reportDb) {
    disp("Downloading required JavaScript libraries...")
    if (!file.exists(file.path(prPath$js,"pace.min.js")))
        #download.file(paste0("https://raw.github.com/HubSpot/pace/",
        #    "v1.0.0/pace.min.js"),
        #    file.path(prPath$js,"pace.min.js"))
        file.copy(from=system.file(package="metaseqR2","pace.min.js"),
            to=file.path(prPath$js,"pace.min.js"),overwrite=TRUE)
    if (!file.exists(file.path(prPath$js,"highcharts.js")))
        #download.file("https://code.highcharts.com/highcharts.js",
        #    file.path(prPath$js,"highcharts.js"))
        file.copy(from=system.file(package="metaseqR2","highcharts.js"),
            to=file.path(prPath$js,"highcharts.js"),overwrite=TRUE)    
    if (!file.exists(file.path(prPath$js,"highcharts-more.js")))
        #download.file("https://code.highcharts.com/highcharts-more.js",
        #    file.path(prPath$js,"highcharts-more.js"))
        file.copy(from=system.file(package="metaseqR2","highcharts-more.js"),
            to=file.path(prPath$js,"highcharts-more.js"),overwrite=TRUE)
    if (!file.exists(file.path(prPath$js,"exporting.js")))
        #download.file("https://code.highcharts.com/modules/exporting.js",
        #    file.path(prPath$js,"exporting.js"))
        file.copy(from=system.file(package="metaseqR2","exporting.js"),
            to=file.path(prPath$js,"exporting.js"),overwrite=TRUE)
    if (!file.exists(file.path(prPath$js,"offline-exporting.js")))
        #download.file(
        #    "https://code.highcharts.com/modules/offline-exporting.js",
        #    file.path(prPath$js,"offline-exporting.js"))
        file.copy(from=system.file(package="metaseqR2","offline-exporting.js"),
            to=file.path(prPath$js,"offline-exporting.js"),overwrite=TRUE)
    if (!file.exists(file.path(prPath$js,"export-data.js")))
        #download.file(
        #    "https://code.highcharts.com/modules/export-data.js",
        #    file.path(prPath$js,"export-data.js"))
        file.copy(from=system.file(package="metaseqR2","export-data.js"),
            to=file.path(prPath$js,"export-data.js"),overwrite=TRUE)
    if (!file.exists(file.path(prPath$js,"canvas2svg.js")))
        #download.file(
        #    "http://jvenn.toulouse.inrae.fr/app/js/canvas2svg.js",
        #    file.path(prPath$js,"canvas2svg.js"))
        file.copy(from=system.file(package="metaseqR2","canvas2svg.js"),
            to=file.path(prPath$js,"canvas2svg.js"),overwrite=TRUE)
    if (!file.exists(file.path(prPath$js,"jvenn.min.js")))
        #download.file(
        #    "http://jvenn.toulouse.inrae.fr/app/js/jvenn.min.js",
        #    file.path(prPath$js,"jvenn.min.js"))
        file.copy(from=system.file(package="metaseqR2","jvenn.min.js"),
            to=file.path(prPath$js,"jvenn.min.js"),overwrite=TRUE)
    if (reportDb == "sqlite") {
        if (!file.exists(file.path(prPath$js,"sql.js")))
            download.file(
        "https://cdnjs.cloudflare.com/ajax/libs/sql.js/0.5.0/js/sql.js",
        file.path(prPath$js,"sql.js"))
    }
    else if (reportDb == "dexie") {
        if (!file.exists(file.path(prPath$js,"dexie.min.js")))
            #download.file(
            #    "https://unpkg.com/dexie@2.0.4/dist/dexie.min.js",
            #    file.path(prPath$js,"dexie.min.js"))
            file.copy(from=system.file(package="metaseqR2","dexie.min.js"),
                to=file.path(prPath$js,"dexie.min.js"),overwrite=TRUE)
    }
}

.formatForReport <- function(x,o=NULL,r=NULL,v=NULL) {
    # Replace seqnames name
    colnames(x)[1] <- "chromosome"
    
    # Link to UCSC genome browser with position
    if (!is.null(o)) {
        x$chromosome <- 
            paste('<a href="http://genome.ucsc.edu/cgi-bin/hgTracks?db=',
                getUcscOrganism(o),'&position=',paste(x$chromosome,":",
                x$start,"-",x$end,sep=""),'" target="_blank">',x$chromosome,
                '</a>',sep="")
    }
    
    # Link to gene pages - only for Ensembl, RefSeq, not custom GTFs! UCSC does
    # not offer a related page.
    if (!is.null(r)) {
        if (r == "ensembl") {
            h <- getHost(o,v)
            if ("gene_id" %in% names(x)) # transLevel = "gene"
                x$gene_id <- paste('<a href="',h,'/Gene/Summary?g=',x$gene_id,
                    '" target="_blank">',x$gene_id,'</a>',sep="")
            # transLevel = "transcript"
            else if ("transcript_id" %in% names(x)) {
                # Stub
            }
            else if ("exon_id" %in% names(x)) { # transLevel = "exon"
                # Stub
            }
        }
        else if (r == "refseq") {
            if ("gene_id" %in% names(x)) # transLevel = "gene"
                x$gene_id <- paste('<a href="https://www.ncbi.nlm.nih.gov/',
                    'nuccore/',x$gene_id,'" target="_blank">',x$gene_id,'</a>',
                    sep="")
            else if ("transcript_id" %in% names(x)) { # transLevel="transcript"
                # Stub
            }
            else if ("exon_id" %in% names(x)) { # transLevel = "exon"
                # Stub
            }
        }
    }
    
    # Round rpgm and fold change
    tmpi1 <- grep("log2_normalized_fold_change_",colnames(x))
    tmpi2 <- grep("rpgm_normalized_mean_counts_",colnames(x))
    tmpi <- c(tmpi1,tmpi2)
    if (length(tmpi) > 0)
        x[,tmpi] <- round(x[,tmpi],digits=4)
    
    # Round p-values and FDR
    tmpi1 <- grep("p-value",colnames(x))
    tmpi2 <- grep("FDR",colnames(x))
    tmpi <- c(tmpi1,tmpi2)
    if (length(tmpi) > 0)
        x[,tmpi] <- round(x[,tmpi],digits=6)
    
    colnames(x) <- gsub("log2_normalized_fold_change_","",colnames(x))
    colnames(x) <- gsub("rpgm_normalized_mean_counts_","",colnames(x))
    tmpi <- grep("p-value",colnames(x))
    if (length(tmpi) > 0)
        colnames(x)[tmpi] <- "p-value"
    tmpi <- grep("FDR",colnames(x))
    if (length(tmpi) > 0)
        colnames(x)[tmpi] <- "FDR"
    
    return(x)
}

metaTest <- function(cpList,metaP=c("simes","bonferroni","fisher","harmonic",
    "dperm_min","dperm_max","dperm_weight","fperm","whitlock","minp","maxp",
    "weight","pandora","none"),counts,sampleList,statistics,statArgs,
    libsizeList,nperm=10000,weight=rep(1/length(statistics),
    length(statistics)),pOffset=NULL,rc=NULL) {
    checkTextArgs("metaP",metaP,c("simes","bonferroni","fisher","harmonic",
    "dperm_min","dperm_max","dperm_weight","fperm","whitlock","minp","maxp",
    "weight","pandora","none"))
    contrast <- names(cpList)
    disp("Performing meta-analysis with ",metaP)
    if (metaP=="pandora")
        metaP <- "weight"
    switch(metaP,
        fisher = {
            sumpList <- cmclapply(cpList,function(x) {
                tmp <- fisherMethod(x,p.corr="none",
                    zeroSub=.Machine$double.xmin)
                rp <- tmp$p.value
                names(rp) <- rownames(x)
                return(rp)
            },rc=rc)
        },
        fperm = {
            sumpList <- cmclapply(cpList,function(x) {
                if (!is.null(rc))
                    tmp <- fisherMethodPerm(x,p.corr="none",B=nperm,
                        mc.cores=getOption("cores"),zeroSub=1e-32)
                else
                    tmp <- fisherMethodPerm(x,p.corr="none",B=nperm,
                        zeroSub=.Machine$double.xmin)
                return(tmp$p.value)
            },rc=rc)
        },
        whitlock = {
            sumpList <- cmclapply(cpList,function(x) 
                return(apply(x,1,combine.test,method="z.transform")),rc=rc)
        },
        simes = {
            sumpList <- cmclapply(cpList,function(x) {
                return(apply(x,1,combineSimes,pOffset))
            },rc=rc)
        },
        bonferroni = {
            sumpList <- cmclapply(cpList,function(x) {
                return(apply(x,1,combineBonferroni,pOffset))
            },rc=rc)
        },
        minp = {
            sumpList <- cmclapply(cpList,function(x) {
                return(apply(x,1,combineMinp))
            },rc=rc)
        },
        maxp = {
            sumpList <- cmclapply(cpList,function(x) {
                return(apply(x,1,combineMaxp))
            },rc=rc)
        },
        harmonic = {
            sumpList <- cmclapply(cpList,function(x) {
                return(apply(x,1,combineHarmonic,weight,pOffset))
            },rc=rc)
        },
        weight = {
            sumpList <- cmclapply(cpList,function(x) {
                return(apply(x,1,combineWeight,weight,pOffset))
            },rc=rc)
        },
        dperm_min = {
            sumpList <- vector("list",length(cpList))
            names(sumpList) <- names(cpList)
            conl <- as.list(contrast)
            names(conl) <- contrast
            tempPList <- cmclapply(conl,metaPerm,
                counts=counts,sampleList=sampleList,
                statistics=statistics,statArgs=statArgs,
                libsizeList=libsizeList,
                nperm=nperm,weight=weight,
                select="min",rc=rc)
            originalPList <- cmclapply(cpList,function(x,m,w=NULL) {
                x[which(is.na(x))] <- 1
                switch(m,
                    min = {
                        return(apply(x,1,min))
                    },
                    max = {
                        return(apply(x,1,max))
                    },
                    weight = {
                        return(apply(x,1,function(p,w) return(prod(p^w)),
                            w))
                    }
                )
            },"min",rc=rc)
            for (cc in names(originalPList)) {
                pc <- cbind(tempPList[[cc]],originalPList[[cc]])
                ly <- ncol(pc)
                sumpList[[cc]] <- apply(pc,1,function(y,m) 
                    return(length(which(y[seq_len(m-1)]<y[m]))/(m-1)),ly)
            }
        },
        dperm_max = {
            sumpList <- vector("list",length(cpList))
            names(sumpList) <- names(cpList)
            conl <- as.list(contrast)
            names(conl) <- contrast
            tempPList <- cmclapply(conl,metaPerm,
                counts=counts,sampleList=sampleList,
                statistics=statistics,statArgs=statArgs,
                libsizeList=libsizeList,
                nperm=nperm,weight=weight,
                select="max",rc=rc)
            originalPList <- cmclapply(cpList,function(x,m,w=NULL) {
                switch(m,
                    min = {
                        return(apply(x,1,min))
                    },
                    max = {
                        return(apply(x,1,max))
                    },
                    weight = {
                        return(apply(x,1,function(p,w) return(prod(p^w)),
                            w))
                    }
                )
            },"max",rc=rc)
            for (cc in names(originalPList)) {
                pc <- cbind(tempPList[[cc]],originalPList[[cc]])
                ly <- ncol(pc)
                sumpList[[cc]] <- apply(pc,1,function(y,m) 
                    return(length(which(y[seq_len(m-1)]<y[m]))/(m-1)),ly)
            }
            #assign("perm.list",tempPList,envir=.GlobalEnv)
            #assign("oList",originalPList,envir=.GlobalEnv)
        },
        dperm_weight = {
            sumpList <- vector("list",length(cpList))
            names(sumpList) <- names(cpList)
            conl <- as.list(contrast)
            names(conl) <- contrast
            tempPList <- cmclapply(conl,metaPerm,
                counts=counts,sampleList=sampleList,
                statistics=statistics,statArgs=statArgs,
                libsizeList=libsizeList,
                nperm=nperm,weight=weight,
                select="weight",rc=rc)
            originalPList <- cmclapply(cpList,function(x,m,w=NULL) {
                switch(m,
                    min = {
                        return(apply(x,1,min))
                    },
                    max = {
                        return(apply(x,1,max))
                    },
                    weight = {
                        return(apply(x,1,function(p,w) {return(prod(p^w))},
                            w))
                    }
                )
            },"weight",weight,rc=rc)
            for (cc in names(originalPList)) {
                pc <- cbind(tempPList[[cc]],originalPList[[cc]])
                ly <- ncol(pc)
                sumpList[[cc]] <- apply(pc,1,function(y,m) 
                    return(length(which(y[seq_len(m-1)]<y[m]))/(m-1)),ly)
            }
        },
        none = {
            # A default value must be there to use with volcanos, we say the one
            # of the first statistic in order of input
            sumpList <- cmclapply(cpList,function(x) return(x[,1]),rc=rc)
        }
    )
    return(sumpList)
}

metaPerm <- function(contrast,counts,sampleList,statistics,statArgs,
    libsizeList,nperm=10000,weight=rep(1/ncol(counts),ncol(counts)),
    select=c("min","max","weight"),replace="auto",rc=NULL) {
    checkTextArgs("select",select,c("min","max","weight"))
    if (replace=="auto") {
        if (ncol(counts)<=6)
            replace=FALSE
        else
            replace=TRUE
    }
    # We will construct relist in a way so that we can assign seeds for random
    # number generation and track progress at the same time
    relist <- vector("list",nperm)
    relist <- cmclapply(seq_along(relist),function(i) {
        return(list(seed=round(1e+6*runif(1)),prog=i))
    },rc=rc)
    disp("  Resampling procedure started...")
    # In this case, we must not use cmclapply as we want to be able to track 
    # progress through mc.preschedule...
    if (!is.null(rc))
        pp <- mclapply(relist,metaWorker,counts,sampleList,contrast,
            statistics,replace,statArgs,libsizeList,select,weight,
            mc.preschedule=FALSE,mc.cores=getOption("cores"))
    else
        pp <- lapply(relist,metaWorker,counts,sampleList,contrast,statistics,
            replace,statArgs,libsizeList,select,weight)
    disp("  Resampling procedure ended...")
    return(do.call("cbind",pp))
}

metaWorker <- function(x,co,sl,cnt,s,r,sa,ll,el,w) {
    disp("    running permutation #",x$prog)
    pl <- makePermutation(co,sl,cnt,r)
    ppmat <- matrix(NA,nrow(co),length(s))
    colnames(ppmat) <- s
    for (alg in s) {
        #disp("      running permutation tests with: ",alg)
        tcl <- makeContrastList(pl$contrast,pl$sampleList)
        switch(alg,
            deseq = {
                pList <- suppressMessages(statDeseq(pl$counts,pl$sampleList,
                    tcl,sa[[alg]]))
            },
            deseq2 = {
                pList <- suppressMessages(statDeseq2(pl$counts,pl$sampleList,
                    tcl,sa[[alg]]))
            },
            edger = {
                pList <- suppressMessages(statEdger(pl$counts,pl$sampleList,
                    tcl,sa[[alg]]))
            },
            noiseq = {
                pList <- suppressMessages(statNoiseq(pl$counts,pl$sampleList,
                    tcl,sa[[alg]]))
            },
            #bayseq = {
            #    pList <- suppressMessages(statBayseq(pl$counts,pl$sampleList,
            #        tcl,sa[[alg]],ll))
            #},
            limma = {
                pList <- suppressMessages(statLimma(pl$counts,pl$sampleList,
                    tcl,sa[[alg]]))
            },
            nbpseq = {
                pList <- suppressMessages(statNbpseq(pl$counts,pl$sampleList,
                    tcl,sa[[alg]],ll))
            },
            absseq = {
                pList <- suppressMessages(statAbsseq(pl$counts,pl$sampleList,
                    tcl,sa[[alg]]))
            },
            dss = {
                pList <- suppressMessages(statDss(pl$counts,pl$sampleList,
                    tcl,sa[[alg]]))
            }
        )
        ppmat[,alg] <- as.numeric(pList[[1]])
    }
    ppmat[which(is.na(ppmat))] <- 1
    switch(el,
        min = {
            pIter <- apply(ppmat,1,min)
        },
        max = {
            pIter <- apply(ppmat,1,max)
        },
        weight = {
            pIter <- apply(ppmat,1,function(p,w) return(prod(p^w)),w)
        }
    )
    return(pIter)
}

combineSimes <- function(p,zerofix=NULL) {
    p <- .zeroFix(p,zerofix)
    m <- length(p)
    y <- sort(p)
    s <- min(m*(y/(seq_len(m))))
    return(min(c(s,1)))
}

combineBonferroni <- function(p,zerofix=NULL) {
    p <- .zeroFix(p,zerofix)
    b <- length(p)*min(p)
    return(min(c(1,b)))
}

combineWeight <- function(p,w,zerofix=NULL) {
    p <- .zeroFix(p,zerofix)
    return(prod(p^w))
}

combineHarmonic <- function(p,w,zerofix=NULL) {
    if (!requireNamespace("harmonicmeanp"))
        stop("R package harmonicmeanp is required!")
    p <- .zeroFix(p,zerofix)
    return(p.hmp(p,w,L=length(p),multilevel=FALSE))
}

combineMinp <- function(p) { return(min(p)) }

combineMaxp <- function(p) { return(max(p)) }

# Copied from ex-CRAN package MADAM and exported. The man pages are copied from
# the original package.
fisherMethod <- function(pvals,method=c("fisher"),p.corr=c("bonferroni","BH",
    "none"),zeroSub=0.00001,na.rm=FALSE,mc.cores=NULL) {
    stopifnot(method %in% c("fisher"))
    stopifnot(p.corr %in% c("none","bonferroni","BH"))
    stopifnot(all(pvals>=0, na.rm=TRUE) & all(pvals<=1, na.rm=TRUE))
    stopifnot(zeroSub>=0 & zeroSub<=1 || length(zeroSub)!=1)
    if(is.null(dim(pvals)))
        stop("pvals must have a dim attribute")
    p.corr <- ifelse(length(p.corr)!=1, "BH", p.corr)
    ##substitute p-values of 0
    pvals[pvals == 0] <- zeroSub
    if(is.null(mc.cores)) {
        fisher.sums <- data.frame(do.call(rbind,apply(pvals,1,fisherSum,
            zeroSub=zeroSub,na.rm=na.rm)))
    } 
    else {
        fisher.sums <- parallel::mclapply(seq_len(nrow(pvals)), function(i) {
            fisherSum(pvals[i,],zeroSub=zeroSub,na.rm=na.rm)
        }, mc.cores=mc.cores)
        fisher.sums <- data.frame(do.call(rbind,fisher.sums))
    }
    
    rownames(fisher.sums) <- rownames(pvals)
    fisher.sums$p.value <- 1-pchisq(fisher.sums$S,df=2*fisher.sums$num.p)
    fisher.sums$p.adj <- switch(p.corr,
        bonferroni = p.adjust(fisher.sums$p.value,"bonferroni"),
        BH = p.adjust(fisher.sums$p.value,"BH"),
        none = fisher.sums$p.value
    )
    return(fisher.sums)
}

# Copied from ex-CRAN package MADAM and exported. The man pages are copied from
# the original package.
fisherMethodPerm <- function(pvals,p.corr=c("bonferroni","BH","none"),
    zeroSub=0.00001,B=10000,mc.cores=NULL,blinker=1000) {
    stopifnot(is.na(blinker) || blinker>0)
    stopifnot(p.corr %in% c("none","bonferroni","BH"))
    stopifnot(all(pvals>=0,na.rm=TRUE) & all(pvals<=1,na.rm=TRUE))
    stopifnot(zeroSub>=0 & zeroSub<=1 || length(zeroSub)!=1)
    if(is.null(dim(pvals)))
        stop("pvals must have a dim attribute")
    p.corr <- ifelse(length(p.corr)!=1,"BH",p.corr)
    pvals[pvals==0] <- zeroSub

    resPerm <- lapply(seq_len(nrow(pvals)),function(i) {
        if(!is.na(blinker) & i%%blinker==0)
        message("=", appendLF=FALSE)
        ##which studies contribute to S (don't have a NA in row i)
        good.p <- which(!is.na(pvals[i,]))
        S.obs= fisherSum(pvals[i,good.p], na.rm=FALSE)
        if(is.null(mc.cores)) {
            Srand <- unlist(lapply(seq_len(B), function(b) {
            ##get non NA p-values from studies contributing to S
            myp <- vapply(good.p, function(pc){
                sample(na.exclude(pvals[,pc]),1)
            },numeric(1))
            fisherSum(myp)$S
        }))
        } else {
        Srand <- unlist(parallel::mclapply(seq_len(B), function(b) {
            ##get non NA p-values from studies contributing to S
            myp <- vapply(good.p, function(pc) {
                sample(na.exclude(pvals[,pc]),1)
            },numeric(1))
            fisherSum(myp)$S
            }, mc.cores=mc.cores))
        }
        p.value <- sum(Srand>=S.obs$S)/B
        data.frame(S=S.obs$S, num.p=S.obs$num.p, p.value=p.value)
    })
    resPerm <- data.frame(do.call(rbind, resPerm))

    if(!is.na(blinker) && blinker>0)
        message()
    ## rownames(resPerm) <- rownames(pvals)
    resPerm$p.adj <- switch(p.corr,
        bonferroni = p.adjust(resPerm$p.value,"bonferroni"),
        BH = p.adjust(resPerm$p.value,"BH"),
        none = resPerm$p.value)
    return(resPerm)
}

# Copied from ex-CRAN package MADAM and exported. The man pages are copied from
# the original package.
fisherSum <- function(p,zeroSub=0.00001,na.rm=FALSE) {
    if(any(p>1, na.rm=TRUE)||any(p<0, na.rm=TRUE))
        stop("You provided bad p-values")
    stopifnot(zeroSub>=0 & zeroSub<=1 || length(zeroSub)!=1)
    p[p==0] <- zeroSub
    if (na.rm)
        p <- p[!is.na(p)]
    S = -2*sum(log(p))
    res <- data.frame(S=S,num.p=length(p))
    return(res)
}

.zeroFix <- function(p,z=NULL) {
    if (!is.null(z) && !is.numeric(z))
        stop("zerofix must be NULL or a numeric greater than 0 and less than ",
            "1!")
    if (!is.null(z) && (z <= 0 || z >= 1))
        stop("When zerofix is not NULL it must be a numeric greater than 0 ",
            "and less than 1!")
    #ze <- which(p==0)
    ze <- which(p < 1e-300) # A very small value as 0 causes also problems...
    if (length(ze)>0) {
        if (!is.null(z))
            p[ze] <- z*min(p[-ze])
        else
            p[ze] <- 0.5*runif(length(ze))*min(p[-ze])
    }
    return(p)
}

metaseqr2 <- function(
    counts,
    sampleList,
    excludeList=NULL,
    fileType=c("auto","sam","bam","bed"),
    path=NULL,
    contrast=NULL,
    libsizeList=NULL,
    embedCols=list(
        idCol=4,
        gcCol=NA,
        nameCol=NA,
        btCol=NA
    ),
    annotation=NULL,
    org=c("hg18","hg19","hg38","mm9","mm10","mm39","rn5","rn6","dm3","dm6",
        "danrer7","pantro4","susscr3","tair10","equcab3"),
    refdb=c("ensembl","ucsc","refseq"),
    version="auto",
    transLevel=c("gene","transcript","exon"),
    countType=c("gene","exon","utr"),
    utrOpts=list(
        frac=1,
        minLength=300,
        downstream=50
    ),
    exonFilters=list(
        minActiveExons=list(
            exonsPerGene=5,
            minExons=2,
            frac=1/5
        )
    ),
    geneFilters=list(
        length=list(
            length=500
        ),
        avgReads=list(
            averagePerBp=100,
            quantile=0.25
        ),
        expression=list(
            median=TRUE,
            mean=FALSE,
            quantile=NA,
            known=NA,
            custom=NA
        ),
        biotype=getDefaults("biotypeFilter",org[1]),
        presence=list(
            frac=0.25,
            minCount=10,
            perCondition=FALSE
        )
    ),
    whenApplyFilter=c("postnorm","prenorm"),
    normalization=c("deseq","deseq2","edaseq","edger","noiseq","nbpseq",
        "absseq","dss","each","none"),
    normArgs=NULL,
    #statistics=c("deseq","deseq2","edger","noiseq","bayseq","limma","nbpseq",
    #    "absseq","dss"),
    statistics=c("deseq","deseq2","edger","noiseq","limma","nbpseq","absseq",
        "dss"),
    statArgs=NULL,
    adjustMethod=sort(c(p.adjust.methods,"qvalue")),
    metaP=if (length(statistics)>1) c("simes","bonferroni","fisher",
        "dperm_min","dperm_max","dperm_weight","fperm","whitlock","minp","maxp",
        "weight","pandora","none") else "none",
    weight=rep(1/length(statistics),length(statistics)),
    nperm=10000,
    pcut=NA,
    logOffset=1,
    pOffset=NULL,
    preset=NULL, # An analysis strictness preset
    qcPlots=c(
        "mds","biodetection","countsbio","saturation","readnoise","filtered",
        "correl","pairwise","boxplot","gcbias","lengthbias","meandiff",
        "meanvar","rnacomp","deheatmap","volcano","biodist","mastat","statvenn",
        "foldvenn","deregulogram"
    ),
    figFormat=c("png","jpg","tiff","bmp","pdf","ps"),
    outList=FALSE,
    exportWhere=NA, # An output directory for the project
    exportWhat=c("annotation","p_value","adj_p_value","meta_p_value",
        "adj_meta_p_value","fold_change","stats","counts","flags"),
    exportScale=c("natural","log2","log10","vst","rpgm"),
    exportValues=c("raw","normalized"),
    exportStats=c("mean","median","sd","mad","cv","rcv"),
    exportCountsTable=FALSE,
    restrictCores=0.6,
    report=TRUE,
    reportTop=0.1,
    reportTemplate="default",
    saveGeneModel=TRUE,
    verbose=TRUE,
    runLog=TRUE,
    reportDb=c("dexie","sqlite"),
    localDb=file.path(system.file(package="metaseqR2"),"annotation.sqlite"),
    offlineReport=TRUE,
    createTracks=FALSE,
    overwriteTracks=FALSE,
    trackExportPath=file.path(exportWhere,"tracks"),
    trackInfo=list(
        stranded=FALSE,
        normTo=1e+9,
        urlBase="http://www.trackserver.me",
        fasta=NULL,
        gtf=NULL,
        hubInfo=list(
            name="MyHub",
            shortLabel="My hub",
            longLabel="My hub long",
            email="someone@example.com"
        )
    ),
    .progressFun=NULL,
    .exportR2C=FALSE,
    ...
) {
    # Save function call for report
    FUN_CALL <- deparse(sys.call())
    
    # Check for older argument names and adjust
    args <- as.list(match.call())
    backCheckedArgs <- .backwardsConvertArgs(args)
    if (backCheckedArgs$mixDetected)
        stop("You must provide either old metaseqR arguments (e.g sample.list)",
            " or new metaseqR2 arguments (e.g. sampleList)! Not both!")
    if (backCheckedArgs$backDetected) {
        argMap <- .backwardsMapOld2New()
        # Check what happens with the new argument embedCols
        if (!is.null(backCheckedArgs$args$embedCols)) {
            embedCols <- backCheckedArgs$args$embedCols
            backCheckedArgs$args$embedCols <- NULL
        }
        for (n in names(backCheckedArgs$args))
            assign(argMap[[n]],eval(backCheckedArgs$args[[n]]))
        # There are some nested arguments that also need correction
        if (!is.null(backCheckedArgs$args$gene.filters)) {
            if (!is.null(geneFilters$avg.reads)) {
                geneFilters$avgReads <- geneFilters$avg.reads
                if (!is.null(geneFilters$avg.reads$average.per.bp)) {
                    geneFilters$avgReads$averagePerBp <-
                        geneFilters$avg.reads$average.per.bp
                    geneFilters$avg.reads$average.per.bp <- NULL
                }
                geneFilters$avg.reads <- NULL
            }
            if (!is.null(geneFilters$presence)) {
                if (!is.null(geneFilters$presence$min.count)) {
                    geneFilters$presence$minCount <-
                        geneFilters$presence$min.count
                    geneFilters$presence$min.count <- NULL
                }
                if (!is.null(geneFilters$presence$per.condition)) {
                    geneFilters$presence$perCondition <-
                        geneFilters$presence$per.condition
                    geneFilters$presence$per.condition <- NULL
                }
            }
        }
        if (!is.null(backCheckedArgs$args$exon.filters)) {
            if (!is.null(exonFilters$min.active.exons)) {
                exonFilters$minActiveExons <- exonFilters$min.active.exons
                if (!is.null(exonFilters$min.active.exons$exons.per.gene)) {
                    exonFilters$minActiveExons$exonsPerGene <-
                        exonFilters$min.active.exons$exons.per.gene
                    exonFilters$min.active.exons$exons.per.gene <- NULL
                }
                if (!is.null(exonFilters$min.active.exons$min.exons)) {
                    exonFilters$minActiveExons$minExons <-
                        exonFilters$min.active.exons$min.exons
                    exonFilters$min.active.exons$min.exons <- NULL
                }
                exonFilters$min.active.exons <- NULL
            }
        }
        if ("venn" %in% qcPlots)
            qcPlots[which(qcPlots == "venn")] <- "statvenn"
        if (annotation == "download")
            annotation <- NULL
    }
    else
        # Check if there are any mispelled or invalid parameters and throw a
        # warning
        checkMainArgs(args)    
    
    # Check essential arguments
    fromRaw <- fromPrevious <- FALSE
    if (missing(counts) && (missing(sampleList) || is.list(sampleList)))
        stop("You must provide a file with genomic region (gene, exon, etc.) ",
            "counts\nor an input targets file to create input from! If the ",
            "counts file is\nmissing, sampleList cannot be missing or it must ",
            "be a targets file\nwith at least three columns! See the ",
            "readTargets function.\ncounts may also be a gene model list ",
            "(see the documentation)")
    sampleListPrev <- NULL
    if (!missing(counts) && !missing(sampleList) && is.character(counts) 
        && file.exists(counts) && length(grep(".RData$",counts))>0) {
        warning("When restoring a previous analysis, if the sampleList ",
            "argument is provided, it precedes the stored one. Also, the ",
            "excludeList argument is ingored...",immediate.=TRUE)
        fromPrevious <- TRUE
        disp("Restoring previous analysis from ",basename(counts))
        tmpEnv <- .backwardsCompatibility(counts)
        sampleListPrev <- tmpEnv$sampleList
        excludeList <- NULL
        if (is.character(sampleList) && file.exists(sampleList)) {
            theList <- readTargets(sampleList,path=path)
            sampleList <- theList$samples
        }
        countType <- tmpEnv$countType
    }
    if (!missing(counts) && missing(sampleList) && is.character(counts) 
        && file.exists(counts) && length(grep(".RData$",counts))>0) {
        # Time to load previous analysis if existing
        fromPrevious <- TRUE
        message("Restoring previous analysis from ",basename(counts))
        tmpEnv <- .backwardsCompatibility(counts)
        sampleList <- tmpEnv$sampleList
        countType <- tmpEnv$countType
    }
    if (missing(sampleList) && !fromPrevious || (!is.list(sampleList) &&
        !file.exists(sampleList)))
        stop("You must provide a list with condition names and sample names ",
            "(same as in the counts file) or an input file to create the ",
            "sample list from!")
    if (!missing(sampleList) && !is.list(sampleList) 
        && file.exists(sampleList) && !missing(counts) && !fromPrevious)
        sampleList <- makeSampleList(sampleList)
    if (!missing(sampleList) && !is.list(sampleList) 
        && file.exists(sampleList) && missing(counts)) {
        counts <- NULL
        theList <- readTargets(sampleList,path=path)
        sampleList <- theList$samples
        fileList <- theList$files
        if (tolower(fileType[1])=="auto")
            fileType <- theList$type
        if (is.null(fileType))
            stop(paste("The type of the input files could not be recognized!",
                "Please specify (BAM or BED)..."))
        fromRaw <- TRUE
    }
    else
        theList <- NULL
    
    # If report requested with SQLite RSQLite must be present,also pander
    if (report) {
        if (!requireNamespace("pander"))
            stop("R package pander is required to build metaseqR2 reports!")
        if ("sqlite" %in% reportDb && !requireNamespace("RSQLite"))
            stop("R package RSQLite is required to build metaseqR reports!")
    }
    
    # Initialize environmental variables
    HOME <- system.file(package="metaseqR2")
    TEMPLATE <- HOME
    # Globalize the project's verbosity and logger
    if (fromRaw)
        PROJECT_PATH <- makeProjectPath(exportWhere)
    else
        PROJECT_PATH <- makeProjectPath(exportWhere,counts)
    assign("VERBOSE",verbose,envir=metaEnv)
    if (runLog) {
        if (file.exists(file.path(PROJECT_PATH$logs,"metaseqr_run.log")))
            unlink(file.path(PROJECT_PATH$logs,"metaseqr_run.log"))
        logger <- create.logger(logfile=file.path(PROJECT_PATH$logs,
            "metaseqr_run.log"),level=2,logformat="%d %c %m")
    }
    else
        logger <- NULL
    assign("LOGGER",logger,envir=metaEnv)
    
    # Check if sample names match in file/df and list, otherwise meaningless to 
    # proceed
    if (!fromRaw && !fromPrevious) {
        if (!is.data.frame(counts) && !is.list(counts)) {
            if (file.exists(counts)) {
                aline <- read.delim(counts,nrows=5) # Read the 1st lines
                aline <- colnames(aline)
            }
            else
                stopwrap("The counts file you provided does not exist!")
        }
        else if (is.data.frame(counts))
            aline <- colnames(counts)
        else if (is.list(counts))
            aline <- names(counts)
        samples <- unlist(sampleList,use.names=FALSE)
        if (length(which(!is.na(match(samples,aline)))) != length(samples))
            stopwrap("The sample names provided in the counts file/list do ",
                "not match with those of the sampleList!")
    }
    
    # If exclude list given, check that it's a subset of sampleList, otherwise
    # just ignore excludeList
    if (!is.null(excludeList) && !is.na(excludeList)) {
        sl <- unlist(sampleList)
        el <- unlist(excludeList)
        if (length(intersect(sl,el)) != length(el)) {
            warnwrap("Some samples in excludeList do not match those in the ",
                "initial sampleList! Ignoring...",now=TRUE)
            excludeList <- NULL
        }
    }

    fileType <- tolower(fileType[1])
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    version <- version[1]
    transLevel <- tolower(transLevel[1])
    countType <- tolower(countType[1])
    whenApplyFilter <- tolower(whenApplyFilter[1])
    normalization <- tolower(normalization[1])
    adjustMethod <- adjustMethod[1]
    metaP <- tolower(metaP[1])
    statistics <- tolower(statistics)
    figFormat <- tolower(figFormat)
    if (!is.null(qcPlots)) 
        qcPlots <- tolower(qcPlots)
    exportWhat <- tolower(exportWhat)
    exportScale <- tolower(exportScale)
    exportValues <- tolower(exportValues)
    exportStats <- tolower(exportStats)
    reportDb <- tolower(reportDb[1])
    if (!is.null(preset)) 
        preset <- tolower(preset[1])
    
    if (fromRaw)
        countsName <- "imported sam/bam/bed files"
    else {
        if (!is.data.frame(counts) && !is.null(counts) && !is.list(counts)) {
            checkFileArgs("counts",counts)
            if (fromPrevious)
                countsName <- "previously stored project"
            else
                countsName <- basename(counts)
        }
        else if (is.list(counts) && !is.data.frame(counts)) {
            countsName <- "previously stored gene model"
        }
        else {
            countsName <- "imported custom data frame"
        }
    }
    
    # annotation cannot be "embedded" when countType is other than "gene" or
    # transLevel is other than "gene"
    if (!is.null(annotation) && !is.list(annotation)
        && is.character(annotation) && annotation == "embedded" 
        && countType != "gene")
        stopwrap("The annotation argument cannot be \"embedded\" when the ",
            "countType argument is other than \"gene\"!")

    # annotation must be a list to be fed to buildCustomAnnotation
    if (is.list(annotation)) {
        # members are checked by buildCustomAnnotation if required
        # We only need to check that the gtfFile exists here
        if (!("gtf" %in% names(annotation)))
            stopwrap("A gtf field must be provided with an existing GTF file ",
                "when providing a list with custom annotation elements!")
        if ("gtf" %in% names(annotation) && is.character(annotation$gtf)
            && !file.exists(annotation$gtf))
            stopwrap("An existing GTF file must be provided when providing a ",
                "list with custom annotation elements!")
    }
    
    if (is.list(counts) && !is.data.frame(counts)
        && (countType=="exon" || countType=="utr")
        && annotation=="embedded") {
        warnwrap("annotation cannot be \"embedded\" when importing a stored ",
            "gene model! Setting to NULL...")
        #annotation <- "download"
        annotation <- NULL
    }

    if (metaP %in% c("weight","pandora","dperm_weight") && 
        abs(1-sum(weight))>1e-5)
        stopwrap("The weights given for p-value combination should sum to 1!")

    checkTextArgs("fileType",fileType,c("auto","sam","bam","bed"),
        multiarg=FALSE)
    if (!is.null(annotation) && !is.list(annotation) 
        && !file.exists(annotation))
        checkTextArgs("annotation",annotation,"embedded",multiarg=FALSE)
    if (is.character(localDb) && file.exists(localDb)) {
        if (!.userOrg(org,localDb) && is.null(annotation))
            checkTextArgs("org",org,c("hg18","hg19","hg38","mm9","mm10","mm39",
                "rn5","rn6","dm3","dm6","danrer7","pantro4","susscr3","tair10",
                "equcab3"),multiarg=FALSE)
    }
    else if (is.null(annotation) 
        && !(is.character(localDb) || file.exists(localDb))) {
        # So only some annotations can be fetched on-the-fly
        checkTextArgs("org",org,c("hg18","hg19","hg38","mm9","mm10","mm39",
            "rn5","rn6","dm3","dm6","danrer7","pantro4","susscr3","tair10",
            "equcab3"),
            multiarg=FALSE)
    }
    if (is.character(localDb) && file.exists(localDb)) {
        if (!.userRefdb(refdb,localDb) && is.null(annotation))
            checkTextArgs("refdb",refdb,c("ensembl","ucsc","refseq"),
                multiarg=FALSE)
    }
    else if (is.null(annotation) 
        && !(is.character(localDb) || file.exists(localDb))) {
        checkTextArgs("refdb",refdb,c("ensembl","ucsc","refseq"),multiarg=FALSE)
    }
    checkTextArgs("transLevel",transLevel,c("gene","transcript","exon"),
        multiarg=FALSE)
    checkTextArgs("countType",countType,c("gene","exon","utr"),
        multiarg=FALSE)
    checkTextArgs("whenApplyFilter",whenApplyFilter,c("postnorm",
        "prenorm"),multiarg=FALSE)
    checkTextArgs("normalization",normalization,c("edaseq","deseq","deseq2",
        "edger","noiseq","nbpseq","absseq","dss","each","none"),multiarg=FALSE)
    if (!any(is.na(statistics)))
        #checkTextArgs("statistics",statistics,c("deseq","deseq2","edger",
        #    "noiseq","bayseq","limma","nbpseq","absseq","dss"),multiarg=TRUE)
        checkTextArgs("statistics",statistics,c("deseq","deseq2","edger",
            "noiseq","limma","nbpseq","absseq","dss"),multiarg=TRUE)
    checkTextArgs("metaP",metaP,c("simes","bonferroni","fisher","dperm_min",
        "dperm_max","dperm_weight","fperm","whitlock","minp","maxp","weight",
        "pandora","none"),multiarg=FALSE)
    checkTextArgs("figFormat",figFormat,c("png","jpg","tiff","bmp","pdf",
        "ps"),multiarg=TRUE)
    checkTextArgs("exportWhat",exportWhat,c("annotation","p_value",
        "adj_p_value","meta_p_value","adj_meta_p_value","fold_change","stats",
        "counts","flags"),multiarg=TRUE)
    checkTextArgs("exportScale",exportScale,c("natural","log2","log10",
        "rpgm","vst"),multiarg=TRUE)
    checkTextArgs("exportValues",exportValues,c("raw","normalized"),
        multiarg=TRUE)
    checkTextArgs("exportStats",exportStats,c("mean","median","sd","mad",
        "cv","rcv"),multiarg=TRUE)
    checkTextArgs("reportDb",reportDb,c("sqlite","dexie"))
    if (!is.null(preset))
        checkTextArgs("preset",preset,c("all_basic","all_normal","all_full",
            "medium_basic","medium_normal","medium_full","strict_basic",
            "strict_normal","strict_full"),multiarg=FALSE)
    if (!is.null(qcPlots))
        checkTextArgs("qcPlots",qcPlots,c("mds","biodetection","countsbio",
            "saturation","readnoise","correl","pairwise","boxplot","gcbias",
            "lengthbias","meandiff","meanvar","rnacomp","deheatmap","volcano",
            "biodist","filtered","mastat","statvenn","foldvenn","deregulogram"),
            multiarg=TRUE)
    if (!is.na(restrictCores)) checkNumArgs("restrictCores",restrictCores,
        "numeric",c(0,1),"botheq")
    if (!is.na(pcut)) 
        checkNumArgs("pcut",pcut,"numeric",c(0,1),"botheq")
    if (!is.null(embedCols$gcCol) && !is.na(embedCols$gcCol)) 
        checkNumArgs("embedCols$gcCol",embedCols$gcCol,"numeric",0,"gt")
    if (!is.null(embedCols$nameCol) && !is.na(embedCols$nameCol)) 
        checkNumArgs("embedCols$nameCol",embedCols$nameCol,"numeric",0,"gt")
    if (!is.null(embedCols$btCol) && !is.na(embedCols$btCol)) 
        checkNumArgs("embedCols$btCol",embedCols$btCol,"numeric",0,"gt")
    if (!is.na(logOffset)) 
        checkNumArgs("logOffset",logOffset,"numeric",0,"gt")
    if (!is.null(pOffset)) 
        checkNumArgs("pOffset",pOffset,"numeric",c(0,1),"botheq")
    checkNumArgs("nperm",nperm,"numeric",10,"gt")
    if (!is.null(reportTop))
        checkNumArgs("reportTop",reportTop,"numeric",c(0,1),"both")
    if (!is.null(contrast)) {
        checkContrastFormat(contrast,sampleList)
        contrast <- unique(contrast)
    }
    #if ("bayseq" %in% statistics) 
    #    libsizeList <- checkLibsize(libsizeList,sampleList)
    # Check the genomic database version argument
    if (is.character(version)) {
        version <- tolower(version)
        checkTextArgs("version",version,c("auto"),multiarg=FALSE)
    }
    else
        checkNumArgs("version",version,"numeric")
    
    # Check utrOpts if countType is "utr"
    if (countType == "utr") {
        utrOptsDef <- getDefaults("utrOpts")
        if (!is.null(utrOpts$frac)) {
            checkNumArgs("utrOpts$frac",utrOpts$frac,"numeric",c(0,1),"both")
            utrOptsDef$frac <- utrOpts$frac
        }
        if (!is.null(utrOpts$minLength)) {
            checkNumArgs("utrOpts$minLength",utrOpts$minLength,"numeric",0,
                "gt")
            utrOptsDef$minLength <- utrOpts$minLength
        }
        if (!is.null(utrOpts$downstream)) {
            checkNumArgs("utrOpts$downstream",utrOpts$downstream,"numeric",0,
                "gte")
            utrOptsDef$downstream <- utrOpts$downstream
        }
        utrOpts <- utrOptsDef
    }
    
    # Check main functionality packages
    checkPackages(metaP,qcPlots)
    # Check the case of embedded annotation, not given gc and gene name columns
    # Checks about countType have been performed before
    if (!is.null(annotation) && annotation == "embedded") {
        if (is.na(embedCols$gcCol) && countType=="gene" 
            && normalization=="edaseq")
            stopwrap("The column that contains the gene GC content ",
                "(\"embedCols$gcCol\") argument is required when ",
                "\"annotation\" is \"embedded\"!")
                
        if (is.na(embedCols$nameCol) && !is.na(geneFilters$expression$known)) {
            warnwrap("The column that contains the HUGO gene symbols ",
                "(\"embedCols$nameCol\") is missing with embedded annotation! ",
                "Gene name expression filter will not be available...")
            geneFilters$expression$known=NA
            if ("volcano" %in% qcPlots)
                warnwrap("The column that contains the HUGO gene symbols ",
                    "(\"embedCols$nameCol\") is missing with embedded ",
                    "annotation! Interactive volcano plots will not contain ",
                    "gene names...")
        }
        
        if (is.na(embedCols$btCol) && countType=="gene") {
            warnwrap("The column that contains the gene biotypes ",
                "(\"embedCols$btCol\") is missing with embedded annotation! ",
                "Biotype filters and certain plots will not be available...")
            geneFilters$biotype=NULL
            toRemove <- match(c("biodetection","countsbio","saturation",
                "biodist","filtered","rnacomp","readnoise"),qcPlots)
            noMatch <- which(is.na(toRemove))
            if (length(noMatch)>0)
                toRemove <- toRemove[-noMatch]
            if (length(toRemove)>0)
                qcPlots <- qcPlots[-toRemove]
        }
    }
    if (org %in% c("hg18","hg38","mm10","mm39") 
        && (refdb %in% c("ucsc","refseq"))) {
        warnwrap("Gene/exon biotypes cannot be retrieved when organism is ",
            "\"hg18\", \"hg38\", \"mm10\", \"mm39\"  and annotation database ",
            "is \"ucsc\" or \"refseq\"!\nBiotype filters and certain plots ",
            "will not be available...")
        geneFilters$biotype=NULL
        toRemove <- match(c("biodetection","countsbio","saturation",
            "biodist","filtered"),qcPlots)
        noMatch <- which(is.na(toRemove))
        if (length(noMatch)>0)
            toRemove <- toRemove[-noMatch]
        if (length(toRemove)>0)
            qcPlots <- qcPlots[-toRemove]
    }
    if (transLevel %in% c("transcript","exon")) {
        warnwrap("When transLevel is \"transcript\" or \"exon\", the GC bias ",
            "plots are not available.")
        if ("gcbias" %in% qcPlots)
            qcPlots <- qcPlots[-which(qcPlots == "gcbias")]
    }
    
    # In case someone wants a simple exon analysis but has not understood well
    # the docs of transLevel and countType...
    if (transLevel == "exon" && countType == "exon") {
        warnwrap("You have chosen transLevel=\"exon\" and countType=\"exon\".",
            "\nYou are probably trying to perform simple differential exon ",
            "analysis. This can be perofmed\nby setting transLevel=\"exon\" ",
            "and countType=\"gene\". Setting for you...")
        countType <- "gene"
    }
    
    # Remove biodist and volcano if no statistics or contrast
    if ("biodist" %in% qcPlots && (any(is.na(statistics)) 
        || is.null(contrast))) {
        warnwrap("The creation of a biodist diagram requires statistical ",
            "testing! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "biodist")]
    }
    if ("volcano" %in% qcPlots && (any(is.na(statistics)) 
        || is.null(contrast))) {
        warnwrap("The creation of a volcano plot requires statistical ",
            "testing! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "volcano")]
    }
    if ("deheatmap" %in% qcPlots && (any(is.na(statistics)) 
        || is.null(contrast))) {
        warnwrap("The creation of a deheatmap plot requires statistical ",
            "testing! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "deheatmap")]
    }
    if ("statvenn" %in% qcPlots && (any(is.na(statistics)) 
        || is.null(contrast))) {
        warnwrap("The creation of a statvenn plot requires statistical ",
            "testing! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "statvenn")]
    }
    if ("foldvenn" %in% qcPlots && (any(is.na(statistics)) 
        || is.null(contrast))) {
        warnwrap("The creation of a foldvenn plot requires statistical ",
            "testing! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "foldvenn")]
    }
    if ("deregulogram" %in% qcPlots && (any(is.na(statistics)) 
        || is.null(contrast))) {
        warnwrap("The creation of a deregulogram plot requires statistical ",
            "testing! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "deregulogram")]
    }
    # mastat only if there is a contrast, otherwise nothing to plot
    if ("mastat" %in% qcPlots && is.null(contrast)) {
        warnwrap("The creation of a mastat plot requires the definition of a ",
            "contrast! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "mastat")]
    }
    # Check if drawing a Venn diagram among tests is possible
    if ("statvenn" %in% qcPlots && length(statistics)==1 
        && !any(is.na(statistics))) {
        warnwrap("The creation of a Venn diagram among different statistical ",
            "tests is possible only when more than\none statistical ",
            "algorithms are used! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "statvenn")]
    }
    # Check if drawing a Venn diagram among contrasts is possible
    if ("foldvenn" %in% qcPlots && !is.null(contrast) &&  length(contrast)==1) {
        warnwrap("The creation of a Venn diagram among different statistical ",
            "comparisons is possible only when more than\none contrast ",
            "defined! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "foldvenn")]
    }
    # Check if drawing a deregulogram for contrast pairs
    if ("deregulogram" %in% qcPlots && !is.null(contrast) 
        && length(contrast)==1) {
        warnwrap("The creation of a deregulogram for pairwise statistical ",
            "comparisons is possible only when more than\none contrast ",
            "defined! Removing from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "deregulogram")]
    }
    # Remove meandiff if there are no replicates 
    if ("meandiff" %in% qcPlots & any(lengths(sampleList) < 2)) {
        warnwrap("The creation of a meandiff plot is not possible for single ",
            "replicate conditions!\nRemoving from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "meandiff")]
    }
    # Remove rnacomp if there are more than 10 samples
    if ("rnacomp" %in% qcPlots & sum(lengths(sampleList)) > 12) {
        warnwrap("The creation of an rnacomp plot is not possible for more ",
            "than 12 samples!\nRemoving from figures list...")
        qcPlots <- qcPlots[-which(qcPlots == "rnacomp")]
    }
    # Check if we have single-replicate conditions and presence$perCondition is
    # enabled
    if (any(lengths(sampleList) < 2) & !is.null(geneFilters$presence)) {
        if (geneFilters$presence$perCondition) {
            warnwrap("geneFilters$presence$perCondition cannot be TRUE when ",
                "there are single-replicate conditions are present\nas this ",
                "will cause sample drops! Setting to FALSE...")
            geneFilters$presence$perCondition <- FALSE
        }        
    }
    # Check if tracks asked but we are in Windows...
    if (createTracks && .Platform$OS.type == "windows") {
        warnwrap("Signal tracks asked but Windows OS detected! Track ",
            "generation is not possible in Windows. Ignoring...")
        createTracks = FALSE
    }
    # Check if tracks asked but not starting with a targets file
    if (createTracks && is.null(theList)) {
        warnwrap("Signal tracks asked but no targets file with BAM file paths ",
            "provided! Ignoring...")
        createTracks <- FALSE
    }
    # Check if tracks asked but organism is not directly supported
    if (createTracks && !(org %in% getSupportedOrganisms()))
        warnwrap("Organism ",org," is not in the list of directly supported ",
            "organisms! If the pipeline fails, please use the\n",
            "createSignalTracks function directly.")
    
    # Check additional input arguments for normalization and statistics
    algArgs <- validateAlgArgs(normalization,statistics,normArgs,statArgs)
    normArgs <- algArgs$normArgs
    statArgs <- algArgs$statArgs
    
    # Override settings if a preset is given
    if (!is.null(preset)) {
        presetOpts <- getPresetOpts(preset,org)
        exonFilters <- presetOpts$exonFilters
        geneFilters <- presetOpts$geneFilters
        pcut <- presetOpts$pcut
        exportWhat <- presetOpts$exportWhat
        exportScale <- presetOpts$exportScale
        exportValues <- presetOpts$exportValues
        exportStats <- presetOpts$exportStats
    }

    if (report) {
        reportMessages <- makeReportMessages("en")
        if (!is.null(qcPlots) && !("png" %in% figFormat)) {
            warnwrap("png format is required in order to build a report! ",
                "Adding to figure output formats...")
            figFormat <- c(figFormat,"png")
        }
    }
    
    # Check for replicates. DESeq2 cannot perform without them and ABSSeq cannot
    # perform complex model analysis if not provided with replicates.
    if ("deseq2" %in% statistics && any(lengths(sampleList) < 2)) { 
        stopwrap("DESeq2 does not support analysis without ",
            "replication. Please choose another algorithm or ",
            "provide replicates.")
    }
    
    if (!is.null(contrast) && "absseq" %in% statistics) {        
        contrastListRepCheck <- makeContrastList(contrast,sampleList)
        for (conName in names(contrastListRepCheck)) {
            con <- contrastListRepCheck[[conName]]
            cons <- unique(unlist(con))
            # Checking if we have more than 2 conditions to compare 
            # simultaneously
            if (length(cons)>2) {
                if (any(lengths(sampleList) < 2))
                    # Checking if there are no replicates
                    stopwrap("ABSSeq cannot perform complex design",
                        "analysis without replicates.")
            }
        }
    }
    
    # Display initialization report
    TB <- Sys.time()
    disp(strftime(Sys.time()),": Data processing started...\n")
    ############################################################################
    if ("dss" %in% statistics && any(lengths(sampleList) < 2)) {
        message=c(
            "\n=================================================\n", 
            "DSS will be run without replicates.\n",
            "Results must be interpreted with caution.\n",
            "===================================================\n"
        )
        disp(message)
    }
    
    disp("Read counts file: ",countsName)
    disp("Conditions: ",paste(names(sampleList),collapse=", "))
    disp("Samples to include: ",paste(unlist(sampleList),collapse=", "))
    if (!is.null(excludeList) && !is.na(excludeList))
        disp("Samples to exclude: ",paste(unlist(excludeList),collapse=", "))
    else
        disp("Samples to exclude: none")
    if (!is.null(contrast))
        disp("Requested contrasts: ",paste(contrast,collapse=", "))
    else
        disp("Requested contrasts: none")
    if (!is.null(libsizeList)) {
        disp("Library sizes: ")
        for (n in names(libsizeList))
            disp("  ",paste(n,libsizeList[[n]],sep=": "))
    }
    if (!is.null(annotation) && !is.list(annotation))
        disp("Annotation: ",annotation)
    if (is.list(annotation))
        disp("Annotation: user provided GTF file")
    disp("Organism: ",org)
    disp("Reference source: ",refdb)
    disp("Count type: ",countType)
    if (countType == "utr") {
        disp("3' UTR fraction: ",utrOpts$frac)
        disp("3' UTR minimum length: ",utrOpts$minLength,"bps")
        disp("3' UTR downstream: ",utrOpts$downstream,"bps")
    }
    if (!is.null(preset))
        disp("Analysis preset: ",preset)
    disp("Transcriptional level: ",transLevel)
    if (!is.null(exonFilters)) {
        disp("Exon filters: ",paste(names(exonFilters),collapse=", "))
        for (ef in names(exonFilters)) {
            disp("  ",ef,": ")
            for (efp in names(exonFilters[[ef]])) {
                if (length(exonFilters[[ef]][[efp]])==1 && 
                    is.function(exonFilters[[ef]][[efp]]))
                    print(exonFilters[[ef]][[efp]])
                else if (length(exonFilters[[ef]][[efp]])==1)
                    disp("    ",paste(efp,exonFilters[[ef]][[efp]],sep=": "))
                else if (length(exonFilters[[ef]][[efp]])>1)
                    disp("    ",paste(efp,paste(exonFilters[[ef]][[efp]],
                        collapse=", "),sep=": "))
            }
        }
    }
    else
        disp("Exon filters: none applied")
    if (!is.null(geneFilters)) {
        disp("Gene filters: ",paste(names(geneFilters),collapse=", "))
        for (gf in names(geneFilters)) {
            disp("  ",gf,": ")
            for (gfp in names(geneFilters[[gf]])) {
                if (length(geneFilters[[gf]][[gfp]])==1 && 
                    is.function(geneFilters[[gf]][[gfp]]))
                    print(geneFilters[[gf]][[gfp]])
                else if (length(geneFilters[[gf]][[gfp]])==1)
                    disp("    ",paste(gfp,geneFilters[[gf]][[gfp]],sep=": "))
                else if (length(geneFilters[[gf]][[gfp]])>1)
                    disp("    ",paste(gfp,paste(geneFilters[[gf]][[gfp]],
                        collapse=", "),sep=": "))
            }
        }
    }
    else
        disp("Gene filters: none applied")
    disp("Filter application: ",whenApplyFilter)
    disp("Normalization algorithm: ",normalization)
    if (!is.null(normArgs)) {
        disp("Normalization arguments: ")
        for (na in names(normArgs)) {
            if (length(normArgs[[na]])==1 && is.function(normArgs[[na]])) {
                disp("  ",na,": ")
                disp(as.character(substitute(normArgs[[na]])))
            }
            else if (length(normArgs[[na]])==1)
                disp("  ",paste(na,normArgs[[na]],sep=": "))
            else if (length(normArgs[[na]])>1)
                disp("  ",paste(na,paste(normArgs[[na]],collapse=", "),
                    sep=": "))
        }
    }
    if (!any(is.na(statistics)))
        disp("Statistical algorithm(s): ",paste(statistics,collapse=", "))
    else
        disp("Statistical algorithm(s): no testing selected")
    if (!is.null(statArgs)) {
        if (!any(is.na(statistics))) {
            disp("Statistical arguments: ")
            for (sa in names(statArgs)) {
                if (length(statArgs[[sa]])==1 && is.function(statArgs[[sa]])) {
                    disp("  ",sa,": ")
                    disp(as.character(substitute(statArgs[[na]])))
                }
                else if (length(statArgs[[sa]])==1)
                    disp("  ",paste(sa,statArgs[[sa]],sep=": "))
                else if (length(statArgs[[sa]])>1)
                    disp("  ",paste(sa,paste(statArgs[[sa]],collapse=", "),
                        sep=": "))
            }
        }
        else
            disp("Statistical arguments: no testing selected")
    }
    disp("Meta-analysis method: ",metaP)
    disp("Multiple testing correction: ",adjustMethod)
    if (!is.na(pcut)) 
        disp("p-value threshold: ",pcut)
    disp("Logarithmic transformation offset: ",logOffset)
    if (!is.null(preset)) 
        disp("Analysis preset: ",preset)
    disp("Quality control plots: ",paste(qcPlots,collapse=", "))
    disp("Figure format: ",paste(figFormat,collapse=", "))
    if (!is.na(exportWhere)) 
        disp("Output directory: ",exportWhere)
    disp("Output data: ",paste(exportWhat,collapse=", "))
    disp("Output scale(s): ",paste(exportScale,collapse=", "))
    disp("Output values: ",paste(exportValues,collapse=", "))
    if ("stats" %in% exportWhat)
        disp("Output statistics: ",paste(exportStats,collapse=", "),"\n")
        
    if (is.function(.progressFun)) {
        text <- paste("Starting the analysis...")
        .progressFun(detail=text)
    }
    ############################################################################
    
    # geneData are always required, unless annotation is embedded, where we
    # need only geneData as embedded annotation is not allowed anywhere else
    disp("Loading gene annotation...")
    if (!is.null(annotation) && annotation == "embedded") {
        # The following should work if annotation elements are arranged 
        # in MeV-like data style
        if (!is.data.frame(counts)) {
            disp("Reading counts file ",countsName,"...")
            geneCounts <- read.delim(counts)
        }
        else
            geneCounts <- counts
        rownames(geneCounts) <- as.character(geneCounts[,embedCols$idCol])
        allCols <- seq_len(ncol(geneCounts))
        samCols <- match(unlist(sampleList),colnames(geneCounts))
        samCols <- samCols[which(!is.na(samCols))]
        annCols <- allCols[-samCols]
        geneData <- geneCounts[,annCols]
        geneCounts <- geneCounts[,samCols]      
        colnames(geneData)[embedCols$idCol] <- "gene_id"
        if (!is.na(embedCols$gcCol)) {
            colnames(geneData)[embedCols$gcCol] <- "gc_content"
            if (max(geneData$gc_content<=1)) # Is already divided
                geneData$gc_content = 100*geneData$gc_content
        }
        if (!is.na(embedCols$nameCol)) 
            colnames(geneData)[embedCols$nameCol] <- "gene_name"
        if (!is.na(embedCols$btCol))
            colnames(geneData)[embedCols$btCol] <- "biotype"
        geneData <- GRanges(geneData)
        names(geneData) <- geneData$gene_id
        
        # Remove saturation if no biotypes
        if ("saturation" %in% qcPlots && (is.na(embedCols$btCol) 
            || length(unique(geneData$biotype)) == 1)) {
            warnwrap("The creation of an saturation plots is available for ",
                "more than one biotype!\nRemoving from figures list...")
            qcPlots <- qcPlots[-which(qcPlots == "saturation")]
        }
    }
    else {
        #sitaType <- ifelse(transLevel=="gene","gene",
        #   ifelse(transLevel=="transcript","transcript",
        #   ifelse(transLevel=="exon","exon")))
        summarized <- ifelse(transLevel %in% c("exon","transcript"),TRUE,FALSE)
        geneData <- tryCatch(loadAnnotation(org,refdb,level=transLevel,
            type="gene",version=version,summarized=summarized,db=localDb,
            rc=restrictCores),
            #               sitadela::loadAnnotation(org,refdb,type=sitaType,
            #version=version,db=localDb,rc=restrictCores)
            error=function(e) {
                # Not found and user based
                if (!is.null(annotation)) { # Has been checked
                    gtfFile <- annotation$gtf
                    metadata <- annotation
                    metadata$gtf <- NULL
                    geneData <- importCustomAnnotation(gtfFile,metadata,
                        "gene","gene")
                    #geneData <- sitadela::importCustomAnnotation(gtfFile,metadata,
                    #    sitaType)
                }
                else
                    stop("Please provide an existing organism or a list ",
                        "with annotation metadata and GTF file!")
            },finally="")
        
        # Remove saturation if no biotypes
        if ("saturation" %in% qcPlots && length(unique(geneData$biotype))==1) {
            warnwrap("The creation of an saturation plots is available for ",
                "more than one biotype!\nRemoving from figures list...")
            qcPlots <- qcPlots[-which(qcPlots == "saturation")]
        }
    }
    
    # Now start examining additional required data per case...
    if (countType=="exon") {
        # We need to load the exon annotation and see what counts have been
        # provided
        if (!fromPrevious) {
            # Load exon annotation
            disp("Loading exon annotation...")
            #sitaType <- ifelse(transLevel=="gene","exon",
            #   ifelse(transLevel=="transcript","transexon",
            #   ifelse(transLevel=="exon","exon")))
            exonData <- tryCatch(loadAnnotation(org,refdb,level=transLevel,
                type=countType,version=version,db=localDb,summarized=TRUE,
                rc=restrictCores),
                                #sitadela::loadAnnotation(org,refdb,type=sitaType,
                #version=version,db=localDb,summarized=TRUE,rc=restrictCores)
            error=function(e) {
                # Not found and user based
                if (!is.null(annotation)) { # Has been checked
                    gtfFile <- annotation$gtf
                    metadata <- annotation
                    metadata$gtf <- NULL
                    exonData <- importCustomAnnotation(gtfFile,metadata,
                        "gene","exon")
                    #exonData <- importCustomAnnotation(gtfFile,metadata,
                    #    sitaType)
                }
                else
                    stop("Please provide an existing organism or a list ",
                        "with annotation metadata and GTF file!")
            },finally="")
            
            # Load/read counts
            if (!is.null(counts)) { # Provided
                if (!is.data.frame(counts) && !is.list(counts)) {
                    disp("Reading counts file ",countsName,"...")
                    exonCounts <- read.delim(counts,row.names=1)

                    ## Re-arrange columns to be in consistency with colData
                    ## used in statDeseq2
                    exonCounts <- exonCounts[,unlist(sampleList,
                        use.names=FALSE),drop=FALSE]
                }
                else { # Already a data frame as input
                    if (is.character(counts[,1])) {
                        exonCounts <- as.matrix(counts[,-1])
                        rownames(exonCounts) <- as.character(counts[,1])

                        ## Re-arrange columns to be in consistency with colData
                        ## used in statDeseq2
                        exonCounts <- exonCounts[,unlist(sampleList,
                            use.names=FALSE),drop=FALSE]
                    }
                    else { # Should be named!
                        if (is.null(rownames(counts)))
                            stopwrap("A counts data frame as input should ",
                                "have rownames!")
                        exonCounts <- counts

                        ## Re-arrange columns to be in consistency with colData
                        ## used in statDeseq2
                        exonCounts <- exonCounts[,unlist(sampleList,
                            use.names=FALSE),drop=FALSE]
                    }
                }
            }
            else { # Coming from read2count
                if (fromRaw) { # Double check
                    r2c <- read2count(theList,exonData,fileType,transLevel,
                        utrOpts,rc=restrictCores)
                    exonCounts <- r2c$counts
                    # Merged exon data! - Not needed anymore
                    #exonData <- r2c$mergedann
                    if (is.null(libsizeList))
                        libsizeList <- r2c$libsize
                    if (exportCountsTable) {
                        exonDataExp <- as.data.frame(exonData)
                        exonDataExp <- exonDataExp[,c(1,2,3,6,7,5,8,9)]
                        names(exonDataExp)[1] <- "chromosome"
                        disp("Exporting raw read counts table to ",
                            file.path(PROJECT_PATH[["lists"]],
                                "raw_counts_table.txt.gz"))
                        resFile <- file.path(PROJECT_PATH[["lists"]],
                            "raw_counts_table.txt.gz")
                        gzfh <- gzfile(resFile,"w")
                        write.table(cbind(
                            exonDataExp[rownames(exonCounts),],
                            exonCounts
                        ),gzfh,sep="\t",row.names=FALSE,quote=FALSE)
                        close(gzfh)
                    }
                }
            }
            exonCounts <- exonCounts[,unlist(sampleList,use.names=FALSE)]

            # Get the exon counts per gene model
            disp("Checking chromosomes in exon counts and gene annotation...")
            geneData <- .reduceGeneData(exonData[rownames(exonCounts)],geneData)
            exonData <- .reduceGeneData(exonData[rownames(exonCounts)],exonData)
            disp("Processing exons...")
            theCounts <- constructGeneModel(exonCounts,exonData,type="exon",
                rc=restrictCores)

            if (saveGeneModel) {
                disp("Saving gene model to ",file.path(PROJECT_PATH[["data"]],
                    "gene_model.RData"))
                save(theCounts,exonData,geneData,sampleList,transLevel,
                    countType,file=file.path(PROJECT_PATH$data,
                    "gene_model.RData"),compress=TRUE)
            }
        }
        else {
            theCounts <- tmpEnv$theCounts
            exonData <- tmpEnv$exonData
            geneData <- tmpEnv$geneData
        }

        # Exclude any samples not wanted (when e.g. restoring a previous project
        # and having determined that some samples are of bad quality OR using
        # a different condition set with a subset of known samples
        if (!is.null(excludeList) && !is.na(excludeList)) {
            for (n in names(excludeList)) {
                sampleList[[n]] <- setdiff(sampleList[[n]],
                    excludeList[[n]])
                if (length(sampleList[[n]])==0) # Removed whole condition
                    sampleList[n] <- NULL
            }
            # Attributes are lost when subsetting
            # https://cran.r-project.org/web/packages/sticky/
            #vignettes/introduction.html
            theCounts <- .excludeSamplesFromList(theCounts,
                unlist(sampleList,use.names=FALSE))
        }
        if (!is.null(sampleListPrev) && !identical(sampleList,sampleListPrev))
            theCounts <- .excludeSamplesFromList(theCounts,
                unlist(sampleList,use.names=FALSE))
        
        # Apply exon filters
        if (!is.null(exonFilters)) {
            exonFilterOut <- filterExons(theCounts,geneData,sampleList,
                exonFilters)
            exonFilterResult <- exonFilterOut$result
            exonFilterFlags <- exonFilterOut$flags
        }
        else
            exonFilterResult <- exonFilterFlags <- NULL
        
        disp("Summarizing count data...")
        theGeneCounts <- theExonLengths <- vector("list",
            length(unlist(sampleList)))
        names(theGeneCounts) <- names(theExonLengths) <- names(theCounts)
        for (n in names(theGeneCounts))
            theGeneCounts[[n]] <- vapply(theCounts[[n]],sum,numeric(1))
        geneCounts <- do.call("cbind",theGeneCounts)
        # Based on the sum of their exon lengths
        lengthList <- attr(theCounts,"lengthList")
        geneLength <- vapply(lengthList,sum,numeric(1))
        # Could also be
        # geneLength <- attr(exonData,"activeLength")
        
        # In case there are small differences between annotation data and  
        # external file, due to e.g. slightly different Ensembl versions
        geneData <- geneData[rownames(geneCounts)]
        totalGeneData <- geneData # We need this for some total stats
    }
    
    if (countType=="utr") {
        # We need to load the utr annotation and see what counts have been
        # provided
        if (!fromPrevious) {
            disp("Loading 3' UTR annotation...")
            #sitaType <- ifelse(transLevel=="gene","utr",
            #   ifelse(transLevel=="transcript","transutr",
            #   ifelse(transLevel=="exon","utr")))
            transcriptData <- tryCatch(loadAnnotation(org,refdb,
                level=transLevel,type=countType,version=version,db=localDb,
                summarized=TRUE,rc=restrictCores),
                #   sitadela::loadAnnotation(org,refdb,
                #type=sitaType,version=version,db=localDb,
                #summarized=TRUE,rc=restrictCores)
                error=function(e) {
                # Not found and user based
                if (!is.null(annotation)) { # Has been checked
                    gtfFile <- annotation$gtf
                    metadata <- annotation
                    metadata$gtf <- NULL
                    transcriptData <- importCustomAnnotation(gtfFile,metadata,
                        transLevel,countType)
                    #transcriptData <- sitadela::importCustomAnnotation(gtfFile,metadata,
                    #    sitaType)
                }
                else
                    stop("Please provide an existing organism or a list with ",
                        "annotation metadata and GTF file!")
            },finally="")
            
            # Load/read counts
            if (!is.null(counts)) { # Otherwise coming ready from read2count
                if (!is.data.frame(counts) && !is.list(counts)) {
                    disp("Reading counts file ",countsName,"...")
                    transcriptCounts <- read.delim(counts, row.names=1)

                    ## Re-arrange columns to be in consistency with colData
                    ## used in statDeseq2
                    transcriptCounts <- transcriptCounts[,unlist(sampleList,
                        use.names= FALSE)]
                }
                else { # Already a data frame as input
                    if (is.character(counts[,1])) {
                        transcriptCounts <- as.matrix(counts[,-1])
                        rownames(transcriptCounts) <- as.character(counts[,1])

                        ## Re-arrange columns to be in consistency with colData
                        ## used in statDeseq2
                        transcriptCounts <- transcriptCounts[,unlist(sampleList,
                            use.names=FALSE),drop=FALSE]
                    }
                    else { # Should be named!
                        if (is.null(rownames(counts)))
                            stopwrap("A counts data frame as input should ",
                                "have rownames!")
                        transcriptCounts <- counts

                        ## Re-arrange columns to be in consistency with colData
                        ## used in statDeseq2
                        transcriptCounts <- transcriptCounts[,unlist(sampleList,
                            use.names=FALSE),drop=FALSE]
                    }
                }
            }
            else { # Coming from read2count
                if (fromRaw) { # Double check
                    r2c <- read2count(theList,transcriptData,fileType,
                        transLevel,utrOpts,rc=restrictCores)
                    transcriptCounts <- r2c$counts
                    if (is.null(libsizeList))
                        libsizeList <- r2c$libsize
                    if (exportCountsTable) {
                        transcriptDataExp <- as.data.frame(transcriptData)
                        transcriptDataExp <- 
                            transcriptDataExp[,c(1,2,3,6,7,5,8,9)]
                        names(transcriptDataExp)[1] <- "chromosome"
                        disp("Exporting raw read counts table to ",
                            file.path(PROJECT_PATH[["lists"]],
                            "raw_counts_table.txt.gz"))
                        resFile <- file.path(PROJECT_PATH[["lists"]],
                            "raw_counts_table.txt.gz")
                        gzfh <- gzfile(resFile,"w")
                        write.table(cbind(
                            transcriptDataExp[rownames(transcriptCounts),],
                            transcriptCounts
                        ),gzfh,sep="\t",row.names=FALSE,quote=FALSE)
                        close(gzfh)
                    }
                }
            }
            transcriptCounts <- transcriptCounts[,unlist(sampleList,
                use.names=FALSE)]

            # Get the transcript counts per gene model
            disp("Checking chromosomes in transcript counts and gene ",
                "annotation...")
            geneData <- 
                .reduceGeneData(transcriptData[rownames(transcriptCounts)],
                    geneData)
            transcriptData <- 
                .reduceGeneData(transcriptData[rownames(transcriptCounts)],
                    transcriptData)
            disp("Processing transcripts...")
            theCounts <- constructGeneModel(transcriptCounts,transcriptData,
                type="utr",rc=restrictCores)
            
            if (saveGeneModel) {
                disp("Saving gene model to ",file.path(PROJECT_PATH[["data"]],
                    "gene_model.RData"))
                save(theCounts,transcriptData,geneData,sampleList,countType,
                    transLevel,file=file.path(PROJECT_PATH$data,
                    "gene_model.RData"),compress=TRUE)
            }
        }
        else {
            theCounts <- tmpEnv$theCounts
            transcriptData <- tmpEnv$transcriptData
            geneData <- tmpEnv$geneData
        }

        # Exclude any samples not wanted (when e.g. restoring a previous project
        # and having determined that some samples are of bad quality
        if (!is.null(excludeList) && !is.na(excludeList)) {
            for (n in names(excludeList)) {
                sampleList[[n]] <- setdiff(sampleList[[n]],
                    excludeList[[n]])
                if (length(sampleList[[n]])==0) # Removed whole condition
                    sampleList[n] <- NULL
            }
            #theCounts <- theCounts[unlist(sampleList)]
            theCounts <- .excludeSamplesFromList(theCounts,
                unlist(sampleList,use.names=FALSE))
        }
        
        disp("Summarizing count data...")
        theGeneCounts <- theTranscriptLengths <- vector("list",
            length(unlist(sampleList)))
        names(theGeneCounts) <- names(theTranscriptLengths) <- 
            names(theCounts)
        for (n in names(theGeneCounts))
            theGeneCounts[[n]] <- vapply(theCounts[[n]],sum,numeric(1))
        geneCounts <- do.call("cbind",theGeneCounts)
        # Based on the sum of their transcript lengths
        lengthList <- attr(theCounts,"lengthList")
        geneLength <- vapply(lengthList,sum,numeric(1))
        # Could also be
        # geneLength <- attr(transcriptData,"activeLength")
        
        # In case there are small differences between annotation data and 
        # external file, due to e.g. slightly different Ensembl versions
        geneData <- geneData[rownames(geneCounts),]
        totalGeneData <- geneData # We need this for some total stats
        
        # No exon filters have been applied, we are at "utr" level
        exonFilterResult <- NULL
    }
    
    if (countType=="gene") {
        # geneData has already been loaded and also geneCounts in the case of
        # embedded annotation
        if (is.null(annotation) || annotation != "embedded") {
            if (!fromPrevious) {
                # Load/read counts
                if (!is.null(counts) 
                    && (is.character(counts) || is.data.frame(counts))) {
                    if (!is.data.frame(counts)) { # Else it's already here
                        disp("Reading counts file ",countsName,"...")
                        geneCounts <- read.delim(counts,row.names=1)

                        ## Re-arrange columns to be in consistency with colData
                        ## used in statDeseq2
                        geneCounts <- geneCounts[,unlist(sampleList,
                            use.names=FALSE),drop=FALSE]
                    }
                    else { # Already a data frame as input
                        if (is.character(counts[,1])) {
                            geneCounts <- as.matrix(counts[,-1])
                            rownames(geneCounts) <- as.character(counts[,1])

                            ## Re-arrange columns to be in consistency with 
                            ## colData used in statDeseq2
                            geneCounts <- geneCounts[,unlist(sampleList,
                                use.names=FALSE),drop=FALSE]
                        }
                        else { # Should be named!
                            if (is.null(rownames(counts)))
                                stopwrap("A counts data frame as input should ",
                                    "have rownames!")
                            geneCounts <- counts

                            ## Re-arrange columns to be in consistency with 
                            ## colData used in statDeseq2
                            geneCounts <- geneCounts[,unlist(sampleList,
                                use.names=FALSE),drop=FALSE]
                        }
                    }
                }
                else { # Coming from read2count
                    if (fromRaw) { # Double check
                        r2c <- read2count(theList,geneData,fileType,transLevel,
                            utrOpts,rc=restrictCores)
                        geneCounts <- r2c$counts
                        if (is.null(libsizeList))
                            libsizeList <- r2c$libsize
                        if (exportCountsTable) {
                            geneDataExp <- as.data.frame(geneData)
                            geneDataExp <- geneDataExp[,c(1,2,3,6,7,5,8,9)]
                            names(geneDataExp)[1] <- "chromosome"
                            disp("Exporting raw read counts table to ",
                                file.path(PROJECT_PATH[["lists"]],
                                "raw_counts_table.txt.gz"))
                            resFile <- file.path(PROJECT_PATH[["lists"]],
                                "raw_counts_table.txt.gz")
                            gzfh <- gzfile(resFile,"w")
                            write.table(cbind(
                                geneDataExp[rownames(geneCounts),],
                                geneCounts
                            ),gzfh,sep="\t",row.names=FALSE,quote=FALSE)
                            close(gzfh)
                        }
                    }
                }
            }
            else {
                geneCounts <- tmpEnv$geneCounts
                geneData <- tmpEnv$geneData
            }
        }
        
        totalGeneData <- geneData # We need this for some total stats
        exonFilterResult <- NULL

        if (transLevel == "gene")
            gValid <- intersect(rownames(geneCounts),geneData$gene_id)
        else if (transLevel == "transcript")
            gValid <- intersect(rownames(geneCounts),geneData$transcript_id)
        else if (transLevel == "exon")
            gValid <- intersect(rownames(geneCounts),geneData$exon_id)
        geneCounts <- geneCounts[gValid,]

        geneData <- geneData[rownames(geneCounts)]
        geneLength <- width(geneData)
        names(geneLength) <- names(geneData)
        
        if (saveGeneModel) {
            disp("Saving gene model to ",file.path(PROJECT_PATH[["data"]],
                "gene_model.RData"))
            save(geneCounts,geneData,sampleList,countType,
                file=file.path(PROJECT_PATH$data,"gene_model.RData"),
                compress=TRUE)
        }
        
        # Exclude any samples not wanted (when e.g. restoring a previous project
        # and having determined that some samples are of bad quality
        if (!is.null(excludeList) && !is.na(excludeList)) {
            for (n in names(excludeList)) {
                sampleList[[n]] <- setdiff(sampleList[[n]],
                    excludeList[[n]])
                if (length(sampleList[[n]])==0) # Removed whole condition
                    sampleList[n] <- NULL
            }
            geneCounts <- geneCounts[,unlist(sampleList,use.names=FALSE)]
        }
    }

    # Transform GC-content and biotype - should never be required in the new
    # annotation strategy
    if (is.null(geneData$gc_content))
        geneData$gc_content <- rep(0.5,length(geneData))
    if (is.null(geneData$biotype))
        geneData$biotype <- rep("gene",length(geneData))
    
    names(geneLength) <- rownames(geneCounts)
    attr(geneData,"geneLength") <- geneLength

    ############################################################################
    # BEGIN FILTERING SECTION
    ############################################################################
    
    if (is.function(.progressFun)) {
        text <- paste("Filtering...")
        .progressFun(detail=text)
    }
    
    # GC bias is NOT alleviated if we do not remove the zeros!!!
    disp("Removing genes with zero counts in all samples...")
    theZeros <- which(apply(geneCounts,1,filterLow,0))
    if (length(theZeros) > 0) {
        # Store the filtered, maybe we do some stats
        geneCountsZero <- geneCounts[theZeros,,drop=FALSE]
        geneDataZero <- geneData[theZeros]
        attr(geneDataZero,"geneLength") <- geneLength[theZeros]
        theZeroNames <- names(geneData)[theZeros]
        # Then remove
        geneCounts <- geneCounts[-theZeros,,drop=FALSE]
        geneData <- geneData[-theZeros]
        attr(geneData,"geneLength") <- geneLength[-theZeros]
    }
    else
        geneCountsZero <- geneDataZero <- theZeroNames <- NULL

    # Store un-normalized gene counts for export purposes
    geneCountsUnnorm <- geneCounts

    # Apply filtering prior to normalization if desired
    if (whenApplyFilter=="prenorm") {
        # However, a first round of normalization has to be performed in
        # order to get proper expression filters
        disp("Prefiltering normalization with: ",normalization)
        switch(normalization,
            edaseq = {
                tempGenes <- normalizeEdaseq(geneCounts,sampleList,
                    normArgs,geneData,output="matrix")
            },
            deseq = {
                tempGenes <- normalizeDeseq(geneCounts,sampleList,normArgs,
                    output="matrix")
            },
            deseq2 = {
                tempGenes <- normalizeDeseq2(geneCounts,sampleList,normArgs,
                    output="matrix")
            },   
            edger = {
                tempGenes <- normalizeEdger(geneCounts,sampleList,normArgs,
                    output="matrix")
            },
            noiseq = {  
                tempGenes <- normalizeNoiseq(geneCounts,sampleList,
                    normArgs,geneData,logOffset,output="matrix")
            },
            nbpseq = {
                tempGenes <- normalizeNbpseq(geneCounts,sampleList,
                    normArgs,libsizeList,output="matrix")
            },
            absseq = {
                tempGenes <- normalizeAbsseq(geneCounts,sampleList,normArgs,
                    output="matrix")
            },
            dss = {
                tempGenes <- normalizeDss(geneCounts,sampleList,normArgs,
                    output="native")
            },
            none = {
                # In case some external normalization is applied (e.g. equal
                # read counts from all samples)
                tempGenes <- geneCounts
            }
        )
        
        # Now filter
        if (!is.null(geneFilters)) {
            geneFilterOut <- filterGenes(tempGenes,geneData,geneFilters,
                sampleList)
            geneFilterResult <- geneFilterOut$result
            geneFilterCutoff <- geneFilterOut$cutoff
            geneFilterFlags <- geneFilterOut$flags
        }
        else
            geneFilterResult <- geneFilterCutoff <- geneFilterFlags <- NULL

        # Unify the filters and filter
        theDeadGenes <- list(
            geneFilterResult$expression$median,
            geneFilterResult$expression$mean,
            geneFilterResult$expression$quantile,
            geneFilterResult$expression$known,
            geneFilterResult$expression$custom
        )
        theDead <- unique(unlist(c(geneFilterResult,exonFilterResult)))
        # Some genes filtered by zero, were present in exon filters, not yet 
        # applied
        if (countType=="exon")
            theDead <- setdiff(theDead,theZeroNames)
        
        # All method specific objects are row-index subsettable
        if (length(theDead)>0) {
            # Store the filtered for later export or some stats
            geneCountsDead <- geneCounts[theDead,]
            geneCountsUnnorm <- geneCountsUnnorm[theDead,]
            geneDataDead <- geneData[theDead]
            attr(geneDataDead,"geneLength") <- attr(geneData,
                "geneLength")[theDead]
            # Now filter
            theDeadInd <- match(theDead,rownames(geneCounts))
            geneCountsExpr <- geneCounts[-theDeadInd,]
            geneDataExpr <- geneData[-theDeadInd]
            attr(geneDataExpr,"geneLength") <- attr(geneData,
                "geneLength")[-theDeadInd]
        }
        else {
            geneCountsExpr <- geneCounts
            geneDataExpr <- geneData
            geneCountsDead <- geneDataDead <- geneCountsUnnorm <- NULL
        }
        
        if (is.function(.progressFun)) {
            text <- paste("Normalizing...")
            .progressFun(detail=text)
        }
        
        disp("Normalizing with: ",normalization)
        switch(normalization,
            edaseq = {
                normGenes <- normalizeEdaseq(geneCountsExpr,sampleList,normArgs,
                    geneDataExpr,output="matrix")
            },
            deseq = {
                normGenes <- normalizeDeseq(geneCountsExpr,sampleList,normArgs,
                    output="native")
            },
            deseq2 = {
                normGenes <- normalizeDeseq2(geneCountsExpr,sampleList,normArgs,
                    output="native")
            },
            edger = {
                normGenes <- normalizeEdger(geneCountsExpr,sampleList,normArgs,
                    output="native")
            },
            noiseq = {
                normGenes <- normalizeNoiseq(geneCountsExpr,sampleList,normArgs,
                    geneDataExpr,logOffset,output="matrix")
            },
            nbpseq = {
                normGenes <- normalizeNbpseq(geneCountsExpr,sampleList,normArgs,
                    libsizeList,output="native")
            },
            absseq = {
                normGenes <- normalizeAbsseq(geneCountsExpr,sampleList,normArgs,
                    output="native")
            },
            dss = {
                normGenes <- normalizeDss(geneCountsExpr,sampleList,normArgs,
                    output="native")
            },   
            none = {
                normGenes <- geneCountsExpr
            }
        )
        normGenesExpr <- normGenes
    }
    else if (whenApplyFilter=="postnorm") {
        if (is.function(.progressFun)) {
            text <- paste("Normalizing...")
            .progressFun(detail=text)
        }
        
        # Apply filtering after normalization if desired (default)
        disp("Normalizing with: ",normalization)   
        switch(normalization,
            edaseq = {
                normGenes <- normalizeEdaseq(geneCounts,sampleList,
                    normArgs,geneData,output="matrix")
            },
            deseq = {
                normGenes <- normalizeDeseq(geneCounts,sampleList,normArgs,
                    output="native")
            },
            deseq2 = {
                normGenes <- normalizeDeseq2(geneCounts,sampleList,
                    normArgs,output="native")
            },
            edger = {
                normGenes <- normalizeEdger(geneCounts,sampleList,normArgs,
                    output="native")
            },
            noiseq = {
                normGenes <- normalizeNoiseq(geneCounts,sampleList,
                    normArgs,geneData,logOffset,output="matrix")
            },
            nbpseq = {
                normGenes <- normalizeNbpseq(geneCounts,sampleList,
                    normArgs,libsizeList,output="native")
            },
            absseq = {
                normGenes <- normalizeAbsseq(geneCounts,sampleList,
                    normArgs,output="native")
            },
            dss = {
                normGenes <- normalizeDss(geneCounts,sampleList,normArgs,
                    output="native")
            },
            none = {
                normGenes <- geneCounts
            }
        )
        
        switch(class(normGenes)[1], 
            .CountDataSet = { # Has been normalized with DESeq
                #tempMatrix <- round(DESeq::counts(normGenes,normalized=TRUE))
                tempMatrix <- round(counts(normGenes,normalized=TRUE))
            },
            DESeqDataSet = { # Has been normalized with DESeq2
                tempMatrix <- round(DESeq2::counts(normGenes,normalized=TRUE))
            },
            DGEList = { # Has been normalized with edgeR
                # Trick found at http://cgrlucb.wikispaces.com/edgeR+spring2013
                scl <- normGenes$samples$lib.size *
                    normGenes$samples$norm.factors
                tempMatrix <- round(t(t(normGenes$counts)/scl)*mean(scl))
            },
            matrix = { # Has been normalized with EDASeq or NOISeq or nothing
                tempMatrix <- normGenes
            },
            data.frame = { # Has been normalized with or nothing
                tempMatrix <- as.matrix(normGenes)
            },
            nbData = { # Has been normalized with NBPSeq and main was "nbpseq"
                tempMatrix <- as.matrix(round(sweep(normGenes$counts,2,
                    normGenes$norm.factors,"*")))
            },
            nbp = { # Has been normalized with NBPSeq and main was "nbsmyth"
                tempMatrix <- as.matrix(round(normGenes$pseudo.counts))
            },
            ABSDataSet = { # Has been normalized with ABSSeq
                tempMatrix <- as.matrix(round(excounts(normGenes)))
            },
            SeqCountSet = { # Has been normalized with DSS
                # Dribble for taking a mtx out of SeqCountSet class
                classes <- asClassVector(sampleList)
                theDesign <- data.frame(condition=classes,
                    row.names=colnames(normGenes)) 
                cds <- newCountDataSet(as.matrix(round(
                    assayData(normGenes)$exprs)),theDesign$condition)
                #DESeq::sizeFactors(cds) <- normalizationFactor(normGenes)
                sizeFactors(cds) <- normalizationFactor(normGenes)
                #tempMatrix <- as.matrix(round(DESeq::counts(cds,
                #    normalized=TRUE)))
                tempMatrix <- as.matrix(round(counts(cds,normalized=TRUE)))
            }
        )

        # Implement gene filters after normalization
        if (!is.null(geneFilters)) {
            geneFilterOut <- filterGenes(tempMatrix,geneData,geneFilters,
                sampleList)
            geneFilterResult <- geneFilterOut$result
            geneFilterCutoff <- geneFilterOut$cutoff
            geneFilterFlags <- geneFilterOut$flags
        }
        else
            geneFilterResult <- geneFilterCutoff <-
                geneFilterFlags <- NULL

        # Unify the filters and filter
        theDeadGenes <- list(
            geneFilterResult$expression$median,
            geneFilterResult$expression$mean,
            geneFilterResult$expression$quantile,
            geneFilterResult$expression$known,
            geneFilterResult$expression$custom
        )
        #when running maually exonFilterResult=NULL
        theDead <- unique(unlist(c(geneFilterResult,exonFilterResult)))
        # Some genes filtered by zero, were present in exon filters, not yet 
        # applied
        if (countType=="exon")
            theDead <- setdiff(theDead,theZeroNames)
        
        # All method specific objects are row-index subsettable
        if (length(theDead) > 0) {
            # Store the filtered for later export or some stats
            geneCountsDead <- tempMatrix[theDead,]
            geneCountsUnnorm <- geneCountsUnnorm[theDead,]
            geneDataDead <- geneData[theDead]
            attr(geneDataDead,"geneLength") <- attr(geneData,
                "geneLength")[theDead]
            # Now filter
            theDeadInd <- match(theDead,rownames(tempMatrix))  
            switch(class(normGenes)[1],                           
                .CountDataSet = { # Has been normalized with DESeq
                    normGenesExpr <- normGenes[-theDeadInd,]
                },
                DESeqDataSet = { # Has been normalized with DESeq2
                    normGenesExpr <- normGenes[-theDeadInd,]
                },
                DGEList = { # edgeR bug???
                    normGenesExpr <- normGenes[-theDeadInd,]
                    normGenesExpr$AveLogCPM <-
                        normGenesExpr$AveLogCPM[-theDeadInd]
                },
                matrix = { # Has been normalized with EDASeq or NOISeq
                    normGenesExpr <- normGenes[-theDeadInd,]
                },
                data.frame = { # Has been normalized with EDASeq or NOISeq
                    normGenesExpr <- as.matrix(normGenes[-theDeadInd,])
                },
                nbData = { # Has been normalized NBPSeq, main.method="nbpseq"
                    normGenesExpr <- normGenes
                    normGenesExpr$counts <-
                        as.matrix(normGenesExpr$counts[-theDeadInd,])
                    normGenesExpr$rel.frequencies <- 
                        normGenesExpr$rel.frequencies[-theDeadInd,]
                    normGenesExpr$tags <-
                        as.matrix(normGenesExpr$tags[-theDeadInd,])
                },
                nbp = {
                    normGenesExpr <- normGenes
                    normGenesExpr$counts <-
                        as.matrix(normGenesExpr$counts[-theDeadInd,])
                    normGenesExpr$pseudo.counts <- 
                        as.matrix(normGenesExpr$pseudo.counts[-theDeadInd,])
                    normGenesExpr$pseudo.libSizes <- 
                        colSums(as.matrix(normGenesExpr$pseudo.counts))*
                            rep(1,dim(normGenesExpr$counts)[2])
                },
                ABSDataSet = { 
                    normGenesExpr <- normGenes
                    ABSSeq::counts(normGenesExpr) <-
                        ABSSeq::counts(normGenesExpr)[-theDeadInd,]
                    excounts(normGenesExpr) <-
                        excounts(normGenesExpr)[-theDeadInd,]
                },
                SeqCountSet = { 
                    normGenesExpr <- normGenes
                    # Subset raw counts table
                    normGenesExpr <- normGenesExpr[-theDeadInd,] 
                    # All the other tools here return a matrix. DSS must do the 
                    # same. --> otherwise error in rbind of line 3192
                    classes <- asClassVector(sampleList)
                    theDesign <- data.frame(condition=classes,
                        row.names=colnames(normGenesExpr)) 
                    cds <- newCountDataSet(as.matrix(round(
                        assayData(normGenesExpr)$exprs)),theDesign$condition)
                    #DESeq::sizeFactors(cds) <- 
                    #    normalizationFactor(normGenesExpr)
                    sizeFactors(cds) <- normalizationFactor(normGenesExpr)
                    #normGenesExpr <- as.matrix(round(DESeq::counts(cds,
                    #    normalized=TRUE)))
                    normGenesExpr <- as.matrix(round(counts(cds,
                        normalized=TRUE)))
                }
            )
            geneCountsExpr <- geneCounts[rownames(normGenesExpr),]
            geneDataExpr <- geneData[-theDeadInd]
            attr(geneDataExpr,"geneLength") <-
                attr(geneData,"geneLength")[-theDeadInd]
            
        }
        else {
            normGenesExpr <- normGenes
            geneCountsExpr <- geneCounts
            geneDataExpr <- geneData
            geneCountsDead <- geneDataDead <- geneCountsUnnorm <- NULL
        }
    }
    
    # Store the final filtered, maybe we do some stats
    geneDataFiltered <- c(geneDataZero,geneDataDead)
    if (!is.null(geneDataFiltered) && length(geneDataFiltered) > 0) {
        disp(length(geneDataFiltered)," genes filtered out")
        if (!is.null(geneDataZero) && length(geneDataZero) > 0)
            attr(geneDataFiltered,"geneLength") <- c(attr(geneDataZero,
                "geneLength"),attr(geneDataDead,"geneLength"))
        else
            attr(geneDataFiltered,"geneLength") <-
                attr(geneDataDead,"geneLength")
    }
    if (!is.null(geneFilters) || !is.null(exonFilters))
        disp(length(geneDataExpr)," genes remain after filtering")

    ############################################################################
    # END FILTERING SECTION
    ############################################################################

    # There is a small case that no genes are left after filtering...
    if(any(dim(normGenesExpr)==0))
        stopwrap("No genes left after gene and/or exon filtering! Try again ",
            "with no filtering or less strict filter rules...")

    if (!is.null(contrast) && !any(is.na(statistics))) {
        if (is.function(.progressFun)) {
            text <- paste("Statistical testing...")
            .progressFun(detail=text)
        }
        
        # Run the statistical test, normGenes is always a method-specific
        # object, handled in the metaseqr.stat.R stat.* functions
        cpList <- vector("list",length(contrast))
        names(cpList) <- contrast
        contrastList <- makeContrastList(contrast,sampleList)
        for (n in names(cpList)) {
            cpList[[n]] <- vector("list",length(statistics))
            names(cpList[[n]]) <- statistics
        }
        for (alg in statistics) {
            disp("Running statistical tests with: ",alg)    
            switch(alg,
                deseq = {
                    pList <- statDeseq(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]])
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                },
                deseq2 = {
                    pList <- statDeseq2(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]])
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                },   
                edger = {
                    pList <- statEdger(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]])
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                },
                noiseq = {
                    pList <- statNoiseq(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]],geneDataExpr,logOffset)
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                },
                #bayseq = {
                #    pList <- statBayseq(normGenesExpr,sampleList,contrastList,
                #        statArgs[[alg]],libsizeList)
                #    if (!is.na(pcut)) {
                #        for (con in names(contrastList))
                #            disp("  Contrast ",con,": found ",
                #                length(which(pList[[con]]<=pcut))," genes")
                #    }
                #},
                limma = {
                    pList <- statLimma(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]])
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                },
                nbpseq = {
                    pList <- statNbpseq(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]],libsizeList)
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                },
                absseq = {
                    pList <- statAbsseq(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]])
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                },
                dss = {
                    pList <- statDss(normGenesExpr,sampleList,contrastList,
                        statArgs[[alg]])
                    # Order pList genes as in normGenesExpr, because dss outputs
                    # gene names in a different order
                    pList[[1]] <- pList[[1]][rownames(normGenesExpr)]
                    if (!is.na(pcut)) {
                        for (con in names(contrastList))
                            disp("  Contrast ",con,": found ",
                                length(which(pList[[con]]<=pcut))," genes")
                    }
                }
            )
            for (n in names(pList))
                cpList[[n]][[alg]] <- pList[[n]]
        }
        for (n in names(cpList))
            cpList[[n]] <- do.call("cbind",cpList[[n]])

        # Create the adjusted p-value matrices (if needed)
        if ("adj_p_value" %in% exportWhat) {
            adjCpList <- cmclapply(cpList,
                function(x,a) return(apply(x,2,p.adjust,a)),adjustMethod,
                rc=restrictCores)
            for (n in names(cpList)) {
                noi <- grep("noiseq",colnames(cpList[[n]]))
                if (length(noi)>0) {
                    # DESeq has not run in this case, FDR cannot be calculated
                    if (length(strsplit(n,"_vs_")[[1]])==2)
                        adjCpList[[n]][,noi] <- rep(NA,nrow(cpList[[n]]))
                }
            }
        }
        else
            adjCpList <- NULL
    }
    else
        # No contrast provided or no staistical tests selected, some dummy 
        # p-value data must exist for export
        cpList <- adjCpList <- NULL
    
    if (!is.null(contrast)) # In case of only statistics=NA
        contrastList <- makeContrastList(contrast,sampleList)

    # At this point, all method-specific objects must become matrices for  
    # exporting and plotting
    switch(class(normGenesExpr)[1],
        .CountDataSet = { # Has been processed with DESeq
            #normGenes <- round(DESeq::counts(normGenes,normalized=TRUE))
            normGenes <- round(counts(normGenes,normalized=TRUE))
            #normGenesExpr <- round(DESeq::counts(normGenesExpr,
            #    normalized=TRUE))
            normGenesExpr <- round(counts(normGenesExpr,normalized=TRUE))
        },
        DESeqDataSet = { # Has been processed with DESeq2
            normGenes <- round(DESeq2::counts(normGenes,normalized=TRUE))
            normGenesExpr <- round(DESeq2::counts(normGenesExpr,
                normalized=TRUE))
        },
        DGEList = { # Has been processed with edgeR
            # Trick found at http://cgrlucb.wikispaces.com/edgeR+spring2013
            sclR <- normGenes$samples$lib.size*normGenes$samples$norm.factors
            normGenes <- round(t(t(normGenes$counts)/sclR)*mean(sclR))
            sclN <- normGenesExpr$samples$lib.size *
                normGenesExpr$samples$norm.factors
            normGenesExpr <- round(t(t(normGenesExpr$counts)/sclN) *
                mean(sclN))
        },
        nbData = {
            normGenes <- as.matrix(round(sweep(normGenes$counts,2,
                normGenes$norm.factors,"*")))
            normGenesExpr <- as.matrix(round(sweep(normGenesExpr$counts,2,
                normGenes$norm.factors,"*")))
        },
        nbp = {
            normGenes <- as.matrix(round(normGenes$pseudo.counts))
            normGenesExpr <- as.matrix(round(normGenesExpr$pseudo.counts))
        },
        ABSDataSet = {
            normGenes <- as.matrix(round(excounts(normGenes)))
            normGenesExpr <- as.matrix(round(excounts(normGenesExpr)))
        },
        SeqCountSet = {
            # Way to take normalized counts out of DSS as a matrix
            classes <- asClassVector(sampleList)
            theDesign <- data.frame(condition=classes,
                row.names=colnames(normGenes)) 
            cds <- newCountDataSet(as.matrix(round(assayData(normGenes)$exprs)),
                theDesign$condition)
            #DESeq::sizeFactors(cds) <- normalizationFactor(normGenes)
            sizeFactors(cds) <- normalizationFactor(normGenes)
            #normGenes <- as.matrix(round(DESeq::counts(cds,normalized=TRUE)))
            normGenes <- as.matrix(round(counts(cds,normalized=TRUE)))
            cdsExpr <- newCountDataSet(as.matrix(round(assayData(
                normGenesExpr)$exprs)),theDesign$condition)
            #DESeq::sizeFactors(cdsExpr) <- normalizationFactor(normGenesExpr)
            sizeFactors(cdsExpr) <- normalizationFactor(normGenesExpr)
            #normGenesExpr <- as.matrix(round(DESeq::counts(cdsExpr,
            #    normalized=TRUE)))
            normGenesExpr <- as.matrix(round(counts(cdsExpr,
                normalized=TRUE)))
        }
        # We don't need the matrix case
    )

    # Now that everything is a matrix, export the normalized counts if asked
    if (exportCountsTable) {
        geneDataExprExp <- as.data.frame(geneDataExpr)
        geneDataExprExp <- geneDataExprExp[,c(1,2,3,6,7,5,8,9)]
        colnames(geneDataExprExp)[1] <- "chromosome"
        geneDataFilteredExp <- as.data.frame(geneDataFiltered)
        geneDataFilteredExp <- geneDataFilteredExp[,c(1,2,3,6,7,5,8,9)]
        colnames(geneDataFilteredExp)[1] <- "chromosome"
        disp("Exporting and compressing normalized read counts table to ",
            file.path(PROJECT_PATH[["lists"]],"normalized_counts_table.txt"))
        expo <- cbind(
            rbind(geneDataExprExp,geneDataFilteredExp),
            rbind(normGenesExpr,geneCountsZero,geneCountsDead)
        )
        resFile <- file.path(PROJECT_PATH[["lists"]],
            "normalized_counts_table.txt.gz")
        gzfh <- gzfile(resFile,"w")
        write.table(expo,gzfh,sep="\t",row.names=FALSE,quote=FALSE)
        close(gzfh)
    }

    # Calculate meta-statistics, if more than one statistical algorithm has been
    # used
    if (!is.null(contrast) && !any(is.na(statistics))) {
        if (length(statistics)>1) {
            sumpList <- metaTest(
                cpList=cpList,
                metaP=metaP,
                counts=normGenesExpr,
                sampleList=sampleList,
                statistics=statistics,
                statArgs=statArgs,
                libsizeList=libsizeList,
                nperm=nperm,
                weight=weight,
                pOffset=pOffset,
                rc=restrictCores
            )
        }
        # We assign the p-values from the only statistic used to sumpList in
        # order to use it for stat plots
        else
            sumpList <- cpList
        # Useless for one statistics but just for safety
        if ("adj_meta_p_value" %in% exportWhat) 
            adjSumpList <- cmclapply(sumpList,
                function(x,a) return(p.adjust(x,a)),adjustMethod,
                    rc=restrictCores)
        else
            adjSumpList <- NULL
    }
    else
        sumpList <- adjSumpList <- NULL

    ############################################################################
    #                            EXPORT SECTION
    ############################################################################

    if (is.function(.progressFun)) {
        text <- paste("Exporting...")
        .progressFun(detail=text)
    }

    # Bind all the flags
    if (countType=="gene")
        flags <- geneFilterFlags
    else if (countType=="utr")
        flags <- geneFilterFlags
    else if (countType=="exon") {
        if (!is.null(exonFilterFlags)) {
            flags <- cbind(geneFilterFlags,
                as.matrix(exonFilterFlags[rownames(geneFilterFlags),]))
            nams <- c(colnames(geneFilterFlags),colnames(exonFilterFlags))
            rownames(flags) <- rownames(geneFilterFlags)
            colnames(flags) <- nams
        }
        else
            flags <- geneFilterFlags
    }
    
    disp("Building output files...")
    if (outList) 
        out <- makeExportList(contrast)
    else 
        out <- NULL
    if (report) 
        html <- makeExportList(contrast) 
    else 
        html <- NULL
    if ("rpgm" %in% exportScale)
        fa <- attr(geneData,"geneLength")
    else
        fa <- NULL
    if ("normalized" %in% exportValues) {
        fac <- fa[rownames(normGenesExpr)]
        normList <- makeTransformation(normGenesExpr,exportScale,fac,logOffset)
    }
    else
        normList <- NULL
    if ("raw" %in% exportValues) {
        fac <- fa[rownames(geneCountsExpr)]
        rawList <- makeTransformation(geneCountsExpr,exportScale,fac,logOffset)
    }
    else
        rawList <- NULL
    if ("flags" %in% exportWhat)
        goodFlags <- flags[rownames(normGenesExpr),]
    else
        goodFlags <- NULL

    if (!is.null(geneCountsZero) || !is.null(geneCountsDead)) {
        geneCountsFiltered <- rbind(geneCountsZero,geneCountsDead)
        geneCountsUnnormFiltered <- rbind(geneCountsZero,
            geneCountsUnnorm)
        if ("normalized" %in% exportValues) {
            fac <- fa[rownames(geneCountsFiltered)]
            normListFiltered <- makeTransformation(geneCountsFiltered,
                exportScale,fac,logOffset)
        }
        else
            normListFiltered <- NULL
        if ("raw" %in% exportValues) {
            fac <- fa[rownames(geneCountsUnnormFiltered)]
            rawListFiltered <- makeTransformation(
                geneCountsUnnormFiltered,exportScale,fac,logOffset)
        }
        else
            rawListFiltered <- NULL
        if ("flags" %in% exportWhat && !is.null(flags))
            allFlags <- rbind(
                matrix(1,nrow(geneCountsZero),ncol(flags)),
                    flags[rownames(geneCountsDead),]
            )
        else
            allFlags <- NULL
    }
    else {
        geneCountsFiltered <- NULL
        geneCountsUnnormFiltered <- NULL
        allFlags <- NULL
    }
    
    # Export structure for SeqCVIBE
    if (.exportR2C) {
        actLen <- c(attr(geneDataExpr,"geneLength"),
            attr(geneDataFiltered,"geneLength"))
        names(actLen) <- c(names(geneDataExpr),names(geneDataFiltered))
        
        disp("Saving more count data to ",file.path(PROJECT_PATH[["data"]],
            "count_data.RData"))
        .createR2CExport(
            rbind(geneCountsExpr,geneCountsUnnormFiltered),
            rbind(normGenesExpr,geneCountsFiltered),
            c(geneDataExpr,geneDataFiltered),
            actLen,
            libsizeList,
            file.path(PROJECT_PATH[["data"]],"count_data.RData")
        )
    }
    
    if (!is.null(contrast)) {
        reportTables <- vector("list",length(contrast))
        names(reportTables) <- contrast
        
        for (cnt in contrast) {
            disp("  Contrast: ",cnt)
            disp("    Adding non-filtered data...")
            if (!any(is.na(statistics)))
                theExport <- buildExport(
                    geneData=geneDataExpr,
                    rawGeneCounts=geneCountsExpr,
                    normGeneCounts=normGenesExpr,
                    flags=goodFlags,
                    sampleList=sampleList,
                    cnt=cnt,
                    statistics=statistics,
                    rawList=rawList,
                    normList=normList,
                    pMat=cpList[[cnt]],
                    adjpMat=adjCpList[[cnt]],
                    sumP=sumpList[[cnt]],
                    adjSumP=adjSumpList[[cnt]],
                    exportWhat=exportWhat,
                    exportScale=exportScale,
                    exportValues=exportValues,
                    exportStats=exportStats,
                    logOffset=logOffset,
                    #report=report
                    report=FALSE
                )
            else
                theExport <- buildExport(
                    geneData=geneDataExpr,
                    rawGeneCounts=geneCountsExpr,
                    normGeneCounts=normGenesExpr,
                    flags=goodFlags,
                    sampleList=sampleList,
                    cnt=cnt,
                    rawList=rawList,
                    normList=normList,
                    exportWhat=exportWhat,
                    exportScale=exportScale,
                    exportValues=exportValues,
                    exportStats=exportStats,
                    logOffset=logOffset,
                    #report=report
                    report=FALSE
                )
            
            # Adjust the export based on what statistics have been done and a 
            # possible p-value cutoff
            export <- theExport$textTable
            colnames(export)[1] <- "chromosome"
            if (report)
                exportHtml <- theExport$htmlTable
            if (!any(is.na(statistics))) { # Ordering based on p-value
                if (!is.na(pcut)) {
                    if (length(statistics)>1) {
                        switch(metaP,
                            fisher = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            fperm = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            whitlock = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            dperm_min = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            dperm_max = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            dperm_weight = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            minp = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            maxp = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            weight = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            pandora = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            simes = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            },
                            none = {
                                cutInd <- which(sumpList[[cnt]]<=pcut)
                            }
                        )
                        pp <- sumpList[[cnt]][cutInd]
                        export <- export[cutInd,]
                        export <- export[order(pp),]
                        if (report) {
                            exportHtml <- exportHtml[cutInd,]
                            exportHtml <- exportHtml[order(pp),]
                        }
                    }
                    else {
                        cutInd <- which(sumpList[[cnt]]<=pcut)
                        pp <- sumpList[[cnt]][cutInd,]
                        export <- export[cutInd,]
                        export <- export[order(pp),]
                        if (report) {
                            exportHtml <- exportHtml[cutInd,]
                            exportHtml <- exportHtml[order(pp),]
                        }
                    }
                }
                else {
                    pp <- sumpList[[cnt]]
                    export <- export[order(pp),]
                    if (report)
                        exportHtml <- exportHtml[order(pp),]
                }
            }
            
            # Final safety trigger
            naInd <- grep("NA",rownames(export))
            if (length(naInd)>0) {
                export <- export[-naInd,]
                if (report) 
                    exportHtml <- exportHtml[-naInd,]
            }

            resFile <- file.path(PROJECT_PATH[["lists"]],
                paste("metaseqr_sig_out_",cnt,".txt.gz",sep=""))
            disp("    Writing output...")
            gzfh <- gzfile(resFile,"w")
            write.table(export,gzfh,quote=FALSE,row.names=FALSE,sep="\t")
            close(gzfh)
            if (outList)
                out[[cnt]] <- export

            if (!is.null(geneCountsZero) || !is.null(geneCountsDead)) {
                disp("    Adding filtered data...")
                theExportFiltered <- buildExport(
                    geneData=geneDataFiltered,
                    rawGeneCounts=geneCountsUnnormFiltered,
                    normGeneCounts=geneCountsFiltered,
                    flags=allFlags,
                    sampleList=sampleList,
                    cnt=cnt,
                    statistics=statistics,
                    rawList=rawListFiltered,
                    normList=normListFiltered,
                    exportWhat=exportWhat,
                    exportScale=exportScale,
                    exportValues=exportValues,
                    exportStats=exportStats,
                    logOffset=logOffset,
                    report=FALSE
                )

                # Now we should be having theExport and theExportFiltered. We do
                # not generate html output for filtered or total results just a
                # compressed text file. We thus have to append 
                # theExport$textTable and theExportFiltered$htmlTable before
                # writing the final output...
                exportAll <- rbind(theExport$textTable,
                    theExportFiltered$textTable)
                # ...and order them somehow... alphabetically according to row
                # names, as the annotation might not have been bundled...
                exportAll <- exportAll[order(rownames(exportAll)),]
                colnames(exportAll)[1] <- "chromosome"
                
                # Here, both filtered and unfiltered genes are passed to the 
                # output list.
                if (outList)
                    out[[cnt]] <- exportAll   
                
                resFile <- file.path(PROJECT_PATH[["lists"]],paste(
                    "metaseqr_all_out_",cnt,".txt.gz",sep=""))
                disp("    Writing output...")
                gzfh <- gzfile(resFile,"w")
                write.table(exportAll,gzfh,quote=FALSE,row.names=FALSE,sep="\t")
                close(gzfh)
            }
            
            # If report requested, build a more condensed summary table, while 
            # the complete tables are available for download
            if (report) {
                if (!any(is.na(statistics))) {
                    if (length(statistics) > 1) {
                        ew <- c("annotation","meta_p_value","adj_meta_p_value",
                            "fold_change","stats")
                        if (is.null(adjSumpList))
                            adjSumpList <- cmclapply(sumpList,
                                function(x,a) return(p.adjust(x,a)),
                                    adjustMethod,rc=restrictCores)
                    }
                    else {
                        ew <- c("annotation","p_value","adj_p_value",
                            "fold_change","stats")
                        if (is.null(adjCpList))
                            adjCpList <- cmclapply(cpList,
                                function(x,a) return(p.adjust(x,a)),
                                    adjustMethod,rc=restrictCores)
                    }
                }
                else
                    ew <- c("annotation","fold_change","stats")
                
                esc <- "rpgm"
                ev <- "normalized"
                est <- "mean"
                
                faR <- attr(geneDataExpr,"geneLength")          
                facR <- faR[rownames(normGenesExpr)]
                normListR <- makeTransformation(normGenesExpr,esc,facR,
                    logOffset)
                
                disp("    Adding report data...")
                if (!any(is.na(statistics))) {
                    reportTables[[cnt]] <- buildExport(
                        geneData=geneDataExpr,
                        rawGeneCounts=geneCountsExpr,
                        normGeneCounts=normGenesExpr,
                        flags=goodFlags,
                        sampleList=sampleList,
                        cnt=cnt,
                        statistics=statistics,
                        rawList=NULL,
                        normList=normListR,
                        pMat=cpList[[cnt]],
                        adjpMat=as.matrix(adjCpList[[cnt]]),
                        sumP=sumpList[[cnt]],
                        adjSumP=adjSumpList[[cnt]],
                        exportWhat=ew,
                        exportScale=esc,
                        exportValues=ev,
                        exportStats=est,
                        logOffset=logOffset,
                        report=FALSE
                    )$textTable
                    if (is.matrix(pp))
                        npp <- rownames(pp)
                    else
                        npp <- names(pp)
                    reportTables[[cnt]] <- 
                        reportTables[[cnt]][npp[order(pp)],,drop=FALSE]
                    
                    if (!is.null(reportTop)) {
                        plasm <- pcut
                        if (is.na(pcut))
                            plasm <- 0.05
                        
                        topi <- ceiling(reportTop*length(which(pp<=plasm)))
                        if (topi == 0)
                            topi <- 10
                        #topi <- ceiling(reportTop*nrow(reportTables[[cnt]]))
                        reportTables[[cnt]] <- 
                            reportTables[[cnt]][seq_len(topi),,drop=FALSE]
                    }
                }
                else {
                    reportTables[[cnt]] <- buildExport(
                        geneData=geneDataExpr,
                        rawGeneCounts=geneCountsExpr,
                        normGeneCounts=normGenesExpr,
                        flags=goodFlags,
                        sampleList=sampleList,
                        cnt=cnt,
                        rawList=NULL,
                        normList=normListR,
                        exportWhat=ew,
                        exportScale=esc,
                        exportValues=ev,
                        exportStats=est,
                        logOffset=logOffset,
                        report=FALSE
                    )$textTable
                    
                    if (!is.null(reportTop)) {
                        ii <- grep("fold_change_",colnames(reportTables[[cnt]]))
                        forOrder <- as.numeric(reportTables[[cnt]][,ii[1]])
                        oo <- sort(abs(forOrder),decreasing=TRUE,
                            index.return=TRUE)
                        topi <- ceiling(reportTop*length(oo$ix))
                        if (topi == 0)
                            topi <- 10  
                        reportTables[[cnt]] <- 
                            reportTables[[cnt]][oo$ix[seq_len(topi)],,
                                drop=FALSE]
                    }
                }
            }
        }
    }
    else {
        disp("  No contrast was specified, exporting only quantifications")
        disp("    Adding non-filtered data...")
        theExport <- buildExport(
            geneData=geneDataExpr,
            rawGeneCounts=geneCountsExpr,
            normGeneCounts=normGenesExpr,
            flags=goodFlags,
            sampleList=sampleList,
            statistics=statistics,
            rawList=rawList,
            normList=normList,
            exportWhat=exportWhat,
            exportScale=exportScale,
            exportValues=exportValues,
            exportStats=exportStats,
            logOffset=logOffset,
            report=FALSE
        )
        
        export <- theExport$textTable
        colnames(export)[1] <- "chromosome"
        if (report)
            exportHtml <- theExport$htmlTable
        
        naInd <- grep("NA",rownames(export))
        if (length(naInd)>0) {
            export <- export[-naInd,]
            if (report) 
                exportHtml <- exportHtml[-naInd,]
        }

        resFile <- file.path(PROJECT_PATH[["lists"]],"metaseqr_sig_out.txt.gz")
        disp("    Writing output...")
        gzfh <- gzfile(resFile,"w")
        write.table(export,gzfh,quote=FALSE,row.names=FALSE,sep="\t")
        close(gzfh)
        if (outList)
            out[[cnt]] <- export
            
        if (!is.null(geneCountsZero) || !is.null(geneCountsDead)) {
            disp("    Adding filtered data...")
            theExportFiltered <- buildExport(
                geneData=geneDataFiltered,
                rawGeneCounts=geneCountsUnnormFiltered,
                normGeneCounts=geneCountsFiltered,
                flags=allFlags,
                sampleList=sampleList,
                statistics=statistics,
                rawList=rawListFiltered,
                normList=normListFiltered,
                exportWhat=exportWhat,
                exportScale=exportScale,
                exportValues=exportValues,
                exportStats=exportStats,
                logOffset=logOffset,
                report=FALSE
            )

            # Now we should be having theExport and theExportFiltered. We do
            # not generate html output for filtered or total results just a
            # compressed text file. We thus have to append 
            # theExport$textTable and theExportFiltered$htmlTable before
            # writing the final output...
            exportAll <- rbind(theExport$textTable,
                theExportFiltered$textTable)
            # ...and order them somehow... alphabetically according to row
            # names, as the annotation might not have been bundled...
            exportAll <- exportAll[order(rownames(exportAll)),]
            colnames(exportAll)[1] <- "chromosome"
            
            # Here, both filtered and unfiltered genes are passed to the 
            # output list.
            if (outList)
                out[[cnt]] <- exportAll   
            
            resFile <- file.path(PROJECT_PATH[["lists"]],
                "metaseqr_all_out.txt.gz")
            disp("    Writing output...")
            gzfh <- gzfile(resFile,"w")
            write.table(exportAll,gzfh,quote=FALSE,row.names=FALSE,sep="\t")
            close(gzfh)
        }
        
        # If report requested, build a more condensed summary table, while 
        # the complete tables are available for download
        if (report) {
            ew <- c("annotation","stats")
            esc <- "rpgm"
            ev <- "normalized"
            est <- "mean"
            
            faR <- attr(geneDataExpr,"geneLength")          
            facR <- faR[rownames(normGenesExpr)]
            normListR <- makeTransformation(normGenesExpr,esc,facR,
                logOffset)
            
            disp("    Adding report data...")
            reportTables <- list(NT=buildExport(
                geneData=geneDataExpr,
                rawGeneCounts=geneCountsExpr,
                normGeneCounts=normGenesExpr,
                flags=goodFlags,
                sampleList=sampleList,
                statistics=statistics,
                rawList=NULL,
                normList=normListR,
                exportWhat=ew,
                exportScale=esc,
                exportValues=ev,
                exportStats=est,
                logOffset=logOffset,
                report=FALSE
            )$textTable)
            
            # Some kind of toping for the table must be done, most probably
            # based on top expressed genes
            if (!is.null(reportTop)) {
                ii <- grep("rpgm",colnames(reportTables[[1]]))
                forOrder <- as.matrix(reportTables[[1]][,ii])
                avgRpgm <- apply(forOrder,1,mean,na.rm=TRUE)
                oo <- sort(avgRpgm,decreasing=TRUE,index.return=TRUE)
                topi <- ceiling(reportTop*length(oo$ix))
                if (topi == 0)
                    topi <- 10  
                reportTables[[1]] <- 
                    reportTables[[1]][oo$ix[seq_len(topi)],,drop=FALSE]
            }
        }
    }

    ############################################################################
    # END EXPORT SECTION
    ############################################################################

    ############################################################################
    # BEGIN PLOTTING SECTION
    ############################################################################
    
    if (is.function(.progressFun)) {
        text <- paste("Plotting...")
        .progressFun(detail=text)
    }
    
    # Check if we have more than 6 samples in total, pairwise plots are not
    # meaningful
    if (length(unlist(sampleList)) > 6 && "pairwise" %in% qcPlots) {
        warnwrap("Pairwise sample comparison plot becomes indistinguishable ",
            "for more than 6 samples! Removing from plots...")
        qcPlots <- qcPlots[-which(qcPlots == "pairwise")]
    }
    
    if (!is.null(qcPlots)) {
        disp("Creating quality control graphs...")
        plots <- list(
            raw=c("mds","biodetection","countsbio","saturation","readnoise",
                "correl","pairwise"),
            norm=c("boxplot","gcbias","lengthbias","meandiff","meanvar",
                "rnacomp"),
            stat=c("deheatmap","volcano","mastat","biodist","deregulogram"),
            other=c("filtered"),
            venn=c("statvenn","foldvenn")
        )
        figRaw <- figUnorm <- figNorm <- figStat <- figOther <- figVenn <- 
            vector("list",length(figFormat))
        names(figRaw) <- names(figUnorm) <- names(figNorm) <-
            names(figStat) <- names(figOther) <- names(figVenn) <-
            figFormat
        if (is.null(contrast))
            contrastList <- NULL
        for (fig in figFormat) {
            disp("Plotting in ",fig," format...")
            figRaw[[fig]] <- metaseqrPlot(geneCounts,sampleList,
                annotation=geneData,plotType=intersect(qcPlots,plots$raw),
                isNorm=FALSE,output=fig,path=PROJECT_PATH$qc)
            
            figUnorm[[fig]] <- metaseqrPlot(geneCounts,sampleList,
                annotation=geneData,plotType=intersect(qcPlots,plots$norm),
                isNorm=FALSE,output=fig,path=PROJECT_PATH$normalization)
            
            # The annotation dimensions change...
            if (whenApplyFilter=="prenorm")
                figNorm[[fig]] <- metaseqrPlot(normGenes,sampleList,
                    annotation=geneDataExpr,plotType=intersect(qcPlots,
                    plots$norm),isNorm=TRUE,output=fig,
                    path=PROJECT_PATH$normalization) 
            else if (whenApplyFilter=="postnorm")
                figNorm[[fig]] <- metaseqrPlot(normGenes,sampleList,
                    annotation=geneData,plotType=intersect(qcPlots,
                    plots$norm),isNorm=TRUE,output=fig,
                    path=PROJECT_PATH$normalization)
            
            figStat[[fig]] <- metaseqrPlot(normGenesExpr,sampleList,
                annotation=geneDataExpr,contrastList=contrastList,
                pList=sumpList,thresholds=list(p=pcut,f=1),
                plotType=intersect(qcPlots,plots$stat),isNorm=TRUE,
                output=fig,path=PROJECT_PATH$statistics)

            if (!is.null(geneDataFiltered))
                figOther[[fig]] <- metaseqrPlot(geneDataFiltered,
                    sampleList,annotation=totalGeneData,
                    plotType=intersect(qcPlots,plots$other),isNorm=FALSE,
                    output=fig,path=PROJECT_PATH$qc)
            else 
                figOther[[fig]] <- NULL
            
            if ("statvenn" %in% qcPlots)
                figVenn[[fig]] <- metaseqrPlot(normGenesExpr,
                    sampleList,annotation=geneDataExpr,
                    contrastList=contrastList,pList=cpList,
                    thresholds=list(p=pcut,f=1),
                    plotType=intersect(qcPlots,plots$venn),
                    output=fig,path=PROJECT_PATH$statistics)
        }
        
        # Create also tracks if asked, all controls have been performed in the
        # beginning, so targets exist
        if (createTracks)
            tLink <- createSignalTracks(theList,org,stranded=trackInfo$stranded,
                normTo=trackInfo$normTo,urlBase=trackInfo$urlBase,
                exportPath=trackExportPath,hubInfo=trackInfo$hubInfo,
                fasta=trackInfo$fasta,gtf=trackInfo$gtf,
                overwrite=overwriteTracks,rc=restrictCores)
        else
            tLink <- NULL
        
        ########################################################################
        #assign("sampleList",sampleList,envir=parent.frame())
        #assign("geneCounts",geneCounts,envir=parent.frame())
        #assign("normGenes",normGenes,envir=parent.frame())
        #assign("normGenesExpr",normGenes,envir=parent.frame())
        #assign("sumpList",sumpList,envir=parent.frame())
        #assign("cpList",cpList,envir=parent.frame())
        #assign("contrastList",contrastList,envir=parent.frame())
        #assign("geneData",geneData,envir=parent.frame())
        #assign("geneDataExpr",geneDataExpr,envir=parent.frame())
        ########################################################################
    }

    ############################################################################
    # END PLOTTING SECTION
    ############################################################################
    
    ############################################################################
    # BEGIN REPORTING SECTION
    ############################################################################
    
    if (report) {
        obj <- list(
            sampleList=sampleList,
            contrastList=contrastList,
            qcPlots=qcPlots,
            geneCounts=geneCounts,
            geneData=geneData,
            geneDataExpr=geneDataExpr,
            geneDataFiltered=geneDataFiltered,
            totalGeneData=totalGeneData,
            normGenes=normGenes,
            normGenesExpr=normGenesExpr,
            cpList=cpList,
            sumpList=sumpList,
            pcut=pcut,
            metaP=metaP,
            whenApplyFilter=whenApplyFilter,
            PROJECT_PATH=PROJECT_PATH
        )
        if (reportDb == "sqlite")
            .createSqlPlotDb(obj)
        else if (reportDb == "dexie")
            .createDexiePlotDb(obj)
        
        disp("Creating HTML report...")
        
        if (!is.null(qcPlots)) {
            # First create zip archives of the figures
            disp("Compressing figures...")
            zipfiles <- file.path(PROJECT_PATH$plots,paste("metaseqr_figures_",
                figFormat,".zip",sep=""))
            names(zipfiles) <- figFormat
            for (f in figFormat) {
                files <- c(
                    dir(PROJECT_PATH$qc,pattern=paste(".",f,sep=""),
                        full.names=TRUE),
                    dir(PROJECT_PATH$normalization,pattern=paste(".",f,sep=""),
                        full.names=TRUE),
                    dir(PROJECT_PATH$statistics,pattern=paste(".",f,sep=""),
                        full.names=TRUE)
                )
                zip(zipfiles[f],files)
            }
            # Then create the final figure variables which brew will find...
            figRaw <- figRaw[["png"]]
            figUnorm <- figUnorm[["png"]]
            figNorm <- figNorm[["png"]]
            figStat <- figStat[["png"]]
            figOther <- figOther[["png"]]
            figVenn <- figVenn[["png"]]
        }
        else
            figRaw <- figUnorm <- figNorm <- figStat <- figOther <-
                figVenn <- NULL
        
        # Then see what is going of if default report changed
        if (tolower(reportTemplate)=="default") {
            if (exists("TEMPLATE")) {
                reportTemplate=list(
                    rmd=file.path(TEMPLATE,"metaseqr2_report.Rmd"),
                    loader=file.path(TEMPLATE,"dna_loader.gif")
                )
            }
            else
                reportTemplate=list(rmd=NULL,loader=NULL)
        }
        if (!is.null(reportTemplate$rmd)) {
            if (file.exists(reportTemplate$rmd)) {
                template <- reportTemplate$rmd
                hasTemplate <- TRUE
            }
            else {
                warnwrap(paste("The template file",reportTemplate$rmd,
                    "was not ","found! The HTML report will NOT be generated."))
                hasTemplate <- FALSE
            }
        }
        else {
            warnwrap(paste("The report option was enabled but no template ",
                "file is provided! The HTML report will NOT be generated."))
            hasTemplate <- FALSE
        }
        if (!is.null(reportTemplate$loader)) {
            if (file.exists(reportTemplate$loader))
                file.copy(from=reportTemplate$loader,to=PROJECT_PATH$media)
            else
                warnwrap(paste("The report loader image",reportTemplate$loader,
                    "was not found!"))
        }
        else
            warnwrap(paste("The report loader image was not provided!"))
        
        # Here we must download all required libraries and put them in the js
        # folder of the report to make it available offline
        if (offlineReport) {
            .downloadJsLibs(PROJECT_PATH,reportDb)
            paceHeader <- paste0("<script type=\"text/javascript\" ",
                "src=\"js/pace.min.js\"></script>")
        }
        else
            paceHeader <- paste0("<script type=\"text/javascript\" ",
                "src=\"https://raw.github.com/HubSpot/pace/v1.0.0/",
                "pace.min.js\"></script>")
        
        if (hasTemplate) {
            execTime <- elap2human(TB)
            REPORT_ENV <- .makeReportEnv(environment())
            
            file.copy(file.path(TEMPLATE,"metaseqr2_report.Rmd"),
                file.path(PROJECT_PATH$main,"metaseqr2_report.Rmd"),
                overwrite=TRUE)
            invisible(knitr::knit_meta(class=NULL,clean=TRUE))
            
            render(
                input=file.path(PROJECT_PATH$main,"metaseqr2_report.Rmd"),
                output_file="index.html",
                output_dir=PROJECT_PATH$main,
                envir=REPORT_ENV
            )
            gc(verbose=FALSE)
            
            # Remove the Rmd file after rendering the report
            unlink(file.path(PROJECT_PATH$main,"metaseqr2_report.Rmd"))
            
            ####################################################################
            # Replace a jQuery incompatibility bug line.. Hopefully to remove...
            # Also add pace.js in header on the top... header-includes does not
            # work...
            L <- readLines(file.path(PROJECT_PATH$main,"index.html"))
            # pace.js
            hi <- grep("<head>",L)
            if (length(hi) > 0)
                L[hi] <- paste0(L[hi],"\n",paceHeader)
            # jQuery bug
            ln <- grep(paste("$(\".menu\").find(\"li[data-target=\" +",
                "window.page + \"]\").trigger(\"click\");"),L,fixed=TRUE)
            if (length(ln) == 1)
                L[ln] <- paste("$(\".menu\").find(\"li[data-target='\" +",
                "window.page + \"']\").trigger(\"click\");")
            cat(L,file=file.path(PROJECT_PATH$main,"index.html"),sep="\n")
            ####################################################################
        }
    }

    ############################################################################
    # END REPORTING SECTION
    ############################################################################
    
    disp("\n",strftime(Sys.time()),": Data processing finished!\n")
    execTime <- elap2human(TB)
    disp("\n","Total processing time: ",execTime,"\n\n")

    if (outList) {
        tmp <- c(geneDataExpr,geneDataFiltered)
        a <- c(attr(geneDataExpr,"geneLength"),
            attr(geneDataFiltered,"geneLength"))
        names(a) <- rownames(tmp)
        attr(tmp,"geneLength") <- a
        if (!is.null(cpList)) {
            for (n in names(cpList)) {
                if (!is.null(geneDataFiltered)) {
                    filler <- matrix(NA,length(geneDataFiltered),
                        ncol(cpList[[n]]))
                    rownames(filler) <- names(geneDataFiltered)
                    colnames(filler) <- colnames(cpList[[n]])
                }
                else {
                    filler <- NULL
                    cpList[[n]] <- rbind(cpList[[n]],filler)
                    cpList[[n]] <- cpList[[n]][rownames(tmp),,drop=FALSE]
                }
            }
        }
        if (!is.null(adjCpList)) {
            for (n in names(adjCpList)) {
                if (!is.null(geneDataFiltered)) {
                    filler <- matrix(NA,length(geneDataFiltered),
                        ncol(adjCpList[[n]]))
                    rownames(filler) <- names(geneDataFiltered)
                    colnames(filler) <- colnames(cpList[[n]])
                }
                else
                    filler <- NULL
                adjCpList[[n]] <- rbind(adjCpList[[n]],filler)
                adjCpList[[n]] <- adjCpList[[n]][rownames(tmp),,drop=FALSE]
            }
        }
        if (!is.null(sumpList)) {
            for (n in names(sumpList)) {
                if (!is.null(geneDataFiltered)) {
                    filler <- rep(NA,length(geneDataFiltered))
                    names(filler) <- names(geneDataFiltered)
                }
                else
                    filler <- NULL
                if (is.matrix(sumpList[[n]])) {
                    # The following if-else was added because some runs had 
                    # filler=NULL while others do not. Sometimes this created a 
                    # problem and the simulations stopped. Now fixed.
                    if (is.null(filler)) {
                        sumpList[[n]] <- rbind(sumpList[[n]],filler)
                        sumpList[[n]] <- sumpList[[n]][rownames(tmp),,
                            drop=FALSE]
                    }
                    else {
                        filler=as.matrix(filler)
                        sumpList[[n]] <- rbind(sumpList[[n]],filler)
                        sumpList[[n]] <- sumpList[[n]][rownames(tmp),,
                            drop=FALSE]
                    }
                }
                else {
                    sumpList[[n]] <- c(sumpList[[n]],filler)
                    sumpList[[n]] <- sumpList[[n]][rownames(tmp)]
                }
            }
        }
        if (!is.null(adjSumpList)) {
            for (n in names(adjSumpList)) {
                if (!is.null(geneDataFiltered)) {
                    filler <- rep(NA,length(geneDataFiltered))
                    names(filler) <- names(geneDataFiltered)
                }
                else
                    filler <- NULL
                adjSumpList[[n]] <- c(adjSumpList[[n]],filler)
                adjSumpList[[n]] <- adjSumpList[[n]][rownames(tmp)]
            }
        }

        complete <- list(
            #call=as.list(match.call()),
            params=list(
                sampleList=sampleList,
                excludeList=excludeList,
                fileType=fileType,
                path=path,
                contrast=contrast,
                libsizeList=libsizeList,
                embedCols=embedCols,
                annotation=annotation,
                org=org,
                refdb=refdb,
                countType=countType,
                exonFilters=exonFilters,
                geneFilters=geneFilters,
                whenApplyFilter=whenApplyFilter,
                normalization=normalization,
                normArgs=normArgs,
                statistics=statistics,
                statArgs=statArgs,
                adjustMethod=adjustMethod,
                metaP=metaP,
                weight=weight,
                nperm=nperm,
                pcut=pcut,
                logOffset=logOffset,
                preset=preset,
                qcPlots=qcPlots,
                figFormat=figFormat,
                outList=outList,
                exportWhere=exportWhere,
                exportWhat=exportWhat,
                exportScale=exportScale,
                exportValues=exportValues,
                exportStats=exportStats,
                exportCountsTable=exportCountsTable,
                restrictCores=restrictCores,
                report=report,
                reportTop=reportTop,
                reportTemplate=reportTemplate,
                saveGeneModel=saveGeneModel,
                verbose=verbose,
                runLog=runLog,
                offlineReport=offlineReport,
                createTracks=createTracks,
                trackExportPath=trackExportPath,
                trackInfo=trackInfo
            ),
            filterCutoffs=list(
                exonFilter=list(
                    minActiveExons=NULL
                ),
                geneFilter=list(
                    length=geneFilters$length$length,
                    avgReads=if (is.null(geneFilterCutoff)) NULL else
                        round(geneFilterCutoff$avgReads,digits=5),
                    expression=list(
                        median=geneFilterCutoff$expression$median,
                        mean=geneFilterCutoff$expression$mean,
                        quantile=geneFilterCutoff$expression$quantile,
                        known=geneFilterCutoff$expression$known,
                        custom=geneFilterCutoff$expression$custom
                    ),
                    biotype=if (is.null(geneFilters$biotype)) NULL else
                        paste(names(geneFilters$biotype)[which(
                            unlist(geneFilters$biotype))],collapse=", ")
                ),
                zeroFiltered=length(theZeros),
                exonFiltered=length(unique(unlist(exonFilterResult))),
                geneFiltered=length(unique(unlist(geneFilterResult))),
                totalFiltered=length(theZeros)+length(theDead)
            ),
            geneData=tmp,
            rawCounts=rbind(geneCountsExpr,geneCountsUnnormFiltered),
            normCounts=rbind(normGenesExpr,geneCountsFiltered),
            flags=rbind(goodFlags,allFlags),
            sampleList=sampleList,
            contrastList=contrast,
            pValue=cpList,
            fdr=adjCpList,
            metaPValue=sumpList,
            metaFdr=adjSumpList
        )

        return(list(data=out,html=html,complete=complete))
    }
    
} # End metaseqr2


constructGeneModel <- function(countData,annoData,type,rc=NULL) {
    # countData is a matrix, must have exact rownames as annoData GRanges
    if (!all(rownames(countData) == names(annoData)))
        stop("The rownames of the provided counts are not the same (or in ",
            " the same order) as the provided annotation GRanges!")
    
    # Messaging...
    msg <- "provided count regions"
    if (!missing(type)) {
        if (type == "exon")
            msg <- "exons"
        else if (type == "utr")
            msg <- "transcripts (UTR regions)"
    }
    
    # Create splitting factor
    theGenes <- factor(annoData$gene_id,levels=unique(annoData$gene_id))
    
    # Split the counts vectors...
    theCounts <- cmclapply(colnames(countData),function(n,f,M) {
        disp("  Separating ",msg," per gene for ",n,"...")
        co <- M[,n]
        names(co) <- rownames(M)
        return(split(co,f))
    },theGenes,countData,rc=rc)
    names(theCounts) <- colnames(countData)
    
    # and the width of the merged exons
    w <- width(annoData)
    if (type == "exon")
        names(w) <- annoData$exon_id
    else if (type == "utr")
        names(w) <- annoData$transcript_id
    lengthList <- split(w,theGenes)
    attr(theCounts,"lengthList") <- lengthList
    # The exon lengths information is essentially redundant in the new version,
    # unless we need it for some QC distribution later on...
    
    return(theCounts)
}

.reduceGeneData <- function(exonData,geneData) {
    exonChrs <- unique(as.character(seqnames(exonData)))
    geneChrs <- unique(as.character(seqnames(geneData)))
    if (length(exonChrs) != length(geneChrs)) {
        m <- as.integer(match(seqnames(geneData),exonChrs))
        geneData <- geneData[which(!is.na(m))]
    }
    return(geneData)
}

.userOrg <- function(org,db=NULL) {
    ua <- getUserAnnotations(db)
    if (nrow(ua) == 0)
        return(FALSE)
    orgs <- unique(ua$organism)
    return(org %in% orgs)
}

.userRefdb <- function(refdb,db=NULL) {
    us <- getUserAnnotations(db)
    if (nrow(us) == 0)
        return(FALSE)
    refdbs <- unique(us$source)
    return(refdb %in% refdbs)
}

.createR2CExport <- function(raw,norm,gr,al,lib,output) {
    # Ensure the same order
    nameOrder <- names(gr)[order(names(gr))]
    
    # Reorder
    raw <- raw[nameOrder,]
    norm <- norm[nameOrder,]
    gr <- gr[nameOrder]
    al <- al[nameOrder]
    
    # Active length and annotation data frame
    attr(gr,"geneLength") <- al
    ann <- as.data.frame(gr)
    ann <- ann[,c(1,2,3,6,7,5,8,9)]
    colnames(ann)[1] <- "chromosome"
    
    # Create the list to be exported, naming violation with dot(.) in order
    # not to break current SeqCVIBE... Will correct in the future.
    b2c.out <- list(
        counts=raw,
        norm=norm,
        annotation=ann,
        length=al,
        libsize=lib
    )
    save(b2c.out,file=output,compress=TRUE)
}

.backwardsCompatibility <- function(dataFile) {
    the.counts <- count.type <- sample.list <- gene.data <- transcript.data <-
        exon.data <- NULL
    tmpEnv <- new.env()
    load(dataFile,tmpEnv)
    if (!is.null(tmpEnv$the.counts)) { # Old file
        # Copy to new
        tmpEnv$theCounts <- tmpEnv$the.counts
        tmpEnv$countsType <- tmpEnv$count.type
        tmpEnv$sampleList <- tmpEnv$sampleList
        tmpEnv$geneData <- tmpEnv$gene.data
        if (!is.null(tmpEnv$transcript.data))
            tmpEnv$transcriptData <- tmpEnv$transcript.data
        if (!is.null(tmpEnv$exon.data))
            tmpEnv$exonData <- tmpEnv$exon.data
        
        # Delete old
        if (exists("the.counts",envir=tmpEnv)) {
            rm(the.counts,count.type,sample.list,gene.data,envir=tmpEnv)
            if (!is.null(tmpEnv$transcript.data))
                rm(transcript.data,envir=tmpEnv)
            if (!is.null(tmpEnv$exon.data))
                rm(exon.data,envir=tmpEnv)
        }        
    }
    return(tmpEnv)
}

.excludeSamplesFromList <- function(L,s) {
    gl <- attr(L,"lengthList")
    L <- L[s]
    attr(L,"lengthList") <- gl
    return(L)
}

test_metaseqr <- function() {
    data("mm9GeneData",package="metaseqR2")
    exDir <- tempdir()

    result1 <- metaseqr2(
        counts=mm9GeneCounts,
        sampleList=sampleListMm9,
        contrast=c("adult_8_weeks_vs_e14.5"),
        libsizeList=libsizeListMm9,
        annotation="embedded",
        org="mm9",
        countType="gene",
        normalization="edger",
        statistics=c("edger","limma"),
        metaP="simes",
        preset="medium_basic",
        qcPlots="mds",
        figFormat="png",
        exportWhere=exDir,
        outList=TRUE,
        restrictCores=0.01
    )
    checkTrue(file.exists(file.path(exDir,"index.html")))
    checkTrue(file.exists(file.path(exDir,"plots","qc","mds.png")))
    checkTrue(file.exists(file.path(exDir,"lists")))
    checkTrue(nrow(result1[[1]][[1]])>0)
    checkEqualsNumeric(ncol(result1[[1]][[1]]),16)
}

getUcscTableNameUtr <- function(org,refdb) {
    switch(refdb,
        ucsc = {
            switch(org,
                hg18 = {
                    return("knownGene")
                },
                hg19 = {
                    return("knownGene")
                },
                hg38 = {
                    return("knownGene")
                },
                mm9 = {
                    return("knownGene")
                },
                mm10 = {
                    return("knownGene")
                },
                mm39 = {
                    return("knownGene")
                },
                rn5 = {
                    return("mgcGenes")
                },
                rn6 = {
                    return("mgcGenes")
                },
                dm3 = {
                    return("flyBaseGene")
                },
                dm6 = {
                    warning("No UCSC Genome annotation for Drosophila ",
                        "melanogaster v6! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                danrer7 = {
                    return("mgcGenes")
                },
                danrer10 = {
                    return("mgcGenes")
                    
                },
                danrer11 = {
                    warning("No UCSC Genome annotation for Danio rerio v11! ",
                        "Will use RefSeq instead...",immediate.=TRUE)
                    return("refGene")
                },
                pantro4 = {
                    warning("No UCSC Genome annotation for Pan ",
                        "troglodytes! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                pantro5 = {
                    warning("No UCSC Genome annotation for Pan ",
                        "troglodytes! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                susscr3 = {
                    warning("No UCSC Genome annotation for Sus ",
                        "scrofa v3! Will use RefSeq instead...",
                        immediate.=TRUE)
                        return("refGene")
                },
                susscr11 = {
                    warning("No UCSC Genome annotation for Sus ",
                        "scrofa v11! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                },
                equcab3 = {
                    warning("No UCSC Genome annotation for Equus ",
                        "caballus v3! Will use RefSeq instead...",
                        immediate.=TRUE)
                    return("refGene")
                }
            )
        },
        refseq = {
            return("refGene")
        }
    )
}

getUcscDbl <- function(org,refdb="ucsc") {
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    checkTextArgs("refdb",refdb,getSupportedUcscDbs())
    
    if (!requireNamespace("RSQLite"))
        stop("R package RSQLite is required to use annotation from UCSC!")

    httpBase <- paste("http://hgdownload.soe.ucsc.edu/goldenPath/",
        getUcscOrganism(org),"/database/",sep="")
    tableDefs <- getUcscTabledef(org,refdb,"fields")
    fileList <- vector("list",length(tableDefs))
    names(fileList) <- names(tableDefs)
    for (n in names(fileList))
        fileList[[n]] <- paste(httpBase,n,".txt.gz",sep="")
        
    # Fill the fields for each table
    drv <- dbDriver("SQLite")
    dbTmp <- tempfile()
    con <- dbConnect(drv,dbname=dbTmp)
    message("  Retrieving tables for temporary SQLite ",refdb," ",org,
        " subset database")
    for (n in names(fileList)) {
        message("    Retrieving table ",n)
        download.file(fileList[[n]],file.path(tempdir(),
            paste(n,".txt.gz",sep="")),quiet=TRUE)
        sqlDf <- read.delim(file.path(tempdir(),paste(n,".txt.gz",sep="")),
            row.names=NULL,header=FALSE,strip.white=TRUE)
        names(sqlDf) <- tableDefs[[n]]
        dbWriteTable(con,n,sqlDf,row.names=FALSE)
    }
    dbDisconnect(con)
    return(dbTmp)
}

getUcscTabledef <- function(org,refdb="ucsc",what="queries") {
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    what <- tolower(what[1])
    checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    checkTextArgs("refdb",refdb,getSupportedUcscDbs())
    checkTextArgs("what",what,c("queries","fields"))
    switch(refdb,
        ucsc = {
            return(.getUcscTabledefUcsc(org,what))
        },
        refseq = {
            return(.getUcscTabledefRefseq(org,what))
        }
    )
}

getUcscTblTpl <- function(tab,what="queries") {
    if (what=="queries") {
        switch(tab,
            knownCanonical = {
                return(paste(
                    "CREATE TABLE",
                    "`knownCanonical` (",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`chromStart` INTEGER NOT NULL DEFAULT '0',",
                    "`chromEnd` INTEGER NOT NULL DEFAULT '0',",
                    "`clusterId` INTEGER NOT NULL DEFAULT '0',",
                    "`transcript` TEXT NOT NULL DEFAULT '',",
                    "`protein` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            knownGene = {
                return(paste(
                    "CREATE TABLE",
                    "`knownGene` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`strand` TEXT NOT NULL DEFAULT '',",
                    "`txStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`txEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonCount` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL,",
                    "`proteinID` TEXT NOT NULL DEFAULT '',",
                    "`alignID` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            knownToRefSeq = {
                return(paste(
                    "CREATE TABLE",
                    "`knownToRefSeq` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`value` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            refFlat = {
                return(paste("CREATE TABLE",
                    "`refFlat` (",
                    "`geneName` TEXT NOT NULL,",
                    "`name` TEXT NOT NULL,",
                    "`chrom` TEXT NOT NULL,",
                    "`strand` TEXT NOT NULL,",
                    "`txStart` UNSIGNED INTEGER NOT NULL,",
                    "`txEnd` UNSIGNED INTEGER NOT NULL,",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL,",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL,",
                    "`exonCount` UNSIGNED INTEGER NOT NULL,",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            refGene = {
                return(paste("CREATE TABLE",
                    "`refGene` (",
                    "`bin` UNSIGNED INTEGER NOT NULL,",
                    "`name` TEXT NOT NULL,",
                    "`chrom` TEXT NOT NULL,",
                    "`strand` TEXT NOT NULL,",
                    "`txStart` UNSIGNED INTEGER NOT NULL,",
                    "`txEnd` UNSIGNED INTEGER NOT NULL,",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL,",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL,",
                    "`exonCount` UNSIGNED INTEGER NOT NULL,",
                    "`exonStarts` TEXT NOT NULL,",
                    "`score` INTEGER NOT NULL,",
                    "`exonEnds` TEXT NOT NULL",
                    "`name2` TEXT NOT NULL,",
                    "`cdsStartStat` TEXT NOT NULL,",
                    "`cdsEndStat` TEXT NOT NULL,",
                    "`exonFrames` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            knownToEnsembl = {
                return(paste(
                    "CREATE TABLE",
                    "`knownToEnsembl` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`value` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            ensemblSource = {
                return(paste(
                    "CREATE TABLE",
                    "`ensemblSource` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`source` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            mgcGenes = {
                return(paste(
                    "CREATE TABLE `mgcGenes` (",
                    "`bin` UNSIGNED INTEGER NOT NULL,",
                    "`name` TEXT NOT NULL,",
                    "`chrom` TEXT NOT NULL,",
                    "`strand` TEXT NOT NULL,",
                    "`txStart` UNSIGNED INTEGER NOT NULL,",
                    "`txEnd` UNSIGNED INTEGER NOT NULL,",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL,",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL,",
                    "`exonCount` UNSIGNED INTEGER NOT NULL,",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL,",
                    "`score` INTEGER DEFAULT NULL,",
                    "`name2` TEXT NOT NULL,",
                    "`cdsStartStat` TEXT NOT NULL,",
                    "`cdsEndStat` TEXT NOT NULL,",
                    "`exonFrames` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            ensemblToGeneName = {
                return(paste(
                    "CREATE TABLE",
                    "`knownToGeneName` (",
                    "`name` TEXT NOT NULL,",
                    "`value` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            flyBaseCanonical = {
                return(paste(
                    "CREATE TABLE",
                    "`flyBaseCanonical` (",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`chromStart` INTEGER NOT NULL DEFAULT '0',",
                    "`chromEnd` INTEGER NOT NULL DEFAULT '0',",
                    "`clusterId` INTEGER unsigned NOT NULL DEFAULT '0',",
                    "`transcript` TEXT NOT NULL DEFAULT '',",
                    "`protein` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            },
            flyBaseGene = {
                return(paste(
                    "CREATE TABLE",
                    "`flyBaseGene` (",
                    "`bin` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`chrom` TEXT NOT NULL DEFAULT '',",
                    "`strand` TEXT NOT NULL DEFAULT '',",
                    "`txStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`txEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsStart` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`cdsEnd` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonCount` UNSIGNED INTEGER NOT NULL DEFAULT '0',",
                    "`exonStarts` TEXT NOT NULL,",
                    "`exonEnds` TEXT NOT NULL",
                    ")",collapse=" "
                ))
            },
            flyBaseToRefSeq = {
                return(paste(
                    "CREATE TABLE",
                    "`flyBaseToRefSeq` (",
                    "`name` TEXT NOT NULL DEFAULT '',",
                    "`value` TEXT NOT NULL DEFAULT ''",
                    ")",collapse=" "
                ))
            }
        )
    }
    else if (what=="fields") {
        switch(tab,
            knownCanonical = {
                return(c("chrom","chromStart","chromEnd","clusterId",
                "transcript","protein"))
            },
            knownGene = {
                return(c("name","chrom","strand","txStart","txEnd","cdsStart",
                    "cdsEnd","exonCount","exonStarts","exonEnds","proteinID",
                    "alignID"))
            },
            knownToRefSeq = {
                return(c("name","value"))
            },
            refFlat = {
                return(c("geneName","name","chrom","strand","txStart","txEnd",
                    "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds"))
            },
            refGene = {
                return(c("bin","name","chrom","strand","txStart","txEnd",
                "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds","score",
                "name2","cdsStartStat","cdsEndStat","exonFrames"))
            },
            knownToEnsembl = {
                return(c("name","value"))
            },
            ensemblSource = {
                return(c("name","source"))
            },
            mgcGenes = {
                return(c("bin","name","chrom","strand","txStart","txEnd",
                    "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds",
                    "score","name2","cdsStartStat","cdsEndStat","exonFrames"
                ))
            },
            ensemblToGeneName = {
                return(c("name","value"))
            },
            flyBaseCanonical = {
                return(c("chrom","chromStart","chromEnd","clusterId",
                    "transcript","protein"))
            },
            flyBaseGene = {
                return(c("bin","name","chrom","strand","txStart","txEnd",
                    "cdsStart","cdsEnd","exonCount","exonStarts","exonEnds"))
            },
            flyBaseToRefSeq = {
                return(c("name","value"))
            }
        )
    }
}

getUcscQuery <- function(org,type,refdb="ucsc") {
    type <- tolower(type[1])
    org <- tolower(org[1])
    refdb <- tolower(refdb[1])
    checkTextArgs("type",type,c("gene","exon","transcript"))
    checkTextArgs("org",org,getSupportedOrganisms(),multiarg=FALSE)
    checkTextArgs("refdb",refdb,c("ucsc","refseq"))
    switch(type,
        gene = {
            switch(refdb,
                ucsc = {
                    return(.getUcscQueryUcscGene(org))
                },
                refseq = {
                    return(.getUcscQueryRefseqGene(org))
                }
            )
        },
        exon = {
            switch(refdb,
                ucsc = {
                    return(.getUcscQueryUcscExon(org))
                },
                refseq = {
                    return(.getUcscQueryRefseqExon(org))                    
                }
            )
        },
        transcript = {
            switch(refdb,
                ucsc = {
                    return(.getUcscQueryUcscTranscript(org))
                },
                refseq = {
                    return(.getUcscQueryRefseqTranscript(org))
                }
            )
        }
    )
}

.getUcscTabledefUcsc <- function(org,what="queries") {
    switch(org,
        hg18 = {
            return(list(
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownGene=getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                refFlat=getUcscTblTpl("refFlat",what)
            ))
        },
        hg19 = {
            return(list(
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownGene=getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what),
                refFlat=getUcscTblTpl("refFlat",what)
            ))
        },
        hg38 = {
            return(list(
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownGene=getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                refFlat=getUcscTblTpl("refFlat",what)
            ))
        },
        mm9 = {
            return(list(
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownGene=getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what),
                refFlat=getUcscTblTpl("refFlat",what)
            ))
        },
        mm10 = {
            return(list(
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownGene=getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what),
                refFlat=getUcscTblTpl("refFlat",what)
            ))
        },
        mm39 = {
            return(list(
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownGene=getUcscTblTpl("knownGene",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what),
                refFlat=getUcscTblTpl("refFlat",what)
            ))
        },
        rn5 = {
            return(list(
                mgcGenes=getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        rn6 = {
            return(list(
                mgcGenes=getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        dm3 = {
            return(list(
                flyBaseCanonical=
                    getUcscTblTpl("flyBaseCanonical",what),
                flyBaseGene=
                    getUcscTblTpl("flyBaseGene",what),
                flyBaseToRefSeq=
                    getUcscTblTpl("flyBaseToRefSeq",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what),
                refFlat=getUcscTblTpl("refFlat",what)
            ))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer7 = {
            return(list(
                mgcGenes=getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer10 = {
            return(list(
                mgcGenes=getUcscTblTpl("mgcGenes",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        #pantro6 = {
        #    warning("No UCSC Genome annotation for Pan ",
        #        "troglodytes! Will use RefSeq instead...",
        #        immediate.=TRUE)
        #    return(list(
        #        refFlat=getUcscTblTpl("refFlat",what),
        #        ensemblToGeneName=
        #            getUcscTblTpl("ensemblToGeneName",what),
        #        ensemblSource=
        #            getUcscTblTpl("ensemblSource",what)
        #    ))
        #},
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        }
    )
}

.getUcscTabledefRefseq <- function(org,what="queries") {
    switch(org,
        hg18 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what)
            ))
        },
        hg19 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        hg38 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what)
            ))
        },
        mm9 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        mm10 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        mm39 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                knownToRefSeq=
                    getUcscTblTpl("knownToRefSeq",what),
                knownCanonical=
                    getUcscTblTpl("knownCanonical",what),
                knownToEnsembl=
                    getUcscTblTpl("knownToEnsembl",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        rn5 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        rn6 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        dm3 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        dm6 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer7 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        danrer10 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro4 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        pantro5 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        #pantro6 = {
        #    return(list(
        #        refFlat=getUcscTblTpl("refFlat",what),
        #        ensemblToGeneName=
        #            getUcscTblTpl("ensemblToGeneName",what),
        #        ensemblSource=
        #            getUcscTblTpl("ensemblSource",what)
        #    ))
        #},
        susscr3 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        susscr11 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        },
        equcab3 = {
            return(list(
                refFlat=getUcscTblTpl("refFlat",what),
                ensemblToGeneName=
                    getUcscTblTpl("ensemblToGeneName",what),
                ensemblSource=
                    getUcscTblTpl("ensemblSource",what)
            ))
        }
    )
}

.getUcscQueryUcscGene <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY gene_id",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should have been like hg19 but it's like hg18
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY gene_id",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
            #   "`chromStart` AS `start`,",
            #   "`chromEnd` AS `end`,",
            #   "`transcript` AS `gene_id`,",
            #   "0 AS `gc_content`,",
            #   "knownGene.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownCanonical` INNER JOIN `knownGene`", 
            #   "ON knownCanonical.transcript=knownGene.name",
            #   "INNER JOIN `knownToRefSeq`", 
            #   "ON knownCanonical.transcript=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat`",
            #   "ON knownToRefSeq.value=refFlat.name",
            #   "GROUP BY `gene_id`",
            #   "ORDER BY `chromosome`,`start`"))
            ## No Ensembl source...
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY gene_id",
                "ORDER BY `chromosome`,`start`"))
        },
        mm39 = {
            #return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
            #   "`chromStart` AS `start`,",
            #   "`chromEnd` AS `end`,",
            #   "`transcript` AS `gene_id`,",
            #   "0 AS `gc_content`,",
            #   "knownGene.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownCanonical` INNER JOIN `knownGene`", 
            #   "ON knownCanonical.transcript=knownGene.name",
            #   "INNER JOIN `knownToRefSeq`", 
            #   "ON knownCanonical.transcript=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat`",
            #   "ON knownToRefSeq.value=refFlat.name",
            #   "GROUP BY `gene_id`",
            #   "ORDER BY `chromosome`,`start`"))
            ## No Ensembl source...
            return(paste("SELECT knownCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownCanonical` INNER JOIN `knownGene`",
                "ON knownCanonical.transcript=knownGene.name",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY gene_id",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm3 = {
            return(paste("SELECT flyBaseCanonical.chrom AS `chromosome`,",
                "`chromStart` AS `start`,",
                "`chromEnd` AS `end`,",
                "`transcript` AS `gene_id`,",
                "0 AS `gc_content`,",
                "flyBaseGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `flyBaseCanonical` INNER JOIN `flyBaseGene`", 
                "ON flyBaseCanonical.transcript=flyBaseGene.name",
                "INNER JOIN `flyBaseToRefSeq`",
                "ON flyBaseCanonical.transcript=flyBaseToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON flyBaseToRefSeq.value=refFlat.name",
                "INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`", 
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
            #    "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "0 AS `gc_content`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            warning("No UCSC Genome annotation for Danio rerio v11! Will use ",
                "RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan troglodytes v4! Will ",
                "use RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v5! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

.getUcscQueryRefseqGene <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT refFlat.chrom AS `chromosome`,",
            #   "refFlat.txStart AS `start`,",
            #   "refFlat.txEnd AS `end`,",
            #   "refFlat.name AS `gene_id`,",
            #   "0 AS `gc_content`,",
            #   "refFlat.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `refFlat` INNER JOIN `knownToRefSeq`",
            #   "ON refFlat.name=knownToRefSeq.value",
            #   "INNER JOIN `knownCanonical`",
            #   "ON knownToRefSeq.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "GROUP BY refFlat.name",
            #   "ORDER BY `chromosome`,`start`"))
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm39 = {
            #return(paste("SELECT refFlat.chrom AS `chromosome`,",
            #   "refFlat.txStart AS `start`,",
            #   "refFlat.txEnd AS `end`,",
            #   "refFlat.name AS `gene_id`,",
            #   "0 AS `gc_content`,",
            #   "refFlat.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `refFlat` INNER JOIN `knownToRefSeq`",
            #   "ON refFlat.name=knownToRefSeq.value",
            #   "INNER JOIN `knownCanonical`",
            #   "ON knownToRefSeq.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "GROUP BY refFlat.name",
            #   "ORDER BY `chromosome`,`start`"))
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`gene_id`,",
                "`gc_content`,`strand`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "refFlat.name AS `gene_id`,",
                "0 AS `gc_content`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

.getUcscQueryUcscExon <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`", 
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`",
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT knownGene.chrom AS `chromosome`,",
            #   "knownGene.exonStarts AS `start`,",
            #   "knownGene.exonEnds AS `end`,",
            #   "knownGene.name AS `exon_id`,",
            #   "knownGene.strand AS `strand`,",
            #   "`transcript` AS `gene_id`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownGene` INNER JOIN `knownCanonical`",
            #   "ON knownGene.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToRefSeq`",
            #   "ON knownCanonical.transcript=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat`",
            #   "ON knownToRefSeq.value=refFlat.name",
            #   "GROUP BY knownGene.name",
            #   "ORDER BY `chromosome`,`start`"))
            ## No Ensembl source...
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm39 = {
            #return(paste("SELECT knownGene.chrom AS `chromosome`,",
            #   "knownGene.exonStarts AS `start`,",
            #   "knownGene.exonEnds AS `end`,",
            #   "knownGene.name AS `exon_id`,",
            #   "knownGene.strand AS `strand`,",
            #   "`transcript` AS `gene_id`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownGene` INNER JOIN `knownCanonical`",
            #   "ON knownGene.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToRefSeq`",
            #   "ON knownCanonical.transcript=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat`",
            #   "ON knownToRefSeq.value=refFlat.name",
            #   "GROUP BY knownGene.name",
            #   "ORDER BY `chromosome`,`start`"))
            ## No Ensembl source...
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.exonStarts AS `start`,",
                "knownGene.exonEnds AS `end`,",
                "knownGene.name AS `exon_id`,",
                "knownGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownCanonical`", 
                "ON knownGene.name=knownCanonical.transcript",
                "INNER JOIN `knownToRefSeq`",
                "ON knownCanonical.transcript=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`", 
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn6 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
            #    "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`" ,
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`", 
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm3 = {
            return(paste("SELECT flyBaseCanonical.chrom AS `chromosome`,",
                "flyBaseGene.exonStarts AS `start`,",
                "flyBaseGene.exonEnds AS `end`,",
                "`transcript` AS `exon_id`,",
                "flyBaseGene.strand AS `strand`,",
                "`transcript` AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `flyBaseCanonical` INNER JOIN `flyBaseGene` ON",
                "flyBaseCanonical.transcript=flyBaseGene.name",
                "INNER JOIN `flyBaseToRefSeq`",
                "ON flyBaseCanonical.transcript=flyBaseToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON flyBaseToRefSeq.value=refFlat.name",
                "INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
            #    "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #    "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            #return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
            #    "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
            #    "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
            #    "mgcGenes.chrom AS `chromosome`,",
            #    "`exonStarts` AS `start`,",
            #    "`exonEnds` AS `end`,",
            #    "mgcGenes.name AS `exon_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "mgcGenes.name AS `gene_id`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `gene_name`",
            #   "ORDER BY `chromosome`,`start`) AS tmp"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "mgcGenes.chrom AS `chromosome`,",
                "`exonStarts` AS `start`,",
                "`exonEnds` AS `end`,",
                "mgcGenes.name AS `exon_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_id`,",
                "mgcGenes.name AS `gene_name`,",
                "'exon' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            warning("No UCSC Genome annotation for Danio rerio v11! Will use ",
                "RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan troglodytes v4! Will ",
                "use RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v5! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v2! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

.getUcscQueryRefseqExon <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT  refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT refFlat.chrom AS `chromosome`,",
            #   "refFlat.exonStarts AS `start`,",
            #   "refFlat.exonEnds  AS `end`,",
            #   "refFlat.name AS `exon_id`,",
            #   "refFlat.strand AS `strand`,",
            #   "refFlat.name AS `gene_id`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
            #   "ON refFlat.name=knownToRefSeq.value",
            #   "INNER JOIN `knownCanonical`",
            #   "ON knownToRefSeq.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "GROUP BY refFlat.name",
            #   "ORDER BY `chromosome`,`start`"))
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm39 = {
            #return(paste("SELECT refFlat.chrom AS `chromosome`,",
            #   "refFlat.exonStarts AS `start`,",
            #   "refFlat.exonEnds  AS `end`,",
            #   "refFlat.name AS `exon_id`,",
            #   "refFlat.strand AS `strand`,",
            #   "refFlat.name AS `gene_id`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
            #   "ON refFlat.name=knownToRefSeq.value",
            #   "INNER JOIN `knownCanonical`",
            #   "ON knownToRefSeq.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "GROUP BY refFlat.name",
            #   "ORDER BY `chromosome`,`start`"))
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds  AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`", 
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "GROUP BY refFlat.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        rn6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        dm6 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer7 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer10 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        danrer11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro4 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        pantro5 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        susscr11 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        },
        equcab3 = {
            return(paste("SELECT `chromosome`,`start`,`end`,`exon_id`,",
                "`strand`,`gene_id`,`gene_name`,`biotype` FROM",
                "(SELECT MAX(`txEnd` - `txStart`) AS `width`,",
                "refFlat.chrom AS `chromosome`,",
                "refFlat.exonStarts AS `start`,",
                "refFlat.exonEnds AS `end`,",
                "refFlat.name AS `exon_id`,",
                "refFlat.strand AS `strand`,",
                "refFlat.name AS `gene_id`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name", 
                "GROUP BY `gene_name`",
                "ORDER BY `chromosome`,`start`) AS tmp"))
        }
    )
}

.getUcscQueryUcscTranscript <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`", 
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg19 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`", 
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownGene.name=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat` ON", 
                "knownToRefSeq.value=refFlat.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is like hg18
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`",
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm9 = {
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`",
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `knownToEnsembl`",
                "ON knownGene.name=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "INNER JOIN `refFlat` ON",
                "knownToRefSeq.value=refFlat.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT knownGene.chrom AS `chromosome`,",
            #   "knownGene.txStart AS `start`,",
            #   "knownGene.txEnd AS `end`,",
            #   "knownGene.name AS `transcript_id`,",
            #   "knownGene.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownGene` INNER JOIN `knownToRefSeq`", 
            #   "ON knownGene.name=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownGene.name=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat` ON",
            #   "knownToRefSeq.value=refFlat.name",
            #   "GROUP BY `transcript_id`",
            #   "ORDER BY `chromosome`,`start`"))
            ## No ensemblSource...
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`",
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm39 = {
            #return(paste("SELECT knownGene.chrom AS `chromosome`,",
            #   "knownGene.txStart AS `start`,",
            #   "knownGene.txEnd AS `end`,",
            #   "knownGene.name AS `transcript_id`,",
            #   "knownGene.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `knownGene` INNER JOIN `knownToRefSeq`", 
            #   "ON knownGene.name=knownToRefSeq.name",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownGene.name=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "INNER JOIN `refFlat` ON",
            #   "knownToRefSeq.value=refFlat.name",
            #   "GROUP BY `transcript_id`",
            #   "ORDER BY `chromosome`,`start`"))
            ## No ensemblSource...
            return(paste("SELECT knownGene.chrom AS `chromosome`,",
                "knownGene.txStart AS `start`,",
                "knownGene.txEnd AS `end`,",
                "knownGene.name AS `transcript_id`,",
                "knownGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `knownGene` INNER JOIN `knownToRefSeq`",
                "ON knownGene.name=knownToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON knownToRefSeq.value=refFlat.name",
                "GROUP BY knownGene.name",
                "ORDER BY `chromosome`,`start`"))
        },
        rn5 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'transcript' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn6 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`",
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'gene' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm3 = {
            return(paste("SELECT flyBaseGene.chrom AS `chromosome`,",
                "flyBaseGene.txStart AS `start`,",
                "flyBaseGene.txEnd AS `end`,",
                "flyBaseGene.name AS `transcript_id`,",
                "flyBaseGene.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `flyBaseGene` INNER JOIN `flyBaseToRefSeq`",
                "ON flyBaseGene.name=flyBaseToRefSeq.name",
                "INNER JOIN `refFlat`",
                "ON flyBaseToRefSeq.value=refFlat.name",
                "INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            warning("No UCSC Genome annotation for Drosophila ",
                "melanogaster v6! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer7 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'transcript' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer10 = {
            #return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
            #    "`txStart` AS `start`,",
            #    "`txEnd` AS `end`,",
            #    "mgcGenes.name AS `transcript_id`,",
            #    "mgcGenes.strand AS `strand`,",
            #    "`name2` AS `gene_name`,",
            #    "`source` AS `biotype`",
            #    "FROM `mgcGenes` INNER JOIN `ensemblToGeneName`", 
            #    "ON mgcGenes.name2=ensemblToGeneName.value",
            #    "INNER JOIN `ensemblSource`",
            #    "ON ensemblToGeneName.name=ensemblSource.name",
            #    "GROUP BY `transcript_id`",
            #    "ORDER BY `chromosome`,`start`"))
            ## Name 2 is empty - 11/6/2020
            return(paste("SELECT mgcGenes.chrom AS `chromosome`,",
                "`txStart` AS `start`,",
                "`txEnd` AS `end`,",
                "mgcGenes.name AS `transcript_id`,",
                "mgcGenes.strand AS `strand`,",
                "mgcGenes.name AS `gene_name`,",
                "'transcript' AS `biotype`",
                "FROM `mgcGenes`",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer11 = {
            warning("No UCSC Genome annotation for Danio rerio v11! Will use ",
                "RefSeq instead...",immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro4 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v4! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro5 = {
            warning("No UCSC Genome annotation for Pan ",
                "troglodytes v5! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr3 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr11 = {
            warning("No UCSC Genome annotation for Sus ",
                "scrofa v11! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab3 = {
            warning("No UCSC Genome annotation for Equus ",
                "caballus v3! Will use RefSeq instead...",
                immediate.=TRUE)
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`",
                "ON ensemblToGeneName.value=refFlat.geneName",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        }
    )
}

.getUcscQueryRefseqTranscript <- function(org) {
    switch(org,
        hg18 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "ORDER BY `chromosome`, `start`"))
        },
        hg19 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "ORDER BY `chromosome`,`start`"))
        },
        hg38 = {
            # Should be the same as hg19 but is as hg18
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "ORDER BY `chromosome`, `start`"))
        },
        mm9 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "INNER JOIN `knownToEnsembl`",
                "ON knownCanonical.transcript=knownToEnsembl.name",
                "INNER JOIN `ensemblSource`",
                "ON knownToEnsembl.value=ensemblSource.name",
                "ORDER BY `chromosome`,`start`"))
        },
        mm10 = {
            #return(paste("SELECT refFlat.chrom AS `chromosome`,",
            #   "refFlat.txStart AS `start`,",
            #   "refFlat.txEnd AS `end`,",
            #   "refFlat.name AS `transcript_id`,",
            #   "refFlat.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `refFlat` INNER JOIN `knownToRefSeq`",
            #   "ON refFlat.name=knownToRefSeq.value",
            #   "INNER JOIN `knownCanonical`",
            #   "ON knownToRefSeq.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "ORDER BY `chromosome`,`start`"))
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "ORDER BY `chromosome`, `start`"))
        },
        mm39 = {
            #return(paste("SELECT refFlat.chrom AS `chromosome`,",
            #   "refFlat.txStart AS `start`,",
            #   "refFlat.txEnd AS `end`,",
            #   "refFlat.name AS `transcript_id`,",
            #   "refFlat.strand AS `strand`,",
            #   "`geneName` AS `gene_name`,",
            #   "`source` AS `biotype`",
            #   "FROM `refFlat` INNER JOIN `knownToRefSeq`",
            #   "ON refFlat.name=knownToRefSeq.value",
            #   "INNER JOIN `knownCanonical`",
            #   "ON knownToRefSeq.name=knownCanonical.transcript",
            #   "INNER JOIN `knownToEnsembl`",
            #   "ON knownCanonical.transcript=knownToEnsembl.name",
            #   "INNER JOIN `ensemblSource`",
            #   "ON knownToEnsembl.value=ensemblSource.name",
            #   "ORDER BY `chromosome`,`start`"))
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "'NA' AS `biotype`",
                "FROM `refFlat` INNER JOIN `knownToRefSeq`",
                "ON refFlat.name=knownToRefSeq.value",
                "INNER JOIN `knownCanonical`",
                "ON knownToRefSeq.name=knownCanonical.transcript",
                "ORDER BY `chromosome`, `start`"))
        },
        rn5 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        rn6 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm3 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        dm6 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer7 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer10 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        danrer11 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro4 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        pantro5 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr3 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        susscr11 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        },
        equcab3 = {
            return(paste("SELECT refFlat.chrom AS `chromosome`,",
                "refFlat.txStart AS `start`,",
                "refFlat.txEnd AS `end`,",
                "refFlat.name AS `transcript_id`,",
                "refFlat.strand AS `strand`,",
                "`geneName` AS `gene_name`,",
                "`source` AS `biotype`",
                "FROM `refFlat` INNER JOIN `ensemblToGeneName`", 
                "ON refFlat.geneName=ensemblToGeneName.value",
                "INNER JOIN `ensemblSource`",
                "ON ensemblToGeneName.name=ensemblSource.name",
                "GROUP BY `transcript_id`",
                "ORDER BY `chromosome`,`start`"))
        }
    )
}

.getUcscCredentials <- function() {
    return(c(
        host="genome-mysql.cse.ucsc.edu",
        user="genome",
        password=""
    ))
}

..getAllUcsc <- function(to) {
    if (missing(to))
        stop("Please provide a path to put the sqlite databases!")
    if (!dir.exists(to))
        dir.create(to,recursive=TRUE)
        
    for (refdb in getSupportedUcscDbs())
        for (org in getSupportedOrganisms())
            ..getUcscSqlite(org,refdb,to)
}

..getUcscSqlite <- function(org,refdb,to) {
    if (missing(to))
        stop("Please provide a path to put the sqlite databases!")
    if (!dir.exists(to))
        dir.create(to,recursive=TRUE)
    
    message("RETRIEVING ",org," FROM ",refdb)
    dbTmp <- getUcscDbl(org=org,refdb=refdb)
    name <- paste(org,"_",refdb,".sqlite",sep="")
    file.copy(dbTmp,file.path(to,name),recursive=TRUE)
    return()
}

.localTblDef <- function() {
    return(list(
        enable_fkey="PRAGMA foreign_keys=1;",
        content=paste(
            "CREATE TABLE IF NOT EXISTS content (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "source TEXT,",
            "organism TEXT,",
            "version INTEGER,",
            "type TEXT,",
            "user INTEGER",
            ");"
        ),
        seqinfo=paste(
            "CREATE TABLE IF NOT EXISTS seqinfo (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "length INTEGER,",
            #"source TEXT,",
            #"organism TEXT,",
            #"version INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        gene=paste(
            "CREATE TABLE IF NOT EXISTS gene (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "gene_id TEXT,",
            "gc_content REAL,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        transcript=paste(
            "CREATE TABLE IF NOT EXISTS transcript (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        exon=paste(
            "CREATE TABLE IF NOT EXISTS exon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        transexon=paste(
            "CREATE TABLE IF NOT EXISTS transexon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "transcript_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        utr=paste(
            "CREATE TABLE IF NOT EXISTS utr (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_transcript=paste(
            "CREATE TABLE IF NOT EXISTS summarized_transcript (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_exon=paste(
            "CREATE TABLE IF NOT EXISTS summarized_exon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_transcript_exon=paste(
            "CREATE TABLE IF NOT EXISTS summarized_transcript_exon (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "exon_id TEXT,",
            "transcript_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_3utr=paste(
            "CREATE TABLE IF NOT EXISTS summarized_3utr (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        summarized_3utr_transcript=paste(
            "CREATE TABLE IF NOT EXISTS summarized_3utr_transcript (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "chromosome TEXT,",
            "start INTEGER,",
            "end INTEGER,",
            "transcript_id TEXT,",
            "gene_id TEXT,",
            "strand TEXT,",
            "gene_name TEXT,",
            "biotype TEXT,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_length=paste(
            "CREATE TABLE IF NOT EXISTS active_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_utr_length=paste(
            "CREATE TABLE IF NOT EXISTS active_utr_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_trans_utr_length=paste(
            "CREATE TABLE IF NOT EXISTS active_trans_utr_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        ),
        active_trans_exon_length=paste(
            "CREATE TABLE IF NOT EXISTS active_trans_exon_length (",
            "_id INTEGER PRIMARY KEY AUTOINCREMENT,",
            "name TEXT,",
            "length INTEGER,",
            "content_id INTEGER NOT NULL,",
            "FOREIGN KEY(content_id) REFERENCES content(_id) ON DELETE CASCADE",
            ");"
        )
    ))
}

.makeAnnotationQuerySet <- function(t,i,j=NULL) {
    mainQuery <- paste("SELECT * FROM ",t," WHERE content_id=",i,sep="")
    seqInfoQuery <- paste("SELECT * FROM seqinfo WHERE content_id=",i,sep="")
    activeQuery <- NULL
    if (t == "summarized_exon" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_length WHERE content_id=",
            j,sep="")
    else if (t == "summarized_3utr" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_utr_length WHERE ",
            "content_id=",j,sep="")
    else if (t == "summarized_3utr_transcript" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_trans_utr_length WHERE ",
            "content_id=",j,sep="")
    else if (t == "summarized_transcript_exon" && !is.null(j))
        activeQuery <- paste("SELECT * FROM active_trans_exon_length WHERE ",
            "content_id=",j,sep="")
    return(list(
        main=mainQuery,
        seqinfo=seqInfoQuery,
        active=activeQuery
    ))
}

.insertContent <- function(con,o,s,v,t,u=0) {
    query <- paste(
        "INSERT INTO content (source, organism, version, type, user) ",
        "VALUES (",paste("'",s,"', ","'",o,"', ",v,", '",t,"', ",u,sep=""),")",
        sep=""
    )
    nr <- dbExecute(con,query)
    return(nr)
}

.browseContent <- function(con) {
    return(dbGetQuery(con,"SELECT * FROM content"))
}

.browseUserContent <- function(con) {
    return(dbGetQuery(con,"SELECT * FROM content WHERE user=1"))
}

.annotationExists <- function(con,o,s,v=NULL,t=NULL,out=c("tf","nr","id")) {
    out <- out[1]
    query <- paste("SELECT _id FROM content WHERE source='",s,
        "' AND organism='",o,"'",sep="")
    if (!is.null(v))
        query <- paste(query," AND version=",v,sep="")
    if (!is.null(t))
        query <- paste(query," AND type='",t,"'",sep="")
    res <- dbGetQuery(con,query)
    if (out == "tf")
        return(nrow(res) > 0)
    else if (out=="nr")
        return(nrow(res))
    else if (out == "id") {
        if (nrow(res) > 0)
            return(res[1,1])
    }
}

.dropAnnotation <- function(con,o,s,v,t) {
    # At least organism must exist
    if (missing(o))
        stop("At least an organism name must be provided for deletion!")
    # A basic deletion query based on organism. Since the content table is
    # connected with the rest through foreing keys, deletion from there should
    # be enough.
    query <- paste("DELETE FROM content WHERE organism='",o,"'",sep="")
    # Augment according to given arguments.
    if (!missing(s))
        query <- paste(query," AND source='",s,"'",sep="")
    if (!missing(v))
        query <- paste(query," AND version=",v,sep="")
    if (!missing(t))
        query <- paste(query," AND type='",t,"'",sep="")
    # Execute
    nr <- dbExecute(con,query)
    return(nr)
}

.installedVersions <- function(con,o,s) {
    query <- paste(
        "SELECT version FROM content WHERE source='",s,"' AND organism='",o,"'",
        sep=""
    )
    res <- dbGetQuery(con,query)
    if (nrow(res) > 0)
        return(as.numeric(res[,1]))
    else
        return(NA)
}

estimateAufcWeights <- function(counts,normalization,statistics,nsim=10,
    N=10000,samples=c(3,3),ndeg=c(500,500),top=500,modelOrg="mm9",fcBasis=1.5,
    drawFpc=FALSE,rc=NULL,...) {
    if (!requireNamespace("zoo"))
        stopwrap("R pacakage zoo is required in order to estimate AUFC ",
            "weights!")

    if (ncol(counts)<4)
        stopwrap("Cannot estimate AUFC weights with an initial dataset with ",
            "less than 4 samples!")
    else if (ncol(counts)>=4 && ncol(counts)<10) {
        reind <- sample(seq_len(ncol(counts)),20,replace=TRUE)
        counts <- counts[,reind]
    }
    parList <- estimateSimParams(counts,...)

    disp("Running simulations... This procedure requires time... Please ",
        "wait...")
    simResults <- cmclapply(seq_len(nsim),function(x,normalization,statistics,N,
        parList,samples,ndeg,fcBasis,modelOrg) {
        D <- makeSimDataSd(N=N,param=parList,samples=samples,ndeg=ndeg,
            fcBasis=fcBasis,modelOrg=modelOrg)
        dd <- D$simdata
        
        if (!is.null(modelOrg)) {
            tmp <- metaseqr2(
                counts=dd,
                sampleList=list(G1=paste("G1_rep",seq_len(samples[1]),sep=""),
                    G2=paste("G2_rep",seq_len(samples[2]),sep="")),
                contrast=c("G1_vs_G2"),
                annotation="embedded",
                embedCols=list(
                    idCol=4,
                    gcCol=5,
                    nameCol=7,
                    btCol=8
                ),
                org=modelOrg,
                countType="gene",
                normalization=normalization,
                statistics=statistics,
                metaP="simes",
                figFormat="png",
                preset="all_basic",
                exportWhere=tempdir(),
                restrictCores=rc,
                qcPlots=NULL,
                report=FALSE,
                runLog=FALSE,
                outList=TRUE
            )
        }
        else {
            tmp <- metaseqr2(
                counts=dd,
                sampleList=list(G1=paste("G1_rep",seq_len(samples[1]),sep=""),
                    G2=paste("G2_rep",seq_len(samples[2]),sep="")),
                contrast=c("G1_vs_G2"),
                annotation="embedded",
                embedCols=list(
                    idCol=4,
                    gcCol=5,
                    nameCol=7,
                    btCol=8
                ),
                countType="gene",
                normalization=normalization,
                statistics=statistics,
                metaP="simes",
                figFormat="png",
                preset="all_basic",
                exportWhere=tempdir(),
                qcPlots=NULL,
                report=FALSE,
                runLog=FALSE,
                outList=TRUE,
                restrictCores=rc
            )
        }

        # Retrieve several p-values
        pList <- vector("list",length(statistics))
        for (s in statistics) {
            field <- paste("p-value",s,sep="_")
            pList[[s]] <- tmp$data[[1]][,field]
            names(pList[[s]]) <- rownames(tmp$data[[1]])
        }
        pMatrix <- do.call("cbind",pList)
        return(list(simdata=D,pvalues=pMatrix))
    },normalization,statistics,N,parList,samples,ndeg,fcBasis,modelOrg,rc=rc)

    disp("Estimating AUFC weights... Please wait...")
    fpcObj <- cmclapply(simResults,function(x) {
        trueDe <- x$simdata$truedeg
        names(trueDe) <- rownames(x$simdata$simdata)
        pMatrix <- x$pvalues
        trueDe <- trueDe[rownames(pMatrix)]
        fdc <- diagplotFtd(trueDe,pMatrix,type="fpc",draw=FALSE)
    },rc=rc)
    avgFpc <- diagplotAvgFtd(fpcObj,draw=drawFpc)

    x <- seq_len(top)
    aufc <- apply(avgFpc$avgFtdr$means[seq_len(top),],2,function(x,i) {
        return(sum(diff(i)*rollmean(x,2)))
    },x)
    weightAufc <- (sum(aufc)/aufc)/sum(sum(aufc)/aufc)
    return(weightAufc)
}

makeSimDataTcc <- function(...) {
    if (suppressWarnings(!requireNamespace("TCC")))
        stopwrap("Bioconductor package TCC is required to create ",
            "simulated data!")
    #tcc <- simulateReadCounts(Ngene=Ngene,PDEG=PDEG,DEG.assign=DEG.assign,
    #    DEG.foldchange=DEG.foldchange,replicates=replicates)
    tcc <- TCC::simulateReadCounts(...)
    n <- nrow(tcc$count)
    # Now we have to simulate annotation
    chromosome <- paste("chr",1+round(20*runif(n)),sep="")
    start <- 1 + round(1e+6*runif(n))
    end <- start + 250 + round(1e+6*runif(n))
    gene_id <- gene_name <- rownames(tcc$count)
    gc_content <- runif(n)
    strand <- sample(c("+","-"),n,replace=TRUE)
    biotype <- sample(paste("biotype",seq_len(10)),n,replace=TRUE)
    simData <- data.frame(
        chromosome=chromosome,
        start=start,
        end=end,
        gene_id=gene_id,
        gc_content=gc_content,
        strand=strand,
        gene_name=gene_name,
        biotype=biotype
    )
    simData <- cbind(simData,tcc$count)
    return(list(simdata=simData,simparam=tcc$simulation))
}

makeSimDataSd <- function(N,param,samples=c(5,5),ndeg=rep(round(0.1*N),2),
    fcBasis=1.5,libsizeRange=c(0.7,1.4),libsizeMag=1e+7,modelOrg=NULL,
    simLengthBias=FALSE) {
    if (!is.null(modelOrg)) {
        modelOrg <- tolower(modelOrg)
        checkTextArgs("modelOrg",modelOrg,c("hg18","hg19","mm9","mm10","mm39",
            "rno5","dm3","rn5","rn6","danrer7","pantro4","tair10"),
            multiarg=FALSE)
        ann <- getAnnotation(modelOrg,"gene")
        realGc <- as.numeric(ann$gc_content)
        realStart <- as.numeric(ann$start)
        realEnd <- as.numeric(ann$end)
        realStrand <- as.character(ann$strand)
    }
    muHat <- param$muHat
    phiHat <- param$phiHat

    if (simLengthBias) {
        sind <- sort(muHat,index.return=TRUE)$ix
        muHat <- muHat[sind]
        phiHat <- phiHat[sind]
        if (length(muHat)>=N)
            ii <- sort(sample(seq_len(length(muHat)),N))
        else
            ii <- sort(sample(seq_len(length(muHat)),N,replace=TRUE))
    }
    else {
        if (length(muHat)>=N)
            ii <- sample(seq_len(length(muHat)),N)
        else
            ii <- sample(seq_len(length(muHat)),N,replace=TRUE)
    }

    s1 <- samples[1]
    s2 <- samples[2]
    L1 <- round(libsizeMag*runif(s1,min=libsizeRange[1],
        max=libsizeRange[2]))
    L2 <- round(libsizeMag*runif(s2,min=libsizeRange[1],
        max=libsizeRange[2]))

    lambda1 <- do.call("cbind",rep(list(muHat[ii]),s1))
    mu1 <- sweep(lambda1,2,L1/sum(lambda1[,1]),"*")
    sim1 <- matrix(0,N,s1)
    for (j in seq_len(s1))
        sim1[,j] <- rnbinom(N,size=1/phiHat[ii],mu=mu1[,j])

    v <- numeric(N)
    if (sum(ndeg)>0) {
        iUpdown <- sample(seq_len(length(v)),sum(ndeg))
        regDir <- rep(c(1,-1),c(ndeg[1],ndeg[2]))
        v[iUpdown] <- regDir
        lambda2 <- ((fcBasis + rexp(N))^v)*lambda1
        mu2 <- sweep(lambda2,2,L2/sum(lambda2[,1]),"*")
        sim2 <- matrix(0,N,s2)
        for (j in seq_len(s2))
            sim2[,j] <- rnbinom(N,size=1/phiHat[ii],mu=mu2[,j])
    }
    else {
        lambda2 <- lambda1
        mu2 <- sweep(lambda2,2,L2/sum(lambda2[,1]),"*")
        sim2 <- matrix(0,N,s2)
        for (j in seq_len(s2))
            sim2[,j] <- rnbinom(N,size=1/phiHat[ii],mu=mu2[,j])
    }

    # Now we have to simulate annotation
    chromosome <- paste("chr",1+round(20*runif(N)),sep="")
    gene_id <- gene_name <- paste("gene",seq_len(N),sep="_")
    if (!is.null(modelOrg)) {
        if (length(realGc)>=N)
            sampleInd <- sample(seq_len(length(realGc)),N)
        else
            sampleInd <- sample(seq_len(length(realGc)),N,replace=TRUE)
        gc_content <- realGc[sampleInd]
        start <- realStart[sampleInd]
        end <- realEnd[sampleInd]
        strand <- realStrand[sampleInd]
        if (simLengthBias) {
            lenix <- sort(end-start,index.return=TRUE)$ix
            start <- start[lenix]
            end <- end[lenix]
            gc_content <- gc_content[lenix]
            strand <- strand[lenix]
        }
    }
    else {
        gc_content <- runif(N)
        start <- 1 + round(1e+6*runif(N))
        end <- start + 250 + round(1e+6*runif(N))
        strand <- sample(c("+","-"),N,replace=TRUE)
        if (simLengthBias) {
            lenix <- sort(end-start,index.return=TRUE)$ix
            start <- start[lenix]
            end <- end[lenix]
            gc_content <- gc_content[lenix]
            strand <- strand[lenix]
        }
    }
    biotype <- sample(paste("biotype",seq_len(10)),N,replace=TRUE)
    simData <- data.frame(
        chromosome=chromosome,
        start=start,
        end=end,
        gene_id=gene_id,
        gc_content=gc_content,
        strand=strand,
        gene_name=gene_name,
        biotype=biotype
    )
    colnames(sim1) <- paste("G1_rep",seq_len(s1),sep="")
    colnames(sim2) <- paste("G2_rep",seq_len(s2),sep="")
    rownames(sim1) <- rownames(sim2) <- names(v) <- gene_id

    return(list(simdata=cbind(simData,sim1,sim2),truedeg=v))
}

estimateSimParams <- function(realCounts,libsizeGt=3e+6,rowmeansGt=5,
    eps=1e-11,rc=NULL,draw=FALSE) {
    if (is.data.frame(realCounts))
        mat <- as.matrix(realCounts)
    else if (is.matrix(realCounts))
        mat <- realCounts
    else if (file.exists(realCounts)) {
        realData <- read.delim(realCounts,row.names=1)
        mat <- as.matrix(realData)
    }
    else
        stopwrap("The input count data must be either a file, a matrix or a ",
            "data frame!")
    
    lowLib <- which(apply(mat,2,sum)<libsizeGt)
    if (length(lowLib)==ncol(mat))
        stopwrap("Cannot estimate simulation parameters as the library sizes ",
            "are too small! Try lowering the value of the libsizeGt ",
            "parameter...")
    if (length(lowLib)>0)
        mat <- mat[,-lowLib]
    disp("Downsampling counts...")
    dmat <- downsampleCounts(mat)
    lowCo <- which(apply(dmat,1,
        function(x) if (mean(x)<5) TRUE else FALSE))
    if (length(lowCo)>0)
        dmat <- dmat[-lowCo,]
    muHat <- apply(dmat,1,mean)
    disp("Estimating initial dispersion population...")
    phiEst <- apply(dmat,1,function(x) {
        m <- mean(x)
        v <- var(x)
        phi <- (v-m)/m^2
        return(phi)
    })
    phiInd <- which(phiEst>0)
    phiEst <- phiEst[phiInd]
    dmat <- dmat[phiInd,]
    disp("Estimating dispersions using log-likelihood...\n")
    init <- cmclapply(seq_along(seq_len(nrow(dmat))),function(i,d,p) {
        list(y=d[i,],h=p[i])
    },dmat,phiEst,rc=rc)
    phiHat <- unlist(cmclapply(init,function(x,eps) {
        optimize(mlfo,c(x$h-1e-2,x$h+1e-2),y=x$y,tol=eps)$minimum
    },eps,rc=rc))
    if (draw) {
        dev.new()
        plot(log10(muHat[phiInd]),log10(phiHat),col="blue",pch=20,cex=0.5,
            xlab="",ylab="")
        title(xlab="mean",ylab="dispesion",font=2,cex=0.9)
        grid()
    }
    return(list(muHat=muHat[phiInd],phiHat=phiHat))
}

downsampleCounts <- function(counts) {
    libSizes <- apply(counts,2,sum)
    targetSize <- min(libSizes)
    toRemove <- libSizes-targetSize
    ii <- which(toRemove>0)
    dcounts <- counts
    for (i in ii) {
        tmp <- round(toRemove[i]*(counts[,i]/sum(counts[,i])))
        victimSize <- sum(tmp)
        if (victimSize>toRemove[i]) {
            dif <- victimSize - toRemove[i]
            #victims <- sample(seq_len(length(tmp)),dif)
            victims <- sort(tmp,decreasing=TRUE,
                index.return=TRUE)$ix[seq_len(dif)]
            tmp[victims] <- tmp[victims] - 1
        }
        else if (victimSize<toRemove[i]) {
            dif <- toRemove[i] - victimSize
            #victims <- sample(seq_len(length(tmp)),dif)
            victims <- sort(tmp,decreasing=TRUE,
                index.return=TRUE)$ix[seq_len(dif)]
            tmp[victims] <- tmp[victims] + 1
        }
        dcounts[,i] <- dcounts[,i] - tmp
    }
    return(dcounts)
}

mlfo <- function(phi,y) {
    N <- length(y)
    mu <- mean(y)
    -(sum(lgamma(y+1/phi)) - N*lgamma(1/phi) - sum(lgamma(y+1)) + 
        sum(y*log(mu*phi/(1+mu*phi))) - (N/phi)*log(1+mu*phi))
}

makePermutation <- function(counts,sampleList,contrast,repl=FALSE) {
    cnts <- strsplit(contrast,"_vs_")[[1]]
    virtualContrast <- paste(paste("VirtCond",seq_len(length(cnts)),sep=""),
        collapse="_vs_")
    vitualSampleList <- vector("list",length(sampleList))
    names(vitualSampleList) <- paste("VirtCond",seq_along(sampleList),sep="")
    # Avoid the extreme case of returning a vector with all samples the same
    if (repl) {
        resample <- rep(1,ncol(counts))
        while(length(unique(resample))==1)
            resample <- sample(seq_len(ncol(counts)),ncol(counts),replace=repl)
    }
    else
        resample <- sample(seq_len(ncol(counts)),ncol(counts),replace=repl)
    virtualCounts <- counts[,resample]
    samples <- paste("VirtSamp",seq_len(ncol(counts)),sep="")
    colnames(virtualCounts) <- samples
    nsample <- vapply(sampleList,length,numeric(1))
    virtualSamples <- split(samples,rep(seq_len(length(nsample)),nsample))
    names(virtualSamples) <- names(vitualSampleList)
    for (n in names(vitualSampleList))
        vitualSampleList[[n]] <- virtualSamples[[n]]
    return(list(counts=virtualCounts,sampleList=vitualSampleList,
        contrast=virtualContrast))
}

calcOtr <- function(truth,p,sig=0.05) {
    if (is.list(p))
        pmat <- do.call("cbind",p)
    else if (is.data.frame(p))
        pmat <- as.matrix(p)
    else if (is.matrix(p))
        pmat <- p
    if (is.null(colnames(pmat)))
        colnames(pmat) <- paste("p",seq_len(ncol(pmat)),sep="_")

    sigGenes <- trueIsects <- missed <- vector("list",ncol(pmat))
    names(sigGenes) <- names(trueIsects) <- names(missed) <- colnames(pmat)
    for (n in colnames(pmat)) {
        sigGenes[[n]] <- names(which(pmat[,n]<sig))
        trueIsects[[n]] <- intersect(sigGenes[[n]],names(which(truth!=0)))
        missed[[n]] <- setdiff(names(which(truth!=0)),trueIsects[[n]])
    }
    result <- data.frame(
        P=vapply(sigGenes,length,numeric(1)),
        TP=vapply(trueIsects,length,numeric(1)),
        FN=vapply(missed,length,numeric(1))
    )
    result$FP <- result$P - result$TP
    otr <- result$TP/(result$FP+result$FN)
    names(otr) <- rownames(result)
    return(list(result=result,otr=otr))
}

calcF1Score <- function(truth,p,sig=0.05) {
    if (is.list(p))
        pmat <- do.call("cbind",p)
    else if (is.data.frame(p))
        pmat <- as.matrix(p)
    else if (is.matrix(p))
        pmat <- p
    if (is.null(colnames(pmat)))
        colnames(pmat) <- paste("p",seq_len(ncol(pmat)),sep="_")

    sigGenes <- trueIsects <- missed <- vector("list",ncol(pmat))
    names(sigGenes) <- names(trueIsects) <- names(missed) <- colnames(pmat)
    for (n in colnames(pmat)) {
        sigGenes[[n]] <- names(which(pmat[,n]<sig))
        trueIsects[[n]] <- intersect(sigGenes[[n]],names(which(truth!=0)))
        missed[[n]] <- setdiff(names(which(truth!=0)),trueIsects[[n]])
    }
    result <- data.frame(
        P=vapply(sigGenes,length,numeric(1)),
        TP=vapply(trueIsects,length,numeric(1)),
        FN=vapply(missed,length,numeric(1))
    )
    result$FP <- result$P - result$TP
    f1 <- 2*result$TP/(2*result$TP+result$FP+result$FN)
    names(f1) <- rownames(result)
    return(list(result=result,f1=f1))
}

test_estimate_aufc_weights <- function() {
    data("mm9GeneData",package="metaseqR2")
    set.seed(42)
    weights <- estimateAufcWeights(
       counts=as.matrix(mm9GeneCounts[,9:12]),
       normalization="edaseq",
       statistics=c("edger","limma"),
       nsim=1,N=10,ndeg=c(2,2),top=4,modelOrg=NULL,#modelOrg="mm39",
       rc=0.01,libsizeGt=1e+5
    )
    #checkEqualsNumeric(weights,c(0.5384615,0.4615385),tolerance=1e-5)
    #checkEqualsNumeric(weights,c(0.5,0.5),tolerance=1e-5)
    checkEqualsNumeric(weights,c(0.6666667,0.3333333),tolerance=1e-5)
    checkEqualsNumeric(sum(weights),1,tolerance=1e-9)
}

read2count <- function(targets,annotation,fileType=targets$type,
    transLevel="gene",utrOpts=list(frac=1,minLength=300,downstream=50),
    interFeature=FALSE,rc=NULL) {
    if (missing(targets))
        stopwrap("You must provide the targets argument!")
    if (missing(annotation))
        stopwrap("You must provide an annotation GenomicRanges or data frame!")
    if (!requireNamespace("GenomicRanges"))
        stopwrap("The Bioconductor package GenomicRanges is required to ",
            "proceed!")
    if (fileType=="bed" && !requireNamespace("rtracklayer"))
        stopwrap("The Bioconductor package rtracklayer is required to process ",
            "BED files!")
    if (fileType %in% c("sam","bam")) {
        if (!requireNamespace("Rsamtools"))
            stopwrap("The Bioconductor package Rsamtools is required to ",
                "process BAM files!")
        if (!requireNamespace("GenomeInfoDb"))
            stopwrap("The Bioconductor package GenomeInfoDb is required to ",
                "continue with BAM files!")
    }
    if (!is.list(targets) && file.exists(targets))
        targets <- readTargets(targets)
    else if (!is.list(targets) && !file.exists(targets))
        stopwrap("You must provide a targets list or a valid targets file!")
    
    # Convert annotation to GRanges. If there are no sufficient columns to
    # create GRanges, then it will fail anyway. All the reducing/merging
    # procedures, would have already taken place.
    if (!is(annotation,"GenomicRanges"))
        annotationGr <- GRanges(annotation)
    else
        annotationGr <- annotation
    
    # Determine internal count type
    if ((length(grep("MET",annotation$transcript_id[1])) > 0
        || length(grep("MEU",annotation$transcript_id[1])) > 0)
        && transLevel!="transcript")
        countType <- "utr"
    else if (length(grep("MEX",annotation$exon_id[1])) > 0
        || length(grep("MTE",annotation$exon_id[1])) > 0)
        countType <- "exon"
    else
        countType <- "gene"
    
    # If the count type is "exon", we must reduce overlapping exons belonging to
    # multiple transcripts, so as to avoid inflating the final read count when
    # summing all exons. Included here mostly for backwards compatibility, until
    # new version stabilizes.
    backList <- .backPreprocessAnnotation(annotation,annotationGr,transLevel)
    annotationGr <- backList$annotationGr
    mergedAnnotation <- backList$mergedAnnotation
    
    if (countType == "utr" && !is.null(utrOpts)) {
        disp("Resizing transcript 3' UTRs...")
        w <- width(annotationGr)
        # 1. Percentage of 3' UTR only (upstream of end), ensuring min length
        nw <- utrOpts$frac*width(annotationGr)
        nw <- ifelse(nw < utrOpts$minLength,utrOpts$minLength,nw)
        annotationGr <- resize(annotationGr,width=nw,fix="end")
        # 2. Extend 3' UTR downstream
        annotationGr <- resize(annotationGr,
            width=width(annotationGr) + utrOpts$downstream,fix="start")
    }
    
    # Continue
    filesList <- targets$files
    sampleNames <- unlist(lapply(filesList,names),use.names=FALSE)
    sampleFiles <- unlist(filesList,use.names=FALSE)
    names(sampleFiles) <- sampleNames
    if (!is.null(targets$paired)) {
        paired <- unlist(targets$paired,use.names=FALSE)
        names(paired) <- sampleNames
    }
    else
        paired <- NULL
    if (!is.null(targets$stranded)) {
        stranded <- unlist(targets$stranded,use.names=FALSE)
        names(stranded) <- sampleNames
    }
    else
        stranded <- NULL
        
    # Initialize libsize
    libsize <- vector("list",length(sampleNames))
    names(libsize) <- sampleNames
    
    if (fileType=="bed") {
        # Initialize counts
        counts <- matrix(0,nrow=length(annotationGr),ncol=length(sampleNames))
        rownames(counts) <- names(annotationGr)
        colnames(counts) <- sampleNames
        
        retVal <- cmclapply(sampleNames,function(n,sampleFiles) {
            disp("Reading bed file ",basename(sampleFiles[n]),
                " for sample with name ",n,". This might take some time...")
            bed <- import.bed(sampleFiles[n],trackLine=FALSE)
            disp("  Checking for chromosomes not present in the annotation...")
            bed <- bed[which(!is.na(match(as(seqnames(bed),"character"),
                seqlevels(annotationGr))))]
            libsize <- length(bed)
            if (length(bed) > 0) {
                disp("  Counting reads overlapping with given annotation...")
                counts <- countOverlaps(annotationGr,bed)
            }
            else
                warnwrap(paste("No reads left after annotation chromosome ",
                    "presence check for sample ",n,sep=""))
            gc(verbose=FALSE)
            return(list(counts=counts,libsize=libsize))
        },sampleFiles,rc=rc)
        disp("  Finished counting!")
    }
    else if (fileType %in% c("sam","bam")) {
        if (fileType=="sam")
            sampleFiles <- .convertSam(sampleFiles)
        
        # Adjust seqlevelsStyle locally, only for counting, the finally reported
        # style is UCSC as several downstream steps depend on it
        annotationGr <- .adjustSeqlevelsStyle(annotationGr,sampleFiles[1])
        
        # We must now compare the Seqinfos of annotation and a BAM file to check
        # whether they match. If they do not match, then for now, we drop the
        # Seqinfo of annotation. In a future release, it will be converted to
        # the BAMs Seqinfo derived from the header.
        annotationGr <- .adjustOrDropSeqlevels(annotationGr,sampleFiles[1])
        
        # Initialize counts, after fixing annotationGr
        counts <- matrix(0,nrow=length(annotationGr),ncol=length(sampleNames))
        rownames(counts) <- names(annotationGr)
        colnames(counts) <- sampleNames
        
        retVal <- cmclapply(sampleNames,function(n,sampleFiles,paired,
            stranded) {
            disp("Reading bam file ",basename(sampleFiles[n])," for sample ",
                "with name ",n,". This might take some time...")
            if (!is.null(paired)) {
                p <- tolower(paired[n])
                if (p=="single") {
                    singleEnd <- TRUE
                    fragments <- FALSE
                    asMates <- FALSE
                }
                else if (p=="paired") {
                    singleEnd <- FALSE
                    fragments <- FALSE
                    asMates <- TRUE
                }
                else if (p=="mixed") {
                    singleEnd <- FALSE
                    fragments <- TRUE
                    asMates <- TRUE
                }
                else {
                    warnwrap("Information regarding single- or paired-end ",
                        "reads is not correctly provided! Assuming single...")
                    singleEnd <- TRUE
                    fragments <- FALSE
                    asMates <- FALSE
                }
            }
            else {
                singleEnd <- TRUE
                fragments <- FALSE
                asMates <- FALSE
            }
            if (!is.null(stranded)) {
                s <- tolower(stranded[n])
                if (s %in% c("forward","reverse"))
                    ignoreStrand <- FALSE
                else if (s=="no")
                    ignoreStrand <- TRUE
                else {
                    warnwrap("Information regarding strandedness of the reads ",
                        "is not correctly provided! Assuming unstranded...")
                    ignoreStrand <- TRUE
                }
            }
            else
                ignoreStrand <- TRUE
            # Check remoteness
            if (length(grep("^(http|ftp)",sampleFiles[n],perl=TRUE))>=1) {
                reads <- as(readGAlignments(file=sampleFiles[n]),"GRanges")
                libsize <- length(reads)
                isRemote <- TRUE
            }
            else {
                reads <- BamFile(sampleFiles[n],asMates=asMates)
                libsize <- countBam(reads,
                    #param=ScanBamParam(scanBamFlag(isUnmappedQuery=FALSE))
                    param=ScanBamParam(scanBamFlag(isUnmappedQuery=FALSE),
                        mapqFilter=20)
                )$records
                isRemote <- FALSE
            }
            if (libsize>0) {
                disp("  Counting reads overlapping with given annotation...")
                if (singleEnd & !fragments)
                    disp("    ...for single-end reads...")
                else if (!singleEnd & !fragments)
                    disp("    ...for paired-end reads...")
                else if (!singleEnd & fragments)
                    disp("    ...for mixed single- and paired-end reads...")
                if (ignoreStrand)
                    disp("    ...ignoring strandedness...")
                else {
                    disp("    ...assuming ",s," sequenced reads...")
                    if (s=="reverse")
                        strand(annotationGr) <- ifelse(strand(
                            annotationGr)=="+","-","+")
                }
                if (isRemote)
                    disp("    ...for remote BAM file... might take longer...")
                pa=ScanBamParam(scanBamFlag(isUnmappedQuery=FALSE),
                    mapqFilter=20)
                counts <- summarizeOverlaps(annotationGr,reads,
                    singleEnd=singleEnd,fragments=fragments,
                    ignore.strand=ignoreStrand,inter.feature=interFeature,
                    param=pa)
                counts <- assays(counts)$counts
            }
            else
                warnwrap(paste("No reads left after annotation chromosome ",
                    "presence check for sample ",n,sep=""))
            gc(verbose=FALSE)
            return(list(counts=counts,libsize=libsize))
        },sampleFiles,paired,stranded,rc=rc)
        disp("  Finished counting!")
    }
    for (i in seq_along(retVal)) {
        counts[,i] <- retVal[[i]]$counts
        libsize[[i]] <- retVal[[i]]$libsize
    }
    
    return(list(counts=counts,libsize=libsize,mergedann=mergedAnnotation))
}

readTargets <- function(input,path=NULL) {
    if (missing(input) || !file.exists(input))
        stopwrap("The targets file should be a valid existing text file!")
    # Test if the input file is YAML    
    if (!requireNamespace("yaml"))
        stopwrap("The R packge yaml is required to proceed!")
    # If the output of read_yaml is a single character of length 1, then it is
    # probably a tab-delimited file
    test <- read_yaml(input)
    if (is(test,"character") && length(test)==1)
        return(.readTabTargets(input,path=path))
    else
        return(.readYamlTargets(test,path=path))
}

.readYamlTargets <- function(input,path=NULL) {
    # The targets can be a nested field in a greater YAML configuration file
    # (useful for integration in a wider CWL workflow). In this case it MUST be
    # named targets OR metaseqR2_targets. Otherwise, crash is inevitable.
    if (is.list(input) && !is.null(input$targets))
        input <- input$targets
    if (is.list(input) && !is.null(input$metaseqR2_targets))
        input <- input$metaseqR2_targets
    # We avoid fieldnames to provide the previous flexibility
    samples <- as.character(input[[1]])
    conditions <- unique(as.character(input[[3]]))
    rawfiles <- as.character(input[[2]])
    if (!is.null(path)) {
        tmp <- dirname(rawfiles) # Test if there is already a path
        if (any(tmp=="."))
            rawfiles <- file.path(path,basename(rawfiles))
    }
    # Check if they exist!!!
    for (f in rawfiles) {
        if (!file.exists(f))
            stopwrap("Raw reads input file ",f," does not exist! Please check!")
    }
    if (length(samples) != length(unique(samples)))
        stopwrap("Sample names must be unique for each sample!")
    if (length(rawfiles) != length(unique(rawfiles)))
        stopwrap("File names must be unique for each sample!")
    sampleList <- vector("list",length(conditions))
    names(sampleList) <- conditions
    for (n in conditions)
        sampleList[[n]] <- samples[which(as.character(input[[3]])==n)]
    fileList <- vector("list",length(conditions))
    names(fileList) <- conditions
    for (n in conditions) {
        fileList[[n]] <- rawfiles[which(as.character(input[[3]])==n)]
        names(fileList[[n]]) <- samples[which(as.character(input[[3]])==n)]
    }
    if (length(input)>3) { # Has info about single- or paired-end reads / strand
        if (length(input)==4) { # Stranded or paired
            whats <- tolower(as.character(input[[4]]))
            if (!all(whats %in% c("yes","no","forward","reverse",
                "single","paired")))
                stopwrap("Unknown options for paired-end reads and/or ",
                    "strandedness in targets file.")
            what <- whats[1]
            if (what %in% c("single","paired")) {
                hasPairedInfo <- TRUE
                hasStrandedInfo <- FALSE
            }
            else {
                if (what %in% c("yes","no")) {
                    .deprecatedWarning("readTargets")
                    tmp <- as.character(input[[4]])
                    tmp[tmp=="yes"] <- "forward"
                    input[[4]] <- tmp
                    hasPairedInfo <- FALSE
                    hasStrandedInfo <- TRUE
                }
                if (what %in% c("forward","reverse","no")) {
                    hasPairedInfo <- FALSE
                    hasStrandedInfo <- TRUE
                }
            }
        }
        if (length(input)==5) { # Both
            whatsPaired <- tolower(as.character(input[[4]]))
            if (!all(whatsPaired %in% c("single","paired","mixed")))
                stopwrap("Unknown option for type of reads (single, paired, ",
                    "mixed) in targets file.")
            whatsStrand <- tolower(as.character(input[[5]]))
            if (!all(whatsStrand %in% c("yes","no","forward","reverse")))
                stopwrap("Unknown option for read strandedness in targets file")
            if (any(whatsStrand=="yes")) {
                .deprecatedWarning("readTargets")
                tmp <- as.character(input[[5]])
                tmp[tmp=="yes"] <- "forward"
                input[[5]] <- tmp
            }
            hasPairedInfo <- TRUE
            hasStrandedInfo <- TRUE
        }
        if (hasPairedInfo && !hasStrandedInfo) {
            pairedList <- vector("list",length(conditions))
            names(pairedList) <- conditions
            for (n in conditions) {
                pairedList[[n]] <- character(length(sampleList[[n]]))
                names(pairedList[[n]]) <- sampleList[[n]]
                for (nn in names(pairedList[[n]]))
                    pairedList[[n]][nn] <- as.character(input[[4]][which(
                        as.character(input[[1]])==nn)])
            }
        }
        else
            pairedList <- NULL
        if (hasStrandedInfo && !hasPairedInfo) {
            strandedList <- vector("list",length(conditions))
            names(strandedList) <- conditions
            for (n in conditions) {
                strandedList[[n]] <- character(length(sampleList[[n]]))
                names(strandedList[[n]]) <- sampleList[[n]]
                for (nn in names(strandedList[[n]]))
                    strandedList[[n]][nn] <- as.character(input[[4]][which(
                        as.character(input[[1]])==nn)])
            }
        }
        else
            strandedList <- NULL
        if (hasStrandedInfo && hasPairedInfo) {
            strandedList <- vector("list",length(conditions))
            names(strandedList) <- conditions
            for (n in conditions) {
                strandedList[[n]] <- character(length(sampleList[[n]]))
                names(strandedList[[n]]) <- sampleList[[n]]
                for (nn in names(strandedList[[n]]))
                    strandedList[[n]][nn] <- as.character(input[[5]][which(
                        as.character(input[[1]])==nn)])
            }
            pairedList <- vector("list",length(conditions))
            names(pairedList) <- conditions
            for (n in conditions) {
                pairedList[[n]] <- character(length(sampleList[[n]]))
                names(pairedList[[n]]) <- sampleList[[n]]
                for (nn in names(pairedList[[n]]))
                    pairedList[[n]][nn] <- as.character(input[[4]][which(
                        as.character(input[[1]])==nn)])
            }
        }
    }
    else
        pairedList <- strandedList <- NULL
    # Guess file type based on only one of them
    tmp <- fileList[[1]][1]
    if (length(grep("\\.bam$",tmp,ignore.case=TRUE,perl=TRUE))>0)
        type <- "bam"
    else if (length(grep("\\.sam$",tmp,ignore.case=TRUE,perl=TRUE))>0)
        type <- "sam"
    else if (length(grep("\\.bed$",tmp,ignore.case=TRUE,perl=TRUE)>0))
        type <- "bed"
    else
        type <- NULL
    return(list(samples=sampleList,files=fileList,paired=pairedList,
        stranded=strandedList,type=type))
}

.readTabTargets <- function(input,path=NULL) {
    tab <- read.delim(input,strip.white=TRUE)
    samples <- as.character(tab[,1])
    conditions <- unique(as.character(tab[,3]))
    rawfiles <- as.character(tab[,2])
    if (!is.null(path)) {
        tmp <- dirname(rawfiles) # Test if there is already a path
        if (any(tmp=="."))
            rawfiles <- file.path(path,basename(rawfiles))
    }
    # Check if they exist!!!
    for (f in rawfiles) {
        if (!file.exists(f))
            stopwrap("Raw reads input file ",f," does not exist! Please check!")
    }
    if (length(samples) != length(unique(samples)))
        stopwrap("Sample names must be unique for each sample!")
    if (length(rawfiles) != length(unique(rawfiles)))
        stopwrap("File names must be unique for each sample!")
    sampleList <- vector("list",length(conditions))
    names(sampleList) <- conditions
    for (n in conditions)
        sampleList[[n]] <- samples[which(as.character(tab[,3])==n)]
    fileList <- vector("list",length(conditions))
    names(fileList) <- conditions
    for (n in conditions) {
        fileList[[n]] <- rawfiles[which(as.character(tab[,3])==n)]
        names(fileList[[n]]) <- samples[which(as.character(tab[,3])==n)]
    }
    if (ncol(tab)>3) { # Has info about single- or paired-end reads / strand
        if (ncol(tab)==4) { # Stranded or paired
            whats <- tolower(as.character(tab[,4]))
            if (!all(whats %in% c("yes","no","forward","reverse",
                "single","paired")))
                stopwrap("Unknown options for paired-end reads and/or ",
                    "strandedness in targets file.")
            what <- whats[1]
            if (what %in% c("single","paired")) {
                hasPairedInfo <- TRUE
                hasStrandedInfo <- FALSE
            }
            else {
                if (what %in% c("yes","no")) {
                    .deprecatedWarning("readTargets")
                    tmp <- as.character(tab[,4])
                    tmp[tmp=="yes"] <- "forward"
                    tab[,4] <- tmp
                    hasPairedInfo <- FALSE
                    hasStrandedInfo <- TRUE
                }
                if (what %in% c("forward","reverse","no")) {
                    hasPairedInfo <- FALSE
                    hasStrandedInfo <- TRUE
                }
            }
        }
        if (ncol(tab)==5) { # Both
            whatsPaired <- tolower(as.character(tab[,4]))
            if (!all(whatsPaired %in% c("single","paired","mixed")))
                stopwrap("Unknown option for type of reads (single, paired, ",
                    "mixed) in targets file.")
            whatsStrand <- tolower(as.character(tab[,5]))
            if (!all(whatsStrand %in% c("yes","no","forward","reverse")))
                stopwrap("Unknown option for read strandedness in targets file")
            if (any(whatsStrand=="yes")) {
                .deprecatedWarning("readTargets")
                tmp <- as.character(tab[,5])
                tmp[tmp=="yes"] <- "forward"
                tab[,5] <- tmp
            }
            hasPairedInfo <- TRUE
            hasStrandedInfo <- TRUE
        }
        if (hasPairedInfo && !hasStrandedInfo) {
            pairedList <- vector("list",length(conditions))
            names(pairedList) <- conditions
            for (n in conditions) {
                pairedList[[n]] <- character(length(sampleList[[n]]))
                names(pairedList[[n]]) <- sampleList[[n]]
                for (nn in names(pairedList[[n]]))
                    pairedList[[n]][nn] <- as.character(tab[which(as.character(
                        tab[,1])==nn),4])
            }
        }
        else
            pairedList <- NULL
        if (hasStrandedInfo && !hasPairedInfo) {
            strandedList <- vector("list",length(conditions))
            names(strandedList) <- conditions
            for (n in conditions) {
                strandedList[[n]] <- character(length(sampleList[[n]]))
                names(strandedList[[n]]) <- sampleList[[n]]
                for (nn in names(strandedList[[n]]))
                    strandedList[[n]][nn] <- 
                        as.character(tab[which(as.character(tab[,1])==nn),4])
            }
        }
        else
            strandedList <- NULL
        if (hasStrandedInfo && hasPairedInfo) {
            strandedList <- vector("list",length(conditions))
            names(strandedList) <- conditions
            for (n in conditions) {
                strandedList[[n]] <- character(length(sampleList[[n]]))
                names(strandedList[[n]]) <- sampleList[[n]]
                for (nn in names(strandedList[[n]]))
                    strandedList[[n]][nn] <- as.character(tab[which(
                        as.character(tab[,1])==nn),5])
            }
            pairedList <- vector("list",length(conditions))
            names(pairedList) <- conditions
            for (n in conditions) {
                pairedList[[n]] <- character(length(sampleList[[n]]))
                names(pairedList[[n]]) <- sampleList[[n]]
                for (nn in names(pairedList[[n]]))
                    pairedList[[n]][nn] <- as.character(tab[which(as.character(
                        tab[,1])==nn),4])
            }
        }
    }
    else
        pairedList <- strandedList <- NULL
    # Guess file type based on only one of them
    tmp <- fileList[[1]][1]
    if (length(grep("\\.bam$",tmp,ignore.case=TRUE,perl=TRUE))>0)
        type <- "bam"
    else if (length(grep("\\.sam$",tmp,ignore.case=TRUE,perl=TRUE))>0)
        type <- "sam"
    else if (length(grep("\\.bed$",tmp,ignore.case=TRUE,perl=TRUE)>0))
        type <- "bed"
    else
        type <- NULL
    return(list(samples=sampleList,files=fileList,paired=pairedList,
        stranded=strandedList,type=type))
}

.writeTargets <- function(targets,outfile) {
    out <- data.frame(
        samplename=unlist(targets$samples,use.names=FALSE),
        filename=unlist(targets$files,use.names=FALSE),
        condition=rep(names(targets$samples),lengths(targets$samples))
    )
    if (!is.null(targets$paired))
        out$paired <- unlist(targets$paired,use.names=FALSE)
    if (!is.null(targets$stranded))
        out$stranded <- unlist(targets$stranded,use.names=FALSE)
    write.table(out,file=outfile,sep="\t",row.names=FALSE,quote=FALSE)
    return(out)
}

.convertSam <- function(sampleFiles) {
    for (n in names(sampleFiles)) {
        dest <- file.path(dirname(sampleFiles[n]),n)
        disp("Converting sam file ",basename(sampleFiles[n]),
            " to bam file ",basename(dest),"...")
        asBam(file=sampleFiles[n],destination=dest,overwrite=TRUE)
        sampleFiles[n] <- paste(dest,"bam",sep=".")
    }
    return(sampleFiles)
}

.adjustOrDropSeqlevels <- function(gr,f) {
    ciAnn <- .chromInfoFromSeqinfo(seqinfo(gr))
    ciBam <- .chromInfoFromBAM(f)
    rows <- intersect(rownames(ciAnn),rownames(ciBam))
    norows <- setdiff(rownames(ciAnn),rownames(ciBam))
    
    if (length(norows) > 0)
        gr <- dropSeqlevels(gr,value=norows,pruning.mode="coarse")
    
    if (length(rows) > 0) {
        ciAnn <- ciAnn[rows,,drop=FALSE]
        ciBam <- ciBam[rows,,drop=FALSE]
        if (!identical(ciAnn,ciBam)) # Different lengths, drop from annotation
            seqlengths(gr) <- rep(NA,length(seqlengths(gr)))
    }
    
    return(gr)
}

.adjustSeqlevelsStyle <- function(gr,f) {
    b <- BamFile(f)
    
    # Check seqlevels styles of the loaded genome for local counting use and
    # change the style of the local annotation.
    # ! Does not work for non-supported organisms !
    tryCatch({
        sb <- seqlevelsStyle(b)
        sg <- seqlevelsStyle(gr)
        if (!any(sg %in% sb))
            seqlevelsStyle(gr) <- sb[1]
    },error=function(e) {
        #warning("Caught error: ",e)
        # Silence, we know the source
    },finally="")
    
    return(gr)
}

.backPreprocessAnnotation <- function(annotation,annotationGr,transLevel,
    rc=NULL) {
    if (length(grep("exon",names(mcols(annotationGr)))) > 0) {
        # countType is exon
        if (length(grep("MEX",annotation$exon_id[1]))) # Retrieved from previous
            mergedAnnotation <- annotation
        else {
            if (transLevel=="gene") {
                disp("Merging exons to create unique gene models...")
                tmp <- .reduceExonsOld(annotationGr)
                annotationGr <- tmp$model
            }
            #else if (transLevel=="transcript") {
            #   disp("Merging exons to create unique gene models...")
            #   annotationGr <- reduceExonsTranscript(annotationGr,rc=rc)
            #}
            #mergedAnnotation <- as.data.frame(annotationGr) # Bug?
            mergedAnnotation <- data.frame(
                chromosome=as.character(seqnames(annotationGr)),
                start=start(annotationGr),
                end=end(annotationGr),
                exon_id=if (!is.null(annotationGr$exon_id))
                    as.character(annotationGr$exon_id) else
                    as.character(annotationGr$name),
                gene_id=if (!is.null(annotationGr$gene_id))
                    as.character(annotationGr$gene_id) else
                    as.character(annotationGr$name),
                strand=as.character(strand(annotationGr)),
                gene_name=if (!is.null(annotationGr$gene_name))
                    as.character(annotationGr$gene_name) else 
                    if (!is.null(annotationGr$symbol))
                    as.character(annotationGr$name) else NULL,
                biotype=if (!is.null(annotationGr$biotype))
                    as.character(annotationGr$biotype) else NULL
            )
            rownames(mergedAnnotation) <- 
                as.character(mergedAnnotation$exon_id)
            names(annotationGr) <- annotationGr$exon_id
        }
    }
    else if (length(grep("transcript",names(mcols(annotationGr))))>0) {
        # countType may be utr
        if (length(grep("MET",annotation$transcript_id[1]))
            || length(grep("MEU",annotation$transcript_id[1])))
            # Retrieved from previous
            mergedAnnotation <- annotation
        else {
            if (transLevel=="gene") {
                disp("Merging transcript 3' UTRs to create unique ",
                    "gene models...")
                annotationGr <- .reduceTranscriptsOld(annotationGr,rc=rc)
            }
            if (transLevel=="transcript") {
                disp("Merging transcript 3' UTRs to create unique ",
                    "transcript models...")
                annotationGr <- 
                    .reduceTranscriptsUtrOld(annotationGr,rc=rc)
            }
            
            #mergedAnnotation <- as.data.frame(annotationGr) # Bug?
            mergedAnnotation <- data.frame(
                chromosome=as.character(seqnames(annotationGr)),
                start=start(annotationGr),
                end=end(annotationGr),
                transcript_id=if (!is.null(annotationGr$transcript_id))
                    as.character(annotationGr$transcript_id) else
                    as.character(annotationGr$name),
                gene_id=if (!is.null(annotationGr$gene_id))
                    as.character(annotationGr$gene_id) else
                    as.character(annotationGr$name),
                strand=as.character(strand(annotationGr)),
                gene_name=if (!is.null(annotationGr$gene_name))
                    as.character(annotationGr$gene_name) else 
                    if (!is.null(annotationGr$symbol))
                    as.character(annotationGr$name) else NULL,
                biotype=if (!is.null(annotationGr$biotype))
                    as.character(annotationGr$biotype) else NULL
            )
            rownames(mergedAnnotation) <- 
                as.character(mergedAnnotation$transcript_id)
            names(annotationGr) <- annotationGr$transcript_id
        }
        interFeature = FALSE # Quant-Seq
    }
    else
        mergedAnnotation <- NULL
    return(list(annotationGr=annotationGr,mergedAnnotation=mergedAnnotation))
}

.reduceExonsOld <- function(gr,rc=NULL) {
    gene <- unique(as.character(gr$gene_id))
    if (!is.null(gr$gene_name))
        gn <- gr$gene_name
    else
        gn <- NULL
    if (!is.null(gr$biotype))
        bt <- gr$biotype   
    else
        bt <- NULL
    redList <- cmclapply(gene,function(x,a,g,b) {
        tmp <- a[a$gene_id==x]
        if (!is.null(g))
            gena <- as.character(tmp$gene_name[1])
        if (!is.null(b))
            btty <- as.character(tmp$biotype[1])
        merged <- reduce(tmp)
        n <- length(merged)
        meta <- DataFrame(
            exon_id=paste(x,"MEX",seq_len(n),sep="_"),
            gene_id=rep(x,n)
        )
        if (!is.null(g))
            meta$gene_name <- rep(gena,n)
        if (!is.null(b))
            meta$biotype <- rep(btty,n)
        mcols(merged) <- meta
        return(merged)
    },gr,gn,bt,rc=rc)
    len <- unlist(cmclapply(redList,function(x) {
        return(sum(width(x)))
    },rc=rc))
    names(len) <- names(redList)
    return(list(model=do.call("c",redList),length=len))
}

.reduceTranscriptsOld <- function(gr,rc=NULL) {
    gene <- unique(as.character(gr$gene_id))
    if (!is.null(gr$gene_name))
        gn <- gr$gene_name
    else
        gn <- NULL
    if (!is.null(gr$biotype))
        bt <- gr$biotype   
    else
        bt <- NULL
    redList <- cmclapply(gene,function(x,a,g,b) {
        tmp <- a[a$gene_id==x]
        if (!is.null(g))
            gena <- as.character(tmp$gene_name[1])
        if (!is.null(b))
            btty <- as.character(tmp$biotype[1])
        merged <- reduce(tmp)
        n <- length(merged)
        meta <- DataFrame(
            transcript_id=paste(x,"MET",seq_len(n),sep="_"),
            gene_id=rep(x,n)
        )
        if (!is.null(g))
            meta$gene_name <- rep(gena,n)
        if (!is.null(b))
            meta$biotype <- rep(btty,n)
        mcols(merged) <- meta
        return(merged)
    },gr,gn,bt,rc=rc)
    return(do.call("c",redList))
}

.reduceTranscriptsUtrOld <- function(gr,rc=NULL) {
    trans <- unique(as.character(gr$transcript_id))
    if (!is.null(gr$gene_name))
        gn <- gr$gene_name
    else
        gn <- NULL
    if (!is.null(gr$biotype))
        bt <- gr$biotype   
    else
        bt <- NULL
    redList <- cmclapply(trans,function(x,a,g,b) {
        tmp <- a[a$transcript_id==x]
        if (!is.null(g))
            gena <- as.character(tmp$gene_name[1])
        if (!is.null(b))
            btty <- as.character(tmp$biotype[1])
        merged <- reduce(tmp)
        n <- length(merged)
        meta <- DataFrame(
            transcript_id=paste(x,"MEU",seq_len(n),sep="_"),
            gene_id=rep(x,n)
        )
        if (!is.null(g))
            meta$gene_name <- rep(gena,n)
        if (!is.null(b))
            meta$biotype <- rep(btty,n)
        mcols(merged) <- meta
        return(merged)
    },gr,gn,bt,rc=rc)
    return(do.call("c",redList))
}

################################################################################

#read2countOld <- function(targets,annotation,fileType=targets$type,
#    transLevel="gene",utrFlank=500,interFeature=FALSE,rc=NULL) {
#    if (missing(targets))
#        stopwrap("You must provide the targets argument!")
#    if (missing(annotation))
#        stopwrap("You must provide an annotation data frame!")
#    if (!require(GenomicRanges))
#        stopwrap("The Bioconductor package GenomicRanges is required to ",
#            "proceed!")
#    if (fileType=="bed" && !require(rtracklayer))
#        stopwrap("The Bioconductor package rtracklayer is required to ",
#            "process BED files!")
#    if (fileType %in% c("sam","bam")) {
#        if (!require(Rsamtools))
#            stopwrap("The Bioconductor package Rsamtools is required to ",
#                "process BAM files!")
#    }
#    if (!is.list(targets) && file.exists(targets))
#        targets <- readTargets(targets)
#    else if (!is.list(targets) && !file.exists(targets))
#        stopwrap("You must provide a targets list or a valid targets file!")
    
#    # Convert annotation to GRanges
#    annotationGr <- makeGRangesFromDataFrame(
#        df=annotation,
#        keep.extra.columns=TRUE,
#        seqnames.field="chromosome"
#    )
#    # annotationGr <- GRanges(annotation) also works for our structure
    
#    # If the count type is "exon", we must reduce overlapping exons belonging 
#    # to multiple transcripts, so as to avoid inflating the final read count 
#    # when summing all exons
#    if (length(grep("exon",colnames(annotation)))>0) { # countType is exon
#        # Retrieved from previous
#        if (length(grep("MEX",annotation$exon_id[1]))) 
#            mergedAnnotation <- annotation
#        else {
#            if (transLevel=="gene") {
#                disp("Merging exons to create unique gene models...")
#                annotationGr <- reduceExons(annotationGr,rc=rc)
#            }
#            #else if (transLevel=="transcript") {
#            #   disp("Merging exons to create unique gene models...")
#            #   annotationGr <- reduceExonsTranscript(annotationGr,rc=rc)
#            #}
#            #mergedAnnotation <- as.data.frame(annotationGr) # Bug?
#            mergedAnnotation <- data.frame(
#                chromosome=as.character(seqnames(annotationGr)),
#                start=start(annotationGr),
#                end=end(annotationGr),
#                exon_id=if (!is.null(annotationGr$exon_id))
#                    as.character(annotationGr$exon_id) else
#                    as.character(annotationGr$name),
#                gene_id=if (!is.null(annotationGr$gene_id))
#                    as.character(annotationGr$gene_id) else
#                    as.character(annotationGr$name),
#                strand=as.character(strand(annotationGr)),
#                gene_name=if (!is.null(annotationGr$gene_name))
#                    as.character(annotationGr$gene_name) else 
#                    if (!is.null(annotationGr$symbol))
#                    as.character(annotationGr$name) else NULL,
#                biotype=if (!is.null(annotationGr$biotype))
#                    as.character(annotationGr$biotype) else NULL
#            )
#            rownames(mergedAnnotation) <- 
#                as.character(mergedAnnotation$exon_id)
#        }
#    }
#    else if (length(grep("transcript",colnames(annotation)))>0) {
#        # countType may be utr
#        if (length(grep("MET",annotation$transcript_id[1]))
#            || length(grep("MEU",annotation$transcript_id[1]))) 
#            # Retrieved from previous
#            mergedAnnotation <- annotation
#        else {
#            if (transLevel=="gene") {
#                disp("Merging transcript 3' UTRs to create unique ",
#                    "gene models...")
#                annotationGr <- reduceTranscriptsUtr(annotationGr,rc=rc)
#            }
#            if (transLevel=="transcript") {
#                disp("Merging transcript 3' UTRs to create unique ",
#                    "transcript models...")
#                annotationGr <- 
#                    reduceTranscriptsUtrTranscript(annotationGr,rc=rc)
#            }
#            if (utrFlank > 0) {
#                disp("Flanking merged transcript 3' UTRs per ",utrFlank,
#                    "bp...")
#                w <- width(annotationGr)
#                annotationGr <- promoters(annotationGr,upstream=utrFlank,
#                    downstream=0)
#                annotationGr <- resize(annotationGr,width=w+2*utrFlank)
#            }
#            #mergedAnnotation <- as.data.frame(annotationGr) # Bug?
#            mergedAnnotation <- data.frame(
#                chromosome=as.character(seqnames(annotationGr)),
#                start=start(annotationGr),
#                end=end(annotationGr),
#                transcript_id=if (!is.null(annotationGr$transcript_id))
#                    as.character(annotationGr$transcript_id) else
#                    as.character(annotationGr$name),
#                gene_id=if (!is.null(annotationGr$gene_id))
#                    as.character(annotationGr$gene_id) else
#                    as.character(annotationGr$name),
#                strand=as.character(strand(annotationGr)),
#                gene_name=if (!is.null(annotationGr$gene_name))
#                    as.character(annotationGr$gene_name) else 
#                    if (!is.null(annotationGr$symbol))
#                    as.character(annotationGr$name) else NULL,
#                biotype=if (!is.null(annotationGr$biotype))
#                    as.character(annotationGr$biotype) else NULL
#            )
#            rownames(mergedAnnotation) <- 
#                as.character(mergedAnnotation$transcript_id)
#        }
#        interFeature = FALSE # Quant-Seq
#    }
#    else
#        mergedAnnotation <- NULL
#    # Continue
#    filesList <- targets$files
#    sampleNames <- unlist(lapply(filesList,names),use.names=FALSE)
#    sampleFiles <- unlist(filesList,use.names=FALSE)
#    names(sampleFiles) <- sampleNames
#    if (!is.null(targets$paired)) {
#        paired <- unlist(targets$paired,use.names=FALSE)
#        names(paired) <- sampleNames
#    }
#    else
#        paired <- NULL
#    if (!is.null(targets$stranded)) {
#        stranded <- unlist(targets$stranded,use.names=FALSE)
#        names(stranded) <- sampleNames
#    }
#    else
#        stranded <- NULL
#    counts <- matrix(0,nrow=length(annotationGr),ncol=length(sampleNames))
#    if (length(grep("exon",colnames(annotation)))>0)
#        rownames(counts) <- as.character(annotationGr$exon_id)
#    else if (length(grep("transcript",colnames(annotation)))>0)
#        rownames(counts) <- as.character(annotationGr$transcript_id)
#    else
#        rownames(counts) <- as.character(annotationGr$gene_id)
#    colnames(counts) <- sampleNames
#    libsize <- vector("list",length(sampleNames))
#    names(libsize) <- sampleNames
#    if (fileType=="bed") {
#        retVal <- cmclapply(sampleNames,function(n,sampleFiles) {
#            disp("Reading bed file ",basename(sampleFiles[n]),
#                " for sample with name ",n,". This might take some time...")
#            bed <- import.bed(sampleFiles[n],trackLine=FALSE)
#            disp("  Checking for chromosomes not present in the annotation...")
#            bed <- bed[which(!is.na(match(as(seqnames(bed),"character"),
#                seqlevels(annotationGr))))]
#            libsize <- length(bed)
#            if (length(bed)>0) {
#                disp("  Counting reads overlapping with given annotation...")
#                counts <- countOverlaps(annotationGr,bed)
#            }
#            else
#                warnwrap(paste("No reads left after annotation chromosome ",
#                    "presence check for sample ",n,sep=""))
#            gc(verbose=FALSE)
#            return(list(counts=counts,libsize=libsize))
#        },sampleFiles,rc=rc)
#    }
#    else if (fileType %in% c("sam","bam")) {
#        if (fileType=="sam") {
#            for (n in sampleNames) {
#                dest <- file.path(dirname(sampleFiles[n]),n)
#                disp("Converting sam file ",basename(sampleFiles[n]),
#                    " to bam file ",basename(dest),"...")
#                asBam(file=sampleFiles[n],destination=dest,overwrite=TRUE)
#                sampleFiles[n] <- paste(dest,"bam",sep=".")
#            }
#        }
#        retVal <- cmclapply(sampleNames,function(n,sampleFiles,paired,
#            stranded) {
#            disp("Reading bam file ",basename(sampleFiles[n])," for sample ",
#                "with name ",n,". This might take some time...")
#            if (!is.null(paired)) {
#                p <- tolower(paired[n])
#                if (p=="single") {
#                    singleEnd <- TRUE
#                    fragments <- FALSE
#                    asMates <- FALSE
#                }
#                else if (p=="paired") {
#                    singleEnd <- FALSE
#                    fragments <- FALSE
#                    asMates <- TRUE
#                }
#                else if (p=="mixed") {
#                    singleEnd <- FALSE
#                    fragments <- TRUE
#                    asMates <- TRUE
#                }
#                else {
#                    warnwrap("Information regarding single- or paired-end ",
#                        "reads is not correctly provided! Assuming single...")
#                    singleEnd <- TRUE
#                    fragments <- FALSE
#                    asMates <- FALSE
#                }
#            }
#            else {
#                singleEnd <- TRUE
#                fragments <- FALSE
#                asMates <- FALSE
#            }
#            if (!is.null(stranded)) {
#                s <- tolower(stranded[n])
#                if (s %in% c("forward","reverse"))
#                    ignoreStrand <- FALSE
#                else if (s=="no")
#                    ignoreStrand <- TRUE
#                else {
#                   warnwrap("Information regarding strandedness of the reads ",
#                        "is not correctly provided! Assuming unstranded...")
#                    ignoreStrand <- TRUE
#                }
#            }
#            else
#                ignoreStrand <- TRUE
#            # Check remoteness
#            if (length(grep("^(http|ftp)",sampleFiles[n],perl=TRUE))>=1) {
#                reads <- as(readGAlignments(file=sampleFiles[n]),"GRanges")
#                libsize <- length(reads)
#                isRemote <- TRUE
#            }
#            else {
#                reads <- BamFile(sampleFiles[n],asMates=asMates)
#                libsize <- countBam(reads,
#                param=ScanBamParam(scanBamFlag(isUnmappedQuery=FALSE)))$records
#                isRemote <- FALSE
#            }
#            if (libsize>0) {
#                disp("  Counting reads overlapping with given annotation...")
#                if (singleEnd & !fragments)
#                    disp("    ...for single-end reads...")
#                else if (!singleEnd & !fragments)
#                    disp("    ...for paired-end reads...")
#                else if (!singleEnd & fragments)
#                    disp("    ...for mixed single- and paired-end reads...")
#                if (ignoreStrand)
#                    disp("    ...ignoring strandedness...")
#                else {
#                    disp("    ...assuming ",s," sequenced reads...")
#                    if (s=="reverse")
#                        strand(annotationGr) <- ifelse(strand(
#                            annotationGr)=="+","-","+")
#                }
#                if (isRemote)
#                    disp("    ...for remote BAM file... might take longer...")
#                counts <- summarizeOverlaps(annotationGr,reads,
#                    singleEnd=singleEnd,fragments=fragments,
#                    ignore.strand=ignoreStrand,inter.feature=interFeature)
#                counts <- assays(counts)$counts
#            }
#            else
#                warnwrap(paste("No reads left after annotation chromosome ",
#                    "presence check for sample ",n,sep=""))
#            gc(verbose=FALSE)
#            return(list(counts=counts,libsize=libsize))
#        },sampleFiles,paired,stranded,rc=rc)
#    }
#    for (i in seq_along(retVal)) {
#        counts[,i] <- retVal[[i]]$counts
#        libsize[[i]] <- retVal[[i]]$libsize
#    }
    
#    return(list(counts=counts,libsize=libsize,mergedann=mergedAnnotation))
#}

checkMainArgs <- function(mainArgs) {
    inArgs <- names(mainArgs)[-1] # 1st member name of calling function
    validArgs <- .getValidArgs()
    invalid <- setdiff(inArgs,validArgs)
    if (length(invalid) > 0) {
        for (i in seq_len(length(invalid)))
            warnwrap("Unknown input argument to metaseqr pipeline: ",invalid[i],
                " ...Ignoring...",now=TRUE)
    }
}

checkTextArgs <- function(argName,argValue,argList,multiarg=FALSE) {
    if (multiarg) {
        argValue <- tolower(argValue)
        if (!all(argValue %in% argList))
            stopwrap("\"",argName,"\""," parameter must be one or more of ",
                paste(paste("\"",argList,sep=""),collapse="\", "),"\"!")
    }
    else {
        argSave <- argValue[1]
        argValue <- tolower(argValue[1])
        # An exception must be added for annotation because it can be an  
        # external file too
        #if (argName=="annotation") { 
        #    if (!(argValue %in% argList) && !file.exists(argSave))
        #        stopwrap("\"",argName,"\""," parameter must be one of ",
        #            paste(paste("\"",argList,sep=""),collapse="\", "),
        #                "\" or an existing file!")
        #}
        #else {
            if (!(argValue %in% argList))
                stopwrap("\"",argName,"\""," parameter must be one of ",
                    paste(paste("\"",argList,sep=""),collapse="\", "),"\"!")
        #}
    }
}

checkNumArgs <- function(argName,argValue,argType,argBounds,direction) {
    switch(argType,
        numeric = {
            if (!is.numeric(argValue))
                stopwrap("\"",argName,"\"",
                    " parameter must be a numeric value!")
            if (!missing(argBounds)) {
                switch(direction,
                    both = {
                        if (argValue<argBounds[1] ||
                            argValue>argBounds[2])
                            stopwrap("\"",argName,"\""," parameter must be a ",
                                "numeric ","value larger than or equal to ",
                                argBounds[1]," and smaller than or equal to ",
                                argBounds[2],"!")
                    },
                    botheq = {
                        if (argValue<=argBounds[1] || argValue>=argBounds[2])
                            stopwrap("\"",argName,"\""," parameter must be a ",
                                "numeric value larger than ",argBounds[1],
                                " and smaller than ",argBounds[2],"!")
                    },
                    gt = {
                        if (argValue<=argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be a ",
                                "numeric value greater than ",argBounds[1],"!")
                    },
                    lt = {
                        if (argValue>=argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be a ",
                                "numeric value lower than ",argBounds[1],"!")
                    },
                    gte = {
                        if (argValue<argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be a ",
                                "numeric value greater than or equal to ",
                                argBounds[1],"!")
                    },
                    lte = {
                        if (argValue>argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be a ",
                                "numeric value lower than or equal to ",
                                argBounds[1],"!")
                    }
                )
            }
        },
        integer = {
            if (!is.integer(argValue))
                stopwrap("\"",argName,"\""," parameter must be an integer!")
            if (!missing(argBounds)) {
                switch(direction,
                    both = {
                        if (argValue<argBounds[1] ||
                            argValue>argBounds[2])
                            stopwrap("\"",argName,"\""," parameter must be ",
                                "an integer larger than or equal to ",
                                argBounds[1]," and smaller than or equal to ",
                                argBounds[2],"!")
                    },
                    botheq = {
                        if (argValue<=argBounds[1] || argValue>=argBounds[2])
                            stopwrap("\"",argName,"\""," parameter must be ",
                                "an integer larger than ",argBounds[1],
                                " and smaller than ",argBounds[2],"!")
                    },
                    gt = {
                        if (argValue<=argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be ",
                                "an integer greater than ",argBounds[1],"!")
                    },
                    lt = {
                        if (argValue>=argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be ",
                                "an integer lower than ",argBounds[1],"!")
                    },
                    gte = {
                        if (argValue<argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be ",
                                "an integer greater than or equal to ",
                                argBounds[1],"!")
                    },
                    lte = {
                        if (argValue>argBounds[1])
                            stopwrap("\"",argName,"\""," parameter must be ",
                                "an integer lower than or equal to ",
                                argBounds[1],"!")
                    }
                )
            }
        }
    )
}

checkFileArgs <- function(argName,argValue) {
    if (!file.exists(argValue))
        stopwrap("\"",argName,"\""," parameter must be an existing file!")
}

checkContrastFormat <- function(cnt,sampleList) {
    # This function will break cnt and check that all contrast counter parts are
    # members of the names of the sampleList and contain the string "_vs_" as 
    # many times as the names of the sampleList minus 1. If satisfied return 
    # TRUE else error.
    cnts <- strsplit(cnt,"_vs_")
    #if (length(unique(unlist(cnts))) != length(names(sampleList)))
    if (!all(unique(unlist(cnts)) %in% names(sampleList)))
        stopwrap("Condition names in sample list and contrast list do not ",
            "match! Check if the contrasts follow the appropriate format (e.g.",
            " \"_vs_\" separating contrasting conditions...")
    if (length(unique(cnt))!=length(cnt))
        warnwrap("Duplicates found in the contrasts list! Duplicates will be ",
            "ignored...")
}

checkLibsize <- function(libsizeList,sampleList) {
    if (!is.null(libsizeList)) {
        if (length(intersect(names(libsizeList),unlist(sampleList,
            use.names=FALSE)))!=length(unlist(sampleList,
            use.names=FALSE))) {
            warnwrap("Sample names in \"libsizeList\" and \"sampleList\" do ",
                "not match! Library sizes will be estimated from count data...")
            return(NULL)
        }
        else return(libsizeList)
    }
    else
        return(NULL)
}

checkPackages <- function(m,p) {
    # Check meta-analysis packages
    if ("whitlock" %in% m && !requireNamespace("survcomp"))
        stopwrap("Bioconductor package survcomp is required for \"whitlock\" ",
            "p-value meta analysis!")
    if ("venn" %in% p && !requireNamespace("VennDiagram"))
        stopwrap("R package VennDiagram is required for some of the selected ",
            "QC plots!")
}

.backwardsConvertArgs <- function(args) {
    args <- args[-1] # 1st member name of calling function
    inArgs <- names(args)
    oldArgs <- .getBackwardsValidArgs()
    newArgs <- .getValidArgs()
    old <- intersect(inArgs,oldArgs)
    
    # Detect mixed arguments by excluding the common and then checking the rest
    # A mix can be detected if we exclude the common ones (e.g. 'normalization')
    # from both newArgs and oldArgs and intersect the rest
    common <- intersect(newArgs,oldArgs)
    restNew <- setdiff(newArgs,common)
    restOld <- setdiff(oldArgs,common)
    if (any(inArgs %in% restNew) && any(inArgs %in% restOld))
        return(list(backDetected=FALSE,mixDetected=TRUE,args=NULL))
    
    # Continue, as the mix has been covered. The call with old arguments 
    # contains also the common
    oldInCall <- setdiff(intersect(inArgs,oldArgs),common)
    if (length(oldInCall) > 0) {
        backArgs <- list()
        backArgs$backDetected <- TRUE
        backArgs$mixDetected <- FALSE
        backArgs$args <- args[oldInCall]
        
        # If old call contains id.col, gc.col, name.col or bt.col, these must be
        # grouped. We define anyway and fill them as needed.
        backArgs$args$embedCols <- list(
            idCol=NA,
            gcCol=NA,
            nameCol=NA,
            btCol=NA
        )
        if ("id.col" %in% oldInCall) {
            backArgs$args$embedCols$idCol <- args$id.col
            backArgs$args$id.col <- NULL
        }
        if ("gc.col" %in% oldInCall) {
            backArgs$args$embedCols$gcCol <- args$gc.col
            backArgs$args$gc.col <- NULL
        }
        if ("name.col" %in% oldInCall) {
            backArgs$args$embedCols$nameCol <- args$name.col
            backArgs$args$name.col <- NULL
        }
        if ("bt.col" %in% oldInCall) {
            backArgs$args$embedCols$btCol <- args$bt.col
            backArgs$args$bt.col <- NULL
        }
        
        # Similar check for utr.flank
        backArgs$args$utrOpts <- list(
            frac=1,
            minLength=300,
            downstream=50
        )
        if ("utr.flank" %in% oldInCall) {
            backArgs$args$utrOpts$downstream <- args$utr.flank
            backArgs$args$utr.flank <- NULL
        }
        
        return(backArgs)
    }
    else
        return(list(backDetected=FALSE,mixDetected=FALSE,args=NULL))
}

.backwardsMapOld2New <- function() {
    oldArgNames <- .getBackwardsValidArgs()
    newArgNames <- .getValidArgs()
    
    # The main differences are: i) there is no gene.file in the new args,
    # ii) the idCol, gcCol, nameCol and btCol arguments are within embedCols
    map <- list()
    
    # 1. Remove "id.col","gc.col","name.col","bt.col" from old arguments so as
    # to group them. Remove also gene.file and utr.flank.
    r1 <- match(c("id.col","gc.col","name.col","bt.col","gene.file",
        "utr.flank"),oldArgNames)
    oldArgNames <- oldArgNames[-r1]
    # 2. Remove embedCols from new arguments as they can't be mapped directly
    r2 <- match("embedCols",newArgNames)
    newArgNames <- newArgNames[-r2]
    
    # We can now map
    map[oldArgNames] <- newArgNames
    return(map)
}

.getValidArgs <- function() {
    return(c(
        "counts","sampleList","excludeList","fileType","path","contrast",
        "libsizeList","embedCols","annotation","org","transLevel","countType",
        "utrOpts","exonFilters","geneFilters","whenApplyFilter",
        "normalization","normArgs","statistics","statArgs","adjustMethod",
        "metaP","weight","nperm","reprod","pcut","logOffset","pOffset",
        "preset","qcPlots","figFormat","outList","exportWhere","exportWhat",
        "exportScale","exportValues","exportStats","exportCountsTable",
        "restrictCores","report","refdb","reportTop","reportTemplate","verbose",
        "runLog","reportDb","saveGeneModel","version","localDb","offlineReport",
        "createTracks","overwriteTracks","trackExportPath","trackInfo",
        ".progressFun",".exportR2C"
    ))
}

.getBackwardsValidArgs <- function() {
    return(c(
        "counts","sample.list","exclude.list","file.type","path","contrast",
        "libsize.list","id.col","gc.col","name.col","bt.col","annotation",
        "gene.file","org","trans.level","count.type","utr.flank","exon.filters",
        "gene.filters","when.apply.filter","normalization","norm.args",
        "statistics","stat.args","adjust.method","meta.p","weight","nperm",
        "reprod","pcut","log.offset","preset","qc.plots","fig.format",
        "out.list","export.where","export.what","export.scale","export.values",
        "export.stats","export.counts.table","restrict.cores","report","refdb",
        "report.top","report.template","verbose","run.log","save.gene.model",
        "version","local.db.home"
    ))
}

## ----install-0, eval=FALSE, echo=TRUE-----------------------------------------
# if(!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# BiocManager::install("metaseqR2")

## ----load-0, eval=TRUE, echo=FALSE, tidy=FALSE, message=FALSE, warning=FALSE----
library(metaseqR2)

## ----example-1, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
library(metaseqR2)

buildDir <- file.path(tempdir(),"test_anndb")
dir.create(buildDir)

# The location of the custom database
myDb <- file.path(buildDir,"testann.sqlite")

# Since we are using Ensembl, we can also ask for a version
organisms <- list(mm39=115)
sources <- ifelse(.Platform$OS.type=="unix",c("ensembl","refseq"),"ensembl")

# If the example is not running in a multicore system, rc is ignored
#buildAnnotationDatabase(organisms,sources,forceDownload=FALSE,db=myDb,rc=0.5)

## ----example-2, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# # Load standard annotation based on gene body coordinates
# genes <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="gene",
#     db=myDb)
# genes
# 
# # Load standard annotation based on 3' UTR coordinates
# utrs <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="utr",
#     db=myDb)
# utrs
# 
# # Load summarized exon annotation based used with RNA-Seq analysis
# sumEx <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="exon",
#     summarized=TRUE,db=myDb)
# sumEx
# 
# # Load standard annotation based on gene body coordinates from RefSeq
# if (.Platform$OS.type=="unix") {
#     refGenes <- loadAnnotation(genome="mm39",refdb="refseq",level="gene",
#         type="gene",db=myDb)
#     refGenes
# }

## ----example-3, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# # Load standard annotation based on gene body coordinates
# genes <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="gene",
#     db=myDb,asdf=TRUE)
# head(genes)

## ----example-4, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# # Setup a temporary directory to download files etc.
# customDir <- file.path(tempdir(),"test_custom")
# dir.create(customDir)
# 
# # Convert from GenePred to GTF - Unix/Linux only!
# if (.Platform$OS.type == "unix" && !grepl("^darwin",R.version$os)) {
#     # Download data from UCSC
#     goldenPath="http://hgdownload.cse.ucsc.edu/goldenPath/"
#     # Gene annotation dump
#     download.file(paste0(goldenPath,"eboVir3/database/ncbiGene.txt.gz"),
#         file.path(customDir,"eboVir3_ncbiGene.txt.gz"))
#     # Chromosome information
#     download.file(paste0(goldenPath,"eboVir3/database/chromInfo.txt.gz"),
#         file.path(customDir,"eboVir3_chromInfo.txt.gz"))
# 
#     # Prepare the build
#     chromInfo <- read.delim(file.path(customDir,"eboVir3_chromInfo.txt.gz"),
#         header=FALSE)
#     chromInfo <- chromInfo[,1:2]
#     rownames(chromInfo) <- as.character(chromInfo[,1])
#     chromInfo <- chromInfo[,2,drop=FALSE]
# 
#     # Coversion from genePred to GTF
#     genePredToGtfEnv <- Sys.getenv("GENEPREDTOGTF_BINARY")
#     if (genePredToGtfEnv == "") {
#         genePredToGtf <- file.path(customDir,"genePredToGtf")
#     } else {
#         genePredToGtf <- file.path(genePredToGtfEnv)
#     }
#     if (!file.exists(genePredToGtf)) {
#         download.file(
#         "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
#             genePredToGtf
#         )
#         system(paste("chmod 775",genePredToGtf))
#     }
#     gtfFile <- file.path(customDir,"eboVir3.gtf")
#     tmpName <- file.path(customDir,paste(format(Sys.time(),"%Y%m%d%H%M%S"),
#         "tgtf",sep="."))
#     command <- paste0(
#         "zcat ",file.path(customDir,"eboVir3_ncbiGene.txt.gz"),
#         " | ","cut -f2- | ",genePredToGtf," file stdin ",tmpName,
#         " -source=eboVir3"," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",
#         gtfFile
#     )
#     system(command)
# 
#     # Build with the metadata list filled (you can also provide a version)
#     buildCustomAnnotation(
#         gtfFile=gtfFile,
#         metadata=list(
#             organism="eboVir3_test",
#             source="ucsc_test",
#             chromInfo=chromInfo
#         ),
#         db=myDb
#     )
# 
#     # Try to retrieve some data
#     eboGenes <- loadAnnotation(genome="eboVir3_test",refdb="ucsc_test",
#         level="gene",type="gene",db=myDb)
#     eboGenes
# }

## ----example-5, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# if (.Platform$OS.type == "unix") {
#     # Gene annotation dump
#     download.file(paste0(goldenPath,"gadMor1/database/augustusGene.txt.gz"),
#         file.path(customDir,"gadMori1_augustusGene.txt.gz"))
#     # Chromosome information
#     download.file(paste(goldenPath,"gadMor1/database/chromInfo.txt.gz",sep=""),
#         file.path(customDir,"gadMori1_chromInfo.txt.gz"))
# 
#     # Prepare the build
#     chromInfo <- read.delim(file.path(customDir,"gadMori1_chromInfo.txt.gz"),
#         header=FALSE)
#     chromInfo <- chromInfo[,1:2]
#     rownames(chromInfo) <- as.character(chromInfo[,1])
#     chromInfo <- chromInfo[,2,drop=FALSE]
# 
#     # Coversion from genePred to GTF
#     genePredToGtf <- file.path(customDir,"genePredToGtf")
#     if (!file.exists(genePredToGtf)) {
#         download.file(
#         "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
#             genePredToGtf
#         )
#         system(paste("chmod 775",genePredToGtf))
#     }
#     gtfFile <- file.path(customDir,"gadMori1.gtf")
#     tmpName <- file.path(customDir,paste(format(Sys.time(),"%Y%m%d%H%M%S"),
#         "tgtf",sep="."))
#     command <- paste0(
#         "zcat ",file.path(customDir,"gadMori1_augustusGene.txt.gz"),
#         " | ","cut -f2- | ",genePredToGtf," file stdin ",tmpName,
#         " -source=gadMori1"," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",
#         gtfFile
#     )
#     system(command)
# 
#     # Build with the metadata list filled (you can also provide a version)
#     buildCustomAnnotation(
#         gtfFile=gtfFile,
#         metadata=list(
#             organism="gadMor1_test",
#             source="ucsc_test",
#             chromInfo=chromInfo
#         ),
#         db=myDb
#     )
# 
#     # Try to retrieve some data
#     gadGenes <- loadAnnotation(genome="gadMor1_test",refdb="ucsc_test",
#         level="gene",type="gene",db=myDb)
#     gadGenes
# }

## ----example-6, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# # Gene annotation dump from Ensembl
# download.file(paste0("ftp://ftp.ensembl.org/pub/release-98/gtf/",
#     "dasypus_novemcinctus/Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#     file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"))
# 
# # Chromosome information will be provided from the following BAM file
# # available from Ensembl. We have noticed that when using Windows as the OS,
# # a remote BAM files cannot be opened by scanBamParam, so for this example,
# # chromosome length information will not be available when running in Windows.
# bamForInfo <- NULL
# if (.Platform$OS.type == "unix")
#     bamForInfo <- paste0("ftp://ftp.ensembl.org/pub/release-98/bamcov/",
#         "dasypus_novemcinctus/genebuild/Dasnov3.broad.Ascending_Colon_5.1.bam")
# 
# # Build with the metadata list filled (you can also provide a version)
# buildCustomAnnotation(
#     gtfFile=file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#     metadata=list(
#         organism="dasNov3_test",
#         source="ensembl_test",
#         chromInfo=bamForInfo
#     ),
#     db=myDb
# )
# 
# # Try to retrieve some data
# dasGenes <- loadAnnotation(genome="dasNov3_test",refdb="ensembl_test",
#     level="gene",type="gene",db=myDb)
# dasGenes

## ----example-7, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE----
# organisms <- list(
#     hg18=54,
#     hg19=75,
#     hg38=114:115,
#     mm9=54,
#     mm10=102,
#     mm39=114:115,
#     rn5=77,
#     rn6=114:115,
#     dm3=77,
#     dm6=114:115,
#     danrer7=77,
#     danrer10=80,
#     danrer11=114:115,
#     pantro4=80,
#     pantro5=114:115,
#     susscr3=80,
#     susscr11=114:115,
#     equcab2=77
# )
# 
# sources <- c("ensembl","ucsc","refseq")
# 
# buildAnnotationDatabase(organisms,sources,forceDownload=FALSE,rc=0.5)

## ----pseudo-1, eval=TRUE, echo=TRUE, message=TRUE, warning=FALSE--------------
annotation <- list(
    gtf="PATH_TO_GTF",
    organism="ORGANISM_NAME",
    source="SOURCE_NAME",
    chromInfo="CHROM_INFO"
)

## ----si-1, eval=TRUE, echo=TRUE-----------------------------------------------
sessionInfo()

################################################################################

# classes.R

setClass("Biodetection",representation(dat="list"))
setClass("CD",representation(dat="list"))
setClass("CountsBio",representation(dat="list"))
setClass("GCbias",representation(dat="list"))
setClass("lengthbias",representation(dat="list"))
setClass("Saturation",representation(dat="list"))
setClass("PCA",representation(dat="list"))

setGeneric("explo.plot",function(object,...) standardGeneric("explo.plot"))

setMethod("explo.plot","Biodetection",
    function(object,samples=c(1,2),plottype=c("persample","comparison"),
        toplot="protein_coding",...)
        biodetection.plot(object@dat,samples=samples,plottype=plottype,
            toplot=toplot,...))

setMethod("explo.plot","CD",function(object,samples=NULL,...) 
    cd.plot(object@dat,samples=samples,...))

setMethod("explo.plot","CountsBio",
    function(object,samples=c(1,2),toplot="global",
        plottype=c("barplot","boxplot"),...)
        countsbio.plot(object@dat,samples,toplot,plottype,...))

#setMethod("explo.plot","GCbias",
#    function(object,samples=NULL,toplot="global",...)
#    GC.plot(object@dat,samples=samples,toplot=toplot,...))

#setMethod("explo.plot","lengthbias",
#    function(object,samples=NULL,toplot="global",...)
#    length.plot(object@dat,samples=samples,toplot=toplot,...))

#setMethod("explo.plot","Saturation",
#    function(object,samples=NULL,toplot=1,yleftlim=NULL,yrightlim=NULL,...)
#    saturation.plot(object@dat,samples=samples,toplot=toplot,yleftlim=yleftlim,
#        yrightlim=yrightlim,...))

#setMethod("explo.plot","PCA",
#    function(object,samples=c(1,2),plottype="scores",factor=NULL)
#    PCA.plot(object@dat,samples=samples,plottype=plottype,factor=factor))


# Show methods for exploration objects
setMethod("show","Biodetection",function(object) {
    cat("\n Reference genome: \n==========\n")
    names(dimnames(object@dat$genome))=NULL
    print(object@dat$genome)
    for (i in seq_len(length(object@dat$biotables))) {
        cat("\n",names(object@dat$biotables)[i],"\n==========\n")
        print(object@dat$biotables[[i]])
    }
})

setMethod("show","CD",function(object) {
    cat("\n Confidence intervals for median of M to compare each sample ",
        "to reference:\n=======\n")
    print(object@dat$DiagnosticTest)
    cat("\n Reference sample is:\n=======\n")
    print(object@dat$refColumn)
})

setMethod("show","CountsBio",function(object) {
    cat("\n Summary: \n============\n")              
    print(object@dat$summary[[1]])            
})

setMethod("show","GCbias",function(object) {
    x <- object@dat$RegressionModels
    for (i in seq_len(length(x))) {
        print(names(x)[i])
        print(summary(x[[i]]))
    }
})

setMethod("show","lengthbias",function(object) {
    x <- object@dat$RegressionModels
    for (i in seq_len(length(x))) {
        print(names(x)[i])
        print(summary(x[[i]]))
    }
})

setMethod("show","Saturation",function(object) {
    x <- dat2save(object)
    cat("\n Number of detected features at each sequencing ",
        "depth: \n============\n")
    for (i in seq_len(length(x))) {
        print(names(x)[i])
        print(x[[i]])
    }            
})

setMethod("show","PCA",function(object) {
    x <- object$result$var.exp
    x <- round(x*100,4)
    colnames(x) <- c("%Var","Cum %Var")
    rownames(x) <- paste("PC",seq_len(nrow(x)))
    cat("\n Percentage of total variance explained by each component: ",
        "\n============\n")
    print(x)
})


# Coercion methods for exploration objects
setGeneric("dat2save",function(object) standardGeneric("dat2save"))
setMethod("dat2save","Biodetection",function(object) object@dat)
setMethod("dat2save","CD",function(object) object@dat)
setMethod("dat2save","CountsBio",function(object) object@dat$summary)
setMethod("dat2save","GCbias",function(object) object@dat$data2plot)
setMethod("dat2save","lengthbias",function(object) object@dat$data2plot)
setMethod("dat2save","Saturation",function(object) {
    muestras=vector("list",length=length(object@dat$depth))
    names(muestras)=names(object@dat$depth)
    for (i in seq_len(length(muestras))) {
        muestras[[i]] <- object@dat$depth[[i]]
        for (j in seq_len(length(object@dat$saturation))) {
            muestras[[i]] <- cbind(muestras[[i]],
                object@dat$saturation[[j]][[i]])
        }
        colnames(muestras[[i]]) <- c("depth",names(object@dat$saturation))
    }

    muestras
})

setMethod("dat2save","PCA",function(object) object@dat$result)

setClass("myInfo",representation(method="character",k="numeric",
    lc="numeric",factor="vector",v="numeric",nss="numeric",pnr="numeric",
    comparison="vector",replicates="character"))
setClass("Output",representation(results="list"),contains="myInfo")

setValidity("Output",function(object) {
    if (!(is.character(object@method))) {
        return(paste("Method must be a string"))
    } else if (!(is.numeric(object@k))) {
        return(paste("k must be numeric"))
    } else if (!(is.numeric(object@lc))) {
        return(paste("lc must be numeric"))
    } else if (!(is.vector(object@factor))) {
        return(paste("Factor must be a vector of strings"))
    } else if (!(is.numeric(object@v))) {
        return(paste("v must be numeric"))
    } else if (!(is.numeric(object@nss))) {
        return(paste("nss must be numeric"))
    } else if (!(is.numeric(object@pnr))) {
        return(paste("pnr must be numeric"))
    } else if (!(is.vector(object@comparison))) {
        return(paste("Comparison must be a vector of strings"))
    } else if (!(is.list(object@results))) {
        return(paste("Results must be a list of data.frames"))
    } else {
        return(TRUE)
    }
})

Output <- function (data,method,k,lc,factor,v,nss,pnr,comparison,replicates) {
    new("Output",results=data,method=method,k=k,lc=lc,factor=factor,v=v,nss=nss,
        pnr=pnr,comparison=comparison,replicates=replicates)
}

setMethod("show","Output",function(object) {
    if (object@method == "n")
        object@method="none"
    for (i in seq_len(length(object@results))) {
        cat("\nSummary",i,"\n=========\n")
        cat("\nYou are comparing",object@comparison[i],"from",object@factor[i],
            "\n\n")
        print(head(object@results[[i]][order(object@results[[i]][,5],
            decreasing=TRUE),]))
    }
    cat("\nNormalization\n")
    cat("\tmethod:",object@method,"\n")
    cat("\tk:",object@k,"\n")
    cat("\tlc:",object@lc,"\n")

    # Simulated samples
    if (object@replicates == "no") {
        cat("\nYou are working with simulated replicates:\n")
        cat("\tpnr:",object@pnr,"\n")
        cat("\tnss:",object@nss,"\n")
        cat("\tv:",object@v,"\n")
    }
    # With biological or technical replicates
    else {
        cat("\nYou are working with",object@replicates,"replicates\n")
    }           
})

################################################################################

################################################################################

# allMD.R

allMD <- function (input,factor,conditions,k=0.5,replicates,norm="rpkm",pnr=0.2,
    nss=5,v=0.02,lc=0) {
    # Check if the factor introduced is already defined
    # If the factor introduced is defined and has more than 2 conditions,it will
    # check if the conditions specified are defined too
    condition_fac <- FALSE
    condition_lev <- FALSE

    datos1 <- datos2 <- matrix()

    for (i in colnames(pData(input))) {
        if (factor == i) {
            condition_fac <- TRUE
            if (!is.factor(pData(input)[,i])) 
                pData(input)[,i]=as.factor(pData(input)[,i])
            if (length(levels(pData(input)[,i])) == 2) {
                if (!is.null(assayData(input)$exprs)) {
                    datos1 <- 
                        assayData(input)$exprs[,which(
                            pData(input)[,i]==levels(pData(input)[,i])[1]),
                            drop=FALSE]
                    datos2 <- 
                        assayData(input)$exprs[,which(
                            pData(input)[,i] ==levels(pData(input)[,i])[2]),
                            drop=FALSE]
                } 
                else {
                    datos1 <- 
                        assayData(input)$counts[,which(
                            pData(input)[,i]==levels(pData(input)[,i])[1]),
                            drop=FALSE]
                    datos2 <- assayData(input)$counts[,which(
                        pData(input)[,i]==levels(pData(input)[,i])[2]),
                            drop=FALSE]
                }

                # Define the comparison string
                comparison <- paste(levels(pData(input)[,i])[1],
                    levels(pData(input)[,i])[2],sep=" - ")

                condition_lev <- TRUE
            }

            else {
                if (is.null(conditions))
                    stop("Error. You must specify which conditions you wish ",
                        "to compare when the factor has two or more ",
                        "conditions.\n")
                if (length(conditions) != 2)
                    stop("Error. The argument conditions must contain the 2 ",
                        "conditions you wish to compare.")
            
                l <- conditions %in% pData(input)[,i]
                # If they are defined,they will be TRUE
                if (l[1] == TRUE && l[2] == TRUE) {
                    if (!is.null(assayData(input)$exprs)) {
                        datos1 <- assayData(input)$exprs[,which(
                            pData(input)[,i] == conditions[1]),drop=FALSE]
                        datos2 <- assayData(input)$exprs[,which(
                            pData(input)[,i] == conditions[2]),drop=FALSE]
                    }
                    else {
                        datos1 <- assayData(input)$counts[,which(
                            pData(input)[,i] == conditions[1]),drop=FALSE]
                        datos2 <- assayData(input)$counts[,which(
                            pData(input)[,i] == conditions[2]),drop=FALSE]
                    }
                    # Define the comparison string
                    comparison <- paste(conditions[1],conditions[2],sep=" - ")
                    condition_lev <- TRUE
                }
            }
        }
    }

    if (!condition_fac)
        stop("The factor you have written does not correspond with any of the ",
            "ones you have defined.")
            
    if (!condition_lev)
        stop("The conditions you have written don't exist in the factor ",
            "specified.\n")

    # Correction to make it work when there are simulated samples
    if (replicates == "no")
        replicates <- "technical"

    n1 <- ncol(as.matrix(datos1))
    n2 <- ncol(as.matrix(datos2))

    if (norm == "n") { # no normalization
        datos1 <- round(datos1,100)
        datos2 <- round(datos2,100)
    }

    if (is.null(k)) {
        m1 <- min(datos1[noceros(datos1,num=FALSE)],na.rm=TRUE)
        m2 <- min(datos2[noceros(datos2,num=FALSE)],na.rm=TRUE)
        mm <- min(m1,m2)
        k <- mm/2    
    } 

    # Total counts for each gene:
    suma1 <- rowSums(as.matrix(datos1))
    suma2 <- rowSums(as.matrix(datos2))

    # All genes
    todos <- rownames(as.matrix(datos1))

    # Genes with counts in any condition
    concounts <- names(which(suma1+suma2 > 0))

    long <- 1000
    g.sinL <- NULL

    if (!is.null(featureData(input)@data$Length)) {
        g.sinL <- names(which(is.na(featureData(input)@data$Length)))
        if (any(!is.na(featureData(input)@data$Length)) == TRUE) 
        long <- featureData(input)@data[concounts,"Length"]
    }

    if (replicates == "technical") { ### technical replicates
        suma1 <- suma1[concounts]
        suma2 <- suma2[concounts]    

        #----------------------------------------------------------------------#
        # Normalization of counts for each condition (aggregating replicates)

        if (norm == "rpkm") {      # RPKM
            suma1.norm <- noirpkm(suma1,long=long,k=k,lc=lc)
            suma2.norm <- noirpkm(suma2,long=long,k=k,lc=lc)
        }

        if (norm == "uqua") {
            suma.norm <- noiuqua(cbind(suma1,suma2),long=long,lc=lc,k=k)
            suma1.norm <- as.matrix(suma.norm[ ,1])
            suma2.norm <- as.matrix(suma.norm[ ,2])
        }

        if (norm == "tmm") {
            suma.norm <- noitmm(as.matrix(cbind(suma1,suma2)),long=long,lc=lc,
                k=k)
            suma1.norm <- as.matrix(suma.norm[ ,1])
            suma2.norm <- as.matrix(suma.norm[ ,2])
        }
    }

    #-------------------------------------------------------------------------#

    ## Noise distribution

    if ((n1+n2)>2) {   # with real samples
        datitos <- cbind(datos1,datos2)
        datitos <- datitos[concounts,]

        gens.sin0 <- setdiff(concounts,g.sinL)

        if (norm == "n") {       # no normalization
            datitos.0 <- sinceros(datitos,k=k)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        if (norm == "rpkm") {      # RPKM
            datitos.0 <- noirpkm(datitos,long=long,k=k,lc=lc)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        if (norm == "uqua") {      # Upper Quartile
            datitos.0 <- noiuqua(datitos,long=long,lc=lc,k=k)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        if (norm == "tmm") {
            datitos.0 <- noitmm(datitos,long=long,lc=lc,k=k)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        datos1.norm <- datitos.norm[ ,seq_len(n1)]
        datos2.norm <- datitos.norm[ ,(n1+1):(n1+n2)]

        if (n1 > 1) {
            MD1 <- MD(dat=datos1.norm)
        } 
        else { 
            MD1 <- NULL 
        }

        if (n2 > 1) {
            MD2 <- MD(dat=datos2.norm)
        } 
        else { 
            MD2 <- NULL 
        }
    } 
    else { # with simulated samples
        if (nss == 0) {
            nss <- 5
        }
        
        datos.sim <- sim.samples(counts1=sinceros(suma1,k=k),
            counts2=sinceros(suma2,k=k),pnr=pnr,nss=nss,v=v)

        nn <- vapply(datos.sim,ncol,numeric(1))
        dat.sim.norm <- vector("list",length=2)
        datitos <- cbind(datos.sim[[1]],datos.sim[[2]])
        rownames(datitos)=names(suma1)

        sumita <- rowSums(datitos)
        g.sin0 <- names(which(sumita > 0))
        gens.sin0 <- setdiff(g.sin0,g.sinL)

        if (norm == "n") {       # no normalization
            datitos.0 <- sinceros(datitos,k=k)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        if (norm == "rpkm") {      # RPKM
            datitos.0 <- noirpkm(datitos,long=long,k=k,lc=lc)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        if (norm == "uqua") {      # Upper Quartile
            datitos.0 <- noiuqua(datitos,long=long,lc=lc,k=k)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        if (norm == "tmm") {
            datitos.0 <- noitmm(datitos,long=long,lc=lc,k=k)
            datitos.norm <- datitos.0[gens.sin0,]
        }

        dat.sim.norm[[1]] <- datitos.norm[,seq_len(nn[1])]
        dat.sim.norm[[2]] <- datitos.norm[,(nn[1]+1):sum(nn)]

        MD1 <- MD(dat=dat.sim.norm[[1]])
        MD2 <- MD(dat=dat.sim.norm[[2]])
    }

    Mr <- c(as.numeric(MD1$M),as.numeric(MD2$M))
    Dr <- c(as.numeric(MD1$D),as.numeric(MD2$D))

    #-------------------------------------------------------------------------#

    ## M and D for different experimental conditions

    if (replicates == "technical" & norm != "n") {
        MDs <- MD(dat=cbind(suma1.norm,suma2.norm))

        lev1 <- suma1.norm[,1]
        lev1 <- lev1[todos]
        
        lev2 <- suma2.norm[,1]
        lev2 <- lev2[todos]
    } 
    else {
        if ((n1+n1) == 2) {
            datos1.norm <- sinceros(as.matrix(datos1)[concounts,],k=k)
            datos2.norm <- sinceros(as.matrix(datos2)[concounts,],k=k)      
        }  

        resum1.norm <- rowMeans(as.matrix(datos1.norm))
        resum2.norm <- rowMeans(as.matrix(datos2.norm))

        lev1 <- resum1.norm[todos]
        lev2 <- resum2.norm[todos]

        MDs <- MD(dat=cbind(resum1.norm,resum2.norm))
    }

    ## Completing M and D
    names(lev1) <- names(lev2) <- todos

    Ms <- as.numeric(MDs$M)
    names(Ms) <- rownames(MDs$M)
    Ms <- Ms[todos]
    names(Ms) <- todos

    Ds <- as.numeric(MDs$D)
    names(Ds) <- rownames(MDs$D)
    Ds <- Ds[todos]
    names(Ds) <- todos

    ## Results
    list("k"=k,"comp"=comparison,"Level1"=lev1,"Level2"=lev2,"Ms"=Ms,"Ds"=Ds,
        "Mn"=Mr,"Dn"=Dr)
}

allMDbio <- function(input,factor,conditions,k=0.5,norm="rpkm",lc=1,r=10,
    a0per=0.9,nclust=15,filter=1,depth=NULL,cv.cutoff=0,cpm=1) {

    # Check if the factor introduced is already defined
    # If the factor introduced is defined and has more than 2 conditions,
    # it will check if the conditions specified are defined too
    condition_fac <- FALSE
    condition_lev <- FALSE

    datos1 <- datos2 <- matrix()

    for (i in colnames(pData(input))) {
        if (factor == i) {
            condition_fac <- TRUE

            if (!is.factor(pData(input)[,i])) 
                pData(input)[,i] <- as.factor(pData(input)[,i])

            if (length(levels(pData(input)[,i])) == 2) {
                if (!is.null(assayData(input)$exprs)) {
                    datos1 <- assayData(input)$exprs[,which(
                        pData(input)[,i]==levels(pData(input)[,i])[1])]
                    datos2 <- assayData(input)$exprs[,which(
                        pData(input)[,i] ==levels(pData(input)[,i])[2])]
                } 
                else {
                    datos1 <- assayData(input)$counts[,which(
                        pData(input)[,i] ==levels(pData(input)[,i])[1])]
                    datos2 <- assayData(input)$counts[,which(
                        pData(input)[,i] ==levels(pData(input)[,i])[2])]
                }

            # Define the comparison string
            comparison <- paste(levels(pData(input)[,i])[1],
                levels(pData(input)[,i])[2],sep=" - ")

            condition_lev <- TRUE

            if (!((ncol(datos1) > 1) && (ncol(datos2) > 1)))
                stop("Error. NOISeqBIO needs at least 2 biological replicates ",
                    "per condition.\n")
            }
            else {
                if (is.null(conditions))
                    stop("Error. You must specify which conditions you wish ",
                        "to compare when the factor has two or more ",
                        "conditions.\n")
                if (length(conditions) != 2)
                    stop("Error. The argument conditions must contain the 2 ",
                        "conditions you wish to compare.")
            
                l <- conditions %in% pData(input)[,i]
                # If they are defined,they will be TRUE
                if (l[1] == TRUE && l[2] == TRUE) {
                    if (!is.null(assayData(input)$exprs)) {
                        datos1 <- assayData(input)$exprs[,which(
                            pData(input)[,i] == conditions[1])]
                        datos2 <- assayData(input)$exprs[,which(
                            pData(input)[,i] == conditions[2])]
                    }
                    else {
                        datos1 <- assayData(input)$counts[,which(
                            pData(input)[,i] == conditions[1])]
                        datos2 <- assayData(input)$counts[,which(
                            pData(input)[,i] == conditions[2])]
                    }
                    
                    # Define the comparison string
                    comparison <- paste(conditions[1],conditions[2],sep=" - ")
                    condition_lev <- TRUE
                }
            }
        }
    }

    if (condition_fac == FALSE)
        stop("The factor specified does not correspond with any of the ",
            "ones you have defined.") 
    if (condition_lev == FALSE)
        stop("The conditions specified don't exist for the factor specified.\n")

    #--------------------------------------------------------------------------#

    # Number of observations within each condition 
    n1 <- ncol(as.matrix(datos1))
    n2 <- ncol(as.matrix(datos2))
    if (max(n1,n2) == 1) 
        stop("There is only one replicate per condition. Please,use ",
            "NOISeq instead of NOISeqBIO.\n")

    # Rounding off data
    if (norm == "n") {      # no normalization
        datos1 <- round(datos1,10)
        datos2 <- round(datos2,10)
    }

    # Computing k
    if (is.null(k)) {
        m1 <- min(datos1[noceros(datos1,num=FALSE)],na.rm=TRUE)
        m2 <- min(datos2[noceros(datos2,num=FALSE)],na.rm=TRUE)
        k <- min(m1,m2)/2          
    }

    # Total counts for each gene:
    suma1 <- rowSums(as.matrix(datos1))
    suma2 <- rowSums(as.matrix(datos2))

    # Genes with counts in any condition
    concounts <- names(which(suma1+suma2 > 0))

    # All genes
    todos <- rownames(as.matrix(datos1))

    # Gene length
    long <- 1000
    g.sinL <- NULL  # genes with no length defined

    if (!is.null(featureData(input)@data$Length)) {
        g.sinL <- names(which(is.na(featureData(input)@data$Length)))
        if (any(!is.na(featureData(input)@data$Length)) == TRUE) 
            long <- featureData(input)@data[concounts,"Length"]
    }

    # Genes with counts and with length
    gens.sin0 <- setdiff(concounts,g.sinL)

    # cond1 and cond2 in the same matrix
    datitos <- cbind(datos1,datos2)  
    datitos <- datitos[concounts,]  # selecting only genes with counts

    # Sequencing depth when filtering method=3
    if (filter == 3 && is.null(depth)) 
        depth <- colSums(datitos)

    #-------------------------------------------------------------------------#
    #-------------------------------------------------------------------------#

    ## Normalization  
    if (norm == "n") {       # no normalization
        datitos.0 <- sinceros(datitos,k=k)
        datitos.norm <- datitos.0[gens.sin0,]
    }

    if (norm == "rpkm") {      # RPKM
        datitos.0 <- noirpkm(datitos,long=long,k=k,lc=lc)
        datitos.norm <- datitos.0[gens.sin0,]
    }

    if (norm == "uqua") {      # Upper Quartile
        datitos.0 <- noiuqua(datitos,long=long,lc=lc,k=k)
        datitos.norm <- datitos.0[gens.sin0,] 
    }

    if (norm == "tmm") {
        datitos.0 <- noitmm(datitos,long=long,lc=lc,k=k)
        datitos.norm <- datitos.0[gens.sin0,]       
    }

    #-------------------------------------------------------------------------#

    ## Filtering out low count features
    if (filter != 0) {
        datos.filt <- filtered.data(dataset=datitos.norm,
            factor=c(rep("cond1",n1),rep("cond2",n2)),norm=TRUE,depth=depth,
            method=filter,cv.cutoff=cv.cutoff,cpm=cpm)
    } 
    else {
        datos.filt <- datitos.norm
    }

    datos1.filt <- datos.filt[,seq_len(n1)]
    datos2.filt <- datos.filt[,(n1+1):(n1+n2)]   

    #-------------------------------------------------------------------------#

    ## Noise distribution

    Zr <- NULL

    if (n1+n2 <= 9) {  # sharing information within clusters
        Zr=share.info(mydata=datos.filt,n1=n1,n2=n2,r=r,nclust=nclust)
    } 
    else {   # r permutations
        for (i in seq_len(r)) {
            print(paste("r =",i))
            mipermu <- sample(seq_len(n1+n2))
            mipermu <- datos.filt[,mipermu]
            mean1 <- rowMeans(mipermu[,seq_len(n1)])
            mean2 <- rowMeans(mipermu[,(n1+1):(n1+n2)])
            sd1 <- apply(mipermu[,seq_len(n1)],1,sd)
            sd2 <- apply(mipermu[,(n1+1):(n1+n2)],1,sd)
            
            myparam <- list("n"=c(n1,n2),"sd"=cbind(sd1,sd2))
            MDperm <- MDbio(dat=cbind(mean1,mean2),param=myparam,a0per=a0per)
            Zr <- cbind(Zr,myDfunction(mydif=MDperm$D,myrat=MDperm$M,stat=1,
                coef=0.5))
        }
    }

    #-------------------------------------------------------------------------#

    ## Z-score for different experimental conditions (SIGNAL)

    mean1 <- rowMeans(as.matrix(datos1.filt))
    mean2 <- rowMeans(as.matrix(datos2.filt))        

    sd1 <- apply(as.matrix(datos1.filt),1,sd)
    sd2 <- apply(as.matrix(datos2.filt),1,sd)

    myparam=list("n"=c(n1,n2),"sd"=cbind(sd1,sd2))

    MDs <- MDbio(dat=cbind(mean1,mean2),param=myparam,a0per=a0per)

    Zs <- myDfunction(mydif=MDs$D,myrat=MDs$M,stat=1,coef=0.5)

    #-------------------------------------------------------------------------#

    ## Completing M and D (in signal)

    lev1 <- mean1[todos]
    lev2 <- mean2[todos]
    names(lev1) <- names(lev2) <- todos

    Zs <- as.numeric(Zs)
    names(Zs) <- rownames(MDs$M)

    Zs <- Zs[todos]
    names(Zs) <- todos

    ## Computing Zn
    Zn <- as.numeric(Zr)

    #-------------------------------------------------------------------------#

    ## Results
    list("k"=k,"comp"=comparison,"Level1"=lev1,"Level2"=lev2,"Zs"=Zs,"Zn"=Zn)
}

## Function to summarize difference and ratio information (D and D0)
myDfunction <- function (mydif,myrat,stat,coef) {
    if (stat == 1) { # linear combination of difference and ratio
        myDvalues=coef*mydif + (1-coef)*myrat
    }
    if (stat == 2) { # distance to origin from (ratio,difference)
        myDvalues=sign(mydif) * sqrt((mydif)^2 + (myrat)^2)  
    }
    myDvalues
}

################################################################################

################################################################################

# ARSyNcomponents.R

# This program selects the number of components that explain more than the 
# Variability% If Variability="average" the number of components will be those 
# that explain more than the average variation of the principal components
# For residuals model the number of components selected are 
# beta*average-variability.

ARSyNcomponents <- function(asca=asca,Variability=0.75,beta=2) {
    MODEL<-asca[-length(asca)]
    M<-length(MODEL)-1
    output<-NULL

    if (Variability=="average") {
        for (i in seq_len(M)) { 
            lim <- 1/rankMatrix(MODEL[[i]]$X)[1]    
            t <- table(MODEL[[i]]$var.exp[,1]>lim)
            if (length(t)==1) {
                t[2]=0
            }
            t <- t[2]
            names(t)<-names(MODEL)[i]
            output<-c(output,t)
        }
    }

    if (Variability!="average") {
        lim<-Variability
        for (i in seq_len(M)) { 
            t <- which(MODEL[[i]]$var.exp[,2]>lim)[1]
            names(t) <- names(MODEL)[i]
            output <- c(output,t)
        }
    }
    
    ### Residuals model
    
    i <- M+1
    lim <- beta*1/rankMatrix(MODEL[[i]]$X)[1]   
    t <- table(MODEL[[i]]$var.exp[,1]>lim)
    if (length(t)==1) {
        t[2]=0
    }
    t <- t[2]
    names(t)<-names(MODEL)[i]
    output <- c(output,t)

    output
}

################################################################################

################################################################################

# ARSyNmodel.R

ARSyNmodel <- function(X=X,Factors=2,Designa=Designa,Designb=Designb,
    Designc=Designc,Variability="average",Join=TRUE,Interaction=TRUE,beta=2) {

    if (Factors==1) {
        Fac0 <- c(1,2)
        names(Fac0) <- c("Model.a","Model.res")
        asca0 <- ASCA.1f(X=X,Designa=Designa,Fac=Fac0)
        Fac <- ARSyNcomponents(asca0,Variability=Variability,beta=beta)
        for (i in seq_len(length(Fac))) {
            Fac0[names(Fac[i])]<-Fac[names(Fac[i])]
        }
        asca <- ASCA.1f(X=X,Designa=Designa,Fac=Fac0)
    }



    if (Factors==2) {
        Fac0 <- c(1,2,2,2)
        names(Fac0) <- c("Model.a","Model.b","Model.ab","Model.res")
        if (Join[1]) {
            names(Fac0)[3] <- c("Model.bab")
        }
        asca0 <- ASCA.2f(X=X,Designa=Designa,Designb=Designb,Fac=Fac0,Join=Join,
            Interaction=Interaction)
        Fac <- ARSyNcomponents(asca0,Variability=Variability,beta=beta)
        for (i in seq_len(length(Fac))) {
            Fac0[names(Fac[i])]<-Fac[names(Fac[i])]
        }
        asca <- ASCA.2f(X=X,Designa=Designa,Designb=Designb,Fac=Fac0,Join=Join,
            Interaction=Interaction)
    }

    if (Factors==3) {
        Fac0 <- c(0,2,2,2,2,2,2,2)
        names(Fac0) <- c("Model.a","Model.b","Model.c","Model.ab","Model.ac",
            "Model.bc","Model.abc","Model.res")
        if (Join[1]) {
            names(Fac0)[4]<-c("Model.bab")
        }
        if (Join[2]) {
            names(Fac0)[5]<-c("Model.cac")
        }
        asca0 <- ASCA.3f(X=X,Designa=Designa,Designb=Designb,Designc=Designc,
            Fac=Fac0,Join=Join,Interaction=Interaction)
        Fac<-ARSyNcomponents(asca0,Variability=Variability,beta=beta)
        for (i in seq_len(length(Fac))) {
            Fac0[names(Fac[i])]<-Fac[names(Fac[i])]
        }
        asca <- ASCA.3f(X=X,Designa=Designa,Designb=Designb,Designc=Designc,
            Fac=Fac0,Join=Join,Interaction=Interaction)
    }

    Input <- asca$Input
    asca$Input <- c(Input,Factors)
    names(asca$Input) <- c(names(Input),"Factors")

    output <- asca
    output
}

################################################################################

################################################################################

# ARSyNseq.R

ARSyNseq <- function(data,factor=NULL,batch=FALSE,norm="rpkm",logtransf=FALSE,
    Variability=0.75,beta=2) {

    Join <- TRUE
    Interaction <- TRUE

    dat <- as.matrix(assayData(data)$exprs)
    long <- featureData(data)@data[rownames(dat),"Length"]
    if (is.null(long)) 
        long <- 1000

    if (norm == "rpkm") {
        dat <- noirpkm(dat,long=long,k=0,lc=1)
    }

    if (norm == "uqua") {
        dat <- noiuqua(dat,long=long,lc=1,k=0)
    }

    if (norm == "tmm") {
        dat <- noitmm(dat,long=long,lc=1,k=0)      
    }

    if (!logtransf)   dat <- log(dat + 1)

    X <- t(dat)         #conditions x genes

    #----------------------------------

    if (is.null(factor)) {
        Covariates <- t(pData(data))
        Num.factors <- nrow(Covariates)
        labels.factors <- rownames(Covariates)
        Design <- list(NULL,NULL,NULL)
        for (i in seq_len(Num.factors)) {
            x <- as.character(Covariates[i,])
            Design[[i]] <- make.ASCA.design(x)
        } 
    }
    else {
        Covariates <- pData(data)[,factor]
        Num.factors <- 1
        labels.factors <- factor
        Design <- list(NULL,NULL,NULL)
        x <- as.character(Covariates)
        Design[[1]] <- make.ASCA.design(x)
    }

    ####################################
    ### --- Execute ASCAmodel 
    ####################################

    my.asca <- ARSyNmodel(Factors=Num.factors,X=X,Designa=Design[[1]],
        Designb=Design[[2]],Designc=Design[[3]],Join=Join,
        Interaction=Interaction,Variability=Variability,beta=beta)

    #################################### 
    ### --- Writing filtered matrix 
    ####################################

    X.filtered <- X
    M <- length(my.asca)-1

    if (!batch) {
        X.filtered <- X.filtered-my.asca[[M]]$TP
    }

    if(batch) {
        X.filtered <- X.filtered-my.asca[[1]]$TP 
    }

    data.filtered <- t(X.filtered)

    if (!logtransf)
        data.filtered <- exp(data.filtered)+1
    
    exprs(data) <- data.filtered

    return(data)
}

################################################################################

# ASCA1f.R

ASCA.1f <- function(X=X,Designa=Designa,Designb=NULL,Designc=NULL,Fac=c(1,2),
    Join=NULL,Interaction=NULL) {

    n <- ncol(X)
    p <- nrow(X)
    I <- ncol(Designa)

    Faca <- Fac[1] # number components Model a (time)
    Facres <- Fac[2] # number components Residues

    #----------------------- Calculate Overall Mean ----------------------------

    offset <- apply(X,2,mean)
    Xoff <- X-(cbind(matrix(1,nrow=p,ncol=1))%*%rbind(offset))

    #-----------------------  PART I: Submodel a (TIME) ------------------------

    Model.a<-ASCAfun1(Xoff,Designa,Faca)
    Xres<-Xoff-Model.a$X

    #------------------------Collecting models ---------------------------------

    models <- ls(pattern="Model")
    output <- vector(mode="list")
    Xres <- Xoff
    for (i in seq_len(length(models))) {
        mymodel <- get(models[i], envir=environment())
        output <- c(output, list(mymodel))
        Xres <- Xres - mymodel$X
        rm(mymodel)
        gc()
    }
    names(output) <- models

    #------------------------- PART III: Submodel res --------------------------

    Model.res <- ASCAfun.res(Xres,Facres)

    Model.res <- list(Model.res)
    names(Model.res) <- c("Model.res")
    output <- c(output,Model.res)

    #------------------------- Add Input data to the Output --------------------

    Input <- list(X,Designa,Designb,Designc,Fac,Join,Interaction)
    names(Input)<-c("X", "Designa", "Designb", "Designc", "Fac", "Join",
        "Interaction")
    Input <- list(Input)
    names(Input) <- "Input"
    output <- c(output,Input)
    
    output
}

################################################################################

################################################################################

# ASCA2f.R

ASCA.2f <- function(X=X,Designa=Designa,Designb=Designb,Designc=NULL,
    Fac=c(1,2,2,2),Join = TRUE,Interaction=TRUE) {

    n <- ncol(X)
    p <- nrow(X)
    I <- ncol(Designa)
    J <- ncol(Designb)

    Faca <- Fac[1] # number components Model a (time)
    Facb <- Fac[2] # number components Model b  (second factor)
    Facab <- Fac[3] # number components Model ab (interaction) 
    Facres <- Fac[4] # number components Residues

    #----------------------- Calculate Overall Mean ----------------------------

    offset <- apply(X,2,mean)
    Xoff <- X-(cbind(matrix(1,nrow=p,ncol=1))%*%rbind(offset))

    #-----------------------  PART I: Submodel a (TIME) ------------------------

    Model.a <- ASCAfun1(Xoff,Designa,Faca)
    Xres <- Xoff-Model.a$X

    #-------------------------- PART II: Submodel b and ab ---------------------

    if (!Join) {
        Model.b<-ASCAfun1(Xoff,Designb,Facb)
        if (Interaction) {
            Model.ab<-ASCAfun2(Xoff,Designa,Designb,Facab)
        }
    }

    if (Join) {
        Model.bab <- ASCAfun12(Xoff,Designa,Designb,Facab)
    }

    #------------------------Collecting models ---------------------------------

    models <- ls(pattern="Model")
    output <- vector(mode="list")
    Xres <- Xoff
    for (i in seq_len(length(models))) {
        mymodel <- get(models[i],envir=environment())
        output <- c(output,list(mymodel))
        Xres <- Xres - mymodel$X
        rm(mymodel)
        gc()    
    }

    names(output) <- models

    #------------------------- PART III: Submodel res --------------------------

    Model.res <- ASCAfun.res(Xres,Facres)

    Model.res <- list(Model.res)
    names(Model.res) <- c("Model.res")
    output <- c(output,Model.res)

    #------------------------- Add Input data to the Output --------------------

    Input <- list(X,Designa,Designb,Designc,Fac,Join,Interaction,Xoff)
    names(Input) <- c("X","Designa","Designb","Designc","Fac","Join",
        "Interaction","Xoff")
    Input <- list(Input)
    names(Input) <- "Input"
    output <- c(output,Input)

    output
}

################################################################################

################################################################################

# ASCA3f.R

ASCA.3f <- function(X=X,Designa=Designa,Designb=Designb,Designc=Designc,
    Fac=c(1,2,2,2,2,2,2,2),Join=c(TRUE,TRUE),
    Interaction=c(TRUE,TRUE,TRUE,TRUE)) {

    n <- ncol(X)
    p <- nrow(X)
    I <- ncol(Designa)
    J <- ncol(Designb)
    K <- ncol(Designc)

    Faca <- Fac[1]# number components Model a (time)
    Facb <- Fac[2] # number components Model b  (second factor)
    Facc <- Fac[3] # number components Model c  (third factor)
    Facab <- Fac[4] # number components Model ab (interaction) 
    Facac <- Fac[5] # number components Model ac (interaction) 
    Facbc <- Fac[6] # number components Model bc (interaction) 
    Facabc <- Fac[7]  # number components Model abc (interaction) 
    Facres <- Fac[8]
    
    #----------------------- Calculate Overall Mean ----------------------------

    offset <- apply(X,2,mean)
    Xoff <- X-(cbind(matrix(1,nrow=p,ncol=1))%*%rbind(offset))

    #-----------------------  PART I: Submodel a (TIME) ------------------------

    Model.a <- ASCAfun1(Xoff,Designa,Faca)
    Xres <- Xoff-Model.a$X

    #-------------------------- PART II.1: Submodel b.ab------------------------
    if(!Join[1]) {
        Model.b <- ASCAfun1(Xoff,Designb,Facb)
        if (Interaction[1]) {
            Model.ab <- ASCAfun2(Xoff,Designa,Designb,Facab)
        }
    }

    if(Join[1]) {
        Model.bab<-ASCAfun12(Xoff,Designa,Designb,Facab)
    }
    
    #-------------------------- PART II.2: Submodel (c.ac) ---------------------
    if(!Join[2]) {
        Model.c <- ASCAfun1(Xoff,Designc,Facc)
        if (Interaction[2]) {
            Model.ac<-ASCAfun2(Xoff,Designa,Designc,Facac)
        }
    }

    if(Join[2]) {
        Model.cac<-ASCAfun12(Xoff,Designa,Designc,Facac)
    }
    
    #-------------------------- PART II.3: Submodel (bc) -----------------------

    if (Interaction[3]) {
        Model.bc<-ASCAfun2(Xoff,Designb,Designc,Facbc)
    }
    
    #-------------------------- PART II.4: Submodel (abc) ----------------------

    if (Interaction[4]) {
        Model.abc<-ASCAfun.triple(Xoff,Designa,Designb,Designc,Facabc)
    }
    #------------------------Collecting models ---------------------------------

    models <- ls(pattern="Model")
    output <- vector(mode="list")
    Xres <- Xoff
    for (i in seq_len(length(models))) {
        mymodel <- get(models[i], envir=environment())
        output <- c(output, list(mymodel))
        Xres <- Xres - mymodel$X
        rm(mymodel)
        gc()
    }
    names(output) <- models

    #------------------------- PART III: Submodel res --------------------------

    Model.res <- ASCAfun.res(Xres,Facres)

    Model.res <- list(Model.res)
    names(Model.res) <- c("Model.res")
    output <- c(output,Model.res)

    #------------------------- Add Input data to the Output --------------------

    Input <- list(X, Designa, Designb, Designc, Fac, Join,Interaction,Xoff)
    names(Input) <- c("X","Designa","Designb","Designc","Fac","Join",
        "Interaction","Xoff")
    Input <- list(Input)
    names(Input) <- "Input"
    output<-c(output,Input)

    output
}

################################################################################

################################################################################

# ASCAfun1.R

ASCAfun1 <- function(X,Design,Fac) {
    n <- ncol(X) # number of genes
    I <- ncol(Design) # number of levels in the factor

    NK <- NULL
    XK <- matrix(NA,nrow=I,ncol=n)

    for (i in seq_len(I)) {
        sub <- X[Design[,i]==1,]
        if (is.null(nrow(sub))) { #when there isn't replicates
            NK[i] <- 1
            XK[i,] <- sub 
        }
        else {
        NK[i]<-nrow(sub)
        XK[i,]<-apply(sub,2,mean) }
    }
    NK<-sqrt(NK)

    # Weigh the data of the Submodel with the corresponding number of 
    # measurement occasions

    XKw <- NK*XK

    PCA <- PCA.GENES(XKw)
    scw <- PCA$scores[,seq_len(Fac)]
    ld <- PCA$loadings[,seq_len(Fac)]
    ssq <- PCA$var.exp
    if (Fac==1) {
        scw <- as.matrix(scw)
        ld <- as.matrix(ld)
    }      
    if (Fac==0) {
        scw<-as.matrix(rep(0,I))
        ld<-as.matrix(rep(0,n))
    }        
    
    # Re-weigth the scores
    sc <- scw/NK
    XKrec <- sc%*%t(ld)

    Xa <- NULL
    TPa <- NULL
    for (i in seq_len(nrow(X))) {
        position<-which(Design[i,]==1)
        Xa<-rbind(Xa,XK[position,])
        TPa<-rbind(TPa,XKrec[position,])
    }

    Ea <- Xa-TPa

    #Leverage & SPE 
    leverage<-apply(ld^2,1,sum)
    SPE<-apply(Ea^2,2,sum)

    output <- list(XK,sc,ld,ssq,Xa,TPa,Ea,leverage,SPE)
    names(output) <- c("data","scores","loadings","var.exp","X","TP","E",
        "leverage","SPE")
    
    output
}

################################################################################

################################################################################

# ASCAfun2.R

ASCAfun2 <- function(X,Desa,Desb,Fac) {
    n <- ncol(X) # number of genes
    I <- ncol(Desa) # number of levels in the factor TIME
    J <- ncol(Desb) # number of levels in the other factor

    XK1 <- matrix(NA,nrow=I,ncol=n)

    for (i in seq_len(I)) {
        sub <- X[Desa[,i]==1,]
        if (is.null(nrow(sub))) { #when there isn't replicates
            XK1[i,] <- sub 
        }
        else {
            XK1[i,] <- apply(sub,2,mean) 
        }
    }

    XK2 <- matrix(NA,nrow=J,ncol=n)

    for (j in seq_len(J)) {
        sub <- X[Desb[,j]==1,]
        if (is.null(nrow(sub))) { #when there isn't replicates
            XK2[j,] <- sub 
        }
        else {
            XK2[j,]<-apply(sub,2,mean) 
        }
    }

    NK <- matrix(NA,nrow=I,ncol=J)
    XK <- matrix(NA,nrow=I*J,ncol=n)

    k=1
    for (j in seq_len(J)){
        for (i in seq_len(I)) {
            sub <- X[(Desa[,i]+Desb[,j])==2,]
            if (is.null(nrow(sub))) {  #when there isn't replicates
                NK[i,j] <- 1
                XK[k,] <- sub-XK1[i,]-XK2[j,]
            }
            else {
                NK[i,j] <- sqrt(nrow(sub))
                XK[k,] <- apply(sub,2,mean)-XK1[i,]-XK2[j,] 
            }
            k <- k+1
        }
    }

    XKw <- XK*(as.numeric(NK))

    PCA <- PCA.GENES(XKw)
    scw <- PCA$scores[,seq_len(Fac)]
    ld <- PCA$loadings[,seq_len(Fac)]
    ssq <- PCA$var.exp
    if (Fac==1) {
        scw <- as.matrix(scw)
        ld <- as.matrix(ld) 
    }
    if (Fac==0) {
        scw <- as.matrix(rep(0,I*J))
        ld <- as.matrix(rep(0,n))
    }
    
    # Re-weigth the scores
    sc <- scw/(as.numeric(NK))
    
    XKrec <- sc%*%t(ld)

    Xab <- NULL
    TPab <- NULL

    for (i in seq_len(nrow(X))) {
        position1 <- which(Desa[i,]==1)
        position2 <- which(Desb[i,]==1)
        Xab <- rbind(Xab,XK[I*(position2-1)+position1,])
        TPab <- rbind(TPab,XKrec[I*(position2-1)+position1,])
    }
    
    Eab <- Xab-TPab

    leverage <- apply(ld^2,1,sum)
    SPE <- apply(Eab^2,2,sum)

    output <- list(XK,sc,ld,ssq,Xab,TPab,Eab,leverage,SPE)
    names(output)<-c("data","scores","loadings","var.exp","X","TP","E",
        "leverage","SPE")
    
    output
}

################################################################################

################################################################################

# ASCAfun12.R

ASCAfun12 <- function (X,Desa,Desb,Fac) {
    n <- ncol(X) # number of genes
    I <- ncol(Desa) # number of levels in the factor TIME
    J <- ncol(Desb) # number of levels in the other factor

    XK1 <- matrix(NA,nrow=I,ncol=n)

    for (i in seq_len(I)) {
        sub <- X[Desa[,i]==1,]
        XK1[i,] <- apply(sub,2,mean)
    }
    
    NK <- matrix(NA,nrow=I,ncol=J)
    XK <- matrix(NA,nrow=I*J,ncol=n)

    k <- 1
    for (j in seq_len(J)) {
        for (i in seq_len(I)) {
            sub <- X[(Desa[,i]+Desb[,j])==2,]
            if (is.null(nrow(sub))) {  #when there isn't replicates
                NK[i,j] <- 1
                XK[k,] <- sub-XK1[i,]
            }
            else {
                NK[i,j] <- sqrt(nrow(sub))
                XK[k,] <- apply(sub,2,mean)-XK1[i,] 
            }
            k <- k+1
        }
    }

    XKw <- XK*(as.numeric(NK))

    PCA <- PCA.GENES(XKw)
    scw <- PCA$scores[,seq_len(Fac)]
    ld <- PCA$loadings[,seq_len(Fac)]
    ssq <- PCA$var.exp

    if (Fac==1) {
        scw <- as.matrix(scw)
        ld <- as.matrix(ld) 
    }
    if (Fac==0) {
        scw <- as.matrix(rep(0,I*J))
        ld <- as.matrix(rep(0,n))
    }
    
    # Re-weigth the scores
    sc <- scw/(as.numeric(NK))
    
    XKrec <- sc%*%t(ld)

    Xab <- NULL
    TPab <- NULL
    for (i in seq_len(nrow(X))){
        position1 <- which(Desa[i,]==1)
        position2 <- which(Desb[i,]==1)
        Xab <- rbind(Xab,XK[I*(position2-1)+position1,])
        TPab <- rbind(TPab,XKrec[I*(position2-1)+position1,])
    }
    Eab <- Xab-TPab

    #Leverage & SPE 
    leverage <- apply(ld^2,1,sum)
    SPE <- apply(Eab^2,2,sum)
    output <- list(XK,sc,ld,ssq,Xab,TPab,Eab,leverage,SPE)
    names(output) <- c("data","scores","loadings","var.exp","X","TP","E",
        "leverage","SPE")
    
    output
}

################################################################################

################################################################################

# ASCAfunres.R

ASCAfun.res <- function (X,Fac) {
    PCA <- PCA.GENES(X)
    sc <- PCA$scores[,seq_len(Fac)]
    ld <- PCA$loadings[,seq_len(Fac)]
    ssq <- PCA$var.exp
    if (Fac==1) {
        sc <- as.matrix(sc)
        ld <- as.matrix(ld)
    }
    TPres<-sc%*%t(ld)
    if (Fac==0) {
        sc <- 0
        ld <- 0
        TPres<-matrix(0,nrow(X),ncol(X))
    }      
    Eres<-X-TPres
    output <- list(sc,ld,ssq,X,TPres,Eres)
    names(output) <- c("scores","loadings","var.exp","X","TP","E")
    output
}

################################################################################

################################################################################

# ASCAfun-triple.R

ASCAfun.triple <- function(X,Desa,Desb,Desc,Fac) {

    n <- ncol(X) # number of genes
    I <- ncol(Desa) # number of levels in the factor TIME
    J <- ncol(Desb) # number of levels in the other factor
    H <- ncol(Desc) # number of levels in the other factor

    #Matrices con medias efectos individuales
    XK1 <- matrix(NA,nrow=I,ncol=n)
    for (i in seq_len(I)) {
        sub <- X[Desa[,i]==1,]
        XK1[i,] <- apply(sub,2,mean)
    }

    XK2 <- matrix(NA,nrow=J,ncol=n)
    for (j in seq_len(J)) {
        sub <- X[Desb[,j]==1,]
        XK2[j,] <- apply(sub,2,mean)
    }

    XK3 <- matrix(NA,nrow=H,ncol=n)
    for (h in seq_len(H)) {
        sub <- X[Desc[,h]==1,]
        XK3[h,] <- apply(sub,2,mean)
    }

    #Matrices con medias de efectos simples

    XK12 <- matrix(NA,nrow=I*J,ncol=n)
    k <- 1
    for (j in seq_len(J)) {
        for (i in seq_len(I)) {
            sub <- X[(Desa[,i]+Desb[,j])==2,]
            XK12[k,] <- apply(sub,2,mean)
            k <- k+1
        }
    }

    XK13 <- matrix(NA,nrow=I*H,ncol=n)
    k <- 1
    for (h in seq_len(H)) {
        for (i in seq_len(I)) {
            sub <- X[(Desa[,i]+Desc[,h])==2,]
            XK13[k,] <- apply(sub,2,mean)
            k <- k+1
        }
    }

    XK23 <- matrix(NA,nrow=J*H,ncol=n)
    k <- 1
    for (h in seq_len(H)){
        for (j in seq_len(J)){
            sub <- X[(Desb[,j]+Desc[,h])==2,]
            XK23[k,] <- apply(sub,2,mean)
            k <- k+1
        }
    }

    NK <- matrix(NA,nrow=I,ncol=J*H)
    XK <- matrix(NA,nrow=I*J*H,ncol=n)
    k <- 1

    for (h in seq_len(H)) {
        for (j in seq_len(J)) {
            for (i in seq_len(I)) {
                sub <- as.matrix(rbind(X[(Desa[,i]+Desb[,j]+Desc[,h])==3,]))
                NK[i,(h-1)*J+j] <- sqrt(nrow(sub))
                XK[k,] <- apply(sub,2,mean)+XK1[i,]+
                    XK2[j,]+XK3[h,]-XK12[(j-1)*I+i,]-XK13[(h-1)*I+i,]-
                    XK23[(h-1)*J+j,]
                k <- k+1
            }
        }
    }

    XKw <- XK*(as.numeric(NK))

    PCA <- PCA.GENES(XKw)
    scw <- PCA$scores[,seq_len(Fac)]
    ld <- PCA$loadings[,seq_len(Fac)]
    ssq <- PCA$var.exp
    if (Fac==1) {
        scw <- as.matrix(scw)
        ld <- as.matrix(ld) 
    }
    if (Fac==0) {
        scw <- as.matrix(rep(0,I*J*H))
        ld <- as.matrix(rep(0,n))
    }
    # Re-weigth the scores
    sc <- scw/(as.numeric(NK))
    
    XKrec <- sc%*%t(ld)

    Xabc <- NULL
    TPabc <- NULL

    for (i in seq_len(nrow(X))){
        position1 <- which(Desa[i,]==1)
        position2 <- which(Desb[i,]==1)
        position3 <- which(Desc[i,]==1)
        Xabc <- rbind(Xabc,XK[I*(position2-1)+I*J*(position3-1)+position1,])
        TPabc <- rbind(TPabc,XKrec[I*(position2-1)+
            I*J*(position3-1)+position1,])
    }
    Eabc <- Xabc-TPabc

    #Leverage & SPE 
    leverage <- apply(ld^2,1,sum)
    SPE <- apply(Eabc^2,2,sum)

    output <- list(XK,sc,ld,ssq,Xabc,TPabc,Eabc,leverage,SPE)
    names(output) <- c("data","scores","loadings","var.exp","X","TP","E",
        "leverage","SPE")
    
    output
}

################################################################################

################################################################################

# auxiliar.R

busca <- function (x,S) {
    which(S[,1] == x[1] & S[,2] == x[2])
}

int.mult <- function(lista,todos=NULL) {
    if (is.null(todos)) {
        todos <- unlist(lista)
    }
    comunes <- todos
    for (i in seq_len(length(lista))) {
        comunes <- intersect(comunes,lista[[i]])
    }
    comunes
}

n.menor <- function (x,S1,S2) {
    length(which(S1 <= x[1] &  S2 <= x[2]))
}

noceros <- function (x,num=TRUE,k=0) {
    nn <- length(which(x > k))
    if (num) {
        nn 
    } 
    else {
        if(nn > 0) { 
            which(x > k) 
        } 
        else { 
            NULL 
        }
    }
}

sinceros <- function (datos,k) {
    datos <- as.matrix(datos)
    datos0 <- as.matrix(datos)
    
    if (is.null(k)) {
        mini0 <- min(datos[noceros(datos,num=FALSE,k=0)])
        kc <- mini0/2
        datos0[datos0 == 0] <- kc
    } 
    else {
        datos0[datos0 == 0] <- k
    }
    datos0
}

sim.samples <- function(counts1,counts2=NULL,pnr=1,nss=5,v=0.02) {
    seqdep <- c(sum(counts1),sum(counts2))
    num.reads1 <- (pnr + c(-v,v))*seqdep[1]   
    muestras <- vector("list")
    muestras$c1 <- NULL
    for (s in seq_len(nss)) {
        tama <- round(runif(1,num.reads1[1],num.reads1[2]),0)
        muestras$c1 <- cbind(muestras$c1,rmultinom(1,size=tama,prob=counts1))
    }
    
    if (!is.null(counts2)) {
        num.reads2 <- (pnr + c(-v,v))*seqdep[2]
        muestras$c2 <- NULL
        for (s in seq_len(nss)) {
            tama <- round(runif(1,num.reads2[1],num.reads2[2]),0)
            muestras$c2 <- cbind(muestras$c2,
                rmultinom(1,size=tama,prob=counts2))
        }
    }
    muestras
}

ranking <- function(results) {
    M <- results$M
    D <- results$D
    prob <- results$prob
    
    ## Changing NA by 0
    M[is.na(M)] <- 0
    D[is.na(D)] <- 0
    prob[is.na(prob)] <- 0
    
    
    ## Ranking
    ranking5 <- sqrt(M*M + D*D)*sign(M)
    
    ## Ranking results
    theranking <- data.frame(rownames(results),ranking5)
    rownames(theranking) <- NULL
    colnames(theranking) <- c("ID","statistic")
    
    theranking
}

plot.y2 <- function(x,yright,yleft,yrightlim=range(yright,na.rm=TRUE),
    yleftlim=range(yleft,na.rm=TRUE),xlim=range(x,na.rm=TRUE),xlab=NULL,
    yylab=c("",""),lwd=c(2,2),pch=c(1,2),col=c(1,2),type=c("o","o"),
    linky=TRUE,smooth=0,bg=c("white","white"),lwds=1,length=10,...,x2=NULL,
    yright2=NULL,yleft2=NULL,col2=c(3,4)) {

    ## Plotting RIGHT axis data
    plot(x,yright,axes=FALSE,ylab="",xlab=xlab,ylim=yrightlim,xlim=xlim,
        pch=pch[1],type=type[1],lwd=lwd[1],col=col[1],...)

    axis(4,pretty(yrightlim,length),col=1,col.axis=1)

    if (is.null(yright2) == FALSE) {
        points(x2,yright2,type=type[1],pch=pch[1],lwd=lwd[1],col=col2[1],...)
    }

    if (smooth != 0) 
        lines(supsmu(x,yright,span=smooth),col=col[1],lwd=lwds,...)

    if (yylab[1]=="") {
        mtext(deparse(substitute(yright)),side=4,outer=FALSE,line=2,col=1,...)
    } 
    else {
        mtext(yylab[1],side=4,outer=FALSE,line=2,col=1,...)
    }

    par(new=TRUE)

    ## Plotting LEFT axis data
    plot(x,yleft,axes=FALSE,ylab="" ,xlab=xlab,ylim=yleftlim,xlim=xlim,bg=bg[1],
        pch=pch[2],type=type[2],lwd=lwd[2],col=col[2],...)
    box()
    axis(2,pretty(yleftlim,length),col=1,col.axis=1)

    if (is.null(yleft2) == FALSE) {
        points(x2,yleft2,type=type[2],pch=pch[2],bg=bg[2],lwd=lwd[2],
            col=col2[2],...)
    }

    if (smooth != 0) 
        lines(supsmu(x,yleft,span=smooth),col=col[2],lwd=lwds,...)

    if(yylab[2] == "") {
        mtext(deparse(substitute(yleft)),side=2,outer=FALSE,line=2,col=1,...)
    } 
    else {
        mtext(yylab[2],side=2,outer=FALSE,line=2,col=1,...)
    }

    ## X-axis
    axis(1,at=pretty(xlim,length))
}

logscaling <- function(data,base=2,k=1) {
    logmaximo <- round(max(data,na.rm=TRUE),0)
    numceros <- nchar(logmaximo)-1
    etiquetas <- c(0,10^(seq_len(numceros)))
    donde <- log(etiquetas + k,base=base)
    data <- log(data + k,base=base)
    list("data"=data,"at"=donde,"labels"=etiquetas)
}

miscolores <- colors()[c(554,89,111,512,17,586,132,428,601,568,86,390)]

################################################################################

################################################################################

# biodetection.plot.R

biodetection.dat <- function(input,factor=NULL,k=0) {
    if (inherits(input,"eSet") == FALSE)
        stop("Error. The input data must be an eSet object\n")

    if (any(!is.na(featureData(input)@data$Biotype)) == FALSE)
        stop ("No biological classification was provided.\nPlease run ",
            "addData() function to add this information\n")

    if (!is.null(assayData(input)$exprs)) {
        dat <- as.matrix(assayData(input)$exprs)
        mysamples <- colnames(assayData(input)$exprs)
    } 
    else {
        dat <- as.matrix(assayData(input)$counts)
        mysamples <- colnames(assayData(input)$counts)
    }
    
    numgenes <- nrow(dat)

    if (is.null(factor)) {  # per sample  
        cat("Biotypes detection is to be computed for:\n")
        print(colnames(dat))
        biotablas <- vector("list",length=NCOL(dat))
        names(biotablas) <- colnames(dat)
    } 
    else {  # per condition
        mifactor <- pData(input)[,factor]
        niveles <- levels(mifactor)
        cat("Biotypes detection is to be computed for:\n")
        print(niveles)
        biotablas <- vector("list",length=length(niveles))
        names(biotablas) <- niveles
        dat <- vapply(niveles,
            function (k) rowSums(as.matrix(dat[,grep(k,mifactor)])),numeric(1))
    }

    infobio <- as.character(featureData(input)@data$Biotype)

    genome <- 100*table(infobio)/sum(table(infobio))
    ordre <- order(genome,decreasing=TRUE)

    for (i in seq_len(length(biotablas))) {
        detect <- dat[,i] > k
        perdet1 <- genome*table(infobio,detect)[names(genome),"TRUE"]/
            table(infobio)[names(genome)]
        perdet2 <- 100*table(infobio,detect)[names(genome),"TRUE"] /
            sum(table(infobio,detect)[,"TRUE"])
        biotablas[[i]] <- as.matrix(rbind(perdet1[ordre],perdet2[ordre]))
        rownames(biotablas[[i]]) <- c("detectionVSgenome","detectionVSsample")
    }

    mybiotable <- list("genome"=genome[ordre],"biotables"=biotablas,
        "genomesize"=numgenes)

    mybiotable
}

biodetection.plot <- function(dat,samples=c(1,2),
    plottype=c("persample","comparison"),toplot="protein_coding",
    toreport=FALSE,...) {

    mypar=par(no.readonly=TRUE)
    plottype=match.arg(plottype)

    if (length(samples) > 2) {
        stop("ERROR: This function cannot generate plots for more than 2 ",
            "samples.\nPlease,use it as many times as needed to generate the ",
            "plots for all your samples.\n")
    }
    
    if (is.numeric(samples)) 
        samples=names(dat$biotables)[samples]
    biotable1 <- rbind(dat$genome,dat$biotables[[samples[1]]],
        rep(0,length(dat$genome)))

    # Computing ylim for left and right axis         
    if (ncol(biotable1) >= 3) {
        ymaxL <- ceiling(max(biotable1[,c(1,2,3)],na.rm=TRUE))
        ymaxR <- max(biotable1[,-c(1,2,3)],na.rm=TRUE)                
    } 
    else {
        ymaxL <- ceiling(max(biotable1,na.rm=TRUE))
        ymaxR=0 
    }

    if (length(samples) == 2) {
        biotable2 <- rbind(dat$genome,dat$biotables[[samples[2]]],
            rep(0,length(dat$genome)))
        if (ncol(biotable2) >= 3) {
            ymax2 <- ceiling(max(biotable2[,c(1,2,3)],na.rm=TRUE))
            ymax2sin <- max(biotable2[,-c(1,2,3)],na.rm=TRUE)
            ymaxR <- ceiling(max(ymaxR,ymax2sin))
        } 
        else {
            ymax2 <- ceiling(max(biotable2,na.rm=TRUE))            
        }
        ymaxL=max(ymaxL,ymax2)     
    }

    # Rescaling biotables (datos2)
    if (length(samples) == 2) {                
        if (ncol(biotable2) >= 3) 
            biotable2[,-c(1,2,3)] <- biotable2[,-c(1,2,3)]*ymaxL/ymaxR
    }

    # Rescaling biotables (datos1)    
    if (ncol(biotable1) >= 3) 
        biotable1[,-c(1,2,3)] <- biotable1[,-c(1,2,3)]*ymaxL/ymaxR

    ## PLOTS

    if (length(samples) == 1) {     # Plot (1 sample) - 2 scales                
        par(mar=c(11,4,2,2))
        barplot(biotable1[c(1,3),],main=samples[1],xlab=NULL,ylab="%features",
            axis.lty=1,legend=FALSE,beside=TRUE,col=c("grey",2),las=2,
            ylim=c(0,ymaxL),border=c("grey",2))
        barplot(biotable1[c(2,4),],main=samples[1],xlab=NULL,ylab="%features",
            axis.lty=1,legend=FALSE,beside=TRUE,col=c(2,1),las=2,density=30,
            ylim=c(0,ymaxL),border=2,add=TRUE)

        # if number of biotypes >= 3 so we have left and right axis
        if (ymaxR > 0) {
            axis(side=4,at=pretty(c(0,ymaxL),n=5),
                labels=round(pretty(c(0,ymaxL),n=5)*ymaxR/ymaxL,1))
            abline(v=9.5,col=3,lwd=2,lty=2)                
        }        
        
        legend(x="topright",bty="n",horiz=FALSE,fill=c("grey",2,2),
            density=c(NA,30,NA),border=c("grey",2,2),
            legend=c("% in genome","detected","% in sample"))
    } 
    else {     # Plot (2 samples)
        par(mar=c(11,4,2,2))
        if (plottype == "persample") {
            barplot(biotable1[c(1,3),],main=samples[1],xlab=NULL,
                ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,
                col=c("grey",2),las=2,ylim=c(0,ymaxL),border=c("grey",2))
            barplot(biotable1[c(2,4),],main=samples[1],xlab=NULL,
                ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,
                col=c(2,1),las=2,density=30,ylim=c(0,ymaxL),border=2,add=TRUE)
            
            # if number of biotypes >= 3 so we have left and right axis
            if (ymaxR > 0) {
                axis(side=4,at=pretty(c(0,ymaxL),n=5),
                    labels=round(pretty(c(0,ymaxL),n=5)*ymaxR/ymaxL,1))
                abline(v=9.5,col=3,lwd=2,lty=2)
            }
            
            legend(x="topright",bty="n",horiz=FALSE,fill=c("grey",2,2),
                density=c(NA,30,NA),border=c("grey",2,2),
                legend=c("% in genome","detected","% in sample"))

            # Datos2                
            barplot(biotable2[c(1,3),],main=samples[2],xlab=NULL,
                ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,
                col=c("grey",4),las=2,ylim=c(0,ymaxL),border=c("grey",4))
            barplot(biotable2[c(2,4),],main=samples[2],xlab=NULL,
                ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,col=c(4,1),
                las=2,density=30,ylim=c(0,ymaxL),border=4,add=TRUE)
            
            # if number of biotypes >= 3 so we have left and right axis
            if (ymaxR > 0) {
                axis(side=4,at=pretty(c(0,ymaxL),n=5),
                    labels=round(pretty(c(0,ymaxL),n=5)*ymaxR/ymaxL,1))
                abline(v=9.5,col=3,lwd=2,lty=2)
            }

            legend(x="topright",bty="n",horiz=FALSE,fill=c("grey",4,4),
                density=c(NA,30,NA),border=c("grey",4,4),
                legend=c("% in genome","detected","% in sample"))
        }
        
        # A plot comparing two samples with regard to genome and for % in sample
        if (plottype == "comparison") {
            lefttable <- rbind(100*dat$biotables[[samples[1]]][1,]/dat$genome,
                100*dat$biotables[[samples[2]]][1,]/dat$genome)
            righttable <- rbind(dat$biotables[[samples[1]]][2,],
                dat$biotables[[samples[2]]][2,])

            if (length(toplot) > 1) {
                toplot <- toplot[1]
                print("WARNING: More than one biotype was provided, the ",
                    "proportion test will only by applied to the first ",
                    "biotype.")
            }
            
            if ((toplot != 1) && (toplot != "global")) {
                numgenes <- dat$genomesize
                myx <- round(righttable[,toplot]*numgenes/100,0)
                mytest <- prop.test(x=myx,n=rep(numgenes,2),
                    alternative="two.sided")
                if (is.numeric(toplot)) 
                    toplot <- colnames(righttable)[toplot]
            }
            
            asumar <- colSums(righttable)
            asumar <- which(asumar < 0.25)
            if (length(asumar) > 1) {
                righttable <- cbind(righttable[,-asumar],
                    rowSums(righttable[,asumar]))
                colnames(righttable)[ncol(righttable)] <- "Others"
            }

            # Detection in the genome
            bbb <- barplot(lefttable,main="Biotype detection over genome total",
                xlab=NULL,ylab="% detected features",axis.lty=1,legend=FALSE,
                cex.names=0.8,beside=TRUE,col=c(2,4),las=2,density=80,
                border=c(2,4),ylim=c(0,100))
            bbb <- colSums(bbb)/2
            lines(bbb,dat$genome,pch=20,type="o",lwd=2)

            # %detection in the sample                
            barplot(righttable,main="Relative biotype abundance in sample",
                xlab=NULL,ylab="Relative % biotypes",axis.lty=1,legend=FALSE,
                beside=TRUE,col=c(2,4),las=2,border=c(2,4))

            legend(x="topright",bty="n",horiz=FALSE,pch=c(15,15,20),
                lwd=c(NA,NA,1),legend=c(samples,"% in genome"),col=c(2,4,1))

            if ((toplot != 1) && (toplot != "global")) {
                print(paste("Percentage of",toplot,"biotype in each sample:"))
                names(mytest$estimate) <- samples
                print(round(mytest$estimate*100,4))
                print(paste("Confidence interval at 95% for the difference of",
                    "percentages:",samples[1],"-",samples[2]))
                print(round(mytest$conf.int[c(1,2)]*100,4))
                if (mytest$p.value < 0.05) {
                    print(paste("The percentage of this biotype is",
                        "significantly DIFFERENT for these two samples",
                        "(p-value =",signif(mytest$p.value,4),")."))
                } 
                else {
                    print(paste("The percentage of this biotype is NOT",
                        "significantly different for these two samples",
                        "(p-value =",signif(mytest$p.value,4),")."))
                }
            }
        }
    }
    
    # Reset with the default values
    if (!toreport) 
        par(mypar)
}

################################################################################

################################################################################

# cd.plot.R

cd.dat <- function (input,norm=FALSE,refColumn=1) {
    if (inherits(input,"eSet") == FALSE)
        stop("ERROR: The input data must be an eSet object.\n")
    
    if (!is.null(assayData(input)$exprs)) {
        if (ncol( assayData(input)$exprs) < 2)
            stop("ERROR: The input object should have at least two samples.\n")

        datos <- assayData(input)$exprs
    } 
    else {
        if (ncol(assayData(input)$counts) < 2)
            stop("ERROR: The input object should have at least two samples.\n")
        datos <- assayData(input)$counts
    }

    ceros = which(rowSums(datos) == 0)
    hayceros = (length(ceros) > 0)

    if (hayceros) {
        print(paste("Warning:",length(ceros),"features with 0 counts in all",
            "samples are to be removed for this analysis."))
        datos <- datos[-ceros,]
    }
    
    ## scaling data and/or changing 0 to k
    if (norm)
        datos <- sinceros(datos, k = NULL)
    else
        datos <- noirpkm(datos,long=1000,lc=1,k=0.5)

    ## to plot
    data2plot <- log2(datos/datos[,refColumn])

    if (is.numeric(refColumn)) 
        refColumn <- colnames(datos)[refColumn]
    
    print(paste("Reference sample is:",refColumn))

    MsinRef <- as.matrix(data2plot[,-match(refColumn, colnames(data2plot))])
    colnames(MsinRef) <- colnames(data2plot)[-match(refColumn,
        colnames(data2plot))]
    alpha <- 0.05
    alpha <- alpha/ncol(MsinRef)
    nperm <- 10^3

    bootmed <- vapply(seq_len(nperm),function(k) {
        permut <- sample(seq_len(nrow(MsinRef)),replace=TRUE,nrow(MsinRef))
        permut <- as.matrix(MsinRef[permut,])
        permut <- apply(permut,2,median)   
        permut
    },numeric(ncol(MsinRef)))

    if (is.null(dim(bootmed))) 
        bootmed <- t(as.matrix(bootmed))

    bootmed <- t(apply(bootmed,1,quantile,probs=round(c(alpha/2,1-alpha/2),4)))
    diagno <- apply(bootmed,1,function (x) {
        ddd <- (x[1] <= 0) * (0 <= x[2])
        if (ddd == 1)
            ddd = "PASSED" 
        else
            ddd = "FAILED"
        ddd
    })
    bootmed <- cbind(bootmed,diagno)

    rownames(bootmed) <- colnames(MsinRef)
    colnames(bootmed)[3] <- "Diagnostic Test"
    print("Confidence intervals for median of M:")
    print(bootmed)

    if ("FAILED" %in% bootmed[,3])
        print(paste0("Diagnostic test: FAILED. Normalization is required ",
            "to correct this bias."))
    else
        print("Diagnostic test: PASSED.")

    list("data2plot"=data2plot,"refColumn"=refColumn,"DiagnosticTest"=bootmed)
}

cd.plot <- function (dat,samples=NULL,...) {
    refColumn <- dat$refColumn
    dat <- dat$data2plot
    if (is.null(samples)) 
        samples <- seq_len(ncol(dat))
    if (is.numeric(samples))
        samples <- colnames(dat)[samples]
    samples <- setdiff(samples,refColumn)

    if(length(samples) > 12) 
        stop("Please select 12 samples or less to be plotted (excluding ",
            "reference).")
    
    dat <- as.matrix(dat[,samples])
    dat.dens <- apply(dat,2,density,adjust=1.5)
    limY <- c(0,max(vapply(dat.dens,
        function(x) max(x$y,na.rm=TRUE),numeric(1))))

    plot(dat.dens[[1]],xlab="M = log2(sample/refsample)",ylab="Density",
        lwd=2,ylim=limY,type="l",col=miscolores[1],
        main=paste("Reference sample:",refColumn),...)
    abline(v=median(dat[,1],na.rm=TRUE),col=miscolores[1],lty=2)
    if (length(samples) > 1) {
        for (i in 2:length(samples)) {
            lines(dat.dens[[i]],col=miscolores[i],lwd=2)
            abline(v=median(dat[,i],na.rm=TRUE),col=miscolores[i],lty=i+1)
        }
    }
    legend("topleft",legend=samples,
        text.col=miscolores[seq_len(length(samples))],bty="n",lty=1,lwd=2,
        col=miscolores[seq_len(length(samples))])
}

################################################################################

################################################################################

# countsbio.plot.R

countsbio.dat <- function (input,biotypes=NULL,factor=NULL,norm=FALSE)  {
    if (inherits(input,"eSet") == FALSE)
        stop("Error. You must give an eSet object\n")

    if (!is.null(assayData(input)$exprs))
        datos <- assayData(input)$exprs
    else
        datos <- assayData(input)$counts

    depth <- round(colSums(datos)/10^6,1)
    names(depth) <- colnames(datos)
    ceros <- which(rowSums(datos) == 0)
    hayceros <- length(ceros) > 0
        
    if (hayceros) {
        warning("Warning:",length(ceros),"features with 0 counts in all ",
            "samples are to be removed for this analysis.")
        datos0=datos[-ceros,,drop=FALSE]
    } 
    else { 
        datos0=datos
    }
    
    nsam <- NCOL(datos)
    
    if (nsam == 1) {
        datos <- as.matrix(datos)
        datos0 <- as.matrix(datos0)
    }

    # Per condition
    if (is.null(factor)) {    # per sample    
        print("Count distributions are to be computed for:")
        print(colnames(datos))
    } 
    else {    # per condition
        mifactor <- pData(input)[,factor]
        niveles <- levels(mifactor)
        print("Counts per million distributions are to be computed for:")
        print(niveles)
        if (norm) {
            datos <- vapply(niveles,function (k) { 
                rowMeans(as.matrix(datos[,grep(k,mifactor)]))
            },numeric(1))
            datos0 <- vapply(niveles,function (k) { 
                rowMeans(as.matrix(datos0[,grep(k,mifactor)]))
            },numeric(1))
        } 
        else {
            datos <- vapply(niveles,function (k) { 
                10^6*rowMeans(t(t(datos[,grep(k,mifactor)])/colSums(as.matrix(
                    datos[,grep(k,mifactor)]))))
            },numeric(1))
            datos0 <- vapply(niveles,function (k) { 
                10^6*rowMeans(t(t(datos0[,grep(k,mifactor)])/colSums(as.matrix(
                    datos0[,grep(k,mifactor)]))))
            },numeric(1))            
        }
        colnames(datos) <- colnames(datos0) <- niveles
        depth <- vapply(niveles,function (k) {
            paste(range(depth[grep(k,mifactor)]),collapse="-")
        },character(1))
    }

    # Biotypes
    if (!is.null(featureData(input)$Biotype)) {
        if (hayceros) {
            infobio0 <- as.character(featureData(input)$Biotype)[-ceros] 
        } 
        else { 
            infobio0 <- as.character(featureData(input)$Biotype) 
        }
        infobio <- as.character(featureData(input)$Biotype)
    } 
    else { 
        infobio0 <- NULL 
        infobio <- NULL 
    }

    if (!is.null(infobio)) {        
        if(is.null(biotypes)) {
            biotypes <- unique(infobio)
            names(biotypes) <- biotypes        
        }     
        # which genes belong to each biotype
        biog <- lapply(biotypes,function(x) { 
            which(is.element(infobio0,x)) 
        })
        names(biog) <- biotypes
        bionum <- c(NROW(datos0),vapply(biog,length,integer(1)))
        names(bionum) <- c("global",names(biotypes)) 
        bio0 <- which(bionum == 0)
        if (length(bio0) > 0) 
            bionum <- bionum[-bio0]
    } 
    else { 
        biotypes <- NULL
        bionum <- NULL 
    }
    
    # Create the summary matrix information
    if (is.null(bionum)) {
        resumen <- vector("list",length=1)
        names(resumen) <- "global"
    } 
    else {
        resumen <- vector("list",length=length(bionum))
        names(resumen) <- names(bionum)
    }

    cuentas <- c(0,1,2,5,10)

    if (is.null(factor)) {
        if (norm) {
            datosCPM <- datos
        } 
        else {
            datosCPM=10^6 * t(t(datos)/colSums(as.matrix(datos)))
        }
    } else { datosCPM=datos }

    for (i in seq_len(length(resumen))) {
        if (i == 1) {
            datosR=datosCPM             
        } else {             
            if(!is.null(infobio)) {
                datosR <- datosCPM[which(infobio == names(resumen)[i]),]
                if (!is(datosR,"matrix")) {
                    datosR <- t(as.matrix(datosR))
                }
            }
        }

        nfeatures <- nrow(datosR)
        datosR <- datosR[which(rowSums(datosR) > 0),]
        if (!is(datosR,"matrix")) {
            datosR <- t(as.matrix(datosR)) 
        }

        myglobal <- NULL
        mypersample <- NULL

        for (kk in seq_len(length(cuentas))) {
            mypersample <- rbind(mypersample,apply(datosR,2,function (x) { 
                length(which(x > cuentas[kk])) 
            }))
            myglobal <- c(myglobal,sum(apply(datosR,1,function (x) { 
                max(x) > cuentas[kk] 
            })))
        } 

        mypersample <- round(100*mypersample/nfeatures,1)
        mypersample <- rbind(mypersample,depth)
        rownames(mypersample) <- seq_len(nrow(mypersample))
        myglobal <- c(round(100*myglobal/nfeatures,1),nfeatures)

        resumen[[i]] <- data.frame(c(paste("CPM >",cuentas),"depth"),
            mypersample,"total"=myglobal)
        colnames(resumen[[i]])[1] <- names(resumen)[i]
        colnames(resumen[[i]])[2:(ncol(resumen[[i]])-1)] <- colnames(datosR)
    }

    ## results
    cosas <- list("result"=datos0,"bionum"=bionum,"biotypes"=infobio0,
        "summary"=resumen)
    
    cosas
}

countsbio.plot <- function(dat,samples=c(1,2),toplot="global",
    plottype=c("barplot","boxplot"),toreport=FALSE,...) {

    # dat: Data coming from countsbio.dat function
    # samples: Samples to be plotted. If NULL,all samples are plotted.
    # toplot: Name of biotype (including "global") to be plotted.

    mypar <- par(no.readonly=TRUE)
    plottype <- match.arg(plottype)

    ## Preparing data
    if (is.null(samples)) { 
        if (NCOL(dat$result) == 1) {
            samples=1
        } else {
            samples <- seq_len(NCOL(dat$result)) 
        }        
    }
    if (is.numeric(toplot)) 
        toplot <- names(dat$summary)[toplot]
    if (is.numeric(samples) && !is.null(colnames(dat$result))) 
        samples <- colnames(dat$result)[samples]

    if (plottype == "barplot") {
        if ((exists("ylab") && !is.character(ylab)) || !exists("ylab")) 
            ylab <- ""

        datos <- dat$summary[[toplot]]
        mytotal <- as.numeric(datos[,"total"])
        datos <- as.matrix(datos[,samples])
        rownames(datos) <- as.character(dat$summary[[toplot]][,1])

        par(mar=c(6,4,4,2))
        barplot(as.numeric(datos[1,]),col=miscolores[1],las=2,main="",ylab="",
            density=70,ylim=c(0,100),cex.axis=0.8,names.arg="",...)
        for (i in 2:(length(mytotal)-2)) {
            barplot(as.numeric(datos[i,]),col=miscolores[i],las=2,main="",
                ylab="",add=TRUE,density=70,ylim=c(0,100),cex.axis=0.8,
                names.arg="",...)
        }
        
        bp <- barplot(as.numeric(datos[(length(mytotal)-1),]),
            col=miscolores[(length(mytotal)-1)],las=2,
            main=paste(toupper(toplot)," (",mytotal[length(mytotal)],")",
            sep=""),ylab="Sensitivity (%)",add=TRUE,names.arg=colnames(datos),
            cex.axis=0.8,density=70,ylim=c(0,100),cex.names=0.8,...)
        
        for (j in seq_len((length(mytotal)-1))) 
            abline(h=mytotal[j],col=miscolores[j],lwd=2)
        if (length(samples) <= 10) {
            mtext(side=3,text=datos["depth",],adj=0.5,at=bp,cex=0.8)
        } 
        else {
            mtext(side=3,text=datos["depth",],at=bp,cex=0.7,las=2)
        }
        legend("top",rownames(datos)[-length(mytotal)],fill=miscolores,
            density=70,bty="n",ncol=3)
        par(mar=c(5,4,4,4) + 0.1)
    }

    if (plottype == "boxplot") {
        conteos <- as.matrix(dat$result[,samples])
        if (is.numeric(samples)) 
            colnames(conteos) <- colnames(dat$result)[samples]
        else 
            colnames(conteos) <- samples
        num <- dat$bionum[toplot]
        if (is.null(num)) {
            if (toplot == "global") {
                num=nrow(conteos)
            } else {
                num=0
            }
        } 
        infobio <- dat$biotypes

        if (num == 0 && toplot != "global") 
            stop("Error: No data available. Please,change toplot parameter.")

        if ((exists("ylab") && !is.character(ylab)) || !exists("ylab"))
            ylab <- "Expression values"

        ## Plots            
        if (length(samples) == 1) {
            escala <- logscaling(conteos,base=2)
            if (is.null(infobio)) {
                boxplot(escala$data,col=miscolores[1],ylab=ylab,main="",
                    yaxt="n",...)
            } 
            else {
                par(mar=c(10,4,4,2))    
                boxplot(escala$data ~ infobio,col=miscolores,ylab=ylab,
                    main=colnames(conteos),las=2,cex.axis=0.8,cex.lab=0.9,
                    yaxt="n",...)
                cuantos <- dat$bionum[-1]
                cuantos <- cuantos[sort(names(cuantos))]
                mtext(cuantos,3,at=seq_len(length(cuantos)),cex=0.6,las=2)
            }
        } 
        else {
            if (toplot != "global") 
                conteos=conteos[which(infobio == toplot),]
            
            escala <- logscaling(conteos,base=2)
            main <- paste(toupper(toplot)," (",num,")",sep="")

            par(mar=c(6,4,2,2))    
            boxplot(escala$data,col=miscolores,ylab=ylab,main=main,las=2,
                cex.lab=0.9,cex.axis=0.8,yaxt="n",...)
        }        
        axis(side=2,at=escala$at,labels=escala$labels)
    }

    if (!toreport)
        par(mypar)
}

################################################################################

################################################################################

# dat.R

dat <- function(input,type = c("biodetection","cd","countsbio","GCbias",
    "lengthbias","saturation","PCA"),k=0,ndepth=6,factor=NULL,norm=FALSE,
    refColumn=1,logtransf = FALSE) {

    type <- match.arg(type)
    if (type == "biodetection") {
        output <- new("Biodetection",
            dat=biodetection.dat(input,factor=factor,k=k))        
    }
    
    if (type == "cd") {
        output <- new("CD",dat=cd.dat(input,norm=norm,refColumn=refColumn))
    }

    if (type == "countsbio") {
        output <- new("CountsBio",
            dat=countsbio.dat(input,factor=factor,norm=norm))
    }

    #if (type == "GCbias") {
    #    output <- new("GCbias",
    #        dat=GC.dat(input,factor=factor,norm=norm))
    #}

    #if (type == "lengthbias") {
    #    output <- new("lengthbias",
    #        dat=length.dat(input,factor=factor,norm=norm))
    #}
    
    if (type == "saturation") {
        output <- new("Saturation",dat=saturation.dat(input,k=k,ndepth=ndepth))
    }
    
    #if (type == "PCA") {
    #    output=new("PCA",dat=PCA.dat(input,norm=norm,logtransf=logtransf))
    #}
    
    output
}

################################################################################

################################################################################

# DE.plot.R

DE.plot <- function(output,q=NULL,graphic=c("MD","expr","chrom","distr"),
    pch=20,cex=0.5,col=1,pch.sel=1,cex.sel=0.6,col.sel=2,log.scale=TRUE,
    chromosomes=NULL,join=FALSE,...) {
    mypar <- par(no.readonly=TRUE)

    if (!is(output,"Output"))
        stop("Error. Output argument must contain an object generated by ",
            "noiseq or noiseqbio functions.\n")

    graphic <- match.arg(graphic)    
    noiseqbio <- "theta" %in% colnames(output@results[[1]])[c(1,2,3,4)]
    
    ## MD plot
    if (graphic == "MD") {
        if (noiseqbio) {   
            M <- output@results[[1]][,"log2FC"]
            D <- abs(output@results[[1]][,1] - output@results[[1]][,2])+1
            names(M) <- names(D) <- rownames(output@results[[1]])
            
            plot(M,D,pch=pch,xlab="M",ylab="D",cex=cex,col=col,log="y",...)
            
            if(!is.null(q)) {
                mySelection <- rownames(degenes(output,q))
                points(M[mySelection],D[mySelection],col=col.sel,pch=pch.sel,
                    cex=cex.sel)
            }
        }
        else {
            plot(output@results[[1]][,"M"],(1+output@results[[1]][,"D"]),
                pch=pch,xlab="M",ylab="D",cex=cex,col=col,log="y",...)
            if (!is.null(q)) {
                mySelection <- rownames(degenes(output,q))
                points(output@results[[1]][mySelection,"M"],
                    output@results[[1]][mySelection,"D"]+1,col=col.sel,
                    pch=pch.sel,cex=cex.sel)
            }
        }
    }
    
    ## Expression plot: Condition1 vs Condition2
    else if (graphic == "expr") {
        data <- cbind(output@results[[1]][1],output@results[[1]][2])
        rownames(data) <- rownames(output@results[[1]])
        colnames(data) <- colnames(output@results[[1]][c(1,2)])

        if (log.scale) {
            k <- min(data,na.rm=TRUE)
            escala <- logscaling(data,base=2,k=k)
            
            plot(escala$data,pch=pch,cex=cex,col=col,yaxt="n",xaxt="n",...)
            axis(side=1,at=escala$at,labels=escala$labels)
            axis(side=2,at=escala$at,labels=escala$labels)
            
            if(!is.null(q)) {
                mySelection <- rownames(degenes(output,q))    
                points(escala$data[mySelection,],col=col.sel,pch=pch.sel,
                    cex=cex.sel)
            }
        } 
        else {
            plot(data,pch=pch,cex=cex,col=col,...)            
            if(!is.null(q)) {
                mySelection <- rownames(degenes(output,q))
                points(data[mySelection,],col=col.sel,pch=pch.sel,cex=cex.sel)
            }
        }
    }
    
    ## MANHATTAN PLOT
    else if (graphic == "chrom") {
        mydata <- data.frame(as.character(output@results[[1]][,"Chrom"]),
            output@results[[1]][,c("GeneStart","GeneEnd")],
            output@results[[1]][,c(1,2)])
        mydata <- na.omit(mydata)
        colnames(mydata) <- c("chr","start","end",colnames(mydata)[-c(1,2,3)])
        
        if (is.null(chromosomes)) {    # todos los cromosomas
            chromosomes <- unique(mydata$chr)
            chromosomes <- sort(chromosomes)
        }
        
        print("REMEMBER. You are plotting these chromosomes and in this order:")
        print(chromosomes)
        
        # logarithmic scale
        if (log.scale) {
            if (min(mydata[,-c(1,2,3)],na.rm=TRUE) < 1) {
                kk <- -min(mydata[,-c(1,2,3)],na.rm=TRUE)+1
            } 
            else { 
                kk <- 0 
            }
            mydata[,-c(1,2,3)] <- log2(mydata[,-c(1,2,3)]+kk)
        }
        
        # Selecting chromosomes and ordering positions
        ordenat <- NULL
        for (cromo in chromosomes) {
            myselec <- mydata[mydata[,"chr"] == cromo,]
            myselec <- myselec[order(myselec[,"start"]),]
            ordenat <- rbind(ordenat,myselec)
        }
        
        sel.ord <- NULL
        
        if (!is.null(q)) {
            # up-regulated in first condition
            cond1 <- rownames(degenes(output,q,M="up"))
            # up-regulated in second condition
            cond2 <- rownames(degenes(output,q,M="down"))
            cond1 <- (rownames(ordenat) %in% cond1)*1
            cond2 <- (rownames(ordenat) %in% cond2)*1
            sel.ord <- cbind(cond1,cond2)
            rownames(sel.ord) <- rownames(ordenat)
        }
        
        chr.long <- aggregate(ordenat[,"end"],
            by=list(as.character(ordenat[,"chr"])),max)
        chr.long <- chr.long[match(chromosomes,chr.long[,1]),]
        
        if (join) { # si todos los cromosomas van en el mismo plot
            total.long <- sum(as.numeric(chr.long$x))
            chr.start <- cumsum(c(1,chr.long$x[-length(chr.long$x)]))
            names(chr.start) <- chr.long[,1]
            plot(c(1,total.long),c(-max(ordenat[,5]),max(ordenat[,4])),
                type="n",xlab="",ylab="Expression data",xaxt="n")
            axis(side=1,at=chr.start,labels=chr.long[,1],font=2)
            abline(h=0,lty=2,lwd=0.5)        
            
            for (ch in chromosomes) {
                dat.chr <- ordenat[which(ordenat[,"chr"] == ch),]
                dat.chr[,c("start","end")] <- dat.chr[,c("start","end")] +
                    chr.start[ch] - 1
                rect(xleft=dat.chr[,"start"],ybottom=0,xright=dat.chr[,"end"],
                    ytop=dat.chr[,4],col="grey",border=NA)
                rect(xleft=dat.chr[,"start"],ybottom=-dat.chr[,5],
                    xright=dat.chr[,"end"],ytop=0,col="grey",border=NA)

                if (!is.null(q)) {                    
                    aux <- which(rownames(sel.ord) %in% rownames(dat.chr))
                    sel.chr1 <- dat.chr[,4]*sel.ord[aux,1]
                    sel.chr2 <- -dat.chr[,5]*sel.ord[aux,2]
                    rect(xleft=dat.chr[,"start"],ybottom=0,
                        xright=dat.chr[,"end"],ytop=sel.chr1,col=2,border=NA)
                    rect(xleft=dat.chr[,"start"],ybottom=sel.chr2,
                        xright=dat.chr[,"end"],ytop=0,col=3,border=NA)
                }

                segments(x0=dat.chr[,"start"],y0=0,x1=dat.chr[,"end"],y1=0,
                    col=4,lwd=0.5) # annotated genes
            }
            text(c(1,1),0.9*c(max(ordenat[,4]),-max(ordenat[,5])),
                colnames(mydata)[4:5],font=3,adj=0,col=2:3)
            # a plot for each chromosome
        } 
        else {        
            num.chr <- length(chromosomes)
            k <- 20
            long.prop <- round((chr.long$x / min(chr.long$x))*k,0)
            while (max(long.prop)*num.chr > 500 | max(long.prop) > 50) {
                k <- k-1
                long.prop <- round((chr.long$x / min(chr.long$x))*k,0)
            }
            forlayout <- matrix(0,num.chr,max(long.prop))
            for (i in seq_len(num.chr)) {
                forlayout[i,seq_len(long.prop[i])] <- i
            }
            layout(forlayout)
            
            if (num.chr > 1)
                par(mar=c(2,4.1,0.1,0.1))
            miylim <- c(-max(ordenat[,5],na.rm=TRUE),
                max(ordenat[,4],na.rm=TRUE))
            
            for (i in seq_len(num.chr)) {
                apintar <- ordenat[which(ordenat[,"chr"] == chromosomes[i]),2:5]
                plot(c(1,chr.long[i,2]),miylim,type="n",xlab="",
                    ylab=chromosomes[i],xaxt="n",font.lab=2,cex.lab=1.3)
                rect(xleft=apintar[,"start"],ybottom=0,xright=apintar[,"end"],
                    ytop=apintar[,3],col="grey",border=NA)
                rect(xleft=apintar[,"start"],ybottom=-apintar[,4],
                    xright=apintar[,"end"],ytop=0,col="grey",border=NA)
                
                if (!is.null(q)) {
                    aux <- which(rownames(sel.ord) %in% rownames(apintar))
                    asel <- sel.ord[aux,]*apintar[,3:4]
                    rect(xleft=apintar[,"start"],ybottom=0,
                        xright=apintar[,"end"],ytop=asel[,1],col=2,border=NA)
                    rect(xleft=apintar[,"start"],ybottom=-asel[,2],
                        xright=apintar[,"end"],ytop=0,col=3,border=NA)
                }            
                abline(h=0,lty=2,lwd=0.5)
                segments(x0=apintar[,"start"],y0=0,x1=apintar[,"end"],y1=0,
                    col=4,lwd=0.5) # annotated genes
                etiq <- mypretty(c(1,chr.long[i,2]),n=10)
                axis(side=1,at=etiq,labels=etiq)
                text(c(1,1),0.9*miylim,colnames(mydata)[5:4],font=3,adj=0,
                    col=3:2)
            }
        }
    }
    
    ## DEG distribution across biotypes/chromosomes
    else if (graphic == "distr") {        
        if(!is.null(q)) { # Computing DEG
            mySelection <- rownames(degenes(output,q))
            detect <- rownames(output@results[[1]]) %in% mySelection
        }
        else {
            stop("You must specify a valid value for q\n")
        }

        if ("Chrom" %in% colnames(output@results[[1]])) {
            numplot <- 1
            infobio <- output@results[[1]][,"Chrom"]
            genome <- 100*table(infobio)/sum(table(infobio))
            ordre <- order(genome,decreasing=TRUE)            
            perdet1 <- genome*table(infobio,detect)[names(
                genome),2]/table(infobio)[names(genome)]
            perdet2 <- 100*table(infobio,detect)[names(
                genome),2]/sum(table(infobio,detect)[,2])
            ceros <- rep(0,length(genome))
            biotable1 <- as.matrix(rbind(genome[ordre],perdet1[ordre],
                perdet2[ordre],ceros))
            rownames(biotable1) <- c("genome","degVSgenome","deg","ceros")
            if (!is.null(chromosomes))
                biotable1=biotable1[,chromosomes]
            ymaxL1 <- ceiling(max(biotable1,na.rm=TRUE))
        } 
        else { 
            numplot <- 0 
        }
        
        if ("Biotype" %in% colnames(output@results[[1]])) {
            numplot <- c(numplot,1)
            infobio <- output@results[[1]][,"Biotype"]
            genome <- 100*table(infobio)/sum(table(infobio))
            ordre <- order(genome,decreasing=TRUE)            
            perdet1 <- genome*table(infobio,detect)[names(
                genome),2]/table(infobio)[names(genome)]
            perdet2 <- 100*table(infobio,detect)[names(
                genome),2]/sum(table(infobio,detect)[,2])
            ceros <- rep(0,length(genome))
            biotable2 <- as.matrix(rbind(genome[ordre],perdet1[ordre],
                perdet2[ordre],ceros))
            rownames(biotable2) <- c("genome","degVSgenome","deg","ceros")
            higher2 <- which(biotable2[1,] > 2)
            lower2 <- which(biotable2[1,] <= 2)
            if (length(higher2) > 0) {       
                ymaxL2 <- ceiling(max(biotable2[,higher2],na.rm=TRUE)) 
                if (length(lower2) > 0) {
                    ymaxR2 <- ceiling(max(biotable2[,lower2],na.rm=TRUE))
                    biotable2[,lower2] <- biotable2[,lower2]*ymaxL2/ymaxR2
                }
                else { 
                    ymaxR2 <- ymaxL2 
                }
                
            } 
            else { 
                ymaxR2 <- ceiling(max(biotable2[,lower2],na.rm=TRUE))
                ymaxL2 <- ymaxR2                             
            }
        } 
        else { 
            numplot <- c(numplot,0) 
        }
        
        # Plot
        if (sum(numplot) == 0) 
            stop("Biotype or chromosome information is needed for this plot\n")
        
        if (sum(numplot) == 1) {    # 1 Plot 
            par(mar=c(10,4,2,2))
            if (numplot[1] == 1) {
                barplot(biotable1[c(1,3),],
                    main="DEG distribution across chromosomes",
                    xlab=NULL,ylab="%features",axis.lty=1,legend=FALSE,
                    beside=TRUE,col=c("grey",2),las=2,ylim=c(0,ymaxL1),
                    border=c("grey",2))
                barplot(biotable1[c(2,4),],
                    main="DEG distribution across chromosomes",
                    xlab=NULL,ylab="%features",axis.lty=1,legend=FALSE,
                    beside=TRUE,col=c(2,1),las=2,density=30,ylim=c(0,ymaxL1),
                    xborder=2,add=TRUE)
                legend(x="topright",bty="n",horiz=FALSE,fill=c("grey",2,2),
                    density=c(NA,30,NA),border=c("grey",2,2),
                    legend=c("%Chrom in genome","%DEG in Chrom",
                    "%Chrom in DEG"))
            }
            
            if (numplot[2] == 1) {
                barplot(biotable2[c(1,3),],
                    main="DEG distribution across biotypes",
                    xlab=NULL,ylab="%features",axis.lty=1,legend=FALSE,
                    beside=TRUE,col=c("grey",4),las=2,ylim=c(0,ymaxL2),
                    border=c("grey",4))
                barplot(biotable2[c(2,4),],
                    main="DEG distribution across biotypes",
                    xlab=NULL,ylab="%features",axis.lty=1,legend=FALSE,
                    beside=TRUE,col=c(4,1),las=2,density=30,ylim=c(0,ymaxL2),
                    border=4,add=TRUE)
                axis(side=4,at=pretty(c(0,ymaxL2),n=5),
                    labels=round(pretty(c(0,ymaxL2),n=5)*ymaxR2/ymaxL2,1))
                if (ymaxR2 != ymaxL2) {
                    abline(v=3*length(higher2) + 0.5,col=3,lwd=2,lty=2)
                }
                legend(x="topright",bty="n",horiz=FALSE,fill=c("grey",4,4),
                    density=c(NA,30,NA),border=c("grey",4,4),
                    legend=c("%Biotype in genome","%DEG in Biotype",
                    "%Biotype in DEG"))
            }
        }
        
        if (sum(numplot) == 2) {    # 2 Plots 
            par(mar=c(10,4,2,2),mfrow=c(1,2))
            
            # Chromosomes
            barplot(biotable1[c(1,3),],
                main="DEG distribution across chromosomes",
                xlab=NULL,ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,
                col=c("grey",2),las=2,ylim=c(0,ymaxL1),border=c("grey",2))
            barplot(biotable1[c(2,4),],
                main="DEG distribution across chromosomes",xlab=NULL,
                ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,col=c(2,1),
                las=2,density=30,ylim=c(0,ymaxL1),border=2,add=TRUE)
            legend(x="topright",bty="n",horiz=FALSE,fill=c("grey",2,2),
                density=c(NA,30,NA),border=c("grey",2,2),
                legend=c("%Chrom in genome","%DEG in Chrom","%Chrom in DEG"))
            
            # Biotypes
            barplot(biotable2[c(1,3),],main="DEG distribution across biotypes",
                xlab=NULL,ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,
                col=c("grey",4),las=2,ylim=c(0,ymaxL2),border=c("grey",4))
            barplot(biotable2[c(2,4),],main="DEG distribution across biotypes",
                xlab=NULL,ylab="%features",axis.lty=1,legend=FALSE,beside=TRUE,
                col=c(4,1),las=2,density=30,ylim=c(0,ymaxL2),border=4,add=TRUE)
            axis(side=4,at=pretty(c(0,ymaxL2),n=5),
                labels=round(pretty(c(0,ymaxL2),n=5)*ymaxR2/ymaxL2,1))
            if (ymaxR2 != ymaxL2) {
                abline(v=3*length(higher2) + 0.5,col=3,lwd=2,lty=2)
            }        
            legend(x="topright",bty="n",horiz=FALSE,fill=c("grey",4,4),
                density=c(NA,30,NA),border=c("grey",4,4),
                legend=c("%Biotype in genome","%DEG in Biotype",
                    "%Biotype in DEG"))
        }
    }
    par(mypar)
}

## Auxiliar function
mypretty <- function(x,n=5) {    
    mywidth <- diff(x)
    mybin <- ceiling(mywidth/(n-1))
    mylabels0 <- x[1] + mybin*(0:(n-1))
    ndig <- nchar(as.character(mylabels0))
    mylabels <- mylabels0
    k <- 0
    for (i in 2:(n-1)) {
        mylabels[i] <- (round(mylabels0[i]/10^(ndig[i]-1),k))*10^(ndig[i]-1)
        while (mylabels[i] == mylabels[i-1]) {
            k <- k+1
            mylabels[i] <- (round(mylabels0[i]/10^(ndig[i]-1),k))*10^(ndig[i]-1)
        }
    }
    mylabels
}

################################################################################

################################################################################

# degenes.R

degenes <- function (object,q=0.95,M=NULL) {
    if (!is(object,"Output"))
        stop("You must give the object returned by the noiseq function\n")

    x <- object@results[[1]]
    noiseqbio="theta" %in% colnames(x)[seq_len(4)]

    if (noiseqbio) {
        y <- na.omit(x[c("theta","prob")])
        colnames(y)[1]="M"
    } else {
        y <- na.omit(x[c("M","D","prob")])
    }

    if (is.null(M)) {
        losdeg <- y[y[,"prob"] > q,]
        print(paste(dim(losdeg)[1],"differentially expressed features"))

    }
    else if (M == "up") {
        estos <- y[y[,"M"] > 0,]
        losdeg <- estos[estos[,"prob"] > q,]
        print(paste(dim(losdeg)[1],
            "differentially expressed features (up in first condition)"))
    } 
    else if (M == "down") {
        estos <- y[y[,"M"] < 0,]
        losdeg <- estos[estos[,"prob"] > q,]
        print(paste(dim(losdeg)[1],
            "differentially expressed features (down in first condition)"))
    }
    else {
        stop("ERROR! Value for parameter M is not valid. Please,choose ",
            "between NULL,'up' or 'down'")
    }
    
    # Restore the object with the same "results" structure
    losdeg <- x[rownames(losdeg),]
    losdeg[order(losdeg[,"prob"],decreasing=TRUE),]
}

################################################################################

################################################################################

# fewrepliactes.R

share.info = function (mydata, n1, n2, r, nclust)   {
    # clustering data by k-means algorithm
    # 1.a) Normalized data
    gc()
    cl <- suppressWarnings(kmeans(mydata,nclust,nstart=25,iter.max=nclust + 30))

    cat("...k-means clustering done\n")
    cat(paste("Size of",nclust,"clusters:\n"))
    print(cl$size)
                        
    # Creating pseudo-data
    cluster.data <- lapply(seq_len(nclust),function (k) { 
        mydata[cl$cluster == k,]
    })
        
    # Resampling
    npermu <- cl$size*r
    npermu <- vapply(npermu,function (x) min(x,1000),numeric(1))

    cat("Resampling cluster...")

    myres=vector("list",length=nclust)

    for (i in seq_len(nclust)) {
        print(i)

        #if (cl$size[i] > 1) {     # OPTION 2.A
        # OPTION 2.C: small clusters
        if (cl$size[i] > 1 && cl$size[i] < 1000) {
            #myres[[i]]=t(sapply(seq_len(npermu[i]),function (j) {
            #    permu=sample(cluster.data[[i]])
            #    nn1=n1*cl$size[i]
            #    nn2=n2*cl$size[i]
            #    mean1=mean(permu[seq_len(nn1)])
            #    mean2=mean(permu[(nn1+1):(nn1+nn2)])
            #    sd1=sd(as.numeric(permu[seq_len(nn1)]))
            #    sd2=sd(as.numeric(permu[(nn1+1):(nn1+nn2)]))
            #    data.frame("M"=log2(mean1/mean2),"D"=mean1-mean2,
            #        "M.sd"=sqrt(sd1^2/(mean1^2*log(2)^2*nn1) + 
            #        sd2^2/(mean2^2*log(2)^2*nn2)),
            #        "D.sd"=sqrt(sd1^2/sqrt(nn1) + sd2^2/sqrt(nn2)))
            #}))
            myres[[i]]=t(vapply(seq_len(npermu[i]),function (j) {
                permu=sample(cluster.data[[i]])
                nn1=n1*cl$size[i]
                nn2=n2*cl$size[i]
                mean1=mean(permu[seq_len(nn1)])
                mean2=mean(permu[(nn1+1):(nn1+nn2)])
                sd1=sd(as.numeric(permu[seq_len(nn1)]))
                sd2=sd(as.numeric(permu[(nn1+1):(nn1+nn2)]))
                data.frame("M"=log2(mean1/mean2),"D"=mean1-mean2,
                    "M.sd"=sqrt(sd1^2/(mean1^2*log(2)^2*nn1) + 
                    sd2^2/(mean2^2*log(2)^2*nn2)),
                    "D.sd"=sqrt(sd1^2/sqrt(nn1) + sd2^2/sqrt(nn2)))
            },vector("list",4)))
        }
        
        if (cl$size[i] >= 1000) {    # OPTION 2.C & 2.D: big clusters
            # Option 2.D: clustering big clusters
            cl2 <- kmeans(cluster.data[[i]],nclust,nstart=25,
                iter.max=nclust + 20)
            cat(paste("Size of",nclust,"subclusters of cluster:",i)); cat("\n")
            print(cl2$size)
            subcluster.data <- lapply(seq_len(nclust),function (k) { 
                cluster.data[[i]][cl2$cluster == k,] 
            })
            npermu2 <- cl2$size*r
            npermu2 <- vapply(npermu2,function (x) min(x,1000),numeric(1))
            ## modified to reduce the number of permutations to be done
            myres2 <- vector("list",length=nclust)
            
            for (h in seq_len(nclust)) {
                if (cl2$size[h] > 1) {
                    #myres2[[h]] <- t(sapply(seq_len(npermu2[h]),function (j) {
                    #    permu=sample(subcluster.data[[h]])
                    #    nn1 <- n1*cl2$size[h]
                    #    nn2 <- n2*cl2$size[h]             
                    #    mean1 <- mean(permu[seq_len(nn1)])
                    #    mean2 <- mean(permu[(nn1+1):(nn1+nn2)])
                    #    sd1 <- sd(as.numeric(permu[seq_len(nn1)]))
                    #    sd2 <- sd(as.numeric(permu[(nn1+1):(nn1+nn2)]))
                    #    data.frame("M"=log2(mean1/mean2),"D"=mean1-mean2,
                    #        "M.sd"=sqrt(sd1^2/(mean1^2*log(2)^2*nn1) +
                    #        sd2^2/(mean2^2*log(2)^2*nn2)),
                    #        "D.sd"=sqrt(sd1^2/sqrt(nn1) + sd2^2/sqrt(nn2)))
                    #}))
                    myres2[[h]] <- t(vapply(seq_len(npermu2[h]),function (j) {
                        permu=sample(subcluster.data[[h]])
                        nn1 <- n1*cl2$size[h]
                        nn2 <- n2*cl2$size[h]             
                        mean1 <- mean(permu[seq_len(nn1)])
                        mean2 <- mean(permu[(nn1+1):(nn1+nn2)])
                        sd1 <- sd(as.numeric(permu[seq_len(nn1)]))
                        sd2 <- sd(as.numeric(permu[(nn1+1):(nn1+nn2)]))
                        data.frame("M"=log2(mean1/mean2),"D"=mean1-mean2,
                            "M.sd"=sqrt(sd1^2/(mean1^2*log(2)^2*nn1) +
                            sd2^2/(mean2^2*log(2)^2*nn2)),
                            "D.sd"=sqrt(sd1^2/sqrt(nn1) + sd2^2/sqrt(nn2)))
                    },vector("list",4)))
                }
            }
            myres[[i]] <- do.call("rbind",myres2)
        }
    }                                        

    # Computing Zr for noise
    cat("Computing Z for noise...\n")

    # 4.A) a0: Global for all R*G permutations
    myres <- do.call("rbind",myres)
    
    a0.M <- quantile(as.numeric(myres[,"M.sd"]),probs=0.9,na.rm=TRUE)
    a0.D <- quantile(as.numeric(myres[,"D.sd"]),probs=0.9,na.rm=TRUE)
    
    M <- as.numeric(myres[,"M"])/(a0.M + as.numeric(myres[,"M.sd"]))
    D <- as.numeric(myres[,"D"])/(a0.D + as.numeric(myres[,"D.sd"]))
    (M + D)/2
}

################################################################################

# filter.low.counts.R

CV = function(data) { 
    100*sd(data,na.rm=TRUE)/mean(data,na.rm=TRUE)
}

filtered.data <- function(dataset,factor,norm=TRUE,depth=NULL,method=1,
    cv.cutoff=100,cpm=1,p.adj="fdr") {    
    dataset0=dataset[rowSums(dataset) > 0,]
    dataset=dataset0

    if ((method == 3) && (norm)) {
        if (is.null(depth)) {
            stop("ERROR: Sequencing depth for each column in dataset must ",
                "be provided.\n")
        }
        dataset <- t(t(dataset0)/(colSums(dataset0)/depth))
    }
    
    if ((method < 3) && (!norm)) {
        dataset <- 10^6*t(t(dataset0)/colSums(dataset0))
    }

    grupos <- unique(factor)
    cumple <- NULL

    cat("Filtering out low count features...\n")

    for (gg in grupos) {        
        datos <- as.matrix(dataset[,grep(gg,factor)])

        if (method == 1) {
            if (ncol(datos) == 1) {
                cumplecond=(datos > cpm)
            } 
            else {
                cumplecond <- 
                    (apply(datos,1,CV) < cv.cutoff)*(rowMeans(datos) > cpm)
                cumplecond[which(is.na(cumplecond) == TRUE)] <- 0 
            }
            cumple <- cbind(cumple,cumplecond)
        }

        if (method == 2) {
            if (ncol(datos) == 1) 
                stop("ERROR: At least 2 replicates per condition are ",
                    "required to apply this method.")
            mytest <- apply(datos,1,function (x) { 
                suppressWarnings(wilcox.test(x,alternative="greater",
                    conf.int=FALSE,mu=0))$"p.value" 
            })
            mytest <- p.adjust(mytest,method=p.adj)
            cumple <- cbind(cumple,1*(mytest < 0.05))    
        }
        
        if (method == 3) {                     
            p0=cpm/10^6                                
            mytest <- apply(datos,1,function (x) 
                suppressWarnings(prop.test(sum(x),n=sum(datos),p=p0,
                    alternative="greater"))$"p.value")
            mytest <- p.adjust(mytest,method=p.adj)
            cumple <- cbind(cumple,1*(mytest < 0.05))                     
        }
    }

    cumple <- which(rowSums(as.matrix(cumple)) >= 1) 

    cat(paste(length(cumple),"features are to be kept for differential ",
        "expression analysis with filtering method",method))
    cat("\n")

    dataset0[cumple,]
}

################################################################################

################################################################################

# makeASCAdesign.R

make.ASCA.design <- function(x) {
    x <- as.factor(x)
    levels <- unique(x)
    n <- length(x)
    p <- length(levels)
    Design <- matrix(0,nrow=n,ncol=p)
    for (i in seq_len(n)){
        for (j in seq_len(p)){
            if (x[i]==levels[j]) {
                Design[i,j]=1
            }
        }
    }
    colnames(Design)<-levels
    output<-Design
    output
}

################################################################################

################################################################################

# MD.R

MD <- function (dat=dat,selec=c(seq_len(nrow(dat)))) {
    pares <- as.matrix(combn(ncol(dat), 2))
    if (NCOL(pares) > 30) {  
        sub30 <- sample(seq_len(NCOL(pares)),size=30,replace=FALSE)
        pares <- pares[,sub30]
    }
    mm <- NULL
    dd <- NULL
    for (i in seq_len(ncol(pares))) {
        a <- dat[selec,pares[1,i]]
        b <- dat[selec,pares[2,i]]
        mm <- cbind(mm,log(a/b,2))
        dd <- cbind(dd,abs(a-b))
    }
    list("M"=mm,"D"=dd)
}

MDbio <- function (dat=dat,selec= c(seq_len(nrow(dat))),param=NULL,a0per=0.9) {
    pares <- as.matrix(combn(ncol(dat), 2))
    mm <- NULL
    dd <- NULL
    for (i in seq_len(ncol(pares))) {
        a <- dat[selec,pares[1,i]]
        b <- dat[selec,pares[2,i]]
        mm <- cbind(mm, log(a/b, 2))
        dd <- cbind(dd, (a-b))  
    }

    ## Correcting (M,D) 
    sd.M <- sqrt(param$sd[,1]^2/(dat[,1]^2*log(2)^2*param$n[1]) + 
        param$sd[,2]^2/(dat[,2]^2*log(2)^2*param$n[2]))

    sd.D = sqrt(param$sd[,1]^2/param$n[1] + param$sd[,2]^2/param$n[2])

    if(is.null(a0per)) {
        a0.M = a0.D = 0
    } 
    else {
        if (a0per == "B") {
            B <- 100
            a0.M <- B*max(sd.M,na.rm=TRUE)
            a0.D <- B*max(sd.D,na.rm=TRUE)
        }
        else {
            a0per <- as.numeric(a0per)
            a0.M <- quantile(sd.M, probs = a0per, na.rm = TRUE)
            a0.D <- quantile(sd.D, probs = a0per, na.rm = TRUE)
        }
    }

    mm <- mm / (a0.M + sd.M)
    dd <- dd / (a0.D + sd.D)

    # Results
    list("M"=mm,"D"=dd)
}

################################################################################

################################################################################

# noiseq.R

noiseq <- function (input,k=0.5,norm=c("rpkm","uqua","tmm","n"),
    replicates=c("technical","biological","no"),factor=NULL,conditions=NULL,
    pnr=0.2,nss=5,v=0.02,lc=0) {
    if (inherits(input,"eSet") == FALSE)
        stop("Error. You must give an object generated by the readData ",
            "function\n")

    if (is.null(factor))
        stop("Error. You must specify the factor to know which conditions ",
            "you wish to compare. Please,give the argument 'factor'.\n")

    replicates <- match.arg(replicates)
    if (replicates == "biological") 
        print("WARNING: Your experiment has biological replicates. You ",
            "should consider using NOISeqBIO instead of NOISeq.")
    norm <- match.arg(norm)
    
    # (M,D) for signal and noise
    print("Computing (M,D) values...")
    miMD <- allMD(input,factor,k=k,replicates=replicates,norm=norm,
        conditions=conditions,pnr=pnr,nss=nss,v=v,lc=lc)

    ## Probability of differential expression

    print("Computing probability of differential expression...")
    prob.concounts <- probdeg(miMD$Ms,miMD$Ds,miMD$Mn,miMD$Dn)$prob

    if (!is.null(assayData(input)$exprs))
        todos <- rownames(as.matrix(assayData(input)$exprs))
    else
        todos <- rownames(as.matrix(assayData(input)$counts))
    
    prob <- prob.concounts[todos]
    names(prob) <- todos

    ## Results
    resultat <- data.frame("level_1"=miMD$Level1,"level_2"=miMD$Level2,
        "M"=miMD$Ms,"D"=miMD$Ds,"prob"=prob)
    rownames(resultat) <- todos

    # We change the name of the conditions to "name_mean"
    colnames(resultat)[1] <- 
        paste(unlist(strsplit(miMD$comp," "))[1],"mean",sep="_")
    colnames(resultat)[2] <- 
        paste(unlist(strsplit(miMD$comp," "))[3],"mean",sep="_")

    resultat <- data.frame(resultat,"ranking"=ranking(resultat)$statistic)
    if (!is.null(featureData(input)@data$Length))
        resultat <- data.frame(resultat,
            "Length"=as.numeric(as.character(featureData(input)@data[todos,
            "Length"])),stringsAsFactors=FALSE)
    if (!is.null(featureData(input)@data$GC))
        resultat <- data.frame(resultat,"GC"=as.numeric(as.character(
            featureData(input)@data[todos,"GC"])),stringsAsFactors=FALSE)
    if (!is.null(featureData(input)@data$Chromosome))
        resultat <- data.frame(resultat,"Chrom"=as.character(
            featureData(input)@data$Chromosome),"GeneStart"=as.numeric(
            as.character(featureData(input)@data$GeneStart)),
            "GeneEnd"=as.numeric(as.character(featureData(input)@data$GeneEnd)),
            stringsAsFactors=FALSE)
    if (!is.null(featureData(input)@data$Biotype))
        resultat <- data.frame(resultat,"Biotype"=as.character(
            featureData(input)@data[todos,"Biotype"]),stringsAsFactors=FALSE)
    #resultat[order(resultat[,5],decreasing=TRUE),]
    
    Output(data=list(resultat),method=norm,k=miMD$k,lc=lc,factor=factor,v=v,
        nss=nss,pnr=pnr,comparison=miMD$comp,replicates=replicates)
}

################################################################################

################################################################################

# noiseqbio.R

noiseqbio <- function (input,k=0.5,norm=c("rpkm","uqua","tmm","n"),nclust=15,
    plot=FALSE,factor=NULL,conditions=NULL,lc=0,r=50,adj=1.5,a0per=0.9,filter=1,
    depth=NULL,cv.cutoff=500,cpm=1) {
    if (inherits(input,"eSet") == FALSE)
        stop("ERROR: You must give an object generated by the readData ",
            "function\n")

    if (is.null(factor))
        stop("ERROR: You must specify the factor to know which conditions ",
            "you wish to compare. Please,give the argument 'factor'.\n")

    if (min(table(input@phenoData@data[,factor])) < 2)
        stop("ERROR: To run NOISeqBIO at least two replicates per condition ",
            "are needed. Please, run NOISeq if there are not enough ",
            "replicates in your experiment.\n")

    norm <- match.arg(norm)

    # Z-scores for signal and noise
    cat("Computing Z values...\n")
    miMD <- allMDbio(input,factor,k=k,norm=norm,conditions=conditions,lc=lc,
        r=r,a0per=a0per,nclust=nclust,filter=filter,depth=depth,
        cv.cutoff=cv.cutoff,cpm=cpm)

    ##### Probability of differential expression
    cat("Computing probability of differential expression...\n")

    ## KDE estimators of f0 and f    
    desde <- min(c(miMD$Zs,miMD$Zn),na.rm=TRUE)
    hasta <- max(c(miMD$Zs,miMD$Zn),na.rm=TRUE)

    fdens <- density(miMD$Zs,adjust=adj,n=5000,from=desde,to=hasta,na.rm=TRUE)
    f <- approxfun(fdens)

    f0dens <- density(miMD$Zn,adjust=adj,n=5000,from=desde,to=hasta,na.rm=TRUE)
    f0 <- approxfun(f0dens)

    if (f0(0)/f(0) < 1) 
        print("WARNING: f0(0)/f(0) < 1 => FP with Z~0 will be detected.")

    f0.f <- f0(miMD$Zs)/f(miMD$Zs)

    ## f0,f plot
    if (plot) {
        plot(f0dens,lwd=2,col=4,main=paste("r=",r,"; a0per=",a0per,sep=""),
            xlab="theta")
        lines(fdens,lwd=2,col=1)
        legend("topleft",c("f","f0"),col=c(1,4),lwd=2)
    }

    ## ESTIMATION of p0
    p0 <- min(1/f0.f,na.rm=TRUE)
    p0=min(p0,1)

    cat(paste("p0 =",p0)); cat("\n")

    ## PROBABILITY of DIFFERENTIAL EXPRESSION
    myprob <- 1 - p0*f0.f
    if (min(myprob,na.rm=TRUE) < 0) {
        print(summary(f0.f))
    }
    names(myprob) <- names(miMD$Zs)

    cat("Probability\n")
    print(summary(myprob))

    if (!is.null(assayData(input)$exprs))
        todos <- rownames(as.matrix(assayData(input)$exprs))
    else
        todos <- rownames(as.matrix(assayData(input)$counts))

    myprob <- myprob[todos]
    names(myprob) <- todos

    ## Results
    resultat <- data.frame("level_1"=miMD$Level1,"level_2"=miMD$Level2,
        "theta"=miMD$Zs,"prob"=myprob,"log2FC"=log2(miMD$Level1/miMD$Level2))
    rownames(resultat) <- todos

    colnames(resultat)[1] <- 
        paste(unlist(strsplit(miMD$comp," "))[1],"mean",sep="_")
    colnames(resultat)[2] <- 
        paste(unlist(strsplit(miMD$comp," "))[3],"mean",sep="_")

    # resultat <- data.frame(resultat,"ranking"=ranking(resultat)$statistic)
    if (!is.null(featureData(input)@data$Length))
        resultat <- data.frame(resultat,"length"=as.numeric(as.character(
            featureData(input)@data[todos,"Length"])),stringsAsFactors=FALSE)
    if (!is.null(featureData(input)@data$GC))
        resultat <- data.frame(resultat,"GC"=as.numeric(as.character(
            featureData(input)@data[todos,"GC"])),stringsAsFactors=FALSE)
    if (!is.null(featureData(input)@data$Chromosome))
        resultat <- data.frame(resultat,"Chrom"=(as.character(
            featureData(input)@data$Chromosome)),
            "GeneStart"=as.numeric(as.character(featureData(
            input)@data$GeneStart)),"GeneEnd"=as.numeric(as.character(
            featureData(input)@data$GeneEnd)),stringsAsFactors=FALSE)

    if (!is.null(featureData(input)@data$Biotype))
        resultat <- data.frame(resultat,"Biotype"=as.character(
            featureData(input)@data[todos,"Biotype"]),stringsAsFactors=FALSE)
    
    Output(data=list(resultat),method=norm,k=miMD$k,lc=lc,factor=factor,
        comparison=miMD$comp,replicates="biological",v=0,nss=0,pnr=0)
}

################################################################################

################################################################################

# normalization.R

noitmm <- function (datos,long=1000,lc=0,k=0,refColumn=1,logratioTrim=.3,
    sumTrim=0.05,doWeighting=TRUE,Acutoff=-1e10) {
    if (!is.null(ncol(long))) {
        mynames <- long[,1]
        long <- long[,2]
        names(long) <- mynames
    }

    L <- (long/1000)^lc
    datos <- datos/L

    total <- colSums(as.matrix(datos))

    datos0 <- sinceros(datos,k)

    if (ncol(as.matrix(datos)) > 1) {
        fk <- .calcNormFactors(as.matrix(datos),refColumn=refColumn,
            method="TMM",logratioTrim=logratioTrim,sumTrim=sumTrim,
            doWeighting=doWeighting,Acutoff=Acutoff)

        fk <- fk * (total/mean(total))
        datos.norm <- t(t(datos0)/fk)
    } 
    else {
        datos.norm <- datos0/L
    }
    na.omit(datos.norm)    
    
}

noirpkm <- function (datos,long=1000,lc=1,k=0) {
    if (!is.null(ncol(long))) {
        mynames=long[,1]
        long=long[,2]
        names(long)=mynames
    }

    total <- colSums(as.matrix(datos))
    datos0 <- sinceros(datos,k)                                     
    datos.norm <- (t(t(datos0)/total)*10^6)/((long/1000)^lc)
    na.omit(datos.norm)     
}

noiuqua <- function (datos,long=1000,lc=0,k=0) {
    if (!is.null(ncol(long))) {
        mynames=long[,1]
        long=long[,2]
        names(long)=mynames
    }

    L <- (long/1000)^lc
    datos=datos/L

    datos0 <- sinceros(datos,k)

    if (ncol(as.matrix(datos)) > 1) {
        sumatot <- rowSums(datos)
        supertot <- sum(sumatot)
        counts0 <- which(sumatot == 0)

        if (length(counts0) > 0) {
            datitos <- datos[-counts0,]
        } else {
            datitos <- datos
        }

        q3 <- apply(datitos,2,quantile,probs=0.75)
        d <- q3*supertot/sum(q3)

        datos.norm <- t(t(datos0)/d)*10^6

    } else {

        datos.norm <- datos0/L

    }
    
    na.omit(datos.norm)
}

## Taken from the edgeR package with minor modifications
.calcNormFactors <- function(object,method=c("TMM","quantile"),refColumn=NULL,
    logratioTrim=.3,sumTrim=0.05,doWeighting=TRUE,Acutoff=-1e10,quantile=0.75) {
    method <- match.arg(method)
    if( is.matrix(object) ) {
        if(is.null(refColumn))
            refColumn <- 1
        data <- object
        libsize <- colSums(data)
    } 
    else {
        stop(".calcNormFactors() only operates on 'matrix' objects")
    }

    f <- switch(method,
        TMM = apply(data,2,.calcFactorWeighted,ref=data[,refColumn],
            logratioTrim=logratioTrim,sumTrim=sumTrim,doWeighting=doWeighting,
            Acutoff=Acutoff),quantile=.calcFactorQuantile(data,libsize,
            q=quantile))
    f <- f/exp(mean(log(f)))

    return(f)
}

.calcFactorQuantile <- function(data,lib.size,q=0.75) {
    y <- t(t(data)/lib.size)
    f <- apply(y,2,function(x) quantile(x,p=q))
    f/exp(mean(log(f)))
}

.calcFactorWeighted <- function(obs,ref,logratioTrim=.3,sumTrim=0.05,
    doWeighting=TRUE,Acutoff=-1e10) {
    if( all(obs==ref) )
        return(1)

    obs <- as.numeric(obs)
    ref <- as.numeric(ref)

    nO <- sum(obs)
    nR <- sum(ref)
    # log ratio of expression,accounting for library size
    logR <- log2((obs/nO)/(ref/nR)) 
    absE <- (log2(obs/nO) + log2(ref/nR))/2 # absolute expression
    v <- (nO-obs)/nO/obs + (nR-ref)/nR/ref # estimated asymptotic variance

    # remove infinite values,cutoff based on A
    fin <- is.finite(logR) & is.finite(absE) & (absE > Acutoff)

    logR <- logR[fin]
    absE <- absE[fin]
    v <- v[fin]

    # taken from the original mean() function
    n <- sum(fin)
    loL <- floor(n * logratioTrim) + 1
    hiL <- n + 1 - loL
    loS <- floor(n * sumTrim) + 1
    hiS <- n + 1 - loS
    
    #keep <- (rank(logR) %in% loL:hiL) & (rank(absE) %in% loS:hiS)
    # a fix from leonardo ivan almonacid cardenas,since rank() can return
    # non-integer values when there are a lot of ties
    keep <- (rank(logR)>=loL & rank(logR)<=hiL) & (rank(absE)>=loS 
        & rank(absE)<=hiS)
    
    if (doWeighting) 
        2^(sum(logR[keep]/v[keep],na.rm=TRUE) / sum(1/v[keep],na.rm=TRUE))
    else
        2^(mean(logR[keep],na.rm=TRUE))
}

################################################################################

################################################################################

# PCA.GENES.R

PCA.GENES <- function(X) {
    n <- ncol(X)
    p <- nrow(X)
    offset <- apply(X,2,mean)
    Xoff <- X-(cbind(matrix(1,p,1))%*%rbind(offset))

    eigen <- eigen(Xoff%*%t(Xoff)/(p-1))
    var <- cbind(eigen$values/sum(eigen$values),
        cumsum(eigen$values/sum(eigen$values)))

    loadings2 <- eigen$vectors
    scores2 <- t(Xoff)%*%loadings2

    normas2 <- sqrt(apply(scores2^2,2,sum))

    scores1 <- loadings2%*%diag(normas2)
    loadings1 <- scores2%*%diag(1/normas2)

    output <- list(eigen,var,scores1,loadings1)
    names(output) <- c("eigen","var.exp","scores","loadings")
    output
}

################################################################################

################################################################################

# probdeg.R

probdeg <- function (Mg,Dg,Mn,Dn,prec=2) {
    tot <- length(Mn)   # number of points in noise distribution
    gens <- names(Mg)
    Mruido <- abs(round(Mn,prec))
    Druido <- round(Dn,prec)
    Mgen <- abs(round(Mg,prec))
    Dgen <- round(Dg,prec)
    MDgen <- na.omit(cbind(Mgen, Dgen))
    MDunic <- unique(MDgen)
    Nres <- apply(MDunic,1,n.menor,S1=Mruido,S2=Druido)
    lugares <- apply(MDgen,1,busca,S=MDunic)
    Nconj <- Nres[lugares]
    names(Nconj) <- names(lugares)
    laprob <- Nconj / tot
    laprob <- laprob[gens]
    names(laprob) <- gens
    Nconj <- Nconj[gens]
    names(Nconj) <- gens
    laprob <- list("prob"=laprob,"numDE"=Nconj,"numNOISE"=tot)
    laprob
}

################################################################################

################################################################################

# readData.R

readData <- function(data=NULL,factors=NULL,length=NULL,biotype=NULL,
    chromosome=NULL,gc=NULL) {
    if (is.null(data))
        stop("Expression information must be provided to the readData function")

    if (is.null(factors))
        stop("Condition information must be provided to the readData funcion")

    if (!is.null(length) && !is.vector(length) && !is.data.frame(length)
        && !is.matrix(length))
        stop( "The length info should be a vector or a data.frame/matrix.")

    if (!is.null(gc) && !is.vector(gc) && !is.data.frame(gc) && !is.matrix(gc))
        stop( "The GC content info should be a vector or a data.frame/matrix.")
    
    if (!is.null(chromosome) && ncol(chromosome) != 3)
        stop( "The chromosome object should be a matrix or data.frame with ",
            "3 columns: chromosome,start position and end position.")

    if (!is.null(biotype) && !is.vector(biotype) && !is.data.frame(biotype) 
        && !is.matrix(biotype))
        stop( "The biotype info should be a vector or a data.frame/matrix.")

    countData <- as.matrix(data)

    rowNames <- rownames(countData)
    
    if (nrow(factors) == ncol(countData)) {
        rownames(factors) <- colnames(countData)
    }
    else {
        stop ("Number of rows in factors must be equal to number of ",
            "columns in data.\n")
    }

    pheno <- AnnotatedDataFrame(data=as.data.frame(factors))
    input <- ExpressionSet(assayData=countData,phenoData=pheno)

    if (!is.null(length))
        input <- addData(data=input,length=length)

    if (!is.null(gc))
        input <- addData(data=input,gc=gc)

    if (!is.null(biotype))
        input <- addData(data=input,biotype=biotype)

    if (!is.null(chromosome))
        input <- addData(data=input,chromosome=chromosome)

    input
}

addData <- function(data,length=NULL,biotype=NULL,chromosome=NULL,factors=NULL,
    gc=NULL) {
    if (!inherits(data,"eSet"))
        stop("Error. You must give an eSet object.")

    if (!is.null(length) && !is.vector(length) && !is.data.frame(length) 
        && !is.matrix(length))
        stop( "The length info should be a vector or a data.frame/matrix.")

    if (!is.null(gc) && !is.vector(gc) && !is.data.frame(gc) && !is.matrix(gc))
        stop( "The GC content info should be a vector or a data.frame/matrix.")

    if (!is.null(biotype) && !is.vector(biotype) && !is.data.frame(biotype) 
        && !is.matrix(biotype))
        stop( "The biotype info should be a vector or a data.frame/matrix.")
    
    if (!is.null(chromosome) && ncol(chromosome) != 3)
        stop( "The chromosome object should be a matrix or data.frame with ",
            "3 columns: chromosome,start position and end position.")

    if (!is.null(assayData(data)$exprs))
        rowNames <- rownames(assayData(data)$exprs)
    else
        rowNames <- rownames(assayData(data)$counts)

    # If exists length
    if (!is.null(length)) {
        Length <- rep(NA,length(rowNames))
        names(Length) <- rowNames
        if (is.vector(length)) {
            Length[rowNames] <- as.numeric(as.character(length[rowNames]))
        } 
        else if (is.data.frame(length) || is.matrix(length)) {
            if (ncol(length) == 2) {
                # We assume that the feature names are in the first column and 
                # the length in the second
                rownames(length) <- length[,1]
                Length[rowNames] <- as.numeric(as.character(length[rowNames,2]))
            } 
            else if (ncol(length) == 1) {
                # We assume that the length are in the first column and the 
                # feature names in the rownames
                Length[rowNames] <- as.numeric(as.character( 
                    length[rowNames,1]))
            } 
            else {
                stop( "The length matrix/data.frame contains more columns ",
                    "than expected.")
            }
        }
        featureData(data)@data <- cbind(featureData(data)@data,Length)
    }

    # If exists gc
    if (!is.null(gc)) {
        GC <- rep(NA,length(rowNames))
        names(GC) <- rowNames
        if (is.vector(gc)) {
            GC[rowNames] <- as.numeric(as.character(gc[rowNames]))
        } 
        else if (is.data.frame(gc) || is.matrix(gc)) {
            if (ncol(gc) == 2) {
                # We assume that the feature names are in the first column and 
                # the GC content in the second
                rownames(gc) <- gc[,1]
                GC[rowNames] <- as.numeric(as.character(gc[rowNames,2]))
            } 
            else if (ncol(gc) == 1) {
                # We assume that the GC contents are in the first column 
                # and the feature names in the rownames
                GC[rowNames] <- as.numeric(as.character(gc[rowNames,1]))
            } 
            else {
                stop( "The GC matrix/data.frame contains more columns ",
                    "than expected.")
            }
        }
        featureData(data)@data <- cbind(featureData(data)@data,GC)
    }
    
    # If exists biotype
    if (!is.null(biotype)) {
        Biotype <- rep(NA,length(rowNames))
        names(Biotype) <- rowNames
        if (is.vector(biotype)) {
            Biotype[rowNames] <- as.character(biotype[rowNames])
        } 
        else if (is.data.frame(biotype) || is.matrix(biotype)) {
            if (ncol(biotype) == 2) {
                # We assume that the feature names are in the first column and 
                # the biotypes in the second
                rownames(biotype) <- biotype[,1]
                Biotype[rowNames] <- as.character( biotype[rowNames,2])
            } 
            else if (ncol(biotype) == 1) {
                # We assume that the biotypes are in the first column and 
                # the feature names in the rownames
                Biotype[rowNames] <- as.character( biotype[rowNames,1]    )
            } 
            else {
                stop("The biotype matrix/data.frame contains more ",
                    "columns than expected.")
            }
        }        
        featureData(data)@data <- cbind(featureData(data)@data,Biotype)
        featureData(data)@data$Biotype <- 
            as.character(featureData(data)@data$Biotype)
    }
    
    # If exists chromosome
    if (!is.null(chromosome)) {
        Chromosome <- GeneStart <- GeneEnd <- rep(NA,length(rowNames))
        names(Chromosome) <- names(GeneStart) <- names(GeneEnd) <- rowNames
        Chromosome[rowNames] <- as.character(chromosome[rowNames,1])
        GeneStart[rowNames] <- as.numeric(as.character(chromosome[rowNames,2]))
        GeneEnd[rowNames] <- as.numeric(as.character(chromosome[rowNames,3]))
        featureData(data)@data <- cbind(featureData(data)@data,Chromosome,
            GeneStart,GeneEnd)
    }

    # If exists new factors
    if (!is.null(factors))
        phenoData(data)@data <- cbind(phenoData(data)@data,factors)
    
    data
}

################################################################################

################################################################################

# saturation.plot.R

#### SATURATION PLOTS

## Data for saturation plot (with or without biotypes)

saturation.dat <- function (input,k=0,biotypes=NULL,ndepth=6) {
    if (inherits(input,"eSet") == FALSE)
        stop("Error. You must give an eSet object\n")

    if (!is.null(assayData(input)$exprs))
        datos <- assayData(input)$exprs
    else
        datos <- assayData(input)$counts
    
    if (!is.null(featureData(input)$Biotype))
        infobio <- featureData(input)$Biotype        
    else
        infobio <- NULL

    nsam <- NCOL(datos)

    if (!is.null(infobio)) {        
        if(is.null(biotypes)) {
            biotypes <- unique(infobio)
            names(biotypes) <- biotypes                 
        }     
    } else { biotypes=NULL }     

    satura <- vector("list",length=length(biotypes)+1)
    names(satura) <- c("global",names(biotypes))

    ndepth1 <- ceiling(ndepth/2)

    datos <- round(datos,0)
    datos0 <- datos + 0.2
    
    # Random subsamples for each sample
    submuestras <- seq.depth <- vector("list",length=nsam)
    names(submuestras) <- names(seq.depth) <- colnames(datos)

    for (n in seq_len(nsam)) { # simulating subsamples for each sample
        total <- sum(datos[,n]) # total counts in sample n
        # simulation for each depth and real depth
        varias <- vector("list",length=ndepth+1)

        for (i in seq_len(ndepth1)) { # simulating depths < real depth
            muestra <- rmultinom(10,size=round(total/(ndepth1+1),0),
                prob=datos[,n])
            if (i == 1)
                varias[[i]] <- muestra 
            else
                varias[[i]] <- varias[[i-1]] + muestra
        }

        varias[[ndepth1+1]] <- as.matrix(datos[,n])

        for (i in (ndepth1+2):(ndepth+1)) { # simulating depths < real depth
            muestra <- rmultinom(10,size=round(total/(ndepth1+1),0),
                prob=datos0[,n])
            if (i == ndepth1+2)
                varias[[i]] <- matrix(varias[[i-1]],ncol=10,
                    nrow=nrow(varias[[i-1]])) + muestra
            else
                varias[[i]] <- varias[[i-1]] + muestra
        }

        submuestras[[n]] <- varias
        seq.depth[[n]] <- c(round(total/(ndepth1+1),0)*seq_len(ndepth1),total,
            round(total/(ndepth1+1),0)*((ndepth1+2):(ndepth+1)))
    }

    # Global saturation
    satura[[1]] <- vector("list",length=nsam)
    names(satura[[1]]) <- colnames(datos)

    for (n in seq_len(nsam)) { # for each sample
        satura[[1]][[n]] <- vapply(submuestras[[n]],function(x) { 
            mean(apply(x,2,noceros,k=k)) 
        },numeric(1))
    }

    # Per biotypes
    if (!is.null(infobio)) { # if biotypes available
        biog <- lapply(biotypes,function(x) { 
            which(is.element(infobio,x)) 
        })
        names(biog) <- names(biotypes)
        for (j in 2:length(satura)) { # for each biotype
            satura[[j]] <- vector("list",length=nsam)
            names(satura[[j]]) <- colnames(datos)
            for (n in seq_len(nsam)) { # for each sample
                conbio <- lapply(submuestras[[n]],function(x) { 
                    as.matrix(x[biog[[j-1]],]) 
                })
                satura[[j]][[n]] <- vapply(conbio,function(x) { 
                    mean(apply(x,2,noceros,k=k)) 
                },numeric(1))
            }
        }
    } 
    else
        biog <- NULL

    # computing detection increasing per million reads
    newdet <- vector("list",length=1+length(biotypes))
    names(newdet) <- c("global",names(biotypes))
    
    for (j in seq_len(length(newdet))) {
        newdet[[j]] <- vector("list",length=nsam)
        names(newdet[[j]]) <- colnames(datos)
        for (n in seq_len(nsam)) {
            puntos <- data.frame("x"=seq.depth[[n]],"y"=satura[[j]][[n]])
            pendi <- NULL
            for(i in 2:nrow(puntos))                
                pendi <- c(pendi,
                    (puntos$y[i]-puntos$y[i-1])/(puntos$x[i]-puntos$x[i-1]))
            newdet[[j]][[n]] <- c(NA,pendi*1000000)
        }
    }

    bionum <- c(NROW(datos),vapply(biog,length,integer(1)))
    names(bionum) <- c("global",names(biog))
    
    # Results at real sequencing depth
    real <- vector("list",length=length(satura))
    names(real) <- names(satura)
    realdepth <- vapply(seq.depth,function (x) x[ndepth1+1],numeric(1))/10^6
    for (i in seq_len(length(real))) {
        real[[i]] <- data.frame("depth"=realdepth,
            "detec"=vapply(satura[[i]],function (x) x[ndepth1+1],numeric(1)))
        rownames(real[[i]]) <- colnames(datos)
    }
    
    # Results
    satura <- list("saturation"=satura,"bionum"=bionum,"depth"=seq.depth,
        "newdet"=newdet,"real"=real)
    
    satura
}

################################################################################

metaseqrPlot <- function(object,sampleList,annotation=NULL,contrastList=NULL,
    pList=NULL,thresholds=list(p=0.05,f=1),plotType=c("mds","biodetection",
    "countsbio","saturation","readnoise","rnacomp","correl","pairs","boxplot",
    "gcbias","lengthbias","meandiff","meanvar","deheatmap","volcano","biodist",
    "filtered","mastat","deregulogram","statvenn","foldvenn"),isNorm=FALSE,
    output="x11",path=NULL,...) {
    if (is(object,"GenomicRanges")) {
        object <- as.data.frame(object)
        object <- object[,c(1,2,3,6,7,5,8,9)]
        colnames(object)[1] <- "chromosome"
    }
    if (!is.matrix(object) && !is.data.frame(object))
        stopwrap("object argument must be a matrix or data frame!")
    if (is.null(annotation) && any(plotType %in% c("biodetection",
        "countsbio","saturation","rnacomp","readnoise","biodist","gcbias",
        "lengthbias","filtered")))
        stopwrap("annotation argument is needed when plotType is\n",
            "\"biodetection\", \"countsbio\",\"saturation\",\"rnacomp\", ",
            "\"readnoise\", \"biodist\", \"gcbias\", \"lengthbias\", ",
            "\"filtered\", \"statvenn\" or \"foldvenn\"!")
    if (any(plotType %in% c("deheatmap","volcano","biodist","deregulogram",
        "statvenn","foldvenn"))) {
        if (is.null(contrastList))
            stopwrap("contrastList argument is needed when plotType is ",
                "\"deheatmap\",\"volcano\", \"biodist\", \"deregulogram\", ",
                "\"mastat\", \"statvenn\" or \"foldvenn\"!")
        if (is.null(pList))
            stopwrap("The p argument which is a list of p-values for each ",
                "contrast is needed when plotType is\n\"deheatmap\", ",
                "\"volcano\", \"mastat\", \"biodist\", \"deregulogram\", ",
                "\"statvenn\" or \"foldvenn\"!")
        if (is.na(thresholds$p) || is.null(thresholds$p) || thresholds$p==1) {
            warnwrap(paste("The p-value threshold when plotType is ",
            "\"deheatmap\", \"volcano\", \"biodist\", \"mastat\", ",
            "\"deregulogram\", \"statvenn\" or \"foldvenn\"! must allow ",
            "the normal plotting of DEG diagnostic plots! Setting to 0.05..."))
            thresholds$p <- 0.05
        }
    }
    if (is.null(path)) 
        path <- getwd()
    if (is.data.frame(object) && !("filtered" %in% plotType)) 
        object <- as.matrix(object)
    
    if (is(annotation,"GenomicRanges")) {
        annotation <- as.data.frame(annotation)
        annotation <- annotation[,c(1,2,3,6,7,5,8,9)]
        colnames(annotation)[1] <- "chromosome"
    }
    
    if (any(plotType %in% c("biodetection","countsbio","saturation",
        "rnacomp","biodist","readnoise")))
        covars <- list(
            data=object,
            length=annotation$end - annotation$start,
            gc=annotation$gc_content,
            chromosome=annotation[,c(1,2,3)],
            factors=data.frame(class=asClassVector(sampleList)),
            biotype=annotation$biotype,
            gene_name=as.character(annotation$gene_name)
        )

    rawPlots <- c("mds","biodetection","countsbio","saturation","readnoise",
        "correl","pairwise")
    normPlots <- c("boxplot","gcbias","lengthbias","meandiff","meanvar",
        "rnacomp")
    statPlots <- c("deheatmap","volcano","mastat","biodist","deregulogram")
    otherPlots <- c("filtered")
    vennPlots <- c("statvenn","foldvenn")
    files <- list()
    
    for (p in plotType) {
        disp("  Plotting ",p,"...")
        if (p %in% rawPlots && !isNorm) {
            switch(p,
                mds = {
                    files$mds <- diagplotMds(object,sampleList,output=output,
                        path=path)
                },
                biodetection = {
                    files$biodetection <- diagplotNoiseq(object,sampleList,
                        covars,whichPlot=p,output=output,path=path,...)
                },
                countsbio = {
                    files$countsbio <- diagplotNoiseq(object,sampleList,
                        covars,whichPlot=p,output=output,path=path,...)
                },
                saturation = {
                    fil <- diagplotNoiseq(object,sampleList,covars,
                        whichPlot=p,output=output,path=path,...)
                    files$saturation$biotype <- fil[["biotype"]]
                    files$saturation$sample <- fil[["sample"]]
                },
                readnoise = {
                    files$readnoise <- diagplotNoiseq(object,sampleList,
                        covars,whichPlot=p,output=output,path=path,...)
                },
                correl = {
                    files$correl$heatmap <- diagplotCor(object,type="heatmap",
                        output=output,path=path,...)
                    files$correl$correlogram <- diagplotCor(object,
                        type="correlogram",output=output,path=path,...)
                },
                pairwise = {
                    files$pairwise <- diagplotPairs(object,output=output,
                        path=path)
                }
            )
        }
        if (p %in% normPlots) {
            switch(p,
                boxplot = {
                    files$boxplot <- diagplotBoxplot(object,name=sampleList,
                        isNorm=isNorm,output=output,path=path,...)
                },
                gcbias = {
                    files$gcbias <- diagplotEdaseq(object,sampleList,
                        covar=annotation$gc_content,isNorm=isNorm,
                        whichPlot=p,output=output,path=path,...)
                },
                lengthbias = {
                    files$lengthbias <- diagplotEdaseq(object,sampleList,
                        covar=annotation$end-annotation$start,isNorm=isNorm,
                        whichPlot=p,output=output,path=path,...)
                },
                meandiff = {
                    fil <- diagplotEdaseq(object,sampleList,isNorm=isNorm,
                        whichPlot=p,output=output,path=path,...)
                    for (n in names(fil)) {
                        if (!is.null(fil[[n]]))
                            files$meandiff[[n]] <- unlist(fil[[n]])
                    }
                },
                meanvar = {
                    fil <- diagplotEdaseq(object,sampleList,isNorm=isNorm,
                        whichPlot=p,output=output,path=path,...)
                    for (n in names(fil)) {
                        if (!is.null(fil[[n]]))
                            files$meanvar[[n]] <- unlist(fil[[n]])
                    }
                },
                rnacomp = {
                    files$rnacomp <- diagplotNoiseq(object,sampleList,covars,
                        whichPlot=p,output=output,isNorm=isNorm,path=path,...)
                }
            )
        }
        if (p %in% statPlots && isNorm) {
            for (cnt in names(contrastList)) {
                disp("  Contrast: ",cnt)                
                samples <- names(unlist(contrastList[[cnt]]))
                mat <- as.matrix(object[,match(samples,colnames(object))])
                switch(p,
                    deheatmap = {
                        files$deheatmap[[cnt]] <- c(
                            diagplotDeHeatmap(mat,scale="asis",cnt,
                                output=output,path=path),
                            diagplotDeHeatmap(mat,scale="zscore",cnt,
                                output=output,path=path)
                        )
                    },
                    volcano = {
                        fc <- log2(makeFoldChange(cnt,sampleList,object,1))
                        for (contrast in colnames(fc)) {
                            files$volcano[[contrast]] <- diagplotVolcano(
                                fc[,contrast],pList[[cnt]],contrast,
                                fcut=thresholds$f,pcut=thresholds$p,
                                output=output,path=path)
                        }
                    },
                    mastat = {
                        m <- log2(makeFoldChange(cnt,sampleList,object,1))
                        a <- makeA(cnt,sampleList,object,1)
                        for (contrast in colnames(m)) {
                            files$mastat[[contrast]] <- diagplotMa(
                                m[,contrast],a[,contrast],pList[[cnt]],
                                contrast,fcut=thresholds$f,pcut=thresholds$p,
                                output=output,path=path)
                        }
                    },
                    biodist = {
                        files$biodist[[cnt]] <- diagplotNoiseq(object,
                            sampleList,covars,whichPlot=p,output=output,
                            biodistOpts=list(p=pList[[cnt]],
                            pcut=thresholds$p,name=cnt),path=path,...)
                    }
                )
            }
            if ("deregulogram" %in% plotType) {
                cntPairs <- combn(names(contrastList),2)
                files$deregulogram <- character(ncol(cntPairs))
                for (i in seq_len(ncol(cntPairs))) {
                    fmat <- cbind(
                        log2(makeFoldChange(cntPairs[1,i],sampleList,
                            object,1))[,1,drop=FALSE],
                        log2(makeFoldChange(cntPairs[2,i],sampleList,
                            object,1))[,1,drop=FALSE]
                    )
                    pmat <- do.call("cbind",pList[c(cntPairs[1,i],
                        cntPairs[2,i])])
                    colnames(pmat) <- colnames(fmat)
                    files$deregulogram[i] <- diagplotDeregulogram(fmat,pmat,
                        fcut=thresholds$f,pcut=thresholds$p,output=output,
                        path=path) 
                }
            }
        }
        if (p %in% otherPlots) {
            switch(p,
                filtered = {
                    files$filtered <- diagplotFiltered(object,annotation,
                        output=output,path=path)
                }
            )
        }
        if (p %in% vennPlots) {
            switch(p,
                statvenn = {
                    for (cnt in names(contrastList)) {
                        disp("  Contrast: ",cnt)
                        if (!is.null(annotation)) {
                            altNames <- as.character(annotation$gene_name)
                            names(altNames) <- rownames(annotation)
                        }
                        else
                            altNames <- NULL
                        files$venn[[cnt]] <- diagplotVenn(pList[[cnt]],
                            pcut=thresholds$p,nam=cnt,output=output,path=path,
                            altNames=altNames)
                    }
                }
            )
        }
    }
    
    return(files)
}

diagplotBoxplot <- function(mat,name=NULL,logIt="auto",yLim="default",
    isNorm=FALSE,output="x11",path=NULL,altNames=NULL,...) {
    if (is.null(path)) 
        path <- getwd()
    if (isNorm)
        status<- "normalized"
    else
        status<- "raw"
    # Need to log?
    if (logIt=="auto") {
        if (diff(range(mat,na.rm=TRUE))>1000)
            mat <- log2disp(mat)
    }
    else if (logIt=="yes")
        mat <- log2disp(mat)
    # Define the axis limits based on user input
    if (!is.numeric(yLim) && yLim=="default") {
        minY <- floor(min(mat))
        maxY <- ceiling(max(mat))
    }
    else if (is.numeric(yLim)) {
        minY <- yLim[1]
        maxY <- yLim[2]
    }
    grouped <- FALSE
    if (is.null(name)) {
        if (is.null(colnames(mat)))
            nams <- paste("Sample",seq_len(ncol(mat)),sep=" ")
        else
            nams <- colnames(mat)
    }
    else if (length(name)==1 && name=="none")
        nams <- rep("",ncol(mat))
    else if (is.list(name)) { # Is sampleList
        nams <- unlist(name)
        grouped <- TRUE
    }
    cols <- c("red3","green3","blue2","gold","skyblue","orange3","burlywood",
        "red","blue","green","orange","darkgrey","green4","black","pink",
        "brown","magenta","yellowgreen","pink4","seagreen4","darkcyan")
    if (grouped) {
        tmp <- as.numeric(factor(asClassVector(name)))
        bCols <- cols[tmp]
    }
    else bCols <- cols
    matList <- list()
    for (i in seq_len(ncol(mat)))
        matList[[i]] <- mat[,i]
    names(matList) <- nams
    if (output != "json") {
        fil <- file.path(path,paste("boxplot_",status,".",output,sep=""))
        graphicsOpen(output,fil)
        if (!is.numeric(yLim) && yLim=="default")
            b <- boxplot(matList,names=nams,col=bCols,las=2,main=paste(
                "Boxplot ",status,sep=""),...)
        else
            b <- boxplot(matList,names=nams,col=bCols,ylim=c(minY,maxY),
                las=2,main=paste("Boxplot ",status,sep=""),...)
        graphicsClose(output)
        return(fil)
    }
    else {
        # Create boxplot object
        b <- boxplot(matList,plot=FALSE)
        colnames(b$stat) <- nams
        # Locate the outliers
        oList <- lapply(names(matList),function(x,M,b) {
            v <- b[,x]
            o <- which(M[[x]]<v[1] | M[[x]]>v[5])
            if (length(o)>0)
                return(M[[x]][o])
            else
                return(NULL)
        },matList,b$stat)
        # Create output object
        obj <- list(
            x=NULL,
            y=NULL,
            plot=b,
            samples=name,
            ylims=c(minY,maxY),
            xlims=NULL,
            status=status,
            pcut=NULL,
            fcut=NULL,
            altnames=altNames,
            user=oList
        )
        json <- boxplotToJSON(obj)
        #fil <- file.path(path,paste("boxplot_",status,".json",sep=""))
        #disp("Writing ",fil)
        #write(json,fil)
        return(json)
    }
}

diagplotMds <- function(x,sampleList,method="spearman",logIt=TRUE,
    output="x11",path=NULL,...) {
    if (is.null(path)) path <- getwd()
    classes <- as.factor(asClassVector(sampleList))
    design <- as.numeric(classes)
    colspace <- c("red","blue","yellowgreen","orange","aquamarine2",
        "pink2","seagreen4","brown","purple","chocolate")
    pchspace <- c(20,17,15,16,8,3,2,0,1,4)
    if (ncol(x)<3) {
        warnwrap("MDS plot cannot be created with less than 3 samples! ",
            "Skipping...")
        return(NULL)
    }
    if (logIt)
        y <- nat2log(x,base=2)
    else
        y <- x
    d <- as.dist(0.5*(1-cor(y,method=method)))
    mdsObj <- cmdscale(d,eig=TRUE,k=2)
    xr <- diff(range(min(mdsObj$points[,1]),max(mdsObj$points[,1])))
    yr <- diff(range(min(mdsObj$points[,2]),max(mdsObj$points[,2])))
    xlim <- c(min(mdsObj$points[,1])-xr/10,max(mdsObj$points[,1])+xr/10)
    ylim <- c(min(mdsObj$points[,2])-yr/10,max(mdsObj$points[,2])+yr/10)
    if (output!="json") {
        fil <- file.path(path,paste("mds.",output,sep=""))
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=9,height=7)
        else
            graphicsOpen(output,fil,width=1024,height=768)     
        plot(mdsObj$points[,1],mdsObj$points[,2],
            col=colspace[seq_len(length(levels(classes)))][design],
            pch=pchspace[seq_len(length(levels(classes)))][design],
            xlim=xlim,ylim=ylim,
            main="MDS plot",xlab="MDS 1",ylab="MDS 2",
            cex=0.9,cex.lab=0.9,cex.axis=0.9,cex.main=0.9)
        text(mdsObj$points[,1],mdsObj$points[,2],labels=colnames(x),pos=3,
            cex=0.7)
        grid()
        graphicsClose(output)
        return(fil)
    }
    else {
        # Create output object
        xx <- mdsObj$points[,1]
        yy <- mdsObj$points[,2]
        names(xx) <- names(yy) <- unlist(sampleList)
        obj <- list(
            x=xx,
            y=yy,
            plot=NULL,
            samples=sampleList,
            ylim=ylim,
            xlim=xlim,
            status=NULL,
            pcut=NULL,
            fcut=NULL,
            altnames=NULL,
            user=NULL
        )
        json <- mdsToJSON(obj)
        return(json)
        #fil <- file.path(path,"mds.json")
        #disp("Writing ",fil)
        #write(json,fil)
    }
}

#.diagplotMdsGg <- function(x,sampleList,method="spearman",logIt=TRUE,...) {
#    classes <- as.factor(asClassVector(sampleList))
#    design <- as.numeric(classes)
#    if (ncol(x)<3) {
#        warnwrap("MDS plot cannot be created with less than 3 samples! ",
#            "Skipping...")
#        return("MDS plot cannot be created with less than 3 samples!")
#    }
#    if (logIt)
#        y <- nat2log(x,base=2)
#    else
#        y <- x
#    d <- as.dist(0.5*(1-cor(y,method=method)))
#    mdsObj <- cmdscale(d,eig=TRUE,k=2)
#    
#    xr <- diff(range(min(mdsObj$points[,1]),max(mdsObj$points[,1])))
#    yr <- diff(range(min(mdsObj$points[,2]),max(mdsObj$points[,2])))
#    xlims <- c(min(mdsObj$points[,1])-xr/10,max(mdsObj$points[,1])+xr/10)
#    ylims <- c(min(mdsObj$points[,2])-yr/10,max(mdsObj$points[,2])+yr/10)
#   
#   plotData <- data.frame(
#       x=mdsObj$points[,1],
#       y=mdsObj$points[,2],
#       Condition=classes
#   )
#   rownames(plotData) <- colnames(x)
#   
#   mds <- ggplot() +
#       geom_point(data=plotData,mapping=aes(x=x,y=y,colour=Condition,
#           shape=Condition),size=4) +
#       xlim(xlims[1],xlims[2]) + 
#       ylim(ylims[1],ylims[2]) +
#       ggtitle("MDS plot") +
#       xlab("\nPrincipal coordinate 1") +
#       ylab("Principal coordinate 2\n") +
#       theme_bw() +
#       theme(axis.title.x=element_text(size=12,face="bold"),
#           axis.title.y=element_text(size=12,face="bold"),
#           axis.text.x=element_text(size=11,face="bold"),
#           axis.text.y=element_text(size=11,face="bold"),
#           strip.text.x=element_text(size=11,face="bold"),
#           strip.text.y=element_text(size=11,face="bold"),
#           legend.position="bottom",
#           legend.title=element_text(size=10,face="bold"),
#           legend.text=element_text(size=9),
#           legend.key=element_blank()) +
#       geom_text(data=plotData,mapping=aes(x=x,y=y,
#           label=rownames(plotData)),size=4,hjust=-0.15,vjust=0)
#   return(mds)
#    #return(fil)
#}

diagplotPairs <- function(x,output="x11",altNames=NULL,path=NULL,...) {    
    x <- as.matrix(x)
    x <- nat2log(x)
    n <- ncol(x)
    if (!is.null(colnames(x)))
        nams <- colnames(x)
    else
        nams <- paste("Sample_",seq_len(ncol(x)),sep="")

    if (!is.null(path))
        fil <- file.path(path,paste("correlation_pairs",output,sep="."))
    else
        fil <- paste("correlation_pairs",output,sep=".")
    
    if (output != "json") {
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=12,height=12)
        else {
            if (ncol(x)<=5)
                graphicsOpen(output,fil,width=800,height=800,res=100)
            else
                graphicsOpen(output,fil,width=1024,height=1024,res=150)
        }
        
        # Setup the grid
        par(mfrow=c(n,n),mar=c(1,1,1,1),oma=c(1,1,0,0),mgp=c(2,0.5,0),
            cex.axis=0.6,cex.lab=0.6)

        # Plot
        for (i in seq_len(n)) {
            for (j in seq_len(n)) {
                if (i==j) { # Diagonal
                    plot(0:10,0:10,type="n",xaxt="n",yaxt="n",xlab="",ylab="")
                    text(c(3,5,3),c(9.5,5,1),c("X-Y plots",nams[i],"M-D plots"),
                        cex=c(0.8,1,0.8))
                    arrows(6,9.5,9.5,9.5,angle=20,length=0.1,lwd=0.8,cex=0.8)
                    arrows(0.2,3.2,0.2,0.2,angle=20,length=0.1,lwd=0.8,cex=0.8)
                }
                else if (i<j) { # XY plot
                    plot(x[,i],x[,j],pch=20,col="blue",cex=0.4,xlab=nams[i],
                        ylab=nams[j],...)
                    lines(lowess(x[,i],x[,j]),col="red")
                    cc <- paste("cor:",formatC(cor(x[,i],x[,j]),digits=3))
                    text(3,max(x[,j]-1),labels=cc,cex=0.7,)
                    #grid()
                }
                else if (i>j) { # MD plot
                    plot((x[,i]+x[,j])/2,x[,j]-x[,i],pch=20,col="blue",
                        cex=0.4,...)
                    lines(lowess((x[,i]+x[,j])/2,x[,j]-x[,i]),col="red")
                    #grid()
                }
            }
        }
        
        if (output != "x11") {
            graphicsClose(output)
            return(fil)
        }
    }
    else {
        jsonList <- vector("list",2)
        names(jsonList) <- c("xy","md")
        jsonList$xy <- jsonList$md <- vector("list",n*(n-1)/2)
        
        nams <- colnames(x)
        plotNames <- character(n*(n-1)/2)
        counter <- 0
        for (i in seq_len((ncol(x)-1))) {
            for (j in (i+1):ncol(x)) {
                counter <- counter + 1
                plotNames[counter] <- paste(nams[i],nams[j],sep="_vs_")
                
                #disp("  creating ",plotNames[counter])
                
                # XY
                obj <- list(
                    x=x[,i],
                    y=x[,j],
                    plot=NULL,
                    samples=c(nams[i],nams[j]),
                    ylim=NULL,
                    xlim=NULL,
                    status=NULL,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=altNames,
                    user=list(counts=NULL,covar=NULL,covarname="X-Y")
                )
                #jsonList$xy[[counter]] <- scatterToJSON(obj,out="list")
                jsonList$xy[[counter]] <- scatterToJSON(obj)
                
                # MD
                obj <- list(
                    x=(x[,i]+x[,j])/2,
                    y=x[,j]-x[,i],
                    plot=NULL,
                    samples=c(nams[i],nams[j]),
                    ylim=NULL,
                    xlim=NULL,
                    status=NULL,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=altNames,
                    user=list(counts=NULL,covar=NULL,
                        covarname="Mean-Difference")
                )
                #jsonList$md[[counter]] <- scatterToJSON(obj,out="list")
                jsonList$md[[counter]] <- scatterToJSON(obj)
            }
        }
        names(jsonList$xy) <- names(jsonList$md) <- plotNames
        return(jsonList)
    }
}

diagplotCor <- function(mat,type=c("heatmap","correlogram"),output="x11",
    path=NULL,...) {
    x <- as.matrix(mat)
    type <- tolower(type[1])
    checkTextArgs("type",type,c("heatmap","correlogram"))
    if (!requireNamespace("corrplot") && type=="correlogram")
        stop("R package corrplot is required!")
    corMat <- cor(mat)
    if (!is.null(colnames(mat)))
        colnames(corMat) <- colnames(mat)
    if (!is.null(path))
        fil <- file.path(path,paste("correlation_",type,".",output,sep=""))
    else
        fil <- paste("correlation_",type,".",output,sep="")
    if (output %in% c("pdf","ps","x11"))
        graphicsOpen(output,fil,width=7,height=7)
    else
        graphicsOpen(output,fil,width=640,height=640,res=100)
    if (type=="correlogram")
        corrplot(corMat,method="ellipse",order="hclust",...)
    else if (type=="heatmap") {
        n <- dim(corMat)[1]
        labs <- matrix(NA,n,n)
        for (i in seq_len(n))
            for (j in seq_len(n))
                labs[i,j] <- sprintf("%.2f",corMat[i,j])
        if (n <= 5)
            notecex <- 1.2
        else if (n > 5 & n < 10)
            notecex <- 0.9
        else
            notecex <- 0.7
        heatmap.2(corMat,col=colorRampPalette(c("yellow","grey","blue")),
            revC=TRUE,trace="none",symm=TRUE,Colv=TRUE,cellnote=labs,
            keysize=1,density.info="density",notecex=notecex,cexCol=0.9,
            cexRow=0.9,font.lab=2)
    }
    graphicsClose(output)
    return(fil)
}

diagplotEdaseq <- function(x,sampleList,covar=NULL,isNorm=FALSE,
    whichPlot=c("meanvar","meandiff","gcbias","lengthbias"),output="x11",
    altNames=NULL,path=NULL,...) {
    if (is.null(path)) path <- getwd()
    checkTextArgs("whichPlot",whichPlot,c("meanvar","meandiff","gcbias",
        "lengthbias"),multiarg=TRUE)
    if (is.null(covar) && whichPlot %in% c("gcbias","lengthbias"))
        stopwrap("\"covar\" argument is required when \"whichPlot\" is ",
            "\"gcbias\ or \"lengthbias\"!")
    if (isNorm)
        status <- "normalized"
    else
        status <- "raw"
    if (whichPlot=="gcbias" && !is.null(covar)) {
        if (max(covar) > 1) # 0-100 scale
            covar <- covar/100
    }
    if (is.null(covar)) 
        covar <- rep(NA,nrow(x))
    #if (!is.null(names(covar)))
    #   x <- x[names(covar),,drop=FALSE]
    s <- newSeqExpressionSet(as.matrix(x),phenoData=AnnotatedDataFrame(
        data.frame(conditions=factor(asClassVector(sampleList)),
        row.names=colnames(x))),featureData=AnnotatedDataFrame(data.frame(
        gc=covar,length=covar,row.names=rownames(x))))
    switch(whichPlot,
        meandiff = {
            if (output!="json") {
                fil <- vector("list",length(sampleList))
                names(fil) <- names(sampleList)
                for (n in names(sampleList)) {
                    if (length(sampleList[[n]])==1) {
                        warnwrap("Cannot create a mean-difference plot with ",
                            "one sample per condition! Skipping...")
                        next
                    }
                    pairMatrix <- combn(seq_len(length(sampleList[[n]])),2)
                    fil[[n]] <- vector("list",ncol(pairMatrix))
                    for (i in seq_len(ncol(pairMatrix))) {
                        s1 <- sampleList[[n]][pairMatrix[1,i]]
                        s2 <- sampleList[[n]][pairMatrix[2,i]]
                        fil[[n]][[i]] <- file.path(path,paste(whichPlot,"_",
                            status,"_",n,"_",s1,"_",s2,".",output,sep=""))
                        names(fil[[n]][i]) <- paste(s1,"vs",s2,sep="_")
                        graphicsOpen(output,fil[[n]][[i]])
                        MDPlot(s,y=pairMatrix[,i],main=paste("MD plot for ",n,
                            " ",status," samples ",s1," and ",s2,sep=""),
                            cex.main=0.9)
                        graphicsClose(output)
                    }
                }
                return(fil)
            }
            else {
                json <- vector("list",length(sampleList))
                names(json) <- names(sampleList)
                for (n in names(sampleList)) {
                    if (length(sampleList[[n]])==1) {
                        warnwrap("Cannot create a mean-difference plot with ",
                            "one sample per condition! Skipping...")
                        next
                    }
                    pairMatrix <- combn(seq_len(length(sampleList[[n]])),2)
                    json[[n]] <- vector("list",ncol(pairMatrix))
                    for (i in seq_len(ncol(pairMatrix))) {
                        s1 <- sampleList[[n]][pairMatrix[1,i]]
                        s2 <- sampleList[[n]][pairMatrix[2,i]]
                        xx <- rowMeans(log(x[,pairMatrix[,i]] + 0.1))
                        yy <- log(x[,pairMatrix[2,i]] + 0.1) - 
                            log(x[,pairMatrix[1,i]] + 0.1)
                        obj <- list(
                            x=xx,
                            y=yy,
                            plot=NULL,
                            samples=c(s1,s2),
                            ylim=NULL,
                            xlim=NULL,
                            status=status,
                            pcut=NULL,
                            fcut=NULL,
                            altnames=altNames,
                            user=list(counts=NULL,covar=NULL,
                                covarname="Mean-Difference")
                        )
                        #json[[n]][[i]] <- scatterToJSON(obj,out="list")
                        json[[n]][[i]] <- scatterToJSON(obj)
                        names(json[[n]])[i] <- paste(s1,"_vs_",s2,sep="")
                    }
                }
                return(json)
            }
        },
        meanvar = {
            if (output!="json") {
                fil <- file.path(path,paste(whichPlot,"_",status,".",output,
                    sep=""))
                graphicsOpen(output,fil)
                suppressWarnings(meanVarPlot(s,main=paste("MV plot for",status,
                    "samples"),cex.main=0.9))
                graphicsClose(output)
                return(fil)
            }
            else {
                m <- apply(x,1,mean)
                v <- apply(x,1,var)
                mm <- m[m<=quantile(m,probs=0.9)]
                vv <- v[m<=quantile(m,probs=0.9)]
                obj <- list(
                    x=mm,
                    y=vv,
                    plot=NULL,
                    samples=NULL,
                    ylim=NULL,
                    xlim=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=altNames,
                    user=list(counts=NULL,covar=NULL,covarname="Mean-Variance")
                )
                json <- scatterToJSON(obj)
                return(json)
            }
        },
        gcbias = {
            if (!output=="json") {
                fil <- file.path(path,paste(whichPlot,"_",status,".",output,
                    sep=""))
                graphicsOpen(output,fil)
                biasPlot(s,"gc",xlim=c(0.1,0.9),log=TRUE,ylim=c(0,15),
                    main=paste("Expression - GC content ",status,sep=""))
                grid()
                graphicsClose(output)
                return(fil)
            }
            else {
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=sampleList,
                    ylim=NULL,
                    xlim=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=NULL,
                    user=list(counts=x,covar=covar,covarname="GC content")
                )
                json <- biasPlotToJSON(obj)
                #fil <- file.path(path,paste(whichPlot,"_",status,".json",
                #    sep=""))
                #disp("Writing ",fil)
                #write(json,fil)
                return(json)
            }
        },
        lengthbias = {
            if (output!="json") {
                fil <- file.path(path,paste(whichPlot,"_",status,".",output,
                    sep=""))
                graphicsOpen(output,fil)
                biasPlot(s,"length",log=TRUE,ylim=c(0,10),
                    main=paste("Expression - Gene length ",status,sep=""))
                grid()
                graphicsClose(output)
                return(fil)
            }
            else {
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=sampleList,
                    ylim=NULL,
                    xlim=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=NULL,
                    user=list(counts=x,covar=covar,
                        covarname="Gene/transcript length")
                )
                json <- biasPlotToJSON(obj)
                #fil <- file.path(path,paste(whichPlot,"_",status,".json",
                #    sep=""))
                #disp("Writing ",fil)
                #write(json,fil)
                return(json)
            }
        }
    )
}

diagplotNoiseq <- function(x,sampleList,covars,whichPlot=c("biodetection",
    "countsbio","saturation","rnacomp","readnoise","biodist"),output="x11",
    biodistOpts=list(p=NULL,pcut=NULL,name=NULL),path=NULL,isNorm=FALSE,
    ...) {
    if (is.null(path)) path <- getwd()
    # covars is a list of gc-content, factors, length, biotype, chromosomes, 
    # factors, basically copy of the noiseq object
    whichPlot <- tolower(whichPlot[1])
    checkTextArgs("whichPlot",whichPlot,c("biodetection","countsbio",
        "saturation","readnoise","rnacomp","biodist"),multiarg=FALSE)
    if (missing(covars))
        stopwrap("\"covars\" argument is required with NOISeq specific plots!")
    else {
        if (is.null(covars$gc))
            covars$gc <- rep(0.5,nrow(x))
        covars$biotype <- as.character(covars$biotype)
        names(covars$length) <- names(covars$gc) <-
            rownames(covars$chromosome) <- names(covars$biotype) <-
            rownames(x)
    }
    if (whichPlot=="biodist") {
        if (is.null(biodistOpts$p))
            stopwrap("A p-value must be provided for the \"biodist\" plot!")
        if (is.null(biodistOpts$pcut) || is.na(biodistOpts$pcut)) 
            biodistOpts$pcut=0.05
    }
    if (isNorm)
        status<- "normalized"
    else
        status<- "raw"
    
    # All of these plots are NOISeq specific so we need a local NOISeq object
    if (any(is.na(unique(covars$biotype))))
        covars$biotype=NULL # Otherwise, it will probably crash
    #localObj <- NOISeq::readData(
    localObj <- readData(
        data=x,
        length=covars$geneLength,
        gc=covars$gcContent,
        chromosome=covars$chromosome,
        #factors=data.frame(class=covars$factors),
        factors=covars$factors,
        biotype=covars$biotype
    )
    switch(whichPlot,
        biodetection = {
            #diagplotData <- NOISeq::dat(localObj,type=whichPlot)
            diagplotData <- dat(localObj,type=whichPlot)
            samples <- unlist(sampleList)
            if (output!="json") {
                fil <- character(length(samples))
                names(fil) <- samples
                for (i in seq_len(length(samples))) {
                    fil[samples[i]] <- file.path(path,paste(whichPlot,"_",
                        samples[i],".",output,sep=""))
                    if (output %in% c("pdf","ps","x11"))
                        graphicsOpen(output,fil[samples[i]],width=9,height=7)
                    else
                        graphicsOpen(output,fil[samples[i]],width=1024,
                            height=768)
                    explo.plot(diagplotData,samples=i)
                    graphicsClose(output)
                }
                return(fil)
            }
            else {
                #diagplotDataSave = NOISeq::dat2save(diagplotData)
                diagplotDataSave = dat2save(diagplotData)
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=sampleList,
                    ylims=NULL,
                    xlims=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=covars$gene_name,
                    user=list(plotdata=diagplotDataSave,covars=covars)
                )
                #json <- bioDetectionToJSON(obj,out="list")
                json <- bioDetectionToJSON(obj)
                #names(json) <- samples
                #fil <- character(length(samples))
                #names(fil) <- samples
                #for (i in seq_len(length(samples))) {
                #    fil[samples[i]] <- file.path(path,
                #        paste(whichPlot,"_",samples[i],".json",sep=""))
                #    disp("Writing ",fil[samples[i]])
                #    write(json[[i]],fil[samples[i]])
                #}
                return(json)
            }
        },
        countsbio = {
            samples <- unlist(sampleList)
            if (output!="json") {
                #diagplotData <- NOISeq::dat(localObj,type=whichPlot,
                #    factor=NULL)
                diagplotData <- dat(localObj,type=whichPlot,factor=NULL)
                fil <- character(length(samples))
                names(fil) <- samples
                for (i in seq_len(length(samples))) {
                    fil[samples[i]] <- file.path(path,paste(whichPlot,"_",
                        samples[i],".",output,sep=""))
                    if (output %in% c("pdf","ps","x11"))
                        graphicsOpen(output,fil[samples[i]],width=9,height=7)
                    else
                        graphicsOpen(output,fil[samples[i]],width=1024,
                            height=768)
                    explo.plot(diagplotData,samples=i,plottype="boxplot")
                    graphicsClose(output)
                }
                return(fil)
            }
            else {
                colnames(x) <- unlist(sampleList)
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=sampleList,
                    ylims=NULL,
                    xlims=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=covars$gene_name,
                    user=list(counts=nat2log(x),covars=covars)
                )
                # Write JSON by sample
                #fil <- vector("list",2)
                #names(fil) <- c("sample","biotype")
                #fil[["sample"]] <- character(length(samples))
                #names(fil[["sample"]]) <- samples
                jsonList <- vector("list",2)
                names(jsonList) <- c("sample","biotype")
                jsonList[["sample"]] <- vector("list",length(samples))
                names(jsonList[["sample"]]) <- samples
                bts <- unique(as.character(obj$user$covars$biotype))
                #fil[["biotype"]] <- character(length(bts))
                #names(fil[["biotype"]]) <- bts
                jsonList[["biotype"]] <- vector("list",length(bts))
                #json <- countsBioToJSON(obj,by="sample")
                jsonList[["sample"]] <- 
                    #countsBioToJSON(obj,by="sample",out="list")
                    countsBioToJSON(obj,by="sample")
                #for (i in seq_len(length(samples))) {
                #    fil[["sample"]][samples[i]] <- file.path(path,
                #        paste(whichPlot,"_",samples[i],".json",sep=""))
                #    disp("Writing ",fil[["sample"]][samples[i]])
                #    write(json[[i]],fil[["sample"]][samples[i]])
                #}
                #json <- countsBioToJSON(obj,by="biotype")
                jsonList[["biotype"]] <- 
                    #countsBioToJSON(obj,by="biotype",out="list")
                    countsBioToJSON(obj,by="biotype")
                names(jsonList[["biotype"]]) <- bts
                #names(json) <- samples
                #for (i in seq_len(length(bts))) {
                #    fil[["biotype"]][bts[i]] <- file.path(path,
                #        paste(whichPlot,"_",bts[i],".json",sep=""))
                #    disp("Writing ",fil[["biotype"]][bts[i]])
                #    write(json[[i]],fil[["biotype"]][bts[i]])
                #}
                return(jsonList)
            }
        },
        saturation = {
            # For 10 saturation points
            #diagplotData <- NOISeq::dat(localObj,k=0,ndepth=9,type=whichPlot)
            diagplotData <- dat(localObj,k=0,ndepth=9,type=whichPlot)
            #d2s <- NOISeq::dat2save(diagplotData)
            d2s <- dat2save(diagplotData)
            if (output != "json") {
                fil <- diagplotNoiseqSaturation(d2s,output,covars$biotype,
                    path=path)
                return(fil)
            }
            else {
                samples <- unlist(sampleList)
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=sampleList,
                    ylims=NULL,
                    xlims=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=covars$gene_name,
                    user=list(plotdata=d2s)
                )
                # Write JSON by sample
                #fil <- vector("list",2)
                #names(fil) <- c("sample","biotype")
                #fil[["sample"]] <- character(length(samples))
                #names(fil[["sample"]]) <- samples
                jsonList <- vector("list",2)
                names(jsonList) <- c("sample","biotype")
                #json <- bioSaturationToJSON(obj,by="sample")
                jsonList[["sample"]] <- 
                    #bioSaturationToJSON(obj,by="sample",out="list")
                    bioSaturationToJSON(obj,by="sample")
                names(jsonList[["sample"]]) <- samples
                #for (i in seq_len(length(samples))) {
                #    fil[["sample"]][samples[i]] <- file.path(path,
                #        paste(whichPlot,"_",samples[i],".json",sep=""))
                #    disp("Writing ",fil[["sample"]][samples[i]])
                #    write(json[[i]],fil[["sample"]][samples[i]])
                #}
                #json <- bioSaturationToJSON(obj,by="biotype")
                jsonList[["biotype"]] <- 
                    #bioSaturationToJSON(obj,by="biotype",out="list")
                    bioSaturationToJSON(obj,by="biotype")
                #names(jsonList[["sample"]]) <- samples
                #fil[["biotype"]] <- character(length(json))
                #names(fil[["biotype"]]) <- names(json)
                #for (n in names(json)) {
                #    fil[["biotype"]][n] <- file.path(path,
                #        paste(whichPlot,"_",n,".json",sep=""))
                #    disp("Writing ",fil[["biotype"]][n])
                #    write(json[[n]],fil[["biotype"]][n])
                #}
                return(jsonList)
            }
        },
        rnacomp = {
            if (ncol(localObj)<3) {
                warnwrap("RNA composition plot cannot be created with less ",
                    "than 3 samples! Skipping...")
                return(NULL)
            }
            if (ncol(localObj)>12) {
                warnwrap("RNA composition plot cannot be created with more ",
                    "than 12 samples! Skipping...")
                return(NULL)
            }
            #diagplotData <- NOISeq::dat(localObj,type="cd")
            diagplotData <- dat(localObj,type="cd")
            if (output!="json") {
                fil <- file.path(path,paste(whichPlot,"_",status,".",output,
                    sep=""))
                graphicsOpen(output,fil)
                explo.plot(diagplotData)
                grid()
                graphicsClose(output)
            }
            else {
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=sampleList,
                    ylims=NULL,
                    xlims=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=covars$gene_name,
                    user=list(plotdata=diagplotData@dat)
                )
                json <- rnacompToJSON(obj)
                return(json)
            }
        },
        readnoise = {
            D <- cddat(localObj)
            if (output!="json") {
                fil <- file.path(path,paste(whichPlot,".",output,sep=""))
                graphicsOpen(output,fil)
                cdplot(D,main="RNA-Seq reads noise")
                grid()
                graphicsClose(output)
                return(fil)
            }
            else {
                colnames(D$data2plot)[2:ncol(D$data2plot)] <- 
                    unlist(sampleList)
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=sampleList,
                    xlim=NULL,
                    ylim=NULL,
                    status=NULL,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=NULL,
                    user=D$data2plot
                )
                json <- readNoiseToJSON(obj)
                #fil <- file.path(path,paste(whichPlot,".json",sep=""))
                #disp("Writing ",fil)
                #write(json,fil)
                return(json)
            }
        },
        biodist = { # We have to fake a noiseq object
            p <- biodistOpts$p
            if (is.matrix(p)) p <- p[,1]
            dummy <- new("Output",
                comparison=c("Dummy1","Dummy2"),
                factor=c("class"),
                k=1,
                lc=1,
                method="n",
                replicates="biological",
                results=list(
                    data.frame(
                        Dummy1=rep(1,length(p)),
                        Dummy2=rep(1,length(p)),
                        M=rep(1,length(p)),
                        D=rep(1,length(p)),
                        prob=as.numeric(p),
                        ranking=rep(1,length(p)),
                        Length=rep(1,length(p)),
                        GC=rep(1,length(p)),
                        Chrom=as.character(covars$chromosome[,1]),
                        GeneStart=covars$chromosome[,2],
                        GeneEnd=covars$chromosome[,3],
                        Biotype=covars$biotype
                    )
                ),
                nss=5,
                pnr=0.2,
                v=0.02
            )
            if (output!="json") {
                if (!is.null(biodistOpts$name))
                    fil <- file.path(path,paste(whichPlot,"_",biodistOpts$name,
                        ".",output,sep=""))
                else
                    fil <- file.path(path,paste(whichPlot,".",output,sep=""))
                if (output %in% c("pdf","ps","x11"))
                    graphicsOpen(output,fil,width=10,height=6)
                else
                    graphicsOpen(output,fil,width=1024,height=640)
                tryCatch( # Many times, there is a problem with this function
                    DE.plot(dummy,chromosomes=NULL,q=biodistOpts$pcut,
                        graphic="distr"),
                    error=function(e) {
                        disp("      Known problem with NOISeq and external ",
                            "p-values  detected! Trying to make a plot with ",
                            "alternative p-values  (median of p-value ",
                            "distribution)...")
                        fil="error"
                        tryCatch(
                            DE.plot(dummy,chromosomes=NULL,
                                q=quantile(biodistOpts$p,0.5),
                                graphic="distr"),
                            error=function(e) {
                                disp("      Cannot create DEG biotype plot! ",
                                    "This is not related to any problem with ",
                                    "the results. Excluding...")
                                fil="error"
                            },
                            finally=""
                        )
                    },
                    finally=""
                )
                graphicsClose(output)
                return(fil)
            }
            else {
                dataList <- deplot(dummy,chromosomes=NULL,q=biodistOpts$pcut)
                
                colnames(x) <- unlist(sampleList)
                obj <- list(
                    x=NULL,
                    y=NULL,
                    plot=NULL,
                    samples=NULL,
                    ylims=NULL,
                    xlims=NULL,
                    status=status,
                    pcut=NULL,
                    fcut=NULL,
                    altnames=covars$gene_name,
                    user=list(plotdata=dataList)
                )
                
                jsonList <- vector("list",2)
                names(jsonList) <- c("chromosome","biotype")
                jsonList[["chromosome"]] <- 
                    #biodistToJSON(obj,by="chromosome",out="list")
                    biodistToJSON(obj,by="chromosome")
                jsonList[["biotype"]] <- 
                    #biodistToJSON(obj,by="biotype",out="list")
                    biodistToJSON(obj,by="biotype")
                return(jsonList)
            }
        }
    )
    
}

diagplotNoiseqSaturation <- function(x,o,tb,path=NULL) {
    if (is.null(path)) path <- getwd()
    if (length(unique(tb))==1) {
        warnwrap("Saturation plot cannot be created with only one biotype! ",
            "Skipping...")
        return(NULL)
    }
    totalBiotypes <- table(tb)
    theBiotypes <- names(tb)
    biotypes <- colnames(x[[1]][,2:ncol(x[[1]])])
    colspace <- c("red3","green4","blue2","orange3","burlywood","lightpink4",
        "gold","skyblue","red2","green2","firebrick3","orange4","yellow4",
        "skyblue3","tan4","gray40","brown2","darkgoldenrod","cyan3","coral2",
        "cadetblue","bisque3","blueviolet","chocolate3","darkkhaki",
        "dodgerblue")
    pchspace <- c(rep(c(15,16,17,18),6),15)

    # Plot all biotypes per sample
    fSample <- character(length(names(x)))
    names(fSample) <- names(x)
    for (n in names(x)) {
        fSample[n] <- file.path(path,paste("saturation_",n,".",o,sep=""))
        if (o %in% c("pdf","ps","x11"))
            graphicsOpen(o,fSample[n],width=10,height=7)
        else
            graphicsOpen(o,fSample[n],width=1024,height=800)
        y <- x[[n]]
        sep <- match(c("global","protein_coding"),colnames(y))
        yab <- cbind(y[,"depth"],y[,sep])
        ynab <- y[,-sep]
        colnames(yab)[1] <- colnames(ynab)[1] <- "depth"
        xlim <- range(y[,"depth"])
        ylimAb <- range(yab[,2:ncol(yab)])
        ylimNab <- range(ynab[,2:ncol(ynab)])
        par(cex.axis=0.9,cex.main=1,cex.lab=0.9,font.lab=2,font.axis=2,pty="m",
            lty=2,lwd=1.5,mfrow=c(1,2))
        plot.new()
        plot.window(xlim,ylimNab)
        axis(1,at=pretty(xlim,10),labels=as.character(pretty(xlim,10)/1e+6))
        axis(2,at=pretty(ylimNab,10))
        title(main="Non abundant biotype detection saturation",
            xlab="Depth in millions of reads",ylab="Detected features")
        co <- 0
        for (b in biotypes) {
            co <- co + 1
            if (b=="global" || b=="protein_coding") {
                # Silently do nothing
            }
            else {
                lines(ynab[,"depth"],ynab[,b],col=colspace[co])
                points(ynab[,"depth"],ynab[,b],pch=pchspace[co],
                    col=colspace[co],cex=1)
            }
        }
        grid()
        graphics::legend(
            x="topleft",legend=colnames(ynab)[2:ncol(ynab)],xjust=1,yjust=0,
            box.lty=0,x.intersp=0.5,cex=0.6,text.font=2,
            col=colspace[seq_len(ncol(ynab)-1)],
                pch=pchspace[seq_len(ncol(ynab)-1)]
        )
        plot.new()
        plot.window(xlim,ylimAb)
        axis(1,at=pretty(xlim,10),labels=as.character(pretty(xlim,10)/1e+6))
        axis(2,at=pretty(ylimAb,10))
        title(main="Abundant biotype detection saturation",
            xlab="Depth in millions of reads",ylab="Detected features")
        co <- 0
        for (b in c("global","protein_coding")) {
            co <- co + 1
            lines(yab[,"depth"],yab[,b],col=colspace[co])
            points(yab[,"depth"],yab[,b],pch=16,col=colspace[co],cex=1.2)
        }
        grid()
        graphics::legend(
            x="topleft",legend=c("global","protein_coding"),xjust=1,yjust=0,
            box.lty=0,lty=2,x.intersp=0.5,cex=0.7,text.font=2,
            col=colspace[c(1,2)],pch=pchspace[c(1,2)]
        )
        mtext(n,side=3,line=-1.5,outer=TRUE,font=2,cex=1.3)
        graphicsClose(o)
    }

    # Plot all samples per biotype
    g <- makeGrid(length(biotypes))
    fAll <- file.path(path,paste("biotype_saturation.",o,sep=""))
    if (o %in% c("pdf","ps"))
        graphicsOpen(o,fAll,width=14,height=14)
    else
        graphicsOpen(o,fAll,width=1600,height=1600,res=150)
    par(cex.axis=0.8,cex.main=0.9,cex.lab=0.8,pty="m",lty=2,lwd=1.5,mfrow=g,
        mar=c(3,3,1,1),oma=c(1,1,0,0),mgp=c(2,0.5,0))
    for (b in biotypes) {
        y <- depth <- vector("list",length(x))
        names(y) <- names(depth) <- names(x)
        for (n in names(x)) {
            y[[n]] <- x[[n]][,b]
            depth[[n]] <- x[[n]][,"depth"]
        }
        y <- do.call("cbind",y)
        xlim <- range(do.call("c",depth))
        ylim <- range(y)
        plot.new()
        plot.window(xlim,ylim)
        axis(1,at=pretty(xlim,5),labels=as.character(pretty(xlim,5)/1e+6),
            line=0.5)
        axis(2,at=pretty(ylim,5),line=0.5)
        title(main=b,xlab="Depth in millions of reads",
            ylab="Detected features")
        co <- 0
        for (n in colnames(y)) {
            co <- co + 1
            lines(depth[[n]],y[,n],col=colspace[co])
            points(depth[[n]],y[,n],pch=pchspace[co],col=colspace[co])
        }
        grid()
        graphics::legend(
            x="bottomright",legend=colnames(y),xjust=1,yjust=0,
            box.lty=0,x.intersp=0.5,
            col=colspace[seq_len(length(colnames(y)))],
            pch=pchspace[seq_len(length(colnames(y)))]
        )
    }
    graphicsClose(o)

    return(list(sample=fSample,biotype=fAll))
}

diagplotVolcano <- function(f,p,con=NULL,fcut=1,pcut=0.05,altNames=NULL,
    output="x11",path=NULL,...) { # output can be json here...
    if (is.null(path)) path <- getwd()
    if (is.null(con))
        con <- conn <- ""
    else {
        conn <- con
        con <- paste("for ",con)
    }
    fil <- file.path(path,paste("volcano_plot_",conn,".",output,sep=""))
    if (output!="json") {
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=8,height=10)
        else
            graphicsOpen(output,fil,width=768,height=1024,res=100)
    }
    rem <- which(is.na(p))
    if (length(rem)>0) {
        p <- p[-rem]
        f <- f[-rem]
        if (!is.null(altNames))
            altNames <- altNames[-rem]
    }
    # Fix problem with extremely low p-values, only for display purposes though
    pZero <- which(p==0)
    if (length(pZero)>0)
        p[pZero] <- runif(length(pZero),0,1e-256)
    xlim <- c(-max(abs(f)),max(abs(f)))
    ylim <- c(0,ceiling(-log10(min(p))))
    up <- which(f>=fcut & p<pcut)
    down <- which(f<=-fcut & p<pcut)
    u <- union(up,down)
    altNamesNeutral <- NULL
    if (length(u)>0) {
        ff <- f[-u]
        pp <- p[-u]
        if (!is.null(altNames))
            altNamesNeutral <- altNames[-u]
    }
    else {
        ff <- f
        pp <- p
        if (!is.null(altNames))
            altNamesNeutral <- altNames
    }
    if (output!="json") {
        par(cex.main=1.1,cex.lab=1.1,cex.axis=1.1,font.lab=2,font.axis=2,
            pty="m",lwd=1.5)
        plot.new()
        plot.window(xlim,ylim)
        axis(1,at=pretty(xlim,10),labels=as.character(pretty(xlim,10)))
        axis(2,at=pretty(ylim,10))
        title(paste(main="Volcano plot",con),
            xlab="Fold change",ylab="-log10(p-value)")
        points(ff,-log10(pp),pch=20,col="blue2",cex=0.9)
        points(f[down],-log10(p[down]),pch=20,col="green3",cex=0.9)
        points(f[up],-log10(p[up]),pch=20,col="red2",cex=0.9)
        abline(h=-log10(pcut),lty=4)
        abline(v=-fcut,lty=2)
        abline(v=fcut,lty=2)
        grid()
        graphics::legend(
            x="topleft",
            legend=c("up-regulated","down-regulated","unregulated",
                "p-value threshold","fold change threshold"),
            col=c("red2","green3","blue1","black","black"),
            pch=c(20,20,20,NA,NA),lty=c(NA,NA,NA,4,2),
            xjust=1,yjust=0,box.lty=0,x.intersp=0.5,cex=0.8,text.font=2
        )
        graphicsClose(output)
        return(fil)
    }
    else {
        obj <- list(
            x=f,
            y=p,
            plot=NULL,
            samples=NULL,
            xlim=xlim,
            ylim=ylim,
            status=NULL,
            pcut=pcut,
            fcut=fcut,
            altnames=altNames,
            user=list(up=up,down=down,unf=ff,unp=pp,ualt=altNamesNeutral,
                con=con)
        )
        #json <- volcanoToJSON(obj,out="list")
        json <- volcanoToJSON(obj)
        #fil <- file.path(path,paste("volcano_",con,".json",sep=""))
        #write(json,fil)
        return(json)
    }
}

diagplotMa <- function(m,a,p=NULL,con=NULL,fcut=1,pcut=0.05,altNames=NULL,
    output="x11",path=NULL,...) { # output can be json here...
    if (is.null(path)) path <- getwd()
    if (is.null(con))
        con <- conn <- ""
    else {
        conn <- con
        con <- paste("for ",con)
    }
    fil <- file.path(path,paste("mastat_plot_",conn,".",output,sep=""))
    if (output!="json") {
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=10,height=8)
        else
            graphicsOpen(output,fil,width=1024,height=768,res=100)
    }
    
    if (!is.null(p)) {
        rem <- which(is.na(p))
        if (length(rem)>0) {
            p <- p[-rem]
            m <- m[-rem]
            a <- a[-rem]
            if (!is.null(altNames))
                altNames <- altNames[-rem]
        }
        # Fix problem with extremely low p-values, only for display purposes 
        # though
        pZero <- which(p==0)
        if (length(pZero)>0)
            p[pZero] <- runif(length(pZero),0,1e-256)
        
        upstat <- which(m>=fcut & p<pcut)
        downstat <- which(m<=-fcut & p<pcut)
        up <- which(m>=fcut & p>=pcut)
        down <- which(m<=-fcut & p>=pcut)
        poor <- which(p<pcut & abs(m)<fcut)
        neutral <- setdiff(seq_len(length(a)),
            Reduce("union",list(upstat,downstat,up,down,poor)))
    }
    else {
        upstat <- downstat <- poor <- integer(0)
        up <- which(m>=fcut)
        down <- which(m<=-fcut)
        neutral <- setdiff(seq_len(length(a)),union(up,down))
    }
        
    ylim <- c(-max(abs(m)),max(abs(m)))
    xlim <- c(min(a),max(a))
    
    if (output!="json") {
        par(cex.main=1.1,cex.lab=1.1,cex.axis=1.1,font.lab=2,font.axis=2,
            pty="m",lwd=1.5)
        plot.new()
        plot.window(xlim,ylim)
        axis(1,at=pretty(xlim,10),labels=as.character(pretty(xlim,10)))
        axis(2,at=pretty(ylim,10))
        title(paste(main="MA plot",con),xlab="Average expression",
            ylab="Fold change")
            
        points(a[neutral],m[neutral],pch=20,col="gray50",cex=0.5)
        points(a[poor],m[poor],pch=20,col="orange",cex=0.5)
        points(a[up],m[up],pch=20,col="red3",cex=0.5)
        points(a[down],m[down],pch=20,col="green3",cex=0.5)
        if (length(upstat) > 0)
            points(a[upstat],m[upstat],pch=20,col="red",cex=0.5)
        if (length(downstat) > 0)
            points(a[downstat],m[downstat],pch=20,col="green",cex=0.5)
        abline(h=-fcut,lty=2)
        abline(h=fcut,lty=2)
        grid()
        graphics::legend(
            x="topright",
            legend=c("significantly up-regulated",
                "significantly down-regulated","up-regulated","down-regulated",
                "poorly regulated","neutral","fold change threshold"),
            col=c("red","green","red3","green3","orange","gray50","black"),
            pch=c(20,20,20,20,20,20,NA),lty=c(NA,NA,NA,NA,NA,NA,2),
            xjust=1,yjust=0,box.lty=0,x.intersp=0.5,cex=0.7,text.font=2
        )
        graphicsClose(output)
        return(fil)
    }
    else {
        obj <- list(
            x=a,
            y=m,
            plot=NULL,
            samples=NULL,
            xlim=xlim,
            ylim=ylim,
            status=NULL,
            pcut=pcut,
            fcut=fcut,
            altnames=altNames,
            user=list(p=p,con=conn)
        )
        #json <- maStatToJSON(obj,out="list")
        json <- maStatToJSON(obj)
        return(json)
    }
}

diagplotDeregulogram <- function(fmat,pmat,fcut=0.5,pcut=0.05,altNames=NULL,
    output="x11",path=NULL,...) { # output can be json here...
    if (is.null(path)) path <- getwd()
    # colnames of at least one of fmat, pmat must be given
    if (is.null(colnames(fmat)) && is.null(colnames(pmat)))
        stopwrap("At least fmat or pmat must have column names!")
    if (is.null(colnames(fmat)) && !is.null(colnames(pmat)))
        colnames(fmat) <- colnames(pmat)
    if (!is.null(colnames(fmat)) && is.null(colnames(pmat)))
        colnames(pmat) <- colnames(fmat)
    
    fil <- file.path(path,paste("deregulogram_",
        paste(colnames(fmat),collapse="__"),".",output,sep=""))
    if (output!="json") {
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=10,height=8)
        else
            graphicsOpen(output,fil,width=1024,height=768,res=100)
    }
    
    # Ensure order/subsets
    fmat <- fmat[rownames(pmat),]
    # Fix problem with extremely low p-values, only for display purposes though
    pZero <- which(pmat==0)
    if (length(pZero)>0)
        pmat[pZero] <- runif(length(pZero),0,1e-256)
        
    # Remove non-signicant always
    allbad <- which(apply(pmat,1,function(x) all(x>=pcut)))
    if (length(allbad) > 0) {
        pmat <- pmat[-allbad,]
        fmat <- fmat[-allbad,]
    }
        
    ylim <- c(-max(abs(fmat[,2])),max(abs(fmat[,2])))
    xlim <- c(-max(abs(fmat[,1])),max(abs(fmat[,1])))
    
    if (output!="json") {
        # red
        upupstat <- which(apply(fmat,1,function(x) all(x >= fcut)) &
            apply(pmat,1,function(x) all(x < pcut)))
        # green
        downdownstat <- which(apply(fmat,1,function(x) all(x <= -fcut)) &
            apply(pmat,1,function(x) all(x < pcut)))
        # red3
        upup <- which(apply(fmat,1,function(x) all(x >= fcut)) &
            apply(pmat,1,function(x) any(x >= pcut)))
        # green3
        downdown <- which(apply(fmat,1,function(x) all(x <= -fcut)) &
            apply(pmat,1,function(x) any(x >= pcut)))
        # orange
        updownstat <- which(apply(fmat,1,
            function(x) x[1] >= fcut & x[2] <= -fcut) &
                apply(pmat,1,function(x) all(x < pcut)))
        # blue
        downupstat <- which(apply(fmat,1,
            function(x) x[1] <= -fcut & x[2] >= fcut) &
                apply(pmat,1,function(x) all(x < pcut)))
        # orange3
        updown <- which(apply(fmat,1,function(x) x[1] >= fcut & x[2] <= -fcut) &
            apply(pmat,1,function(x) any(x >= pcut)))
        # blue3
        downup <- which(apply(fmat,1,function(x) x[1] <= -fcut & x[2] >= fcut) &
            apply(pmat,1,function(x) any(x >= pcut)))
        # black
        poor <- which(apply(pmat,1,function(x) all(x < pcut)) &
            apply(fmat,1,function(x) any(abs(x) < fcut)))
        # gray70
        nones <- which(apply(pmat,1,function(x) any(x >= pcut)) &
            apply(fmat,1,function(x) all(abs(x) < fcut)))
        # gray40
        neutral <- setdiff(seq_len(nrow(fmat)),
            Reduce("union",list(upupstat,downdownstat,upup,downdown,updownstat,
                downupstat,updown,downup,poor,nones)))
        
        par(cex.main=1.1,cex.lab=1.1,cex.axis=1.1,font.lab=2,font.axis=2,
            pty="m",lwd=1.5)
        plot.new()
        plot.window(xlim,ylim)
        axis(1,at=pretty(xlim,10),labels=as.character(pretty(xlim,10)))
        axis(2,at=pretty(ylim,10))
        title(paste(main="Deregulogram",colnames(fmat)[1],"and",
            colnames(fmat)[2]),xlab=paste("Fold change",colnames(fmat)[1]),
            ylab=paste("Fold change",colnames(fmat)[2]))
        
        if (length(neutral) > 0)
            points(fmat[neutral,1],fmat[neutral,2],pch=20,col="gray40",cex=0.5)
        if (length(nones) > 0)
            points(fmat[nones,1],fmat[nones,2],pch=20,col="gray70",cex=0.5)
        if (length(poor) > 0)
            points(fmat[poor,1],fmat[poor,2],pch=20,col="black",cex=0.5)
        if (length(downup) > 0)
            points(fmat[downup,1],fmat[downup,2],pch=20,col="blue3",cex=0.5)
        if (length(updown) > 0)
            points(fmat[updown,1],fmat[updown,2],pch=20,col="orange3",cex=0.5)
        if (length(downupstat) > 0)
            points(fmat[downupstat,1],fmat[downupstat,2],pch=20,col="blue",
                cex=0.5)
        if (length(updownstat) > 0)
            points(fmat[updownstat,1],fmat[updownstat,2],pch=20,col="orange",
                cex=0.5)
        if (length(downdown) > 0)
            points(fmat[downdown,1],fmat[downdown,2],pch=20,col="green3",
                cex=0.5)
        if (length(upup) > 0)
            points(fmat[upup,1],fmat[upup,2],pch=20,col="red3",cex=0.5)
        if (length(downdownstat) > 0)
            points(fmat[downdownstat,1],fmat[downdownstat,2],pch=20,col="green",
                cex=0.5)
        if (length(upupstat) > 0)
            points(fmat[upupstat,1],fmat[upupstat,2],pch=20,col="red",cex=0.5)
        
        abline(h=-fcut,lty=2)
        abline(h=fcut,lty=2)
        abline(v=-fcut,lty=2)
        abline(v=fcut,lty=2)
        
        grid()
        graphics::legend(
            x="topleft",
            legend=c("significantly both up-regulated",
                "significantly both down-regulated","both up-regulated",
                "both down-regulated","significantly up-down-regulated",
                "significantly down-up-regulated","up-down-regulated",
                "down-up-regulated","poorly regulated","no regulated","neutral",
                "fold change threshold"),
            col=c("red","green","red3","green3","orange","blue","orange3",
                "blue3","black","gray70","gray40","black"),
            pch=c(rep(20,11),NA),lty=c(rep(NA,11),2),xjust=1,yjust=0,
            box.lty=0,x.intersp=0.5,cex=0.7,text.font=2
        )
        graphicsClose(output)
        return(fil)
    }
    else {
        obj <- list(
            x=NULL,
            y=NULL,
            plot=NULL,
            samples=NULL,
            xlim=xlim,
            ylim=ylim,
            status=NULL,
            pcut=pcut,
            fcut=fcut,
            altnames=altNames,
            user=list(fmat=fmat,pmat=pmat)
        )
        json <- dereguloToJSON(obj)
        return(json)
    }
}

diagplotDeHeatmap <- function(x,scale=c("asis","zscore"),con=NULL,output="x11",
    path=NULL,...) {
    if (is.null(path)) path <- getwd()
    if (is.null(con))
        con <- conn <- ""
    else {
        conn <- con
        con <- paste("for ",con)
    }
    scale <- scale[1]
    y <- nat2log(x,2,1)
    if (scale == "zscore") {
        y <- t(scale(t(y)))
        if (any(is.infinite(y)))
            y[which(is.infinite(y))] <- min(y)
        if (any(is.na(y)))
            y[which(is.na(y))] <- 0
    }
        
    # First plot the normal image
    fil <- file.path(path,paste("de_heatmap_",conn,"_",scale,".",output,sep=""))
    if (output %in% c("pdf","ps","x11"))
        graphicsOpen(output,fil,width=10,height=10)
    else
        graphicsOpen(output,fil,width=800,height=800)
    heatmap.2(y,trace="none",col=bluered(16),labRow="",cexCol=0.9,keysize=1,
        font.lab=2,main=paste("DEG heatmap",con),cex.main=0.9)
    graphicsClose(output)
    return(fil)
}

diagplotFiltered <- function(x,y,output="x11",path=NULL,...) {
    if (output !="json") {
        if (is.null(path)) path <- getwd()
        fil <- file.path(path,paste("filtered_genes.",output,sep=""))
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=12,height=8)
        else
            graphicsOpen(output,fil,width=1200,height=800,res=100)
        chr <- table(as.character(x$chromosome))
        bt <- table(as.character(x$biotype))
        chrAll <- table(as.character(y$chromosome))
        btAll <- table(as.character(y$biotype))
        barlabChr <- as.character(chr)
        barlabBt <- as.character(bt)
        perChr <- chr/chrAll[names(chr)]
        perBt <- bt/btAll[names(bt)]
        # Some bug...
        perChr[perChr>1] <- 1
        perBt[perBt>1] <- 1
        #
        suppressWarnings(perChrLab <- paste(formatC(100*perChr,digits=1,
            format="f"),"%",sep=""))
        suppressWarnings(perBt.lab <- paste(formatC(100*perBt,digits=1,
            format="f"),"%",sep=""))

        par(mfrow=c(2,2),mar=c(1,4,2,1),oma=c(1,1,1,1))

        # Chromosomes
        barxChr <- barplot(chr,space=0.5,
            ylim=c(0,max(chr)+ceiling(max(chr)/10)),yaxt="n",xaxt="n",
            plot=FALSE)
        plot.new()
        plot.window(xlim=c(0,ceiling(max(barxChr))),
            ylim=c(0,max(chr)+ceiling(max(chr)/10)),mar=c(1,4,1,1))
        axis(2,at=pretty(0:(max(chr)+ceiling(max(chr)/10))),cex.axis=0.9,padj=1,
            font=2)
        text(x=barxChr,y=chr,label=barlabChr,cex=0.7,font=2,col="green3",
            adj=c(0.5,-1.3))
        title(main="Filtered genes per chromosome",cex.main=1.1)
        mtext(side=2,text="Number of genes",line=2,cex=0.9,font=2)
        grid()
        barplot(chr,space=0.5,ylim=c(0,max(chr)+ceiling(max(chr)/10)),
            col="blue3",border="yellow3",yaxt="n",xaxt="n",font=2,add=TRUE)

        # Biotypes
        barxBt <- barplot(bt,space=0.5,ylim=c(0,max(bt)+ceiling(max(bt)/10)),
            yaxt="n",xaxt="n",plot=FALSE)
        plot.new()
        plot.window(xlim=c(0,ceiling(max(barxBt))),
            ylim=c(0,max(bt)+ceiling(max(bt)/10)),mar=c(1,4,1,1))
        axis(2,at=pretty(0:(max(bt)+ceiling(max(bt)/10))),cex.axis=0.9,padj=1,
            font=2)
        text(x=barxBt,y=bt,label=barlabBt,cex=0.7,font=2,col="blue",
            adj=c(0.5,-1.3),xpd=TRUE)
        title(main="Filtered genes per biotype",cex.main=1.1)
        mtext(side=2,text="Number of genes",line=2,cex=0.9,font=2)
        grid()
        barplot(bt,space=0.5,ylim=c(0,max(bt)+ceiling(max(bt)/10)),col="red3",
            border="yellow3",yaxt="n",xaxt="n",font=2,add=TRUE)

        # Chromosome percentage
        barxPerChr <- barplot(perChr,space=0.5,ylim=c(0,max(perChr)),
            yaxt="n",xaxt="n",plot=FALSE)
        plot.new()
        par(mar=c(9,4,1,1))
        plot.window(xlim=c(0,max(barxPerChr)),ylim=c(0,max(perChr)))
        #axis(1,at=barxPerChr,labels=names(perChr),cex.axis=0.9,font=2,
        #    tcl=-0.3,col="lightgrey",las=2)
        axis(1,at=barxPerChr,labels=FALSE,tcl=-0.3,col="lightgrey")
        axis(2,at=seq(0,max(perChr),length.out=5),labels=formatC(seq(0,
            max(perChr),length.out=5),digits=2,format="f"),cex.axis=0.9,padj=1,
            font=2)
        text(barxPerChr,par("usr")[3]-max(perChr)/17,labels=names(perChr),
            srt=45,adj=c(1,1.1),xpd=TRUE,cex=0.9,font=2)
        text(x=barxPerChr,y=perChr,label=perChrLab,cex=0.7,font=2,
            col="green3",adj=c(0.5,-1.3),xpd=TRUE)
        mtext(side=2,text="fraction of total genes",line=2,cex=0.9,font=2)
        grid()
        barplot(perChr,space=0.5,ylim=c(0,max(perChr)),col="blue3",
            border="yellow3",yaxt="n",xaxt="n",font=2,add=TRUE)

        # Biotype percentage
        barxPerBt <- barplot(perBt,space=0.5,ylim=c(0,max(perBt)),yaxt="n",
            xaxt="n",plot=FALSE)
        plot.new()
        par(mar=c(9,4,1,1))
        plot.window(xlim=c(0,max(barxPerBt)),ylim=c(0,max(perBt)))
        #axis(1,at=barxPerBt,labels=names(perBt),cex.axis=0.9,font=2,
        #    tcl=-0.3,col="lightgrey",las=2)
        axis(1,at=barxPerBt,labels=FALSE,tcl=-0.3,col="lightgrey")
        axis(2,at=seq(0,max(perBt),length.out=5),
            labels=formatC(seq(0,max(perBt),length.out=5),digits=2,format="f"),
            cex.axis=0.9,padj=1,font=2)
        text(barxPerBt,par("usr")[3]-max(perBt)/17,
            labels=gsub("prime","'",names(perBt)),srt=45,adj=c(1,1.1),
            xpd=TRUE,cex=0.9,font=2)
        text(x=barxPerBt,y=perBt,label=perBt.lab,cex=0.7,font=2,col="blue",
            adj=c(0.5,-1.3),xpd=TRUE)
        mtext(side=2,text="fraction of total genes",line=2,cex=0.9,font=2)
        grid()
        barplot(perBt,space=0.5,ylim=c(0,max(perBt)),col="red3",
            border="yellow3",yaxt="n",xaxt="n",font=2,add=TRUE)
        
        graphicsClose(output)
        
        return(fil)
    }
    else {
        if (is(x,"GenomicRanges")) {
            x <- as.data.frame(x)
            x <- x[,c(1,2,3,6,7,5,8,9)]
            colnames(x)[1] <- "chromosome"
        }
        if (is(y,"GenomicRanges")) {
            y <- as.data.frame(y)
            y <- y[,c(1,2,3,6,7,5,8,9)]
            colnames(y)[1] <- "chromosome"
        }
        obj <- list(
            x=NULL,
            y=NULL,
            plot=NULL,
            samples=NULL,
            xlim=NULL,
            ylim=NULL,
            status=NULL,
            pcut=NULL,
            fcut=NULL,
            altnames=NULL,
            user=list(filtered=x,total=y)
        )
        #fil <- list(chromosome=NULL,biotype=NULL)
        jsonList <- vector("list",2)
        names(jsonList) <- c("chromosome","biotype")
        #json <- filteredToJSON(obj,by="chromosome")
        #fil$chromosome <- file.path(path,"filtered_genes_chromosome.json")
        #write(json,fil$chromosome)
        jsonList[["chromosome"]] <- filteredToJSON(obj,by="chromosome")
        #json <- filteredToJSON(obj,by="biotype")
        jsonList[["biotype"]] <- filteredToJSON(obj,by="biotype")
        #fil$biotype <- file.path(path,"filtered_genes_biotype.json")
        #write(json,fil$biotype)
        return(jsonList)
    }
}

diagplotVenn <- function(pmat,fcmat=NULL,pcut=0.05,fcut=0.5,
    direction=c("dereg","up","down"),nam=as.character(round(1000*runif(1))),
    output="x11",path=NULL,altNames=NULL,...) {
    checkTextArgs("direction",direction,c("dereg","up","down"))
    if (is.na(pcut) || is.null(pcut) || pcut==1)
        warnwrap("Invalid pcut argument! Using the default (0.05)")
    algs <- colnames(pmat)
    if (is.null(algs))
        stopwrap("The p-value matrices must have the colnames attribute ",
            "(names of statistical algorithms)!")
    if (!is.null(fcmat) && (is.null(colnames(fcmat)) ||
        length(intersect(colnames(pmat),colnames(fcmat)))!=length(algs)))
        stopwrap("The fold change matrices must have the colnames attribute ",
            "(names of statistical algorithms) and must be the same as in the ",
            "p-value matrices!")
    nalg <- length(algs)
    if(nalg>5) {
        warnwrap(paste("Cannot create a Venn diagram for more than 5 result ",
            "sets! ",nalg,"found, only the first 5 will be used..."))
        algs <- algs[seq_len(5)]
        nalg <- 5
    }
    
    lenalias <- c("two","three","four","five")
    aliases <- toupper(letters[seq_len(nalg)])
    names(algs) <- aliases
    genes <- rownames(pmat)
    pairs <- makeVennPairs(algs)
    morePairs <- makeMoreVennPairs(algs)
    areas <- makeVennAreas(length(algs))
    counts <- makeVennCounts(length(algs))
    # Initially populate the results and counts lists so they can be used to 
    # create the rest of the intersections
    results <- vector("list",nalg+length(pairs)+length(morePairs))
    names(results)[seq_len(nalg)] <- aliases
    names(results)[(nalg+1):length(results)] <- c(names(pairs),names(morePairs))
    if (is.null(fcmat)) {
        for (a in aliases) {
            results[[a]] <- genes[which(pmat[,algs[a]]<pcut)]
            counts[[areas[[a]]]] <- length(results[[a]])
        }
    }
    else {
        switch(direction,
            dereg = {
                for (a in aliases) {
                    results[[a]] <-
                        genes[which(pmat[,algs[a]]<pcut & abs(
                        fcmat[,algs[a]])>=fcut)]
                    counts[[areas[[a]]]] <- length(results[[a]])
                }
            },
            up = {
                for (a in aliases) {
                    results[[a]] <-
                        genes[which(pmat[,algs[a]]<pcut &
                        fcmat[,algs[a]]>=fcut)]
                    counts[[areas[[a]]]] <- length(results[[a]])
                }
            },
            down = {
                for (a in aliases) {
                    results[[a]] <-
                        genes[which(pmat[,algs[a]]<pcut &
                        fcmat[,algs[a]]<=-fcut)]
                    counts[[areas[[a]]]] <- length(results[[a]])
                }
            }
        )
    }
    
    # Now, perform the intersections
    for (p in names(pairs)) {
        a = pairs[[p]][1]
        b = pairs[[p]][2]
        results[[p]] <- intersect(results[[a]],results[[b]])
        counts[[areas[[p]]]] <- length(results[[p]])
    }
    # and the setdiffs
    for (mp in names(morePairs)) {
        a = morePairs[[mp]][1]
        b = morePairs[[mp]][2]
        results[[mp]] <- setdiff(results[[a]],results[[b]])
        #counts[[areas[[mp]]]] <- length(results[[mp]])
    }
    
    # And now, the Venn diagrams must be constructed
    colorScheme <- makeVennColorscheme(length(algs))
    fil <- file.path(path,paste("venn_plot_",nam,".",output,sep=""))
    if (output %in% c("pdf","ps","x11"))
        graphicsOpen(output,fil,width=8,height=8)
    else
        graphicsOpen(output,fil,width=800,height=800,res=100)
    switch(lenalias[length(algs)-1],
        two = {
            v <- draw.pairwise.venn(
                area1=counts$area1,
                area2=counts$area2,
                cross.area=counts$cross.area,
                category=paste(algs," (",aliases,")",sep=""),
                lty="blank",
                fill=colorScheme$fill,
                cex=1.5,
                cat.cex=1.3,
                cat.pos=c(0,0),
                cat.col=colorScheme$font,
                #cat.dist=0.07,
                cat.fontfamily=rep("Bookman",2)
            )
        },
        three = {
            #overrideTriple <<- TRUE
            v <- draw.triple.venn(
                area1=counts$area1,
                area2=counts$area2,
                area3=counts$area3,
                n12=counts$n12,
                n13=counts$n13,
                n23=counts$n23,
                n123=counts$n123,
                category=paste(algs," (",aliases,")",sep=""),
                lty="blank",
                fill=colorScheme$fill,
                cex=1.5,
                cat.cex=1.3,
                #cat.pos=c(0,0,180),
                cat.col=colorScheme$font,
                #cat.dist=0.07,
                cat.fontfamily=rep("Bookman",3)
            )
        },
        four = {
            v <- draw.quad.venn(
                area1=counts$area1,
                area2=counts$area2,
                area3=counts$area3,
                area4=counts$area4,
                n12=counts$n12,
                n13=counts$n13,
                n14=counts$n14,
                n23=counts$n23,
                n24=counts$n24,
                n34=counts$n34,
                n123=counts$n123,
                n124=counts$n124,
                n134=counts$n134,
                n234=counts$n234,
                n1234=counts$n1234,
                category=paste(algs," (",aliases,")",sep=""),
                lty="blank",
                fill=colorScheme$fill,
                cex=1.5,
                cat.cex=1.3,
                c(0.1,0.1,0.05,0.05),
                cat.col=colorScheme$font,
                cat.fontfamily=rep("Bookman",4)
            )
        },
        five = {
            v <- draw.quintuple.venn(
                area1=counts$area1,
                area2=counts$area2,
                area3=counts$area3,
                area4=counts$area4,
                area5=counts$area5,
                n12=counts$n12,
                n13=counts$n13,
                n14=counts$n14,
                n15=counts$n15,
                n23=counts$n23,
                n24=counts$n24,
                n25=counts$n25,
                n34=counts$n34,
                n35=counts$n35,
                n45=counts$n45,
                n123=counts$n123,
                n124=counts$n124,
                n125=counts$n125,
                n134=counts$n134,
                n135=counts$n135,
                n145=counts$n145,
                n234=counts$n234,
                n235=counts$n235,
                n245=counts$n245,
                n345=counts$n345,
                n1234=counts$n1234,
                n1235=counts$n1235,
                n1245=counts$n1245,
                n1345=counts$n1345,
                n2345=counts$n2345,
                n12345=counts$n12345,
                category=paste(algs," (",aliases,")",sep=""),
                lty="blank",
                fill=colorScheme$fill,
                cex=1.5,
                cat.cex=1.3,
                cat.dist=0.1,
                cat.col=colorScheme$font,
                cat.fontfamily=rep("Bookman",5)
            )
        }
    )
    graphicsClose(output)

    # Now do something with the results
    if (!is.null(path)) {
        resultsEx <- vector("list",length(results))
        names(resultsEx) <- names(results)
        if (!is.null(altNames)) {
            for (n in names(results))
                resultsEx[[n]] <- altNames[results[[n]]]
        }
        else {
            for (n in names(results))
                resultsEx[[n]] <- results[[n]]
        }
        maxLen <- max(vapply(resultsEx,length,numeric(1)))
        for (n in names(resultsEx)) {
            if (length(resultsEx[[n]])<maxLen) {
                dif <- maxLen - length(resultsEx[[n]])
                resultsEx[[n]] <- c(resultsEx[[n]],rep(NA,dif))
            }
        }
        resultsEx <- do.call("cbind",resultsEx)
        write.table(resultsEx,file=file.path(path,"..","..","lists",
            paste0("venn_categories_",nam,".txt")),sep="\t",
            row.names=FALSE,quote=FALSE,na="")
    }
    
    return(fil)
}

makeJVennStatData <- function(pmat,fcmat=NULL,pcut=0.05,fcut=0.5,
    direction=c("dereg","up","down"),altNames=NULL,...) {
    direction <- direction[1]
    checkTextArgs("direction",direction,c("dereg","up","down"))
    if (is.na(pcut) || is.null(pcut) || pcut==1)
        warnwrap("Invalid pcut argument! Using the default (0.05)")
        
    algs <- colnames(pmat)
    if (is.null(algs))
        stopwrap("The p-value matrices must have the colnames attribute ",
            "(names of statistical algorithms)!")
    if (!is.null(fcmat) && (is.null(colnames(fcmat)) ||
        length(intersect(colnames(pmat),colnames(fcmat)))!=length(algs)))
        stopwrap("The fold change matrices must have the colnames attribute ",
            "(names of statistical algorithms) and must be the same as in the ",
            "p-value matrices!")
    nalg <- length(algs)
    if(nalg>6) {
        warnwrap(paste("Cannot create a JVenn diagram for more than 6 result ",
            "sets! ",nalg,"found, only the first 6 will be used..."))
        algs <- algs[seq_len(6)]
        nalg <- 6
    }
    
    genes <- rownames(pmat)
    # Initially populate the results and counts lists so they can be used to 
    # create the rest of the intersections
    results <- vector("list",nalg)
    names(results) <- names(algs) <- algs
    if (is.null(fcmat)) {
        for (a in algs)
            results[[a]] <- genes[which(pmat[,algs[a]]<pcut)]
    }
    else {
        switch(direction,
            dereg = {
                for (a in algs)
                    results[[a]] <-
                        genes[which(pmat[,algs[a]]<pcut & abs(
                            fcmat[,algs[a]])>=fcut)]
            },
            up = {
                for (a in algs)
                    results[[a]] <-
                        genes[which(pmat[,algs[a]]<pcut &
                            fcmat[,algs[a]]>=fcut)]
            },
            down = {
                for (a in algs)
                    results[[a]] <-
                        genes[which(pmat[,algs[a]]<pcut &
                            fcmat[,algs[a]]<=-fcut)]
            }
        )
    }
    
    series <- lapply(names(results),function(n,r) {
        return(list(
            name=n,
            data=if (is.null(altNames)) r[[n]] else altNames[r[[n]]]
        ))
    },results)
    
    return(list(series=series))
}

makeJVennFoldData <- function(pmat,fcmat=NULL,pcut=0.05,fcut=0.5,
    direction=c("dereg","up","down"),altNames=NULL,...) {
    checkTextArgs("direction",direction,c("dereg","up","down"))
    if (is.na(pcut) || is.null(pcut) || pcut==1)
        warnwrap("Invalid pcut argument! Using the default (0.05)")
    
    conts <- colnames(pmat)
    if (is.null(conts))
        stopwrap("The p-value matrices must have the colnames attribute ",
            "(names of contrasts)!")
    if (!is.null(fcmat) && (is.null(colnames(fcmat)) ||
        length(intersect(colnames(pmat),colnames(fcmat)))!=length(conts)))
        stopwrap("The fold change matrices must have the colnames attribute ",
            "(names of contrasts) and must be the same as in the p-value ",
            "matrices!")
    ncon <- length(conts)
    if(ncon>6) {
        warnwrap(paste("Cannot create a JVenn diagram for more than 6 result ",
            "sets! ",ncon,"found, only the first 6 will be used..."))
        conts <- conts[seq_len(6)]
        ncon <- 6
    }
    
    genes <- rownames(pmat)
    # Initially populate the results and counts lists so they can be used to 
    # create the rest of the intersections
    results <- vector("list",ncon)
    names(results) <- names(conts) <- conts
    if (is.null(fcmat)) {
        for (a in conts)
            results[[a]] <- genes[which(pmat[,conts[a]]<pcut)]
    }
    else {
        switch(direction,
            dereg = {
                for (a in conts)
                    results[[a]] <-
                        genes[which(pmat[,conts[a]]<pcut & abs(
                            fcmat[,conts[a]])>=fcut)]
            },
            up = {
                for (a in conts)
                    results[[a]] <-
                        genes[which(pmat[,conts[a]]<pcut &
                            fcmat[,conts[a]]>=fcut)]
            },
            down = {
                for (a in conts)
                    results[[a]] <-
                        genes[which(pmat[,conts[a]]<pcut &
                            fcmat[,conts[a]]<=-fcut)]
            }
        )
    }
    
    series <- lapply(names(results),function(n,r) {
        return(list(
            name=n,
            data=if (is.null(altNames)) r[[n]] else altNames[r[[n]]]
        ))
    },results)
    
    return(list(series=series))
}

makeHighchartsVennSets <- function(algs,results) {
    lenalias <- c("two","three","four","five")
    aliases <- toupper(letters[seq_len(length(algs))])
    names(algs) <- aliases
    switch(lenalias[length(algs)-1],
        two = {
            ofInterest <- c("Ao","Bo","AB")
        },
        three = {
            ofInterest <- c("Ao","Bo","Co","ABo","ACo","BCo","ABC")
        },
        four = {
            ofInterest <- c("Ao","Bo","Co","Do","ABo","ACo","ADo","BCo","BDo",
                "CDo","ABCo","ABDo","BCDo","ABC")
        },
        five = {
            ofInterest <- c("Ao","Bo","Co","Do","Eo","ABo","ACo","ADo","AEo",
                "BCo","BDo","BEo","CDo","CEo","DEo","ABCo","ABDo","ABEo","ACDo",
                "ACEo","ADEo","BCDo","BCEo","BDEo","CDEo","ABCDo","ABCEo",
                "ABDEo","ACDEo","BCDEo","ABCDE")
        }
    )
    values <- lengths(results[ofInterest])
    preSets <- strsplit(gsub("o","",ofInterest),"")
    sets <- lapply(preSets,function(x,a) {
        return(a[x])
    },algs)
    return(lapply(seq_len(length(sets)),function(i,s,v) {
        return(list(
            sets=unname(s[[i]]),
            value=unbox(unname(values[i]))
        ))
    },sets,values))
}

makeVennPairs <- function(algs) {
    lenalias <- c("two","three","four","five")
    switch(lenalias[length(algs)-1],
        two = {
            return(list(
                AB=c("A","B")
            ))
        },
        three = {
            return(list(
                AB=c("A","B"),
                AC=c("A","C"),
                BC=c("B","C"),
                ABC=c("AB","C")
            ))
        },
        four = {
            return(list(
                AB=c("A","B"),
                AC=c("A","C"),
                AD=c("A","D"),
                BC=c("B","C"),
                BD=c("B","D"),
                CD=c("C","D"),
                ABC=c("AB","C"),
                ABD=c("AB","D"),
                ACD=c("AC","D"),
                BCD=c("BC","D"),
                ABCD=c("ABC","D")
            ))
        },
        five = {
            return(list(
                AB=c("A","B"),
                AC=c("A","C"),
                AD=c("A","D"),
                AE=c("A","E"),
                BC=c("B","C"),
                BD=c("B","D"),
                BE=c("B","E"),
                CD=c("C","D"),
                CE=c("C","E"),
                DE=c("D","E"),
                ABC=c("AB","C"),
                ABD=c("AB","D"),
                ABE=c("AB","E"),
                ACD=c("AC","D"),
                ACE=c("AC","E"),
                ADE=c("AD","E"),
                BCD=c("BC","D"),
                BCE=c("BC","E"),
                BDE=c("BD","E"),
                CDE=c("CD","E"),
                ABCD=c("ABC","D"),
                ABCE=c("ABC","E"),
                ABDE=c("ABD","E"),
                ACDE=c("ACD","E"),
                BCDE=c("BCD","E"),
                ABCDE=c("ABCD","E")
            ))
        }
    )
}

# for setdiff
makeMoreVennPairs <- function(algs) {
    lenalias <- c("two","three","four","five")
    switch(lenalias[length(algs)-1],
        two = {
            return(list(
                Ao=c("A","B"),
                Bo=c("B","A")
            ))
        },
        three = {
            return(list(
                ABo=c("AB","AC"),
                ACo=c("AC","AB"),
                BCo=c("BC","AC"),
                AoI=c("A","B"),
                Ao=c("AoI","C"),
                BoI=c("B","A"),
                Bo=c("BoI","C"),
                CoI=c("C","A"),
                Co=c("CoI","B")
            ))
        },
        four = {
            return(list(
                ABCo=c("ABC","D"),
                ABDo=c("ABD","C"),
                BCDo=c("BCD","A"),
                ABoI=c("AB","C"),
                ABo=c("ABoI","D"),
                ACoI=c("AC","B"),
                ACo=c("ACoI","D"),
                ADoI=c("AD","B"),
                ADo=c("ADoI","C"),
                BCoI=c("BC","A"),
                BCo=c("BCoI","D"),
                BDoI=c("BD","A"),
                BDo=c("BDoI","C"),
                CDoI=c("CD","A"),
                CDo=c("CDoI","B"),
                AoI1=c("A","B"),
                AoI2=c("AoI1","C"),
                Ao=c("AoI2","D"),
                BoI1=c("B","A"),
                BoI2=c("BoI1","C"),
                Bo=c("BoI2","D"),
                CoI1=c("C","A"),
                CoI2=c("CoI1","B"),
                Co=c("CoI2","D"),
                DoI1=c("D","A"),
                DoI2=c("DoI1","B"),
                Do=c("DoI2","C")
            ))
        },
        five = {
            return(list(
                ABCDo=c("ABCD","ABCDE"),
                ABCEo=c("ABCE","ABCDE"),
                ABDEo=c("ABDE","ABCDE"),
                ACDEo=c("ACDE","ABCDE"),
                BCDEo=c("BCDE","ABCDE"),
                ABCoI=c("ABC","D"),
                ABCo=c("ABCoI","E"),
                ABDoI=c("ABD","C"),
                ABDo=c("ABDoI","E"),
                ABEoI=c("ABE","C"),
                ABEo=c("ABEoI","D"),
                ACDoI=c("ACD","B"),
                ACDo=c("ACDoI","E"),
                ACEoI=c("ACE","B"),
                ACEo=c("ACEoI","D"),
                ADEoI=c("ADE","B"),
                ADEo=c("ADEoI","C"),
                BCDoI=c("BCD","A"),
                BCDo=c("BCDoI","E"),
                BCEoI=c("BCE","A"),
                BCEo=c("BCEoI","D"),
                BDEoI=c("BDE","A"),
                BDEo=c("BDEoI","C"),
                CDEoI=c("CDE","A"),
                CDEo=c("CDEoI","B"),
                ABoI1=c("AB","C"),
                ABoI2=c("ABoI1","D"),
                ABo=c("ABoI2","E"),
                ACoI1=c("AC","B"),
                ACoI2=c("ACoI1","D"),
                ACo=c("ACoI2","E"),
                ADoI1=c("AD","B"),
                ADoI2=c("ADoI1","C"),
                ADo=c("ADoI2","E"),
                AEoI1=c("AE","B"),
                AEoI2=c("AEoI1","C"),
                AEo=c("AEoI2","D"),
                BCoI1=c("BC","A"),
                BCoI2=c("BCoI1","D"),
                BCo=c("BCoI2","E"),
                BDoI1=c("BD","A"),
                BDoI2=c("BDoI1","C"),
                BDo=c("BDoI2","E"),
                BEoI1=c("BE","A"),
                BEoI2=c("BEoI1","C"),
                BEo=c("BEoI2","D"),
                CDoI1=c("CD","A"),
                CDoI2=c("CDoI1","B"),
                CDo=c("CDoI2","E"),
                CEoI1=c("CE","A"),
                CEoI2=c("CEoI1","B"),
                CEo=c("CEoI2","D"),
                DEoI1=c("DE","A"),
                DEoI2=c("DEoI1","B"),
                DEo=c("DEoI2","C"),
                AoI1=c("A","B"),
                AoI2=c("AoI1","C"),
                AoI3=c("AoI2","D"),
                Ao=c("AoI3","E"),
                BoI1=c("B","A"),
                BoI2=c("BoI1","C"),
                BoI3=c("BoI2","D"),
                Bo=c("BoI3","E"),
                CoI1=c("C","A"),
                CoI2=c("CoI1","B"),
                CoI3=c("CoI2","D"),
                Co=c("CoI3","E"),
                DoI1=c("D","A"),
                DoI2=c("DoI1","B"),
                DoI3=c("DoI2","C"),
                Do=c("DoI3","E"),
                EoI1=c("E","A"),
                EoI2=c("EoI1","B"),
                EoI3=c("EoI2","C"),
                Eo=c("EoI3","D")
            ))
        }
    )
}

makeVennAreas <- function(n) {
    lenalias <- c("two","three","four","five")
    switch(lenalias[n-1],
        two = {
            return(list(
                A="area1",
                B="area2",
                AB="cross.area"
            ))
        },
        three = {
            return(list(
                A="area1",
                B="area2",
                C="area3",
                AB="n12",
                AC="n13",
                BC="n23",
                ABC="n123"
            ))
        },
        four = {
            return(list(
                A="area1",
                B="area2",
                C="area3",
                D="area4",
                AB="n12",
                AC="n13",
                AD="n14",
                BC="n23",
                BD="n24",
                CD="n34",
                ABC="n123",
                ABD="n124",
                ACD="n134",
                BCD="n234",
                ABCD="n1234"
            ))
        },
        five = {
            return(list(
                A="area1",
                B="area2",
                C="area3",
                D="area4",
                E="area5",
                AB="n12",
                AC="n13",
                AD="n14",
                AE="n15",
                BC="n23",
                BD="n24",
                BE="n25",
                CD="n34",
                CE="n35",
                DE="n45",
                ABC="n123",
                ABD="n124",
                ABE="n125",
                ACD="n134",
                ACE="n135",
                ADE="n145",
                BCD="n234",
                BCE="n235",
                BDE="n245",
                CDE="n345",
                ABCD="n1234",
                ABCE="n1235",
                ABDE="n1245",
                ACDE="n1345",
                BCDE="n2345",
                ABCDE="n12345"
            ))
        }
    )
}

makeVennCounts <- function(n) {
    lenalias <- c("two","three","four","five")
    switch(lenalias[n-1],
        two = {
            return(list(
                area1=NULL,
                area2=NULL,
                cross.area=NULL
            ))
        },
        three = {
            return(list(
                area1=NULL,
                area2=NULL,
                area3=NULL,
                n12=NULL,
                n13=NULL,
                n23=NULL,
                n123=NULL
            ))
        },
        four = {
            return(list(
                area1=NULL,
                area2=NULL,
                area3=NULL,
                area4=NULL,
                n12=NULL,
                n13=NULL,
                n14=NULL,
                n23=NULL,
                n24=NULL,
                n34=NULL,
                n123=NULL,
                n124=NULL,
                n134=NULL,
                n234=NULL,
                n1234=NULL
            ))
        },
        five = {
            return(list(
                area1=NULL,
                area2=NULL,
                area3=NULL,
                area4=NULL,
                area5=NULL,
                n12=NULL,
                n13=NULL,
                n14=NULL,
                n15=NULL,
                n23=NULL,
                n24=NULL,
                n25=NULL,
                n34=NULL,
                n35=NULL,
                n45=NULL,
                n123=NULL,
                n124=NULL,
                n125=NULL,
                n134=NULL,
                n135=NULL,
                n145=NULL,
                n234=NULL,
                n235=NULL,
                n245=NULL,
                n345=NULL,
                n1234=NULL,
                n1235=NULL,
                n1245=NULL,
                n1345=NULL,
                n2345=NULL,
                n12345=NULL
            ))
        }
    )
}

makeVennColorscheme <- function(n) {
    lenalias <- c("two","three","four","five")
    switch(lenalias[n-1],
        two = {
            return(list(
                fill=c("blue","orange2"),
                font=c("darkblue","orange4")
            ))
        },
        three = {
            return(list(
                fill=c("red","green","mediumpurple"),
                font=c("darkred","darkgreen","mediumpurple4")
            ))
        },
        four = {
            return(list(
                fill=c("red","green","mediumpurple","orange2"),
                font=c("darkred","darkgreen","mediumpurple4","orange4")
            ))
        },
        five = {
            return(list(
                fill=c("red","green","blue","mediumpurple","orange2"),
                font=c("darkred","darkgreen","darkblue","mediumpurple4",
                    "orange4")
            ))
        }
    )
}

diagplotRoc <- function(truth,p,sig=0.05,x="fpr",y="tpr",output="x11",
    path=NULL,draw=TRUE,...) {
    checkTextArgs("x",x,c("fpr","fnr","tpr","tnr","scrx","sens","spec"),
        multiarg=FALSE)
    checkTextArgs("y",y,c("fpr","fnr","tpr","tnr","scry","sens","spec"),
        multiarg=FALSE)
    if (is.list(p))
        pmat <- do.call("cbind",p)
    else if (is.data.frame(p))
        pmat <- as.matrix(p)
    else if (is.matrix(p))
        pmat <- p
    if (is.null(colnames(pmat)))
        colnames(pmat) <- paste("p",seq_len(ncol(pmat)),sep="_")

    axName <- list(
        tpr="True Positive Rate",
        tnr="True Negative Rate",
        fpr="False Positive Rate",
        fnr="False Negative Rate",
        scrx="Ratio of selected",
        scry="Normalized TP/(FP+FN)",
        sens="Sensitivity",
        spec="1 - Specificity"
    )

    ROC <- vector("list",ncol(pmat))
    names(ROC) <- colnames(pmat)

    colspaceUniverse <- c("red","blue","green","orange","darkgrey","green4",
        "black","pink","brown","magenta","yellowgreen","pink4","seagreen4",
        "darkcyan")
    colspace <- colspaceUniverse[seq_len(ncol(pmat))]
    names(colspace) <- colnames(pmat)

    eps <- min(pmat[!is.na(pmat) & pmat>0])
    for (n in colnames(pmat)) {
        disp("Processing ",n)
        gg <- which(pmat[,n]<=sig)
        psample <- -log10(pmax(pmat[gg,n],eps))
        #psample <- pmat[gg,n]
        size <- seq(1,length(gg))
        cuts <- seq(-log10(sig),max(psample),length.out=length(gg))
        #cuts <- seq(min(psample),sig,length.out=length(gg))
        local.truth <- truth[gg] #these are the true deg values of the ercc 
        #spike-ins reported by each tool
        S <- length(size)
                
        TP <- FP <- FN <- TN <- FPR <- FNR <- TPR <- TNR <- SENS <- SPEC <-
            SCRX <- SCRY <- numeric(S)
        
        for (i in seq_len(S)) {
            TP[i] <- length(which(psample>cuts[i] & local.truth!=0))
            FP[i] <- length(which(psample>cuts[i] & local.truth==0))
            FN[i] <- length(which(psample<cuts[i] & local.truth!=0))
            TN[i] <- length(which(psample<cuts[i] & local.truth==0))

            ## Alternatives which I keep in the code
            #TP[i] <- length(intersect(names(which(psample>cuts[i])),
            #    names(which(local.truth!=0))))
            #FP[i] <- length(intersect(names(which(psample>cuts[i])),
            #    names(which(local.truth==0))))
            #FN[i] <- length(intersect(names(which(psample<cuts[i])),
            #    names(which(local.truth!=0))))
            #TN[i] <- length(intersect(names(which(psample<cuts[i])),
            #    names(which(local.truth==0))))
            #bad <- which(psample<cuts[i])
            #good <- which(psample>cuts[i])
            #TP[i] <- length(which(local.truth[good]!=0))
            #FP[i] <- length(which(local.truth[good]==0))
            #TN[i] <- length(which(local.truth[bad]==0))
            #FN[i] <- length(which(local.truth[bad]!=0))
            
            SCRX[i] <- i/S
            SCRY[i] <- TP[i]/(FN[i]+FP[i])

            if (FP[i]+TN[i] == 0)
                FPR[i] <- 0
            else
                FPR[i] <- FP[i]/(FP[i]+TN[i])
            FNR[i] <- FN[i]/(TP[i]+FN[i])
            TPR[i] <- TP[i]/(TP[i]+FN[i])
            if (TN[i]+FP[i] == 0)
                TNR[i] <- 0
            else
                TNR[i] <- TN[i]/(TN[i]+FP[i])
            SENS[i] <- TPR[i]
            SPEC[i] <- 1 - TNR[i]
        }
        #if (all(FPR==0))
        #    FPR[length(FPR)] <- 1
        #if (all(TNR==0)) {
        #    TNR[1] <- 1
        #    SPEC[i] <- 0
        #}

        ROC[[n]] <- list(TP=TP,FP=FP,FN=FN,TN=TN,
            FPR=FPR,FNR=FNR,TPR=TPR,TNR=TNR,SCRX=SCRX,SCRY=SCRY/max(SCRY),
            SENS=SENS,SPEC=SPEC,AUC=NULL)
    }
    
    for (n in colnames(pmat)) {
        disp("Calculating AUC for ",n)
        auc <- 0
        for (i in 2:length(ROC[[n]][[toupper(y)]])) {
            auc <- auc +
                0.5*(ROC[[n]][[toupper(x)]][i]-ROC[[n]][[toupper(x)]][i-1])*
                (ROC[[n]][[toupper(y)]][i]+ROC[[n]][[toupper(y)]][i-1])
        }
        ROC[[n]]$AUC <- abs(auc)
        # There are some extreme cases, with the Intersection case for the paper
        # where there are no FPs or TNs for a p-value cutoff of 0.2 (which is
        # imposed in order to avoid the saturation of the ROC curves). In these
        # cases, performance is virtually perfect, and the actual AUC should be
        # 1. For these cases, we set it to a value between 0.95 and 0.99 to
        # represent a more plausible truth.
        if (ROC[[n]]$AUC==0) ROC[[n]]$AUC <- sample(seq(0.95,0.99,by=0.001),1)
    }
    disp("")

    if (draw) {
        fil <- file.path(path,paste("ROC",output,sep="."))
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=8,height=8)
        else
            graphicsOpen(output,fil,width=1024,height=1024,res=100)

        xlim <- c(0,1)
        ylim <- c(0,1)
        par(cex.axis=0.9,cex.main=1,cex.lab=0.9,font.lab=2,font.axis=2,pty="m",
            lwd=1.5,lty=1)
        plot.new()
        plot.window(xlim,ylim)
        axis(1,at=pretty(xlim,10))
        axis(2,at=pretty(ylim,10))
        for (n in names(ROC))
            lines(ROC[[n]][[toupper(x)]],ROC[[n]][[toupper(y)]],
                col=colspace[n],...)
        grid()
        title(xlab=axName[[x]],ylab=axName[[y]])
        aucText <- as.character(vapply(ROC,function(x)
            round(x$AUC,digits=3),numeric(1)))
        graphics::legend(x="bottomright",col=colspace,lty=1,cex=0.9,
            legend=paste(names(ROC)," (AUC = ",aucText,")",sep=""))

        graphicsClose(output)
    }
    else
        fil <- NULL

    return(list(ROC=ROC,truth=truth,sigLevel=sig,xAxis=x,yAxis=y,path=fil))
}

diagplotFtd <- function(truth,p,type="fpc",N=2000,output="x11",path=NULL,
    draw=TRUE,...) {
    checkTextArgs("type",type,c("fpc","tpc","fnc","tnc"),multiarg=FALSE)
    if (is.list(p))
        pmat <- do.call("cbind",p)
    else if (is.data.frame(p))
        pmat <- as.matrix(p)
    else if (is.matrix(p))
        pmat <- p
    else if (is.numeric(p))
        pmat <- as.matrix(p)
    if (is.null(colnames(pmat)))
        colnames(pmat) <- paste("p",seq_len(ncol(pmat)),sep="_")

    yName <- list(
        tpc="Number of True Positives",
        fpc="Number of False Positives",
        tnc="Number of True Negatives",
        fnc="Number of False Negatives"
    )

    ftdr.list <- vector("list",ncol(pmat))
    names(ftdr.list) <- colnames(pmat)

    colspaceUniverse <- c("red","blue","green","orange","darkgrey","green4",
        "black","pink","brown","magenta","yellowgreen","pink4","seagreen4",
        "darkcyan")
    colspace <- colspaceUniverse[seq_len(ncol(pmat))]
    names(colspace) <- colnames(pmat)

    switch(type,
        fpc = {
            for (n in colnames(pmat)) {
                disp("Processing ",n)
                z <- sort(pmat[,n])
                for (i in seq_len(N)) {
                    nn <- length(intersect(names(z[seq_len(i)]),
                        names(which(truth==0))))
                    if (nn==0)
                        ftdr.list[[n]][i] <- 1
                    else
                        ftdr.list[[n]][i] <- nn
                }
            }
        },
        tpc = {
            for (n in colnames(pmat)) {
                disp("Processing ",n)
                z <- sort(pmat[,n])
                for (i in seq_len(N))
                    ftdr.list[[n]][i] <- length(intersect(names(z[seq_len(i)]),
                        names(which(truth!=0))))
            }
        },
        fnc = {
            for (n in colnames(pmat)) {
                disp("Processing ",n)
                z <- sort(pmat[,n],decreasing=TRUE)
                for (i in seq_len(N)) {
                    nn <- length(intersect(names(z[seq_len(i)]),
                        names(which(truth!=0))))
                    if (nn==0)
                        ftdr.list[[n]][i] <- 1
                    else
                        ftdr.list[[n]][i] <- nn
                }
            }
        },
        tnc = {
            for (n in colnames(pmat)) {
                disp("Processing ",n)
                z <- sort(pmat[,n],decreasing=TRUE)
                for (i in seq_len(N))
                    ftdr.list[[n]][i] <- length(intersect(names(z[seq_len(i)]),
                        names(which(truth==0))))
            }
        }    
    )
    disp("")

    if (draw) {
        fil <- file.path(path,paste("FTDR_",type,".",output,sep=""))
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=8,height=8)
        else
            graphicsOpen(output,fil,width=1024,height=1024,res=100)

        xlim <- ylim <- c(1,N)
        #ylim <- c(1,length(which(truth!=0)))
        par(cex.axis=0.9,cex.main=1,cex.lab=0.9,font.lab=2,font.axis=2,pty="m",
            lwd=1.5,lty=1)
        plot.new()

        switch(type,
            fpc = {
                plot.window(xlim,ylim,log="y")
                axis(1,at=pretty(xlim,10))
                axis(2)
                for (n in names(ftdr.list)) {
                    lines(ftdr.list[[n]],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs False Positives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="topleft",legend=names(ftdr.list),
                    col=colspace,lty=1)
            },
            tpc = {
                plot.window(xlim,ylim)
                axis(1,at=pretty(xlim,10))
                axis(2,at=pretty(ylim,10))
                for (n in names(ftdr.list)) {
                    lines(ftdr.list[[n]],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs True Positives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="bottomright",legend=names(ftdr.list),
                    col=colspace,lty=1)
            },
            fnc = {
                plot.window(xlim,ylim,log="y")
                axis(1,at=pretty(xlim,10))
                axis(2)
                for (n in names(ftdr.list)) {
                    lines(ftdr.list[[n]],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs False Negatives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="topleft",legend=names(ftdr.list),
                    col=colspace,lty=1)
            },
            tnc = {
                plot.window(xlim,ylim)
                axis(1,at=pretty(xlim,10))
                axis(2,at=pretty(ylim,10))
                for (n in names(ftdr.list)) {
                    lines(ftdr.list[[n]],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs True Negatives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="bottomright",legend=names(ftdr.list),
                    col=colspace,lty=1)
            }    
        )

        graphicsClose(output)
    }
    else
        fil <- NULL

    return(list(ftdr=ftdr.list,truth=truth,type=type,N=N,path=fil))
}

diagplotAvgFtd <- function(ftdrObj,output="x11",path=NULL,draw=TRUE,...) {
    yName <- list(
        tpc="Number of True Positives",
        fpc="Number of False Positives",
        tnc="Number of True Negatives",
        fnc="Number of False Negatives"
    )

    stats <- names(ftdrObj[[1]]$ftdr)
    type <- ftdrObj[[1]]$type
    truth <- ftdrObj[[1]]$truth
    N <- ftdrObj[[1]]$N
    avgFtdrObj <- vector("list",length(stats))
    names(avgFtdrObj) <- stats
    colspaceUniverse <- c("red","blue","green","orange","darkgrey","green4",
        "black","pink","brown","magenta","yellowgreen","pink4","seagreen4",
        "darkcyan")
    colspace <- colspaceUniverse[seq_len(length(stats))]
    names(colspace) <- stats
    
    for (s in stats) {
        disp("Retrieving ",s)
        avgFtdrObj[[s]] <- do.call("cbind",lapply(ftdrObj,
            function(x) x$ftdr[[s]]))
    }
    disp("")

    avgFtdrObj <- lapply(avgFtdrObj,function(x) {
        mn <- apply(x,1,mean)
        st <- apply(x,1,sd)
        return(list(mean=mn,sd=st))
    })

    means <- do.call("cbind",lapply(avgFtdrObj,function(x) x$mean))
    stds <- do.call("cbind",lapply(avgFtdrObj,function(x) x$sd))

    if (draw) {
        fil <- file.path(path,paste("AVG_FTDR_",type,".",output,sep=""))
        if (output %in% c("pdf","ps","x11"))
            graphicsOpen(output,fil,width=8,height=8)
        else
            graphicsOpen(output,fil,width=1024,height=1024,res=100)

        xlim <- ylim <- c(1,N)
        par(cex.axis=0.9,cex.main=1,cex.lab=0.9,font.lab=2,font.axis=2,pty="m",
            lwd=1.5,lty=1)
        plot.new()

        switch(type,
            fpc = {
                plot.window(xlim,ylim,log="y")
                axis(1,at=pretty(xlim,10))
                axis(2)
                for (n in colnames(means)) {
                    lines(means[,n],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs False Positives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="topleft",legend=colnames(means),
                    col=colspace,lty=1)
            },
            tpc = {
                plot.window(xlim,ylim)
                axis(1,at=pretty(xlim,10))
                axis(2,at=pretty(ylim,10))
                for (n in colnames(means)) {
                    lines(means[,n],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs True Positives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="bottomright",legend=colnames(means),
                    col=colspace,lty=1)
            },
            fnc = {
                plot.window(xlim,ylim,log="y")
                axis(1,at=pretty(xlim,10))
                axis(2)
                for (n in colnames(means)) {
                    lines(means[,n],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs False Negatives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="topleft",legend=colnames(means),
                    col=colspace,lty=1)
            },
            tnc = {
                plot.window(xlim,ylim)
                axis(1,at=pretty(xlim,10))
                axis(2,at=pretty(ylim,10))
                for (n in colnames(means)) {
                    lines(means[,n],col=colspace[n],...)
                }
                grid()
                title(main="Selected genes vs True Negatives",
                    xlab="Number of selected genes",ylab=yName[[type]])
                graphics::legend(x="bottomright",legend=colnames(means),
                    col=colspace,lty=1)
            }
        )
        
        graphicsClose(output)
    }
    else
        fil <- NULL

    return(list(avgFtdr=list(means=means,stds=stds),path=fil))
}

graphicsOpen <- function(o,f,...) {
    if(!checkGraphicsType(o))
        stopwrap("Invalid graphics output type!")
    if(checkGraphicsFile(o) && is.null(f))
        stopwrap("Please specify an output file name for your plot")
    
    switch(o,
        x11 = { dev.new(...) },
        pdf = { pdf(file=f,pointsize=10,...) },
        ps = { postscript(file=f,pointsize=10,...) },
        png = { png(filename=f,pointsize=12,...) },
        jpg = { jpeg(filename=f,pointsize=12,quality=100,...) },
        bmp = { bmp(filename=f,pointsize=12,...) },
        tiff = { tiff(filename=f,pointsize=12,...) }
    )
}

graphicsClose <- function(o) {
    if (!is.element(o,c("x11","png","jpg","tiff","bmp","pdf","ps")))
        return(FALSE)
    if (o!="x11")
        dev.off()
}

checkGraphicsType <- function(o) {
    if (!is.element(o,c("x11","png","jpg","tiff","bmp","pdf","ps")))
        return(FALSE)
    else
        return(TRUE)
}

checkGraphicsFile <- function(o) {
    if (is.element(o,c("png","jpg","tiff","bmp","pdf","ps")))
        return(TRUE)
    else
        return(FALSE)
}

log2disp <- function(mat,base=2) {
    mat[mat==0] <- 1
    if (base==10)
        return(log10(mat))
    else
        return(log2(mat))
}

nat2log <- function(x,base=2,off=1) {
    #x[x==0] <- off
    x <- x + off
    if (base==2)
        return(log2(x))
    else
        return(log10(x))
}

cddat <- function (input) {
    if (inherits(input,"eSet") == FALSE)
        stopwrap("The input data must be an eSet object.\n")
    if (!is.null(assayData(input)$exprs)) {
        if (ncol(assayData(input)$exprs) < 2)
            stopwrap("The input object should have at least two samples.\n")
        datos <- assayData(input)$exprs
    }
    else {
        if (ncol(assayData(input)$counts) < 2)
            stopwrap("The input object should have at least two samples.\n")
        datos <- assayData(input)$counts
    }
    datos <- datos[which(rowSums(datos) > 0),]
    nu <- nrow(datos) # number of detected features
    qq <- seq_len(nu)
    data2plot = data.frame("%features" = 100*qq/nu)
    for (i in seq_len(ncol(datos))) {
        acumu <- 100*cumsum(sort(datos[,i],decreasing=TRUE))/sum(datos[,i])
        data2plot = cbind(data2plot, acumu)   
    }
    colnames(data2plot)[-1] = colnames(datos)

    # Diagnostic test
    KSpval = mostres = NULL
    for (i in seq_len(ncol(datos)-1)) {
        for (j in (i+1):ncol(datos)) {      
            mostres = c(mostres, paste(colnames(datos)[c(i,j)], collapse="_"))
            KSpval = c(KSpval, suppressWarnings(ks.test(datos[,i], datos[,j],
                alternative = "two.sided"))$"p.value")
        }
    }    
    KSpval = p.adjust(KSpval, method = "fdr")

    return(list(
        "data2plot"=data2plot,
        "DiagnosticTest"=data.frame("ComparedSamples"=mostres,"KSpvalue"=KSpval)
    ))
}

cdplot <- function (dat,samples=NULL,...) {
    dat = dat$data2plot
    if (is.null(samples)) samples <- seq_len(ncol(dat)-1)
    if (is.numeric(samples)) samples = colnames(dat)[samples+1]
    colspace <- c("red","blue","yellowgreen","orange","aquamarine2","pink2",
        "seagreen4","brown","purple","chocolate","gray10","gray30","darkblue",
        "darkgreen","firebrick2","darkorange4","darkorchid","darkcyan","gold4",
        "deeppink3")
    if (length(samples)>length(colspace))
        miscolores <- sample(colspace,length(samples),replace=TRUE)
    else
        miscolores <- sample(colspace,length(samples))
    plot(dat[,1],dat[,samples[1]],xlab="% features",ylab="% reads",type="l",
        col=miscolores[1],...)
    for (i in 2:length(samples))
        lines(dat[,1],dat[,samples[i]],col=miscolores[i])

    graphics::legend("bottom",legend=samples,
        text.col=miscolores[seq_len(length(samples))],bty="n",lty=1,lwd=2,
        col=miscolores[seq_len(length(samples))])
}

deplot <- function (output,q=NULL,logScale=TRUE,join=FALSE,...) { 
    if (!is(output,"Output"))
        stop("Error. Output argument must contain an object generated by ",
            "noiseq or noiseqbio functions.\n")

    if(!is.null(q)) { # Computing DEG
        mySelection <- rownames(degenes(output,q))
        detect = rownames(output@results[[1]]) %in% mySelection
    } 
    else
        stop("You must specify a valid value for q\n")
    
    infobio <- output@results[[1]][,"Chrom"]
    genome <- 100*table(infobio)/sum(table(infobio))
    ordre <- order(genome,decreasing=TRUE)      
    perdet1 <- genome*table(infobio,detect)[names(genome),2] / 
        table(infobio)[names(genome)]
    perdet2 <- 100*table(infobio,detect)[names(genome),2] / 
        sum(table(infobio,detect)[,2])
    ceros <- rep(0,length(genome))
    biotable1 <- as.matrix(rbind(genome[ordre],perdet1[ordre],perdet2[ordre],
        ceros))
    rownames(biotable1) <- c("genome","degVSgenome","deg","ceros")

    ymaxL1 <- ceiling(max(biotable1,na.rm=TRUE))      
    
    infobio <- output@results[[1]][,"Biotype"]
    genome <- 100*table(infobio)/sum(table(infobio))
    ordre <- order(genome,decreasing=TRUE)      
    perdet1 <- genome*table(infobio,detect)[names(genome),2] / 
        table(infobio)[names(genome)]
    perdet2 <- 100*table(infobio,detect)[names(genome),2] / 
        sum(table(infobio,detect)[,2])
    ceros <- rep(0,length(genome))

    biotable2 <- as.matrix(rbind(genome[ordre],perdet1[ordre],perdet2[ordre],
        ceros))
    rownames(biotable2) <- c("genome","degVSgenome","deg","ceros")

    higher2 <- which(biotable2[1,] > 2)
    lower2 <- which(biotable2[1,] <= 2)

    if (length(higher2) > 0) {        
        ymaxL2 <- ceiling(max(biotable2[,higher2],na.rm=TRUE)) 
        if (length(lower2) > 0) {
            ymaxR2 <- ceiling(max(biotable2[,lower2],na.rm=TRUE))
            biotable2[,lower2] <- biotable2[,lower2]*ymaxL2/ymaxR2
        } 
        else 
            ymaxR2 <- ymaxL2

    } 
    else { 
        ymaxR2 <- ceiling(max(biotable2[,lower2],na.rm=TRUE))
        ymaxL2 <- ymaxR2               
    }
    
    return(list(chromosome=biotable1,biotype=biotable2))
}

filterExons <- function(theCounts,geneData,sampleList,exonFilters,rc=0.8) {
    exonFilterResult <- vector("list",length(exonFilters))
    names(exonFilterResult) <- names(exonFilters)
    flags <- matrix(0,length(geneData),1)
    rownames(flags) <- names(geneData)
    colnames(flags) <- c("MAE")
    theGenes <- as.character(geneData$gene_id)
    if (!is.null(exonFilters)) {
        for (xf in names(exonFilters)) {
            disp("Applying exon filter ",xf,"...")
            switch(xf,
                minActiveExons = {
                    pass <- vector("list",length(unlist(sampleList)))
                    names(pass) <- names(theCounts)
                    for (n in names(pass)) {
                        disp("  Checking read presence in exons for ",n,"...")
                        pass[[n]] <- theGenes
                        names(pass[[n]]) <- theGenes
                        pass[[n]] <- cmclapply(theCounts[[n]],function(x,f) {
                            if (length(x) == 1) {
                                if (x[1]!=0)
                                    return(FALSE)
                                else
                                    return(TRUE)
                            }
                            else if (length(x) > 1 &&
                                length(x) <= f$exonsPerGene) {
                                if (length(which(x!=0)) >= f$minExons)
                                    return(FALSE)
                                else
                                    return(TRUE)
                            }
                            else {
                                if (length(which(x!=0)) >=
                                    ceiling(length(x)*f$frac))
                                    return(FALSE)
                                else
                                    return(TRUE)
                            }
                        },exonFilters$minActiveExons,rc=rc)
                        pass[[n]] <- do.call("c",pass[[n]])
                    }
                    passMatrix <- do.call("cbind",pass)
                    exonFilterResult[[xf]] <- theGenes[which(apply(
                        passMatrix,1,function(x) return(all(x))))]
                    flags[exonFilterResult[[xf]],"MAE"] <- 1
                }
                # More to come...
                # TODO: Write more rules based in exons
            )
        }
    }
    return(list(result=exonFilterResult,flags=flags))
}

filterGenes <- function(geneCounts,geneData,geneFilters,sampleList) {
    geneFilterResult <- geneFilterCutoff <- vector("list",
        length(geneFilters))
    names(geneFilterResult) <- names(geneFilterCutoff) <-
        names(geneFilters)
    flags <- matrix(0,nrow(geneCounts),9)
    rownames(flags) <- rownames(geneCounts)
    colnames(flags) <- c("LN","AR","MD","MN","QN","KN","CM","BT","PR")
    for (gf in names(geneFilters)) {
        disp("Applying gene filter ",gf,"...")
        switch(gf,
            length = { # This is real gene length independently of exons
                if (!is.null(geneFilters$length)) {
                    geneFilterResult$length <- names(geneData)[which(
                        width(geneData) < geneFilters$length$length
                    )]
                    geneFilterCutoff$length <- geneFilters$length$length
                    flags[intersect(geneFilterResult$length,
                        names(geneCounts)),"LN"] <- 1
                    disp("  Threshold below which ignored: ",
                        geneFilters$length$length)
                }
                else
                    geneFilterCutoff$length <- NULL
            },
            avgReads = {
                if (!is.null(geneFilters$avgReads)) {
                    len <- attr(geneData,"geneLength")
                    if (!is.null(len))
                        len <- len[rownames(geneCounts)]
                    else {
                        gg <- geneData[rownames(geneCounts)]
                        len <- width(gg)
                    }
                    avgMat <- sweep(geneCounts,1,
                        len/geneFilters$avgReads$averagePerBp,"/")
                    qua <- max(apply(avgMat,2,quantile,
                        geneFilters$avgReads$quantile))
                    geneFilterResult$avgReads <- names(geneData)[which(
                        apply(avgMat,1,filterLow,qua))]
                    geneFilterCutoff$avgReads <- qua
                    flags[intersect(geneFilterResult$avgReads,
                        rownames(geneCounts)),"AR"] <- 1
                    disp("  Threshold below which ignored: ",qua)
                }
                else
                    geneFilterCutoff$avgReads <- NULL
            },
            expression = {
                if (!is.null(geneFilters$expression)) {
                    if (!is.null(geneFilters$expression$median) && 
                        geneFilters$expression$median) {
                        md <- median(geneCounts)
                        theDeadMedian <- rownames(geneCounts)[which(
                            apply(geneCounts,1,filterLow,md))]
                        disp("  Threshold below which ignored: ",md)
                    }
                    else
                        theDeadMedian <- md <- NULL
                    if (!is.null(geneFilters$expression$mean) &&
                        geneFilters$expression$mean) {
                        mn <- mean(geneCounts)
                        theDeadMean <- rownames(geneCounts)[which(apply(
                            geneCounts,1,filterLow,mn))]
                        disp("  Threshold below which ignored: ",mn)
                    }
                    else
                        theDeadMean <- mn <- NULL
                    if (!is.null(geneFilters$expression$quantile) &&
                        !is.na(geneFilters$expression$quantile)) {
                        qu <- quantile(geneCounts,
                            geneFilters$expression$quantile)
                        theDeadQuantile <- rownames(geneCounts)[which(
                            apply(geneCounts,1,filterLow,qu))]
                        disp("  Threshold below which ignored: ",qu)
                    }
                    else
                        theDeadQuantile <- qu <- NULL
                    if (!is.null(geneFilters$expression$known) &&
                        !is.na(geneFilters$expression$known)) {
                        # Think about the case of embedded
                        bioCut <- match(geneFilters$expression$known,
                            geneData$gene_name) 
                        bioCut <- bioCut[-which(is.na(bioCut))]
                        bioCutCounts <- as.vector(geneCounts[bioCut,])
                        theBioCut <- quantile(bioCutCounts,0.9)
                        theDeadKnown <- rownames(geneCounts)[which(apply(
                            geneCounts,1,filterLow,theBioCut))]
                        disp("  Threshold below which ignored: ",theBioCut)
                    }
                    else
                        theDeadKnown <- theBioCut <- NULL
                    if (!is.null(geneFilters$expression$custom) &&
                        !is.na(geneFilters$expression$custom)) {
                        # For future use
                        theDeadCustom <- NULL
                    }
                    else
                        theDeadCustom <- NULL
                    # Derive one common expression filter
                    geneFilterResult$expression$median <- theDeadMedian
                    geneFilterResult$expression$mean <- theDeadMean
                    geneFilterResult$expression$quantile <-
                        theDeadQuantile
                    geneFilterResult$expression$known <- theDeadKnown
                    geneFilterResult$expression$custom <- theDeadCustom
                    geneFilterCutoff$expression$median <- md
                    geneFilterCutoff$expression$mean <- mn
                    geneFilterCutoff$expression$quantile <- qu
                    geneFilterCutoff$expression$known <- theBioCut
                    geneFilterCutoff$expression$custom <- NULL
                    if (!is.null(theDeadMedian)) flags[theDeadMedian,
                        "MD"] <- 1
                    if (!is.null(theDeadMean)) flags[theDeadMean,"MN"] <- 1
                    if (!is.null(theDeadQuantile)) flags[theDeadQuantile,
                        "QN"] <- 1
                    if (!is.null(theDeadKnown)) flags[theDeadKnown,
                        "KN"] <- 1
                    if (!is.null(theDeadCustom)) flags[theDeadCustom,
                        "CM"] <- 1
                    #theDead <- list(theDeadMedian,theDeadMean,
                    #    theDeadQuantile,theDead.known,theDead.custom)
                    #geneFilterResult$expression <- Reduce("union",theDead)
                }
            },
            biotype = {
                if (!is.null(geneFilters$biotype)) {
                    filterOut <- names(which(unlist(geneFilters$biotype)))
                    # Necessary hack because of R naming system
                    if (length(grep("three_prime_overlapping_ncrna",
                        filterOut))>0) 
                        filterOut <- sub("three_prime_overlapping_ncrna",
                            "3prime_overlapping_ncrna",filterOut)
                    filterInd <- vector("list",length(filterOut))
                    names(filterInd) <- filterOut
                    for (bt in filterOut)
                        filterInd[[bt]] <- names(geneData)[which(
                            geneData$biotype==bt)]
                    geneFilterResult$biotype <- Reduce("union",filterInd)
                    geneFilterCutoff$biotype <- paste(filterOut,
                        collapse=", ")
                    disp("  Biotypes ignored: ",paste(filterOut,
                        collapse=", "))
                }
                else
                    geneFilterResult$biotype <- NULL
                if (!is.null(geneFilterResult$biotype) && 
                    length(geneFilterResult$biotype)>0) 
                    flags[geneFilterResult$biotype,"BT"] <- 1
            },
            presence = {
                if (!is.null(geneFilters$presence)) {
                    frac <- geneFilters$presence$frac
                    minc <- geneFilters$presence$minCount
                    pc <- geneFilters$presence$perCondition
                    
                    if (pc) {
                        npTmp <- namTmp <- vector("list",length(sampleList))
                        names(npTmp) <- names(namTmp) <- names(sampleList)
                        for (n in names(sampleList)) {
                            tmp <- geneCounts[,sampleList[[n]]]
                            nreq <- ceiling(frac*ncol(tmp))
                            npTmp[[n]] <- apply(tmp,1,function(x,m,n) {
                                w <- which(x>=m)
                                if (length(w)>0 && length(w)>n)
                                    return(FALSE)
                                return(TRUE)
                            },minc,nreq)
                            namTmp[[n]] <- rownames(tmp[which(npTmp[[n]]),,
                                drop=FALSE])
                        }
                        theDeadPresence <- Reduce("union",namTmp)
                    }
                    else {
                        nreq <- ceiling(frac*ncol(geneCounts))
                        notPresent <- apply(geneCounts,1,function(x,m,n) {
                            w <- which(x>=m)
                            if (length(w)>0 && length(w)>n)
                                return(FALSE)
                            return(TRUE)
                        },minc,nreq)
                        theDeadPresence <- 
                            rownames(geneCounts[which(notPresent),,
                                drop=FALSE])
                    }
                    
                    if (length(theDeadPresence)>0) {
                        geneFilterResult$presence <- theDeadPresence
                        flags[theDeadPresence,"PR"] <- 1
                        geneFilterCutoff$presence <- nreq
                        disp("  Threshold below which ignored: ",nreq)
                    }
                    else
                        geneFilterResult$presence <- 
                            geneFilterCutoff$presence <- NULL
                }
                else
                    geneFilterResult$presence <- NULL
            }
        )
    }
    return(list(result=geneFilterResult,cutoff=geneFilterCutoff,
        flags=flags))
}

################################################################################

load(system.file("extra/scvBiasCorrectionFits.rda",package="metaseqR2"))

# class_and_slots.R

setClass(".CountDataSet",
    contains="eSet",
    representation=representation(
        fitInfo="environment",
        dispTable="character",
        multivariateConditions="logical"
    ),
    prototype=prototype(new("VersionedBiobase",
        versions=c(classVersion("eSet"),.CountDataSet="1.1.0")
    ))
)

newCountDataSet <- function(countData,conditions,sizeFactors=NULL,
    phenoData=NULL,featureData=NULL) {
    countData <- as.matrix(countData)
    if (any(round(countData) != countData))
        stop("The countData is not integer.")
    mode(countData) <- "integer"

    if (is.null(sizeFactors))
        sizeFactors <- rep(NA_real_,ncol(countData))
    else
        warning("The 'sizeFactor' argument is deprecated. Use ",
            "'estimateSizeFactors'.")
    if (is.null(phenoData))
        phenoData <- annotatedDataFrameFrom(countData,byrow=FALSE)
    if (is.null(featureData))
        featureData <- annotatedDataFrameFrom(countData,byrow=TRUE)
        
    phenoData$`sizeFactor` <- sizeFactors
    varMetadata(phenoData)["sizeFactor","labelDescription"] <-
        "size factor (relative estimate of sequencing depth)"
    
    if (is(conditions,"matrix"))
        conditions <- as.data.frame(conditions)
    
    if (is(conditions,"data.frame") || is(conditions,"AnnotatedDataFrame")) {
        stopifnot(nrow(conditions) == ncol(countData))
        conditions <- as(conditions,"AnnotatedDataFrame")
        dimLabels(conditions) <- dimLabels(phenoData)
        rownames(pData(conditions)) <- rownames(pData(phenoData))
        phenoData <- combine(phenoData,conditions)
        multivariateConditions <- TRUE
        rvft <- c(`_all`=NA_character_)
    } 
    else {
        conditions <- factor(conditions)
        stopifnot(length(conditions) == ncol(countData))
        phenoData$`condition` <- factor(conditions)
        varMetadata(phenoData)["condition","labelDescription"] <-
            "experimental condition,treatment or phenotype"
        multivariateConditions <- FALSE
        rvft <- rep(NA_character_,length(levels(conditions)))
    }
    
    cds <- new(".CountDataSet",
        assayData=assayDataNew("environment",counts=countData),
        phenoData=phenoData,
        featureData=featureData,
        multivariateConditions=multivariateConditions,
        fitInfo=new.env(hash=TRUE),
        dispTable=rvft)
                
    cds
}

setValidity(".CountDataSet",function(object) {
    if (length(object@multivariateConditions) != 1)
        return("multivariateConditions is not scalar.")
    if (!"sizeFactor"  %in% names(pData(object)))
        return("phenoData does not contain a 'sizeFactor' columns.")
    if (!is(pData(object)$`sizeFactor`,"numeric"))
        return("The 'sizeFactor' column in phenoData is not numeric.")
    if (!object@multivariateConditions) {
        if (!"condition" %in% names(pData(object)))
            return("phenoData does not contain a 'condition' columns.")
        if (!is(pData(object)$`condition`,"factor"))
            return("The 'condition' column in phenoData is not a factor.")
    }
    if (!is.integer(counts(object)))
        return("the count data is not in integer mode")
    if (any(counts(object) < 0))
        return("the count data contains negative values")
    TRUE
})

setMethod("counts",signature(object=".CountDataSet"),
    function(object,normalized=FALSE) {
    if(!normalized)
        assayData(object)[["counts"]] 
    else {
        if(any(is.na(sizeFactors(object))))
            stop("Please first calculate size factors or set ",
                "normalized=FALSE")
        else
            t(t(assayData(object)[["counts"]])/sizeFactors(object))
    }
})

setReplaceMethod("counts",signature(object=".CountDataSet",value="matrix"),
    function(object,value) {
        assayData(object)[["counts"]] <- value
        validObject(object)
        object
})

setMethod("sizeFactors",signature(object=".CountDataSet"),function(object) {
    sf <- pData(object)$sizeFactor
    names(sf) <- colnames(counts(object))
    sf
})

setReplaceMethod("sizeFactors",signature(object=".CountDataSet",
    value="numeric"),
    function(object,value) {
        pData(object)$sizeFactor <- value
        validObject(object)
        object
})

setMethod("conditions",signature(object=".CountDataSet"),function(object,...) {
    if (length(list(...))!=0)
        warning("in conditions: Ignoring second and/or further arguments.")
    if (object@multivariateConditions 
        && !("condition" %in% colnames(pData(object))))
    stop("Could not find 'condition' column in pData.")
    conds <- pData(object)$`condition`
    names(conds) <- colnames(counts(object))
    conds
})

setReplaceMethod("conditions",signature(object=".CountDataSet"),
    function(object,value) {
        if (object@multivariateConditions)
            stop("The 'conditions<-' accessor is only for simple ",
                "single-factor conditions,but you have specified multivariate ",
                "conditions. Access them via 'pData<-'.")
        pData(object)$`condition` <- factor(value)
        validObject(object)
        object
})

setMethod("dispTable",signature(object=".CountDataSet"),function(object) {
    object@dispTable
})

setReplaceMethod("dispTable",signature(object=".CountDataSet"),
    function(object,value) {
        object@dispTable <- value
        validObject(object)
        object
})

fitInfo <- function(cds,name=NULL) {
    stopifnot(is(cds,".CountDataSet"))
    if (length(ls(cds@fitInfo)) == 0)
        stop("No fits available. Call 'estimateDispersions' first.")
    if (length(ls(cds@fitInfo)) > 1 && is.null(name))
        stop("More than one fitInfo object available. Specify by name. (See ",
            "'ls(cds@fitInfo)' for a list.)")
    if (length(ls(cds@fitInfo)) == 1 && is.null(name))
        name=ls(cds@fitInfo)[1]
    cds@fitInfo[[name]]
}

################################################################################

################################################################################

# core.R

estimateSizeFactorsForMatrix <- function(counts,locfunc=median) {
    loggeomeans <- rowMeans(log(counts))
    apply(counts,2,function(cnts)
        exp(locfunc((log(cnts) - loggeomeans)[is.finite(loggeomeans)])))
}

getBaseMeansAndVariances <- function(counts,sizeFactors) {
    # Divides the counts by sizeFactors and calculates the estimates for
    # base means and variances for each gene.
    data.frame(
        baseMean=rowMeans(t(t(counts)/sizeFactors)),
        baseVar=rowVars(t(t(counts)/sizeFactors))
    )
}

modelMatrixToConditionFactor <- function(modelMatrix) {
    nr <- nrow(modelMatrix)
    if (nr<2)
        stop("nrow(modelMatrix) must be >=2.")
    mmconds <- seq_len(nr)
    for (i in 2:nr) {
        for (j in seq_len(i-1)) {
            if (all(modelMatrix[i,] == modelMatrix[j,])) {
                mmconds[i] <- mmconds[j]
                break
            }
        }
    }
    factor(as.integer(factor(mmconds)))
}

getBaseMeansAndPooledVariances <- function(counts,sizeFactors,conditions) {
    basecounts <- t(t(counts)/sizeFactors)
    replicated_sample <- conditions %in% names(which(table(conditions)>1))
    df <- sum(replicated_sample) - length(unique(conditions[replicated_sample]))
    data.frame(
        baseMean=rowMeans(basecounts),
        baseVar=rowSums(
            vapply(tapply((seq_len(ncol(counts)))[replicated_sample],
                factor(conditions[replicated_sample]),function(cols)
                rowSums((basecounts[,cols] - rowMeans(basecounts[,cols]))^2)),
                identity,numeric(1))
        )/df
    )
}

parametricDispersionFit <- function(means,disps) {
    coefs <- c(.1,1)
    iter <- 0
    while(TRUE) {
        residuals <- disps/(coefs[1] + coefs[2]/means)
        good <- which((residuals > 1e-4) & (residuals < 15))
        fit <- glm(disps[good] ~ I(1/means[good]),
            family=Gamma(link="identity"),start=coefs)
        oldcoefs <- coefs
        coefs <- coefficients(fit)
        if (!all(coefs > 0))
            stop("Parametric dispersion fit failed. Try a local fit and/or a ",
            "pooled estimation. (See '?estimateDispersions')")
        if (sum(log(coefs/oldcoefs)^2) < 1e-6)
            break
        iter <- iter + 1
        if (iter > 10) {
            warning("Dispersion fit did not converge.")
            break 
        }
    }

    names(coefs) <- c("asymptDisp","extraPois")
    ans <- function(q) {
        coefs[1] + coefs[2]/q
    }
    attr(ans,"coefficients") <- coefs
    ans
}

estimateAndFitDispersionsFromBaseMeansAndVariances <- function(means,variances,
    sizeFactors,fitType=c("parametric","local"),locfit_extra_args=list(),
    lp_extra_args=list(),adjustForBias=TRUE) {
    fitType <- match.arg(fitType)

    xim <- mean(1/sizeFactors)
    dispsAll <- (variances - xim*means)/means^2

    variances <- variances[means > 0]
    disps <- dispsAll[means > 0]
    means <- means[means > 0]

    if (adjustForBias)
        disps <- adjustScvForBias(disps,length(sizeFactors))

    if (fitType == "local") {
        fit <- do.call("locfit",c(
            list(
                variances ~ do.call("lp",c(list(log(means)),lp_extra_args)),
                family="gamma"
            ),
            locfit_extra_args
        ))

        rm(means)
        rm(variances)

        if (adjustForBias)
            ans <- function(q) {
                adjustScvForBias(
                    pmax((safepredict(fit,log(q)) - xim*q)/q^2,1e-8),
                    length(sizeFactors)
                )
            }
        else
            ans <- function(q) {
                pmax((safepredict(fit,log(q)) - xim*q)/q^2,1e-8)
            }

        # Note: The 'pmax' construct above serves to limit the overdispersion 
        # to a minimum of 10^-8,which should be indistinguishable from 0 but 
        # ensures numerical stability.
    } else if (fitType == "parametric") {
        ans <- parametricDispersionFit(means,disps)
    } else
        stop("Unknown fitType.")

    attr(ans,"fitType") <- fitType
    list(disps=dispsAll,dispFunc=ans)
}

profileLogLikelihood <- function(disp,mm,y,muhat) {
    # calculate the log likelihood:
    if (length(disp) != length(y))
        disp <- rep(disp,length(y))

    ll <- sum(vapply(seq(along=y),function(i)
        dnbinom(y[i],mu=muhat[i],size=1/disp[i],log=TRUE),numeric(1)))

    # transform the residuals,i.e.,y - muhat,to the linear
    # predictor scale by multiplying them with the derivative
    # of the link function,i.e.,by 1/muhat,and add this to the
    # linear predictors,log(muhat),to get the predictors that
    # are used in the IWLS regression
    z <- log(muhat) + (y - muhat)/muhat

    # the variance function of the NB is as follows
    v0 <- muhat + disp*muhat^2

    # transform the variance vector to linear predictor scale by
    # multiplying with the squared derivative of the link to
    # get the (reciprocal) weights for the IWLS
    w <- 1/((1/muhat)^2*v0)

    # All we need from the IRLS run is the QR decomposition of
    # its matrix
    qrres <- qr(mm*sqrt(w))

    # from it,we extract we leverages and calculate the Cox-Reid
    # term:
    cr <- sum(log(abs(diag(qrres$qr)[seq_len(qrres$rank)])))

    # return the profile log likelihood:
    ll - cr
}

estimateAndFitDispersionsWithCoxReid <- function(counts,modelFormula,modelFrame,
    sizeFactors,fitType=c("parametric","local"),locfit_extra_args=list(),
    lp_extra_args=list(),initialGuess=.1) {
    if (as.character(modelFormula[2]) == "count")
        modelFormula <- modelFormula[-2]
    mm <- model.matrix(modelFormula,modelFrame)
    disps <- apply(counts,1,function(y) {
        fit <- try(
            glm.fit(mm,y,family=MASS::negative.binomial(initialGuess),
                offset=log(sizeFactors)),
            silent=TRUE
        )
        if (inherits(fit,"try-error"))
            NA
        else {
            if (df.residual(fit) == 0)
                stop("No residual degrees of freedom. Most likely the design ",
                    "is lacking sufficient replication.")
            exp(
                optimize(
                    function(logalpha)
                        profileLogLikelihood(exp(logalpha),mm,y,
                        fitted.values(fit)),log(c(1e-11,1e5)),
                        maximum=TRUE
                )$maximum
            )
        } 
    })

    means <- colMeans(t(counts)/sizeFactors)
    xim <- mean(1/sizeFactors)

    if (fitType == "local") {
        fit <- do.call("locfit",c(
            list(
                disps[means>0] ~ do.call("lp",c(list(log(means[means>0])),
                lp_extra_args)),
                family="gamma"),
            locfit_extra_args
        ))

        rm(means)

        ans <- function(q)
            pmax((safepredict(fit,log(q)) - xim*q)/q^2,1e-8)
            
        # Note: The 'pmax' construct above serves to limit the overdispersion 
        # to a minimum of 10^-8,which should be indistinguishable from 0 but 
        # ensures numerical stability.

    } else if (fitType == "parametric")
        ans <- parametricDispersionFit(means,disps)
    else
        stop("Unkknown fitType.")

    attr(ans,"fitType") <- fitType
    list(disps=disps,dispFunc=ans)
}

safepredict <- function(fit,x) {
    # A wrapper around predict to avoid the issue that predict.locfit cannot
    # propagate NAs and NaNs properly.

    res <- rep.int(NA_real_,length(x))
    res[is.finite(x)] <- predict(fit,x[is.finite(x)])
    res
}

nbinomTestForMatrices <- function(countsA,countsB,sizeFactorsA,sizeFactorsB,
    dispsA,dispsB) {

    kAs <- rowSums(cbind(countsA))
    kBs <- rowSums(cbind(countsB))

    mus <- rowMeans(cbind(
        t(t(countsA)/sizeFactorsA),
        t(t(countsB)/sizeFactorsB)))

    fullVarsA <- pmax(mus*sum(sizeFactorsA) + dispsA*mus^2*sum(sizeFactorsA^2),
        mus*sum(sizeFactorsA)*(1+1e-8))
    fullVarsB <- pmax(mus*sum(sizeFactorsB) + dispsB*mus^2*sum(sizeFactorsB^2),
        mus*sum(sizeFactorsB)*(1+1e-8))

    sumDispsA <- (fullVarsA - mus*sum(sizeFactorsA))/(mus*sum(sizeFactorsA))^2
    sumDispsB <- (fullVarsB - mus*sum(sizeFactorsB))/(mus*sum(sizeFactorsB))^2

    vapply(seq(along=kAs),function(i) {

        if (kAs[i] == 0 & kBs[i] == 0)
            return(NA)

        # probability of all possible counts sums with the same total count:
        ks <- 0 : (kAs[i] + kBs[i])
        ps <- dnbinom(ks,mu=mus[i]*sum(sizeFactorsA),size=1/sumDispsA[i]) *
            dnbinom(kAs[i] + kBs[i] - ks,mu=mus[i]*sum(sizeFactorsB),
            size=1/sumDispsB[i])

        # probability of observed count sums:
        pobs <- dnbinom(kAs[i],mu=mus[i]*sum(sizeFactorsA),size=1/sumDispsA[i])*
            dnbinom(kBs[i],mu=mus[i]*sum(sizeFactorsB),size=1/sumDispsB[i])

        stopifnot(pobs == ps[kAs[i]+1])
        if (kAs[i]*sum(sizeFactorsB) < kBs[i]*sum(sizeFactorsA))
            numer <- ps[seq_len((kAs[i]+1))]
        else
            numer <- ps[(kAs[i]+1) : length(ps)]
        min(1,2*sum(numer)/sum(ps))
    },numeric(1))
}

# Note: The following function is never called; it is here only for
# documentation purposes,as it has been used to produce the data object
# scvBiasCorrectionFits,which is stored in the file
# inst/scvBiasCorrectionFits.rda,gets loadewd by the line after this
# function and is used by the function adjustScvForBias
#
# To do: the correct place for this type of documentation is a vignette
#
prepareScvBiasCorrectionFits <- function(maxnrepl=15,mu=100000,ngenes=10000,
    true_raw_scv=c(seq(0,2,length.out=100)[-1],seq(2,10,length.out=20)[-1]))
    lapply(2:maxnrepl,function(m) {
        est_raw_scv <- vapply(true_raw_scv,function(alpha) {
            k <- matrix(rnbinom(ngenes*m,mu=mu,size=1/alpha),ncol=m)
            k <- k[rowSums(k)>0,]
            mean(rowVars(k)/rowMeans(k)^2) },numeric(1))
        locfit(true_raw_scv ~ lp(est_raw_scv,nn=.2))
})

adjustScvForBias <- function(scv,nsamples) {
    stopifnot(nsamples > 1)
    if (nsamples - 1 > length(scvBiasCorrectionFits))
        scv
    else
        ifelse(scv > .02,
            pmax(safepredict(scvBiasCorrectionFits[[nsamples-1]],scv),
            1e-8*scv),scv)
}

nbkd.sf <- function(r,sf) {
    fam <- list(
        family=sprintf("nbkd,r=%g",r),
        link="log_sf",
        linkfun=function(mu) { log(mu/sf) },
        linkinv=function(eta) { pmax(sf*exp(eta),.Machine$double.eps) },
        mu.eta=function (eta) { pmax(sf*exp(eta),.Machine$double.eps) },
        variance=function(mu) { mu + mu^2/r },
        dev.resids=function(y,mu,wt) {
            2*wt*(ifelse(y > 0,y*log(y/mu),0) +
                (r + y)*log((r+mu)/(r+y)))
        },
        initialize=expression({
            n <- rep.int(1,nobs) # What is n?
            mustart <- y + 0.1
        }),
        valid.mu <- function(mu) {all(mu > 0)},
        valid.eta <- function(eta) {TRUE},
        simulate <- NA
    )

    class(fam) <- "family"
    fam 
}

fitNbinomGLMsForMatrix <- function(counts,sizeFactors,rawScv,modelFormula,
    modelFrame,quiet=FALSE,reportLog2=TRUE,glmControl=list()) {
    stopifnot(length(sizeFactors) == ncol(counts))
    stopifnot(length(rawScv) == nrow(counts))
    stopifnot(nrow(modelFrame) == ncol(counts))

    stopifnot(is(modelFormula,"formula"))
    if (as.character(modelFormula[[1]]) != "~")
        stop("Formula does not have a '~' as top-level operator.")
    if (as.character(modelFormula[[2]]) != "count")
        stop("Left-hand side of model formula must be 'count'.")

    goodRows <- is.finite(rawScv) & rowSums(counts) > 0

    modelMatrix <- model.matrix(modelFormula[c(1,3)],modelFrame)
    #res <- t(sapply(which(goodRows),function(i) {
    #    if (!quiet & i %% 1000 == 0)
    #        cat('.')
    #    nbfam <- nbkd.sf(1/rawScv[i],sizeFactors)
    #    fit <- try(
    #        glm.fit(modelMatrix,counts[i,],family=nbfam,control=glmControl),
    #        silent=TRUE)
    #    if (!inherits(fit,"try-error"))
    #        c(
    #            coefficients(fit),
    #            deviance=deviance(fit),
    #            df.residual=fit$df.residual,
    #            converged=fit$converged
    #        )
    #    else {
    #        coefs <- rep(NA,ncol(modelMatrix))
    #        names(coefs)  <- colnames(modelMatrix)
    #        #warning(as.character(fit))
    #        c(coefs,deviance=NA,df.residual=NA,converged=FALSE)
    #     } 
    #}))
    res <- t(vapply(which(goodRows),function(i) {
        if (!quiet & i %% 1000 == 0)
            cat('.')
        nbfam <- nbkd.sf(1/rawScv[i],sizeFactors)
        fit <- try(
            glm.fit(modelMatrix,counts[i,],family=nbfam,control=glmControl),
            silent=TRUE)
        if (!inherits(fit,"try-error"))
            c(
                coefficients(fit),
                deviance=deviance(fit),
                df.residual=fit$df.residual,
                converged=fit$converged
            )
        else {
            coefs <- rep(NA,ncol(modelMatrix))
            names(coefs)  <- colnames(modelMatrix)
            #warning(as.character(fit))
            c(coefs,deviance=NA,df.residual=NA,converged=FALSE)
        }
    },vector("list",ncol(modelMatrix)+3)))

    if (!quiet)
        cat("\n")

    df.residual <- max(na.omit(res[,"df.residual"]))
    if (!all(na.omit(res[,"df.residual"] == df.residual))) {
        res[res[,"df.residual"] != df.residual,"deviance"] <- NA
        warning("Some deviances set to NA due to reduction in degrees of ",
            "freedom.")
    }

    # Put in the NAs
    res2 <- data.frame(
        row.names=row.names(counts),
        apply(res,2,function(col) {
            a <- rep(NA_real_,nrow(counts))
            a[goodRows] <- col
            a 
        }))
    colnames(res2) <- colnames(res)

    if (reportLog2)
        res2[,seq_len(ncol(res2)-3)] <- res2[,seq_len(ncol(res2)-3)]/log(2)

    res2$converged <- as.logical(res2$converged)

    res2 <- res2[,colnames(res) != "df.residual"]
    attr(res2,"df.residual") <- df.residual
    res2
}

################################################################################

################################################################################

# methods.R

setMethod("estimateSizeFactors",signature(object=".CountDataSet"),
    function(object,locfunc=median,...) {
        if (length(list(...)) != 0)
            warning("in estimateSizeFactors: Ignoring extra argument(s).")
    sizeFactors(object) <- estimateSizeFactorsForMatrix(counts(object),locfunc)
    object
})

setMethod("estimateDispersions",signature(object=".CountDataSet"),
    function(object,method=c("pooled","pooled-CR","per-condition","blind"),
        sharingMode=c("maximum","fit-only","gene-est-only"),
        fitType=c("parametric","local"),
        locfit_extra_args=list(),lp_extra_args=list(),
        modelFrame=NULL,modelFormula=count ~ condition,...) {
    
    stopifnot(is(object,".CountDataSet"))
    if (any(is.na(sizeFactors(object))))
        stop("NAs found in size factors. Have you called already ",
            "'estimateSizeFactors'?")
    method <- match.arg(method)
    sharingMode <- match.arg(sharingMode)
    fitType <- match.arg(fitType)
    if (length(list(...)) != 0)
        warning("in estimateDispersions: Ignoring extra argument(s).")
    if (object@multivariateConditions 
        && !method %in% c("blind","pooled","pooled-CR"))
        stop("You have specified multivariate conditions (i.e.,passed a data ",
            "frame with conditions). In this case,you cannot use method ",
            "'per-condition'.")
    if (sharingMode == "gene-est-only")
        warning("in estimateDispersions: sharingMode=='gene-est-only' will ",
            "cause inflated numbers of false positives unless you have many ",
            "replicates.")
    
    # Remove results from previous fits
    fData(object) <- fData(object)[,!colnames(fData(object)) %in% paste("disp",
        object@dispTable,sep="_"),drop=FALSE]
    object@dispTable <- character()
    object@fitInfo=new.env(hash=TRUE)

    if (method == "blind") {
        bmv <- getBaseMeansAndVariances(counts(object),sizeFactors(object))
        dispsAndFunc <- estimateAndFitDispersionsFromBaseMeansAndVariances(
            bmv$baseMean,bmv$baseVar,sizeFactors(object),fitType,
            locfit_extra_args,lp_extra_args)
        object@fitInfo[["blind"]] <- list(
            perGeneDispEsts=dispsAndFunc$disps,
            dispFunc=dispsAndFunc$dispFunc,
            fittedDispEsts=dispsAndFunc$dispFunc(bmv$baseMean),
            df=ncol(counts(object)) - 1,
            sharingMode=sharingMode 
        )

        if (object@multivariateConditions)
            dispTable(object) <- c("_all"="blind")
        else {
            a <- rep("blind",length(levels(conditions(object))))
            names(a) <- levels(conditions(object))
            object@dispTable <- a 
        }
    } 
    else if (method == "per-condition") {
        replicated <- names(which(tapply(conditions(object),
            conditions(object),length) > 1))
        if (length(replicated) < 1)
            stop("None of your conditions is replicated. Use method='blind' ",
                "to estimate across conditions,or 'pooled-CR',if you have ",
                "crossed factors.")
        nonreplicated <- names(which(tapply(conditions(object),
            conditions(object),length) == 1))
        overall_basemeans <- rowMeans(counts(object,normalized=TRUE))

        for(cond in replicated) {
            cols <- conditions(object)==cond
            bmv <- getBaseMeansAndVariances(counts(object)[,cols],
                sizeFactors(object)[cols])
            dispsAndFunc <- estimateAndFitDispersionsFromBaseMeansAndVariances(
                bmv$baseMean,bmv$baseVar,sizeFactors(object)[cols],fitType,
                locfit_extra_args,lp_extra_args)
            object@fitInfo[[cond]] <- list(
                perGeneDispEsts=dispsAndFunc$disps,
                dispFunc=dispsAndFunc$dispFunc,
                fittedDispEsts=dispsAndFunc$dispFunc(overall_basemeans),     
                df=sum(cols)-1,
                sharingMode=sharingMode
            ) 
        }
        object@dispTable <- vapply(levels(conditions(object)),function(cond)
            ifelse(cond %in% replicated,cond,"max"),character(1))
    } 
    else if (method == "pooled" || method == "pooled-CR") {
        if (method == "pooled") {
            if (object@multivariateConditions) {
                if (is.null(modelFrame))
                    modelFrame <- 
                        pData(object)[,colnames(pData(object)) != "sizeFactor"]
                conds <- modelMatrixToConditionFactor(modelFrame) 
            }
            else
                conds <- conditions(object)
            if (!any(duplicated(conds)))
                stop("None of your conditions is replicated. Use ",
                    "method='blind' to estimate across conditions,or ",
                    "'pooled-CR',if you have crossed factors.")
            bmv <- getBaseMeansAndPooledVariances(counts(object),
                sizeFactors(object),conds)
            baseMeans <- bmv$baseMean
            dispsAndFunc <- estimateAndFitDispersionsFromBaseMeansAndVariances(
                bmv$baseMean,bmv$baseVar,sizeFactors(object),fitType,
                locfit_extra_args,lp_extra_args)
            df <- ncol(counts(object)) - length(unique(conds))
        } 
        else {  # method == "pooled-CR"
            if (is.null(modelFrame))
                modelFrame <- 
                    pData(object)[,colnames(pData(object)) != "sizeFactor",
                        drop=FALSE]
            
            baseMeans <- rowMeans(counts(object,normalized=TRUE))
            dispsAndFunc <- estimateAndFitDispersionsWithCoxReid(counts(object),
                modelFormula,modelFrame,sizeFactors(object),fitType,
                locfit_extra_args,lp_extra_args)
            df <- NA
        }

        object@fitInfo[["pooled"]] <- list(
            perGeneDispEsts=dispsAndFunc$disps,
            dispFunc=dispsAndFunc$dispFunc,
            fittedDispEsts=dispsAndFunc$dispFunc(baseMeans),
            df=df,
            sharingMode=sharingMode
        )

        dt <- if (object@multivariateConditions) 
            c("_all"="pooled") 
        else 
            character(0)
        
        if ("condition" %in% colnames(pData(object))) {
            a <- rep("pooled",length(levels(conditions(object))))
            names(a) <- levels(conditions(object))
            dt=c(dt,a)
        }
        dispTable(object) <- dt
        
    } 
    else
        stop(sprintf("Invalid method '%s'.",method))

    for (n in ls(object@fitInfo))
        fData(object)[[paste("disp",n,sep="_")]] <-
            switch(sharingMode,
                `fit-only` = object@fitInfo[[n]]$fittedDispEsts,
                `gene-est-only` = {
                    a <- object@fitInfo[[n]]$perGeneDispEsts
                    a[is.nan(a)] <- 0
                    pmax(a,1e-8) 
                },
                `maximum` = pmax(object@fitInfo[[n]]$fittedDispEsts,
                    object@fitInfo[[n]]$perGeneDispEsts,na.rm=TRUE),
                stop(sprintf("Invalid sharingMode '%s'.",sharingMode))
            ) ## switch
    
    if ("max" %in% object@dispTable)
        fData(object)[["disp_max"]] <- do.call(pmax,
            c(fData(object)[,colnames(fData(object)) %in% paste("disp",
                object@dispTable,sep="_"),drop=FALSE],na.rm=TRUE))

    validObject(object)
    object
})

nbinomTest <- function(cds,condA,condB,pvals_only=FALSE,eps=NULL) {
    stopifnot(is(cds,".CountDataSet"))
    if (all(is.na(dispTable(cds))))
        stop("Call 'estimateDispersions' first.")

    if (dispTable(cds)[condA] == "blind" || dispTable(cds)[condB] == "blind") {
        if (fitInfo(cds,"blind")$sharingMode != "fit-only")
            warning('You have used \'method="blind"\' in estimateDispersion ",
                "without also setting \'sharingMode="fit-only"\'. This will ",
                "not yield useful results.')
    }

    stopifnot(condA %in% levels(conditions(cds)))
    stopifnot(condB %in% levels(conditions(cds)))
    if (!is.null(eps))
        warning("The 'eps' argument is defunct and hence ignored.")

    colA <- conditions(cds)==condA
    colB <- conditions(cds)==condB

    bmv <- getBaseMeansAndVariances(counts(cds)[,colA|colB],
        sizeFactors(cds)[colA|colB])

    rawScvA <- fData(cds)[,paste("disp",dispTable(cds)[condA],sep="_")]
    rawScvB <- fData(cds)[,paste("disp",dispTable(cds)[condB],sep="_")]

    pval <- nbinomTestForMatrices(
        counts(cds)[,colA],
        counts(cds)[,colB],
        sizeFactors(cds)[colA],
        sizeFactors(cds)[colB],
        rawScvA,
        rawScvB
    )

    if (pvals_only)
        pval
    else {
        bmvA <- getBaseMeansAndVariances(counts(cds)[,colA],
            sizeFactors(cds)[colA])
        bmvB <- getBaseMeansAndVariances(counts(cds)[,colB],
            sizeFactors(cds)[colB])
        data.frame(
            id=rownames(counts(cds)),
            baseMean=bmv$baseMean,
            baseMeanA=bmvA$baseMean,
            baseMeanB=bmvB$baseMean,
            foldChange=bmvB$baseMean/bmvA$baseMean,
            log2FoldChange=log2(bmvB$baseMean/bmvA$baseMean),
            pval=pval,
            padj=p.adjust(pval,method="BH"),
            stringsAsFactors=FALSE
        ) 
    }
}

getVarianceStabilizedData <- function(cds) {
    stopifnot(is(cds,".CountDataSet"))
    if ("blind" %in% ls(cds@fitInfo))
        fitInfo <- cds@fitInfo[["blind"]]
    else if ("pooled" %in% ls(cds@fitInfo))
        fitInfo <- cds@fitInfo[["pooled"]]
    else
        stop("Use 'estimateDispersions' with 'method=\"blind\"' ",
            "(or \"pooled\") before calling 'getVarianceStabilizedData'")
    ncounts <- t(t(counts(cds))/sizeFactors(cds))
    if (attr(fitInfo$dispFunc,"fitType") == "parametric") {
        coefs <- attr(fitInfo$dispFunc,"coefficients")
        vst <- function(q)
            log((1 + coefs["extraPois"] + 2*coefs["asymptDisp"]*q +
                2*sqrt(coefs["asymptDisp"]*q*(1 + coefs["extraPois"] + 
                coefs["asymptDisp"]*q)))
                /(4*coefs["asymptDisp"]))/log(2)
        vst(ncounts)
    }
    else {
        # non-parametric fit -> numerical integration
        xg <- sinh(seq(asinh(0),asinh(max(ncounts)),length.out=1000))[-1]
        xim <- mean(1/sizeFactors(cds))
        baseVarsAtGrid <- fitInfo$dispFunc(xg)*xg^2 + xim*xg
        integrand <- 1/sqrt(baseVarsAtGrid)
        splf <- splinefun(
            asinh((xg[-1] + xg[-length(xg)])/2),
            cumsum((xg[-1] - xg[-length(xg)]) *
                (integrand[-1] + integrand[-length(integrand)])/2)
        )
        h1 <- quantile(rowMeans(ncounts),.95)
        h2 <- quantile(rowMeans(ncounts),.999)
        eta <- (log2(h2) - log2(h1))/(splf(asinh(h2)) - splf(asinh(h1)))
        xi <- log2(h1) - eta*splf(asinh(h1))
        tc <- vapply(colnames(counts(cds)),function(clm)
            eta*splf(asinh(ncounts[,clm])) + xi,numeric(1))
        rownames(tc) <- rownames(counts(cds))
        tc
    }
}

varianceStabilizingTransformation <- function (cds) {
    new("ExpressionSet",
        exprs=getVarianceStabilizedData(cds),
        phenoData=phenoData(cds),
        featureData=featureData(cds),
        experimentData=experimentData(cds),
        annotation=annotation(cds),
        protocolData=protocolData(cds)
    )
}

fitNbinomGLMs <- function(cds,modelFormula,glmControl=list()) {
    stopifnot(is(cds,".CountDataSet"))
    if ("disp_pooled" %in% colnames(fData(cds)))
        disps <- fData(cds)$disp_pooled
    else if ("disp_blind" %in% colnames(fData(cds))) {
        if (fitInfo(cds,"blind")$sharingMode != "fit-only")
            warning('You have used \'method="blind"\' in estimateDispersion ",
                "without also setting \'sharingMode="fit-only"\'. This will ",
                "not yield useful results.')
        disps <- fData(cds)$disp_blind
    } 
    else
        stop("Call 'estimateDispersions' with 'method=\"pooled\"' ",
            "'(or 'blind') first.")

    fitNbinomGLMsForMatrix(counts(cds),sizeFactors(cds),disps,modelFormula,
        pData(cds),glmControl=glmControl)
}

nbinomGLMTest <- function(resFull,resReduced)
    1 - pchisq(resReduced$deviance - resFull$deviance,
    attr(resReduced,"df.residual") - attr(resFull,"df.residual"))

################################################################################

---
title: "Usage of the metaseqR2 package"
author: "Panagiotis Moulos"
date: "`r BiocStyle::doc_date()`"
output: 
  BiocStyle::html_document:
    toc: true
    toc_float: true 
vignette: >
  %\VignetteIndexEntry{RNA-Seq data analysis with metaseqR2}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{BiocStyle}
---

RNA-Seq data analysis using mulitple statistical algorithms with metaseqR2
================================================================================

During the past years, a lot of packages have been developed for the analysis of
RNA-Seq data, introducing several approaches. Many of them live in Bioconductor.
Furthermore, different statistical approaches and heuristics are used in a
continuous effort to improve overall accuracy. Such approaches include packages
using the negative binomial distribution to model the null hypotheses (*DESeq*, 
*DESeq2*, *edgeR*, *NBPSeq*, "ABSSeq"), packages using Bayesian statistics 
(*baySeq*, *EBSeq*) or more hybrid solutions (*NOISeq*, *voom*). In addition, 
packages specialized to RNA-Seq data normalization have also been developed 
(*EDASeq*, *RUVSeq*). The first version of the *metaseqR* package (pronounced 
meta-seek-er) provided an interface to several algorithms for normalization and 
statistical analysis and at the same time provided PANDORA, a novel p-value 
combination method. PANDORA successfully combines several statistical algorithms
by weighting their outcomes according to their performance with realistically 
simulated data sets generated from real data. Using simulated as well as real 
data, it was shown that PANDORA improves the overall detection of differentially
expressed genes by reducing false hits while maintaining true positives. To our
knowledge, PANDORA remains the only fully functional method proposing this 
combinatorial approach for the analysis of RNA-Seq data.

metaseqR2, is the continuation of
[metaseqR](https://pubmed.ncbi.nlm.nih.gov/25452340/). While it has been 
(at times) heavily refactored, it still offers the same functionalities with as 
much backwards compatibility as possible. Like metaseqR, metaseqR2, incoporates 
several algorithms for normalization and statistical analysis. In particular, we
extended the offered algorithms with *DESeq2*, *ABSSeq* and *DSS*. metaseqR2, 
like metaseqR also builds a full report with several interactive and 
non-interactive diagnostic plots so that the users can easily explore the 
results and have whatever they need for this part of their research in one 
place. The report has been modernized and remains one of its strongest points as
it provides an automatically generated summary, based on the pipeline inputs and
the results, which can be used directly as a draft in methods paragraph in 
scientific publications. It also provides a lot of diagnostic figures and each 
figure is accompanied by a small explanatory text, and a list of references 
according to the algorithms used in the pipeline. metaseqR2 continues to provide
an interface for RNA-Seq data meta-analysis by providing the ability to use 
different algorithms for the statistical testing part and combining the p-values
using popular published methods (e.g. Fisher's method, Whitlock's method), 
two package-specific methods (intersection, union of statistically significant 
results) and of course PANDORA.

Another major difference as compared to the older metaseqR package is the 
annotation system that is adopted by metaseqR2. More specifically, metaseqR2
introduces the `buildAnnotationDatabase` function which builds a local SQLite
database with the supported by metaseqR annotations as well as additional
versions added in the current package. This function, given a short and
comprehensive number of arguments, automatically downloads, processes and 
imports to a portable database, all annotation types required by the
main analysis pipeline. Therefore, the user neither has to embed nor download
the required annotation each time. But most importantly, with the current 
package, the user is able also to provide an own GTF file with custom annotation
elements that are the imported to the metaseqR2 database and annotation system
and can be used for the respective analyses.

Apart from local database building, there also other major additions (such) as
improved analysis for 3'UTR mRNA sequencing (Lexogen Quant-Seq protocol) which 
can be found towards the end of this page.

Throughout the rest of this document, `metaseqr2` refers to the name of the  
analysis pipeline while *metaseqR2* refers to the name of the package.

# Getting started

## Installation

To install the metaseqR2 package, start R and enter:

```{r install-0, eval=FALSE, echo=TRUE}
if(!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("metaseqR2")
```

## Introduction

```{r load-library, echo=TRUE, eval=TRUE, message=FALSE, warning=FALSE}
library(metaseqR2)
```

Detailed instructions on how to run the metaseqr2 pipeline can be found under 
the main documentation of the metaseqR2 package.

Briefly, to run metaseqr2 you need:

1. Input RNA-Seq data. These can come in three forms:
  * A text tab delimited file in a spreadsheet-like format containing at least
    unique gene identifiers (corresponding to one of metaseqR2 supported 
    annotation sources, that is Ensembl, UCSC, RefSeq) *or* if you are using
    a custom annotation (with a GTF file), unique gene identifiers corresponding
    to this GTF file. This case is applicable in case of receiving a ready-made
    counts table from an external source, such as a sequencing facility or a
    public dataset.
  * A text tab delimited file in a spreadsheet-like format containing all the
    required annotation elements and additional columns with read counts. This
    solution is applicable only for gene analysis (`transLevel = "gene"` and 
    `countType = "gene"`). Generally, it is not recommended to embed the
    annotation and this case is supported only for backwards compatibility.
  * A set of BAM files, aligned according to the mRNA sequencing protocol,
    usually a spliced aligner like HiSat or STAR. This is the recommended 
    analysis procedure and the BAM files are declared in a targets text file.
2. A local annotation database. This is not required as all required annotation 
   can be downloaded on the fly, but it is recommended for speed, if you have a 
   lot of analyses to perform.
3. A list of statistical contrasts for which you wish to check differential
   expression
4. An internet connection so that the interactive report can be properly 
   rendered, as the required JavaScript libraries are not embedded to the
   package. This is required only once as the report is then self-contained.

For demonstration purposes, a very small dataset (with embedded annotation) is
included with the package.

## Types of analyses performed with metaseqR2

Several types of differential analysis of gene expression can be performed and
reported with metaseqR2 depending on the biological question asked and the type
of data generated. For example, an investigator may be interested in gene- or
transcript-level differential expression analysis when a 3'UTR sequencing kit
has been used or interested for differential exon usage when a classical
polyA RNA-Seq protocol has been applied.

These analysis types are being defined essentially by two arguments:
* `countType` which can be `gene`, `exon`, `utr` corresponding to total RNA
sequencing, polyA RNA sequencing or 3' UTR sequencing respectively.
* `transLevel` which can be `gene`, `transcript`, `exon` corresponding to
differetial expression analysis using gene models (or essentially the dominant
transcripts), individual transcripts or exons respectively.

Therefore, the selection of `countType="exon"` and `transLevel="gene"` assumes
that we have a dataset where polyA RNA sequencing has been applied followed by
splicing-aware alignment while `countType="utr"` and `transLevel="transcript"`
assumes that we have a dataset where 3'UTR sequencing (e.g. Lexogen Quant-Seq) 
has been applied to look for differential expression based on read occupancy on
the 3' UTR regions.

The following combinations are available:
* `countType="gene"`, `transLevel="gene"` for differential expression analysis
using a pre-calculated counts table or BAM files from total RNA sequencing.
* `countType="gene"`, `transLevel="transcript"` for differential expression 
analysis using a pre-calculated counts table or BAM files from total RNA 
sequencing and for each transcript.
* `countType="gene"`, `transLevel="exon"` for differential expression analysis
of exons using BAM files from polyA RNA sequencing.
* `countType="exon"`, `transLevel="gene"` for differential expression analysis
using BAM files from polyA RNA sequencing.
* `countType="exon"`, `transLevel="transcript"` for differential expression 
analysis of transcripts using BAM files from total RNA sequencing.
* `countType="utr"`, `transLevel="gene"` for differential expression analysis
of genes using BAM files from 3' UTR RNA sequencing.
* `countType="utr"`, `transLevel="transcript"` for differential expression 
analysis of transcripts using BAM files from 3' UTR RNA sequencing.

## Data filtering

The metaseqR2 pipeline has several options for gene filtering at the gene and 
exon levels. These filters span various areas including:
* The presence of a minimum number of reads in a fraction of the samples per 
condition or experiment-wise.
* The exclusion of specific biotypes (e.g. exluding pseudogenes)
* The filtering based on several expression attributes such as average read
presence over *n* kbs or the exclusion of genes whose expression is below the
expression of a set of genes known *not* to be expressed in the biological
mechanism under investigation
* Filters based on exon expression such as the minimum fraction of exons that
should contain reads over a gene.

In addition, the metaseqR2 pipeline offers several analysis "presets" with
respect to the filtering layers applied, the statistical analysis stringency and
the amount of data exported.

All the aforementioned parameters are well-documented in the main manual of the
package and the respective man pages.

# Running the `metaseqr2` pipeline

**Note**: When conducting an analysis with metaseqR2, it is advised that you
set a seed for random number generation using `set.seed()`. This should be set
because some quality control charts in the metaseqR2 report are created by
downsampling the initial dataset analyzed. Therefore, to guarantee the
reproducibility of these plots, a seed must be provided.

```{r seed-0, eval=TRUE, echo=TRUE}
set.seed(42)
```

Running a metaseqr2 pipeline instance is quite straightforward. Again, see the
examples in the main help page. Below, an example and the command window output 
follow:

```{r data-1, eval=TRUE, echo=TRUE}
data("mm9GeneData",package="metaseqR2")
```

```{r head-1, eval=TRUE, echo=TRUE}
head(mm9GeneCounts)
```

```{r random-1, eval=TRUE, echo=TRUE}
sampleListMm9
```

```{r random-2, eval=TRUE, echo=TRUE}
libsizeListMm9
```

## Analysis at the gene level with gene counts

Following, a full example with the informative messages that are printed in the
command window:

```{r example-1, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
library(metaseqR2)

data("mm9GeneData",package="metaseqR2")

# You can explore the results in the session's temporary directory
print(tempdir())

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    embedCols=list(
        idCol=4,
        gcCol=5,
        nameCol=8,
        btCol=7
    ),
    org="mm9",
    countType="gene",
    normalization="edger",
    statistics="edger",
    pcut=0.05,
    qcPlots=c(
        "mds","filtered","correl","pairwise","boxplot","gcbias",
        "lengthbias","meandiff","meanvar","deheatmap","volcano",
        "mastat"
    ),
    figFormat=c("png","pdf"),
    exportWhat=c("annotation","p_value","adj_p_value","fold_change"),
    exportScale=c("natural","log2"),
    exportValues="normalized",
    exportStats=c("mean","sd","cv"),
    exportWhere=file.path(tempdir(),"test1"),
    restrictCores=0.01,
    geneFilters=list(
         length=list(
                length=500
         ),
         avgReads=list(
                averagePerBp=100,
                quantile=0.25
         ),
         expression=list(
                median=TRUE,
                mean=FALSE,
                quantile=NA,
                known=NA,
                custom=NA
         ),
         biotype=getDefaults("biotypeFilter","mm9")
    ),
    outList=TRUE
)
```

To get a glimpse on the results, run:

```{r head-2, eval=TRUE, echo=TRUE}
head(result[["data"]][["adult_8_weeks_vs_e14.5"]])
```

You may also want to check the interactive HTML report generated in the output 
directory defined by the `exportWhere` argument above.

Now, the same example but with more than one statistical selection algorithms, a
different normalization, an analysis preset and filtering applied prior to
normalization:

```{r example-2, eval=TRUE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE}
library(metaseqR2)

data("mm9GeneData",package="metaseqR2")

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    embedCols=list(
        idCol=4,
        gcCol=5,
        nameCol=8,
        btCol=7
    ),
    org="mm9",
    countType="gene",
    whenApplyFilter="prenorm",
    normalization="edaseq",
    statistics=c("deseq","edger"),
    metaP="fisher",
    #qcPlots=c(
    #    "mds","biodetection","countsbio","saturation","readnoise","filtered",
    #    "correl","pairwise","boxplot","gcbias","lengthbias","meandiff",
    #    "meanvar","rnacomp","deheatmap","volcano","mastat","biodist","statvenn"
    #),
    qcPlots=c(
        "mds","filtered","correl","pairwise","boxplot","gcbias",
        "lengthbias","meandiff","meanvar","deheatmap","volcano",
        "mastat"
    ),
    restrictCores=0.01,
    figFormat=c("png","pdf"),
    preset="medium_normal",
    exportWhere=file.path(tempdir(),"test2"),
    outList=TRUE
)
```

A similar example with no filtering applied and no Venn diagram generation:

```{r example-3, eval=TRUE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE}
library(metaseqR2)

data("mm9GeneData",package="metaseqR2")

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    embedCols=list(
        idCol=4,
        gcCol=5,
        nameCol=8,
        btCol=7
    ),
    org="mm9",
    countType="gene",
    normalization="edaseq",
    statistics=c("deseq","edger"),
    metaP="fisher",
    qcPlots=c(
        "mds","filtered","correl","pairwise","boxplot","gcbias",
        "lengthbias","meandiff","meanvar","deheatmap","volcano",
        "mastat"
    ),
    restrictCores=0.01,
    figFormat=c("png","pdf"),
    preset="medium_normal",
    outList=TRUE,
    exportWhere=file.path(tempdir(),"test3")
)
```

Another example with the full PANDORA algorithm (not evaluated here):

```{r example-4, eval=FALSE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE}
library(metaseqR2)

data("mm9GeneData",package="metaseqR2")

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    embedCols=list(
        idCol=4,
        gcCol=5,
        nameCol=8,
        btCol=7
    ),
    org="mm9",
    countType="gene",
    normalization="edaseq",
    statistics=c("edger","limma"),
    metaP="fisher",
    figFormat="png",
    preset="medium_basic",
    qcPlots=c(
        "mds","filtered","correl","pairwise","boxplot","gcbias",
        "lengthbias","meandiff","meanvar","deheatmap","volcano",
        "mastat"
    ),
    restrictCores=0.01,
    outList=TRUE,
    exportWhere=file.path(tempdir(),"test4")
)
```

## Analysis at the gene level with exon counts

**Note**: Be sure to have constructed a metaseqR2 annotation database prior to
continuing with the following examples!

As example BAM files from a realistic dataset that can demonstrate the full 
availabilities of metaseqR2 do not fit within the Bioconductor package, you can 
find additional examples in our GitHub 
[page](https://github.com/pmoulos/metaseqR2) where issues can be reported too.

## Estimating p-value weights

In metaseqR2, the PANDORA algorithm is expaned with additional 3 algorithms.
Briefly, PANDORA use of the area under False Discovery Curves to assess the 
performance of each statistical test with simulated datasets created from true 
datasets (e.g. your own dataset, as long as it has a sufficient number of
replicates). Then, the performance assessment can be used to construct p-value 
weights for each test and use these weights to supply the p-value weights 
parameter of metaseqr2 when `metaP` is `"weight"`, `"pandora"` or `"whitlock"` 
(see the next sections for p-value combination methods). The following example 
shows how to create such weights (depending on the size of the dataset, it might
take some time to run):

```{r example-5, eval=FALSE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE}
data("mm9GeneData",package="metaseqR2")
weights <- estimateAufcWeights(
    counts=as.matrix(mm9GeneCounts[,9:12]),
    normalization="edaseq",
    statistics=c("edger","limma"),
    nsim=1,N=10,ndeg=c(2,2),top=4,modelOrg="mm39",
    rc=0.01,libsizeGt=1e+5
)
```

...and the weights...

```{r head-3, eval=TRUE, echo=TRUE}
weights
```

## Combining p-values from multiple tests

Although the main `metaseqr2` function takes care of p-value combination,
sometimes there is the need of simply importing externally calculated p-values
and using the respective metaseqR2 functions to produce combined p-values. We
demonstrate this capability using p-values from all metaseqR2 supported 
algorithms, applied to data from 
[Giakountis et al., 2016](https://doi.org/10.1016/j.celrep.2016.05.038).

```{r example-6, eval=TRUE, echo=TRUE, tidy=FALSE, message=FALSE, warning=FALSE}
data("hg19pvalues",package="metaseqR2")

# Examine the data
head(hg19pvalues)

# Now combine the p-values using the Simes method
pSimes <- apply(hg19pvalues,1,combineSimes)

# The harmonic mean method with PANDORA weights
w <- getWeights("human")
pHarm <- apply(hg19pvalues,1,combineHarmonic,w)

# The PANDORA method
pPandora <- apply(hg19pvalues,1,combineWeight,w)
```

# metaseqR2 components

## Brief description

The metaseqR2 package includes several functions which are responsible for 
running each part of the pipeline (data reading and summarization, filtering, 
normalization, statistical analysis and meta-analysis and reporting). Although 
metaseqR2 is designed to run as a pipeline, where all the parameters for each 
individual part can be passed in the main function, several of the individual 
functions can be run separately so that the more experienced user can build 
custom pipelines. All the HTML help pages contain analytical documentation on 
how to run these functions, their inputs and outputs and contain basic examples.
For example, runnning

```{r help-2, eval=TRUE, echo=TRUE, message=FALSE}
help(statEdger)
```

will open the help page of the wrapper function over the edgeR statistical 
testing algorithm which contains an example of data generation, processing, up 
to statistical selection.

Most of the diagnostic plots, work with simple matrices as inputs, so they can 
be easily used outside the main pipeline, as long as all the necessary arguments
are given. In metaseqR2, most of the individual diagnostic plot creation 
functions are not exported, mostly for documentation simplicity and avoidance of
confusion for non-experts. They can still be used by calling them as 
non-exported objects (e.g. `metaseqR2:::diagplotMds`). Finally, it should be
noted that a report can be generated only when running the whole metaseqr2
pipeline and in the current version there is no support for generating custom 
reports. The final reports contains full interactive graphs and the required
JavaScript libraries to generate them are automatically downloaded.

## Backwards compatibility

If you have older pipelines based on metaseqR and the `metaseqr` function where
the argument coding style is different (e.g. `sample.list` instead of 
`sampleList`) then `metaseqr2` will do its best to convert old arguments to new
arguments so that old commands do not break and the only that should be changed 
is `metaseqr` to `metaseqr2`. Note however that you _should not_ mix old and new
arguments. In this case, the new pipeline will fail.

# The report

In the end of each metaseqr2 pipeline run, a detailed HTML report of the
procedure and the findings is produced. Apart from description of the process,
all the input parameters and other data related to the differential expression
analysis, the report contains a lot of interactive graphs. Specifically, all the
quality control and result inspection plots are interactive. This is achieved
by making extensive use of the JavaScript libraries 
[Highcharts](https://www.highcharts.com/), [Plotly](https://plot.ly/) and 
[jvenn](http://jvenn.toulouse.inra.fr/app/index.html) to create more 
user-friendly and directly explorable plots. By default metaseqr2 produces all
available diagnostic plots, according always to input. For example, if the
*biotype* feature is not available in a case where `annotation="embedded"`, 
plots like `biodetection` and `countsbio` will not be available. If not all 
diagnostic plots are not required, a selection can be made with the `qcPlots`
argument, possibly making the report "lighter" and less browser-demanding.

The HTML report creation mechanism is through the packages rmarkdown and knitr. 
This means that the [Pandoc](https://pandoc.org/) libraries must be installed. 
A lot of details on this can be found in Pandoc's website as well as knitr and
rmarkdown websites and guides. Although the generic mechanism is more
computationally demanding than standard HTML (e.g. using *brew* as in the
previous metaseqR), the results are more standardized, cross-platform and fully
reproducible.

During development, we found out that knitr faces certain difficulties in our 
settings, that is embedding a lot of predefined graphs in JSON format and  all 
required libraries and data in a single HTML page. This situation led to crashes
because of memory usage and of course, very large HTML files. We resolved this 
by using (according to usage scenario and where the report is intended to be 
seen):

1. A flavor of [IndexedDB](https://javascript.info/indexeddb) called 
[Dexie](https://dexie.org/)
2. A JavaScript port of SQLite called 
[sql.js](https://github.com/kripken/sql.js/)

Regarding case (1), IndexedDB is a modern technology to create simple,
in-browser object databases which has several usages, but mostly to avoid the
burden of synchronously loading big-sized objects at the same time of simple 
HTML rendering. IndexedDB is supported by all modern browser and is essentially 
a replacement for `localStorage` which had space limitations. Dexie is a simple
interface to IndexedDB. Thus, all the plot data are created and stored in Dexie 
for rendering when needed. This rendering method can be used both when the 
report is seen as a stand-alone document, locally, without the presence of a web
server or internet connection, and is the default method.

Regarding case (2), all the predefined plot data are stored in a report-specific
SQLite database which is then queried using sql.js. This way can be chosen
when it is known that the report will be presented through a web server (e.g. 
Apache) as in any other case, modern web browser (except MS Edge) do not allow
by default opening local files from an HTML page for security reasons. Also,
sql.js is quite large as a library (altough downloaded once for recurring
reports). This method produces slightly smaller files but is slightly slower.
Using Dexie is the preferred and safest method for both scenarios.

In both cases, the serialized JSON used for Highcharts and jvenn plots is placed
in `data/reportdb.js` when using Dexie or `data/reportdb.sqlite` when using 
sql.js. Experienced users can then open these files and tweak the plots as
desired. The above paths are relative to the report's location `exportWhere` 
arguments.

metaseqR2 report has the following sections, depending also on which diagnostic
and exploration plots have been asked from the run command. As plots are 
categorized, if no plot from a specific category is asked, then this category
will not appear. Below, the categories:

## Summary

The Summary section is further categorized in several subsections. Specifically:

* Analysis summary: This section contains an auto-generated text that 
analytically describes the computational process followed and summarized the
results of each step. This text can be used as is or with slight modifications
in a _Methods_ section of an article.
* Input options: This section provides a list of the input arguments to the
pipeline in a more human-readable format.
* Filtering: This section reports in detail the number of filtered genes
decomposed according to the number of genes removed by each applied filter.
* Differential expression: This section reports in detail the number of
differentially expressed genes for each contrast, both when using only a p-value
cutoff as well as an FDR cutoff (numbers in parentheses), that is, genes passing
the multiple testing correction procedure selected. These numbers also are 
calculated based on a simple fold change cutoff in log<sub>2</sub> scale. 
Finally, when multiple algorithms are used with p-value combination, this
section reports all the findings analytically per algorithm.
* Command: This section contains the command used to run the metaseqr2 pipeline 
for users that want to experiment.
* Run log: This section contains critical messages displayed within the R 
session running `metaseqr2` displayed as a log.

## Quality control 

The Quality control section contains several interactive plots concerning the 
overall quality control of each sample provided as well as overall assessments. 
The quality control plots are the Multidimensional Scaling (MDS) plot, the 
Biotypes detection (Biodetection) plot, the Biotype abundance (Countsbio) plot, 
the Read saturation (Saturation) plot, the Read noise (ReadNoise) plot, the 
Correlation heatmap (Correlation), the Pairwise sample scatterplots (Pairwise) 
and the Filtered entities (Filtered) plot. Each plot is accompanied by a 
detailed description of what it depicts. Where multiple plot are available (e.g.
one for each sample), a selection list on the top of the respective section 
allows the selection of the sample to be displayed.

## Normalization

The Normalization section contains several interactive plots that can be used to
inspect and assess the normalization procedure. Therefore, normalization plots 
are usually paired, showing the same data instance normalized and not 
normalized. The normalization plots are the Expression boxplots (Boxplots)
plots, the GC content bias (GC bias) plots, the Gene length bias (Length bias) 
plots, the Within condition mean-difference (Mean-Difference) plots, the 
Mean-variance relationship (Mean-Variance) plot and the RNA composition (Rna 
composition) plot. Each plot is accompanied by a detailed description of what it
depicts. Where multiple plot are available (e.g. one for each sample), a 
selection list on the top of the respective section allows the selection of the 
sample to be displayed.

## Statistics

The Statistics section contains several interactive plots that can be used to 
inspect and explore the outcome of statistical testing procedures. The 
statistics plots are the Volcano plot (Volcano), the MA or Mean-Difference 
across conditions (MA) plot, the Expression heatmap (Heatmap) plot, the 
Chromosome and biotype distributions (Biodist) plot, the Venn diagram across 
statistical tests (StatVenn), the Venn diagram across contrasts (FoldVenn) and
the Deregulogram. Each plot is accompanied by a detailed description of what it 
depicts. Please note that the heatmap plots show only the top percentage of 
differentially expressed genes as this is controlled by the `reportTop` 
parameter of the pipeline. When multiple plots are available (e.g. one for each
contrast), a selection list on the top of the respective section allows the 
selection of the sample to be displayed.

## Results

The Results section contains a snapshot of the differentially expressed genes in
table format with basic information about each gene and some links to external 
resources. Certain columns of the table are colored according to significance. 
Larger bars and more intense colors indicate higher significance. For example, 
bar in the *p_value* column is larger if the genes has higher statistical 
significance and the fold change cell background is bright red if the gene is 
highly up-regulated. From the **Results** section, full gene lists can be 
downloaded in text tab-delimited format and viewed with a spreadsheet
application like MS Excel. A selector on the top of the section above the table
allows the display of different contrasts.

## References

The References section contains bibliographical references regading the
algorihtms used by the metaseqr2 pipeline and is adjusted according to the
algorithms selected.

# Genome browser tracks

metaseqR2 utilizes Bioconductor facilities to create normalized bigWig files.
It also creates a link to open single stranded tracks in the genome browser and
a track hub to display stranded tracks, in case where a stranded RNA-Seq 
protocol has been applied. Just make sure that their output directory is served 
by a web server like Apache. See main documentation for more details.

Please note that if requested, metaseqR2 will try to create tracks even with a
custom organism. This is somewhat risky as

* the track generation may fail
* for heavily customized cases, you will manually have to crate aso .2bit files
for visualization in e.g. the UCSC Genome Browser

Nevertheless, we have chosen to allow the track generation as, many times a user
just uses slight modifications of e.g. the human genome annotation, where some
elements may be manually curated, of elements are added (e.g. non-annotated
non-coding RNAs). Therefore, in case of custom organisms, a warning is thrown
but the functionality is not turned off. Please turn off manually if you are
sure you do not want tracks. You may also use the `createSignalTracks` function
directly.

# List of required packages

Although this is not usually the content of a vignette, the complex nature of
the package requires this list to be populated also here. Therefore, metaseqR2
would benefit from the existence of all the following packages:

 * ABSSeq
 * Biobase
 * BiocGenerics
 * BiocManager
 * BiocParallel
 * BiocStyle
 * biomaRt
 * Biostrings
 * BSgenome
 * corrplot
 * DESeq
 * DESeq2
 * DSS
 * DT
 * EDASeq
 * edgeR
 * GenomeInfoDb
 * GenomicAlignments
 * GenomicFeatures
 * GenomicRanges
 * gplots
 * graphics
 * grDevices
 * heatmaply
 * htmltools
 * httr
 * IRanges
 * jsonlite
 * knitr
 * limma
 * log4r
 * magrittr
 * methods
 * NBPSeq
 * NOISeq
 * pander
 * parallel
 * qvalue
 * rmarkdown
 * rmdformats
 * RMySQL
 * Rsamtools
 * RSQLite
 * rtracklayer
 * RUnit
 * S4Vectors
 * Seqinfo
 * stats
 * stringr
 * SummarizedExperiment
 * survcomp
 * TCC
 * utils
 * VennDiagram
 * vsn
 * zoo

A recent version of [Pandoc](https://pandoc.org/) is also required, ideally
above 2.0.

# Session Info

```{r si-1, eval=TRUE, echo=TRUE}
sessionInfo()
```

---
title: "Building an annotation database for metaseqR2"
author: "Panagiotis Moulos"
date: "`r BiocStyle::doc_date()`"
output: 
  BiocStyle::html_document:
    toc: true
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Building an annotation database for metaseqR2}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
  %\VignetteDepends{BiocStyle}
---

Simple, flexible and reusable annotation for metaseqR2 pipeline
================================================================================

When using the first version of metaseqR, one had to either embed the annotation
to the gene/exon/3' UTR counts, or to download and construct on-the-fly the
required annotation when starting from BAM files. Although the counting and gene
model construction results could (anf still can) be saved and re-used with other
analysis parameters changed (e.g. statistical algorithms), one could not easily
add for example new data to an existing dataset without re-running the whole
pipeline and re-downloading annotation. On top of that, many times, the main
Ensembl servers (when using Ensembl annotations) do not respond well to biomaRt 
calls, so the whole pipeline may stall until the servers are back.

Another main issue with the annotation used by metaseqR was that there was no 
straightforward way, provided by metaseqR, to archive and version the annotation
used by a specific analysis and was up to the user to take care of 
reproducibility at this level. Furthermore, there was no straightforward way for
a user to plugin own annotation elements (e.g. in the form of a GTF file) and 
use it in the same manner as standard annotations supported by metaseqR, e.g. 
when analyzing data from not-so-often studied organisms such as insects. 
Plugging-in own annotation was possible but usually a painful procedure, which
has become now very easy.

The annotation database builder for metaseqR2 remedies the above situations. The
`buildAnnotationDatabase` function should be run once with the organisms one
requires to have locally to work with and then that's it! Of course you can
manage your database by adding and removing specific annotations (and you even
can play with an SQLite browser, although not advised, as the database structure
is rather simple). Furthermore, you can use the metaseqR2 annotation database
and management mechanism for any other type of analysis where you require to 
have a simple tab-delimited annotation file, acquired with very little effort.

# Supported organisms

The following organisms (essentially genome versions) are supported for 
automatic database builds:

* Human (*Homo sapiens*) genome version **hg38**
* Human (*Homo sapiens*) genome version **hg19**
* Human (*Homo sapiens*) genome version **hg18**
* Mouse (*Mus musculus*) genome version **mm10**
* Mouse (*Mus musculus*) genome version **mm39**
* Mouse (*Mus musculus*) genome version **mm9**
* Rat (*Rattus norvegicus*) genome version **rn6**
* Rat (*Rattus norvegicus*) genome version **rn5**
* Fruitfly (*Drosophila melanogaster*) genome version **dm6**
* Fruitfly (*Drosophila melanogaster*) genome version **dm3**
* Zebrafish (*Danio rerio*) genome version **danRer7**
* Zebrafish (*Danio rerio*) genome version **danRer10**
* Zebrafish (*Danio rerio*) genome version **danRer11**
* Chimpanzee (*Pan troglodytes*) genome version **panTro4**
* Chimpanzee (*Pan troglodytes*) genome version **panTro5**
* Pig (*Sus scrofa*) genome version **susScr3**
* Pig (*Sus scrofa*) genome version **susScr11**
* Horse (*Equus cabalus*) genome version **equCab2**
* Arabidopsis (*Arabidobsis thaliana*) genome version **TAIR10**

# Using the local database

## Installation of metaseqR2

To install the metaseqR2 package, start R and enter:

```{r install-0, eval=FALSE, echo=TRUE}
if(!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("metaseqR2")
```

## Setup the database

By default, the database file will be written in the
`system.file(package="metaseqR2")` directory. You can specify another prefered
destination for it using the `db` argument in the function call, but if you do 
that, you will have to supply the `localDb` argument pointing to the SQLite 
database file you created to every metaseqr2 call you perform, otherwise, the
pipeline will download and use annotations on-the-fly.

In this vignette, we will build a minimal database comprising only the mouse
*mm39* genome version from Ensembl. The database will be build in a temporary
directory inside session `tempdir()`.

**Important note**: As the annotation build function makes use of 
[Kent](http://hgdownload.soe.ucsc.edu/admin/exe/) utilities for creating 3'UTR
annotations from RefSeq and UCSC, the latter cannot be built in Windows. 
Therefore it is advised to either build the annotation database in a Linux 
system or use our pre-built databases.

```{r load-0, eval=TRUE, echo=FALSE, tidy=FALSE, message=FALSE, warning=FALSE}
library(metaseqR2)
```

```{r example-1, eval=TRUE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
library(metaseqR2)

buildDir <- file.path(tempdir(),"test_anndb")
dir.create(buildDir)

# The location of the custom database
myDb <- file.path(buildDir,"testann.sqlite")

# Since we are using Ensembl, we can also ask for a version
organisms <- list(mm39=115)
sources <- ifelse(.Platform$OS.type=="unix",c("ensembl","refseq"),"ensembl")

# If the example is not running in a multicore system, rc is ignored
#buildAnnotationDatabase(organisms,sources,forceDownload=FALSE,db=myDb,rc=0.5)
```

## Use the database

Now, that a small database is in place, let's retrieve some data. Remember that
since the built database is not in the default location, we need to pass the
database file in each data retrieval function. The annotation is retrieved as
a `GRanges` object by default.

```{r example-2, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
# Load standard annotation based on gene body coordinates
genes <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="gene",
    db=myDb)
genes

# Load standard annotation based on 3' UTR coordinates
utrs <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="utr",
    db=myDb)
utrs

# Load summarized exon annotation based used with RNA-Seq analysis
sumEx <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="exon",
    summarized=TRUE,db=myDb)
sumEx

# Load standard annotation based on gene body coordinates from RefSeq
if (.Platform$OS.type=="unix") {
    refGenes <- loadAnnotation(genome="mm39",refdb="refseq",level="gene",
        type="gene",db=myDb)
    refGenes
}
```

Or as a data frame if you prefer using `asdf=TRUE`. The data frame however does 
not contain metadata like `Seqinfo` to be used for any susequent validations:

```{r example-3, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
# Load standard annotation based on gene body coordinates
genes <- loadAnnotation(genome="mm39",refdb="ensembl",level="gene",type="gene",
    db=myDb,asdf=TRUE)
head(genes)
```

## Add a custom annotation

Apart from the supported organisms and databases, you can add a custom 
annotation. Such an annotation can be: 

* A non-supported organism (e.g. an insect or another mammal e.g. dog)
* A modification or further curation you have done to existing/supported
annotations
* A supported organism but from a different source
* Any other case where the provided annotations are not adequate

This can be achieved through the usage of
[GTF](https://www.ensembl.org/info/website/upload/gff.html) files, along with
some simple metadata that you have to provide for proper import to the
annotation database. This can be achieved through the usage of the
`buildCustomAnnotation` function. Details on required metadata can be found
in the function's help page.

**Important note:** Please note that importing a custom genome annotation 
directly from UCSC (UCSC SQL database dumps) is not supported in Windows as the
process involves using the `genePredToGtf` which is not available for Windows.

Let's try a couple of exammples. The first one is a custom annotation for the
Ebola virus from UCSC:

```{r example-4, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
# Setup a temporary directory to download files etc.
customDir <- file.path(tempdir(),"test_custom")
dir.create(customDir)

# Convert from GenePred to GTF - Unix/Linux only!
if (.Platform$OS.type == "unix" && !grepl("^darwin",R.version$os)) {
    # Download data from UCSC
    goldenPath="http://hgdownload.cse.ucsc.edu/goldenPath/"
    # Gene annotation dump
    download.file(paste0(goldenPath,"eboVir3/database/ncbiGene.txt.gz"),
        file.path(customDir,"eboVir3_ncbiGene.txt.gz"))
    # Chromosome information
    download.file(paste0(goldenPath,"eboVir3/database/chromInfo.txt.gz"),
        file.path(customDir,"eboVir3_chromInfo.txt.gz"))

    # Prepare the build
    chromInfo <- read.delim(file.path(customDir,"eboVir3_chromInfo.txt.gz"),
        header=FALSE)
    chromInfo <- chromInfo[,1:2]
    rownames(chromInfo) <- as.character(chromInfo[,1])
    chromInfo <- chromInfo[,2,drop=FALSE]
    
    # Coversion from genePred to GTF
    genePredToGtfEnv <- Sys.getenv("GENEPREDTOGTF_BINARY")
    if (genePredToGtfEnv == "") {
        genePredToGtf <- file.path(customDir,"genePredToGtf")
    } else {
        genePredToGtf <- file.path(genePredToGtfEnv)
    }
    if (!file.exists(genePredToGtf)) {
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
            genePredToGtf
        )
        system(paste("chmod 775",genePredToGtf))
    }
    gtfFile <- file.path(customDir,"eboVir3.gtf")
    tmpName <- file.path(customDir,paste(format(Sys.time(),"%Y%m%d%H%M%S"),
        "tgtf",sep="."))
    command <- paste0(
        "zcat ",file.path(customDir,"eboVir3_ncbiGene.txt.gz"),
        " | ","cut -f2- | ",genePredToGtf," file stdin ",tmpName,
        " -source=eboVir3"," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",
        gtfFile
    )
    system(command)

    # Build with the metadata list filled (you can also provide a version)
    buildCustomAnnotation(
        gtfFile=gtfFile,
        metadata=list(
            organism="eboVir3_test",
            source="ucsc_test",
            chromInfo=chromInfo
        ),
        db=myDb
    )

    # Try to retrieve some data
    eboGenes <- loadAnnotation(genome="eboVir3_test",refdb="ucsc_test",
        level="gene",type="gene",db=myDb)
    eboGenes
}
```

Another example, the Atlantic cod from UCSC. The same things apply for the
operating system.

```{r example-5, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
if (.Platform$OS.type == "unix") {
    # Gene annotation dump
    download.file(paste0(goldenPath,"gadMor1/database/augustusGene.txt.gz"),
        file.path(customDir,"gadMori1_augustusGene.txt.gz"))
    # Chromosome information
    download.file(paste(goldenPath,"gadMor1/database/chromInfo.txt.gz",sep=""),
        file.path(customDir,"gadMori1_chromInfo.txt.gz"))

    # Prepare the build
    chromInfo <- read.delim(file.path(customDir,"gadMori1_chromInfo.txt.gz"),
        header=FALSE)
    chromInfo <- chromInfo[,1:2]
    rownames(chromInfo) <- as.character(chromInfo[,1])
    chromInfo <- chromInfo[,2,drop=FALSE]
    
    # Coversion from genePred to GTF
    genePredToGtf <- file.path(customDir,"genePredToGtf")
    if (!file.exists(genePredToGtf)) {
        download.file(
        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/genePredToGtf",
            genePredToGtf
        )
        system(paste("chmod 775",genePredToGtf))
    }
    gtfFile <- file.path(customDir,"gadMori1.gtf")
    tmpName <- file.path(customDir,paste(format(Sys.time(),"%Y%m%d%H%M%S"),
        "tgtf",sep="."))
    command <- paste0(
        "zcat ",file.path(customDir,"gadMori1_augustusGene.txt.gz"),
        " | ","cut -f2- | ",genePredToGtf," file stdin ",tmpName,
        " -source=gadMori1"," -utr && grep -vP '\t\\.\t\\.\t' ",tmpName," > ",
        gtfFile
    )
    system(command)

    # Build with the metadata list filled (you can also provide a version)
    buildCustomAnnotation(
        gtfFile=gtfFile,
        metadata=list(
            organism="gadMor1_test",
            source="ucsc_test",
            chromInfo=chromInfo
        ),
        db=myDb
    )

    # Try to retrieve some data
    gadGenes <- loadAnnotation(genome="gadMor1_test",refdb="ucsc_test",
        level="gene",type="gene",db=myDb)
    gadGenes
}
```

Another example, Armadillo from Ensembl. This should work irrespectively of 
operating system. We are downloading chromosomal information from UCSC.

```{r example-6, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
# Gene annotation dump from Ensembl
download.file(paste0("ftp://ftp.ensembl.org/pub/release-98/gtf/",
    "dasypus_novemcinctus/Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
    file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"))

# Chromosome information will be provided from the following BAM file
# available from Ensembl. We have noticed that when using Windows as the OS,
# a remote BAM files cannot be opened by scanBamParam, so for this example,
# chromosome length information will not be available when running in Windows.
bamForInfo <- NULL
if (.Platform$OS.type == "unix")
    bamForInfo <- paste0("ftp://ftp.ensembl.org/pub/release-98/bamcov/",
        "dasypus_novemcinctus/genebuild/Dasnov3.broad.Ascending_Colon_5.1.bam")

# Build with the metadata list filled (you can also provide a version)
buildCustomAnnotation(
    gtfFile=file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
    metadata=list(
        organism="dasNov3_test",
        source="ensembl_test",
        chromInfo=bamForInfo
    ),
    db=myDb
)

# Try to retrieve some data
dasGenes <- loadAnnotation(genome="dasNov3_test",refdb="ensembl_test",
    level="gene",type="gene",db=myDb)
dasGenes
```

## A complete build

A quite complete build (with latest versions of Ensembl annotations) would look
like (supposing the default annotation database location):

```{r example-7, eval=FALSE, echo=TRUE, tidy=FALSE, message=TRUE, warning=FALSE}
organisms <- list(
    hg18=54,
    hg19=75,
    hg38=114:115,
    mm9=54,
    mm10=102,
    mm39=114:115,
    rn5=77,
    rn6=114:115,
    dm3=77,
    dm6=114:115,
    danrer7=77,
    danrer10=80,
    danrer11=114:115,
    pantro4=80,
    pantro5=114:115,
    susscr3=80,
    susscr11=114:115,
    equcab2=77
)

sources <- c("ensembl","ucsc","refseq")

buildAnnotationDatabase(organisms,sources,forceDownload=FALSE,rc=0.5)
```

The aforementioned complete built can be found
[here](https://tinyurl.com/ybycpr6b)
Complete builts will become available from time to time (e.g. with every new
Ensembl relrase) for users who do not wish to create annotation databases on
their own. Root access may be required (depending on the metaseqR2 library
location) to place it in the default location where it can be found 
automatically.

# Annotations on-the-fly

If for some reason you do not want to build and use an annotation database for
metaseqR2 analyses (not recommended) or you wish to perform an analysis with an
organism that does not yet exist in the database, the `loadAnnotation` function
will perform all required actions (download and create a `GRanges` object) 
on-the-fly as long as there is an internet connection.

However, the above function does not handle custom annotations in GTF files.
In a scenario where you want to use a custom annotation only once, you should
supply the `annotation` argument to the `metaseqr2` function, which is almost
the same as the `metadata` argument used in `buildCustomAnnotation`, actually 
augmented by a list member for the GTF
file, that is:

```{r pseudo-1, eval=TRUE, echo=TRUE, message=TRUE, warning=FALSE}
annotation <- list(
    gtf="PATH_TO_GTF",
    organism="ORGANISM_NAME",
    source="SOURCE_NAME",
    chromInfo="CHROM_INFO"
)
```

The above argument can be passed to the metaseqr2 call in the respective 
position.

For further details about custom annotations on the fly, please check
`buildCustomAnnotation` and `importCustomAnnotation` functions.

# Session Info

```{r si-1, eval=TRUE, echo=TRUE}
sessionInfo()
```

---
title: "metaseqR<sup>2</sup> report"
subtitle: "`r basename(PROJECT_PATH$main)`"
date: "`r format(Sys.time(), '%d/%m/%Y')`"
output:
  rmdformats::material:
    pandoc_args: ["+RTS","-K2048m","-RTS"]
    fig_width: 7
    fig_height: 7
    fig_caption: true
    highlight: kate
    lightbox: true
    thumbnails: true
    gallery: true
    cards: true
    use_bookdown: false
    mathjax: null
    self_contained: false
---

```{r knitr_init, echo=FALSE, cache=FALSE}
library(knitr)
library(rmdformats)

# Global options
options(max.print="75")
opts_chunk$set(
    echo=FALSE,
    cache=FALSE,
    prompt=FALSE,
    tidy=TRUE,
    comment=NA,
    message=FALSE,
    warning=FALSE,
    results="asis",
    eval=TRUE
)
opts_knit$set(width=75)
```

```{r load_js}
if (offlineReport) {
    #cat("<script type=\"text/javascript\" src=\"js/pace.min.js\"></script>",
    #   sep="")
    cat("<script type=\"text/javascript\" src=\"js/highcharts.js\"></script>",
        sep="")
    cat("<script type=\"text/javascript\" src=\"js/highcharts-more.js\">",
        "</script>",sep="")
    cat("<script type=\"text/javascript\" src=\"js/exporting.js\"></script>",
        sep="")
    cat("<script type=\"text/javascript\" src=\"js/offline-exporting.js\">",
        "</script>",sep="")
    cat("<script type=\"text/javascript\" src=\"js/export-data.js\"></script>",
        sep="")
    cat("<script type=\"text/javascript\" src=\"js/canvas2svg.js\"></script>",
        sep="")
    cat("<script type=\"text/javascript\" src=\"js/jvenn.min.js\"></script>",
        sep="")
    if (reportDb == "sqlite") {
        cat("<script>Module = { TOTAL_MEMORY: 134217728 };</script>")
        cat("<script type=\"text/javascript\" src=\"js/sql.js\"></script>",
            sep="")
    } else if (reportDb == "dexie") {
       cat("<script type=\"text/javascript\" src=\"js/dexie.min.js\"></script>",
            sep="")
    }
} else {
    #cat("<script type=\"text/javascript\" src=\"https://raw.github.com/",
    #   "HubSpot/pace/v1.0.0/pace.min.js\"></script>",sep="")
    cat("<script type=\"text/javascript\" src=\"https://code.highcharts.com/",
        "highcharts.js\"></script>",sep="")
    cat("<script type=\"text/javascript\" src=\"https://code.highcharts.com/",
        "highcharts-more.js\"></script>",sep="")
    cat("<script type=\"text/javascript\" src=\"https://code.highcharts.com/",
        "modules/exporting.js\"></script>",sep="")
    cat("<script type=\"text/javascript\" src=\"https://code.highcharts.com/",
        "modules/offline-exporting.js\"></script>",sep="")
    cat("<script type=\"text/javascript\" src=\"https://code.highcharts.com/",
        "modules/export-data.js\"></script>",sep="")
    cat("<script type=\"text/javascript\ ",
        "src=\"http://jvenn.toulouse.inra.fr/app/js/canvas2svg.js\">",
        "</script>",sep="")
    cat("<script type=\"text/javascript\" ",
        "src=\"http://jvenn.toulouse.inra.fr/app/js/jvenn.min.js\">",
        "</script>",sep="")
    if (reportDb == "sqlite") {
        cat("<script>Module = { TOTAL_MEMORY: 134217728 };</script>")
        cat("<script type=\"text/javascript\" src=\"",
        "https://cdnjs.cloudflare.com/ajax/libs/sql.js/0.5.0/js/sql.js\">",
        "</script>",sep="")
    } else if (reportDb == "dexie") {
        cat("<script type=\"text/javascript\" ",
            "src=\"https://unpkg.com/dexie@2.0.4/dist/dexie.min.js\">",
            "</script>",sep="")
    }
}
if (reportDb == "dexie")
    cat("<script type=\"text/javascript\" src=\"data/reportdb.js\"></script>")
```

```{js pace_opts}
Pace.options = {
    ajax: false,
    restartOnPushState: false,
    restartOnRequestAfter: false
}
```

<div id="loader" class="loading"></div>

```{r decide_dexie}
if (reportDb == "dexie") {
    code <- paste0(
        "var DEXIE = true;\n",
        "var db = new Dexie('plotdb');\n",
        "db.version(1).stores({\n",
        "    plots: '++_id,[name+type+subtype]'\n",
        "});\n",
        "db.plots.clear().then(console.log('Previous plots cleared!'));\n",
        "db.plots.bulkAdd(plotData);"
    )
} else if (reportDb == "sqlite") {
    code <- "var DEXIE = false;"
}
cat("<script>\n",code,"\n</script>")
```

```{js define_getplot_funs}
getPlotFromDb = function(query,renderId) {
    if (DEXIE) {
        $("#loader").show();
        db.plots.get({
            name: query.name,
            type: query.type,
            subtype: query.subtype
        },function(res) {
            var chartData = res.json;
            chartData.chart.renderTo = renderId;
            chartData.chart.events = {
                load: function() {
                    $("#loader").hide();
                }
            };
            var chart = new Highcharts.Chart(chartData);
        });
    }
    else {
        var xhr = new XMLHttpRequest();
        xhr.open('GET', './data/reportdb.sqlite',true);
        xhr.responseType = 'arraybuffer';
        xhr.onload = function(e) {
            var sqlQuery = "SELECT json FROM plot WHERE name='" + query.name + 
                "'" +" AND type='" + query.type + "'" + " AND subtype='" + 
                query.subtype + "';";
            var uInt8Array = new Uint8Array(this.response);
            var db = new SQL.Database(uInt8Array);
            var contents = db.exec(sqlQuery);
            var chartData = JSON.parse(contents[0].values[0][0]);
            //console.log(chartData)
            chartData.chart.renderTo = renderId;
            chartData.chart.events = {
                load: function() {
                    $("#loader").hide();
                }
            };
            var chart = new Highcharts.Chart(chartData);
        };
        xhr.send();
        $("#loader").show();
    }
}

getPlotWithFunFromDb = function(query,renderId) {
    if (DEXIE) {
        $("#loader").show();
        db.plots.get({
            name: query.name,
            type: query.type,
            subtype: query.subtype
        },function(res) {
            var chartData = res.json;
            chartData.chart.renderTo = renderId;
            
            // Deserialize the stored functions - only in a couple of cases
            if (query.type === "countsbio") {
                chartData.yAxis.labels.formatter = new Function('return ' +
                    chartData.yAxis.labels.formatter)()
                    
                chartData.plotOptions.scatter.tooltip.pointFormatter = 
                    new Function('return ' +
                        chartData.plotOptions.scatter.tooltip.pointFormatter)()
                
                chartData.plotOptions.boxplot.events.legendItemClick = 
                    new Function('return ' + 
                        chartData.plotOptions.boxplot.events.legendItemClick)()
            }
            if (query.type === "boxplot") {
                chartData.plotOptions.boxplot.tooltip.pointFormatter = 
                    new Function('return ' +
                        chartData.plotOptions.boxplot.tooltip.pointFormatter)()
                
                chartData.plotOptions.boxplot.events.legendItemClick = 
                    new Function('return ' + 
                        chartData.plotOptions.boxplot.events.legendItemClick)()
            }
            
            chartData.chart.events = {
                load: function() {
                    $("#loader").hide();
                }
            };
            var chart = new Highcharts.Chart(chartData);
        });
    }
    else {
        var xhr = new XMLHttpRequest();
        xhr.open('GET', './data/reportdb.sqlite',true);
        xhr.responseType = 'arraybuffer';
        xhr.onload = function(e) {
            var sqlQuery = "SELECT json FROM plot WHERE name='" + query.name + 
                "'" + " AND type='" + query.type + "'" + " AND subtype='" + 
                query.subtype + "';";
            var uInt8Array = new Uint8Array(this.response);
            var db = new SQL.Database(uInt8Array);
            var contents = db.exec(sqlQuery);
            var chartData = JSON.parse(contents[0].values[0][0]);
            
            // Deserialize the stored functions - only in a couple of cases
            if (query.type === "countsbio") {
                chartData.yAxis.labels.formatter = new Function('return ' +
                    chartData.yAxis.labels.formatter)()
                    
                chartData.plotOptions.scatter.tooltip.pointFormatter = 
                    new Function('return ' +
                        chartData.plotOptions.scatter.tooltip.pointFormatter)()
                
                chartData.plotOptions.boxplot.events.legendItemClick = 
                    new Function('return ' + 
                        chartData.plotOptions.boxplot.events.legendItemClick)()
            }
            if (query.type === "boxplot") {
                chartData.plotOptions.boxplot.tooltip.pointFormatter = 
                    new Function('return ' +
                        chartData.plotOptions.boxplot.tooltip.pointFormatter)()
                
                chartData.plotOptions.boxplot.events.legendItemClick = 
                    new Function('return ' + 
                        chartData.plotOptions.boxplot.events.legendItemClick)()
            }
            
            chartData.chart.renderTo = renderId;
            chartData.chart.events = {
                load: function() {
                    $("#loader").hide();
                }
            };
            var chart = new Highcharts.Chart(chartData);
        };
        xhr.send();
        $("#loader").show();
    }
}

getVennFromDb = function(query,renderId) {
    if (DEXIE) {
        $("#loader").show();
        db.plots.get({
            name: query.name,
            type: query.type,
            subtype: query.subtype
        },function(res) {
            var chartData = res.json;
            chartData.fnClickCallback = function() {
                var value = '';
                if (this.listnames.length == 1) {
                    value += 'Elements only in ';
                } else {
                    value += 'Common elements in ';
                }
                for (name in this.listnames) {
                    value += this.listnames[name] + ' ';
                }
                value += ':\n';
                for (val in this.list) {
                    value += this.list[val] + '\n';
                }
                if (query.subtype === "stat_dereg" 
                    || query.subtype === "stat_up"
                    || query.subtype === "stat_down") {
                    $('#jvenn_stat_list').text(value);
                }
                if (query.subtype === "fold_dereg" 
                    || query.subtype === "fold_up"
                    || query.subtype === "fold_down") {
                    $('#jvenn_fold_list').text(value);
                }
            }
            $('#'+renderId).jvenn(chartData);
            $("#loader").hide();
        });
    }
    else {
        var xhr = new XMLHttpRequest();
        xhr.open('GET', './data/reportdb.sqlite',true);
        xhr.responseType = 'arraybuffer';
        xhr.onload = function(e) {
            var sqlQuery = "SELECT json FROM plot WHERE name='" + query.name + 
                "'" +" AND type='" + query.type + "'" + " AND subtype='" + 
                query.subtype + "';";
            
            var uInt8Array = new Uint8Array(this.response);
            var db = new SQL.Database(uInt8Array);
            var contents = db.exec(sqlQuery);
            var chartData = JSON.parse(contents[0].values[0][0]);
            chartData.fnClickCallback = function() {
                var value = '';
                if (this.listnames.length == 1) {
                    value += 'Elements only in ';
                } else {
                    value += 'Common elements in ';
                }
                for (name in this.listnames) {
                    value += this.listnames[name] + ' ';
                }
                value += ':\n';
                for (val in this.list) {
                    value += this.list[val] + '\n';
                }
                if (query.subtype === "stat_dereg" 
                    || query.subtype === "stat_up"
                    || query.subtype === "stat_down") {
                    $('#jvenn_stat_list').text(value);
                }
                if (query.subtype === "fold_dereg" 
                    || query.subtype === "fold_up"
                    || query.subtype === "fold_down") {
                    $('#jvenn_fold_list').text(value);
                }
            }
            $('#'+renderId).jvenn(chartData);
            $("#loader").hide();
        };
        xhr.send();
        $("#loader").show();
    }
}
```

<style>
p {
    margin: 0 0 10px;
    text-align: justify;
}

.detable-container {
    font-size: 0.8em;
}

.header-panel {
    background-color: #0E2B8A;
    min-height: 144px;
    position: relative;
    z-index: 3;
}

.header-panel h1.subtitle {
    font-size: 16px;
    font-weight: 600;
}

.figure-hint {
    text-align: justify;
    font-size: 0.9em;
}

.hc-rect-sm {
    width: 640px;
    height: 640px;
    margin: auto;
}

.hc-rect-xs {
    width: 320px;
    height: 320px;
    margin: auto;
}

.hc-selector-label {
    color: #0030CE;
    font-size: 0.9em;
    font-weight: 600;
}

.deheatmap-container {
    width: 740px;
    height: 860px;
    text-align: center;
    margin: 25px auto 5px;
}

.coheatmap-container {
    width: 640px;
    height: 640px;
    text-align: center;
    margin: auto;
}

.link-box {
    background-color: #D5EEEB; 
    border-style: solid; 
    border-color: #00887A; 
    border-width: 1px; 
    width: 80%; 
    padding: 10px;
    margin-top: 20px;
    -moz-border-radius: 5px;
    -webkit-border-radius: 5px;
    -khtml-border-radius: 5px;
    border-radius: 5px;
}

.hidden-element {
    visibility: hidden;
    width: 10px;
    height: 10px;
}

.jvenn_list {
    font-size: 0.9em;
    font-wight: 600;
    width: 100%;
    height: 300px;
}

#summary_log_container {
    font-family: "Courier New", Courier, monospace;
    font-size: 0.9em;
    height: 400px;
    overflow-y: scroll;
    overflow-x: hidden;
    overflow-wrap: normal;
}

/******************************* Chart loading ********************************/
.loading {
  display: none;
  position: fixed;
  z-index: 999;
  height: 100%;
  width: 100%;
  overflow: show;
  margin: auto;
  top: 0;
  left: 0;
  bottom: 0;
  right: 0;
  background: radial-gradient(rgba(76, 51, 204, 0.7), rgba(0, 0, 0, .8));
  background: -webkit-radial-gradient(rgba(76, 51, 204, 0.7), rgba(0, 0, 0,.8));
}

.loading:before {
  content: 'Loading plots...';
  font-size: 92px;
  font-weight: 400;
  color: white;
  display: block;
  position: fixed;
  top: 40%;
  left: 35%;
  width: 100%;
  height: 100%;
}

/********************************** Pace.js ***********************************/
.pace {
  -webkit-pointer-events: none;
  pointer-events: none;

  -webkit-user-select: none;
  -moz-user-select: none;
  user-select: none;
}

.pace.pace-inactive .pace-progress {
  display: none;
}

.pace .pace-progress {
  position: fixed;
  z-index: 2000;
  top: 50%;
  left: 50%;
  height: 240px;
  width: 240px;
  margin-top: -120px; /* Negative half of height. */
  margin-left: -220px; /* Negative half of width. */

  -webkit-transform: translate3d(0, 0, 0) !important;
  -ms-transform: translate3d(0, 0, 0) !important;
  transform: translate3d(0, 0, 0) !important;
}

.pace .pace-progress:after {
  display: block;
  position: absolute;
  top: 0;
  content: attr(data-progress-text);
  font-family: "Helvetica Neue", sans-serif;
  font-weight: 200;
  font-size: 240px;
  line-height: 1;
  text-align: right;
  color: rgba(0, 0, 220, 0.3);
}

.pace .pace-progress:before {
  display: block;
  position: absolute;
  top: 0;
  content: url(media/dna_loader.gif);
  margin-left: -300px;
  margin-top: -120px;
}

.pace-running > *:not(.pace) {
  opacity:0;
}
/******************************** End Pace.js *********************************/
</style>

```{r report_init}
library(DT)
library(GenomicRanges)
library(gplots)
library(heatmaply)
library(htmltools)
library(magrittr)
library(pander)

makeFoldChange <- metaseqR2:::makeFoldChange
nat2log <- metaseqR2:::nat2log
asClassVector <- metaseqR2:::asClassVector
makeReportMessages <- metaseqR2:::makeReportMessages
.formatForReport <- metaseqR2:::.formatForReport

# Initialize some variables used throughout the report
samples <- unlist(sampleList,use.names=FALSE)
biotypes <- unique(as.character(geneData$biotype))
reportMessages <- makeReportMessages("en")

qualPlots <- c("mds","biodetection","countsbio","saturation","readnoise",
    "correl","pairwise","filtered")
normPlots <- c("boxplot","gcbias","lengthbias","meandiff","meanvar","rnacomp")
statPlots <- c("deheatmap","volcano","mastat","biodist","statvenn","foldvenn",
    "deregulogram")

covarsRaw <- covarsStat <- NULL
if (any(qcPlots %in% c("biodetection","countsbio","saturation","rnacomp",
    "readnoise"))) {
    covarsRaw <- list(
        data=geneCounts,
        length=width(geneData),
        gc=as.numeric(geneData$gc_content),
        chromosome=data.frame(
            chromosome=as.character(seqnames(geneData)),
            start=start(geneData),
            end=end(geneData)
        ),
        factors=data.frame(class=asClassVector(sampleList)),
        biotype=as.character(geneData$biotype),
        gene_name=as.character(geneData$gene_name)
    )
}

if ("biodist" %in% qcPlots) {
    covarsStat <- list(
        data=normGenesExpr,
        length=width(geneDataExpr),
        gc=as.numeric(geneDataExpr$gc_content),
        chromosome=data.frame(
            chromosome=as.character(seqnames(geneDataExpr)),
            start=start(geneDataExpr),
            end=end(geneDataExpr)
        ),
        factors=data.frame(class=asClassVector(sampleList)),
        biotype=as.character(geneDataExpr$biotype),
        gene_name=as.character(geneDataExpr$gene_name)
    )
}
```

# Instructions {.tabset .tabset-pills}

Welcome to the **metaseqR2** report! If you are familiar with the metaseqR 
report, then you will find that there are not many differences with respect to 
the presented information. Some diagnostic and exploration plots were added. The 
most notable difference is that __*all*__ plots are interactive. This helps a 
lot with exploration and interpretation but also adds a lot of computational
burden. However, relatively modern systems with recent browser versions should
be capable of rendering all the graphics. The metaseqR2 report has been tested 
with Google Chrome, Mozilla Firefox and Microsoft Edge. It has **not** been 
tested with Internet Explorer, Opera and Safari and most probably will not be. 
Other Chromium browsers (e.g. Brave) should also be fine.

One particular characteristic of the metaseqR2 report is that all plots are
interactive. This is achieved by using the standard graphics underlying data
with libraries including [Highcharts](https://www.highcharts.com/),
[Plotly](https://plot.ly/) and 
[jvenn](http://jvenn.toulouse.inra.fr/app/index.html) to create more 
user-friendly and directly explorable plots. Instructions on the usage of these 
plots follow:

* All plots are interactively **explorable**. This means that if you move your 
mouse **inside** the plot area (a move called *mouse-over*), you can retrieve 
information on each single data point. This applies to all plots. More 
specifically:
  + In **scatterplots**, if you mouse-over each point, information about this 
  point is presented, depending on the type of the plot. The data series from 
  which the point comes is also presented. For example, in a Volcano plot, fold 
  change and significance, as well as the name of the gene and the data category
  (e.g. up-regulated) will be presented.
  + In **barplots**, if you mouse-over each bar, information about this bar is 
  presented, such as the value it represents and the data series from which it
  comes. If the barplot contains groups of bars, then information about each 
  group is displayed. For example, in a Biodetection plot, each bar group
  presents the percentage of a biotype in the examined genome, the percentage
  in the sample and the detected percentage according to read counts.
  + In **boxplots**, if you mouse-over the boxes, the information about the 
  underlying distribution is displayed (maximum, upper quartile, median, lower
  quartile and minimum) as well as the data series. If you mouse-over an 
  outlier, then information on this single point is presented (e.g. value).
  + Some **barplots** have a **double** y-axis system corresponding to 
  **different** measurements or scales. For example in Biodetection barplots,
  the left y-axis presents abundant features while the right y-axis presents
  non-abundant features. In the Filtered barplot, y-axes present different 
  values (numbers and fractions).
  + **Line plots** can be moused-over too. Depending on the plot type, exact 
  values may or may not be shown, depending on how important it is to display 
  them, and to avoid over-crowding the plots. For example in Reads noise plot,
  we are interested in the trend and not so much in exact values.
  + **Heatmaps** can be moused-over too. Information on each heatmap cell will 
  be displayed.
* All scatterplots and heatmaps are **zoomable**. You need to press the left 
mouse button inside a plot area and draw a square area to zoom-in. If you wish 
to reset the zoom, there is a **button** appearing for this when zooming-in.
* Data series in scatterplots, barplots and boxplots can be **toggled** on or 
off by **clicking** on the **legend** name of each data series which is placed 
**below** each plot. For example, in Volcano plots, if you click on the name 
"Unregulated", then the respective data series will stop appearing in the plot. 
You can bring it **back** by clicking the legend again.
* All plots are **exportable**. On the **top right** corner of each scatterplot, 
barplot and boxplot, there is a menu button with several functionalities, 
including **exporting** in various formats and presenting the plot in 
**full-screen** mode. For heatmaps, this functionality is offered by a set of 
small buttons that appear if you mouse-over at the **top** of the heatmap.
* In Venn diagrams, if you **click** on the **number** for each category, the 
respective gene/transcript names will appear in the **box** on the right of the 
diagram.
* All plots can be **downloaded** in static formats (in formats according to 
```metaseqr2``` call) from the Results section.

The metaseqR2 report contains the sections described below, depending on which 
diagnostic and exploration plots have been asked for from the run command. As 
plots are categorized, if no plot from a specific category is asked for, then 
this category will not appear. Below, are the categories:

## Summary

The **Summary** section is further categorized in several subsections. 
Specifically:

* **Analysis summary:** This section contains an auto-generated text that 
analytically describes the computational process followed and summarized the
results of each step. This text can be used as is or with slight modifications
in the <u>Methods</u> section of an article.
* **Input options:** This section provides a list of the input arguments to the
pipeline in a more human-readable format.
* **Filtering:** This section reports in detail the number of filtered genes
decomposed according to the number of genes removed by each applied filter.
* **Differential expression:** This section reports in detail the number of
differentially expressed genes for each contrast, both when using only a p-value
cutoff as well as an FDR cutoff (numbers in parentheses), that is, genes passing
the multiple testing correction procedure selected. These numbers are also
calculated based on a simple fold change cutoff in log<sub>2</sub> scale. 
Finally, when multiple algorithms are used with p-value combination, this
section reports all the findings analytically per algorithm.
* **Command:** This section contains the command used to run the ```metaseqr2```
pipeline for users that want to experiment as well as a critical messages
displayed within the ```R``` session running ```metaseqr2``` displayed as a log.
Finally, if a targets file has been used to perform the analysis, a table
depicting the parameters in the targets files is created and a link to download
the actual targets file, but any relative paths to BAM files are stripped and
the user is responsible to prepend them if the targets file has to be reused in
another location, e.g. locally.
* **Tracks:** This section contains a link which opens a new window to the UCSC
Genome Browser where normalized tracks based on the input BAM files are 
displayed. If stranded tracks have been requested (according to the
sequencing protocol or technology), the a track hub is created to display the
stranded tracks. From this tab, you can also download bigWig files as well as 
copy track lines for manual input to the UCSC Genome Browser.

## Quality control 

The **Quality** control section contains several interactive plots concerning 
the overall quality control of each sample provided as well as overall 
assessments. The quality control plots are the *Multidimensional Scaling* 
(**MDS**) plot, the *Biotypes detection* (**Biodetection**) plot, the 
*Biotype abundance* (**Countsbio**) plot, the *Read saturation* (**Saturation**) 
plot, the *Read noise* (**ReadNoise**) plot, the *Correlation heatmap* 
(**Correlation**), the *Pairwise sample scatterplots* (**Pairwise**) and the 
*Filtered entities* (**Filtered**) plot. Each plot is accompanied by a detailed
description of what it depicts. Where multiple plot are available (e.g. one for
each sample), a selection list on the top of the respective  section allows the
selection of the sample to be displayed.

## Normalization

The **Normalization** section contains several interactive plots that can be 
used to inspect and assess the normalization procedure. Therefore, normalization
plots are usually paired, showing the same data instance normalized and not
normalized. The normalization plots are the *Expression boxplots* (**Boxplots**)
plots, the *GC content bias* (**GC bias**) plots, the *Gene length bias* 
(**Length bias**) plots, the *Within condition mean-difference*
(**Mean-Difference**) plots, the *Mean-variance relationship* 
(**Mean-Variance**) plot and the *RNA composition* (**Rna composition**) plot.
Each plot is accompanied by a detailed description of what it depicts. Where 
multiple plot are available (e.g. one for each sample), a selection list on the 
top of the respective section allows the selection of the sample to be 
displayed.

## Statistics

The **Statistics** section contains several interactive plots that can be used
to inspect and explore the outcome of statistical testing procedures. The 
statistics plots are the *Volcano plot* (**Volcano**), the 
*MA or Mean-Difference across conditions* (**MA**) plot, the 
*Expression heatmap* (**Heatmap**) plot, the 
*Chromosome and biotype distributions* (**Biodist**) plot, the
*Venn diagram across statistical tests* (**StatVenn**), the 
*Venn diagram across contrasts* (**FoldVenn**) and the *Deregulogram*. Each plot 
is accompanied by a detailed description of what it depicts. Please note that 
the heatmap plots only show the top percentage of differentially expressed genes 
as this is controlled by the ```reportTop``` parameter of the ```metaseqr2``` 
pipeline. When multiple plots are available (e.g. one for each contrast), a 
selection list on the top of the respective section allows the selection of the 
sample to be displayed.

## Results

The **Results** section contains a snapshot of differentially expressed genes in 
table format with basic information about each gene and links to external 
resources. Certain columns of the table are colored according to significance. 
Larger bars and more intense colors indicate higher significance. For example, 
the bar in the *p_value* column is larger if the genes has higher statistical 
significance and the fold change cell background is bright red if the gene is 
highly up-regulated. From the **Results** section, full gene lists can be 
downloaded in text tab-delimited format and viewed with a spreadsheet
application such as MS Excel. A selector on the top of the section above the 
table allows the display of different contrasts.

## References

The **References** section contains bibliographical references regading the
algorithms used by the ```metaseqr2``` pipeline and is adjusted according to the
algorithms selected.
 
# Summary {.tabset .tabset-pills}

## Analysis summary

### Analysis summary

```{r summary_analysis_summary}
if (fromRaw) {
    cat("The raw ",fileType," files, one for each RNA-Seq sample, were ",
        "summarized to ")
    if (transLevel == "gene") {
        if (countType=="exon") {
            cat("an exon ") 
        } else if (countType=="gene") {
            cat("a gene ")
        } else if (countType=="utr") {
            cat("a 3'UTR ")
        }
    } else if (transLevel=="transcript") {
        if (countType=="exon") {
            cat("an exon ")
        } else if (countType=="utr") {
            cat("a 3'UTR ")
        }
    } else if (transLevel=="exon") {
        if (countType=="exon")
            cat("an exon ")
    }
    cat("read counts table, ")
    if (fileType=="bam" || fileType=="sam") {
        cat("using the Bioconductor package GenomicRanges. ")
    } else if (fileType=="bed") {
        cat("using the Bioconductor packages rtracklayer and GenomicRanges. ")
    }
    cat("In the final read counts table, each row represented ")
    if (countType=="exon") {
        cat("one exon, ") 
    } else if (countType=="gene") {
        cat("one gene, ")
    }
    cat("each column one RNA-Seq sample and each cell, the corresponding ",
        "read counts associated with each row and column.")
}

if (countType == "exon") {
    if (!is.null(exonFilters)) {
        cat("The exon read counts were filtered for artifacts that could",
            "affect the subsequent normalization and statistical testing",
            "procedures as follows: ")
        msgExonFiltersMinActiveExons <- NULL
        if (!is.null(exonFilters$minActiveExons)) {
            msgExonFiltersMinActiveExons <- paste(
                "if an annotated ",transLevel," had up to ", 
                exonFilters$minActiveExons$exonsPerGene," exons, read ",
                "presence was required in at least ",
                exonFilters$minActiveExons$minExons," of the exons, else if ", 
                "an annotated ",transLevel," had more than ",
                exonFilters$minActiveExons$exonsPerGene," exons, then read ",
                "presence was required in at least ",
                exonFilters$minActiveExons$frac,
                "x&lceil;E&rceil; exons, where &lceil;<sup>.</sup>&rceil; is ",
                "the <em>ceiling</em> mathematical function. The application ",
                "of this filter resulted in the exclusion of ",
                length(exonFilterResult$minActiveExons)," ",transLevel,
                "s from further analysis. ",sep=""
            )
        }
        cat(msgExonFiltersMinActiveExons,sep=", ")
        cat("The total number of ",transLevel,"s excluded due to the ",
            "application of exon filters was ",
             sum(as.numeric(sapply(exonFilterResult,length))),". ",sep="")
    }
    cat("The final read counts for each ",transLevel," model were calculated",
        "as the sums of their exon reads, creating a ",transLevel," counts",
        "table where each row corresponded to an Ensembl",transLevel,
        " model and each column corresponded to an RNA-Seq sample. ")
}

cat("The ",transLevel," counts table was normalized for inherent systematic ",
    "or experimental biases (e.g. sequencing depth, ",transLevel," length, ",
    "GC content bias etc. using the Bioconductor package ",
     reportMessages$norm[[normalization]]," after removing ",transLevel,
    "s that had zero counts over all the RNA-Seq samples (",length(theZeros),
     " ",transLevel,"s). The output of the normalization algorithm ",sep="")
cat("was a table with normalized counts, which can be used for differential",
    "expression analysis with statistical algorithms developed specifically",
    "for count data. ")
    
if (!is.null(geneFilters)) {
    cat("Prior to the statistical testing procedure, the ",transLevel," read ",
    "counts were filtered for possible artifacts that may affect the ",
    "subsequent statistical testing procedures. Genes/transcripts presenting ",
    "any of the following were excluded from further analysis: ")
    
    latinNumbered <- c("i)","ii)","iii)","iv)","v)","vi)","vii)","viii)","ix)",
        "x)","xi)","xii)","xiii)","xiv)","xv)")
    counter <- 0
    msgGeneFiltersLength <- msgGeneFiltersExpressionMedian <- 
        msgGeneFiltersExpressionMean <- msgGeneFiltersExpressionQuantile <-
        msgGeneFiltersExpressionKnown <- msgGeneFiltersExpressionCustom <-
        msgGeneFiltersBiotype <- msgGeneFiltersAvgReads <- 
        msgGeneFiltersPresence <- NULL
    if (!is.null(geneFilters$length)) {
        counter <- counter + 1
        msgGeneFiltersLength <- paste(
            latinNumbered[counter]," ",transLevel,"s with length less than ", 
            geneFilters$length$length,"bp (",length(geneFilterResult$length),
            " genes)",sep=""
        )
    }
    if (!is.null(geneFilters$avgReads)) {
        counter <- counter + 1
        msgGeneFiltersAvgReads <- paste(
            latinNumbered[counter]," ",transLevel,"s whose average numbers of ",
            "reads per ",geneFilters$avgReads$averagePerBp," bp was less than ",
            "the ",100*geneFilters$avgReads$quantile,"<sup>th</sup> quantile ",
            "of the total normalized distribution of average reads per ",
            geneFilters$avgReads$averagePerBp,"bp (",
            length(geneFilterResult$averagePerBp)," ",transLevel,"s with ",
            "cutoff value ",round(geneFilterCutoff$avgReads,digits=5)," ",
            "average reads per ",geneFilters$avgReads$averagePerBp," bp)",sep=""
        )
    }
    if (!is.null(geneFilters$expression)) {
        if (!is.null(geneFilters$expression$median)
            && geneFilters$expression$median) {
            counter <- counter + 1
            msgGeneFiltersExpressionMedian <- paste(
                latinNumbered[counter]," ",transLevel,"s with read counts ",
                "below the median read counts of the total normalized count ",
                "distribution (",length(geneFilterResult$expression$median)," ",
                transLevel,"s with cutoff value ",
                geneFilterCutoff$expression$median," normalized read counts)",
                sep=""
            )
        }
        if (!is.null(geneFilters$expression$mean) 
            && geneFilters$expression$mean) {
            counter <- counter + 1
            msgGeneFiltersExpressionMean <- paste(
                latinNumbered[counter]," ",transLevel,"s with read counts ",
                "below the mean read counts of the total normalized counts ",
                "distribution (",length(geneFilterResult$expression$mean)," ",
                transLevel,"s with cutoff value ",
                geneFilterCutoff$expression$mean," normalized read counts)",
                sep=""
            )
        }
        if (!is.null(geneFilters$expression$quantile) 
            && !is.na(geneFilters$expression$quantile)) {
            counter <- counter + 1
            msgGeneFiltersExpressionQuantile <- paste(
                latinNumbered[counter]," ",transLevel,"s with read counts ",
                "below the ",100*geneFilters$expression$quantile,
                "<sup>th</sup> quantile of the normalized counts distribution ",
                "(",length(geneFilterResult$expression$quantile)," ",transLevel,
                "s with cutoff value ",geneFilterCutoff$expression$quantile," ",
                "normalized read counts)",sep=""
            )
        }
        if (!is.null(geneFilters$expression$known) 
            && !is.na(geneFilters$expression$known)) {
            counter <- counter + 1
            msgGeneFiltersExpressionKnown <- paste(
                latinNumbered[counter]," ",transLevel,"s with read counts ",
                "below the 90<sup>th</sup> quantile of the counts of the ",
                "following ",transLevel,"s, known to not being expressed from ",
                "the related literature: ",
                paste(geneFilters$expression$known,collapse=", "),
                "(",length(geneFilterResult$expression$known)," ",transLevel,
                "s with cutoff value",geneFilterCutoff$expression$known,
                " normalized read counts)",sep=""
            )
        }
        if (!is.null(geneFilters$expression$custom) 
            && !is.na(geneFilters$expression$custom)) {
            counter <- counter + 1
            msgGeneFiltersExpressionCustom <- paste(
                latinNumbered[counter]," genes not passing the user defined ",
                "filter provided to the metaseqr pipeline (",
                length(geneFilterResult$expression$custom)," ",transLevel,
                "s with cutoff value",geneFilterCutoff$expression$custom,")",
                sep=""
            )
        }
    }
    if (!is.null(geneFilters$biotype)) {
        counter <- counter + 1
        msgGeneFiltersBiotype <- paste(
            latinNumbered[counter]," ",transLevel,"s whose biotype matched ",
            "the following: ",
            paste(
                names(geneFilters$biotype)[which(unlist(geneFilters$biotype))],
                collapse=", "
            ),
            " (",length(geneFilterResult$biotype)," ",transLevel,"s)",sep=""
        )
    }
    if (!is.null(geneFilters$presence)) {
        counter <- counter + 1
        msgGeneFiltersPresence <- paste(
            latinNumbered[counter]," ",transLevel,"s which in ",
            geneFilters$presence$frac*100,"% of samples did not exceed ",
            geneFilters$presence$minCount," counts (",
            length(geneFilterResult$presence)," ",transLevel,"s)",sep=""
        )
        if (geneFilters$presence$perCondition) {
            msgGeneFiltersPresence <- 
                paste(msgGeneFiltersPresence," condition-wise",sep="")
        } else {
            msgGeneFiltersPresence <- 
                paste(msgGeneFiltersPresence," across all samples",sep="")
        }
    }
    
    cat(msgGeneFiltersLength,msgGeneFiltersAvgReads,
        msgGeneFiltersExpressionMedian,msgGeneFiltersExpressionMean,
        msgGeneFiltersExpressionQuantile,msgGeneFiltersExpressionKnown,
        msgGeneFiltersExpressionCustom,msgGeneFiltersBiotype,
        msgGeneFiltersPresence,sep=", ")
    cat(". The total number of ",transLevel,"s excluded due to the ",
        "application of ",transLevel," filters was ",
        sum(as.numeric(sapply(geneFilterResult,length))),". ",sep="")
    cat("The total (unified) number of ",transLevel,"s  excluded due to the ",
        "application of all filters was ",length(theZeros) + length(theDead),
        ". ",sep="")
}

cat("The resulting ",transLevel," counts table was subjected to differential ",
    "expression analysis for the contrasts ",
    paste(gsub("_vs_"," versus ",contrast),collapse=", "),sep="")
if (length(statistics)>1) {
    cat(" using the Bioconductor packages ",paste(unlist(
        reportMessages$stat[statistics],use.names=FALSE),collapse=", "),". ",
        sep="")
    if (metaP!="none") {
        cat("In order to combine the statistical significance from multiple ",
            "algorithms and perform meta-analysis, the ")
            cat(reportMessages$meta[[metaP]],"method was applied. ")
    }
} else {
    cat(" using the Bioconductor package ",reportMessages$stat[[statistics]],
        ". ",sep="")
}

if (!is.na(pcut))
    if (pcut==1) plasm <- 0.05 else plasm <- pcut

if (!is.null(contrast)) {
    cat("The final numbers of differentially expressed ",transLevel,"s were ",
        "(per contrast): ",sep="")
    msgContrast <- character(length(contrast))
    names(msgContrast) <- contrast

    for (cnt in contrast) {
        if (!is.na(pcut) && length(which(sumpList[[cnt]]<plasm))==0) {
            msgContrast[cnt] <- paste(
                "for the contrast ",gsub("_vs_"," versus ",cnt),
                ", no differentially expressed ",transLevel,"s were found",
                " with a p-value threshold of ",plasm,sep=""
            )
        } else if (is.na(pcut)) {
            msgContrast[cnt] <- paste(
                "for the contrast ",gsub("_vs_"," versus ",cnt),
                ", no statistical threshold defined",sep=""
            )
        } else {
            if (adjustMethod != "none") {
                addTextP <- paste(
                    " (",length(which(p.adjust(sumpList[[cnt]],
                    adjustMethod)<plasm)),")",sep=""
                )
                hasFdrText <- " (FDR or adjusted p-value)"
            } else {
                addTextP <- hasFdrText <- NULL
            }
            if (length(strsplit(cnt,"_vs_")[[1]]) == 2) {
                tmp <- log2(makeFoldChange(cnt,sampleList,normGenesExpr[which(
                    sumpList[[cnt]]<plasm),,drop=FALSE],logOffset))
                if (adjustMethod != "none") {
                    areThere <- which(p.adjust(sumpList[[cnt]],
                        adjustMethod)<plasm)
                    if (length(areThere) > 0) {
                        if (length(areThere)==1) {
                            tmpF <- log2(makeFoldChange(cnt,sampleList,
                                normGenesExpr[areThere,,drop=FALSE],logOffset))
                            addTextFu <- 
                                paste(" (",length(which(tmpF>=1)),")",sep="")
                            addTextFd <- 
                                paste(" (",length(which(tmpF<=-1)),")",sep="")
                            addTextFn <- 
                                paste(" (",length(which(abs(tmpF)<1)),")",
                                    sep="")
                        } else {
                            tmpF <- log2(makeFoldChange(cnt,sampleList,
                                normGenesExpr[areThere,,drop=FALSE],logOffset))
                            addTextFu <- 
                                paste(" (",length(which(tmpF>=1)),")",sep="")
                            addTextFd <- 
                                paste(" (",length(which(tmpF<=-1)),")",sep="")
                            addTextFn <- 
                                paste(" (",length(which(abs(tmpF)<1)),")",
                                    sep="")
                        }
                    } else {
                        addTextFu <- addTextFd <- addTextFn <- NULL
                    }
                } else {
                    addTextFu <- addTextFd <- addTextFn <- NULL
                }
                msgContrast[cnt] <- paste(
                    "for the contrast ",gsub("_vs_"," versus ",cnt),", ",
                    length(which(sumpList[[cnt]]<plasm)),addTextP,
                    " statistically significant ",transLevel,"s were found ",
                    "with a p-value",hasFdrText," threshold of ",plasm," and ",
                    "of these, ",length(which(tmp>=1)),addTextFu,
                    " were up-regulated, ",length(which(tmp<=-1)),addTextFd,
                    " were down-regulated and ",addTextFn,
                    " were not differentially expressed according to an ",
                    "absolute fold change cutoff value of 1 in ",
                    "log<sub>2</sub> scale",sep=""
                )
            } else {
                msgContrast[cnt] <- paste(
                    "for the contrast ",gsub("_vs_"," versus ",cnt),", ",
                    length(which(sumpList[[cnt]]<plasm)),addTextP,
                    " differentially expressed ",transLevel,"s were found ",
                    "with a p-value",hasFdrText," threshold of ",plasm," at ",
                    "least in one condition",sep=""
                )
            }
        }
    }
    cat(paste(msgContrast,collapse=", "))
} else {
    cat("No statistical testing was performed")
}

cat(". Literature references for all the algorithms used can be found at the ",
    "end of this report.")
```

## Input options

### Input options

**Read counts file: **
```{r summary_input_options_1}
cat(countsName)
```

**Conditions:** 
```{r summary_input_options_2}
cat(paste(names(sampleList),collapse=", "))
```

**Samples included:** 
```{r summary_input_options_3}
cat(paste(unlist(sampleList),collapse=", "))
```

**Samples excluded:** 
```{r summary_input_options_4}
if (!is.null(excludeList) && !is.na(excludeList)) {
    cat(paste(unlist(excludeList),collapse=", ")) 
} else {
    cat("none")
}

```

**Requested contrasts:** 
```{r summary_input_options_5}
if (!is.null(contrast)) {
    cat(paste(contrast,collapse=", "))
} else {
    cat("No contrasts were defined.")
}
```

**Library sizes:**
```{r summary_input_options_6}
if (!is.null(libsizeList)) {
    cat("<ul>")
    for (n in names(libsizeList)) 
        cat("<li>",paste("<em>",n,"</em>: ",libsizeList[[n]],sep=""),"</li>")
    cat("</ul>")
} else {
    cat("not available")
}
```

**Organism:** 
```{r summary_input_options_7}
cat(reportMessages$org[[org]])
```

**Annotation source:** 
```{r summary_input_options_8}
cat(reportMessages$refdb[[refdb]])
```

**Count type:** 
```{r summary_input_options_9}
cat(countType)
```

```{r summary_input_options_10}
if (countType=="utr" && !is.null(utrOpts)) {
    cat("<strong>3' UTR fraction:</strong> ",utrOpts$frac,"<br/>")
    cat("<strong>3' UTR minimum length:</strong> ",utrOpts$minLength,"bps<br/>")
    cat("<strong>3' UTR downstream:</strong> ",utrOpts$downstream,"bps<br/>")
}
```

**Exon filters:**
```{r summary_input_options_11}
if (!is.null(exonFilters)) {
    cat(paste(names(exonFilters),collapse=", "),"<br/>")
    for (ef in names(exonFilters)) {
        cat("<ul>")
        cat("<li><em><span style=\"font-size:1em\">",ef,"</span></em><ul>",
            sep="")
        for (efp in names(exonFilters[[ef]])) {
            if (length(exonFilters[[ef]][[efp]]) == 1 
                && is.function(exonFilters[[ef]][[efp]])) {
                cat("<li>custom function</li>")
            } else if (length(exonFilters[[ef]][[efp]]) == 1) {
                cat("<li>",paste(efp,exonFilters[[ef]][[efp]],sep=": "),
                    "</li>",sep="")
            } else if (length(exonFilters[[ef]][[efp]]) > 1) {
                cat("<li>",paste(efp,paste(exonFilters[[ef]][[efp]],
                    collapse=", "),sep=": "),"</li>",sep="")
            }
        }
        cat("</ul></li></ul>")
    }
} else {
    cat("none applied")
}
```

```{r summary_input_options_12}
if (!is.null(preset))
    cat("<strong>Analysis preset:</strong>",reportMessages$preset[[preset]])
```

**Gene filters:**
```{r summary_input_options_13}
if (!is.null(geneFilters)) {
    cat(paste(names(geneFilters),collapse=", "),"<br/>")
    for (gf in names(geneFilters)) {
        cat("<ul>")
        cat("<li><em><span style=\"font-size:1em\">",gf,"</span></em><ul>",
            sep="")
        for (gfp in names(geneFilters[[gf]])) {
            if (length(geneFilters[[gf]][[gfp]]) == 1 
                && is.function(geneFilters[[gf]][[gfp]])) {
                cat("<li>custom function</li>")
            } else if (length(geneFilters[[gf]][[gfp]]) == 1) {
                cat("<li>",paste(gfp,geneFilters[[gf]][[gfp]],sep=": "),"</li>",
                    sep="")
            } else if (length(geneFilters[[gf]][[gfp]]) > 1) {
                cat("<li>",paste(gfp,paste(geneFilters[[gf]][[gfp]],
                    collapse=", "),sep=": "),"</li>",sep="")
            }
        }
        cat("</ul></li></ul>")
    }
} else {
    cat("none applied")
}
```

**Filter application:** 
```{r summary_input_options_14}
cat(reportMessages$whenfilter[[whenApplyFilter]])
```

**Normalization algorithm:** 
```{r summary_input_options_15}
cat(reportMessages$norm[[normalization]])
```

**Normalization arguments:**
```{r summary_input_options_16}
if (!is.null(normArgs)) {
    if (normalization == "each") {
        for (n in names(normArgs)) {
            cat("<strong>Statistical arguments for",
                reportMessages$norm[[n]],": </strong>",
                paste(names(normArgs[[n]]),collapse=", "))
            if (length(normArgs[[n]]) > 0) {
                cat("<ul>")
                for (na in names(normArgs[[n]])) {
                    if (length(normArgs[[n]][[na]]) == 1 
                        && is.function(normArgs[[n]][[na]])) {
                        cat("<li>",na,": ",
                            as.character(substitute(normArgs[[na]])),
                        "</li>",sep="")
                    } else if (length(normArgs[[n]][[na]]) == 1) {
                        cat("<li>",paste(na,normArgs[[n]][[na]],sep=": "),
                            "</li>")
                    } else if (length(normArgs[[n]][[na]]) > 1) {
                        cat("<li>",paste(na,paste(normArgs[[n]][[na]],
                            collapse=", "),sep=": "),"</li>")
                    }
                }
                cat("</ul>")
            } else {
                cat("not available or not required")
            }
        }
    } else {
        cat(paste(names(normArgs),collapse=", "),"<br/>")
        cat("<ul>")
        for (na in names(normArgs)) {
            if (length(normArgs[[na]])==1 && is.function(normArgs[[na]])) {
                cat("<li>",as.character(substitute(normArgs[[na]])),"</li>",
                    sep="")
            } else if (length(normArgs[[na]]) == 1) {
                cat("<li>",paste(na,normArgs[[na]],sep=": "),"</li>")
            } else if (length(normArgs[[na]]) > 1) {
                cat("<li>",paste(na,paste(normArgs[[na]],collapse=", "),
                    sep=": "),"</li>")
            }
        }
        cat("</ul>")
    }
} else {
    cat("not available")
}
```

**Statistical algorithm(s):** 
```{r summary_input_options_17}
if (!any(is.na(statistics))) {
    cat(paste(unlist(reportMessages$stat[statistics],use.names=FALSE),
        collapse=", "))
} else {
    cat("No statitical testing performed.")
}
```

```{r summary_input_options_18}
if (!is.null(statArgs)) {
    for (s in names(statArgs)) {
        cat("<strong>Statistical arguments for ",reportMessages$stat[[s]],
            ": </strong>",paste(names(statArgs[[s]]),collapse=", "),sep="")
        if (length(statArgs[[s]]) > 0) {
            cat("<ul>")
            for (sa in names(statArgs[[s]])) {
                if (length(statArgs[[s]][[sa]]) == 1 
                    && is.function(statArgs[[s]][[sa]])) {
                    cat("<li>",sa,": ",as.character(substitute(statArgs[[na]])),
                        "</li>",sep="")
                } else if (length(statArgs[[s]][[sa]])==1) {
                    cat("<li>",paste(sa,statArgs[[s]][[sa]],sep=": "),"</li>")
                } else if (length(statArgs[[s]][[sa]])>1) {
                    cat("<li>",paste(sa,paste(statArgs[[s]][[sa]],
                        collapse=", "),sep=": "),"</li>")
                }
            }
            cat("</ul>")
        } else {
            cat("not available or not required<br/>")
        }
    }
} else {
    cat("<strong>Statistical arguments not available</strong>")
}
```

**Meta-analysis method:** 
```{r summary_input_options_19}
cat(reportMessages$meta[[metaP]])
```

**Multiple testing correction:** 
```{r summary_input_options_20}
cat(reportMessages$adjust[[tolower(adjustMethod)]])
```

**p-value threshold:** 
```{r summary_input_options_21}
if (!is.na(pcut)) cat(pcut) else cat("not available")
```

**Logarithmic tranformation offset: ** 
```{r summary_input_options_22}
cat(logOffset)
```

**Analysis preset:** 
```{r summary_input_options_23}
if (!is.null(preset)) cat(preset) else cat("not available")
```

**Quality control plots:**
```{r summary_input_options_24}
cat(paste(unlist(reportMessages$plots[qcPlots],use.names=FALSE),collapse=", "))
```

**Figure format:** 
```{r summary_input_options_25}
    cat(paste(figFormat,collapse=", "))
```

**Output directory:** 
```{r summary_input_options_26}
if (!is.na(exportWhere)) cat(exportWhere) else cat("default")
```

**Output data: **
```{r summary_input_options_27}
cat(paste(unlist(reportMessages$export[exportWhat],use.names=FALSE),
    collapse=", "))
```

**Output scale(s):** 
```{r summary_input_options_28}
cat(paste(unlist(reportMessages$export[exportScale],use.names=FALSE),
    collapse=", "))
```

**Output values:** 
```{r summary_input_options_29}
cat(paste(unlist(reportMessages$export[exportValues],use.names=FALSE),
    collapse=", "))
```

**Output statistics:**
```{r summary_input_options_30}
cat(paste(unlist(reportMessages$export[exportStats],use.names=FALSE),
    collapse=", "))
```

**Total run time:**
```{r summary_input_options_31}
cat(execTime)
```

## Filtering

<h3>Filtered 
```{r summary_filtered_what_1}
cat(transLevel,"s",sep="")
``` 
</h3>

**Number of filtered**
```{r summary_filtered_what_2}
cat("<strong>",transLevel,"s: </strong>",sep="")
```
```{r summary_filtered_howmany_1}
cat(length(theZeros) + length(theDead))
```
*which is the **union** of*

* Filtered because of zero reads:
```{r summary_filtered_howmany_2}
cat(length(theZeros))
```
* Filtered because of exon filters:
```{r summary_filtered_howmany_3}
cat(sum(as.numeric(sapply(exonFilterResult,length))))
if (sum(as.numeric(sapply(exonFilterResult,length))) != 0) {
    cat("<em> which is the <strong>union</strong> of</em>")
    cat("<ul>")
    for (n in names(exonFilterResult)) {
        cat("<li>",n,": ",length(exonFilterResult[[n]]),"</li>")
    }
    cat("</ul>")
}
```
* Filtered because of gene filters:
```{r summary_filtered_howmany_4}
cat(length(unique(unlist(geneFilterResult))))
```
*which is the **union** of*
```{r summary_filtered_howmany_5}
cat("<ul>")
for (n in names(geneFilterResult)) {
    if (!is.list(geneFilterResult[[n]])) {
        if (!is.null(geneFilterResult[[n]]))
            cat("<li>",n,": ",length(unlist(geneFilterResult[[n]])),
                " genes with filter cutoff value ",geneFilterCutoff[[n]],
                "</li>",sep="")
    } else {
        cat("<li>",n,": ",length(unlist(geneFilterResult[[n]])),
            " ",transLevel,"s further decomposed to (filter name, filtered ",
            transLevel,"s, filter cutoff):</li>",sep="")
        cat("<ul>")
        for (nn in names(geneFilterResult[[n]])) {
            if (!is.null(geneFilterResult[[n]][[nn]]))
                cat("<li>",nn,": ",length(unlist(geneFilterResult[[n]][[nn]])),
                " ",transLevel,"s with filter cutoff value ",
                paste(geneFilterCutoff[[n]][[nn]],collapse=", "),"</li>",sep="")
        }
        cat("</ul>")
    }
}
cat("</ul>")
```

## Differential expression

<h3>Differentially expressed 
```{r summary_de_what_1}
cat(transLevel,"s",sep="")
``` 
</h3>

**Number of differentially expressed **
```{r summary_de_what_2}
cat("<strong>",transLevel,"s per contrast: </strong>",sep="")
if (!is.null(contrast) && !any(is.na(statistics))) {
    if (!is.na(pcut) && pcut==1)
        cat("The p-value cutoff during the analysis was set to 1 so as to ",
            "retrieve the total ",transLevel," list which passed the filtering ",
            "procedure. Each ",transLevel," in the list is accompanied by its ",
            "statistical scores (p-value, FDR, etc.)")

    cat("<ul>")
    for (cnt in contrast) {
        cat("<li><strong>",cnt,": </strong>",sep="")
        if (!is.na(pcut) && length(which(sumpList[[cnt]]<pcut)) == 0) {
            cat("no differentially expressed ",transLevel,"s","</li>",sep="")
        } else if (is.na(pcut)) {
            cat("no statistical threshold defined","</li>")
        } else {
            if (!is.na(pcut) && pcut==1) {
                plasm <- 0.05
            } else {
                plasm <- pcut
            }
            if (adjustMethod!="none") {
                addTextP <- paste("(",length(which(p.adjust(sumpList[[cnt]],
                    adjustMethod) < plasm)),")",sep="")
                hasFdrText <- "(FDR or adjusted p-value)"
            } else {
                addTextP <- hasFdrText <- NULL
            }
            if (length(strsplit(cnt,"_vs_")[[1]]) == 2) {
                tmpP <- log2(makeFoldChange(cnt,sampleList,normGenesExpr[which(
                    sumpList[[cnt]] < plasm),,drop=FALSE],logOffset))
                if (adjustMethod != "none") {
                    areThere <- which(p.adjust(sumpList[[cnt]],adjustMethod)<plasm)
                    if (length(areThere)>0) {
                        if (length(areThere) == 1) {
                            tmpF <- log2(makeFoldChange(cnt,sampleList,
                                normGenesExpr[areThere,,drop=FALSE],logOffset))
                            addTextFu <- 
                                paste(" (",length(which(tmpF>=1)),")",sep="")
                            addTextFd <- 
                                paste(" (",length(which(tmpF<=-1)),")",sep="")
                            addTextFn <- 
                                paste(" (",length(which(abs(tmpF)<1)),")",sep="")
                        } else {
                            tmpF <- log2(makeFoldChange(cnt,sampleList,
                                normGenesExpr[areThere,,drop=FALSE],logOffset))
                            addTextFu <- 
                                paste(" (",length(which(tmpF>=1)),")",sep="")
                            addTextFd <- 
                                paste(" (",length(which(tmpF<=-1)),")",sep="")
                            addTextFn <- 
                                paste(" (",length(which(abs(tmpF)<1)),")",sep="")
                        }
                    } else {
                        addTextFu <- addTextFd <- addTextFn <- NULL
                    }
                } else {
                    addTextFu <- addTextFd <- addTextFn <- NULL
                }
                cat(length(which(sumpList[[cnt]]<plasm))," ",addTextP,
                    " statistically significant ",transLevel,"s of which ",
                    length(which(tmpP>=1)),addTextFu," up regulated, ",
                    length(which(tmpP<=-1)),addTextFd," down regulated and ",
                    length(which(abs(tmpP)<1)),addTextFn," not differentially ",
                    "expressed according to a p-value ",hasFdrText," threshold of ",
                    plasm," and an absolute fold change cutoff value of 1 in ",
                    "log<sub>2</sub> scale.",sep="")
            } else {
                cat(length(which(sumpList[[cnt]]<plasm))," ",addTextP,
                    " statistically significant, differentially expressed in at",
                    " least one condition at a p-value ",hasFdrText,
                    " threshold of ",plasm,".",sep="")
            }
            if (length(statistics)>1 && metaP != "none") {
                cat(" These numbers refer to the combined analysis performed by ",
                    "metaseqR2. Per statistical algorithm, the differentially ",
                    "expressed ",transLevel,"s are:",sep="")
                cat("<ul>")
                for (s in statistics) {
                    cat("<li><em>",reportMessages$stat[[s]],": </em>",sep="")
                    if (adjustMethod!="none") {
                        addTextP <- paste("(",length(which(p.adjust(
                            cpList[[cnt]][,s],adjustMethod)<plasm)),")",sep="")
                        hasFdrText <- "(FDR or adjusted p-value)"
                    } else {
                        addTextP <- hasFdrText <- NULL
                    }
                    if (length(strsplit(cnt,"_vs_")[[1]]) == 2) {
                        tmpP <- log2(makeFoldChange(cnt,sampleList,
                            normGenesExpr[which(cpList[[cnt]][,s]<plasm),,
                            drop=FALSE],logOffset))
                        if (adjustMethod != "none") {
                            tmpF <- log2(makeFoldChange(cnt,sampleList,
                                normGenesExpr[which(p.adjust(cpList[[cnt]][,s],
                                adjustMethod)<plasm),,drop=FALSE],logOffset))
                            addTextFu <- 
                                paste("(",length(which(tmpF>=1)),")",sep="")
                            addTextFd <- 
                                paste("(",length(which(tmpF<=-1)),")",sep="")
                            addTextFn <- 
                                paste("(",length(which(abs(tmpF)<1)),")",sep="")
                        } else {
                            addTextFu <- addTextFd <- addTextFn <- NULL
                        }
                        cat(length(which(cpList[[cnt]][,s]<plasm))," ",addTextP,
                            " statistically significant ",transLevel,"s of which ",
                            length(which(tmpP>=1))," ",addTextFu," up regulated, ",
                            length(which(tmpP<=-1)),addTextFd," down regulated ",
                            "and ",length(which(abs(tmpP)<1))," ",addTextFn,
                            " not differentially expressed according to a p-value ",
                            hasFdrText," threshold of ",plasm," and an absolute ",
                            "fold change cutoff value of 1 in log<sub>2</sub> ",
                            "scale.",sep="")
                    } else {
                        cat(length(which(cpList[[cnt]][,s]<plasm))," ",addTextP,
                            " statistically significant, differentially expressed ",
                            "in at least one condition at a p-value ",hasFdrText,
                            " threshold of ",plasm,".",sep="")
                    }
                    cat("</li>")
                }
                cat("</ul>")
            }
        }
    }
    cat("</ul>")
} else {
    cat("No statistical testing was performed. If contrasts were defined,",
        "then fold changes are available in the Results section.")
}
```

## Command

The differential expression analysis and this report were generated using the
following command:

```{r fun_call}
cat("<pre><code>")
cat(paste(FUN_CALL,collapse="<br/>"))
cat("</code></pre>")
```

```{r display_targets}
if (!is.null(theList)) {
    # First write targets
    tarDf <- metaseqR2:::.writeTargets(theList,outfile=file.path(
        PROJECT_PATH$main,"tracks","targets.txt"))
    tarDf$filename <- basename(as.character(tarDf$filename))
    
    # Download link
    cat("<br/>")
    cat("You can download the targets file from <strong>",
      "<a href='tracks/targets.txt' download target='_blank'>here</a>",
      "</strong><br/>",sep="")

    # Then render a table with targets
    cat("<br/>")
    cat("The following table summarizes the targets file used for the",
        "analysis. Do not forget to prepend the path to your BAM files in the",
        "<srong>filename</strong> column (also in the file that can be",
        "downloaded above).<br/>")
    print(kable(tarDf))
}
```

```{r summary_run_log}
if (runLog) {
    cat("<br/>The above command generated the following log output:<br/><br/>")
    cat("<div id=\"summary_log_container\">")
    logString <- paste(readLines(file.path(PROJECT_PATH$logs,
        "metaseqr_run.log")),collapse="---EOL---")
    logString <- gsub("\\$","->",logString)
    cat(gsub("---EOL---","<br/>",logString))
    cat("</div><br/>")
} else {
    cat("<h4>No logging requested</h4>")
}
```

## `r if (createTracks) 'Tracks'`

```{r tracks_1}
if (createTracks) {
    o <- metaseqR2:::getUcscOrganism(org)
    cat("You can use this ")
    if (trackInfo$stranded) {
        cat("<a href='https://genome.ucsc.edu/cgi-bin/hgTracks?genome=",o,
            "&hubUrl=",tLink,"' target='_blank'><strong>link</strong>",
            "</a>",sep="")
    } else {
        cat("<a href='https://genome.ucsc.edu/cgi-bin/hgTracks?db=",o,
            "&hgct_customText=",tLink,"' target='_blank'><strong>link</strong>",
            "</a>",sep="")
    }
    cat(" to load a UCSC Genome Browser session with the tracks derived from ",
        "this analysis. If stranded mode was chosen, a trackhub will be ",
        "loaded, otherwise, simple tracks will be loaded.",sep="")
    cat("<br/>")
}
```

```{r tracks_2}
if (createTracks) {
    if (trackInfo$stranded) {
        o <- metaseqR2:::getUcscOrganism(org)
        plusLinks <- paste0(trackInfo$urlBase,"/",o,"/",samples,"_plus.bigWig")
        minusLinks <- paste0(trackInfo$urlBase,"/",o,"/",samples,
            "_minus.bigWig")
        cat("You can download individual bigWig files, one for each sample,",
            "using the following list:<br/><br/>")
        cat("<strong>Plus (+) strand</strong><br/>")
        for (i in 1:length(plusLinks))
            cat("<a href='",plusLinks[i],"'>",samples[i],"</a><br/>",sep="")
        cat("<br/><strong>Minus (-) strand</strong><br/>")
        for (i in 1:length(minusLinks))
            cat("<a href='",minusLinks[i],"'>",samples[i],"</a><br/>",sep="")
        cat("<br/>")
    } else {
        tLines <- readLines(file.path(trackExportPath,"tracks.txt"))
        tmp <- strsplit(tLines," ")
        tNames <- sapply(tmp,function(x) {
            return(strsplit(x[3],"=")[[1]][2])
        })
        tLinks <- paste0(trackInfo$urlBase,"/",tNames,".bigWig")
        cat("You can download individual bigWig files, one for each sample,",
            "using the following list:<br/>")
        for (i in 1:length(tNames))
            cat("<a href='",tLinks[i],"'>",tNames[i],"</a><br/>",sep="")
        cat("<br/>")
        cat("You can manually load bigWig files in UCSC Genome Browser using ",
            "the following track lines:<br/><br/>")
        cat("<pre><code>")
        cat(paste(tLines,collapse="<br/>"))
        cat("</code></pre><br/>")
    }
}
```

```{r decide_qc}
if (any(qcPlots %in% qualPlots)) {
    pandoc.header("Quality control",level=1)
    pandoc.header("Quality control figures {.tabset .tabset-pills}",level=2)
    cat("The following figures summarize the quality control steps and ",
        "assessment performed by the ```metaseqr2``` pipeline. Each figure ",
        "category is accompanied by an explanatory text. All figures are ",
        "interactive wih additional controls on the top right of the figure.",
        sep="")
}
                
```

```{r analysis_figures_mds}
if ("mds" %in% qcPlots) {
    pandoc.header("MDS",level=3)
    
    ## Get chart data as serialized JSON
    #json <- diagplotMds(geneCounts,sampleList,output="json")
    
    cat("<h4><strong>Multidimensional scaling</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$mds,
        "<br/><br/></div>",sep="")
    
    # Container div
    cat("<div class=\"hc-rect-sm\" id=\"mds_container\"></div>",sep="")
    ## Render the chart
    #cat("<script>$('#mds_container').highcharts(",json,");</script>",sep="")
    # A hidden element to trigger upon page load and fetch the plot
    cat("<button id=\"_mds_trigger\" class=\"hidden-element\"></button>",sep="")
}
```

```{r analysis_figures_biodetection}
if ("biodetection" %in% qcPlots) {
    pandoc.header("Biodetection",level=3)
    
    cat("<h4><strong>Biotype detection</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$biodetection,
        "<br/><br/></div>",sep="")
        
    # Create a selector for each sample
    items <- paste("<option value=",paste("biodetection",samples,sep="_"),">",
        samples,"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a sample to display plot for",
        "</div>",sep="")
    cat("<select id=\"biodetection_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div
    cat("<div class=\"hc-rect-sm\" id=\"biodetection_container\"></div>",sep="")
    
    #if (reportDb == "dexie") {
    #   # Get chart data as list
    #   log <- capture.output({
    #       json <- diagplotNoiseq(geneCounts,sampleList,covars=covarsRaw,
    #           whichPlot="biodetection",output="json")
    #   })
    #   
    #   # Add extra data, serialize and import the plots to Dexie
    #   preImport <- vector("list",length(json))
    #   for (i in 1:length(json)) {
    #       preImport[[i]] <- list(
    #           name=paste("biodetection",samples[i],sep="_"),
    #           type="biodetection",
    #           subtype="generic",
    #           json=json[[samples[i]]]
    #       )
    #   }
    #   toImport <- toJSON(preImport,auto_unbox=TRUE,null="null")
    #   
    #   # Import script
    #   cat("<script type=\"text/javascript\">")
    #   cat("db.plots.bulkAdd(",toImport,");",sep="")
    #   cat("</script>")
    #   
    #   # Initialize first plot with serialized JSON
    #   cat("<script>$('#biodetection_container').highcharts(",
    #       toJSON(preImport[[1]]$json,auto_unbox=TRUE,null="null"),
    #       ");</script>",sep="")
    #}
}
```

```{r analysis_figures_countsbio}
if ("countsbio" %in% qcPlots) {
    pandoc.header("Biocounts",level=3)
    
    cat("<h4><strong>Biotype representation</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$countsbio,
        "<br/><br/></div>",sep="")
    
    cat("<h4><em>Biotypes within samples</em></h4>")
    
    # Create a selector for each sample
    items <- paste("<option value=",paste("countsbio",samples,sep="_"),">",
        samples,"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a sample to display plot for",
        "</div>",sep="")
    cat("<select id=\"countsbio_sample_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div - sample
    cat("<div class=\"hc-rect-sm\" id=\"countsbio_sample_container\"></div>",
        sep="")
    
    cat("<br/><h4><em>Biotype representation across samples</em></h4>")
    
    # Create a selector for each biotype
    items <- paste("<option value=",paste("countsbio",biotypes,sep="_"),">",
        biotypes,"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a biotype to display plot for",
        "</div>",sep="")
    cat("<select id=\"countsbio_biotype_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div - biotype
    cat("<div class=\"hc-rect-sm\" id=\"countsbio_biotype_container\">",
        "</div>",sep="")    
}
```

```{r analysis_figures_saturation}
if ("saturation" %in% qcPlots) {
    pandoc.header("Saturation",level=3)
    
    cat("<h4><strong>Biotype representation</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$saturation,
        "<br/><br/></div>",sep="")
    
    cat("<h4><em>Read saturation per biotype for all samples</em></h4>")
    
    # Create a selector for each sample
    items <- paste("<option value=",paste("saturation",samples,sep="_"),">",
        samples,"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a sample to display plot for",
        "</div>",sep="")
    cat("<select id=\"saturation_sample_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div - sample
    cat("<div class=\"hc-rect-sm\" id=\"saturation_sample_container\"></div>",
        sep="")
    
    cat("<br/><h4><em>Read saturation per sample for all biotypes</em></h4>")
    
    # Create a selector for each biotype
    items <- paste("<option value=",paste("saturation",biotypes,sep="_"),">",
        biotypes,"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a biotype to display plot for",
        "</div>",sep="")
    cat("<select id=\"saturation_biotype_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div - biotype
    cat("<div class=\"hc-rect-sm\" id=\"saturation_biotype_container\">",
        "</div>",sep="")
}
```

```{r analysis_figures_readsnoise}
if ("readnoise" %in% qcPlots) {
    pandoc.header("Reads noise",level=3)
    
    cat("<h4><strong>RNA-Seq reads noise</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$readnoise,
        "<br/><br/></div>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"readsnoise_container\"></div>",
        sep="")
    cat("<button id=\"_readnoise_trigger\" class=\"hidden-element\"></button>",
        sep="")
}
```

```{r analysis_figures_correl}
if ("correl" %in% qcPlots) {
    pandoc.header("Correlation",level=3)
    
    cat("<h4><strong>Pairwise sample correlations</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$correl,
        "<br/><br/></div>",sep="")
    
    corMat <- cor(geneCounts)
    if (!is.null(colnames(geneCounts)))
        colnames(corMat) <- colnames(geneCounts)
    
    n <- dim(corMat)[1]
    labs <- matrix(NA,n,n)
    for (i in 1:n)
        for (j in 1:n)
            labs[i,j] <- sprintf("%.2f",corMat[i,j])
    if (n <= 5) {
        notecex <- 1.2
    } else if (n > 5 & n < 10) {
        notecex <- 0.9
    } else {
        notecex <- 0.7
    }
    
    h <- list(heatmaply(corMat,revC=TRUE,trace="none",symm=TRUE,Colv=TRUE,
        cellnote=labs,colors=colorRampPalette(c("yellow","grey","blue")),
        k_col=length(sampleList),k_row=length(sampleList)) %>%
        layout(width=640,height=480))
    setNames(h,NULL)
    htmltools::tagList(tags$div(id="cormap",class="coheatmap-container",h[[1]]))
    #heatmaply(corMat,revC=TRUE,trace="none",symm=TRUE,Colv=TRUE,
    #   cellnote=labs,colors=colorRampPalette(c("yellow","grey","blue")),
    #   k_col=length(sampleList),k_row=length(sampleList)) %>%
    #   layout(width=640,height=480)
}
```

```{r analysis_figures_pairwise}
if ("pairwise" %in% qcPlots) {
    pandoc.header("Pairwise",level=3)
    
    cat("<h4><strong>Pairwise sample correlations</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$pairwise,
        "<br/><br/></div>",sep="")
        
    nams <- colnames(geneCounts)
    n <- ncol(geneCounts)
    plotNames <- character(n*(n-1)/2)
    counter <- 0
    for (i in 1:(n-1)) {
        for (j in (i+1):n) {
            counter <- counter + 1
            plotNames[counter] <- paste(nams[i],nams[j],sep="_vs_")
        }
    }
    
    # Create a selector for each sample
    items <- paste("<option value=",plotNames,">",gsub("_vs_"," vs ",plotNames),
        "</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a pair to display plot for",
        "</div>",sep="")
    cat("<select id=\"pairwise_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    cat("<h4><em>X-Y plots</em></h4>")
    # Container div - sample
    cat("<div class=\"hc-rect-sm\" id=\"pairwise_xy_container\"></div>",
        sep="")
    
    cat("<br/><h4><em>M-D plots</em></h4>")
    # Container div - sample
    cat("<div class=\"hc-rect-sm\" id=\"pairwise_md_container\"></div>",
        sep="")
}
```

```{r analysis_figures_filtered}
if ("filtered" %in% qcPlots) {
    pandoc.header("Filtered",level=3)
    
    cat("<h4><strong>Chromosome and biotype distribution of filtered ",
        transLevel,"s</strong></h4>",sep="")
    cat("<div class=\"figure-hint\">",reportMessages$explain$filtered,
        "<br/><br/></div>",sep="")
    cat("<h4>Chromosome distribution of filtered",transLevel,"s</h4>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"chromosome_distr_container\"></div>",
        sep="")
    cat("<h4>Biotype distribution of filtered ",transLevel,"s</h4>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"biotype_distr_container\"></div>",
        sep="")
    cat("<button id=\"_filtered_trigger\" class=\"hidden-element\"></button>",
        sep="")
}
```

```{r decide_norm}
if (any(qcPlots %in% normPlots)) {
    pandoc.header("Normalization",level=1)
    pandoc.header("Normalization assessment figures {.tabset .tabset-pills}",
        level=2)
    cat("The following figures allow for the assessment of the ",
        "normalization procedures performed by the ```metaseqr2``` pipeline. ",
        "Each figure category is accompanied by an explanatory text. All ",
        "figures are interactive wih additional controls on the top right ",
        "corner of the figure.",sep="")
}
```

```{r analysis_figures_boxplot}
if ("boxplot" %in% qcPlots) {
    pandoc.header("Boxplots",level=3)
    
    cat("<h4><strong>Boxplots</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$boxplot,
        "<br/><br/></div>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"boxplot_unorm_container\"></div>",
        sep="")
    cat("<div class=\"hc-rect-sm\" id=\"boxplot_norm_container\"></div>",
        sep="")
    cat("<br/><br/>")
    cat("<button id=\"_boxplot_trigger\" class=\"hidden-element\"></button>",
        sep="")
}
```

```{r analysis_figures_gcbias}
if ("gcbias" %in% qcPlots) {
    pandoc.header("GC bias",level=3)
    
    cat("<h4><strong>GC bias assessment plots</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$gcbias,
        "<br/><br/></div>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"gcbias_unorm_container\"></div>",
        sep="")
    cat("<div class=\"hc-rect-sm\" id=\"gcbias_norm_container\"></div>",
        sep="")
    cat("<br/><br/>")
    cat("<button id=\"_gcbias_trigger\" class=\"hidden-element\"></button>",
        sep="")
}
```

```{r analysis_figures_lengthbias}
if ("lengthbias" %in% qcPlots) {
    pandoc.header("Length bias",level=3)
    
    cat("<h4><strong>Length bias assessment plots</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$lengthbias,
        "<br/><br/></div>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"lengthbias_unorm_container\"></div>",
        sep="")
    cat("<div class=\"hc-rect-sm\" id=\"lengthbias_norm_container\"></div>",
        sep="")
    cat("<br/><br/>")
    cat("<button id=\"_lengthbias_trigger\" class=\"hidden-element\"></button>",
        sep="")
}
```

```{r analysis_figures_meandiff}
if ("meandiff" %in% qcPlots) {
    pandoc.header("Mean-Difference",level=3)
    
    cat("<h4><strong>Mean-difference plots for normalization assessment",
        "</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$meandiff,
        "<br/><br/></div>",sep="")
    
    # Create a selector for each sample
    namList <- vector("list",length(sampleList))
    names(namList) <- names(sampleList)
    for (n in names(sampleList)) {
        pairMatrix <- combn(1:length(sampleList[[n]]),2)
        nm <- character(0)
        for (i in 1:ncol(pairMatrix)) {
            s1 <- sampleList[[n]][pairMatrix[1,i]]
            s2 <- sampleList[[n]][pairMatrix[2,i]]
            nm <- c(nm,paste(s1,"_vs_",s2,sep=""))
        }
        namList[[n]] <- nm
    }
    
    html <- ""
    for (n in names(namList)) {
        nams <- namList[[n]]
        items <- paste("<option value=",nams,">",gsub("_vs_"," vs ",nams),
            "</option>",sep="")
        html <- paste(html,"<optgroup label=\"",n,"\">",items,"</optgroup>",
            sep="")
    }
    
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a pair to display plots for",
        "</div>",sep="")
    cat("<select id=\"meandiff_selector\">",html,"</select>",sep="")
    cat("</div>")
    
    # Container div - un-normalized
    cat("<div class=\"hc-rect-sm\" id=\"meandiff_unorm_container\"></div>",
        sep="")
    # Container div - normalized
    cat("<div class=\"hc-rect-sm\" id=\"meandiff_norm_container\"></div>",
        sep="") 
}
```

```{r analysis_figures_meanvar}
if ("meanvar" %in% qcPlots) {
    pandoc.header("Mean-Variance",level=3)
    
    cat("<h4><strong>Mean-variance plot for normalization assessment",
        "</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$meanvar,
        "<br/><br/></div>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"meanvar_unorm_container\"></div>",
        sep="")
    cat("<div class=\"hc-rect-sm\" id=\"meanvar_norm_container\"></div>",
        sep="")
    cat("<button id=\"_meanvar_trigger\" class=\"hidden-element\"></button>",
        sep="")
}
```

```{r analysis_figures_rnacomp}
if ("rnacomp" %in% qcPlots) {
    pandoc.header("Rna composition",level=3)
    
    cat("<h4><strong>RNA composition plot</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$rnacomp,
        "<br/><br/></div>",sep="")
        
    cat("<div class=\"hc-rect-sm\" id=\"rnacomp_unorm_container\"></div>",
        sep="")
    cat("<div class=\"hc-rect-sm\" id=\"rnacomp_norm_container\"></div>",sep="")
    cat("<button id=\"_rnacomp_trigger\" class=\"hidden-element\"></button>",
        sep="")
}
```

```{r decide_stat}
if (any(qcPlots %in% statPlots)) {
    pandoc.header("Statistics",level=1)
    pandoc.header(
        "Differential expression assessment figures {.tabset .tabset-pills}",
        level=2)
    cat("The following figures allow for the assessment of the statistical ",
        "testing procedures performed by the ```metaseqr2``` pipeline. Each ",
        "figure category is accompanied by an explanatory text. All figures ",
        "are interactive wih additional controls on the top right corner of ",
        "the figure.",sep="")
}
```

```{r analysis_figures_volcano}
if ("volcano" %in% qcPlots) {
    pandoc.header("Volcano",level=3)
    
    cat("<h4><strong>Volcano plots</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$volcano,
        "<br/><br/></div>",sep="")
        
    nn <- names(contrastList)
    namc <- character(0)
    for (n in nn) {
        fc <- log2(makeFoldChange(n,sampleList,normGenesExpr[1:3,],1))
        for (contr in colnames(fc))
            namc <- c(namc,contr)
    }
    
    # Create a selector for each contrast
    items <- paste("<option value=",paste("volcano",namc,sep="_"),">",
        gsub("_vs_"," vs ",namc),"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a contrast to display plot ",
        "for </div>",sep="")
    cat("<select id=\"volcano_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div
    cat("<div class=\"hc-rect-sm\" id=\"volcano_container\"></div>",sep="")
}
```

```{r analysis_figures_mastat}
if ("mastat" %in% qcPlots) {
    pandoc.header("MA",level=3)
    
    cat("<h4><strong>Mean-Difference (MA) plots</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$mastat,
        "<br/><br/></div>",sep="")
    
    nn <- names(contrastList)
    namc <- character(0)
    for (n in nn) {
        m <- log2(makeFoldChange(n,sampleList,normGenesExpr[1:3,],1))
        for (contr in colnames(m))
            namc <- c(namc,contr)
    }
    
    # Create a selector for each contrast
    items <- paste("<option value=",paste("mastat",namc,sep="_"),">",
        gsub("_vs_"," vs ",namc),"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a contrast to display plot ",
        "for </div>",sep="")
    cat("<select id=\"mastat_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div
    cat("<div class=\"hc-rect-sm\" id=\"mastat_container\"></div>",sep="")
}
```

```{r analysis_figures_deheatmap}
if ("deheatmap" %in% qcPlots) {
    pandoc.header("Heatmap",level=3)
    
    cat("<h4><strong>Differential expression heatmaps</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$deheatmap,
        "<br/><br/></div>",sep="")
    
    # Create a selector for each contrast
    items <- paste("<option value=",
        paste("deheatmap",gsub("\\.","__",names(contrastList)),sep="_"),">",
        gsub("_vs_"," vs ",names(contrastList)),"</option>",sep="")
    scal <- paste("<option value=",c("asis","zscore"),">",
        c("log2 reads","z-scores"),"</option>",sep="")
    cat("<div class=\"row\"><div class=\"col-sm-12\">",sep="")
    cat("<div class=\"row\"><div class=\"col-sm-6\">")
    cat("<div class=\"hc-selector-label\">Select a contrast to display ",
        "heatmap for </div>",sep="")
    cat("<select id=\"heatmap_contrast_selector\">",items,"</select>",sep="")
    cat("</div><div class=\"col-sm-6\">")
    cat("<div class=\"pull-right\">")
    cat("<div class=\"hc-selector-label\">Select a scale to display heatmap ",
        "for </div>",sep="")
    cat("<select id=\"heatmap_scale_selector\">",scal,"</select>",sep="")
    cat("</div></div></div></div></div>")
    
    hList <- vector("list",length(contrastList)*2)
    hNames <- character(length(contrastList)*2)
    hInd <- 0
    for (cnt in names(contrastList)) {
        s <- names(unlist(contrastList[[cnt]]))
        if (is.na(pcut))
            pcut <- 0.05
        ind <- which(sumpList[[cnt]]<=pcut)
        if (length(ind) > 0) {
            mat <- as.matrix(normGenesExpr[ind,s])
        } else {
            mat <- as.matrix(normGenesExpr[,s])
        }
        if (!is.null(reportTop)) {
            topg <- ceiling(reportTop*nrow(mat))
            mat <- mat[1:topg,,drop=FALSE]
        }
        # As is scale
        hInd <- hInd + 1
        hNames[hInd] <- paste0("deheatmap_",gsub("\\.","__",cnt),"_asis")
        hList[[hInd]] <- heatmaply(nat2log(mat),trace="none",colors=bluered(64),
            labRow=rep(NA,nrow(mat)),
            main=paste("DEG heatmap",cnt,"log2 reads")) %>% 
            layout(width=720,height=840)
        # z-scores
        M <- t(scale(t(nat2log(mat))))
        if (any(is.infinite(M)))
            M[which(is.infinite(M))] <- min(M)
        if (any(is.na(M)))
            M[which(is.na(M))] <- 0
        hInd <- hInd + 1
        hNames[hInd] <- paste0("deheatmap_",gsub("\\.","__",cnt),"_zscore")
        hList[[hInd]] <- heatmaply(M,trace="none",colors=bluered(64),
            labRow=rep(NA,nrow(mat)),
            main=paste("DEG heatmap",cnt,"z-scores")) %>% 
            layout(width=720,height=840)
    }
    setNames(hList,NULL)
    
    # Dirty hack
    toEval <- "htmltools::tagList("
    hItems <- character(length(hList))
    for (i in 1:length(hList))
        hItems[i] <- paste0("tags$div(id='",hNames[i],
            "', class='deheatmap-container',hList[[",i,"]])")
    toEval <- paste0(toEval,paste(hItems,collapse=","),")")
    eval(parse(text=toEval))
    
    #htmltools::tagList(
    #   tags$div(id=paste("deheatmap",gsub("\\.","__",names(contrastList)[1]),
    #       sep="_"),class="deheatmap-container",hList[[1]]),
    #   tags$div(id=paste("deheatmap",gsub("\\.","__",names(contrastList)[2]),
    #       sep="_"),class="deheatmap-container",hList[[2]])
    #)
}
```

```{js deheatmap_render}
// Chunk to initialize the display of the DE heatmaps
var hids = $('.deheatmap-container').map(function(index) {
    return(this.id);
});
$('.deheatmap-container').hide(0);
$('#'+hids[0]).show();
```

```{r analysis_figures_biodist}
if ("biodist" %in% qcPlots) {
    pandoc.header("Biodist",level=3)
    
    cat("<h4><strong>Chromosome and biotype distributions of differentially ",
        "expressed ",transLevel,"s</strong></h4>",sep="")
    cat("<div class=\"figure-hint\">",reportMessages$explain$biodist,
        "<br/><br/></div>",sep="")
    
    nn <- names(contrastList)
    
    # Create a selector for each contrast
    items <- paste("<option value=",paste("biodist",nn,sep="_"),">",
        gsub("_vs_"," vs ",nn),"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a contrast to display plot ",
        "for </div>",sep="")
    cat("<select id=\"biodist_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Containers   
    cat("<h4>Chromosome distribution of differentially expressed ",transLevel,
        "s</h4>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"biodist_chromosome_container\"></div>",
        sep="")
    cat("<h4>Biotype distribution of differentially expressed ",transLevel,
        "s</h4>",sep="")
    cat("<div class=\"hc-rect-sm\" id=\"biodist_biotype_container\"></div>",
        sep="")
}
```

```{r analysis_figures_deregulogram}
if ("deregulogram" %in% qcPlots) {
    pandoc.header("Deregulogram",level=3)
    
    cat("<h4><strong>Deregulograms</strong></h4>")
    cat("<div class=\"figure-hint\">",reportMessages$explain$deregulogram,
        "<br/><br/></div>",sep="")
    
    cntPairs <- combn(names(contrastList),2)
    namc <- character(ncol(cntPairs))
    counter <- 0
    for (i in 1:ncol(cntPairs)) {
        counter <- counter + 1
        namc[counter] <- paste0(cntPairs[1,i],"__",cntPairs[2,i])
    }
    
    # Create a selector for each pair of contrasts
    items <- paste("<option value=",paste("deregulogram",namc,sep="_"),">",
        gsub("__"," against ",namc),"</option>",sep="")
    cat("<div>")
    cat("<div class=\"hc-selector-label\">Select a pair of contrasts to ",
        "display plot for </div>",sep="")
    cat("<select id=\"deregulo_selector\">",items,"</select>",sep="")
    cat("</div>")
    
    # Container div
    cat("<div class=\"hc-rect-sm\" id=\"deregulo_container\"></div>",sep="")
}
```

```{r analysis_figures_statvenn}
if ("statvenn" %in% qcPlots) {
    pandoc.header("StatVenn",level=3)
    
    cat("<h4><strong>Venn diagram of differentially expressed ",transLevel,
        "s</strong></h4>",sep="")
    cat("<div class=\"figure-hint\">",reportMessages$explain$statvenn,
        "<br/><br/></div>",sep="")
        
    # Create a selector for each contrast
    nn <- names(contrastList)
    items <- paste("<option value=",paste("statvenn",nn,sep="_"),">",
        gsub("_vs_"," vs ",nn),"</option>",sep="")
    deupdo <- paste("<option value=",c("dereg","up","down"),">",
        c("Deregulated","Up regulated","Down downregulated"),"</option>",sep="")
    cat("<div class=\"row\"><div class=\"col-sm-8\">")
    cat("<div class=\"row\"><div class=\"col-sm-6\">")
    cat("<div class=\"hc-selector-label\">Select a contrast to display Venn ",
        "diagram for </div>",sep="")
    cat("<select id=\"statvenn_contrast_selector\">",items,"</select>",sep="")
    cat("</div><div class=\"col-sm-6\">")
    cat("<div class=\"hc-selector-label\">Select a direction to display Venn ",
        "diagram for </div>",sep="")
    cat("<select id=\"statvenn_direction_selector\">",deupdo,"</select>",sep="")
    cat("</div></div></div></div>")
    
    cat("<div class=\"row\">")
    cat("<div class=\"col-sm-8\"><div id=\"jvenn_stat_container\"></div></div>")
    cat("<div class=\"col-sm-4\">")
    cat("<div class=\"hc-selector-label\">Click on a number on Venn diagrams ",
        "to display the respective genes</div>",sep="")
    cat("<textarea id=\"jvenn_stat_list\" class=\"jvenn_list\" readonly>",
        "</textarea></div>",sep="")
    cat("</div>")
}

```{r analysis_figures_foldvenn}
if ("foldvenn" %in% qcPlots) {
    pandoc.header("FoldVenn",level=3)
    
    cat("<h4><strong>Venn diagram of differentially expressed ",transLevel,
        "s</strong></h4>",sep="")
    cat("<div class=\"figure-hint\">",reportMessages$explain$foldvenn,
        "<br/><br/></div>",sep="")
        
    # Create a selector for each contrast
    if (length(statistics) > 1) {
        nn <- c(statistics,metaP)
    } else {
        nn <- statistics
    }
    
    items <- paste("<option value=",paste("foldvenn",nn,sep="_"),">",nn,
        "</option>",sep="")
    deupdo <- paste("<option value=",c("dereg","up","down"),">",
        c("Deregulated","Up regulated","Down downregulated"),"</option>",sep="")
    cat("<div class=\"row\"><div class=\"col-sm-8\">")
    cat("<div class=\"row\"><div class=\"col-sm-6\">")
    cat("<div class=\"hc-selector-label\">Select an algorithm to display Venn ",
        "diagram for </div>",sep="")
    cat("<select id=\"foldvenn_algo_selector\">",items,"</select>",sep="")
    cat("</div><div class=\"col-sm-6\">")
    cat("<div class=\"hc-selector-label\">Select a direction to display Venn ",
        "diagram for </div>",sep="")
    cat("<select id=\"foldvenn_direction_selector\">",deupdo,"</select>",sep="")
    cat("</div></div></div></div>")
    
    cat("<div class=\"row\">")
    cat("<div class=\"col-sm-8\"><div id=\"jvenn_fold_container\"></div></div>")
    cat("<div class=\"col-sm-4\">")
    cat("<div class=\"hc-selector-label\">Click on a number on Venn diagrams ",
        "to display the respective genes</div>",sep="")
    cat("<textarea id=\"jvenn_fold_list\" class=\"jvenn_list\" readonly>",
        "</textarea></div>",sep="")
    cat("</div>")
}
```

# Results

## Tables of differentially expressed genes

The following tables allow for a quick exploration of the results of the
statistical analysis performed by the ```metaseqr2``` pipeline. If no
statistical testing or contrasts requested, just ignore any respective texts
and jump to tables or download the results.

```{r results_tables_header}
if (is.null(reportTop)) {
    topText <- paste("all the statistically significant ",  transLevel,"s. ",
        sep="")
} else {
    topText <- paste("the top ",round(100*reportTop),"% statistically ",
        "significant ",transLevel,"s. ",sep="")
}
restText <- paste("Use the download links below each table to retrieve the ",
    "total list of differentially expressed ",transLevel,"s or the whole ",
    transLevel," list of the selected genome irrespective of differential ",
    "expression.",sep="")
cat("Each table presents ",topText,restText,sep="")
cat("Furthermore each table can be searched using the search field on the ",
    "top right and you can also find the following information:<br/><br/>",
    sep="")

cat("<ul>")
cat("<li>The <em>chromosome</em> column is linked to the genomic location of ",
    "the ",transLevel," and opens a new tab/window to the UCSC Genome Browser",
    "</li>",sep="")
cat("<li>The *gene_id* column opens a link to the respective full annotation ",
    "source (only for Ensembl and RefSeq)</li>",sep="")
cat("<li>The background of the *p_value* and *FDR* columns displays a bar ",
    "with length proportional to the significance of each ",transLevel,"</li>",
    sep="")
cat("<li>The background color of the *fold change (_vs_)* column(s) displays ",
    "shows the deregulation of each ",transLevel," and is proportional to ",
    "the deregulation strength (red for up- green for down-regulation)</li>",
    sep="")
cat("<li>The background of the rest columns (condition average expression) ",
    "displays a bar with length proportional to the expression strength ",
    "of each condition</li>",sep="")
cat("</ul>")
```

```{r dt_hack, include=FALSE}
DT::datatable(NULL,extensions="Buttons")
```

```{r results_tables}
# Create a selector for each contrast
if (!is.null(contrast)) {
    items <- paste("<option value=",paste("detable",gsub("\\.","__",
        names(contrastList)),sep="_"),">",gsub("_vs_"," vs ",
        names(contrastList)),"</option>",sep="")
} else {
    items <- "<option value=detable_no_contrast>No contrast</option>"
}
cat("<hr><div>")
cat("<div class=\"hc-selector-label\">Select a contrast to display ",
    "DEG table for </div>",sep="")
cat("<select id=\"detable_selector\">",items,"</select>",sep="")
cat("</div><br/>")

if (!is.null(contrast) && !any(is.na(statistics))) {
    for (cnt in contrast) {
        cntn <- gsub("_vs_"," vs ",cnt)
        #pandoc.header(cntn,level=3)
        
        cat("<div class='detable-helper-wrapper' id='detable_",
            gsub("\\.","__",cnt),"'>",sep="")
        
        cat("<h4>DEG table for the contrast <strong>",cntn,"</strong></h4>",
            sep="")
        cat("<div class=\"figure-hint\">The following table presents ",topText,
            " for the contrast <strong>",cntn,"</strong>.</div>",sep="")
        
        # 15 colors should be enough for most situations...
        sColors <- c("#97F9F9","#A4DEF9","#C1E0F7","#CFBAE1","#C59FC9",
            "#EDEEC9","#DDE7C7","#BFD8BD","#98C9A3","#77BFA3","#7BDFF2",
            "#B2F7EF","#EFF7F6","#F7D6E0","#F2B5D4")
            
        # We do not need GC content really...
        gci <- grep("gc_content",colnames(reportTables[[cnt]]))
        if (length(gci) > 0)
            reportTables[[cnt]] <- reportTables[[cnt]][,-gci]
            
        # Fold change and condition means should always exist in this table 
        # defined by us...
        pi <- grep("p-value",colnames(reportTables[[cnt]]))
        di <- grep("FDR",colnames(reportTables[[cnt]]))
        fi <- grep("_vs_",colnames(reportTables[[cnt]]))
        # If more than one statistics applied, the 1st two from fi are meta 
        # p-value and meta FDR
        if (length(statistics) > 1)
            fi <- fi[-c(1,2)]
        #ci <- numeric(length(sampleList))
        #N <- names(sampleList)
        ci <- numeric(length(contrastList[[cnt]]))
        N <- unique(unlist(contrastList[[cnt]],use.names=FALSE))
        
        if (length(grep("rpgm_normalized_mean_counts_",
            colnames(reportTables[[cnt]]))) > 0) {
            for (i in 1:length(N))
                ci[i] <- grep(paste0("rpgm_normalized_mean_counts_",N[i]),
                    colnames(reportTables[[cnt]]))
        } else {
            for (i in 1:length(N))
                ci[i] <- grep(N[i],colnames(reportTables[[cnt]]),fixed=TRUE)
        }
        
        # Get values before rounding for styling
        brks <- quantile(reportTables[[cnt]][,fi],probs=seq(.05,.95,.05),
            na.rm=TRUE)
        preClrs <- col2rgb(rev(redgreen(length(brks)+1)))
        clrs <- paste("rgb(",apply(preClrs,2,paste0,collapse=","),")",sep="")
        pRangeForColor <- 1-as.numeric(reportTables[[cnt]][,pi])
        fRangeForColor <- 1-as.numeric(reportTables[[cnt]][,di])
        styleBars <- character(length(ci))
        for (i in 1:length(ci))
            styleBars[i] <- styleColorBar(range(reportTables[[cnt]][,ci[i]]),
                sColors[i])
        
        # After setting colors, format the table for the report
        if (!is.numeric(version)) # Not defined in call
            version <- NULL
        reportTables[[cnt]] <- .formatForReport(reportTables[[cnt]],o=org,
            r=refdb,v=version)
        # Hack to prevent the header name taking 2 rows because of 
        # hyphenation(?)
        names(reportTables[[cnt]])[pi] <- "p_value"
        
        # Dirty hack
        # https://stackoverflow.com/questions/32018521/
        # shiny-use-stylecolorbar-with-data-from-two-data-frames
        reportTables[[cnt]] <- 
            cbind(pRangeForColor,fRangeForColor,reportTables[[cnt]])
        jspf <- paste0("function(row,data) {",
            "var value = data[0];",
            "var backgroundValueP =",styleColorBar(pRangeForColor,
                "lightblue")[1],";",
            "var value = data[1];",
            "var backgroundValueF =",styleColorBar(fRangeForColor,
                "lightsalmon")[1],";",
            "$('td',row).eq(",pi-1,").css({",
                "'background': backgroundValueP,",
                "'background-repeat': 'no-repeat',",
                "'background-position': 'center',",
                "'background-size': '98% 88%'",
            "});",
            "$('td',row).eq(",di-1,").css({",
                "'background': backgroundValueF,",
                "'background-repeat': 'no-repeat',",
                "'background-position': 'center',",
                "'background-size': '98% 88%'",
            "});",
            "for (var i=",fi[1],";i<=",fi[length(fi)],";i++) {",
                "var value = data[i+1];",
                "$('td',row).eq(i-1).css({",
                    "'color': '#FFFFFF',",
                    "'font-weight': 'bold',",
                    "'background-color': ",styleInterval(brks,clrs),
                "});",
            "}")
        
        for (i in 1:length(ci)) {
            jsa <- paste0("var value = data[",ci[i]+1,"];",
                "$('td',row).eq(",ci[i]-1,").css({",
                    "'background': ",styleBars[i],",",
                    "'background-repeat': 'no-repeat',",
                    "'background-position': 'center',",
                    "'background-size': '98% 88%'",
                "});")
            jspf <- paste0(jspf,jsa)
        }
        jspf <- paste0(jspf,"}")
        
        theTable <- DT::datatable(
            reportTables[[cnt]],
            rownames=FALSE,
            width="95%",
            height=460,
            style="bootstrap",
            class="table-condensed table-striped table-responsive",
            escape=FALSE,
            elementId=paste("detable_",gsub("\\.","__",cnt),"_",sep=""),
            extensions="Buttons",
            options=list(
                scrollX=TRUE,
                order=list(list(pi+1,"asc")),
                columnDefs=list(
                    list(
                        targets=0:1,
                        visible=FALSE
                    )
                ),
                dom='Bfrtip',
                #buttons=I('colvis'),
                buttons=list(
                    list(
                        extend='colvis',
                        #exclude=0:1,
                        text="Show/Hide columns"
                    )
                ),
                rowCallback=JS(jspf)
            )
        ) #%>% formatStyle(
            #names(reportTables[[cnt]])[fi+2],
            #backgroundColor=styleInterval(brks,clrs),
            #fontWeight="bold",
            #color="#FFFFFF"
        #) #%>% formatStyle(
            #names(reportTables[[cnt]])[pi+2],
            #background=styleColorBar(pRangeForColor,"lightblue"),
            #fontWeight="bold",
            #backgroundSize="98% 88%",
            #backgroundRepeat="no-repeat",
            #backgroundPosition="center"
        #)
        print(htmltools::tagList(tags$div(class="detable-container",theTable)))
        
        cat("<br/><br/><div class=\"figure-hint link-box\">")
        cat("<strong><a href=\"lists/metaseqr_sig_out_",cnt,".txt.gz\"",
            " download>Download</a> the DEG result list for ",cnt,
            ".</strong><br/>",sep="")
        if (!is.null(geneCountsZero) || !is.null(geneCountsDead))
            cat("<strong><a href=\"lists/metaseqr_all_out_",cnt,".txt.gz\"",
            " download>Download</a> the whole result list for ",cnt,
            ".</strong><br/>",sep="")
        cat("</div><br/><br/>")
        
        cat("</div>")
    }
} else if (!is.null(contrast) && any(is.na(statistics))) {
    for (cnt in contrast) {
        cntn <- gsub("_vs_"," vs ",cnt)
        
        cat("<div class='detable-helper-wrapper' id='detable_",
            gsub("\\.","__",cnt),"'>",sep="")
        
        cat("<h4>DEG table for the contrast <strong>",cntn,"</strong></h4>",
            sep="")
        cat("<div class=\"figure-hint\">The following table presents ",topText,
            " for the contrast <strong>",cntn,"</strong>.</div>",sep="")
        
        # 15 colors should be enough for most situations...
        sColors <- c("#97F9F9","#A4DEF9","#C1E0F7","#CFBAE1","#C59FC9",
            "#EDEEC9","#DDE7C7","#BFD8BD","#98C9A3","#77BFA3","#7BDFF2",
            "#B2F7EF","#EFF7F6","#F7D6E0","#F2B5D4")
            
        # We do not need GC content really...
        gci <- grep("gc_content",colnames(reportTables[[cnt]]))
        if (length(gci) > 0)
            reportTables[[cnt]] <- reportTables[[cnt]][,-gci]
            
        # Fold change and condition means should always exist in this table 
        # defined by us...
        fi <- grep("_vs_",colnames(reportTables[[cnt]]))
        #ci <- numeric(length(contrastList[[cnt]]))
        #N <- unique(unlist(contrastList[[cnt]],use.names=FALSE))
        ci <- numeric(length(names(sampleList)))
        N <- names(sampleList)
        
        if (length(grep("rpgm_normalized_mean_counts_",
            colnames(reportTables[[cnt]]))) > 0) {
            for (i in 1:length(N))
                ci[i] <- grep(paste0("rpgm_normalized_mean_counts_",N[i]),
                    colnames(reportTables[[cnt]]))
        } else {
            for (i in 1:length(N))
                ci[i] <- grep(N[i],colnames(reportTables[[cnt]]),fixed=TRUE)
        }
        
        # Get values before rounding for styling
        brks <- quantile(reportTables[[cnt]][,fi],probs=seq(.05,.95,.05),
            na.rm=TRUE)
        preClrs <- col2rgb(rev(redgreen(length(brks)+1)))
        clrs <- paste("rgb(",apply(preClrs,2,paste0,collapse=","),")",sep="")
        styleBars <- character(length(ci))
        for (i in 1:length(ci))
            styleBars[i] <- styleColorBar(range(reportTables[[cnt]][,ci[i]]),
                sColors[i])
        
        # After setting colors, format the table for the report
        if (!is.numeric(version)) # Not defined in call
            version <- NULL
        reportTables[[cnt]] <- .formatForReport(reportTables[[cnt]],o=org,
            r=refdb,v=version)
        
        # Dirty hack
        # https://stackoverflow.com/questions/32018521/
        # shiny-use-stylecolorbar-with-data-from-two-data-frames
        reportTables[[cnt]] <- cbind(reportTables[[cnt]])
        jspf <- paste0("function(row,data) {",
            "for (var i=",fi[1],";i<=",fi[length(fi)],";i++) {",
                "var value = data[i-1];",
                "$('td',row).eq(i-1).css({",
                    "'color': '#FFFFFF',",
                    "'font-weight': 'bold',",
                    "'background-color': ",styleInterval(brks,clrs),
                "});",
            "}")
        
        for (i in 1:length(ci)) {
            jsa <- paste0("var value = data[",ci[i]-1,"];",
                "$('td',row).eq(",ci[i]-1,").css({",
                    "'background': ",styleBars[i],",",
                    "'background-repeat': 'no-repeat',",
                    "'background-position': 'center',",
                    "'background-size': '98% 88%'",
                "});")
            jspf <- paste0(jspf,jsa)
        }
        jspf <- paste0(jspf,"}")
        
        theTable <- DT::datatable(
            reportTables[[cnt]],
            rownames=FALSE,
            width="95%",
            height=460,
            style="bootstrap",
            class="table-condensed table-striped table-responsive",
            escape=FALSE,
            elementId=paste("detable_",gsub("\\.","__",cnt),"_",sep=""),
            extensions="Buttons",
            options=list(
                scrollX=TRUE,
                dom='Bfrtip',
                buttons=list(
                    list(
                        extend='colvis',
                        text="Show/Hide columns"
                    )
                ),
                rowCallback=JS(jspf)
            )
        )
        print(htmltools::tagList(tags$div(class="detable-container",theTable)))
        
        cat("<br/><br/><div class=\"figure-hint link-box\">")
        cat("<strong><a href=\"lists/metaseqr_sig_out_",cnt,".txt.gz\"",
            " download>Download</a> the DEG result list for ",cnt,
            ".</strong><br/>",sep="")
        if (!is.null(geneCountsZero) || !is.null(geneCountsDead))
            cat("<strong><a href=\"lists/metaseqr_all_out_",cnt,".txt.gz\"",
            " download>Download</a> the whole result list for ",cnt,
            ".</strong><br/>",sep="")
        cat("</div><br/><br/>")
        
        cat("</div>")
    }
} else {
    cat("<div class='detable-helper-wrapper' id='detable_no_contrast'>")
    cat("<h4>Quantification table for the run</h4>")
    
    # 15 colors should be enough for most situations...
    sColors <- c("#97F9F9","#A4DEF9","#C1E0F7","#CFBAE1","#C59FC9",
        "#EDEEC9","#DDE7C7","#BFD8BD","#98C9A3","#77BFA3","#7BDFF2",
        "#B2F7EF","#EFF7F6","#F7D6E0","#F2B5D4")
        
    # We do not need GC content really...
    gci <- grep("gc_content",colnames(reportTables[[1]]))
    if (length(gci) > 0)
        reportTables[[1]] <- reportTables[[1]][,-gci]
        
    #ci <- numeric(length(unlist(sampleList)))
    #N <- unique(unlist(sampleList,use.names=FALSE))
    ci <- numeric(length(names(sampleList)))
    N <- names(sampleList)
    
    if (length(grep("rpgm_normalized_mean_counts_",
        colnames(reportTables[[1]]))) > 0) {
        for (i in 1:length(N))
            ci[i] <- grep(paste0("rpgm_normalized_mean_counts_",N[i]),
                colnames(reportTables[[1]]))
    } else {
        for (i in 1:length(N))
            ci[i] <- grep(N[i],colnames(reportTables[[1]]),fixed=TRUE)
    }
    
    # Get values before rounding for styling
    styleBars <- character(length(ci))
    for (i in 1:length(ci))
        styleBars[i] <- styleColorBar(range(reportTables[[1]][,ci[i]]),
            sColors[i])
    
    # After setting colors, format the table for the report
    if (!is.numeric(version)) # Not defined in call
        version <- NULL
    reportTables[[1]] <- .formatForReport(reportTables[[1]],o=org,r=refdb,
        v=version)
    
    # Dirty hack
    # https://stackoverflow.com/questions/32018521/
    # shiny-use-stylecolorbar-with-data-from-two-data-frames
    jspf <- paste0("function(row,data) {")
    for (i in 1:length(ci)) {
        jsa <- paste0("var value = data[",ci[i]-1,"];",
            "$('td',row).eq(",ci[i]-1,").css({",
                "'background': ",styleBars[i],",",
                "'background-repeat': 'no-repeat',",
                "'background-position': 'center',",
                "'background-size': '98% 88%'",
            "});")
        jspf <- paste0(jspf,jsa)
    }
    jspf <- paste0(jspf,"}")
    
    theTable <- DT::datatable(
        reportTables[[1]],
        rownames=FALSE,
        width="95%",
        height=460,
        style="bootstrap",
        class="table-condensed table-striped table-responsive",
        escape=FALSE,
        elementId="detable_no_contrast",
        extensions="Buttons",
        options=list(
            scrollX=TRUE,
            dom='Bfrtip',
            buttons=list(
                list(
                    extend='colvis',
                    text="Show/Hide columns"
                )
            ),
            rowCallback=JS(jspf)
        )
    )
    print(htmltools::tagList(tags$div(class="detable-container",theTable)))
    
    cat("<br/><br/><div class=\"figure-hint link-box\">")
    cat("<strong><a href=\"lists/metaseqr_sig_out.txt.gz\"",
        " download>Download</a> the result list.</strong><br/>",sep="")
    if (!is.null(geneCountsZero) || !is.null(geneCountsDead))
        cat("<strong><a href=\"lists/metaseqr_all_out.txt.gz\"",
        " download>Download</a> the whole result list.</strong><br/>",sep="")
    cat("</div><br/><br/>")
    
    cat("</div>")
}
```

```{js detable_render}
var tids = $('.detable-helper-wrapper').map(function(index) {
    return(this.id);
});
$('.detable-helper-wrapper').hide(0);
$('#'+tids[0]).show();

// The only way to align columns...
var timer = setInterval(function() {
    if ($('#'+tids[0]+"_").find("table").length > 0) {
        if ($('#'+tids[0]+"_").find("table")[1].id !== undefined) {
            var id = $('#'+tids[0]+"_").find("table")[1].id;
            $('#'+id).DataTable().columns.adjust();
            clearInterval(timer);
        }
    }
},100);
```

```{r export_links}
if (exportCountsTable) {
    cat("<br/><div class=\"figure-hint link-box\">")
    if (file.exists(file.path(PROJECT_PATH[["lists"]],
        "raw_counts_table.txt.gz")))
        cat("<strong><a href=\"lists/raw_counts_table.txt.gz\" download>",
            "Download</a> the raw read counts table for the experiment.",
            "</strong><br/>",sep="")
    cat("<strong><a href=\"lists/normalized_counts_table.txt.gz\"",
        " download>Download</a> the normalized read counts table for the ",
        "experiment.</strong><br/>",sep="")
    cat("</div>")
}

if (length(qcPlots) > 0) {
    ltxt <- paste(paste("<a href=\"plots/metaseqr_figures_",figFormat,
        ".zip\" download>",figFormat,"</a>",sep=""),collapse=", ")
    cat("<div class=\"figure-hint link-box\">")
    cat("<strong>Get all the figures in ",ltxt," format.</strong>")
    cat("</div>")
}
```

# References

```{r references}
refs <- unique(c(
    reportMessages$references$main,
    reportMessages$references$filein[[fileType]],
    reportMessages$references$norm[[normalization]],
    unlist(reportMessages$references$stat[statistics],use.names=FALSE),
    unlist(reportMessages$references$figure[qcPlots],use.names=FALSE),
    reportMessages$references$multiple[[adjustMethod]],
    reportMessages$references$meta[[metaP]]
))
cat("<div class=\"figure-hint\" style=\"margin-left:25px;\"><ol>")
for (r in refs)
    cat("<li>",r,"</li>",sep="")
cat("</ol></div>")
```

```{js bind_event_changes}
$("#biodetection_selector").on('change',function() {
    var name = this.value;
    getPlotFromDb({name: name,type: "biodetection",subtype: "generic"},
        "biodetection_container");
});

$("#countsbio_sample_selector").on('change',function() {
    var name = this.value;
    getPlotWithFunFromDb({name: name,type: "countsbio",subtype: "sample"},
        "countsbio_sample_container");
});

$("#countsbio_biotype_selector").on('change',function() {
    var name = this.value;
    getPlotWithFunFromDb({name: name,type: "countsbio",subtype: "biotype"},
        "countsbio_biotype_container");
});

$("#saturation_sample_selector").on('change',function() {
    var name = this.value;
    getPlotWithFunFromDb({name: name,type: "saturation",subtype: "sample"},
        "saturation_sample_container");
});

$("#saturation_biotype_selector").on('change',function() {
    var name = this.value;
    getPlotWithFunFromDb({name: name,type: "saturation",subtype: "biotype"},
        "saturation_biotype_container");
});

$("#pairwise_selector").on('change',function() {
    var name = this.value;
    getPlotFromDb({name: name,type: "pairwise",subtype: "xy"},
        "pairwise_xy_container");
    getPlotFromDb({name: name,type: "pairwise",subtype: "md"},
        "pairwise_md_container");
});

$("#meandiff_selector").on('change',function() {
    var name = this.value;
    getPlotFromDb({name: name,type: "meandiff",subtype: "unorm"},
        "meandiff_unorm_container");
    getPlotFromDb({name: name,type: "meandiff",subtype: "norm"},
        "meandiff_norm_container");
});

$("#volcano_selector").on('change',function() {
    var name = this.value;
    getPlotFromDb({name: name,type: "volcano",subtype: "generic"},
        "volcano_container");
});

$("#mastat_selector").on('change',function() {
    var name = this.value;
    getPlotFromDb({name: name,type: "mastat",subtype: "generic"},
        "mastat_container");
});

$("#deregulo_selector").on('change',function() {
    var name = this.value;
    getPlotFromDb({name: name,type: "deregulogram",subtype: "generic"},
        "deregulo_container");
});

$("#heatmap_contrast_selector").on('change',function() {
    var name = this.value;
    var suffix = $("#heatmap_scale_selector").val();
    var newName = name + "_" + suffix;
    $(".deheatmap-container").hide(0);
    $("#"+newName).show();
});

$("#heatmap_scale_selector").on('change',function() {
    var name = this.value;
    var prefix = $("#heatmap_contrast_selector").val();
    var newName = prefix + "_" + name;
    $(".deheatmap-container").hide(0);
    $("#"+newName).show();
});

$("#biodist_selector").on('change',function() {
    var name = this.value;
    getPlotFromDb({name: name,type: "biodist",subtype: "chromosome"},
        "biodist_chromosome_container");
    getPlotFromDb({name: name,type: "biodist",subtype: "biotype"},
        "biodist_biotype_container");
});

$("#statvenn_contrast_selector").on('change',function() {
    var name = this.value;
    var suffix = $("#statvenn_direction_selector").val();
    var newName = name + "_" + suffix;
    var newSubtype = "stat_" + suffix;
    getVennFromDb({name: newName,type: "statvenn",subtype: newSubtype},
        "jvenn_stat_container");
});

$("#statvenn_direction_selector").on('change',function() {
    var name = this.value;
    var prefix = $("#statvenn_contrast_selector").val();
    var newName = prefix + "_" + name;
    var newSubtype = "stat_" + name;
    getVennFromDb({name: newName,type: "statvenn",subtype: newSubtype},
        "jvenn_stat_container");
});

$("#foldvenn_algo_selector").on('change',function() {
    var name = this.value;
    var suffix = $("#foldvenn_direction_selector").val();
    var newName = name + "_" + suffix;
    var newSubtype = "fold_" + suffix;
    getVennFromDb({name: newName,type: "foldvenn",subtype: newSubtype},
        "jvenn_fold_container");
});

$("#foldvenn_direction_selector").on('change',function() {
    var name = this.value;
    var prefix = $("#foldvenn_algo_selector").val();
    var newName = prefix + "_" + name;
    var newSubtype = "fold_" + name;
    getVennFromDb({name: newName,type: "foldvenn",subtype: newSubtype},
        "jvenn_fold_container");
});

$("#detable_selector").on('change',function() {
    var name = this.value;
    $(".detable-helper-wrapper").hide(0);
    $("#"+name).show();
    
    // Hack to make the table visible (htmlwidget known issue)
    window.dispatchEvent(new Event('resize'));
    // Hack to align the column headers
    var id = $('#'+name+"_").find("table")[1].id;
    $('#'+id).DataTable().columns.adjust();
});

// Hidden buttons
$("#_mds_trigger").on('click',function() {
    getPlotFromDb({name: "MDS",type: "mds",subtype: "generic"},"mds_container");
});
$("#_readnoise_trigger").on('click',function() {
    getPlotFromDb({name: "ReadNoise",type: "readnoise",subtype: "generic"},
        "readsnoise_container");
});
$("#_filtered_trigger").on('click',function() {
    getPlotFromDb({name: "filtered_chromosome",type: "filtered",
        subtype: "chromosome"},"chromosome_distr_container");
    getPlotFromDb({name: "filtered_biotype",type: "filtered",
        subtype: "biotype"},"biotype_distr_container");
});
$("#_boxplot_trigger").on('click',function() {
    getPlotFromDb({name: "Boxplot",type: "boxplot",subtype: "unorm"},
        "boxplot_unorm_container");
    getPlotFromDb({name: "Boxplot",type: "boxplot",subtype: "norm"},
        "boxplot_norm_container");
});
$("#_gcbias_trigger").on('click',function() {
    getPlotFromDb({name: "GCBias",type: "gcbias",subtype: "unorm"},
        "gcbias_unorm_container");
    getPlotFromDb({name: "GCBias",type: "gcbias",subtype: "norm"},
        "gcbias_norm_container");
});
$("#_lengthbias_trigger").on('click',function() {
    getPlotFromDb({name: "LengthBias",type: "lengthbias",subtype: "unorm"},
        "lengthbias_unorm_container");
    getPlotFromDb({name: "LengthBias",type: "lengthbias",subtype: "norm"},
        "lengthbias_norm_container");
});
$("#_meanvar_trigger").on('click',function() {
    getPlotFromDb({name: "MeanVar",type: "meanvar",subtype: "unorm"},
        "meanvar_unorm_container");
    getPlotFromDb({name: "MeanVar",type: "meanvar",subtype: "norm"},
        "meanvar_norm_container");
});
$("#_rnacomp_trigger").on('click',function() {
    getPlotFromDb({name: "RnaComp",type: "rnacomp",subtype: "unorm"},
        "rnacomp_unorm_container");
    getPlotFromDb({name: "RnaComp",type: "rnacomp",subtype: "norm"},
        "rnacomp_norm_container");
});

// Initialize some by changing plots (harmless for dexie, required for sqlite)
$('#_mds_trigger').click();
$("#_readnoise_trigger").click();
$("#_filtered_trigger").click();
$("#_boxplot_trigger").click();
$("#_gcbias_trigger").click();
$("#_lengthbias_trigger").click();
$("#_meanvar_trigger").click();
$("#_rnacomp_trigger").click();

$('#biodetection_selector').trigger('change');
$('#countsbio_sample_selector').trigger('change');
$('#countsbio_biotype_selector').trigger('change');
$('#saturation_sample_selector').trigger('change');
$('#saturation_biotype_selector').trigger('change');
$('#pairwise_selector').trigger('change');
$('#meandiff_selector').trigger('change');
$("#volcano_selector").trigger('change');
$("#mastat_selector").trigger('change');
$("#deregulo_selector").trigger('change');
$("#biodist_selector").trigger('change');
if ($("#statvenn_contrast_selector").length > 0) {
    $("#statvenn_contrast_selector").trigger('change');
}
if ($("#foldvenn_algo_selector").length > 0) {
    $("#foldvenn_algo_selector").trigger('change');
}
```

\name{diagplotCor}
\alias{diagplotCor}
\title{Summarized correlation plots}
\usage{
    diagplotCor(mat, type = c("heatmap", "correlogram"),
        output = "x11", path = NULL, ...)
}
\arguments{
    \item{mat}{the read counts matrix or data frame.}

    \item{type}{create heatmap of correlogram plots.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"} or \code{"ps"}.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filename of the pairwise comparisons plot produced if
    it's a file.
}
\description{
    This function uses the read counts matrix to create
    heatmap or correlogram correlation plots.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
diagplotCor(dataMatrix,type="heatmap")
diagplotCor(dataMatrix,type="correlogram")
}
\author{
    Panagiotis Moulos
}

\name{metaseqrPlot}
\alias{metaseqrPlot}
\title{Diagnostic plots for the metaseqR2 package}
\usage{
    metaseqrPlot(object, sampleList, annotation = NULL,
        contrastList = NULL, pList = NULL,
        thresholds = list(p = 0.05, f = 1),
        plotType = c("mds", "biodetection", "countsbio",
            "saturation", "readnoise", "rnacomp", "correl",
            "pairs", "boxplot", "gcbias", "lengthbias",
            "meandiff", "meanvar", "deheatmap", "volcano",
            "biodist", "filtered", "mastat", "deregulogram",
            "statvenn", "foldvenn"),
        isNorm = FALSE, output = "x11", path = NULL, ...)
}
\arguments{
    \item{object}{a matrix or a data frame containing count
    data derived before or after the normalization procedure,
    filtered or not by the metaseqR2's filters and/or p-value.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{annotation}{a data frame containing annotation
    elements for each row in object. See also Details.}

    \item{contrastList}{the vector of contrasts as defined in 
    the main help page of \code{\link{metaseqr2}}.}

    \item{pList}{a list of p-values for each contrast as
    obtained from any of the \code{stat*} methods of the
    metaseqr package. See also Details.}

    \item{thresholds}{a list with the elements \code{"p"} and
    \code{"f"} which are the p-value and the fold change
    cutoff when \code{diagplotType="volcano"}.}

    \item{plotType}{one or more of the diagnostic plots
    supported in metaseqR2 package. See also Details.}

    \item{isNorm}{a logical indicating whether object
    contains raw or normalized data. It is not essential and
    it serves only plot annotation purposes.}

    \item{output}{one or more R plotting device to direct the
    plot result to. See Details.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    A named list containing the file names of the produced
    plots. Each list member is names according to the
    selected plotting device and is also a named list, whose
    names are the plot types. The final contents are the file
    names in case the plots are written to a physical
    location (not meaningful for \code{"x11"}).
}
\description{
    This is the main function for producing sructured quality
    control and informative graphs base on the results of the
    various steps of the metaseqR package. The graphs
    produced span a variety of issues like good sample
    reproducibility (Multi-Dimensional Scaling plot, biotype
    detection, heatmaps. diagplotMetaseqr, apart from
    implementing certain package-specific plots, is a wrapper
    around several diagnostic plots present in other RNA-Seq
    analysis packages such as EDASeq and NOISeq.
}
\details{
    Regarding \code{object}, the object can be fed to any of 
    the \code{diagplotMetaseqr} plotting systems but not every
    plot is meaningful. For example, it's meaningless to
    create a \code{"biodist"} plot for a count matrix before
    normalization or statistical testing.
    
    Regarding \code{annotation}, usually, it is a subset of 
    the annotation obtained by \code{\link{getAnnotation}} or
    a subset of possibly embedded annotation with the input
    counts table. This parameter is optional and required
    only when diagplotType is any of \code{"biodetection"},
    \code{"countsbio"}, \code{"saturation"},
    \code{"rnacomp"}, \code{"readnoise"}, \code{"biodist"},
    \code{"gcbias"}, \code{"lengthbias"} or
    \code{"filtered"}.
    
    Regarding \code{contrastList}, this parameter is optional 
    and required only when \code{diagplotType} is any of
    \code{"deheatmap"}, \code{"volcano"} or \code{"biodist"}.
    It can also be a named structured list of contrasts as 
    returned by the internal function 
    \code{metaseqR2:::makeContrastList}.
    
    Regarding \code{diagplotType}, many of these plots
    require the presence of additional package, something
    that is checked while running the main metaseqr2 function.
    The supported plots are \code{"mds"}, \code{"biodetection"},
    \code{"countsbio"}, \code{"saturation"}, \code{"rnacomp"},
    \code{"boxplot"}, \code{"gcbias"}, \code{"lengthbias"},
    \code{"meandiff"}, \code{"meanvar"}, \code{"deheatmap"},
    \code{"volcano"}, \code{"biodist"}, \code{"filtered"}, 
    \code{"readnoise"}, \code{"venn"}, \code{"correl"},
    \code{"pairwise"}. For a brief description of these plots 
    please see the main \code{\link{metaseqr2}} help page.
    
    Regarding \code{pList}, this parameter is optional 
    and required only when \code{diagplotType} is any of
    \code{"deheatmap"}, \code{"volcano"} or \code{"biodist"}.
    
    Regarding \code{output}, supported mechanisms are: 
    \code{"png"}, \code{"jpg"}, \code{"bmp"}, \code{"pdf"},
    \code{"ps"} or \code{"json"}. The latter is currently 
    available for the creation of interactive volcano plots 
    only when reporting the output, through the highcharts 
    javascript library. The default plotting (\code{"x11"}) 
    is not supported due to instability in certain devices.
}
\note{
    In order to make the best out of this function, you
    should generally provide the annotation argument as most
    and also the most informative plots depend on this. If
    you don't know what is inside your counts table or how
    many annotation elements you can provide by embedding it,
    it's always best to setup a local databse so as to use
    predefined annotations that work better with the
    functions of the whole package.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
contrast <- "A_vs_B"
metaseqrPlot(dataMatrix,sampleList,plotType=c("mds","boxplot"))

normArgs <- getDefaults("normalization","deseq2")
object <- normalizeDeseq2(dataMatrix,sampleList,normArgs)
metaseqrPlot(object,sampleList,plotType="boxplot")

## More
#p <- statDeseq2(object,sampleList)
#metaseqrPlot(object,sampleList,contrastList=contrast,pList=p,
#    plotType="volcano")
}
\author{
    Panagiotis Moulos
}

\name{getDefaults}
\alias{getDefaults}
\title{Default parameters for several metaseqr functions}
\usage{
    getDefaults(what, method = NULL)
}
\arguments{
    \item{what}{a keyword determining the procedure for which
    to fetch the default settings according to method
    parameter. It can be one of \code{"normalization"},
    \code{"statistics"}, \code{"geneFilter"},
    \code{"exonFilter"} or \code{"biotypeFilter"}.}

    \item{method}{the supported algorithm included in
    metaseqR for which to fetch the default settings.
    Se Details.}
}
\value{
    A list with default setting that can be used directly in
    the call of metaseqr.
}
\description{
    This function returns a list with the default settings
    for each filtering, statistical and normalization
    algorithm included in the metaseqR package. See the
    documentation of the main function and the documentation
    of each statistical and normalization method for details.
}
\details{
    When \code{what} is \code{"normalization"}, method is 
    one of \code{"edaseq"}, \code{"deseq"}, \code{"edger"},
    \code{"noiseq"} or \code{"nbpseq"}. When \code{what} is
    \code{"statistics"}, method is one of \code{"deseq"},
    \code{"edger"}, \code{"noiseq"}, \code{"limma"},
    or \code{"nbpseq"}. When \code{method} is
    \code{"biotypeFilter"}, \code{what} is the input
    organism (see the main \code{\link{metaseqr2}} help page
    for a list of supported organisms).
}
\examples{
normArgsEdaseq <- getDefaults("normalization","edaseq")
statArgsEdger <- getDefaults("statistics","edger")
}
\author{
    Panagiotis Moulos
}

\name{metaTest}
\alias{metaTest}
\title{Meta-analysis using several RNA-Seq statistics}
\usage{
    metaTest(cpList,
        metaP = c("simes", "bonferroni", "fisher", "harmonic",
        "dperm_min", "dperm_max", "dperm_weight", "fperm", 
        "whitlock", "minp", "maxp", "weight", "pandora", 
        "none"), counts, sampleList, statistics, statArgs, 
        libsizeList, nperm = 10000, 
        weight = rep(1/length(statistics), length(statistics)), 
        pOffset = NULL, rc = NULL)
}
\arguments{
    \item{cpList}{a named list whose names are the contrasts
    requested from metaseqr2. Each member is a p-value matrix
    whose colnames are the names of the statistical tests
    applied to the data. See the main \code{\link{metaseqr2}}
    help page.}

    \item{metaP}{the p-value combination method to use. See
    the main \code{\link{metaseqr2}} help page.}

    \item{counts}{the normalized and possibly filtered read
    counts matrix. See the main \code{\link{metaseqr2}} help
    page.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition. See the main
    \code{\link{metaseqr2}} help page.}

    \item{statistics}{the statistical algorithms used in
    metaseqr2. See the main \code{\link{metaseqr2}} help page.}

    \item{statArgs}{the parameters for each statistical
    argument. See the main \code{\link{metaseqr2}} help page.}

    \item{libsizeList}{a list with library sizes. See the
    main \code{\link{metaseqr2}} and the \code{stat*} help
    pages.}

    \item{nperm}{the number of permutations (Monte Carlo
    simulations) to perform.}

    \item{weight}{a numeric vector of weights for each
    statistical algorithm.}
    
    \item{pOffset}{\code{NULL} (default) or a fixed 
    numeric value between 0 and 1. See also the main 
    \code{\link{metaseqr2}} man page.}

    \item{rc}{the fraction of the available cores to use
    in a multicore system.}
}
\value{
    A named list with combined p-values. The names are the
    contrasts and the list members are combined p-value
    vectors, one for each contrast.
}
\description{
    This function calculates the combined p-values when
    multiple statistical algorithms are applied to the input
    dataset. It is a helper and it requires very specific
    arguments so it should not be used individually
}
\details{
    Ideally one would want to create the same set of indices 
    for a given dataset so as to create reproducible p-values. 
    To achieve this, use the \code{set.seed} function prior 
    to any calculations.
}
\examples{
cpList <- list(a=matrix(runif(100),50,2))
metaP <- metaTest(cpList,"simes")
}
\author{
    Panagiotis Moulos
}

\name{loadAnnotation}
\alias{loadAnnotation}
\title{Load a metaseqR2 annotation element}
\usage{
    loadAnnotation(genome, refdb, 
        level = c("gene", "transcript", "exon"),
        type = c("gene", "exon", "utr"), version="auto",
        db = file.path(system.file(package = "metaseqR2"),
            "annotation.sqlite"), summarized = FALSE, 
            asdf = FALSE, rc = NULL)
}
\arguments{
    \item{genome}{a \code{\link{metaseqr2}} supported
    organisms or a custom, imported by the user, name. See 
    also the main \code{\link{metaseqr2}} man page.}

    \item{refdb}{a \code{\link{metaseqr2}} supported
    annotation source or a custom, imported by the user, name.
    See also the main \code{\link{metaseqr2}} man page.}
    
    \item{level}{same as the \code{transLevel} in 
    \code{\link{metaseqr2}}.}
    
    \item{type}{same as the \code{countType} in 
    \code{\link{metaseqr2}}.}
    
    \item{version}{same as the \code{version} in 
    \code{\link{metaseqr2}}.}
    
    \item{db}{same as the \code{db} in 
    \code{\link{buildAnnotationDatabase}}.}
    
    \item{summarized}{if \code{TRUE}, retrieve summarized,
    non-overlaping elements where appropriate (e.g. exons).}
    
    \item{asdf}{return the result as a \code{\link{data.frame}}
    (default \code{FALSE}).}
    
    \item{rc}{same as the \code{rc} in 
    \code{\link{buildAnnotationDatabase}}.}
}
\value{
    The function returns a \code{GenomicRanges} object with
    the requested annotation.
}
\description{
    This function creates loads an annotation element from
    the local annotation database to be used with metaseqr2.
    If the annotation is not found and the organism is 
    supported, the annotation is created on the fly but not
    imported in the local database. Use
    \code{buildAnnotationDatabase} for this purpose.
    
}
\examples{
db <- file.path(system.file(package="metaseqR2"),
    "annotation.sqlite")
if (file.exists(db))
    gr <- loadAnnotation(genome="hg19",refdb="ensembl",
        level="gene",type="gene",db=db)
}
\author{
    Panagiotis Moulos
}

\name{metaseqr2}
\alias{metaseqr2}
\title{The main metaseqr2 pipeline}
\usage{
    metaseqr2(counts, sampleList, excludeList = NULL,
        fileType = c("auto", "sam", "bam", "bed"),
        path = NULL, contrast = NULL, libsizeList = NULL,
        embedCols = list(idCol = 4, gcCol = NA, nameCol = NA, 
            btCol = NA),
        annotation = NULL, org = c("hg18", "hg19", "hg38", "mm9", 
            "mm10", "mm39", "rn5", "rn6", "dm3", "dm6", "danrer7", 
            "pantro4", "susscr3", "tair10", "equcab3"),
        refdb = c("ensembl", "ucsc", "refseq"), version = "auto",
        transLevel = c("gene", "transcript", "exon"),
        countType = c("gene", "exon","utr"),
        utrOpts = list(frac = 1, minLength = 300, downstream = 50),
        exonFilters = list(minActiveExons = list(exonsPerGene = 5, 
            minExons = 2, frac = 1/5)),
        geneFilters = list(length = list(length = 500), 
            avgReads = list(averagePerBp = 100, quantile = 0.25), 
            expression = list(median = TRUE, mean = FALSE,  
                quantile = NA, known = NA, custom = NA), 
            biotype = getDefaults("biotypeFilter", org[1]),
            presence = list(frac = 0.25, minCount = 10,
                perCondition = FALSE)),
        whenApplyFilter = c("postnorm", "prenorm"),
        normalization = c("deseq", "deseq2", "edaseq", "edger",
            "noiseq", "nbpseq", "absseq", "dss", "each", "none"),
        normArgs = NULL,
        statistics = c("deseq", "deseq2", "edger", "noiseq", 
            "limma", "nbpseq", "absseq", "dss"),
        statArgs = NULL,
        adjustMethod = sort(c(p.adjust.methods, "qvalue")),
        metaP = if (length(statistics) > 1) c("simes", 
            "bonferroni", "fisher", "dperm_min", "dperm_max",
            "dperm_weight", "fperm", "whitlock", "minp", "maxp",
            "weight", "pandora", "none") else "none",
        weight = rep(1/length(statistics), length(statistics)),
        nperm = 10000, pcut = NA, logOffset = 1, pOffset = NULL, 
        preset = NULL, qcPlots = c("mds", "biodetection", 
            "countsbio", "saturation", "readnoise","filtered", 
            "correl", "pairwise", "boxplot", "gcbias", 
            "lengthbias", "meandiff", "meanvar", "rnacomp", 
            "deheatmap", "volcano", "biodist", "mastat", 
            "statvenn", "foldvenn", "deregulogram"),
        figFormat = c("png", "jpg", "tiff", "bmp", "pdf", "ps"),
        outList = FALSE, exportWhere = NA,
        exportWhat = c("annotation", "p_value", "adj_p_value", 
            "meta_p_value", "adj_meta_p_value", "fold_change", 
            "stats", "counts","flags"),
        exportScale = c("natural", "log2", "log10", "vst", 
            "rpgm"),
        exportValues = c("raw", "normalized"),
        exportStats = c("mean", "median", "sd", "mad", "cv", 
            "rcv"),
        exportCountsTable = FALSE,
        restrictCores = 0.6, report = TRUE, reportTop = 0.1,
        reportTemplate = "default", saveGeneModel = TRUE,
        verbose = TRUE, runLog = TRUE, 
        reportDb = c("dexie", "sqlite"), 
        localDb = file.path(system.file(package = "metaseqR2"),
            "annotation.sqlite"),
        offlineReport = TRUE, 
        createTracks = FALSE, overwriteTracks = FALSE, 
        trackExportPath = file.path(exportWhere, "tracks"),
        trackInfo = list(stranded = FALSE, normTo = 1e+9, 
            urlBase = "http://www.trackserver.me",
            fasta = NULL, gtf = NULL,
            hubInfo = list(name = "MyHub", shortLabel = "My hub",
            longLabel = "My hub long", 
            email = "someone@example.com")), .progressFun = NULL,
            .exportR2C = FALSE,...)
}
\arguments{
    \item{counts}{a text tab-delimited file containing gene,
    exon or 3'UTR counts in one of the following formats: i) the
    first column contains unique gene or exon identifiers and
    the rest of the columns contain the read counts for each
    sample. ii) The first n columns should contain only **gene**
    annotation elements like chromosomal locations, gene
    accessions, exon accessions, GC content etc. and the rest
    columns should contain gene read counts, iii) \code{counts} 
    can also be an \code{.RData} file with previous analysis 
    elements (see Details) and iv) counts can be a list 
    representing the gene model (see Details). Several 
    restrictions apply im each of the four cases. See Details 
    for analytical descriptions.}

    \item{sampleList}{a list containing condition names and
    the samples under each condition or a small tab-delimited 
    file with the experiment description. Not needed when
    restoring a previous analysis. See Details for analytical 
    description.}
    
    \item{excludeList}{ a list of samples to exclude, in the 
    same (list) format as \code{sampleList} above.}

    \item{path}{an optional path where all the BED/BAM files
    are placed, to be prepended to the BAM/BED file names in
    the targets file. See Details for further information.}

    \item{fileType}{the type of raw input files. It can be
    \code{"auto"} for auto-guessing, \code{"bed"} for BED
    files, \code{"sam"} for SAM files or \code{"bam"} for BAM
    files.}

    \item{contrast}{a character vector of contrasts to be
    tested in the statistical testing step(s) of the pipeline.
    Each element of contrast should STRICTLY have the format
    "ConditionA_vs_ConditionB_vs_...". Special attention is
    needed as fold change calculations are based on this
    argument. If it is \code{NULL}, no statistical testing
    of fold change calculations are performed. See Details 
    for further information.}

    \item{libsizeList}{an optional named list where names
    represent samples (MUST be the same as the samples in
    \code{sample.list}) and members are the library sizes
    (the sequencing depth) for each sample. For example
    \code{libsize.list <- list(Sample_A1=32456913,}
    \code{Sample_A2=4346818)}.}
    
    \item{embedCols}{a named list with column numbers to 
    guide the case of embedded annotation. See Details for
    further information.}

    \item{annotation}{It can be one of i) \code{NULL} 
    (default) to use the existing annotation database 
    or fetch on the fly, ii) \code{"embedded"} if the
    annotation elements are embedded in the read counts file
    (restrictions apply) or iii) a list with a path to a GTF
    file and certain required metadata. See Details for a full
    description.}
    
    \item{org}{the supported organisms by \code{metaseqr2} or
    a user-named organism which has been imported to the 
    database. See Details for more information.} 

    \item{refdb}{the reference annotation repository from
    which to retrieve annotation elements to use with
    metaseqr2. It can be one of \code{"ensembl"} (default),
    \code{"ucsc"} or \code{"refseq"} or a user based one 
    (similar to the \code{org} argument).}
    
    \item{version}{the version of the annotation to use. See
    Details.}
    
    \item{transLevel}{perform differential expression 
    analysis at which transcriptional unit, can be one of
    \code{"gene"} (default), \code{"transcript"} for 
    reporting differential expression at the transcript
    level or \code{"exon"} for exon level.}
    
    \item{countType}{the type of reads inside the counts
    file. It can be one of \code{"gene"}, \code{"exon"} or
    \code{"utr"} for Quant-Seq (Lexogen) protocol. This is
    a very important and mandatory parameter as it defines
    the course of the workflow.}
    
    \item{utrOpts}{a named list with members \code{frac} 
    which is the fraction (0-1) of the 3' UTR region to count 
    reads in, \code{minLength} the minimum acceptable 3'UTR
    length irrespective of \code{frac} and \code{downstream} 
    the number of base pairs to flank the end of the 3' UTR of 
    transcripts when analyzing Quant-Seq data.}

    \item{exonFilters}{a named list whose names are the
    names of the supported exon filters and its members the
    filter parameters. See section "Exon filters" below for
    details.}

    \item{geneFilters}{a named list whose names are the
    names of the supported gene filters and its members the
    filter parameters. See section "Gene filters" below for
    details.}

    \item{whenApplyFilter}{a character string determining
    when to apply the exon and/or gene filters, relative to
    normalization. It can be \code{"prenorm"} to apply apply
    the filters and exclude genes from further processing
    before normalization, or \code{"postnorm"} to apply the
    filters after normalization (default). See also Details.}

    \item{normalization}{the normalization algorithm to be
    applied on the count data. It can be one of
    \code{"edaseq"} for EDASeq normalization, \code{"deseq"}
    for the normalization algorithm in the DESq package 
    (default), \code{"edger"} for the normalization algorithms 
    present in the edgeR package \code{"noiseq"} for the 
    normalization algorithms present in the NOISeq package 
    \code{"nbpseq"} for the normalization algorithms present 
    in the NBPSeq package or \code{"none"} to not normalize
    the data (highly unrecommended). Algorithm specific 
    arguments can be passed through the \code{normArgs} 
    argument).}

    \item{normArgs}{a named list whose names are the names
    of the normalization algorithm parameters and its members
    parameter values. See section "Normalization parameters"
    below for details. Leave \code{NULL} for the defaults of
    \code{normalization}.}

    \item{statistics}{one or more statistical analyses to be
    performed by the metaseqr2 pipeline. It can be one or more
    of \code{"deseq"} (default) to conduct statistical
    test(s) implemented in the DESeq package, \code{"edger"}
    to conduct statistical test(s) implemented in the edgeR
    package, \code{"limma"} to conduct the RNA-Seq version of
    statistical test(s) implemented in the limma package,
    \code{"noiseq"} to conduct statistical test(s)
    implemented in the NOISeq package, \code{"nbpseq"} to 
    conduct statistical test(s) implemented in the NBPSeq 
    package, \code{"deseq2"} to conduct statistical test(s)
    implemented in the DESeq2 package, \code{"dss"} to 
    conduct statistical test(s) implemented in the DSS
    package and \code{"absseq"} to conduct statistical 
    test(s) implemented in the ABSSeq package. In any case 
    individual algorithm parameters are controlled by the 
    contents of the \code{statArgs} list. Finally, it can 
    be \code{NA}. In this case no testing is performed and 
    only fold changes are provided if \code{contrast} is 
    not \code{NULL}.}

    \item{statArgs}{a named list whose names are the names
    of the statistical algorithms used in the pipeline. Each
    member is another named list whose names are the
    algorithm parameters and its members are the parameter
    values. See section "Statistics parameters" below for
    details. Leave \code{NULL} for the defaults of
    \code{statistics}.}

    \item{adjustMethod}{the multiple testing p-value
    adjustment method. It can be one of
    \code{\link{p.adjust.methods}} or \code{"qvalue"} from
    the qvalue Bioconductor package. Defaults to \code{"BH"}
    for Benjamini-Hochberg correction.}

    \item{metaP}{the meta-analysis method to combine
    p-values from multiple statistical tests . It can be 
    one of \code{"simes"} (default), \code{"bonferroni"}, 
    \code{"minp"}, \code{"maxp"}, \code{"weight"}, 
    \code{"pandora"}, \code{"dperm_min"}, \code{"dperm_max"}, 
    \code{"dperm_weight"}, \code{"fisher"}, \code{"fperm"}, 
    \code{"whitlock"} or \code{"none"}. See Details for a full
    description.}

    \item{weight}{a vector of weights with the same length as
    the \code{statistics} vector containing a weight for each
    statistical test. It should sum to 1. \strong{Use with
    caution with the} \code{dperm_weight} \strong{parameter!
    Theoretical background is not yet} \strong{solid and only
    experience shows improved results!}}

    \item{nperm}{the number of permutations performed to
    derive the meta p-value when \code{metaP="fperm"} or
    \code{metaP="dperm"}. It defaults to 10000.}

    \item{pcut}{a p-value cutoff for exporting differentially
    genes, default is to export all the non-filtered genes.}

    \item{logOffset}{an offset to be added to values during
    logarithmic transformations in order to avoid Infinity
    (default is \code{1}).}
    
    \item{pOffset}{a value between \code{0} and \code{1} 
    to multiply potential zero p-values with for the 
    combination methods including weighting or \code{NULL} 
    (default). See also Details.}

    \item{preset}{an analysis strictness preset.
    \code{preset} can be one of \code{"all_basic"},
    \code{"all_normal"}, \code{"all_full"},
    \code{"medium_basic"}, \code{"medium_normal"},
    \code{"medium_full"}, \code{"strict_basic"},
    \code{"strict_normal"} or \code{"strict_full"}, each of
    which control the strictness of the analysis and the
    amount of data to be exported. For an explanation of the
    presets, see the section "Presets" below.}

    \item{qcPlots}{a set of diagnostic plots to show/create.
    It can be one or more of \code{"mds"},
    \code{"biodetection"}, \code{"rnacomp"},
    \code{"countsbio"}, \code{"saturation"},
    \code{"readnoise"}, \code{"filtered"}, \code{"boxplot"},
    \code{"gcbias"}, \code{"lengthbias"}, \code{"meandiff"},
    \code{"meanvar"}, \code{"deheatmap"}, \code{"volcano"},
    \code{"mastat"}, \code{"biodist"}, \code{"statvenn"},
    \code{"foldvenn"}. See also Details.}

    \item{figFormat}{the format of the output diagnostic
    plots. It can be one or more of \code{"png"},
    \code{"jpg"}, \code{"tiff"}, \code{"bmp"}, \code{"pdf"},
    \code{"ps"}. The native format \code{"x11"} (for direct
    display) is not provided as an option as it may not
    render the proper display of some diagnostic plots in
    some devices.}

    \item{outList}{a logical controlling whether to export a
    list with the results in the running environment.}

    \item{exportWhere}{an output directory for the project
    results (report, lists, diagnostic plots etc.)}

    \item{exportWhat}{the content of the final lists. It can
    be one or more of \code{"annotation"}, to bind the
    annoation elements for each gene, \code{"p_value"}, to
    bind the p-values of each method, \code{"adj_p_value"},
    to bind the multiple testing adjusted p-values,
    \code{"meta_p_value"}, to bind the combined p-value from
    the meta-analysis, \code{"adj_meta_p_value"}, to bind the
    corrected combined p-value from the meta-analysis,
    \code{"fold_change"}, to bind the fold changes of each
    requested contrast, \code{"stats"}, to bind several
    statistics calclulated on raw and normalized counts (see
    the \code{exportStats} argument), \code{"counts"}, to
    bind the raw and normalized counts for each sample.}

    \item{exportScale}{export values from one or more
    transformations applied to the data. It can be one or
    more of \code{"natural"}, \code{"log2"}, \code{"log10"},
    \code{"vst"} (Variance Stabilizing Transormation, see the
    documentation of DESeq package) and \code{"rpgm"} which
    is ratio of mapped reads per gene model (either the gene
    length or the sum of exon lengths, depending on 
    \code{countType} argument). Note that this is not RPKM
    as reads are already normalized for library size using 
    one of the supported normalization methods. Also, 
    \code{"rpgm"} might be misleading when \code{normalization} 
    is other than \code{"deseq"}.}

    \item{exportValues}{It can be one or more of
    \code{"raw"} to export raw values (counts etc.) and
    \code{"normalized"} to export normalized counts.}

    \item{exportStats}{calculate and export several
    statistics on raw and normalized counts, condition-wise.
    It can be one or more of \code{"mean"}, \code{"median"},
    \code{"sd"}, \code{"mad"}, \code{"cv"} for the
    Coefficient of Variation, \code{"rcv"} for a robust
    version of CV where the median and the MAD are used
    instead of the mean and the standard deviation.}

    \item{exportCountsTable}{exports also the calculated 
    read counts table when input is read from bam files 
    and exports also the normalized count table in all 
    cases. Defaults to \code{FALSE}.}

    \item{restrictCores}{in case of parallel execution of
    several subfunctions, the fraction of the available cores
    to use. In some cases if all available cores are used
    (\code{restrictCores=1} and the system does not have
    sufficient RAM, the pipeline running machine might
    significantly slow down.}

    \item{report}{a logical value controlling whether to
    produce a summary report or not. Defaults to
    \code{TRUE}.}

    \item{reportTop}{a fraction of top statistically 
    significant genes to append to the HTML report. This 
    helps in keeping the size of the report as small as 
    possible, as appending the total gene list might 
    create a huge HTML file. Users can always retrieve 
    the whole gene lists from the report links. Defaults 
    to \code{0.1} (top 10% of statistically significant 
    genes). Set to \code{NA} or \code{NULL} to append all 
    the statistically significant genes to the HTML report.}

    \item{reportTemplate}{an HTML template to use for the
    report. Do not change this unless you know what you are
    doing.}

    \item{saveGeneModel}{in case of exon analysis, a list
    with exon counts for each gene will be saved to the file
    \code{exportWhere/data/gene_model.RData}. This file can
    be used as input to metaseqR for exon count based 
    analysis, in order to avoid the time consuming step of 
    assembling the counts for each gene from its exons}

    \item{verbose}{print informative messages during
    execution? Defaults to \code{TRUE}.}

    \item{runLog}{write a log file of the \code{metaseqr2}
    run using package log4r. Defaults to \code{TRUE}. The
    filename will be auto-generated.}
    
    \item{reportDb}{database system to use for storing the
    report intereactive graphs. Can be \code{"sqlite"} (default)
    or \code{"dexie"}. See Details for further explanation on
    what should be used.}
    
    \item{localDb}{the metaseqR2 annotaation database location.
    See also \code{link{buildAnnotationDatabase}}.}
    
    \item{offlineReport}{\code{TRUE} (default) to download
    and include the required JavaScript libraries to
    properly view the report offline. Ignored if
    \code{report=FALSE}}.
    
    \item{createTracks}{option to create normalized bigWig 
    files to display in a genome browser (e.g. UCSC). Defaults
    to \code{FALSE}.}
    
    \item{overwriteTracks}{overwrite tracks if they already
    exist? Defaults to \code{FALSE}.}
    
    \item{trackExportPath}{where to export the bigWig files, 
    defaults to \code{file.path(exportWhere,"tracks")}.}
    
    \item{trackInfo}{if \code{createTracks=TRUE}, a list with
    additional required information to create the tracks. See
    Details for further explanation.}
    
    \item{.progressFun}{a function which updates a 
    \code{Progress} object from shiny. This function must
    accept a \code{detail} argument. See 
    http://shiny.rstudio.com/articles/progress.html}
    
    \item{.exportR2C}{export additional RData along with
    \code{saveGeneModel}.}

    \item{...}{further arguments that may be passed to
    plotting functions, related to \code{\link{par}}.}
}
\value{
    If \code{outList} is \code{TRUE}, a named list whose
    length is the same as the number of requested contrasts.
    Each list member is named according to the corresponding
    contrast and contains a data frame of differentially
    expressed genes for that contrast. The contents of the
    data frame are defined by the \code{exportWhat,
    exportScale, exportStats, exportValues} parameters. If
    \code{report} is \code{TRUE}, the output list contains
    two main elements. The first is described above (the
    analysis results) and the second contains the same
    results but in HTML formatted tables.
}
\description{
    This function is the main metaseqr2 workhorse and
    implements the main metaseqr2 workflow which performs data
    read, filtering, normalization and statistical selection,
    creates diagnostic plots and exports the results and a
    report if requested. The metaseqr2 function is responsible
    for assembling all the steps of the metaseqr2 pipeline
    which i) reads the input gene or exon read count table
    ii) performs prelimininary filtering of data by removing
    chrM and other non-essential information for a typical
    differential gene expression analysis as well as a
    preliminary expression filtering based on the exon
    counts, if an exon read count file is provided. iii)
    performs data normalization with one of currently widely
    used algorithms, including EDASeq (Risso et al., 2011),
    DESeq (Anders and Huber, 2010), edgeR (Robinson et al.,
    2010), NOISeq (Tarazona et al., 2012) or no normalization
    iv) performs a second stage of filtering based on the
    normalized gene expression according to several gene
    filters v) performs statistical testing with one or more
    of currently widely used algorithms, including DESeq
    (Anders and Huber, 2010), edgeR (Robinson et al., 2010),
    NOISeq (Tarazona et al., 2012), limma (Smyth et al.,
    2005) for RNA-Seq data vi) in the case of multiple 
    statistical testing algorithms, performs meta-analysis 
    using one of six vailable methods (see the meta.p argument) 
    vii) exports the resulting differentially expressed gene 
    list in text tab-delimited format viii) creates a set of 
    diagnostic plots either available in the aforementioned 
    packages or metaseqr2 specific ones and ix) creates a 
    comprehensive HTML report which summarizes the run 
    information, the results and the diagnostic plots. Certain 
    diagnostic plots (e.g. the volcano plot) can be interactive 
    with the use of the external Highcharts
    (http://www.highcharts.com) JavaScript library for
    interactive graphs. Although the inputs to the metaseqr2
    workflow are many, in practice, setting only very few of
    them and accepting the defaults as the rest can result in
    quite comprehensible results for mainstream organisms
    like mouse, human, fly and rat.
}
\details{
    When \code{counts} is a tab-delimited file, the following
    restrictions apply:
    \itemize{ 
        \item In the case (i) the first cell of each row is a 
        gene or exon accession and the rest are integers 
        representing the counts for that accession. In that 
        case, the \code{annotation} parameter should strictly 
        be \code{NULL} or an external file in GTF format.
        \item In the case (ii) the \code{annotation} parameter 
        can also be \code{"embedded"}. The ideal embedded 
        annotation contains 8 columns, chromosome, gene or 
        exon start, gene or exon end, gene or exon accession, 
        GC-content (fraction or percentage), strand, HUGO 
        gene symbol and gene biotype (e.g. "protein_coding" 
        or "ncRNA"). When the \code{annotation} parameter is 
        \code{"embedded"}, certain of these features are 
        mandatory (co-ordinates and accessions). If they are 
        not present, the pipeline will not run. If additional 
        elements are not present (e.g. GC content or biotypes), 
        certain features of metaseqr2 will not be available. 
        For example, EDASeq normalization will not be 
        performed based on a GC content covariate but based 
        on gene length which is not what the authors of
        EDASeq suggest. If biotypes are not present, a lot of
        diagnostic plots will not be available. If the HUGO 
        gene symbols are missing, the final annotation will 
        contain only gene accessions and thus be less 
        comprehensible. Counts can be a data frame satisfying 
        the above conditions. It is a data frame by default 
        when \code{read2count} is used.
        \item In the case (iii)  the .RData file 
        (output of \code{\link{save}} function contains 
        static input elements (list containing the gene 
        model (exon counts for each gene), gene and exon 
        annotation to avoid re-(down)loading and/or gene 
        counts depending on \code{countType}). This kind of 
        input facilitates the re-analysis of the same 
        experiment, using different filtering, normalization 
        and statistical algorithms. This \code{.RData} file 
        is produced when \code{saveGeneModel=TRUE}.
        \item In the case (iv) \code{counts} can be a list 
        representing the gene model (exon/UTR counts for each 
        gene). This \code{.RData} file can be generated by 
        setting \code{saveGeneModel=TRUE} when performing data 
        analysis for the first time.
    }
    
    Regarding \code{sampleList} it should have the format 
    \code{sampleList <-} \code{list(ConditionA=c("Sample_A1",} 
    \code{"Sample_A2","Sample_A3"),} 
    \code{ConditionB=c("Sample_B1","Sample_B2"),} 
    \code{ConditionC=c("Sample_C1","Sample_C2"))}. The names of 
    the samples in list members MUST match the column names 
    containing the read counts in the counts file. If they do 
    not match, the pipeline will either crash or at best, ignore 
    several of your samples. Alternative, \code{sampleList} can 
    be a small tab-delimited file structured as follows: the 
    first line of the external tab delimited file should contain
    column names (names are not important). The first column MUST
    contain UNIQUE sample names and the second column MUST
    contain the biological condition where each of the samples
    in the first column should belong to. If the \code{counts}
    argument is missing, the \code{sampleList} argument MUST be 
    a targets text tab-delimited file which contains the sample 
    names, the BAM/BED file names and the biological 
    conditions/groups for each sample/file. The file should be 
    text tab-delimited and structured as follows: the first line 
    of the external tab delimited file should contain column 
    names (names are not important). The first column MUST 
    contain UNIQUE sample names. The second column MUST contain
    the raw BAM/BED files WITH their full path. Alternatively, 
    the \code{path} argument should be provided. If \code{path} 
    is not provided and if the files in the second column of 
    the targets file do not contain a path to a directory, the 
    current directory is assumed to be the BAM/BED file 
    container. The third column MUST contain the biological 
    condition where each of the samples in the first column 
    should belong to.
    
    Regarding \code{contrast}, a valid example based on the 
    \code{sampleList} above is 
    \code{contrast <- c("ConditionA_vs_ConditionB",}
    \code{"ConditionA_vs_ConditionC",}
    \code{"ConditionA_vs_ConditionB_vs_ConditionC")}. The
    first element of pairwise contrasts (e.g. "ConditionA"
    above) MUST be the control condition or any reference
    that ConditionB is checked against. metaseqr2 uses this
    convention to properly calculate fold changes.
    
    Regarding \code{embedCols}, this list must contain four
    members, named \code{idCol}, \code{gcCol}, \code{nameCol}
    and \code{btCol}, which hold the position in the delimited
    file with embedded annotation, where unique gene ids, GC
    content, gene names and gene biotypes respectively are
    located. More specifically:
    \itemize{
        \item \code{idCol} is an integer denoting the column 
        number in the file (or data frame) provided with the 
        counts argument, where the unique gene accessions are.
        Default to \code{4} which is the standard feature name
        column in a BED file.
        \item \code{gcCol} is an integer denoting the column 
        number in the file (or data frame) provided with the 
        \code{counts} argument, where each gene's GC content 
        is given. If not provided, GC content normalization 
        provided by EDASeq will not be available.
        \item \code{nameCol} is an integer denoting the column 
        number in the file (or data frame) provided with the 
        counts argument, where the HUGO gene symbols are given. 
        If not provided, it will not be available when reporting
        results. In addition, the \code{"known"} gene filter 
        will not be available for application.
        \item \code{btCol} is an integer denoting the column 
        number in the file (or data frame) provided with the
        counts argument, where the gene biotypes are given. 
        If not provided, the \code{"biodetection"}, 
        \code{"countsbio"}, \code{"saturation"}, 
        \code{"filtered"} and \code{"biodist"} plots will not 
        be available.
    }
    
    Regarding annotation instructs metaseqr2 where to find the
    annotation for the given counts file. It can be one of i)
    \code{"download"} (default) for automatic downloading of
    the annotation for the organism specified by the org
    parameter (using biomaRt), ii) \code{"embedded"} if the
    annotation elements are embedded in the read counts file
    or iv) a file specified by the user which should be as
    similar as possible to the \code{"download"} case, in
    terms of column structure.
    
    Regarding \code{org}, it can be, for human genomes 
    \code{"hg18"}, \code{"hg19"} or \code{"hg38"}, for mouse 
    genomes \code{"mm9"}, \code{"mm10"}, for rat genomes 
    \code{"rn5"} or \code{"rn6"}, for drosophila genome
    \code{"dm3"} or \code{"dm6"}, for zebrafish genome 
    \code{"danrer7"}, \code{"danrer10"} or \code{"danrer11"}, 
    for chimpanzee genome \code{"pantro4"}, \code{"pantro5"}, 
    for pig genome \code{"susscr3"}, \code{"susscr11"}, for 
    Arabidopsis thaliana genome \code{"tair10"} and for 
    Equus caballus genome \code{"equcab2"}. Finally, it can
    be \code{"USER_NAMED_ORG"} with a custom organism which
    has been imported to the annotation database by the user
    using a GTF file. For example \code{org="mm10_p1"}.
    
    Regarding \code{version}, this is an integer denoting the
    version of the annotation to use from the local annotation
    database or fetch on the fly. For Ensembl, it corresponds
    to Ensembl releases, while for UCSC/RefSeq, it is the
    date of creation (locally).
    
    Regarding \code{whenApplyFilter}, in the case of
    \code{whenApplyFilter="prenorm"}, a first normalization
    round is applied to a copy of the gene counts matrix in
    order to derive the proper normalized values that will
    constitute the several expression-based filtering
    cutoffs.
    
    Regarding \code{metaP}, for the \code{"fisher"} and 
    \code{"fperm"} methods, see the documentation of the R 
    package MADAM. For the \code{"whitlock"} method, see the 
    documentation of the survcomp Bioconductor package. With 
    the \code{"maxp"} option, the final p-value is the maximum 
    p-value out of those returned by each statistical test. 
    This is equivalent to an "intersection" of the results 
    derived from each algorithm so as to have a final list with 
    the common genes returned by all statistical tests. 
    Similarly, when \code{meta.p="minp"}, is equivalent to a 
    "union" of the results derived from each algorithm so as 
    to have a final list with all the genes returned by all 
    statistical tests. The latter can be used as a very lose 
    statistical threshold to aggregate results from all methods 
    regardless of their False Positive Rate. With the 
    \code{"simes"} option, the method proposed by Simes 
    (Simes, R. J., 1986) is used. With the \code{"dperm_min"}, 
    \code{"dperm.max"}, \code{"dperm.weight"} options, a 
    permutation procedure is initialed, where \code{nperm} 
    permutations are performed across the samples of the 
    normalized counts matrix, producing \code{nperm} permuted 
    instances of the initital dataset. Then, all the chosen 
    statistical tests are re-executed for each permutation. 
    The final p-value is the number of times that the p-value 
    of the permuted datasets is smaller than the original 
    dataset. The p-value of the original dataset is created 
    based on the choice of one of \code{dperm.min}, 
    \code{dperm.max} or \code{dperm.weight} options. In case of 
    \code{dperm.min}, the intial p-value vector is consists 
    of the minimum p-value resulted from the applied 
    statistical tests for each gene. The maximum p-value 
    is used with the \code{dperm.max} option. With the 
    \code{dperm.weight} option, the \code{weight} 
    weighting vector for each statistical test is used to 
    weight each p-value according to the power of 
    statistical tests (some might work better for a 
    specific dataset). Be careful as the permutation 
    procedure usually requires a lot of time. However, 
    it should be the most accurate. This method will NOT 
    work when there are no replicated samples across 
    biological conditions. In that case, use 
    \code{meta.p="simes"} instead. Finally, there are the 
    \code{"minp"}, \code{"maxp"} and \code{"weight"}
    options which correspond to the latter three methods 
    but without permutations. Generally, permutations 
    would be accurate to use when the experiment includes
    >5 samples per condition (or even better 7-10) which 
    is rather rare in RNA-Seq experiments. Finally, 
    \code{"pandora"} is the same as \code{"weight"} and is
    added to be in accordance with the main algorithm name.
    
    Regarding \code{pOffset}, it is used to correct for
    the case of a p-value which is equal to 0 as a result
    of internal numerical and approximation procedures.
    When \code{NULL}, random numbers greater than 0 and
    less than or equal to 0.5 are used to multiply the
    offending p-values with the lowest provided non-zero
    p-value, maintaining thus a virtual order of 
    significance, avoiding having the same p-values for 
    two tests and assuming that all zero p-values represent
    extreme statistical significance. When a numeric
    between 0 and 1, this number is used for the above
    multiplication instead.
    
    Regarding \code{qcPlots} The \code{"mds"} stands
    for Mutlti-Dimensional Scaling and it creates a PCA-like
    plot but using the MDS dimensionality reduction instead.
    It has been succesfully used for NGS data (e.g. see the
    package htSeqTools) and it shows how well samples from
    the same condition cluster together. For
    \code{"biodetection"}, \code{"countsbio"},
    \code{"saturation"}, \code{"rnacomp"},
    \code{"readnoise"}, \code{"biodist"} see the vignette of
    NOISeq package. The \code{"saturation"} case has been
    rewritten in order to display more samples in a more
    simple way. In addition, the \code{"readnoise"} plots 
    represent an older version or the RNA composition plot 
    included in older versions of NOISeq. For \code{"gcbias"},
    \code{"lengthbias"}, \code{"meandiff"}, \code{"meanvar"} 
    see the vignette of EDASeq package. \code{"lenghtbias"} 
    is similar to \code{"gcbias"} but using the gene length 
    instead of the GC content as covariate. The \code{"boxplot"} 
    option draws boxplots of log2 transformed gene counts. The
    \code{"filtered"} option draws a 4-panel figure with the
    filtered genes per chromosome and per biotype, as
    absolute numbers and as fractions of the genome. See also
    the help page of \code{\link{diagplotFiltered}}. The
    \code{"deheatmap"} option performs hierarchical
    clustering and draws a heatmap of differentially
    expressed genes. In the context of diagnostic plots, it's
    useful to see if samples from the same groups cluster
    together after statistical testing. The \code{"volcano"}
    option draws a volcano plot for each contrast and if a
    report is requested, an interactive volcano plot is
    presented in the HTML report. The \code{"venn"} option
    will draw an up to 5-way Venn diagram depicting the
    common and specific to each statistical algorithm genes
    and for each contrast, when meta-analysis is performed.
    The \code{"correl"} option creates two correlation
    graphs: the first one is a correlation heatmap (a
    correlation matrix which depicts all the pairwise
    correlations between each pair of samples in the counts
    matrix is drawn as a clustered heatmap) and the second
    one is a correlogram plot, which summarizes the
    correlation matrix in the form of ellipses (for an
    explanation please see the vignette/documentation of the
    R package corrplot. Set \code{qcPlots=NULL} if you don't
    want any diagnostic plots created.
    
    Regarding \code{reportDb}, contrary with the first version of
    metaseqR, all graphs in the metaseqR2 report are interactive
    with the usage of the JavaScript libraries Highcharts, plotly
    (heatmaply) and jvenn.js. However, this adds a great burden
    regarding rendering the final HTML file and its content,
    a burden which becomes heavier by the fact the metaseqR2
    report is rendered using \code{knitr} and \code{rmarkdown} 
    instead of raw HTML (previously, \code{brew}). Therefore, the
    pre-calculated JSON objects representing the graphs are 
    stored either in a report-specific IndexedDB 
    (https://javascript.info/indexeddb) flavor called Dexie
    (https://dexie.org/) (default) or in an SQLite database
    and then queried using \code{sql.js} 
    (https://github.com/kripken/sql.js/). Dexie is 
    prefered because it is very efficient and can produce an 
    independent report that does not need to be served through 
    a web-server and can be viewed locally. Although Dexie is 
    very efficient, some caution is required as \code{knitr} and 
    \code{\link{render}} from \code{rmarkdown} are not very 
    memory efficient when rendering larger HTML files. A large 
    HTML file may be produced when analyzing a large dataset
    with a lot of contrasts that may result in a lot of tables.
    In such cases, if the report generation crashes with errors
    related to memory, try lowering the \code{reportTop}
    argument. \code{reportTop} does not affect the final lists
    of differentially expressed genes, only the report tables.
    The same must be applied also if the report takes too much
    time to load. If the report is to be served through a web
    server like Apache (e.g. when the report is provided by a 
    facility to end users), \code{reportDb="sqlite"} may be
    preferred as the total report size will be smaller because
    of an SQLite database hosting all plots which are queried
    when required but from the SQLite database and not from
    the in-browser database (Dexie). **Note** that when using
    an SQLite database, you will **NOT** be able to view the
    report in any browser other than Microsoft Edge because
    of security policies regarding local file access. **Note**
    also that \code{sql.js} is a rather large JavaScript library
    (around 2.5MB).
    
    Regarding \code{trackInfo}, it is a helper list to guide the
    bigWig track creation and has the following members:
    \itemize{
        \item \code{stranded}, which can be \code{TRUE} or
        \code{FALSE} depending on whether you wish to create
        stranded tracks by separating + and - strand reads.
        In the case of stranded tracks, a UCSC Genome Brower
        trackhub is created. Individual tracks can be retrieved
        from the trackhub.
        \item \code{normTo}, which is a large integer, denoting
        the total sum of signal to be used as the normalization
        target. It defaults to \code{1e+9}. This means that if 
        for a particular sample the sum of signal is 1.5e+9
        (\code{sum(sapply(coverage(x),sum)) == 1.5e+9}) then
        this is linearly scaled to 1e+9.
        \item \code{urlBase}, which is a base url appended to
        the bigWig files produced (the base path of the 
        \code{bigDataUrl} in UCSC Genome Browser track lines).
        \item \code{hubInfo}, a list with the track hub 
        description in case of stranded tracks. Please see the
        track hub specifications at the UCSC Genome Browser site.
        \item \code{fasta}, reference genome in FASTA format for 
        the case of analyzing a custom, non-directly supported 
        organism. It will be converted to the .2bit format and 
        written along with a track hub.
        \item \code{gtf}, a GTF file describing gene models in 
        the case of analyzing a custom, non-directly supported 
        organism. It will be converted to the .bigBed format and 
        written along with a track hub. Essentially the same as
        \code{annotation$gtf}.
    }
    All files (bigWig files, track/trackhub info) are written in 
    the \code{tracks} subdirectory of the main path where the
    report and the outputs are written.
    
}
\note{
    Currently only gene and exon annotation from Ensembl
    (http://www.ensembl.org), UCSC and RefSeq are supported. 
    In addition, the user may choose to use own GTF file on
    the fly or import to the backend annotation database (see
    \code{\link{buildAnnotationDatabase}}). Thus, the unique
    gene ids in the counts files should correspond to valid 
    Ensembl, UCSC or RefSeq gene or exon accessions for the 
    organism of interest, or according to the user's GTF. 
    If you are not sure about the source of your counts file or 
    do not know how to produce it, it's better to start from the 
    original BAM/BED files (metaseqr2 will use the
    \code{\link{read2count}} function to create a counts
    file). Keep in mind that in the case of BED files, the
    performance will be significantly lower and the overall
    running time significanlty higher as the R functions
    which are used to read BED files to proper structures
    (GenomicRanges) and calculate the counts are quite slow.
    An alternative way is maybe the easyRNASeq package
    (Delhomme et al, 2012). The \code{\link{read2count}}
    function does not use this package but rather makes use
    of standard Bioconductor functions to handle NGS data. If
    you wish to work outside R, you can work with other
    popular read counters such as the HTSeq read counter
    (http://www-huber.embl.de/users/anders/HTSeq/doc/overview.html).
    Please also note that in the current version, the members
    of the \code{geneFilters} and \code{exonFilters} lists
    are not checked for validity so be careful to supply with
    correct names otherwise the pipeline will crash or at the
    best case scenario, will ignore the filters. Also note
    that when you are supplying metaseqr2 wtih an exon counts
    table, gene annotation is always downloaded so please be
    sure to have a working internet connection. In addition
    to the above, if you have a multiple core system, be very
    careful on how you are using the \code{restrictCores}
    argument and generally how many cores you are using with
    scripts purely written in R. The analysis with exon read
    data can very easily cause memory problems, so unless you
    have more than 64Gb of RAM available, consider setting
    restrict.cores to something like 0.2 when working with
    exon data. Finally, if you do not wish to download the
    same annotation again and again when performing multiple
    analyses, it is best to use the
    \code{\link{buildAnnotationDatabase}} function to download 
    and store the resulting data frames in local SQLite 
    database and then use these files with the 
    \code{org}, \code{refdb} and \code{version} options.

    Please note that the \strong{meta-analysis} feature
    provided by metaseqr2 does not satisfy the strict definition 
    of "meta-analysis", which is the combination of multiple 
    similar datasets under the same statistical methodology. 
    Instead it is the use of mulitple statistical tests applied 
    to the same data. For the Simes method, please consult also
    "Simes, R. J. (1986). "An improved Bonferroni procedure
    for multiple tests of significance". Biometrika 73 (3):
    751–754."
    
    Also, if \code{weight="meta_perm"} ideally one would 
    want to create the same set of indices for a given 
    dataset so as to create reproducible p-values. To 
    achieve this, use the \code{set.seed} function prior to
    any calculations. When metaseqR2 is loaded, the random
    seed is set to \code{42}.
}
\section{Exon filters}{
    The exon filters are a set of filters which are applied
    after the gene models are assembled from the read counts
    of individual exons and before the gene expression is
    summarized from the exons belonging to each gene. These
    filters can be applied when the input read counts file
    contains exon reads. It is not applicable when the input
    file already contains gene counts. Such filters can be
    for example "accept genes where all the exons contain
    more than x reads" or "accept genes where there is read
    presence in at least m/n exons, n being the total exons
    of the gene". Such filters are NOT meant for detecting
    differential splicing as also the whole metaseqr2
    pipeline, thus they should not be used in that context.
    The \code{exonFilters} argument is a named list of
    filters, where the names are the filter names and the
    members are the filter parameters (named lists with
    parameter name, parameter value). See the usage of the
    \code{metaseqr2} function for an example of how these
    lists are structured. The supported exon filters in the
    current version are: i) \code{minActiveExons} which
    implements a filter for demanding m out of n exons of a
    gene to have a certain read presence with parameters
    \code{exonsPerGene}, \code{minExons} and \code{frac}.
    The filter is described as follows: if a gene has up to
    \code{exonsPerGene} exons, then read presence is
    required in at least \code{minExons} of them, else read
    presence is required in a \code{frac} fraction of the
    total exons. With the default values, the filter
    instructs that if a gene has up to 5 exons, read presence
    is required in at least 2, else in at least 20% of the
    exons, in order to be accepted. More filters will be
    implemented in future versions and users are encouraged
    to propose exon filter ideas to the author by mail. See
    \code{metaseqr2} usage for the defaults. Set
    \code{exonFilters=NULL} to not apply any exon filtering.
}

\section{Gene filters}{
    The gene filters are a set of filters applied to gene
    expression as this is manifested through the read
    presence on each gene and are preferably applied after
    normalization. These filters can be applied both when the
    input file or data frame contains exon read counts and
    gene read counts. Such filter can be for example "accept
    all genes above a certain count threshold" or "accept all
    genes with expression above the median of the normalized
    counts distribution" or "accept all with length above a
    certain threshold in kb" or "exclude the 'pseudogene'
    biotype from further analysis". The supported gene
    filters in the current version, which have the same
    structure as the exon filters (named list of lists with
    filter names, parameter names and parameter arguments)
    are: i) \code{length} which implements a length filter
    where genes are accepted for further analysis if they are
    above \code{length} (its parameter) kb. ii)
    \code{avg.reads} which implements a filter where a gene
    is accepted for further analysis if it has more average
    reads than the \code{quantile} of the average count
    distribution per \code{averagePerBp} base pairs. In
    summary, the reads of each gene are averaged per
    \code{averagePerBp} based on each gene's length (in
    case of exons, input the "gene's length" is the sum of
    the lengths of exons) and the \code{quantile} quantile of
    the average counts distribution is calculated for each
    sample. Genes passing the filter should have an average
    read count larger than the maximum of the vector of the
    quantiles calculated above. iii) \code{expression} which
    implements a filter based on the overall expression of a
    gene. The parameters of this filter are: \code{median},
    where genes below the median of the overall count
    distribution are not accepted for further analysis (this
    filter has been used to distinguish between "expressed"
    and "not expressed" genes in several cases, e.g. (Mokry
    et al., NAR, 2011) with a logical as value, \code{mean}
    which is the same as \code{median} but using the mean,
    \code{quantile} which is the same as the previous two but
    using a specific quantile of the total counts
    distribution, \code{known}, where in this case, a set of
    known not-expressed genes in the system under
    investigation are used to estimate an expression cutoff.
    This can be quite useful, as the genes are filtered based
    on a "true biological" cutoff instead of a statistical
    cutoff. The value of this filter is a character vector of
    HUGO gene symbols (MUST be contained in the annotation,
    thus it's better to use \code{annotation="download"})
    whose counts are used to build a "null" expression
    distribution. The 90th quantile of this distribution is
    then the expression cutoff. This filter can be combined
    with any other filter. Be careful with gene names as they
    are case sensitive and must match exactly ("Pten" is
    different from "PTEN"!). iv) \code{biotype} where in this
    case, genes with a certain biotype (MUST be contained in
    the annotation, thus it's better to use
    \code{annotation="download"}) are excluded from the
    analysis. This filter is a named list of logical, where
    names are the biotypes in each genome and values are
    \code{TRUE} or \code{FALSE}. If the biotype should be
    excluded, the value should be \code{TRUE} else
    \code{FALSE}. See the result of
    \code{get.defaults("biotype.filter","hg19")} for an
    example. Finally, in future versions there will be
    support for user-defined filters in the form of a
    function. v) \code{presence} where in this case, a gene
    is further considered for statistical testing if 
    \code{frac} (x100 for a percentage value) have more
    than \code{minCount} reads across all samples
    (\code{perCondition=FALSE}) or across the samples
    of each condition (\code{perCondition=TRUE}).
}

\section{Normalization parameters}{
    The normalization parameters are passed again as a named
    list where the names of the members are the normalization
    parameter names and the values are the normalization
    parameter values. You should check the documentation of
    the packages EDASeq, DESeq, edgeR, NOISeq and NBPSeq for
    the parameter names and parameter values. There are a few
    exceptions in parameter names: in case of
    \code{normalization="edaseq"} the only parameter names
    are \code{within.which} and \code{between.which},
    controlling the withing lane/sample and between
    lanes/samples normalization algorithm. In the case 
    of \code{normalization="nbpseq"}, there is one
    additional parameter called \code{main.method} which can
    take the calues \code{"nbpseq"} or \code{"nbsmyth"}.
    These values correspond to the two different workflows
    available in the NBPSeq package. Please, consult the
    NBPSeq package documentation for further details. For the
    rest of the algorithms, the parameter names are the same
    as the names used in the respective packages. For
    examples, please use the \code{\link{getDefaults}}
    function.
}

\section{Statistics parameters}{
    The statistics parameters as passed to statistical
    algorithms in metaseqr2, exactly with the same way as the
    normalization parametes above. In this case, there is one
    more layer in list nesting. Thus, \code{statArgs} is a
    named list whose names are the names the algorithms used
    (see the \code{statistics} parameter). Each member is
    another named list,with parameters to be used for each
    statistical algorithm. Again, the names of the member
    lists are parameter names and the values of the member
    lists are parameter values. You should check the
    documentations of DESeq, edgeR, NOISeq, limma and
    NBPSeq for these parameters. There are a few exceptions
    in parameter names: In case of \code{statistics="edger"},
    apart from the rest of the edgeR statistical testing
    arguments, there is the argument \code{mainMethod} which
    can be either \code{"classic"} or \code{"glm"}, again
    defining whether the binomial test or GLMs will be used
    for statistical testing. For examples, please use the
    \code{\link{getDefaults}} function. When
    \code{statistics="nbpseq"}, apart from the rest arguments
    of the NBPSeq functions \code{estimate.disp} and
    \code{estimate.dispersion}, there is the argument
    \code{mainMethod} which can be \code{"nbpseq"} or
    \code{"nbsmyth"}. This argument determines the parameters
    to be used by the \code{estimate.dispersion} function or
    by the \code{estimate.disp} function to estimate RNA-Seq
    count dispersions. The difference between the two is that
    they constitute different starting points for the two
    workflows in the package NBPSeq. The first worklfow (with
    \code{mainMethod="nbpseq"} and the
    \code{estimate.dispersion} function is NBPSeq package
    specific, while the second (with
    \code{mainMethod="nbsmyth"} and the \code{estimate.disp}
    function is similar to the workflow of the edgeR package.
    For additional information regarding the statistical
    testing in NBPSeq, please consult the documentation of
    the NBPSeq package.
}

\section{Presets}{
    The analysis presets are a set of keywords (only one can
    be used) that predefine some of the parameters of the
    metaseqr2 pipeline. For the time being they are quite
    simple and they control i) the strictness of filtering
    and statistical thresholding with three basic levels
    ("all", "medium", "strict") and ii) the data columns that
    are exported, again in three basic ways ("basic",
    "normal", "full") controlling the amount of data to be
    exported. These keywords can be combined with a dot in
    the middle (e.g. \code{"all.basic"} to define an analysis
    preset. When using analysis presets, the following
    argumentsof metaseqr2 are overriden: \code{exonFilters},
    \code{geneFilters}, \code{pcut}, \code{exportWhat},
    \code{exportScale}, \code{exportValues},
    \code{exonStats}. If you want to explicitly control the
    above arguments, the \code{preset} argument should be set
    to \code{NULL} (default). Following is a synopsis of the
    different presets and the values of the arguments they
    moderate: 
    \itemize{ 
        \item \code{"all_basic"}: use all genes (do not filter)
        and export all genes and basic annotation and statistics 
        elements. In this case, the above described arguments become: 
        \itemize{ 
            \item \code{exonFilters=NULL} 
            \item \code{geneFilters=NULL}
            \item \code{pcut=1} 
            \item \code{exportWhat=c("annotation","p_value",}
            \code{"adj_p_value","meta_p_value",}
            \code{"adj_meta_p_value","fold_change")} 
            \item \code{exportScale=c("natural","log2")} 
            \item \code{exportValues=c("normalized")} 
            \item \code{exportStats=c("mean")} 
        } 
        \item \code{"all_normal"}: use all genes (do not filter) 
        and export all genes and normal annotation and statistics
        elements. In this case, the above described arguments
        become: 
        \itemize{ 
            \item \code{exonFilters=NULL} 
            \item \code{geneFilters=NULL} 
            \item \code{pcut=1} 
            \item \code{exportWhat=c("annotation","p_value",}
            \code{"adj_p_value","meta_p_value",} 
            \code{"adj_meta_p_value","fold_change","stats",}
            \code{"counts")}
            \item \code{exportScale=c("natural","log2")} 
            \item \code{exportValues=c("normalized")} 
            \item \code{exportStats=c("mean","sd","cv")} 
        }
        \item \code{"all_full"}: use all genes (do not filter) 
        and export all genes and all annotation and statistics
        elements. In this case, the above described arguments 
        become: 
        \itemize{ 
            \item \code{exonFilters=NULL} 
            \item \code{geneFilters=NULL}
            \item \code{pcut=1} 
            \item \code{exportWhat=c("annotation","p_value",
            "adj_p_value","meta_p_value",} 
            \code{"adj_meta_p_value","fold_change","stats","counts")}
            \item \code{exportScale=c("natural","log2","log10","vst")}
            \item \code{exportValues=c("raw","normalized")} 
            \item \code{exportStats=c("mean","median","sd","mad",
            "cv","rcv")}
    } 
    \item \code{"medium_basic"}: apply a medium set of filters
    and and export statistically significant genes and basic
    annotation and statistics elements. In this case, the above
    above described arguments become: 
    \itemize{
        \item \code{exonFilters=list(minActiveExons=}
        \code{list(exonsPerGene=5,minExons=2,frac=1/5))} 
        \item \code{geneFilters=list(length=list(length=500),} 
        \code{avgReads=list(averagePerBp=100,quantile=0.25),} 
        \code{expression=list(median=TRUE,mean=FALSE,quantile=NA,}
        \code{known=NA,custom=NA),} 
        \code{biotype=getDefaults("biotypeFilter",org[1]))}
        \item \code{pcut=0.05} 
        \item \code{exportWhat=c("annotation","p_value",}
        \code{"adj_p_value","meta_p_value",} 
        \code{"adj_meta_p_value","fold_change")} 
        \item \code{exportScale=c("natural","log2")} 
        \item \code{exportValues=c("normalized")} 
        \item \code{exportStats=c("mean")} 
    } 
    \item \code{"medium_normal"}: apply a medium set of filters 
    and export statistically significant genes and normal
    annotation and statistics elements. In this case, the
    above described arguments become: 
    \itemize{ 
        \item \code{exonFilters=list(minActiveExons=}
        \code{list(exonsPerGene=5,minExons=2,frac=1/5))} 
        \item \code{geneFilters=list(length=list(length=500),} 
        \code{avgReads=list(averagePerBp=100,quantile=0.25),} 
        \code{expression=list(median=TRUE,mean=FALSE,}
        \code{quantile=NA,known=NA,custom=NA),} 
        \code{biotype=getDefaults("biotypeFilter",org[1]))}
        \item \code{pcut=0.05} 
        \item \code{export.what=c("annotation","p_value",}
        \code{"adj_p_value","meta_p_value",} 
        \code{"adj_meta_p_value","fold_change",}
        \code{"stats","counts")}
        \item \code{exportScale=c("natural","log2")} 
        \item \code{exportValues=c("normalized")} 
        \item \code{exportStats=c("mean","sd","cv")} 
    } 
    \item \code{"medium_full"}: apply a medium set of filters 
    and export statistically significant genes and full
    annotation and statistics elements. In this case, the
    above described arguments become: 
    \itemize{ 
        \item \code{exonFilters=list(minActiveExons=}
        \code{list(exonsPerGene=5,minExons=2,frac=1/5))} 
        \item \code{geneFilters=list(length=list(length=500),} 
        \code{avgReads=list(averagePerBp=100,quantile=0.25),} 
        \code{expression=list(median=TRUE,mean=FALSE,}
        \code{quantile=NA,known=NA,custom=NA),} 
        \code{biotype=getDefaults("biotypeFilter",org[1]))}
        \item \code{pcut=0.05} 
        \item \code{exportWhat=c("annotation","p_value",}
        \code{"adj_p_value","meta_p_value",} 
        \code{"adj_meta_p_value","fold_change",}
        \code{"stats","counts")}
        \item \code{exportScale=c("natural","log2","log10",}
        \code{"vst")}
        \item \code{exportValues=c("raw","normalized")} 
        \item \code{exportStats=c("mean","median","sd","mad",}
        \code{"cv","rcv")}
    } 
    \item \code{"strict_basic"}: apply a strict set of filters
    and and export statistically significant genes and basic
    annotation and statistics elements. In this case, the
    above described arguments become: 
    \itemize{
        \item \code{exonFilters=list(minActiveExons=}
        \code{list(exonsPerGene=4,minExons=2,frac=1/4))} 
        \item \code{geneFilters=list(length=list(length=750),} 
        \code{avgReads=list(averagePerBp=100,quantile=0.5),} 
        \code{expression=list(median=TRUE,mean=FALSE,}
        \code{quantile=NA,known=NA,custom=NA),} 
        \code{biotype=getDefaults("biotypeFilter",org[1]))}
        \item \code{pcut=0.01} 
        \item \code{exportWhat=c("annotation","p_value",}
        \code{"adj_p_value","meta.p.value",} 
        \code{"adj_meta_p_value","fold_change")} 
        \item \code{exportScale=c("natural","log2")} 
        \item \code{exportValues=c("normalized")} 
        \item \code{exportStats=c("mean")} 
    } 
    \item \code{"strict_normal"}: apply a strict set of filters 
    and export statistically significant genes and normal
    annotation and statistics elements. In this case, the
    above described arguments become: 
    \itemize{ 
        \item \code{exonFilters=list(minActiveExons=}
        \code{list(exonsPerGene=4,minExons=2,frac=1/4))} 
        \item \code{geneFilters=list(length=list(length=750),} 
        \code{avgReads=list(averagePerBp=100,quantile=0.5),} 
        \code{expression=list(median=TRUE,mean=FALSE,}
        \code{quantile=NA,known=NA,custom=NA),} 
        \code{biotype=getDefaults("biotypeFilter",org[1]))}
        \item \code{pcut=0.01} 
        \item \code{exportWhat=c("annotation","p_value",}
        \code{"adj_p_value","meta_p_value",} 
        \code{"adj_meta_p_value","fold_change",}
        \code{"stats","counts")}
        \item \code{exportScale=c("natural","log2")} 
        \item \code{exportValues=c("normalized")} 
        \item \code{exportStats=c("mean","sd","cv")} 
    } 
    \item \code{"strict_full"}: apply a strict set of filters 
    and export statistically significant genes and full
    annotation and statistics elements. In this case, the
    above described arguments become:
    \itemize{ 
        \item \code{exonFilters=list(minActiveExons=}
        \code{list(exonsPerGene=4,minExons=2,frac=1/4))} 
        \item \code{geneFilters=list(length=list(length=750),} 
        \code{avgReads=list(averagePerBp=100,quantile=0.5),} 
        \code{expression=list(median=TRUE,mean=FALSE,}
        \code{quantile=NA,known=NA,custom=NA),} 
        \code{biotype=getDefaults("biotypeFilter",org[1]))}
        \item \code{pcut=0.01} 
        \item \code{exportWhat=c("annotation","p_value",}
        \code{"adj_p_value","meta_p_value",} 
        \code{"adj_meta_p_value","fold_change",}
        \code{"stats","counts")}
        \item \code{exportScale=c("natural","log2","log10",}
        \code{"vst")}
        \item \code{exportValues=c("raw","normalized")} 
        \item \code{exportStats=c("mean","median","sd","mad",}
        \code{"cv","rcv")}
    } }
}
\examples{
# An example pipeline with gene counts
data("mm9GeneData",package="metaseqR2")

result <- metaseqr2(
    counts=mm9GeneCounts,
    sampleList=sampleListMm9,
    contrast=c("adult_8_weeks_vs_e14.5"),
    libsizeList=libsizeListMm9,
    annotation="embedded",
    org="mm9",
    countType="gene",
    normalization="edger",
    statistics="edger",
    pcut=0.05,
    figFormat="png",
    qcPlots="mds",
    exportWhat=c("annotation","p_value","adj_p_value","fold_change"),
    exportScale="natural",
    exportValues="normalized",
    exportStats="mean",
    exportWhere=file.path(tempdir(),"test1"),
    restrictCores=0.01,
    geneFilters=list(
        length=list(
            length=500
        ),
        avgReads=list(
            averagePerBp=100,
            quantile=0.25
        ),
        expression=list(
            median=TRUE,
            mean=FALSE,
            quantile=NA,
            known=NA,
            custom=NA
        ),
        biotype=getDefaults("biotypeFilter","mm9")
    ),
    outList=TRUE
)
head(result$data[["adult_8_weeks_vs_e14.5"]])
}
\author{
    Panagiotis Moulos
}

\name{createSignalTracks}
\alias{createSignalTracks}
\title{Create bigWig signal tracks}
\usage{
    createSignalTracks(targets, org, urlBase = NULL, 
        stranded = FALSE, normTo = 1e+9, exportPath = ".",
        hubInfo = list(name = "MyHub", shortLabel = "My hub",
        longLabel = "My hub", email = "someone@example.com"),
        fasta = NULL, gtf = NULL, forceHub = FALSE, 
        overwrite = FALSE, rc = NULL)
}
\arguments{
    \item{targets}{a tab-delimited file with the experimental 
    description or the output of \code{\link{readTargets}}.
    See also the \code{sampleList} argument in the main
    \code{\link{metaseqr2}} pipeline.}

    \item{org}{See the \code{org} argument in the main
    \code{\link{metaseqr2}} pipeline.}

    \item{urlBase}{a valid URL which is prepended to the created
    bigWig files.}

    \item{stranded}{Separate + and - strands and create separate
    bigWig files.}

    \item{normTo}{the total sum of signal to be used as the 
    normalization target. See also the \code{trackInfo} argument 
    in the main \code{\link{metaseqr2}} pipeline.}

    \item{exportPath}{path to export tracks.}
    
    \item{hubInfo}{information regarding the track hub created
    when \code{stranded=TRUE}. See also the \code{trackInfo} 
    argument in the main \code{\link{metaseqr2}} pipeline.}
    
    \item{overwrite}{overwrite tracks if they exist? Defaults to
    \code{FALSE}.}
    
    \item{fasta}{reference genome in FASTA format for the case
    of analyzing a custom, non-directly supported organism. It
    will be converted to the .2bit format and written along with
    a track hub.}
    
    \item{gtf}{a GTF file describing gene models in the case of
    analyzing a custom, non-directly supported organism. It will
    be converted to the .bigBed format and written along with
    a track hub.}
    
    \item{forceHub}{when \code{stranded=TRUE}, a UCSC Genome
    Browser trackhub is created, otherwise only tracklines 
    describing individual tracks. If \code{TRUE}, a trackhub is
    always created.}
    
    \item{rc}{Fraction of cores to use.}
}
\value{
    A string with the link(s) to the created tracks.
}
\description{
    This function creates bigWig files to be used for exploring
    RNA signal in genome browsers. When strands are separated,
    a UCSC genome browser trackhub is created to group tracks
    for the same sample. A link to the created data is returned.
}
\examples{
dataPath <- system.file("extdata",package="metaseqR2")
targets <- data.frame(samplename=c("C","T"),
    filename=file.path(dataPath,c("C.bam","T.bam")),  
    condition=c("Control","Treatment"),
    paired=c("single","single"),stranded=c("forward","forward"))
path <- tempdir()
write.table(targets,file=file.path(path,"targets.txt"),
    sep="\t",row.names=FALSE,quote=FALSE)
if (.Platform$OS.type == "unix")
    #link <- createSignalTracks(file.path(path,"targets.txt"),"mm9")
}
\author{
    Panagiotis Moulos
}

\name{diagplotBoxplot}
\alias{diagplotBoxplot}
\title{Boxplots wrapper for the metaseqR2 package}
\usage{
    diagplotBoxplot(mat, name = NULL, logIt = "auto",
        yLim = "default", isNorm = FALSE, output = "x11",
        path = NULL, altNames = NULL, ...)
}
\arguments{
    \item{mat}{the count data matrix.}

    \item{name}{the names of the samples plotted on the
    boxplot. See also Details.}

    \item{logIt}{whether to log transform the values of mat
    or not. It can be \code{TRUE}, \code{FALSE} or
    \code{"auto"} for auto-detection. Auto-detection log
    transforms by default so that the boxplots are smooth and
    visible.}

    \item{yLim}{custom y-axis limits. Leave the string
    \code{"default"} for default behavior.}

    \item{isNorm}{a logical indicating whether object
    contains raw or normalized data. It is not essential and
    it serves only plot annotation purposes.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"}, \code{"ps"} or \code{"json"}. The latter is
    currently available for the creation of interactive
    volcano plots only when reporting the output, through the
    highcharts javascript library (JSON for boxplots not yet
    available).}

    \item{path}{the path to create output files.}
    
    \item{altNames}{an optional vector of names, e.g. HUGO
    gene symbols, alternative or complementary to the unique
    rownames of \code{mat} (which must exist!). It is used only 
    in JSON output.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filename of the boxplot produced if it's a file.
}
\description{
    A wrapper over the general boxplot function, suitable for
    matrices produced and processed with the metaseqr
    package. Intended for internal use but can be easily used
    as stand-alone. It can colors boxes based on group
    depending on the name argument.
}
\details{
    Regarding \code{name}, if \code{NULL}, the function check 
    the column names of \code{mat}. If they are also 
    \code{NULL}, sample names are autogenerated. If 
    \code{name="none"}, no sample names are plotted. If name 
    is a list, it should be the sampleList argument provided 
    to the manin metaseqr2 function. In that case, the boxes  
    are colored per group.
}
\examples{
# Non-normalized boxplot
require(DESeq2)
dataMatrix <- metaseqR2:::exampleCountData(2000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
diagplotBoxplot(dataMatrix,sampleList)

# Normalized boxplot
normArgs <- getDefaults("normalization","deseq2")
object <- normalizeDeseq2(dataMatrix,sampleList,normArgs)
diagplotBoxplot(object,sampleList)
}
\author{
    Panagiotis Moulos
}

\name{readTargets}
\alias{readTargets}
\title{Creates sample list and BAM/BED file list from file}
\usage{
    readTargets(input, path = NULL)
}
\arguments{
    \item{input}{a tab-delimited file or a YAML file 
    specifically structured. See Details.}

    \item{path}{an optional path where all the BED/BAM 
    files are placed, to be prepended to the BAM/BED 
    file names in the targets file.}
}
\value{
    A named list with four members. The first member is 
    a named list whose names are the conditions of the 
    experiments and its members are the samples belonging 
    to each condition. The second member is like the 
    first, but this time the members are named vectors 
    whose names are the sample names and the vector 
    elements are full path to BAM/BED files. The third 
    member is like the second, but instead of filenames 
    it contains information about single- or paired-end
    reads (if available). The fourth member is like the 
    second, but instead of filenames it contains 
    information about the strandedness of the reads (if
    available). The fifth member is the guessed type 
    of the input files (SAM/BAM or BED). It will be used 
    if not given in the main \code{\link{read2count}}
    function.
}
\description{
    Create the main sample list and determine the BAM/BED
    files for each sample from an external file.
}
\details{
    Regarding the input file, this can be a simple text 
    tab-delimited file or a YAML file describing the data to
    be analyzed. 
    
    Regarding the tab-delimited version, its columns must
    be structured as follows: the first line of the external
    tab delimited file should contain column names (names
    are not important). The first column MUST contain UNIQUE
    sample names. The second column MUST contain the raw 
    BAM/BED files WITH their full path. Alternatively, the 
    \code{path} argument should be provided (see below). The 
    third column MUST contain the biological condition where 
    each of the samples in the first column should belong to. 
    There is an optional fourth column which should contain 
    the keywords \code{"single"} for single-end reads, 
    \code{"paired"} for paired-end reads or \code{"mixed"} 
    for BAM files that contain both single- and paired-end 
    reads (e.g. after a mapping procedure with two round of 
    alignment). If this column is not provided, single-end 
    reads will be assumed. There is an optional fifth column 
    which stranded read assignment. It should contain the 
    keywords \code{"forward"} for a forward (5'->3') strand 
    library construction protocol, \code{"reverse"} for a 
    reverse (3'->5') strand library construction protocol, 
    or \code{"no"} for unstranded/unknown protocol. If this 
    column is not provided, unstranded reads will be assumed.
    
    Regarding the YAML version, the same instructions apply,
    but this time instead of columns, the data are provided
    as a YAML array under a keyword/top-level field 
    representing the respective header in the tab-delimited
    version. Alternatively, the aforementioned structure
    can be nested under a root level named strictly either
    \code{targets} or \code{metaseqR2_targets}. The latter
    can be especially useful when incorporating the
    metaseqR2 pipeline in a wider pipeline including various
    analyses and described using a workflow language such
    as CWL.
}
\examples{
dataPath <- system.file("extdata",package="metaseqR2")
targets <- data.frame(samplename=c("C","T"),
    filename=file.path(dataPath,c("C.bam","T.bam")),  
    condition=c("Control","Treatment"),
    paired=c("single","single"),stranded=c("forward","forward"))
path <- tempdir()

# Tab delimited case
write.table(targets,file=file.path(path,"targets.txt"),
    sep="\t",row.names=FALSE,quote=FALSE)
theList <- readTargets(file.path(path,"targets.txt"),path=path)
sampleList <- theList$samples
bamfileList <- theList$files

# YAML case
require(yaml)
write_yaml(as.list(targets),file.path(path,"targets.yml"))
theYList <- readTargets(file.path(path,"targets.yml"),path=path)
identical(theList,theYList) # TRUE

# YAML case with nested targets
write_yaml(list(targets=as.list(targets)),
    file.path(path,"targets2.yml"))
theYList2 <- readTargets(file.path(path,"targets2.yml"),path=path)
identical(theYList,theYList2) # TRUE
}
\author{
    Panagiotis Moulos
}

\name{diagplotEdaseq}
\alias{diagplotEdaseq}
\title{Diagnostic plots based on the EDASeq package}
\usage{
    diagplotEdaseq(x, sampleList, covar = NULL,
        isNorm = FALSE,
        whichPlot = c("meanvar", "meandiff", "gcbias", "lengthbias"),
        output = "x11", altNames = NULL, path = NULL, ...)
}
\arguments{
    \item{x}{the count data matrix.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{covar}{The covariate to plot counts against.
    Usually \code{"gc"} or \code{"length"}.}

    \item{isNorm}{a logical indicating whether object
    contains raw or normalized data. It is not essential and
    it serves only plot annotation purposes.}

    \item{whichPlot}{the EDASeq package plot to generate. 
    See Details.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"} or \code{"ps"}.}
    
    \item{altNames}{optional names, alternative or complementary 
    to the rownames of \code{x}. It is used only in JSON output.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filenames of the plot produced in a named list with
    names the which.plot argument. If \code{output="x11"}, no
    output filenames are produced.
}
\description{
    A wrapper around the plotting functions availale in the
    EDASeq normalization Bioconductor package. For analytical
    explanation of each plot please see the vignette of the
    EDASeq package. It is best to use this function through
    the main plotting function
    \code{\link{metaseqrPlot}}.
}
\details{
    Regarding \code{whichPlot}, it can be one or more of 
    \code{"meanvar"}, \code{"meandiff"}, \code{"gcbias"} or
    \code{"lengthbias"}. Please refer to the documentation of
    the EDASeq package for details on the use of these 
    plots. The \code{whichPlot="lengthbias"} case is 
    not covered by EDASeq documentation, however it is 
    similar to the GC-bias plot when the covariate is the 
    gene length instead of the GC content.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
diagplotEdaseq(dataMatrix,sampleList,whichPlot="meandiff")
}
\author{
    Panagiotis Moulos
}

\name{normalizeEdger}
\alias{normalizeEdger}
\title{Normalization based on the edgeR package}
\usage{
    normalizeEdger(geneCounts, sampleList,
        normArgs = NULL, output = c("matrix", "native"))
}
\arguments{
    \item{geneCounts}{a table where each row represents a
    gene and each column a sample. Each cell contains the
    read counts for each gene and sample. Such a table can be
    produced outside metaseqr2 and is imported during the
    basic metaseqr2 workflow.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{normArgs}{a list of edgeR normalization
    parameters. See the result of
    \code{getDefaults("normalization",} \code{"edger")} for
    an example and how you can modify it.}

    \item{output}{the class of the output object. It can be
    \code{"matrix"} (default) for versatility with other
    tools or \code{"native"} for the edgeR native S4 object
    (DGEList). In the latter case it should be handled with
    suitable edgeR methods.}
}
\value{
    A matrix or a DGEList with normalized counts.
}
\description{
    This function is a wrapper over edgeR normalization. It
    accepts a matrix of gene counts (e.g. produced by
    importing an externally generated table of counts to the
    main metaseqr2 pipeline).
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
diagplotBoxplot(dataMatrix,sampleList)

normDataMatrix <- normalizeEdger(dataMatrix,sampleList)
diagplotBoxplot(normDataMatrix,sampleList)
}
\author{
    Panagiotis Moulos
}

\name{statNbpseq}
\alias{statNbpseq}
\title{Statistical testing with NBPSeq}
\usage{
    statNbpseq(object, sampleList, contrastList = NULL,
        statArgs = NULL, libsizeList = NULL)
}
\arguments{
    \item{object}{a matrix or an object specific to each
    normalization algorithm supported by metaseqR2, containing
    normalized counts. See also Details.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{contrastList}{vector of contrasts as defined in the 
    main help page of \code{\link{metaseqr2}}. See also 
    Details.}

    \item{statArgs}{a list of NBPSeq statistical algorithm
    parameters. See the result of
    \code{getDefaults("statistics",} \code{"nbpseq")} for an
    example and how you can modify it. It is not required
    when the input object is already a list from NBPSeq
    normalization as the dispersions are already estimated.}

    \item{libsizeList}{an optional named list where names
    represent samples (MUST be the same as the samples
    \code{in sampleList}) and members are the library sizes
    (the sequencing depth) for each sample. If not provided,
    the default is the column sums of the \code{object}
    matrix.}
}
\value{
    A named list of p-values, whose names are the names of
    the contrasts.
}
\description{
    This function is a wrapper over NBPSeq statistical
    testing. It accepts a matrix of normalized gene counts or
    an S4 object specific to each normalization algorithm
    supported by metaseqR2.
}
\details{
    Regarding \code{object}, apart from \code{matrix} (also 
    for NOISeq), the object can be a \code{SeqExpressionSet} 
    (EDASeq), \code{CountDataSet} (DESeq), \code{DGEList} 
    (edgeR), \code{DESeqDataSet} (DESeq2), \code{SeqCountSet} 
    (DSS) or \code{ABSDataSet} (ABSSeq).
    
    Regarding \code{contrastList} it can also be a named 
    structured list of contrasts as returned by the internal
    function \code{metaseqR2:::makeContrastList}.
}
\note{
    There is currently a problem with the NBPSeq package and
    the workflow that is specific to the NBPSeq package. The
    problem has to do with function exporting as there are
    certain functions which are not recognized from the
    package internally. For this reason and until it is
    fixed, only the Smyth workflow will be available with the
    NBPSeq package.
}
\examples{
require(NBPSeq)
dataMatrix <- metaseqR2:::exampleCountData(2000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
contrast <- "A_vs_B"
normDataMatrix <- normalizeNbpseq(dataMatrix,sampleList)
p <- statNbpseq(normDataMatrix,sampleList,contrast)
}
\author{
    Panagiotis Moulos
}

\name{diagplotFiltered}
\alias{diagplotFiltered}
\title{Diagnostic plot for filtered genes}
\usage{
    diagplotFiltered(x, y, output = "x11", path = NULL, ...)
}
\arguments{
    \item{x}{an annotation data frame like the ones produced
    by \code{\link{getAnnotation}}. \code{x} should be the
    filtered annotation according to metaseqR's filters.}

    \item{y}{an annotation data frame like the ones produced
    by \code{\link{getAnnotation}}. \code{y} should contain
    the total annotation without the application of any
    metaseqr filter.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"} or \code{"ps"}.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filenames of the plots produced in a named list with
    names the \code{which.plot} argument. If
    output=\code{"x11"}, no output filenames are produced.
}
\description{
    This function plots a grid of four graphs depicting: in
    the first row, the numbers of filtered genes per
    chromosome in the first column and per biotype in the
    second column. In the second row, the percentages of
    filtered genes per chromosome related to the whole genome
    in the first columns and per biotype in the second
    column.
}
\examples{
data("mm9GeneData",package="metaseqR2")
y <- mm9GeneCounts[,c(1:6,8,7)]
x <- y[-sample(1:nrow(y),1000),]
diagplotFiltered(x,y)
}
\author{
    Panagiotis Moulos
}

\name{makeSimDataTcc}
\alias{makeSimDataTcc}
\title{Create simulated counts using TCC package}
\usage{
    makeSimDataTcc(...)
}
\arguments{
    \item{...}{parameters to the \code{simulateReadCounts}
    function.}
}
\value{
    A list with the following members: \code{simdata} holding
    the simulated dataset complying with metaseqr2
    requirements, and \code{simparam} holding the simulation
    parameters (see TCC documentation). Note that the produced 
    data are based in an Arabidopsis dataset.
}
\description{
    This function creates simulated RNA-Seq gene expression
    datasets using the \code{simulateReadCounts} function
    from the Bioconductor package TCC and it adds simulated
    annoation elements. For further information please
    consult the TCC package documentation.
}
\examples{
if (require(TCC)) {
dd <- makeSimDataTcc(Ngene=1000,PDEG=0.2,
    DEG.assign=c(0.9,0.1),
    DEG.foldchange=c(5,5),replicates=c(3,3))
head(dd$simdata)
}
}
\author{
    Panagiotis Moulos
}

\name{diagplotRoc}
\alias{diagplotRoc}
\title{Create basic ROC curves}
\usage{
    diagplotRoc(truth, p, sig = 0.05, x = "fpr", 
        y = "tpr", output = "x11", path = NULL,
        draw = TRUE, ...)
}
\arguments{
    \item{truth}{the ground truth differential 
    expression vector. It should contain only 
    zero and non-zero elements, with zero denoting
    non-differentially expressed genes and non-zero, 
    differentially expressed genes. Such a vector 
    can be obtained for example by using the 
    \code{\link{makeSimDataSd}} function, which 
    creates simulated RNA-Seq read counts based on 
    real data.}

    \item{p}{a p-value matrix whose rows correspond 
    to each element in the \code{truth} vector. If 
    the matrix has a \code{colnames} attribute, a 
    legend will be added to the plot using these 
    names, else a set of column names will be 
    auto-generated. \code{p} can also be a list or 
    a data frame.}

    \item{sig}{a significance level (0 < \code{sig} 
    <=1).}

    \item{x}{what to plot on x-axis, can be one of 
    \code{"fpr"}, \code{"fnr"}, \code{"tpr"}, 
    \code{"tnr"} for False Positive Rate, False
    Negative Rate, True Positive Rate and True 
    Negative Rate respectively.}

    \item{y}{what to plot on y-axis, same as 
    \code{x} above.}

    \item{output}{one or more R plotting device to 
    direct the plot result to. Supported mechanisms: 
    \code{"x11"} (default), \code{"png"}, \code{"jpg"}, 
    \code{"bmp"}, \code{"pdf"} or \code{"ps"}.}

    \item{path}{the path to create output files.}

    \item{draw}{boolean to determine whether to
    plot the curves or just return the calculated
    values (in cases where the user wants the
    output for later averaging for example). 
    Defaults to \code{TRUE} (make plots).}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    A named list with two members. The first member is 
    a list containing the ROC statistics: \code{TP} 
    (True Postives), \code{FP} (False Positives), 
    \code{FN} (False Negatives), \code{TN} 
    (True Negatives), \code{FPR} (False Positive Rate), 
    \code{FNR} (False Negative Rate), \code{TPR} (True 
    Positive Rate), \code{TNR} (True Negative Rate), 
    \code{AUC} (Area Under the Curve). The second is 
    the path to the created figure graphic.
}
\description{
    This function creates basic ROC curves using a 
    matrix of p-values (such a matrix can be 
    derived for example from the result table of 
    \code{\link{metaseqr2}} by subsetting the table 
    to get the p-values from several algorithms) 
    given a ground truth vector for differential 
    expression and a significance level.
}
\examples{
p1 <- 0.001*matrix(runif(300),100,3)
p2 <- matrix(runif(300),100,3)
p <- rbind(p1,p2)
rownames(p) <- paste("gene",1:200,sep="_")
colnames(p) <- paste("method",1:3,sep="_")
truth <- c(rep(1,40),rep(-1,40),rep(0,20),rep(1,10),
    rep(2,10),rep(0,80))
names(truth) <- rownames(p)
rocObj <- diagplotRoc(truth,p)
}
\author{
    Panagiotis Moulos
}

\name{buildCustomAnnotation}
\alias{buildCustomAnnotation}
\title{Import custom annotation to the metaseqR2 annotation
    database from GTF file}
\usage{
    buildCustomAnnotation(gtfFile, metadata,
    db = file.path(system.file(package = "metaseqR2"),
        "annotation.sqlite"), rewrite=TRUE)
}
\arguments{
    \item{gtfFile}{a GTF file containing the gene structure
    of the organism to be imported.}

    \item{metadata}{a list with additional information about
    the annotation to be imported. See Details.}
    
    \item{db}{a valid path (accessible at least by the
    current user) where the annotation database will be 
    set up. It defaults to 
    \code{system.file(package = "metaseqR2"),}
    \code{"annotation.sqlite")} that is, the installation
    path of metaseqR2 package. See also Details.}
    
    \item{rewrite}{if custom annotation found, rwrite? 
    (default \code{FALSE}). Set to \code{TRUE} if you wish to 
    update the annotation database for a particular custom
    annotation.}
}
\value{
    The function does not return anything. Only the SQLite 
    database is created or updated.
}
\description{
    This function imports a GTF file with some custom annotation
    to the metaseqR2 annotation database.
}
\details{
    Regarding the \code{metadata} argument, it is a list
    with specific format which instructs 
    \code{buildCustomAnnotation} on importing the custom
    annotation. Such a list may has the following members: 
    \itemize{
        \item \code{organism} a name of the organism which is
        imported (e.g. \code{"my_mm9"}). This is the only
        mandatory member.
        \item \code{source} a name of the source for this
        custom annotation (e.g. \code{"my_mouse_db"}). If
        not given or \code{NULL}, the word \code{"inhouse"}
        is used.
        \item \code{version} a string denoting the version.
        If not given or \code{NULL}, current date is used.
        \item \code{chromInfo} it can be one of the following:
        \itemize{
            \item a tab-delimited file with two columns, the
            first being the chromosome/sequence names and the
            second being the chromosome/sequence lengths.
            \item a BAM file to read the header from and
            obtain the required information
            \item a \code{\link{data.frame}} with one column
            with chromosome lengths and chromosome names as
            \code{rownames}.
        }
    }
    See the examples below for a \code{metadata} example.
    
    Regarding \code{db}, this controls the location of the
    installation database. If the default is used, then there is
    no need to provide the local database path to any function
    that uses the database (e.g. the main \code{metaseqr2}).
    Otherwise, the user will either have to provide this each
    time, or the annotation will have to be downloaded and used
    on-the-fly.
}
\examples{
# Dummy database as example
customDir <- file.path(tempdir(),"test_custom")
dir.create(customDir)

myDb <- file.path(customDir,"testann.sqlite")
chromInfo <- data.frame(length=c(1000L,2000L,1500L),
    row.names=c("A","B","C"))

# Build with the metadata list filled (you can also provide a version)
if (.Platform$OS.type == "unix") {
    buildCustomAnnotation(
        gtfFile=file.path(system.file(package="metaseqR2"),
            "dummy.gtf"),
        metadata=list(
            organism="dummy",
            source="dummy_db",
            version=1,
            chromInfo=chromInfo
        ),
        db=myDb
    )

    # Try to retrieve some data
    myGenes <- loadAnnotation(genome="dummy",refdb="dummy_db",
        level="gene",type="gene",db=myDb)
    myGenes
}

## Real data!
## Setup a temporary directory to download files etc.
#customDir <- file.path(tempdir(),"test_custom")
#dir.create(customDir)

#myDb <- file.path(customDir,"testann.sqlite")

## Gene annotation dump from Ensembl
#download.file(paste0("ftp://ftp.ensembl.org/pub/release-98/gtf/",
#  "dasypus_novemcinctus/Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#  file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"))

## Chromosome information will be provided from the following BAM file
## available from Ensembl
#bamForInfo <- paste0("ftp://ftp.ensembl.org/pub/release-98/bamcov/",
#  "dasypus_novemcinctus/genebuild/Dasnov3.broad.Ascending_Colon_5.1.bam")

## Build with the metadata list filled (you can also provide a version)
#buildCustomAnnotation(
#  gtfFile=file.path(customDir,"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#  metadata=list(
#    organism="dasNov3_test",
#    source="ensembl_test",
#    chromInfo=bamForInfo
#  ),
#  db=myDb
#)

## Try to retrieve some data
#dasGenes <- loadAnnotation(genome="dasNov3_test",refdb="ensembl_test",
#  level="gene",type="gene",db=myDb)
#dasGenes
}
\author{
    Panagiotis Moulos
}

\docType{data}
\name{sampleListMm9}
\alias{sampleListMm9}
\title{Mouse RNA-Seq data with two conditions, four samples}
\format{a named \code{list} with condition and sample names.}
\source{
    ENCODE (http://genome.ucsc.edu/encode/)
}
\description{
    The sample list for \code{mm9GeneCounts}. See the data
    set description.
}
\author{
    Panagiotis Moulos
}
\keyword{datasets}

\name{diagplotVolcano}
\alias{diagplotVolcano}
\title{(Interactive) volcano plots of differentially expressed 
    genes}
\usage{
    diagplotVolcano(f, p, con = NULL, fcut = 1, pcut = 0.05,
        altNames = NULL, output = "x11", path = NULL, ...)
}
\arguments{
    \item{f}{the fold changes which are to be plotted on the
    x-axis.}

    \item{p}{the p-values whose -log10 transformation is
    going to be plotted on the y-axis.}

    \item{con}{an optional string depicting a name (e.g. the
    contrast name) to appear in the title of the volcano
    diagplot.}

    \item{fcut}{a fold change cutoff so as to draw two
    vertical lines indicating the cutoff threshold for
    biological significance.}

    \item{pcut}{a p-value cutoff so as to draw a horizontal
    line indicating the cutoff threshold for statistical
    significance.}

    \item{altNames}{an optional vector of names, e.g. HUGO
    gene symbols, alternative or complementary to the unique
    names of \code{f} or \code{p} (one of them must be
    named!). It is used only in JSON output.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"}, \code{"ps"} or \code{"json"}. The latter is
    currently available for the creation of interactive
    volcano plots only when reporting the output, through the
    highcharts javascript library.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filenames of the plots produced in a named list with
    names the \code{which.plot} argument. If
    \code{output="x11"}, no output filenames are produced.
}
\description{
    This function plots a volcano plot or returns a JSON
    string which is used to render aninteractive in case of
    HTML reporting.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(5000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
contrast <- "A_vs_B"
M <- normalizeEdger(dataMatrix,sampleList)
p <- statEdger(M,sampleList,contrast)
ma <- apply(M[,sampleList$A],1,mean)
mb <- apply(M[,sampleList$B],1,mean)
f <- log2(ifelse(mb==0,1,mb)/ifelse(ma==0,1,ma))
diagplotVolcano(f,p[[1]],con=contrast)
#j <- diagplotVolcano(f,p[[1]],con=contrast,output="json")
}
\author{
    Panagiotis Moulos
}

\name{read2count}
\alias{read2count}
\title{SAM/BAM/BED file reader helper for the metaseqr2 pipeline}
\usage{
    read2count(targets, annotation, fileType = targets$type,
        transLevel = "gene", utrOpts = list(frac = 1, 
        minLength = 300, downstream = 50), interFeature = FALSE, 
        rc = NULL)
}
\arguments{
    \item{targets}{a named list, the output of 
    \code{\link{readTargets}} or an existing file with
    targets. See also the main \code{metaseqr2} man page.}

    \item{annotation}{a \code{GenomicRanges} or 
    \code{data.frame} with genomic coordinates to use for
    read counting. See also \code{\link{getAnnotation}}.}

    \item{fileType}{the type of raw input files. It can be
    \code{"bed"} for BED files or \code{"sam"}, \code{"bam"}
    for SAM/BAM files. See the same argument in the main
    \code{\link{metaseqr2}} function for the case of
    auto-guessing.}
    
    \item{transLevel}{see the \code{transLevel} argument
    in the main \code{\link{metaseqr2}} function.}
    
    \item{utrOpts}{a named list with members \code{frac} 
    which is the fraction (0-1) of the 3' UTR region to count 
    reads in, \code{minLength} the minimum acceptable 3'UTR
    length irrespective of \code{frac} and \code{downstream} 
    the number of base pairs to flank the end of the 3' UTR of 
    transcripts when analyzing Quant-Seq data.}
    
    \item{interFeature}{see the \code{inter.feature} argument
    in \code{summarizeOverlaps}.}

    \item{rc}{the fraction of the available cores to use
    in a multicore system.}
}
\value{
    A data frame with counts for each sample, ready to be
    passed to the main \code{\link{metaseqr2}} pipeline.
}
\description{
    This function is a helper for the \code{metaseqr2}
    pipeline, for reading SAM/BAM or BED files when a read
    counts file is not available. It can also be used
    very easily in an autonomous manner.
}
\examples{
dataPath <- system.file("extdata",package="metaseqR2")
targets <- data.frame(samplename=c("C","T"),
    filename=file.path(dataPath,c("C.bam","T.bam")),  
    condition=c("Control","Treatment"),
    paired=c("single","single"),stranded=c("forward","forward"))
path <- tempdir()
write.table(targets,file=file.path(path,"targets.txt"),
    sep="\t",row.names=FALSE,quote=FALSE)
#geneData <- loadAnnotation("mm10","ensembl","gene")
myTargets <- readTargets(file.path(path,"targets.txt"))
if (.Platform$OS.type == "unix") {
#    r2c <- read2count(targets=myTargets,
#        fileType=myTargets$type,annotation=geneData)
#    geneCounts <- r2c$counts
#    libsizeList <- r2c$libsize
}
}
\author{
    Panagiotis Moulos
}

\name{normalizeEdaseq}
\alias{normalizeEdaseq}
\title{Normalization based on the EDASeq package}
\usage{
    normalizeEdaseq(geneCounts, sampleList,
        normArgs = NULL, geneData = NULL,
        output = c("matrix", "native"))
}
\arguments{
    \item{geneCounts}{a table where each row represents a
    gene and each column a sample. Each cell contains the
    read counts for each gene and sample. Such a table can be
    produced outside metaseqr2 and is imported during the
    basic metaseqr2 workflow.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{normArgs}{a list of EDASeq normalization
    parameters. See the result of
    \code{getDefaults("normalization",} \code{"edaseq")} for
    an example and how you can modify it.}

    \item{geneData}{an optional annotation data frame (such
    the ones produced by \code{getAnnotation}) which
    contains the GC content for each gene and from which the
    gene lengths can be inferred by chromosome coordinates.}

    \item{output}{the class of the output object. It can be
    \code{"matrix"} (default) for versatility with other
    tools or \code{"native"} for the EDASeq native S4 object
    (SeqExpressionSet). In the latter case it should be
    handled with suitable EDASeq methods.}
}
\value{
    A matrix or a SeqExpressionSet with normalized counts.
}
\description{
    This function is a wrapper over EDASeq normalization. It
    accepts a matrix of gene counts (e.g. produced by
    importing an externally generated table of counts to the
    main metaseqr2 pipeline).
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
diagplotBoxplot(dataMatrix,sampleList)

lengths <- round(1000*runif(nrow(dataMatrix)))
starts <- round(1000*runif(nrow(dataMatrix)))
ends <- starts + lengths
gc=runif(nrow(dataMatrix))
geneData <- data.frame(
    chromosome=c(rep("chr1",nrow(dataMatrix)/2),
        rep("chr2",nrow(dataMatrix)/2)),
    start=starts,end=ends,gene_id=rownames(dataMatrix),gc_content=gc,
    row.names=rownames(dataMatrix)
)
normDataMatrix <- normalizeEdaseq(dataMatrix,sampleList,
    geneData=geneData)
diagplotBoxplot(normDataMatrix,sampleList)
}
\author{
    Panagiotis Moulos
}

\name{downsampleCounts}
\alias{downsampleCounts}
\title{Downsample read counts}
\usage{
    downsampleCounts(counts)
}
\arguments{
    \item{counts}{the read counts table 
    which is subjected to downsampling.}
}
\value{
    The downsampled counts matrix.
}
\description{
    This function downsamples the library sizes 
    of a read counts table to the lowest library 
    size, according to the methdology used in 
    (Soneson and Delorenzi, BMC Bioinformatics, 
    2013).
}
\details{
    The downsampling process involves random sampling. 
    For guaranteed reproducibility, be sure to use 
    \code{set.seed} before downsampling. By default, 
    when the metaseqR2 package is loaded, the seed is 
    set to \code{42}.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(5000)
D <- downsampleCounts(dataMatrix)
}
\author{
    Panagiotis Moulos
}

\name{estimateAufcWeights}
\alias{estimateAufcWeights}
\title{Estimate AUFC weights}
\usage{
    estimateAufcWeights(counts, normalization,
        statistics, nsim = 10, N = 10000, 
        samples = c(3, 3), ndeg = c(500, 500),
        top = 500, modelOrg = "mm9", fcBasis = 1.5,
        drawFpc = FALSE, rc = NULL,
        ...)
}
\arguments{
    \item{counts}{the real raw counts table from
    which the simulation parameters will be 
    estimated. It must not be normalized and must
    contain only integer counts, without any other
    annotation elements and unique gene identifiers
    as the rownames attribute.}

    \item{normalization}{same as \code{normalization} 
    in \code{\link{metaseqr2}}.}

    \item{statistics}{same as \code{statistics} in 
    \code{\link{metaseqr2}}.}

    \item{nsim}{the number of simulations to perform
    to estimate the weights. It default to 10.}

    \item{N}{the number of genes to produce. 
    See \code{\link{makeSimDataSd}}.}

    \item{samples}{a vector with 2 integers, which
    are the number of samples for each condition 
    (two conditions currently supported).}

    \item{ndeg}{a vector with 2 integers, which
    are the number of differentially expressed 
    genes to be produced. The first element is 
    the number of up-regulated genes while the 
    second is the number of down-regulated genes.}

    \item{fcBasis}{the minimum fold-change for 
    deregulation.}

    \item{top}{the top \code{top} best ranked 
    (according to p-value) to use, to calculate 
    area under the false discovery curve.}

    \item{modelOrg}{the organism from which the 
    data are derived. It must be one of 
    \code{\link{metaseqr2}} supported organisms.}

    \item{drawFpc}{draw the averaged false 
    discovery curves? Default to \code{FALSE}.}

    \item{rc}{the fraction of the available cores to 
    use in a multicore system.}

    \item{...}{Further arguments to be passed to 
    \code{\link{estimateSimParams}}.}
}
\value{
    A vector of weights to be used in 
    \code{\link{metaseqr2}} with the 
    \code{weights} option.
}
\description{
    This function automatically estimates weights 
    for the \code{"weight"} and \code{"dperm_weight"} 
    options of metaseqR2 for combining p-values from 
    multiple statistical tests. It creates simulated 
    dataset based on real data and then performs 
    statistical analysis with metaseqR2 several times 
    in order to derive False Discovery Curves. Then, 
    the average areas under the false discovery curves
    are used to construct weights for each algorithm,
    according to its performance when using simulated 
    data.
}
\details{
    The weight estimation process involves a lot of 
    random sampling. For guaranteed reproducibility,
    be sure to use \code{set.seed} prior to any
    calculations. By default, when the metaseqR2 package
    is loaded, the seed is set to \code{42}.
}
\examples{
require(zoo)
data("mm9GeneData",package="metaseqR2")
weights <- estimateAufcWeights(
    counts=as.matrix(mm9GeneCounts[sample(nrow(mm9GeneCounts),1000),9:12]),
    normalization="edaseq",
    statistics=c("edger","limma"),
    nsim=1,N=100,ndeg=c(10,10),top=10,modelOrg=NULL,
    rc=0.01,libsizeGt=1e+5
)
}
\author{
    Panagiotis Moulos
}

\docType{data}
\name{hg19pvalues}
\alias{hg19pvalues}
\title{p-values from human RNA-Seq data with two conditions, four samples}
\format{a \code{matrix} with p-values from metaseqR2 supported tests.}
\source{
    Giakountis et al. (https://doi.org/10.1016/j.celrep.2016.05.038)
}
\description{
    This data set contains p-values calculated with each
    of the supported statistical testing algorithms in
    metaseqR2 for 1000 genes. The purpose of this matrix is
    to demonstrate the p-value combination methods as well as
    be used for a playground for other such methods and with
    other metaseqR2 facilities.
}
\author{
    Panagiotis Moulos
}
\keyword{datasets}

\name{diagplotAvgFtd}
\alias{diagplotAvgFtd}
\title{Create average False (or True) Discovery 
    curves}
\usage{
    diagplotAvgFtd(ftdrObj, output = "x11",
        path = NULL, draw = TRUE, ...)
}
\arguments{
    \item{ftdrObj}{a list with outputs from 
    \code{\link{diagplotFtd}}.}

    \item{output}{one or more R plotting 
    device to direct the plot result to.
    Supported mechanisms: \code{"x11"} (default), 
    \code{"png"}, \code{"jpg"}, \code{"bmp"}, 
    \code{"pdf"} or \code{"ps"}.}

    \item{path}{the path to create output files.}

    \item{draw}{boolean to determine whether 
    to plot the curves or just return the 
    calculated values (in cases where the user 
    wants the output for later averaging
    for example). Defaults to \code{TRUE} (make 
    plots).}

    \item{...}{further arguments to be passed to 
    plot devices, such as parameter from 
    \code{\link{par}}.}
}
\value{
    A named list with two members: the first member 
    (\code{avgFtdr}) contains a list with the 
    means and the standard deviations of the averaged
    \code{ftdrObj} and are used to create the plot. 
    The second member (\code{path}) contains the 
    path to the created figure graphic.
}
\description{
    This function creates false (or true) discovery 
    curves using a list containing several outputs 
    from \code{\link{diagplotFtd}}.
}
\examples{
p11 <- 0.001*matrix(runif(300),100,3)
p12 <- matrix(runif(300),100,3)
p21 <- 0.001*matrix(runif(300),100,3)
p22 <- matrix(runif(300),100,3)
p31 <- 0.001*matrix(runif(300),100,3)
p32 <- matrix(runif(300),100,3)
p1 <- rbind(p11,p21)
p2 <- rbind(p12,p22)
p3 <- rbind(p31,p32)
rownames(p1) <- rownames(p2) <- rownames(p3) <-
    paste("gene",1:200,sep="_")
colnames(p1) <- colnames(p2) <- colnames(p3) <-
    paste("method",1:3,sep="_")
truth <- c(rep(1,40),rep(-1,40),rep(0,20),
    rep(1,10),rep(2,10),rep(0,80))
names(truth) <- rownames(p1)
ftdObj1 <- diagplotFtd(truth,p1,N=100,draw=FALSE)
ftdObj2 <- diagplotFtd(truth,p2,N=100,draw=FALSE)
ftdObj3 <- diagplotFtd(truth,p3,N=100,draw=FALSE)
ftdObj <- list(ftdObj1,ftdObj2,ftdObj3)
avgFtdObj <- diagplotAvgFtd(ftdObj)
}
\author{
    Panagiotis Moulos
}

\name{diagplotVenn}
\alias{diagplotVenn}
\title{Venn diagrams when performing meta-analysis}
\usage{
    diagplotVenn(pmat, fcmat = NULL, pcut = 0.05,
        fcut = 0.5, direction = c("dereg", "up", "down"),
        nam = as.character(round(1000 * runif(1))),
        output = "x11", path = NULL, altNames = NULL, ...)
}
\arguments{
    \item{pmat}{a matrix with p-values corresponding to the
    application of each statistical algorithm. See also
    Details.}

    \item{fcmat}{an optional matrix with fold changes
    corresponding to the application of each statistical
    algorithm. See also Details.}

    \item{pcut}{if \code{fcmat} is supplied, an absolute
    fold change cutoff to be applied to \code{fcmat} to
    determine the differentially expressed genes for each
    algorithm.}

    \item{fcut}{a p-value cutoff for statistical
    significance. Defaults to \code{0.05}.}

    \item{direction}{if \code{fcmat} is supplied, a keyword
    to denote which genes to draw in the Venn diagrams with
    respect to their direction of regulation. See Details.}

    \item{nam}{a name to be appended to the output graphics
    file (if \code{"output"} is not \code{"x11"}).}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"} or \code{"ps"}.}

    \item{path}{the path to create output files. If 
    \code{"path"} is not \code{NULL}, a file with the 
    intersections in the Venn diagrams will be produced 
    and written in \code{"path"}.}

    \item{altNames}{an optional named vector of names, e.g.
    HUGO gene symbols, alternative or complementary to the
    unique gene names which are the rownames of \code{pmat}.
    The names of the vector must be the rownames of
    \code{pmat}.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filenames of the plots produced in a named list with
    names the \code{which.plot} argument. If
    output=\code{"x11"}, no output filenames are produced.
}
\description{
    This function uses the R package VennDiagram and plots an
    up to 5-way Venn diagram depicting the common and
    specific to each statistical algorithm genes, for each
    contrast. Mostly for internal use because of its main
    argument which is difficult to construct, but can be used
    independently if the user grasps the logic.
}
\details{
    Regarding \code{pmat}, the p-value matrix must have the 
    colnames attribute and the colnames should correspond to 
    the name of the algorithm used to fill the specific 
    column (e.g. if \code{"statistics"=c("deseq","edger",}
    \code{"nbpseq")} then \code{colnames(pmat) <-}
    \code{c("deseq","edger","nbpseq")}.
    
    Regarding \code{fcmat}, the fold change matrix must have 
    the colnames attribute and the colnames should correspond 
    to the name of the algorithm used to fill the specific 
    column (see the parameter \code{pmat}).
    
    Regarding \code{direction}, it can be one of \code{"dereg"}
    for the total of regulated genes, where 
    \code{abs(fcmat[,n])>=fcut} (default), \code{"up"} for
    the up-regulated genes where \code{fcmat[,n]>=fcut} or
    \code{"down"} for the up-regulated genes where
    \code{fcmat[,n]<=-fcut}.
}
\examples{
require(VennDiagram)
p1 <- 0.01*matrix(runif(300),100,3)
p2 <- matrix(runif(300),100,3)
p <- rbind(p1,p2)
rownames(p) <- paste("gene",1:200,sep="_")
colnames(p) <- paste("method",1:3,sep="_")
vennContents <- diagplotVenn(p)
}
\author{
    Panagiotis Moulos
}

\name{makeSimDataSd}
\alias{makeSimDataSd}
\title{Create simulated counts using the 
    Soneson-Delorenzi method}
\usage{
    makeSimDataSd(N, param, samples = c(5, 5),
        ndeg = rep(round(0.1*N), 2), fcBasis = 1.5,
        libsizeRange = c(0.7, 1.4), libsizeMag = 1e+7,
        modelOrg = NULL, simLengthBias = FALSE)
}
\arguments{
    \item{N}{the number of genes to produce.}

    \item{param}{a named list with negative binomial 
    parameter sets to sample from. The first member is
    the mean parameter to sample from (\code{muHat}) 
    and the second the dispersion (\code{phiHat}). 
    This list can be created with the 
    \code{\link{estimateSimParams}} function.}

    \item{samples}{a vector with 2 integers, 
    which are the number of samples for each 
    condition (two conditions currently supported).}

    \item{ndeg}{a vector with 2 integers, which are 
    the number of differentially expressed genes to 
    be produced. The first element is the number of 
    up-regulated genes while the second is the 
    number of down-regulated genes.}

    \item{fcBasis}{the minimum fold-change for 
    deregulation.}

    \item{libsizeRange}{a vector with 2 numbers 
    (generally small, see the default), as they 
    are multiplied with \code{libsizeMag}. These 
    numbers control the library sized of the 
    synthetic data to be produced.}

    \item{libsizeMag}{a (big) number to multiply 
    the \code{libsizeRange} to produce library 
    sizes.}

    \item{modelOrg}{the organism from which the 
    real data are derived from. It must be one 
    of the supported organisms (see the main 
    \code{\link{metaseqr2}} help page). It is used 
    to sample real values for GC content.}

    \item{simLengthBias}{a boolean to instruct 
    the simulator to create genes whose read counts is
    proportional to their length. This is achieved by 
    sorting in increasing order the mean parameter of 
    the negative binomial distribution (and the 
    dispersion according to the mean) which will cause 
    an increasing gene count length with the sampling. 
    The sampled lengths are also sorted so that in the 
    final gene list, shorter genes have less counts as 
    compared to the longer ones. The default is FALSE.}
}
\value{
    A named list with two members. The first 
    member (\code{simdata}) contains the 
    synthetic dataset 
}
\description{
    This function creates simulated RNA-Seq gene 
    expression datasets using the method presented 
    in (Soneson and Delorenzi, BMC Bioinformatics, 
    2013). For the time being, it creates only 
    simulated datasets with two conditions.
}
\details{
    The simulated data generation involves a lot of 
    random sampling. For guaranteed reproducibility,
    be sure to use \code{set.seed} prior to any
    calculations. By default, when the metaseqR2 package
    is loaded, the seed is set to \code{42}.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
## File "bottomly_read_counts.txt" from the ReCount database
#download.file(paste("http://bowtie-bio.sourceforge.net/recount/",
#    "countTables/bottomly_count_table.txt",sep=""),
#    destfile="~/bottomly_count_table.txt")
N <- 2000
#parList <- estimateSimParams("~/bottomly_read_counts.txt")
parList <- estimateSimParams(dataMatrix,libsizeGt=3e+4)
sim <- makeSimDataSd(N,parList)
synthData <- sim$simdata
trueDeg <- which(sim$truedeg!=0)
}
\author{
    Panagiotis Moulos
}

\name{getInstalledAnnotations}
\alias{getInstalledAnnotations}
\title{Load a metaseqR2 annotation element}
\usage{
    getInstalledAnnotations(obj = NULL)
}
\arguments{
    \item{obj}{\code{NULL} or the path to a metaseqR2
    SQLite annotation database. If \code{NULL}, the
    function will try to guess the location of the 
    SQLite database.}
}
\value{
    The function returns a \code{data.frame} object with
    the installed local annotations.
}
\description{
    This function returns a data frame with information
    on locally installed, supported or custom, annotations.
    
}
\examples{
db <- file.path(system.file(package="metaseqR2"),
    "annotation.sqlite")
if (file.exists(db))
    ig <- getInstalledAnnotations(obj=db)
}
\author{
    Panagiotis Moulos
}

\name{statNoiseq}
\alias{statNoiseq}
\title{Statistical testing with NOISeq}
\usage{
    statNoiseq(object, sampleList, contrastList = NULL,
        statArgs = NULL, geneData = NULL, logOffset = 1)
}
\arguments{
    \item{object}{a matrix or an object specific to each
    normalization algorithm supported by metaseqR2, containing
    normalized counts. See also Details.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{contrastList}{vector of contrasts as defined in the 
    main help page of \code{\link{metaseqr2}}. See also 
    Details.}

    \item{statArgs}{a list of edgeR statistical algorithm
    parameters. See the result of
    \code{getDefaults("statistics",} \code{"noiseq")} for an
    example and how you can modify it.}

    \item{geneData}{an optional annotation data frame (such
    the ones produced by \code{get.annotation} which contains
    the GC content for each gene and from which the gene
    lengths can be inferred by chromosome coordinates.}

    \item{logOffset}{a number to be added to each element of
    data matrix in order to avoid Infinity on log type data
    transformations.}
}
\value{
    A named list of NOISeq q-values, whose names are the
    names of the contrasts.
}
\description{
    This function is a wrapper over NOISeq statistical
    testing. It accepts a matrix of normalized gene counts or
    an S4 object specific to each normalization algorithm
    supported by metaseqR2.
}
\details{
    Regarding \code{object}, apart from \code{matrix} (also 
    for NOISeq), the object can be a \code{SeqExpressionSet} 
    (EDASeq), \code{CountDataSet} (DESeq), \code{DGEList} 
    (edgeR), \code{DESeqDataSet} (DESeq2), \code{SeqCountSet} 
    (DSS) or \code{ABSDataSet} (ABSSeq).
    
    Regarding \code{contrastList} it can also be a named 
    structured list of contrasts as returned by the internal
    function \code{metaseqR2:::makeContrastList}.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(1000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
contrast <- "A_vs_B"
lengths <- round(1000*runif(nrow(dataMatrix)))
starts <- round(1000*runif(nrow(dataMatrix)))
ends <- starts + lengths
gc=runif(nrow(dataMatrix))
biotype=rep("protein_coding",nrow(dataMatrix))
geneData <- data.frame(
    chromosome=c(rep("chr1",nrow(dataMatrix)/2),
    rep("chr2",nrow(dataMatrix)/2)),
        start=starts,end=ends,gene_id=rownames(dataMatrix),
    gc_content=gc,biotype=biotype
)
normArgs <- metaseqR2:::getDefaults("normalization","noiseq")
normDataMatrix <- normalizeNoiseq(dataMatrix,sampleList,normArgs,
    geneData)
p <- statNoiseq(normDataMatrix,sampleList,contrast,
    geneData=geneData)
}
\author{
    Panagiotis Moulos
}

\name{diagplotPairs}
\alias{diagplotPairs}
\title{Massive X-Y, M-D correlation plots}
\usage{
    diagplotPairs(x, output = "x11", altNames = NULL, 
        path = NULL, ...)
}
\arguments{
    \item{x}{the read counts matrix or data frame.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"} or \code{"ps"}.}
    
    \item{altNames}{optional names, alternative or complementary 
    to the rownames of \code{x}. It is used only in JSON output.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filename of the pairwise comparisons plot produced if
    it's a file.
}
\description{
    This function uses the read counts matrix to create
    pairwise correlation plots. The upper diagonal of the
    final image contains simple scatterplots of each sample
    against each other (log2 scale) while the lower diagonal
    contains mean-difference plots for the same samples (log2
    scale). This type of diagnostic plot may not be
    interpretable for more than 10 samples.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
diagplotPairs(dataMatrix)
}
\author{
    Panagiotis Moulos
}

\name{getWeights}
\alias{getWeights}
\title{Get precalculated statistical test 
    weights}
\usage{
    getWeights(org = c("human", "chimpanzee", "mouse",
        "fruitfly", "arabidopsis", "rat"))
}
\arguments{
    \item{org}{\code{"human"}, \code{"chimpanzee"},
    \code{"mouse"}, \code{"fruitfly"}, \code{"arabidopsis"}
    or \code{"rat"}.}
}
\value{
    A named vector of convex weights.
}
\description{
    This function returns pre-calculated weights 
    for human, chimpanzee, mouse, fruitfly and 
    arabidopsis based on the performance of 
    simulated datasets estimated from real data 
    from the ReCount database 
    (\url{http://bowtie-bio.sourceforge.net/recount/}).
    Currently pre-calculated weights are available only 
    when all six statistical tests are used and for 
    normalization with EDASeq. For other combinations, 
    use the \code{\link{estimateAufcWeights}} function.
}
\examples{
wh <- getWeights("human")
}
\author{
    Panagiotis Moulos
}

\name{diagplotMds}
\alias{diagplotMds}
\title{Multi-Dimensinal Scale plots or RNA-Seq samples}
\usage{
    diagplotMds(x, sampleList, method = "spearman",
        logIt = TRUE, output = "x11", path = NULL, ...)
}
\arguments{
    \item{x}{the count data matrix.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{method}{which correlation method to use. Same as
    the method parameter in \code{\link{cor}} function.}

    \item{logIt}{whether to log transform the values of x or
    not.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"}, \code{"ps"} or \code{"json"}. The latter is
    currently available for the creation of interactive
    volcano plots only when reporting the output, through the
    highcharts javascript library.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filename of the MDS plot produced if it's a file.
}
\description{
    Creates a Multi-Dimensional Scale plot for the given
    samples based on the count data matrix. MDS plots are
    very useful for quality control as you can easily see of
    samples of the same groups are clustered together based
    on the whole dataset.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(5000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
diagplotMds(dataMatrix,sampleList)
}
\author{
    Panagiotis Moulos
}

\name{combineMinp}
\alias{combineMinp}
\title{Combine p-values using the minimum p-value}
\usage{
    combineMinp(p)
}
\arguments{
    \item{p}{a p-value matrix (rows are genes, 
    columns are statistical tests).}
}
\value{
    A vector of combined p-values. 
}
\description{
    This function combines p-values from the 
    various statistical tests supported by
    metaseqR by taking the minimum p-value.
}
\examples{
p <- matrix(runif(300),100,3)
pc <- combineMinp(p)
}
\author{
    Panagiotis Moulos
}

\name{getAnnotation}
\alias{getAnnotation}
\title{Annotation downloader}
\usage{
    getAnnotation(org, type, refdb = "ensembl", ver = NULL,
        rc = NULL)
}
\arguments{
    \item{org}{the organism for which to download
    annotation (one of the supported ones).}

    \item{type}{\code{"gene"}, \code{"exon"} or 
    \code{"utr"}. Same as the \code{countType} in 
    \code{\link{metaseqr2}}.}

    \item{refdb}{the online source to use to fetch 
    annotation. It can be \code{"ensembl"} (default), 
    \code{"ucsc"} or \code{"refseq"}. In the later two
    cases, an SQL connection is opened with the UCSC 
    public databases.}
    
    \item{ver}{the version of the annotation to use.}

    \item{rc}{Fraction of cores to use. Same as the 
    \code{rc} in \code{\link{buildAnnotationDatabase}}.}
}
\value{
    A data frame with the canonical (not isoforms!) genes or
    exons of the requested organism. When
    \code{type="genes"}, the data frame has the following
    columns: chromosome, start, end, gene_id, gc_content,
    strand, gene_name, biotype. When \code{type="exon"} the
    data frame has the following columns: chromosome, start,
    end, exon_id, gene_id, strand, gene_name, biotype. When 
    \code{type="utr"} the data frame has the following columns: 
    chromosome, start, end, transcript_id, gene_id, strand, 
    gene_name, biotype. The gene_id and exon_id correspond to 
    Ensembl, UCSC or RefSeq gene, transcript and exon accessions 
    respectively. The gene_name corresponds to HUGO nomenclature 
    gene names.
}
\description{
    For Ensembl based annotations, this function connects to the 
    EBI's Biomart service using the package biomaRt and downloads 
    annotation elements (gene co-ordinates, exon co-ordinates, 
    gene identifications, biotypes etc.) for each of the supported
    organisms. For UCSC/RefSeq annotations, it connects to the 
    respective SQL databases if the package \code{RMySQL} is 
    present, otherwise it downloads flat files and build a 
    temporary SQLite database to make the necessaru build 
    queries. See the help page of \code{\link{metaseqr2}}
    for a list of supported organisms.
}
\note{
    The data frame that is returned contains only "canonical"
    chromosomes for each organism. It does not contain
    haplotypes or random locations and does not contain
    chromosome M.
}
\examples{
#mm10Genes <- getAnnotation("mm10","gene")
}
\author{
    Panagiotis Moulos
}

\name{statDeseq2}
\alias{statDeseq2}
\title{Statistical testing with DESeq2}
\usage{
    statDeseq2(object, sampleList, contrastList = NULL,
        statArgs = NULL)
}
\arguments{
    \item{object}{a matrix or an object specific to each
    normalization algorithm supported by metaseqR2, containing
    normalized counts. See also Details.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{contrastList}{vector of contrasts as defined in the 
    main help page of \code{\link{metaseqr2}}. See also 
    Details.}

    \item{statArgs}{a list of DESeq statistical algorithm
    parameters. See the result of
    \code{getDefaults("statistics",} \code{"deseq")} for an
    example and how you can modify it. It is not required
    when the input object is already a CountDataSet from
    DESeq normalization as the dispersions are already
    estimated.}
}
\value{
    A named list of p-values, whose names are the names of
    the contrasts.
}
\description{
    This function is a wrapper over DESeq statistical
    testing. It accepts a matrix of normalized gene counts or
    an S4 object specific to each normalization algorithm
    supported by metaseqR2.
}
\details{
    Regarding \code{object}, apart from \code{matrix} (also 
    for NOISeq), the object can be a \code{SeqExpressionSet} 
    (EDASeq), \code{CountDataSet} (DESeq), \code{DGEList} 
    (edgeR), \code{DESeqDataSet} (DESeq2), \code{SeqCountSet} 
    (DSS) or \code{ABSDataSet} (ABSSeq).
    
    Regarding \code{contrastList} it can also be a named 
    structured list of contrasts as returned by the internal
    function \code{metaseqR2:::makeContrastList}.
}
\examples{
require(DESeq2)
dataMatrix <- metaseqR2:::exampleCountData(1000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
contrast <- "A_vs_B"
normDataMatrix <- normalizeDeseq2(dataMatrix,sampleList)
p <- statDeseq2(normDataMatrix,sampleList,contrast)
}
\author{
    Dionysios Fanidis
}

\name{importCustomAnnotation}
\alias{importCustomAnnotation}
\title{Import a metaseqR2 custom annotation element}
\usage{
    importCustomAnnotation(gtfFile, metadata,
        level = c("gene", "transcript", "exon"),
        type = c("gene", "exon", "utr"))
}
\arguments{
    \item{gtfFile}{a GTF file containing the gene structure
    of the organism to be imported.}

    \item{metadata}{a list with additional information about
    the annotation to be imported. The same as in the
    \code{\link{buildCustomAnnotation}} man page.}
    
    \item{level}{same as the \code{transLevel} in 
    \code{\link{metaseqr2}}.}
    
    \item{type}{same as the \code{countType} in 
    \code{\link{metaseqr2}}.}
}
\value{
    The function returns a \code{GenomicRanges} object with
    the requested annotation.
}
\description{
    This function creates a local annotation database to be
    used with metaseqr2 so as to avoid long time on the fly 
    annotation downloads and formatting.
}
\examples{
# Dummy GTF as example
chromInfo <- data.frame(length=c(1000L,2000L,1500L),
    row.names=c("A","B","C"))

# Build with the metadata list filled (you can also provide a version)
myGenes <- importCustomAnnotation(
    gtfFile=file.path(system.file(package="metaseqR2"),"dummy.gtf"),
    metadata=list(
        organism="dummy",
        source="dummy_db",
        version=1,
        chromInfo=chromInfo
    ),
    level="gene",type="gene"
)

## Real data!
## Gene annotation dump from Ensembl
#download.file(paste0("ftp://ftp.ensembl.org/pub/release-98/gtf/",
#  "dasypus_novemcinctus/Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#  file.path(tempdir(),"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"))

## Build with the metadata list filled (you can also provide a version)
#dasGenes <- importCustomAnnotation(
#  gtfFile=file.path(tempdir(),"Dasypus_novemcinctus.Dasnov3.0.98.gtf.gz"),
#  metadata=list(
#    organism="dasNov3_test",
#    source="ensembl_test"
#  ),
#  level="gene",type="gene"
#)
}
\author{
    Panagiotis Moulos
}

\name{buildAnnotationDatabase}
\alias{buildAnnotationDatabase}
\title{Build a local annotation database for metaseqR2}
\usage{
    buildAnnotationDatabase(organisms, sources,
    db = file.path(system.file(package = "metaseqR2"),
        "annotation.sqlite"),
    forceDownload = TRUE, rc = NULL)
}
\arguments{
    \item{organisms}{a list of organisms and versions
    for which to download and build annotations. Check 
    the main \code{\link{metaseqr2}} help page for details 
    on supported organisms and the Details section below.}

    \item{sources}{a character vector of public sources
    from which to download and build annotations. Check 
    the main \code{\link{metaseqr2}} help page for details 
    on supported annotation sources.}
    
    \item{db}{a valid path (accessible at least by the
    current user) where the annotation database will be 
    set up. It defaults to 
    \code{system.file(package = "metaseqR2"),}
    \code{"annotation.sqlite")} that is, the installation
    path of metaseqR2 package. See also Details.}
    
    \item{forceDownload}{by default, 
    \code{buildAnnotationDatabase} will not download an
    existing annotation again (\code{FALSE}). Set to 
    \code{TRUE} if you wish to update the annotation 
    database for a particular version.}

    \item{rc}{fraction (0-1) of cores to use in a multicore 
    system. It defaults to \code{NULL} (no parallelization).
    Sometimes used for building certain annotation types.}
}
\value{
    The function does not return anything. Only the SQLite 
    database is created or updated.
}
\description{
    This function creates a local annotation database to be
    used with metaseqr2 so as to avoid long time on the fly 
    annotation downloads and formatting.
}
\details{
    Regarding the \code{organisms} argument, it is a list
    with specific format which instructs 
    \code{buildAnnotationDatabase} on which organisms and
    versions to download from the respective sources. Such
    a list may have the format: 
    \code{organisms=list(hg19=75, mm9=67, mm10=96:97)}
    This is explained as follows:
    \itemize{
        \item A database comprising the human genome versions
        \code{hg19} and the mouse genome versions 
        \code{mm9, mm10} will be constructed.
        \item If \code{"ensembl"} is in \code{sources}, 
        version 75 is downloaded for \code{hg19} and versions 
        \code{67, 96, 97} for \code{mm9, mm10}. 
        \item If \code{"ucsc"} or \code{"refseq"} are in 
        \code{sources}, the latest versions are downloaded
        and marked by the download date. As UCSC and RefSeq
        versions are not accessible in the same way as
        Ensembl, this procedure cannot always be replicated.
    }
    \code{organisms} can also be a character vector with organism
    names/versions (e.g. \code{organisms = c("mm10","hg19")}),
    then the latest versions are downloaded in the case of 
    Ensembl.
    
    Regarding \code{db}, this controls the location of the
    installation database. If the default is used, then there is
    no need to provide the local database path to any function
    that uses the database (e.g. the main \code{metaseqr2}).
    Otherwise, the user will either have to provide this each
    time, or the annotation will have to be downloaded and used
    on-the-fly.
}
\examples{
# Build a test database with one genome
myDb <- file.path(tempdir(),"testann.sqlite")

organisms <- list(mm10=75)
sources <- "ensembl"

# If the example is not running in a multicore system, rc is ignored
#buildAnnotationDatabase(organisms,sources,db=myDb,rc=0.5)

# A more complete case, don't run as example
# Since we are using Ensembl, we can also ask for a version
#organisms <- list(
#    mm9=67,
#    mm10=96:97,
#    hg19=75,
#    hg38=96:97
#)
#sources <- c("ensembl", "refseq")

## Build on the default location (depending on package location, it may
## require root/sudo)
#buildAnnotationDatabase(organisms,sources)

## Build on an alternative location
#myDb <- file.path(path.expand("~"),"my_ann.sqlite")
#buildAnnotationDatabase(organisms,sources,db=myDb)
}
\author{
    Panagiotis Moulos
}

\name{diagplotDeHeatmap}
\alias{diagplotDeHeatmap}
\title{Diagnostic heatmap of differentially expressed genes}
\usage{
    diagplotDeHeatmap(x, scale = c("asis", "zscore"), con = NULL, 
        output = "x11", path = NULL, ...)
}
\arguments{
    \item{x}{the data matrix to create a heatmap for.}
    
    \item{scale}{value scale in the heatmap. As provided 
    (\code{scale="asis"}, default) or Z-scores
    (\code{scale="zscore"})}

    \item{con}{an optional string depicting a name (e.g. the
    contrast name) to appear in the title of the volcano
    plot.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"}, \code{"ps"}.}

    \item{path}{the path to create output files.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filenames of the plots produced in a named list with
    names the \code{whichPlot} argument. If
    \code{output="x11"}, no output filenames are produced.
}
\description{
    This function plots a heatmap of the differentially
    expressed genes produced by the metaseqr workflow, useful
    for quality control, e.g. whether samples belonging to
    the same group cluster together.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
contrast <- "A_vs_B"
M <- normalizeEdger(dataMatrix,sampleList)
p <- statEdger(M,sampleList,contrast)
diagplotDeHeatmap(dataMatrix[p[[1]]<0.05,])
}
\author{
    Panagiotis Moulos
}

\docType{data}
\name{libsizeListMm9}
\alias{libsizeListMm9}
\title{Mouse RNA-Seq data with two conditions, four samples}
\format{a named \code{list} with library sizes.}
\source{
    ENCODE (http://genome.ucsc.edu/encode/)
}
\description{
    The library size list for \code{mm9GeneCounts}. See the
    data set description.
}
\author{
    Panagiotis Moulos
}
\keyword{datasets}

\name{diagplotFtd}
\alias{diagplotFtd}
\title{Create False (or True) Positive (or
    Negative) curves}
\usage{
    diagplotFtd(truth, p, type = "fpc", N = 2000,
        output = "x11", path = NULL, draw = TRUE, ...)
}
\arguments{
    \item{truth}{the ground truth differential 
    expression vector. It should contain only 
    zero and non-zero elements, with zero denoting
    non-differentially expressed genes and non-zero, 
    differentially expressed genes. Such a vector 
    can be obtained for example by using the 
    \code{\link{makeSimDataSd}} function, which 
    creates simulated RNA-Seq read counts based on 
    real data. The elements of \code{truth} MUST 
    be named (e.g. each gene's name).}

    \item{p}{a p-value matrix whose rows correspond 
    to each element in the \code{truth} vector. If 
    the matrix has a \code{colnames} attribute, a 
    legend will be added to the plot using these 
    names, else a set of column names will be 
    auto-generated. \code{p} can also be a list or 
    a data frame. The p-values MUST be named (e.g. 
    each gene's name).}

    \item{type}{what to plot, can be \code{"fpc"} 
    for False Positive Curves (default), 
    \code{"tpc"} for True Positive Curves, 
    \code{"fnc"} for False Negative Curves or 
    \code{"tnc"} for True Negative Curves.}

    \item{N}{create the curves based on the 
    top (or bottom) \code{N} ranked genes 
    (default is 2000) to be used with
    \code{type="fpc"} or \code{type="tpc"}.}

    \item{output}{one or more R plotting device to 
    direct the plot result to. Supported mechanisms: 
    \code{"x11"} (default), \code{"png"}, \code{"jpg"}, 
    \code{"bmp"}, \code{"pdf"} or \code{"ps"}.}

    \item{path}{the path to create output files.}

    \item{draw}{boolean to determine whether to
    plot the curves or just return the calculated
    values (in cases where the user wants the
    output for later averaging for example). 
    Defaults to \code{TRUE} (make plots).}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    A named list with two members: the first member 
    (\code{ftdr}) contains the values used to create 
    the plot. The second member (\code{path}) contains 
    the path to the created figure graphic.
}
\description{
    This function creates false (or true) discovery 
    curves using a matrix of p-values (such a matrix 
    can be derived for example from the result table 
    of \code{\link{metaseqr2}} by subsetting the table 
    to get the p-values from several algorithms) 
    given a ground truth vector for differential 
    expression.
}
\examples{
p1 <- 0.001*matrix(runif(300),100,3)
p2 <- matrix(runif(300),100,3)
p <- rbind(p1,p2)
rownames(p) <- paste("gene",1:200,sep="_")
colnames(p) <- paste("method",1:3,sep="_")
truth <- c(rep(1,40),rep(-1,40),rep(0,20),
    rep(1,10),rep(2,10),rep(0,80))
names(truth) <- rownames(p)
ftdObj <- diagplotFtd(truth,p,N=100)
}
\author{
    Panagiotis Moulos
}

\name{estimateSimParams}
\alias{estimateSimParams}
\title{Estimate negative binomial parameters from 
    real data}
\usage{
    estimateSimParams(realCounts, libsizeGt = 3e+6,
        rowmeansGt = 5,eps = 1e-11, rc = NULL, draw = FALSE)
}
\arguments{
    \item{realCounts}{a text tab-delimited file 
    with real RNA-Seq data. See Details.}

    \item{libsizeGt}{a library size below which 
    samples are excluded from parameter estimation 
    (default: \code{3000000}).}

    \item{rowmeansGt}{a row means (mean counts 
    over samples for each gene) below which 
    genes are excluded from parameter estimation
    (default: 5).}

    \item{eps}{the tolerance for the convergence 
    of \code{\link{optimize}} function. Defaults 
    to 1e-11.}

    \item{rc}{in case of parallel optimization, the 
    fraction of the available cores to use.}

    \item{draw}{boolean to determine whether to 
    plot the estimated simulation parameters 
    (mean and dispersion) or not. Defaults to 
    \code{FALSE} (do not draw a mean-dispersion 
    scatterplot).}
}
\value{
    A named list with two members: \code{muHat}
    which contains negative binomial mean 
    estimates and \code{phiHat} which contains 
    dispersion estimates.
}
\description{
    This function reads a read counts table 
    containing real RNA-Seq data (preferebly 
    with more than 20 samples so as to get as 
    much accurate as possible estimations) and 
    calculates a population of count means and 
    dispersion parameters which can be used to 
    simulate an RNA-Seq dataset with synthetic 
    genes by drawing from a negative binomial 
    distribution. This function works in the 
    same way as described in (Soneson and 
    Delorenzi, BMC Bioinformatics, 2013) and 
    (Robles et al., BMC Genomics, 2012).
}
\details{
    Regarding \code{realCounts}, the file should strictly
    contain a unique gene name (e.g. Ensembl accession) in 
    the first column and all other columns should contain 
    read counts for each gene. Each column must be named
    with a unique sample identifier. See examples in the 
    ReCount database 
    \url{http://bowtie-bio.sourceforge.net/recount/}.
    
    Also, the parameter estimation involves a lot of 
    random sampling. For guaranteed reproducibility,
    be sure to use \code{set.seed} prior to any
    calculations. By default, when the metaseqR2 package
    is loaded, the seed is set to \code{42}.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(2000)
parList <- estimateSimParams(dataMatrix,libsizeGt=3e+4)
}
\author{
    Panagiotis Moulos
}

\name{diagplotNoiseq}
\alias{diagplotNoiseq}
\title{Diagnostic plots based on the NOISeq package}
\usage{
    diagplotNoiseq(x, sampleList, covars,
        whichPlot = c("biodetection", "countsbio", "saturation", 
        "rnacomp", "readnoise", "biodist"),
        output = "x11",
        biodistOpts = list(p = NULL, pcut = NULL, name = NULL),
        path = NULL, isNorm = FALSE, ...)
}
\arguments{
    \item{x}{the count data matrix.}

    \item{sampleList}{the list containing condition names
    and the samples under each condition.}

    \item{covars}{a list (whose annotation elements are
    ideally a subset of an annotation data frame produced by
    \code{\link{getAnnotation}}) with the following members:
    data (the data matrix), length (gene length), gc (the
    gene gc_content), chromosome (a data frame with
    chromosome name and co-ordinates), factors (a factor with
    the experimental condition names replicated by the number
    of samples in each experimental condition) and biotype
    (each gene's biotype as depicted in Ensembl-like
    annotations).}

    \item{whichPlot}{the NOISeq package plot to generate. 
    See Details}

    \item{biodistOpts}{a list with the following members: p
    (a vector of p-values, e.g. the p-values of a contrast),
    pcut (a unique number depicting a p-value cutoff,
    required for the \code{"biodist"} case), name (a name for
    the \code{"biodist"} plot, e.g. the name of the
    contrast.}

    \item{output}{one or more R plotting device to direct the
    plot result to. Supported mechanisms: \code{"x11"}
    (default), \code{"png"}, \code{"jpg"}, \code{"bmp"},
    \code{"pdf"} or \code{"ps"}.}

    \item{path}{the path to create output files.}

    \item{isNorm}{a logical indicating whether object
    contains raw or normalized data. It is not essential and
    it serves only plot annotation purposes.}

    \item{...}{further arguments to be passed to plot
    devices, such as parameter from \code{\link{par}}.}
}
\value{
    The filenames of the plots produced in a named list with
    names the \code{whichPlot} argument. If
    \code{output="x11"}, no output filenames are produced.
}
\description{
    A wrapper around the plotting functions availale in the
    NOISeq Bioconductor package. For analytical explanation
    of each plot please see the vignette of the NOISeq 
    package. It is best to use this function through the 
    main plotting function \code{\link{metaseqrPlot}}.
}
\details{
    Regarding \code{whichPlot}, It can be one or more of 
    \code{"biodetection"}, \code{"countsbio"}, 
    \code{"saturation"}, \code{"rnacomp"}, \code{"readnoise"} 
    or \code{"biodist"}. Please refer to the documentation of 
    the NOISeq package for details on the use of these plots. 
    The \code{whichPlot="saturation"} case is modified to be
    more informative by producing two kinds of plots.
}
\note{
    Please note that in case of \code{"biodist"} plots, the
    behavior of the function is unstable, mostly due to the
    very specific inputs this plotting function accepts in
    the NOISeq package. We have tried to predict unstable
    behavior and avoid exceptions through the use of tryCatch
    but it's still possible that you might run onto an error.
}
\examples{
dataMatrix <- metaseqR2:::exampleCountData(5000)
sampleList <- list(A=c("A1","A2"),B=c("B1","B2","B3"))
lengths <- round(1000*runif(nrow(dataMatrix)))
starts <- round(1000*runif(nrow(dataMatrix)))
ends <- starts + lengths
covars <- list(
    data=dataMatrix,
    length=lengths,
    gc=runif(nrow(dataMatrix)),
    chromosome=data.frame(
        chromosome=c(rep("chr1",nrow(dataMatrix)/2),
        rep("chr2",nrow(dataMatrix)/2)),
        start=starts,
        end=ends
    ),
    factors=data.frame(class=metaseqR2:::asClassVector(sampleList)),
    biotype=c(rep("protein_coding",nrow(dataMatrix)/2),rep("ncRNA",
        nrow(dataMatrix)/2))
)
p <- runif(nrow(dataMatrix))
diagplotNoiseq(dataMatrix,sampleList,covars=covars,
    biodistOpts=list(p=p,pcut=0.1,name="A_vs_B"))
}
\author{
    Panagiotis Moulos
}

\name{combineWeight}
\alias{combineWeight}
\title{Combine p-values using weights}
\usage{
    combineWeight(p, w, zerofix = NULL)
}
\arguments{
    \item{p}{a p-value matrix (rows are genes, 
    columns are statistical tests).}

    \item{w}{a weights vector, must sum to 1.}
    
    \item{zerofix}{\code{NULL} (default) or a fixed 
    numeric value between 0 and 1.}
}
\value{
    A vector of combined p-values. 
}
\description{
    This function combines p-values from the 
    various statistical tests supported by
    metaseqR using p-value weights.
}
\details{
    The argument \code{zerofix} is used to correct for
    the case of a p-value which is equal to 0 as a result
    of internal numerical and approximation procedures.
    When \code{NULL}, random numbers greater than 0 and
    less than or equal to 0.5 are used to multiply the
    offending p-values with the lowest provided non-zero
    p-value, maintaining thus a virtual order of 
    significance, avoiding having the same p-values for 
    two tests and assuming that all zero p-values represent
    extreme statistical significance. When a numeric
    between 0 and 1, this number is used for the above
    multiplication instead.
}
\examples{
p <- matrix(runif(300),100,3)
pc <- combineWeight(p,w=c(0.2,0.5,0.3))
}
\author{
    Panagiotis Moulos
}

\docType{data}
\name{mm9GeneCounts}
\alias{mm9GeneCounts}
\title{Mouse RNA-Seq data with two conditions, four samples}
\format{a \code{data.frame} with gene read counts and some embedded 
annotation, one row per gene.}
\source{
    ENCODE (http://genome.ucsc.edu/encode/)
}
\description{
    This data set contains RNA-Seq gene read counts for 3
    chromosomes. The data were downloaded from the ENCODE
    public repository and are derived from the study of
    Mortazavi et al., 2008 (Mortazavi A, Williams BA, McCue
    K, Schaeffer L, Wold B. Mapping and quantifying mammalian
    transcriptomes by RNA-Seq. Nat Methods. 2008
    Jul;5(7):621-8). In their experiment, the authors studied
    among others genes expression at two developmental stages
    of mouse liver cells. It has two conditions-developmental
    stages (e14.5, adult_8_weeks) and four samples (e14.5_1,
    e14.5_2, a8w_1, a8w_2). It also contains a predefined
    \code{sampleList} and \code{libsizeList} named
    \code{sampleListMm9} and \code{libsizeListMm9}.
}
\author{
    Panagiotis Moulos
}
\keyword{datasets}
